#------------------------------------------------------------------------------
# Copyright (c) 2006, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------

""" 
A wrapper for other colormaps to add a constant or multiplied alpha channel.
"""

import enthought.util.numerix as Numeric

# Major package imports.
from enthought.util.numerix import zeros, Float64

# Enthought library imports.
from enthought.traits.api import Float, Int

# Local imports.
from colormap import AbstractColormap


class AlphaWrapColormap(AbstractColormap):
    """An AlphaWrapColormap wraps an existing colormap and multiplies in an
    alpha value.  This is useful for functions that wish to apply an alpha
    value across an entire plot but don't want to create a new colormap
    that otherwise duplicates the existing one.
    """

    ###########################################################################
    # Traits.
    ###########################################################################

    #### Public ####

    # The maximum value for the mapped data.
    max_value = Float(1.0)

    # The minimum value for the mapped data.
    min_value = Float(0.0)

    # The total number of color steps in the map.
    steps = Int(1)
    

    ###########################################################################
    # Public 'object' interface.
    ###########################################################################
    
    def __init__(self, colormap, alpha, mode='multiply', **traits):
        """ Constructor.

        colormap is a colormap to wrap while adding alpha.  alpha is the
        alpha value to apply, and mode is either 'multiply' or 'replace'.
        If mode is 'replace', the alpha value specified will be the alpha
        value for all entries in the wrapped colormap.  If mode is 'multiply',
        the alpha value from the wrapped colormap and the alpha value
        specified will be multiplied together for the final result.
        """

        self._colormap = colormap
        self.min_value = colormap.min_value
        self.max_value = colormap.max_value
        self.steps = colormap.steps
        self._alpha = alpha
        if not mode in ('multiply', 'replace'):
            raise RuntimeError("Invalid mode in AlphaWrapColormap")
        self._mode = mode
        super(AbstractColormap, self).__init__(**traits)

    ###########################################################################
    # Public 'AbstractColormap' interface.
    ###########################################################################

    def map_value(self, val):
        if self._mode == 'multiply':
            vali = tuple(self._colormap.map_value(val))
            return vali[:3] + (vali[3]*self._alpha,)
        else:
            return tuple(self._colormap.map_value(val))[:3] + (self._alpha,)

    def map_array(self, ary):
        aryi = self._colormap.map_array(ary)
        if aryi.shape[-1] == 3:
            newshape = aryi.shape[:-1] + (4,)
            newary = zeros(newshape, Float64)
            newary[...,:3] = aryi
            newary[3] = self._alpha
            return newary
        else:
            if self._mode == 'replace':
                aryi[...,3] = self._alpha
                return aryi
            else:
                aryi[...,3] *= self._alpha
                return aryi
                
        

