#------------------------------------------------------------------------------
# Copyright (c) 2006, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Copyright (c) 2002-2004 John D. Hunter
# All Rights Reserved.
#------------------------------------------------------------------------------

""" 
A matlab style colormap -- get data from an array of RGB or RGBA values
then map numbers to the value in the nearest array index to the number.
"""

import enthought.util.numerix as Numeric

# Major package imports.
from enthought.util.numerix import add, arange, around, array, asarray, choose, clip, concatenate
from enthought.util.numerix import divide, ones, put, ravel, resize, searchsorted, shape
from enthought.util.numerix import sometrue, sort, take, where, zeros

from types import IntType, FloatType

# Enthought library imports.
from enthought.traits.api import Array, Bool, Dict, Float, Int, Property, Str, Tuple

# Local imports.
from colormap import AbstractColormap


class ArrayColormap(AbstractColormap):
    """ An AbstractColormap which maps to the nearest index in an array.

    ArrayColormap may be constructed from either an RGB or an RGB array.
    ArrayColormap will clip data to [0, len(colormap)-1], then map values
    to the color in the nearest array index.

    """

    ###########################################################################
    # Traits.
    ###########################################################################

    #### Public ####

    # The maximum value for the mapped data.
    max_value = Float(1.0)

    # The minimum value for the mapped data.
    min_value = Float(0.0)

    # The total number of color steps in the map.
    steps = Int(1)
    
    # Whether we map alpha
    has_alpha = Bool(False)
    
    #### Private ####

    _colormap = array([])


    ###########################################################################
    # Public 'object' interface.
    ###########################################################################
    
    def __init__(self, colorarray, **traits):
        """ Constructor.

        colorarray: an array of n by 3 or 4, an array of color values, either
        RGB or RGBA
        """

        self._colormap = array(colorarray)
        self.min_value = 0.0
        self.max_value = float(len(colorarray)-1)
        if colorarray.shape[1] == 4:
            self.has_alpha = True
        else:
            self.has_alpha = False
        self.steps = len(colorarray)

        self._red_lut = self._colormap[:,0]
        self._green_lut = self._colormap[:,1]
        self._blue_lut = self._colormap[:,2]
        if self.has_alpha:
            self._alpha_lut = self._colormap[:,3]
        

        super(AbstractColormap, self).__init__(**traits)
        return

    ###########################################################################
    # Public 'AbstractColormap' interface.
    ###########################################################################

    def map_value(self, val):
        fixedval = round(clip(val, self.min_value, self.max_value))
        return self._colormap[int(fixedval)]

    def map_array(self, ary):
        return self._map(around(clip(ary, self.min_value, self.max_value)))
        

    #### matplotlib ####
    def _map(self, X, alpha=1.0):
        
        alpha = min(alpha, 1.0) # alpha must be between 0 and 1
        alpha = max(alpha, 0.0)
        if type(X) in [IntType, FloatType]:
            vtype = 'scalar'
            xa = array([X])
        else:
            vtype = 'array'
            xa = asarray(X)

        xa = xa.astype(Numeric.Int)
        rgba = zeros(xa.shape+(4,), Numeric.Float)
        rgba[...,0] = take(self._red_lut, xa)
        rgba[...,1] = take(self._green_lut, xa)
        rgba[...,2] = take(self._blue_lut, xa)
        if self.has_alpha:
            rgba[...,3] = take(self._alpha_lut, xa)
        else:
            rgba[...,3] = alpha
            
        if vtype == 'scalar':
            rgba = tuple(rgba[0,:])
            
        return rgba

#### EOF ######################################################################
