""" Helper class to layout and draw sets of axes. """

from axis_info import AxisInfo
from plot_container import BOTTOM, LEFT, RIGHT, TOP

class AxesInfo:
    """ Helper class to layout and draw sets of axes. """

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
 
    def __init__ ( self, gc, index_axis, render_values = None, 
                         process_index = True ):
        self.axes          = []       
        self.index_axis    = index_axis
        self.process_index = process_index
        if render_values is not None:
            self.add_axes( gc, render_values )
            
    #---------------------------------------------------------------------------
    #  Add the axis information for both axes of one or more PlotValue objects:
    #---------------------------------------------------------------------------
 
    def add_axes ( self, gc, render_values ):
        for renderer, values in render_values:
            self.add_axis( renderer, values, False )
            self.add_axis( renderer, values, True )
        self.process_axes( gc )

    #---------------------------------------------------------------------------
    #  Add the axis information for the primary PlotValue axis:
    #---------------------------------------------------------------------------
 
    def add_axis ( self, renderer, values, is_index ):
        value  = values[0]
        origin = value.origin_
        if is_index:
            value = value._index()
            axis  = value.axis_index
            if axis is None:
                axis = value.axis
        else:
            axis = value.axis
        labels     = value._labels
        horizontal = is_index ^ (self.index_axis != 'x')
        if horizontal:
            origin &= (LEFT | RIGHT)
        else:
            origin &= (TOP  | BOTTOM)
        for axis_info in self.axes:
            if ((axis     == axis_info.axis)     and
                (is_index == axis_info.is_index) and
                (origin   == axis_info.origin)   and
                (labels   == axis_info.labels)):
                break
        else:
            axis_info = AxisInfo( axis, is_index, origin, horizontal, labels )
            self.axes.append( axis_info )
        axis_info.add_render_values( renderer, values )

    #---------------------------------------------------------------------------
    #  Factor out all of the index axes:
    #---------------------------------------------------------------------------
 
    def factor_index_axes ( self, axes_info ):
        index_axes = self.axes
        value_axes = []
        for axis_info in axes_info.axes:
            if axis_info.is_index:
                for index_axis_info in index_axes:
                    if ((axis_info.axis   == index_axis_info.axis)   and
                        (axis_info.origin == index_axis_info.origin) and
                        (axis_info.labels == index_axis_info.labels)):
                        index_axis_info.add_axis( axis_info )
                        break
                else:
                    index_axes.append( axis_info )
            else:
                value_axes.append( axis_info )
        axes_info.axes = value_axes

    #---------------------------------------------------------------------------
    #  Process all of the accumulated axis data and come up with a final set of
    #  total values as well as values for each unique axis:
    #---------------------------------------------------------------------------
 
    def process_axes ( self, gc ):
        # Initialize 'first' axis indicators:
        vl1 = vr1 = ht1 = hb1 = True
 
        # Initialize axis 'hangover' maximums:
        max_ldx = max_rdx = max_mdx = max_tdy = max_bdy = max_mdy = 0
 
        # Initialize axis minimum size requirements:
        ldx = rdx = tdy = bdy = 0
        
        # Iterate over all axis information:
        for axis_info in self.axes:
 
            # Skip if it is an index axis and we aren't processing index axes:
            is_index = axis_info.is_index
            if is_index and (not self.process_index):
                continue
        
            # Bind each value to its axis_info objects:
            if is_index:
                for renderer, values in axis_info.render_values:
                    for value in values:
                        value._axis_info_index = axis_info
            else:
                for renderer, values in axis_info.render_values:
                    for value in values:
                        value._axis_info = axis_info
            
            # Calculate the min/max range for the axis:
            axis_info.get_bounds()
            
            # Calculate the axis minimum size info:
            axis_info._min_size( gc )
            
            if axis_info.axis.visible:
                position = axis_info.position
             
                if axis_info.horizontal:
                    # Handle a horizontal axis:
                    if position & TOP:
                        if ht1:
                            ht1 = False
                            tdy += axis_info.bdy
                        else:
                            tdy += axis_info.dy
                    if position & BOTTOM:
                        if hb1:
                            hb1 = False
                            bdy += axis_info.bdy
                        else:
                            bdy += axis_info.dy
                    max_ldx = max( max_ldx, axis_info.ldx )
                    max_rdx = max( max_rdx, axis_info.rdx )
                    max_mdx = max( max_mdx, axis_info.mdx )
                else:
                    # Handle a vertical axis:
                    if position & LEFT:
                        if vl1:
                            vl1 = False
                            ldx += axis_info.ldx
                        else:
                            ldx += axis_info.dx
                    if position & RIGHT:
                        if vr1:
                            vr1 = False
                            rdx += axis_info.ldx
                        else:
                            rdx += axis_info.dx
                    max_tdy = max( max_tdy, axis_info.tdy )
                    max_bdy = max( max_bdy, axis_info.bdy )
                    max_mdy = max( max_mdy, axis_info.mdy )
               
        # Now put them all together:
        self.ldx    = max( ldx, max_ldx )
        self.rdx    = max( rdx, max_rdx )
        self.dx     = self.ldx + self.rdx
        self.tdy    = max( tdy, max_tdy )
        self.bdy    = max( bdy, max_bdy )
        self.dy     = self.tdy + self.bdy
        self.max_dx = max_mdx
        self.max_dy = max_mdy

    #---------------------------------------------------------------------------
    #  Return the list of axes that contain the specified (x,y) point:
    #---------------------------------------------------------------------------
 
    def objects_at ( self, x, y ):
        for axis_info in self.axes:
            if axis_info._isin( x, y ):
                return [ axis_info ]
        return []

    #---------------------------------------------------------------------------
    #  Lay out all of the axes:
    #
    #  Note: (x, y, dx, dy) specifies the bounding box of the plotting area that
    #        the axes surround and (sx,sy) specifies the scaling factors to be
    #        applied to the axes).
    #---------------------------------------------------------------------------
 
    def _layout ( self, x, y, dx, dy, sx, sy ):
        # Initialize the 'first' axis indicators:
        vl1 = vr1 = ht1 = hb1 = True
 
        # Calculate the layout starting points:
        lx = x
        by = y
        ty = y + dy
        rx = x + dx
 
        # Iterate over all axis information:
        for axis_info in self.axes:
            axis_info._layout_grid( x, y, dx, dy, sx, sy )
 
            # If the axis is not visible, skip rest of layout:
            if not axis_info.axis.visible:
                continue
 
            position = axis_info.position
 
            if axis_info.horizontal:
                # Handle a horizontal axis:
                if position & TOP:
                    if ht1:
                        ht1 = False
                        ty -= axis_info.tdy
                    else:
                        ty += axis_info.dy
                    axis_info._layout( TOP, x - axis_info.ldx, ty,
                                       dx + axis_info.ldx + axis_info.rdx,
                                       axis_info.dy )
                if position & BOTTOM:
                    if hb1:
                        hb1 = False
                        by -= axis_info.bdy
                    else:
                        by -= axis_info.dy
                    axis_info._layout( BOTTOM, x - axis_info.ldx, by,
                                       dx + axis_info.ldx + axis_info.rdx,
                                       axis_info.dy )
            else:
                # Handle a vertical axis:
                if position & LEFT:
                    if vl1:
                        vl1 = False
                        lx -= axis_info.ldx
                    else:
                        lx -= axis_info.dx
                    axis_info._layout( LEFT, lx, y - axis_info.tdy,
                                       axis_info.dx,
                                       dy + axis_info.tdy + axis_info.bdy )
                if position & RIGHT:
                    if vr1:
                        vr1 = False
                        rx -= axis_info.rdx
                    else:
                        rx += axis_info.dx
                    axis_info._layout( RIGHT, rx, y - axis_info.tdy,
                                       axis_info.dx,
                                       dy + axis_info.tdy + axis_info.bdy )

    #---------------------------------------------------------------------------
    #  Draw all of the axes:
    #---------------------------------------------------------------------------
 
    def _draw ( self, gc, ox, oy ):
        for axis_info in self.axes:
            if axis_info.horizontal:
                axis_info._draw( gc, ox, 0 )
            else:
                axis_info._draw( gc, 0, oy )

#### EOF ######################################################################
