""" Helper class to layout and draw individual axes. """

import sys

from plot_label import PlotLabel
from plot_base import *
from plot_container import SCALE_LOG
from plot_ticks import auto_ticks
from enthought.util.numerix import around, log10

class AxisInfo:
    """ Helper class to layout and draw individual axes. """

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
 
    def __init__ ( self, axis, is_index, origin, horizontal, labels ):
        self.axis          = axis
        self.is_index      = is_index
        self.origin        = origin
        self.horizontal    = horizontal
        self.labels        = labels
        self.position      = axis.position_
        self.min           = BIG_POS_FLOAT
        self.max           = BIG_NEG_FLOAT
        self.log_scale     = False
        self.render_values = []
 
    #---------------------------------------------------------------------------
    #  Add the range information for a specified PlotValue:
    #---------------------------------------------------------------------------
 
    def add_render_values ( self, renderer, values ):
        self.render_values.append( ( renderer, values ) )
 
    #---------------------------------------------------------------------------
    #  Add the range information from another similar axis:
    #---------------------------------------------------------------------------
 
    def add_axis ( self, axis ):
        self.render_values.extend( axis.render_values )
   
    #---------------------------------------------------------------------------
    #  Calculate the min/max bounds for the axis:
    #---------------------------------------------------------------------------
    
    def get_bounds ( self ):
        if self.is_index:
            for renderer, values in self.render_values:
                amin, amax = renderer._get_bounds_index( values )
                self.min = min( self.min, amin )
                self.max = max( self.max, amax )
        else:
            for renderer, values in self.render_values:
                amin, amax = renderer._get_bounds_value( values )
                self.min = min( self.min, amin )
                self.max = max( self.max, amax )
 
    #---------------------------------------------------------------------------
    #  Return whether the axis contains a specified (x,y) point:
    #---------------------------------------------------------------------------
 
    def _isin ( self, x, y ):
        if hasattr( self, 'tl' ):
            ax, ay, adx, ady = self.tl
            if ((x >= ax) and (y >= ay) and 
                (x < (ax + adx)) and (y < (ay + ady))):
                return True
          
        if hasattr( self, 'br' ):
            ax, ay, adx, ady = self.br
            if ((x >= ax) and (y >= ay) and 
                (x < (ax + adx)) and (y < (ay + ady))):
                return True
        return False
          
    #---------------------------------------------------------------------------
    #  Compute the minimum size needed for the axis:
    #---------------------------------------------------------------------------
 
    def _min_size ( self, gc ):
        axis = self.axis
        
        # Calculate where the ticks should be for this axis:
        if self.labels is not None:
            n     = len( self.labels )
            if not axis.tick_interval == 'auto':
                intermediate = range( 0, n, int(axis.tick_interval) )
            else:
                intermediate = range( 0, n )
            ticks = array( [ min( self.min, -0.5 ) ] + intermediate +
                           [ max( self.max, n - 0.5 ) ] )
        else:
            ticks = auto_ticks( self.min, self.max,
                                min( axis.bound_low, axis.bound_high ),
                                max( axis.bound_low, axis.bound_high ),
                                axis.tick_interval )

        # Do log scaling of ticks (if necessary and allowed):
        if (axis.scale_ == SCALE_LOG) and (ticks[0] >= 0.0):
            self.log_scale = True
            if ticks[0] == 0.0:
                if (self.min > 0.0) and (len( ticks ) >= 2):
                    if self.min < ticks[1]:
                        ticks[0] = self.min
                    elif len( ticks ) > 2:
                        ticks = ticks[ 1: ]
                    else:
                        self.log_scale = False
                else:
                    self.log_scale = False

        self._ticks = ticks
        markers = array( self.axis.markers.keys() )
        self._markers = markers
        
        # If the axis is not visible, set the size info to 0 and exit:
        if not axis.visible:
            self.ldx = self.rdx = self.mdx = self.dx = 0
            self.tdy = self.bdy = self.mdy = self.dy = 0
            return
 
        # Calculate the amount of space used by the axis title:
        title = axis._title_label
        if title.text == '':
            title_dx = title_dy = 0
        else:
            # Make sure the text is always horizontal for the calculation:
            title.rotate = 0
            title_dx, title_dy = title._min_size( gc )

        # Calculate the amount of space used by the axis markers:
        marker_dx, marker_dy = self._get_markers_min_size( gc, axis.markers.values() )
        
        # Calculate the minimum space needed by the axis labels:
        style = axis.label_style_
        if style == 0:
            # No labels, so don't set up any space for them:
            label_ldx = label_mdx = label_rdx = label_dy = 0
        else:
            # Compute the label sizes:
            label_ldx, label_mdx, label_rdx, label_dy = self.make_labels( gc )
            label_dy += axis.label_offset
            if ((self.horizontal and (self.origin & RIGHT)) or
               (((not self.horizontal) and (self.origin & TOP)))):
               label_ldx, label_rdx = label_rdx, label_ldx
 
        # Now, sum/integrate all the values we've computed (assume we are
        # doing it for a horizontal axis):
        margin_height = axis.margin_height
        margin_width  = axis.margin_width
        weight        = axis.line_weight
 
        # Calculate the vertical values:
        if axis.tick_visible:
            tick_in  = axis.tick_in
            tick_out = axis.tick_out
        else:
            tick_in = tick_out = 0
        tdy = margin_height + max( tick_in, weight / 2 )
        bdy = (margin_height + max( tick_out, (weight + 1) / 2 ) +
               label_dy + title_dy)
        dy  = tdy + bdy
 
        # Should probably include the 'tick_weight' in these calculations:
        mdx = max( label_mdx, title_dx )
        ldx = margin_width + label_ldx
        rdx = margin_width + label_rdx
        dx  = ldx + mdx + rdx
 
        # Now, save the values correctly, based on whether this is a
        # horizontal or vertical axis.
        # Note: This assumes the axis is on left/bottom of plot. If it is
        # on the other side (or both) then some of the values need to be
        # interchanged or summed:
        if self.horizontal:
            self.ldx = ldx + marker_dx / 2
            self.rdx = rdx + marker_dx / 2
            self.tdy = tdy 
            self.bdy = bdy + marker_dy
            self.mdy = 0
            if axis.use_min_width:
                self.dx = self.mdx = axis.min_width
            else:
                self.dx  = dx
                self.mdx = mdx
            if axis.use_min_height:
                self.dy = axis.min_height
            else:
                self.dy = dy + marker_dy
        else:
            self.ldx = bdy + marker_dx
            self.rdx = tdy
            self.mdx = 0
            self.tdy = ldx + marker_dy / 2
            self.bdy = rdx + marker_dy / 2
            if axis.use_min_width:
                self.dx = axis.min_width
            else:
                self.dx = dy + marker_dx
            if axis.use_min_height:
                self.dy = self.mdy = axis.min_height
            else:
                self.dy  = dx
                self.mdy = mdx

    #---------------------------------------------------------------------------
    #  Layout the axis:
    #---------------------------------------------------------------------------
 
    def _layout ( self, position, x, y, dx, dy ):
        axis = self.axis
        if not axis.visible:
            return

        mx   = axis.margin_width
        my   = axis.margin_height
 
        # Get title related information:
        title     = axis._title_label
        has_title = (title.text != '')
        if has_title:
            tdx = title_dx = title._frame_min_dx
            tdy = title_dy = title._frame_min_dy
            title.rotate = 90 * (not self.horizontal)
           
        if position & (TOP | LEFT):
            # Lay out the title (if present):
            if has_title:
                self.tl_title_cache = cache = PlotCache()
                
                # Horizontal label.
                if self.horizontal:
                    # The y-origin for the label.
                    layout_y = y + dy - my - tdy
                    
                    # If we're going to clip, left justify the label.
                    if tdx > dx:
                        layout_x = x
                        
                    # Otherwise, center it.
                    else:
                        layout_x = x + (dx - tdx) / 2
                        
                    # Layout the title label accordingly.
                    title._layout( layout_x, layout_y, tdx, tdy, cache )
                    
                # Vertical label.
                else:
                    # The x-origin for the label.
                    layout_x = x + mx
                    
                    # If we're going to clip, justify the label.
                    if tdx > dy:
                        layout_y = y
                    
                    # Otherwise, center it.
                    else:
                        layout_y = y + (dy - tdx) / 2
                    
                    # Layout the title label accordingly.
                    title._layout( layout_x, layout_y, tdy, tdx, cache )
              
            # Cache the information needed to draw the axis line and tick marks:
            self.tl = ( x, y, dx, dy )
        else:
            # Lay out the title (if present):
            if has_title:
                self.br_title_cache = cache = PlotCache()
                if self.horizontal:
                    # The y-origin for the label.
                    layout_y = y + my

                    # If we're going to clip, justify the label.
                    if tdx > dx:
                        layout_x = x
                    
                    # Otherwise, center it.
                    else:
                        layout_x = x + (dx - tdx) / 2
                        
                    # Layout the title label accordingly.
                    title._layout( layout_x, layout_y, tdx, tdy, cache )

                else:
                    # The x-origin for the label.
                    layout_x = x + dx - mx - tdy
                    
                    # If we're going to clip, justify the label.
                    if tdx > dy:
                        layout_y = y

                    # Otherwise, center it.
                    else:
                        layout_y = y + (dy - tdx) / 2
                    
                    # Layout the title label accordingly.
                    title._layout( layout_x, layout_y, tdy, tdx, cache )
                  
            # Cache the information needed to draw the axis line and tick marks:
            self.br = ( x, y, dx, dy )
 
        # Lay out the labels:
        has_labels = hasattr( self, '_labels' )
        if has_labels:
            origin  = self.origin
            caches  = []
            labels  = self._labels
            ticks   = self.ticks
            on_edge = (axis.label_style_ == -2)
            if self.horizontal:
                offset = self.tdy + axis.tick_out + axis.label_offset
                if position & TOP:
                    yt = y + offset
                    self.tl_label_caches = caches
                else:
                    yt = y + dy - offset - labels[0]._frame_min_dy
                    self.br_label_caches = caches
                xmin = -sys.maxint
                xmax =  sys.maxint
                for label in labels[:]:
                    tdx        = label._frame_min_dx
                    tdy        = label._frame_min_dy
                    tick_index = label._tick_index
                    xl         = ticks[ tick_index ]
                    if on_edge:
                        if (tick_index != 0) ^ ((origin & LEFT) == 0):
                            xl -= (tdx + 3)
                        else:
                            xl += 3
                    else:
                        xl -= tdx / 2
                    if (xl < xmin) or ((xl + tdx) > xmax):
                        labels.remove( label )
                    else:
                        cache = PlotCache()
                        caches.append( cache )
                        label._layout( xl, yt, tdx, label._frame_min_dy, cache )
                        if origin & LEFT:
                            xmin = xl + tdx + 4
                        else:
                            xmax = xl - 4
            else:
                offset = self.rdx + axis.tick_out + axis.label_offset
                mul    = ((position & LEFT) != 0)
                if mul:
                    xr = x + dx - offset
                    self.tl_label_caches = caches
                else:
                    xr = x + offset
                    self.br_label_caches = caches
                ymin = -sys.maxint
                ymax =  sys.maxint
                for label in labels[:]:
                    tdx        = label._frame_min_dx
                    tdy        = label._frame_min_dy
                    xl         = xr - mul * tdx
                    tick_index = label._tick_index
                    yt         = ticks[ tick_index ]
                    if on_edge:
                        if (tick_index != 0) ^ ((origin & TOP) == 0):
                            yt -= tdy
                    else:
                        yt -= tdy / 2
                    if (yt < ymin) or ((yt + tdy) > ymax):
                        labels.remove( label )
                    else:
                        cache = PlotCache()
                        caches.append( cache )
                        label._layout( xl, yt, tdx, tdy, cache )
                        if origin & TOP:
                            ymax = yt - 4
                        else:
                            ymin = yt + tdy + 4

        # If we have markers, lay them out.
        if axis.markers is not None and len(axis.markers) > 0:
            origin  = self.origin
            caches  = []
            if self.horizontal:
                offset = self.tdy + axis.tick_out + axis.label_offset
                if position & TOP:
                    if has_labels:
                        offset += tdy
                    yt = y + offset
                    self.tl_marker_caches = caches
                else:
                    if has_labels:
                        offset += tdy
                    yt = y + my + dy - offset - axis.markers.values()[0]._frame_min_dy
                    self.br_marker_caches = caches

                for i, (value, marker) in enumerate(axis.markers.items()):
                    cache = PlotCache()
                    caches.append(cache)

                    marker._layout(
                        self.ss_markers[i] - marker._frame_min_dx / 2, yt, marker._frame_min_dx, marker._frame_min_dy, cache)

            else:
                offset = self.rdx + axis.tick_out + axis.label_offset
                if position & RIGHT:
                    if has_labels:
                        offset += tdx
                    xt = x + offset
                    self.br_marker_caches = caches
                else:
                    if has_labels:
                        offset += tdx
                    xt = x + mx + dx - offset - axis.markers.values()[0]._frame_min_dx
                    self.tl_marker_caches = caches

                for i, (value, marker) in enumerate(axis.markers.items()):
                    cache = PlotCache()
                    caches.append(cache)

                    marker._layout(
                        xt, self.ss_markers[i] - marker._frame_min_dy / 2, marker._frame_min_dx, marker._frame_min_dy, cache)
                    
    #---------------------------------------------------------------------------
    #  Layout the axis grid:
    #---------------------------------------------------------------------------
 
    def _layout_grid ( self, x, y, dx, dy, sx, sy ):
        self._x  = x
        self._y  = y
        self._dx = dx
        self._dy = dy
        self._sx = sx
        self._sy = sy

        ticks = self._ticks
        markers = self._markers
        
        # Do log scaling of ticks if necessary:
        if self.log_scale:
            ticks = log10( ticks )
            markers = log10( markers )
            
        self.tick_min = tmin = ticks[0]
        self.tick_max = tmax = ticks[-1]
        dt            = tmax - tmin
 
        if self.horizontal:
            sxdx1 = sx * dx - 1
            # The rounding here is to align the ticks with the center of a pixel
            # so Agg will render the ticks fully within a pixel and not anti-alias
            # them.
            if self.origin & LEFT:
                self.ticks = around(x + ((ticks - tmin) * (float( sxdx1 ) / dt))) + 0.5
                self.ss_markers = around(x + ((markers - tmin) * (float( sxdx1 ) / dt))) + 0.5
            else:
                self.ticks = around(((x + sxdx1) -
                              ((ticks - tmin) * (float( sxdx1 ) / dt)))) + 0.5
                self.ss_markers = around(((x + sxdx1) -
                                         ((markers - tmin) * (float( sxdx1 ) / dt)))) + 0.5
        else:
            sydy1 = sy * dy - 1
            # The rounding here is to align the ticks with the center of a pixel
            # so Agg will render the ticks fully within a pixel and not anti-alias
            # them.
            if self.origin & TOP:
                self.ticks = around((y + sydy1) -
                                   ((ticks - tmin) * (float( sydy1 ) / dt))) + 0.5
                self.ss_markers = around((y + sydy1) -
                                   ((markers - tmin) * (float( sydy1 ) / dt))) + 0.5
                
            else:
                self.ticks = around(y + ((ticks - tmin) * (float( sydy1 ) / dt))) + 0.
                self.ss_markers = around(y + ((markers - tmin) * (float( sydy1 ) / dt))) + 0.5

    #---------------------------------------------------------------------------
    #  Draw an axis:
    #---------------------------------------------------------------------------
 
    def _draw ( self, gc, ox, oy ):
        
        # If the axis is not visible, return immediately:
        axis = self.axis
        if not axis.visible:
            return
 
        # Get the range of 'visible' ticks to reduce the amount of
        # unnecessary drawing done (Note: This is a major performance
        # enhancement when there are lots of labels and a high 'scale' factor):
        horizontal = self.horizontal
        ticks      = self.ticks
        tick_low   = ticks[0]
        tick_high  = ticks[-1]
        dir        = 2 * (tick_high - tick_low) - 1
        if horizontal:
            pt1 = self._x + ox
            pt2 = pt1 + self._dx - 1
        else:
            pt1 = self._y + oy
            pt2 = pt1 + self._dy - 1
        if dir < 0:
            tick_low, tick_high = tick_high, tick_low
        pt1  = min( max( pt1, tick_low ), tick_high )
        pt2  = min( max( pt2, tick_low ), tick_high )
        low  = bin_search1( ticks, pt1, dir )
        high = bin_search1( ticks, pt2, dir )
        if low > high:
            low, high = high, low
        low  = max( 0, low - 1 )
        high = min( len( ticks ), high + 1 )
 
        # If the grid lines are visible, draw them:
        if axis.grid_visible:
            gc.save_state()
            gc.set_antialias(False)
            gc.clip_to_rect(self._x, self._y, self._dx, self._dy)
            gc.set_stroke_color( axis.grid_color_ )
            gc.set_line_width( axis.grid_weight )
            gc.set_line_dash( axis.grid_style_ )                      
         
            n = min( high, len( ticks ) - 1 )
            if horizontal:
                if ox:
                    ticks = ticks - ox
                y1 = self._y
                y2 = y1 + self._dy
                for i in range( low, n ):
                    x = ticks[i]
                    gc.begin_path()
                    gc.move_to( x, y1 )
                    gc.line_to( x, y2 )
                    gc.stroke_path()
            else:
                if oy:
                    ticks = ticks - oy
                x1 = self._x
                x2 = x1 + self._dx
                for i in range( low, n ):
                    y = ticks[i]
                    gc.begin_path()
                    gc.move_to(x1,y)
                    gc.line_to(x2,y)
                    gc.stroke_path()
             
            gc.restore_state()
 
        # Draw the titles:
        title        = axis._title_label
        title.rotate = 90 * (not horizontal)
        if hasattr( self, 'tl_title_cache' ):
            title._draw( gc, cache = self.tl_title_cache )
        if hasattr( self, 'br_title_cache' ):
            title._draw( gc, cache = self.br_title_cache )
 
        # Check for 'top' or 'left' axis:
        if hasattr( self, 'tl' ):
            x, y, dx, dy = self.tl
         
            # Set up the clipping area:
            gc.save_state()
            gc.clip_to_rect(*self.tl)

            # Draw the labels (if any):
            if hasattr( self, 'tl_label_caches' ):
                labels = self._labels
                caches = self.tl_label_caches
                for i in range( len( labels ) ):
                    label = labels[i]
                    if low <= label._tick_index <= high:
                        label._draw( gc, ox, oy, caches[i] ) 

            # Draw the markers (if any):
            if axis.markers is not None and len(axis.markers) > 0:
                caches = self.tl_marker_caches
                for i, (value, marker) in enumerate(self.axis.markers.items()):
                    marker._draw( gc, ox, oy, caches[i] )
                    
            # Draw the axis line and tick marks:
            self._draw_axis( gc, TOP + LEFT, x, y, dx, dy, ox, oy, 
                             low, high )
         
            # Reset the clipping area:
            gc.restore_state()
                      
            # Draw the 'selection' marker (if requested).  Insure that the
            # selection box does not include the tick marks.
            if axis.selected:
                if self.horizontal:
                    axis._draw_selection( gc, x, y + axis.tick_in, dx, dy - axis.tick_in)
                
                else:
                    axis._draw_selection( gc, x, y, dx - axis.tick_in, dy)
                    
        # Check for 'bottom' or 'right' axis:
        if hasattr( self, 'br' ):
            x, y, dx, dy = self.br
         
            # Set up the clipping area:
            gc.save_state()
            gc.clip_to_rect(*self.br)
             
            # Draw the labels:
            if hasattr( self, 'br_label_caches' ):
                labels = self._labels
                caches = self.br_label_caches
                for i in range( len( labels ) ):
                    label = labels[i]
                    if low <= label._tick_index <= high:
                        label._draw( gc, ox, oy, caches[i] ) 
         
            # Draw the markers (if any):
            if axis.markers is not None and len(axis.markers) > 0:
                caches = self.br_marker_caches
                for i, (value, marker) in enumerate(self.axis.markers.items()):
                    marker._draw( gc, ox, oy, caches[i] )

            # Draw the axis line and tick marks:
            self._draw_axis( gc, BOTTOM + RIGHT, x, y, dx, dy, ox, oy, 
                             low, high )
         
            # Reset the clipping area:
            gc.restore_state()
                      
            # Draw the 'selection' marker (if requested).  Insure that the
            # selection box does not include the tick marks.
            if axis.selected:
                
                if self.horizontal:
                    axis._draw_selection( gc, x, y, dx, dy - axis.tick_in )
            
                else:
                    axis._draw_selection( gc, x + axis.tick_in, y, dx - axis.tick_in, dy )
                    
        return
    
    #---------------------------------------------------------------------------
    #  Draw an axis line and tick marks:
    #---------------------------------------------------------------------------
 
    def _draw_axis ( self, gc, position, x, y, dx, dy, ox, oy, low, high ):
        
        ticks        = self.ticks
        axis         = self.axis
        my           = axis.margin_height
        tick_visible = axis.tick_visible
        if tick_visible:
            tick_in  = axis.tick_in
            tick_out = axis.tick_out
        else:
            tick_in = tick_out = 0
 
        # Draw the axis line:
        gc.set_stroke_color( axis.line_color_ )
        gc.set_line_width( axis.line_weight )
        gc.set_line_dash( axis.line_style_ )                      
        gc.set_antialias( False )
        
        if self.horizontal:
            minx = self._x
            maxx = minx + self._dx - 1
            if position & TOP:
                y1 = y + tick_in + my - 1   ### WHY -1???
            else:
                y1 = y + dy - tick_in - my
            gc.begin_path()
            gc.move_to( minx, y1 )
            gc.line_to( maxx, y1 )
            gc.stroke_path()
        else:
            miny = self._y
            maxy = miny + self._dy - 1
            if position & LEFT:
                x1 = x + dx - tick_in - my
            else:
                x1 = x + tick_in + my - 1   ### WHY -1???
            gc.begin_path()
            gc.move_to( x1, miny )
            gc.line_to( x1, maxy )
            gc.stroke_path()
 
        # Draw the tick marks:
        if tick_visible:
            tick_len = tick_in + tick_out
            gc.set_stroke_color( axis.tick_color_ )
            gc.set_line_width( axis.tick_weight )
            gc.set_line_dash( axis.tick_style_  )                      
            if self.horizontal:
                if ox:
                    ticks = ticks - ox
                if position & TOP:
                    y1 = y  + my
                    y2 = y1 + tick_len
                else:
                    y2 = y + dy - my
                    y1 = y2 - tick_len
                for x1 in ticks[ low: high ]:
                    if (x1 >= minx) and (x1 < maxx):                   
                        gc.begin_path()
                        gc.move_to( x1, y1 )
                        gc.line_to( x1, y2 )
                        gc.stroke_path()
            else:
                if oy:
                    ticks = ticks - oy
                if position & LEFT:
                    x2 = x + dx - my
                    x1 = x2 - tick_len
                else:
                    x1 = x + my
                    x2 = x1 + tick_len
                for y1 in ticks[ low: high ]:
                    if (y1 >= miny) and (y1 < maxy):                    
                        gc.begin_path()
                        gc.move_to( x1, y1 )
                        gc.line_to( x2, y1 )
                        gc.stroke_path()

    #---------------------------------------------------------------------------
    #  Return the closest (x',y') point along the axis line correponding to a
    #  specified (x,y) point:
    #
    #  Note: The coordinate not along the axis line is always returned as 0.
    #---------------------------------------------------------------------------
 
    def closest ( self, x, y ):
        if self.horizontal:
            xl, xr = self.ticks[0], self.ticks[-1]
            if xl > xr:
                xl, xr = xr, xl
            return ( min( max( x, xl ), xr ), 0 )
       
        yt, yb = self.ticks[0], self.ticks[-1]
        if yt > yb:
            yt, yb = yb, yt
        return ( 0, min( max( y, yt ), yb ) ) 
 
    #---------------------------------------------------------------------------
    #  Return the length of the axis line:
    #---------------------------------------------------------------------------
 
    def axis_length ( self ):
        return abs( self.ticks[0] - self.ticks[-1] )
 
    #---------------------------------------------------------------------------
    #  Set a new zoom range for the axis:
    #---------------------------------------------------------------------------
 
    def zoom ( self, x1, x2 ):
 
        # Convert from screen space to data space:
        x0 = self.ticks[0]
        dx = float( abs( self.ticks[-1] - x0 ) )
        if self.log_scale:
            d0 = log10( self._ticks[0] )
            dd = log10( self._ticks[-1] ) - d0
            d1 = 10.0 ** (d0 + ((float( abs( x1 - x0 ) ) / dx) * dd))
            d2 = 10.0 ** (d0 + ((float( abs( x2 - x0 ) ) / dx) * dd))
        else:
            d0 = self._ticks[0]
            dd = self._ticks[-1] - d0
            d1 = d0 + (float( abs( x1 - x0 ) ) / dx) * dd
            d2 = d0 + (float( abs( x2 - x0 ) ) / dx) * dd
 
        # Make sure d1 < d2:
        if d2 < d1:
            d1, d2 = d2, d1
           
        axis = self.axis
 
        # If this is the first zoom, save the original axis bounds:
        if axis._bound_low is None:
            axis._bound_low  = axis.bound_low
            axis._bound_high = axis.bound_high
 
        # Set the new axis bounds:
        axis.bound_low  = d1
        axis.bound_high = d2

    #---------------------------------------------------------------------------
    #  Unzoom the axis (if possible):
    #---------------------------------------------------------------------------
 
    def unzoom ( self ):
        axis = self.axis
        if axis._bound_low is not None:
            axis.bound_low  = axis._bound_low
            axis.bound_high = axis._bound_high
            axis._bound_low = None
            return True
        return False
       
    #---------------------------------------------------------------------------
    #  Create the set of PlotLabels to be drawn with the axis:
    #---------------------------------------------------------------------------
  
    def make_labels ( self, gc ):
        axis         = self.axis
        style        = axis.label_style_
        ticks        = self._ticks
        font         = axis.label_font
        color        = axis.label_color_
        format       = axis.label_format
        self._labels = []
           
        # This code tries to omit the first and/or last labels when they are
        # too close to the next ticks due to 'fit' settings:
        if self.labels is not None:
            first = last = 1
        else:
            first = last = 0
            if len( ticks ) > 3:
               delta = ticks[2] - ticks[1]
               first = ((ticks[1]  - ticks[0])  < .25 * delta)
               last  = ((ticks[-1] - ticks[-2]) < .25 * delta)
  
        if style < 0:
            # Handle only the end ticks being labeled:
#           dx, dy = self.make_label( gc, first, font, color, format, axis )
#           if (first == 0) and (style == -1):
            dx, dy = self.make_label( gc, 0, font, color, format, axis )
            if style == -1:
                left   = dx / 2
                middle = (dx + 1) / 2
            else:
                left   = 0
                middle = dx + 6  # '6' = left/right edge padding (3 pixels each)
#           end = len( ticks ) - 1 - last
#           if end <= first:
#               return ( left, middle, 0, dy )
#           dx, dy2 = self.make_label( gc, end, font, color, format, axis )
#            right   = 0
#            if last == 0:
#                right = dx / 2
#                dx    = (dx + 1) /2
#            middle += dx + 4 * (middle != 0)
            dx, dy2 = self.make_label( gc, len( ticks ) - 1 , font, color, 
                                       format, axis )
            right = 0
            if style == -1:
                right = dx / 2
                dx    = (dx + 1) /2
            middle += dx + 4 * (middle != 0)
            return ( left, middle, right, max( dy, dy2 ) )
  
        # Otherwise, generate labels every style'th tick mark:
        max_dy = 0
        for i in range( first, len( ticks ) - last, style ):
            dx, dy = self.make_label( gc, i, font, color, format, axis )
            max_dy = max( max_dy, dy )
            if i == first:
                if first == 0:
                    left   = dx / 2
                    middle = (dx + 1) / 2
                else:
                    left   = 0
                    middle = dx
        if len( self._labels ) <= 1:
            return ( left, middle, 0, max_dy )
        right = 0
        label = self._labels[-1]
        if label._tick_index == (len( ticks ) - 1):
            right = dx / 2
            dx    = (dx + 1) /2
        middle += dx + 4 * (middle != 0)
        return ( left, middle, right, max_dy )

    #---------------------------------------------------------------------------
    #  Create a single PlotLabel to be drawn with the axis:
    #---------------------------------------------------------------------------
  
    def make_label ( self, gc, index, font, color, format, container ):
       
#       label = PlotLabel( text      = format % self._ticks[ index ],
#                          font      = font,
#                          color     = color,
#                          container = container )
#       label._tick_index = index
  
        # Note: The following code is equivalent to the commented out code 
        # above, but runs a lot faster because it bypasses the trait checking
        # mechanism. We know the traits must be valid, because they are
        # obtained from the 'PlotAxis' object traits, which have already been
        # checked. Speed is important here because some axes generate lots of
        # labels, which are recreated each time '_min_size' is called from the
        # top-level:
        label = PlotLabel()
        ldic  = label.__dict__
        if self.labels is not None:
            if not self.axis.tick_interval == 'auto':
                value = self.labels[ (index-1)*int(self.axis.tick_interval) ]
            else:
                value = self.labels[ index - 1 ]
        else:
            value = format( self._ticks[ index ] )
        ldic[ '_text_list' ]    = [ value ]
        ldic[ 'font'  ]         = font
        ldic[ 'color_' ]        = color
        ldic[ 'container' ]     = container
        ldic[ '_tick_index' ]   = index
        ldic[ 'margin_height' ] = 2
        ldic[ 'margin_width' ]  = 2 
  
        self._labels.append( label )
  
         # Return the minimum size required by the label:
        dx, dy = label._min_size( gc )
        if self.horizontal:
            return ( dx, dy )
        return ( dy, dx )

    def _get_markers_min_size( self, gc, markers ):
        """ Get the minimum sizes required to draw all the markers. """

        dx = 0
        dy = 0
        for marker in markers:
            mdx, mdy = marker._min_size( gc )
            dx = max(dx, mdx)
            dy = max(dy, mdy)

        return (dx, dy)
