""" Style sheet for PlotAxis objects. """


from enthought.enable.traits.api import RGBAColor
from enthought.kiva.traits.kiva_font_trait import KivaFont
from enthought.traits.api import HasStrictTraits, Range, Str, true

from axis_traits import AxisBounds, AxisPosition, AxisScale, LabelFormat
from axis_traits import LabelStyle, TickIn, TickInterval, TickOut
from plot_traits import LineStyle, LineWeight, Margin


class AxisStyle(HasStrictTraits):
    """ Style sheet for PlotAxis objects. """

    # Is the axis visible?
    visible = true(
        desc  = 'whether the axis is visible',
        label = 'Is Visible?'
    )

    # The scaling mode of the axis (linear or log)
    scale = AxisScale(
        desc  = 'the scaling mode of the axis',
        label = 'Scaling'
    )

    # The sides of the plot the axis appears on.
    position = AxisPosition(
        desc = 'the sides of the plot the axis appears on',
        label = 'Position'
    )

    ## FIXME : I think this is probably broken now.
    # The popup menu for the axis.
    menu          = Str(desc  = 'the axis pop-up menu',
                        label = 'Popup Menu Text')

    # The thickness of the axis line (in pixels).
    line_weight = LineWeight(
        desc  = 'the axis line weight (in pixels)',
        label = 'Line Weight'
    )

    # The color of the axis line.
    line_color = RGBAColor(
        'black',
        desc  = 'the axis line color',
        label = 'Line Color'
    )

    # The line style of the axis line.
    line_style = LineStyle(
        desc  = 'the axis line style',
        label = 'Line Style'
    )

    # The font for the axis label.
    label_font = KivaFont(
        'modern 10',
        desc  = 'the axis label font',
        label = 'Label Font'
    )

    # The text color of the axis label.
    label_color = RGBAColor(
        'black',
        desc  = 'the axis label text color',
        label = 'Label Color'
    )

    # The position of the axis label.
    label_style = LabelStyle(
        desc  = 'the location of labels along the axis',
        label = 'Label Positions'
    )

    # The offset of the axis labels from the tick marks.
    label_offset = Range(
        0, 31, value = 3,
        desc  = 'the offset of axis labels from the tick marks',
        label = 'Label Offset'
    )

    # The format to apply to the axis labels.
    label_format = LabelFormat(
        desc  = 'the format to apply to axis labels',
        label = 'Label Format'
    )

    # Are the tick marks visible?
    tick_visible = true(
        desc  = 'whether the axis tick marks are visible',
        label = 'Are Ticks Visible?'
    )

    # The line thickness of the axis tick marks.
    tick_weight = LineWeight(
        desc  = 'the axis tick mark line weight (in pixels)',
        label = 'Tick Line Weight'
    )

    # The color of the axis tick marks.
    tick_color = RGBAColor(
        'black',
        desc  = 'the axis tick mark color',
        label = 'Tick Color'
    )

    # The line style of the axis tick marks.
    tick_style = LineStyle(
        desc  = 'the axis tick mark line style',
        label = 'Tick Line Style'
    )

    # The number of pixels the tick marks tick into the plot.
    tick_in = TickIn(
        desc  = 'the length of axis ticks into the plot (in pixels)',
        label = 'Inside Tick Length'
    )

    # The number of pixels the tick marks tick out of the plot.
    tick_out = TickOut(
        desc  = 'the length of axis ticks out from the plot (in pixels)',
        label = 'Outside Tick Length'
    )

    # The spacking of the tick marks
    tick_interval = TickInterval(
        desc  = 'the spacing of axis tick marks',
        label = 'Tick Interval'
    )

    # Are the grid lines visible?
    grid_visible = true(
        desc  = 'whether the axis grid lines are visible',
        label = 'Are Grids Visible?'
    )

    # The line thickness of the axis grid.
    grid_weight = LineWeight(
        desc  = 'the axis grid line weight (in pixels)',
        label = 'Grid Line weight'
    )

    # The line color of the axis grid.
    grid_color = RGBAColor(
        'grey',
        desc  = 'the axis grid line color',
        label = 'Grid Line Color'
    )

    # The line style of the axis grid.
    grid_style = LineStyle(
        'dot',
        desc  = 'the axis grid line style',
        label = 'Grid Line Style'
    )

    # The margin space around the outside of the axis.
    margin = Margin(
        desc = 'the margin space around the outside of the axis (in pixels)'
    )

    # The margin space outside the top and bottom sides of the axis.
    margin_height = Margin(
        desc = 'the margin space outside the top and bottom sides of the axis '
               '(in pixels)',
        label = 'Margin Height'
    )

    # The margin space outside the left and right sides of the axis.
    margin_width = Margin(
        desc = 'the margin space outside the left and right sides of the axis '
               '(in pixels)',
        label = 'Margin Width'
    )

    # The lower axis bound.
    bound_low = AxisBounds(
        desc = 'the lower axis bound',
        label = 'Lower Bound'
    )

    # The upper axis bound.
    bound_high = AxisBounds(
        desc  = 'the upper axis bound',
        label = 'Upper Bound'
    )

    # Is the axis zoomable?
    zoomable = true(
        desc  = 'whether the axis can be drag zoomed by the user',
        label = 'Is Axis Zoomable?'
    )


#### EOF ######################################################################
