""" A collection of trait definitions which only apply to axes. """

import plot_base

from enthought.traits.api import Range, Trait, TraitType, TraitPrefixList
from enthought.traits.api import TraitPrefixMap, TraitRange, TraitError
from enthought.traits.ui.api import ImageEnumEditor

from plot_base import BOTTOM, LEFT, RIGHT, TOP, BIG_POS_FLOAT

###############################################################################
# Axis Bounds.
###############################################################################

_axis_bounds = ['auto', 'fit']

AxisBounds = Trait(
    'auto',
    TraitPrefixList(_axis_bounds),
    TraitType(0.0)
)

###############################################################################
# Axis Positions.
###############################################################################

_axis_positions = {
    'top':    TOP + RIGHT,
    'right':  TOP + RIGHT,
    'left':   BOTTOM + LEFT,
    'bottom': BOTTOM + LEFT,
    'both':   TOP + LEFT + BOTTOM + RIGHT,
    'all':    TOP + LEFT + BOTTOM + RIGHT
}

_edit_axis_positions = [ 'left', 'right',  'all',  'top',  'bottom', 'both' ]

_axis_position_editor = ImageEnumEditor(
    cols   = 3,
    suffix = '_axis',
    values = _edit_axis_positions,
    module = plot_base
)

AxisPosition = Trait(
    'bottom',
    TraitPrefixMap(_axis_positions),
    editor = _axis_position_editor
)

###############################################################################
# Axis Scale.
###############################################################################

SCALE_LINEAR = 0
SCALE_LOG    = 1

_axis_scaling = {
    'linear': SCALE_LINEAR,
    'log':    SCALE_LOG
}

AxisScale = Trait('linear', TraitPrefixMap(_axis_scaling), cols=2)

###############################################################################
# Label styles.
###############################################################################

_label_styles = {
    'all':   1,  # Labels at every tick mark.
    'end':  -1,  # Labels centered on the outermost tick marks.
    'edge': -2,  # Labels for outermost tick marks justified to plot edge.
    'none':  0   # No labels.
}

_edit_label_styles = [
    'all', 'none', 'end', 'edge', '2', '3', '4'
]

_label_style_editor = ImageEnumEditor(
    cols   = 2,
    suffix = '_labels',
    values = _edit_label_styles,
    module = plot_base
)

LabelStyle = Trait(
    'all',
    TraitPrefixMap(_label_styles),
    TraitRange(1, 10),
    editor = _label_style_editor
)

###############################################################################
# Label format.
###############################################################################

def _format_test(object, name, value):
    """ Test that value is a valid label format.

    Valid label formats are Python conversion strings or functions which
    convert a number to a string.

    NOTE: As this is currently implemented, a side effect of this validator
    is to wrap Python conversion strings in a function before assigning the
    trait.
    """

    # If we have a string, we will test that it is a valid Python conversion
    # string by wrapping it in a function and then invoking the function.
    if isinstance(value, basestring):
        value = lambda n, f = value: f % n

    # Here value is now the function that was passed in or string wrapped
    # up in a function.  Invoke it to ensure it returns a string.
    if isinstance(value(0.0), basestring):
        return value

    raise TraitError

_format_test.info = ( "a standard Python conversion string (e.g. '%%.2f') "
                      "or a function which converts a number to a string" )

LabelFormat = Trait('%g', _format_test)

###############################################################################
# Tick in.
###############################################################################

_tick_in_editor = ImageEnumEditor(
    cols   = 4,
    suffix = '_tick_in',
    values = [ str(x) for x in xrange(0, 32, 2) ],
    module = plot_base
)

TickIn = Range(0, 31, value=5, editor=_tick_in_editor)

###############################################################################
# Tick interval.
###############################################################################

_tick_intervals = [ 'auto' ]

TickInterval = Trait(
    'auto',
    TraitPrefixList(_tick_intervals),
    TraitRange(0.0, BIG_POS_FLOAT),
    TraitRange(-100, -1)
)

###############################################################################
# Tick out.
###############################################################################

_tick_out_editor = ImageEnumEditor(
    cols   = 4,
    suffix = '_tick_out',
    values = [ str(x) for x in xrange(0, 32, 2) ],
    module = plot_base
)

TickOut = Range(0, 31, value=0, editor=_tick_in_editor)



#### EOF ######################################################################
