#-------------------------------------------------------------------------------
#
#  Defines the BackBuffer object, which allows PlotValues to be rendered onto
#  a secondary graphics context in addition to their main drawing window.
#  Also defines a utility function which allows rendering PlotValues onto
#  arbitrary graphics contexts.  
#
#  Written by: Peter Wang
#
#  Date: 06/29/2005
#
#  (c) Copyright 2005 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

from enthought.traits.api import Bool, Enum, HasTraits, Instance
from enthought.kiva.agg import GraphicsContextArray

class BackBuffer(HasTraits):
    """
    Encapsulates the information about a back buffer for PlotValues.
    PlotValues should not share BackBuffer instances.  (However,
    BackBuffers can share a common GC.)
    """
    
    # Draws the front buffer (screen), the back buffer, both buffers, or none.
    mode = Enum("draw_both", "draw_none", "draw_back", "draw_front")
    
    # The graphics context to use as the back buffer; this doesn't have
    # to be unique to this backbuffer.
    gc = Instance(GraphicsContextArray)
    
    # If draw_once is set, then after a PlotValue has rendered onto this
    # BackBuffer, it sets the "is_drawn" flag and it will not draw onto
    # this buffer again, until either "is_drawn" is set False again,
    # or if draw_once is set False.
    draw_once = Bool(False)
    is_drawn = Bool(False)

def gc_plot(plotvalues, top_parent, layout_size, gc=None, pix_format="rgb24"):
    """
    Parameters
    ----------
        
    plotvalues
        list[PlotValue]
    layout_size
        tuple(int, int)
    top_parent
        PlotCanvas, PlotGroup, etc.; the top-level Chaco
        element that draws plotvalue onto the screen and,
        more importantly, is the object responsible for
        laying out plotvalue onto a screen region of size
        layout_size.
    gc
        A graphics context 
        
    If "gc" is supplied, then a new graphics context is not created,
    and the plotvalues are rendered into this gc.  NOTE: it is the caller's
    responsibility to make sure that the supplied gc is the correct size;
    if it is not, then renderings may be clipped.
    
    If "gc" is not supplied, then pix_format specifies the pixel format
    of the new GC to create.  Default is "rgb24".
    
    Note that PlotValues render in the order supplied in "plotvalues".
    
    Description
    -----------
    Takes a list of plotvalues, their parent, and a desired plot layout size,
    and returns a gc that consists of just the plotvalue's rendering.  The
    returned GC will have a size equal to the size of the plotvalue's plot
    area in the layout; this is usually quite smaller than layout_size.
    
    If an appropriate top_parent is not given, then there is no guarantee
    about the returned GC reflecting the contents of the screen.  (A different
    top_parent implies a different layout, which implies different plot_bounds
    for the plotvalue.)
    """
    
    # Create a dummy graphics context and render the plot into that; this
    # way, we can look at the plotvalue's extents and know what the real plot
    # size needs to be.
    dummy_gc = GraphicsContextArray(layout_size)

    top_parent.render(dummy_gc, 0, 0, layout_size[0], layout_size[1])
    
    if gc == None:
        # Now create the real GC we will paint into, using the plot extents cached
        # in the first plotvalue.  (We're OK to use the first plotvalue because
        # presumably the top_parent layout has set the extents appropriately for
        # all the plotvalues.
        gc = GraphicsContextArray((plotvalues[0]._plot_bounds[2],
                                   plotvalues[0]._plot_bounds[3]),
                                  pix_format = pix_format)
    
    for plotvalue in plotvalues:
        plotvalue.backbuffer = BackBuffer(gc=gc, draw_mode="draw_back")
    
    # Draw again
    top_parent.render(dummy_gc, 0, 0, layout_size[0], layout_size[1])

    # Remove the backbuffer
    for plotvalue in plotvalues:
        plotvalue.backbuffer = None
    
    return gc


#EOF
