"""
Color map class for rendering image plots.
"""


from enthought.util.numerix import Float as NumericFloat
from enthought.util.numerix import array, arange, searchsorted, choose, ravel, \
                                   resize, concatenate, zeros
from enthought.traits.api import Trait, HasTraits, Enum, RGBColor, Float, Array, \
                             Int, Property

class AbstractColormap(HasTraits):
    """
    A ColorMap not only maps from scalar values to color values, it also
    serves as the value axis for an ImagePlotValue.
    """
    
    color_depth = Enum('rgba', 'rgb')
    
    def map_value(self, val):
        """
        map_value(val) -> color
        
        Maps a single value to a single color.  Color is represented either as
        a length-3 array or length-4 array depending on the colordepth setting.
        """
        raise NotImplementedError()
    
    def map_array(self, ary):
        """
        map_array(ary) -> color_array
        
        Returns an array of values containing the colors mapping to ary.
        If the input ary is NxM, the returned array is NxMx3 or NxMx4
        depending on what the colordepth of the colormap is set to.
        """
        raise NotImplementedError()

    def map_array_index(self, ary):
        """
        map_array_index(ary) -> index into color_bands
        
        Like map_array(), but returns an array of indices into the colormap's
        color bands instead of an array of colors.  If the input is
        NxM, then output is NxM integer indices.
        
        This function might not apply to all colormaps.  Ones that cannot
        define a static set of color bands (e.g. function-defined colormaps)
        will not be able to implement this function.
        """
        raise NotImplementedError()


class LinearColormap(AbstractColormap):
    
    min_color = Property(Array)   # RGBColor('black')
    max_color = Property(Array)   # RGBColor('white')
    min_value = Float(0.0)
    max_value = Float(1.0)
    steps = Int(10)

    # An array of color values, in order from min_color to max_color.
    # Dimensions are (steps x 3) or (steps x 4) depending on whether
    # colors are 24 or 32 bit.
    color_bands = Property(Array)

    # shadow traits
    _min_color = Array
    _max_color = Array
    
    # _color_band is a sequence of self.steps number of colors, arranged in 
    # order from min_color to max_color.
    _color_bands = Array
    _delta_color = Array
    
    # _value_bands is an array of floats that define the start of intervals
    # corresponding to each band in _color_bands.  basically, the index
    # of a value into _value_bands returned by Numeric.searchsorted()-1 
    # defines which color_band colors that value should get.
    _value_bands = Array
    _delta_value = Float(1.0)
    
    def __init__(self, color_depth='rgba', min_value=0.0, max_value=1.0,
                       min_color=None, max_color=None, steps=10):
        
        self.__constructing = True
        
        default_min_color = { 'rgba': [0.0]*3 + [1.0], 'rgb': [0.0]*3 }
        default_max_color = { 'rgba': [1.0]*4, 'rgb': [1.0]*3 }
        
        self.color_depth = color_depth
        if min_color is None:
            self.min_color = array(default_min_color[color_depth])
        else:
            self.min_color = array(min_color)
        
        if max_color is None:
            self.max_color = array(default_max_color[color_depth])
        else:
            self.max_color = array(max_color)
        
        self.min_value = min_value
        self.max_value = max_value
        self.steps = steps
        
        self.__constructing = False
        
        self._calc_value()
        self._calc_color()

    def map_value(self, val):
        if val > self.max_value:
            val = self.max_value
        elif val < self.min_value:
            val = self.min_value
        return (val - self.min_value) / self._delta_value * self._delta_color + \
                self._min_color

    def map_array(self, ary):
        if self.color_depth == 'rgb':
            bytedepth = 3
        else:
            bytedepth = 4
        output = zeros((len(ravel(ary)) * bytedepth), NumericFloat)
        color_indices = searchsorted(self._value_bands, ravel(ary)) - 1
        for index, color_index in enumerate(color_indices):
            output[index*bytedepth : (index+1)*bytedepth] = \
                self._color_bands[color_index]
        return resize(output, ary.shape + (bytedepth,))


    def map_array_index(self, ary):
        color_indices = searchsorted(self._value_bands, ravel(ary)) - 1
        return resize(color_indices, ary.shape)
    
    #------------------------------------------------------------------------
    # Private recomputation functions
    #------------------------------------------------------------------------
    
    def _calc_color(self):
        "Recomputes the cached color information"
        
        if self.__constructing:
            return
        
        self._delta_color = self._max_color - self._min_color
        
        # calculate the colors for each band of values; since steps indicates
        # the number of colors to use, and not the number of intervals,
        # we have to subtract 1.
        color_step = self._delta_color / (self.steps-1)
        self._color_bands = [i * color_step + self._min_color \
                                for i in arange(0,self.steps)]
    
    def _calc_value(self):
        "Recomputes cached value information"
        
        if self.__constructing:
            return
        
        self._delta_value = (self.max_value - self.min_value) / (self.steps-1)
        self._value_bands = arange(self.min_value, self.max_value, self._delta_value)
        # because we need to map self._min_value to 1, we need to make
        # self._value_bands[0] just a little bit more negative.
        self._value_bands[0] -= self._delta_value / 10.0


    #------------------------------------------------------------------------
    # Color Property getters and setters
    #------------------------------------------------------------------------
    
    def _set_min_color(self, new_val):
        if len(new_val) != len(self.color_depth):
            raise RuntimeError, "Color value must be length %d!" % (len(self._color_depth))
        self._min_color = new_val
        self._calc_color()
    
    def _get_min_color(self):
        return self._min_color

    def _set_max_color(self, new_val):
        if len(new_val) != len(self.color_depth):
            raise RuntimeError, "Color value must be length %d!" % (len(self._color_depth))
        self._max_color = new_val
        self._calc_color()
    
    def _get_max_color(self):
        return self._max_color

    def _get_color_bands(self):
        return self._color_bands

    #------------------------------------------------------------------------
    # Trait event handlers
    #------------------------------------------------------------------------

    def _min_value_changed(self, old_val, new_val):
        if new_val > self.max_value:
            self.min_value = self.max_value 
        self._calc_value()
    
    def _max_value_changed(self, old_val, new_val):
        if new_val < self.min_value:
            self.max_value = self.min_value
        self._calc_value()

    def _steps_changed(self, old, new):
        self._calc_color()
        self._calc_value()
    
    def _color_depth_changed(self, old, new):
        if new == 'rgba':
            # add a 1.0 for the alpha value
            self._min_color = concatenate([self._min_color, [1.0]])
            self._max_color = concatenate([self._max_color, [1.0]])
        elif new == 'rgb':
            self._min_color = self._min_color[:3]
            self._max_color = self._max_color[:3]
        else:
            raise PlotError, "Unknown color depth '%s' in ColorMap!" % new
        self._calc_color()


class LogColormap(AbstractColormap):
    pass




#EOF
