""" A PlotData wrapper to provide contours for an ImageData object. """


from enthought.util.scipyx import arange, array, mgrid, ravel, shape, searchsorted, Float

# Enthought library imports.
from enthought.chaco import AbstractColormap, ImageData, PlotData, PlotValue
from enthought.enable.traits.api import RGBAColor
from enthought.traits.api import Array, Dict, Instance, List, Property

# Contouring extension module.
from contour import Cntr


class ContourData(PlotData):
    """ A PlotData wrapper to provide contours for an ImageData object. """

    # The color of the contour lines.  Used iff a colormap is not specified.
    color = RGBAColor('black')

    # The colormap for the contour lines.
    colormap = Instance(AbstractColormap, event='update')

    # The image data to contour over.
    image_data = Instance(ImageData, event='update')

    # The levels where the contour should be.
    levels = Array(event='update')

    # The names of the valid sources.
    sources = Property(List)

    #### Private Traits ####

    # The underlying trace data returned by the contouring algorithm.
    _traces = Property(Dict)

    # The underlying polygon data returned by the contouring algorithm.
    _polygons = Property(Dict)

    ###########################################################################
    # object interface.
    ###########################################################################

    def __init__(self, data, **traits):
        """ Constructor. """

        super(ContourData, self).__init__(**traits)

        self.image_data = data

        return

    ###########################################################################
    # Public 'ContourData' interface.
    ###########################################################################

    def auto_level(self, n_levels):
        """ Create n_levels evenly spaced contour levels across the data. """

        flat = ravel(self.image_data.value)
        minimum = min(flat)
        maximum = max(flat)

        self.levels = \
            arange(minimum, maximum+0.00000000000001, (maximum-minimum)/(n_levels+2))

        return

    ###########################################################################
    # Public 'PlotData' interface.
    ###########################################################################

    def add_to(self, plottable, source_name='default', line_weight=1):
        """ Add the specified channel to the plot. """

        if source_name == 'regions':
            for level, polygons in self._polygons.items():
                for i, polygon in enumerate(polygons):
                    source = 'region %s,%d' % (str(level), i)
                    color = self._get_color(source)
                    plot_value = PlotValue(type='polygon', fill_color=color, line_color='clear')
                    self.get_source(plot_value, source)
                    plottable.add(plot_value)

        else:
            for level, traces in self._traces.items():
                for i, trace in enumerate(traces):
                    source = 'contour %s,%d' % (str(level), i)
                    color = self._get_color(source)
                    plot_value = PlotValue(type='line', line_color=color, line_weight=line_weight)
                    self.get_source(plot_value, source)
                    plottable.add(plot_value)

        return


    def data(self, source_name):
        """ Get the data for the specified channel name.

        For this class, the channel name is expected to be of the form:
          'contour <level>,<index>'
        Where level is the value of the contour level (e.g., 0.2) and index
        is the index (i.e., an integer) into the set of contour lines at that
        level.
        """

        level, index = self._parse_source_name(source_name)
        if source_name.startswith('region'):
            data = self._polygons[level][int(index)]
        else:
            data = self._traces[level][int(index)]

        return data


    ###########################################################################
    # Private 'ContourData' interface.
    ###########################################################################

    def _get_color(self, source_name):
        """ Get the color for the specified source channel. """

        level, ignore = self._parse_source_name(source_name)
        if self.colormap is not None:
            color = None
            levelindex = self._nearestlevel(float(level))
            try:
                color = tuple(self.colormap.map_value(float(levelindex)))
            except NotImplementedError:
                color = tuple((self.colormap.map_array(array([float(levelindex)])))[0])

            return color

        else:
            # If no colormap was specified, return black.
            return self.color

    def _nearestlevel(self, level):
        nearestlevel = 0
        distance = abs(level-self.levels[0])
        for i in range(1,len(self.levels)):
            if abs(level-self.levels[i])<distance:
                distance = abs(level-self.levels[i])
                nearestlevel = i
        return nearestlevel

    def _parse_source_name(self, source_name):
        """ Parse the source name into the level and index.

        Returns (level, index)
        """

        id = source_name.split(' ')[1]
        return id.split(',')


    #### Property Accessors ####

    def _get_sources(self):
        """ Property accessor for the sources trait. """

        # Create the source list if the shadow trait has not been cached.
        if self._shadow_sources is None:
            self._shadow_sources = []
            for level, traces in self._traces.items():
                for i, trace in enumerate(traces):
                    self._shadow_sources.append('contour %s,%d' % (str(level), i))

            for level, polygons in self._polygons.items():
                for i, polygon in enumerate(polygons):
                    self._shadow_sources.append('region %s,%d' % (str(level), i))

        return self._shadow_sources

    def _calculate_scaling(self):
            rows, cols = shape(self.image_data.value)
            xlower = self.image_data.index[0][0]
            xupper = self.image_data.index[0][-1]
            ylower = self.image_data.index[1][0]
            yupper = self.image_data.index[1][-1]
            xscale = (float(cols-1)/(float(xupper-xlower)))
            yscale = (float(rows-1)/float(yupper-ylower))
            return (xlower, xscale, ylower, yscale)


    def _get__traces(self):
        """ Property accessor for the _traces trait. """

        if self._shadow_traces is None:
            self._shadow_traces = {}
            rows, cols = shape(self.image_data.value)
            xlower, xscale, ylower, yscale = self._calculate_scaling()

            x, y = mgrid[:rows, :cols]
            contour = Cntr(x, y, self.image_data.value)
            for level in self.levels:
                trace = contour.trace(level)
                newtrace = []
                for innertuple in trace:
                    newtuple = []
                    newtuple.append([(x/xscale)+xlower for x in innertuple[0]])
                    newtuple.append([(y/yscale)+ylower for y in innertuple[1]])
                    newtrace.append(tuple(newtuple))
                self._shadow_traces[str(level)] = newtrace

        return self._shadow_traces

    def _get__polygons(self):
        """ Property accessor for the _polygons trait. """

        if self._shadow_polygons is None:
            self._shadow_polygons = {}
            rows, cols = shape(self.image_data.value)
            (xlower, xscale, ylower, yscale) = self._calculate_scaling()
            x, y = mgrid[:rows, :cols]

            contour = Cntr(x, y, self.image_data.value)
            for i, level in enumerate(self.levels[:-1]):
                trace = contour.trace(level, self.levels[i+1])
                newtrace = []
                for innertuple in trace:
                    newtuple = []
                    newtuple.append([(x/xscale)+xlower for x in innertuple[0]])
                    newtuple.append([(y/yscale)+ylower for y in innertuple[1]])
                    newtrace.append(tuple(newtuple))

                self._shadow_polygons[str(level)] = newtrace

        return self._shadow_polygons

    #### Trait Listeners ####

    def _update_changed(self, name):
        """ Called if any of the 'update' traits changes. """

        # Clear out the cached shadow traits.
        self._shadow_traces = None
        self._shadow_sources = None
        self._shadow_polygons = None

        # Then force connected plots to update appropriately.
        super(ContourData, self)._update_changed(name)

        return

#### EOF ######################################################################
