#-------------------------------------------------------------------------------
#  
#  Load an ArcView 'shape' file  
#  
#  Written by: David C. Morrill
#  
#  Date: 02/20/2003
#  
#  (c) Copyright 2003, Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import os.path

from struct                      import unpack
from enthought.traits.api            import Str, Trait #, trait_editors # ui_hack

from enthought.chaco.plot_canvas import PlotCanvas
from enthought.chaco.plot_data   import PlotData
from enthought.chaco.plot_value  import PlotValue


#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# String for formatting source names:
SOURCENAME = '%s #%d'

#-------------------------------------------------------------------------------
#  'ArcView' class:
#-------------------------------------------------------------------------------

class ArcView ( PlotData ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------
   
    file_name  = Str # ( editor = trait_editors().TraitEditorFile() ) # ui_hack
    _file_name = Str
       
    #---------------------------------------------------------------------------
    #  Return the data for a specified source: 
    #---------------------------------------------------------------------------
    
    def data ( self, source_name ):
        self.init()
        try:
            return self._data[ source_name ]
        except:
            return [ 0.0, 1.0 ] 
        
    #---------------------------------------------------------------------------
    #  Make sure we have been initialized:
    #---------------------------------------------------------------------------
     
    def init ( self ):
        if self.file_name != self._file_name:
            try:
                self._file_name = self.file_name
                self.read()
            except:
                pass
    
    #---------------------------------------------------------------------------
    #  Create a new source:
    #---------------------------------------------------------------------------
    
    def source ( self, value, source_name ):
        self.init()
        if source_name == 'default':
            source_name = self._default_source
        return PlotData.source( self, value, source_name )
        
    #---------------------------------------------------------------------------
    #  Add one or more source connections to a plottable object:
    #---------------------------------------------------------------------------
    
    def add_to ( self, plottable, source_name ):
        if isinstance( plottable, PlotValue ):
            self.get_source( plottable, source_name or self.__sources__[0] )
            return
        if source_name is not None:
            value = PlotValue( plot_type = 'polygon' )
            self.get_source( value, source_name )
            plottable.add( value )
            return
        self.init()
        canvas = PlotCanvas( plot_type = self._plot_type )
        for source_name in self._data.keys():
            value = PlotValue()
            self.get_source( value, source_name )
            canvas.add( value )
        plottable.add( canvas )
        
    #---------------------------------------------------------------------------
    #  Read the contents of an ArcView 'shape' file: 
    #---------------------------------------------------------------------------
    
    def read ( self ):
        self._data = {}
        file_name  = os.path.splitext( self.file_name )[0] + '.shp'
        self._fh   = open( file_name, 'rb' )
        try:
            self.read_header()
            while True:
                self.read_record()
        finally:
            self._fh.close()
            if len( self._data ) > 0:
                keys = self._data.keys()
                keys.sort()
                self._default_source = keys[0]
                self.add_trait( 'source_names', Trait( keys ) )
           
    #---------------------------------------------------------------------------
    #  Read the header record from a shape file: 
    #---------------------------------------------------------------------------
      
    def read_header ( self ):
        header = ( unpack( '>i 20x i',    self._fh.read( 28 ) ) +
                   unpack( '<i i 4d 32x', self._fh.read( 72 ) ) )
        if header[0] != 9994:
            print 'Not a shape file'
            raise ValueError
        self._bounds = header[4:8]
        self._type   = [ 'Null', 'Point', 'Unknown', 'Arc', 'Unknown',
                         'Polygon', 'Unknown', 'Unknown', 'MultiPoint' 
                       ][ header[3] ]
        self._plot_type = [ 'Line', 'Scatter', 'Line', 'Line', 'Line',
                         'Polygon', 'Line', 'Line', 'Scatter' 
                       ][ header[3] ]
        
    #---------------------------------------------------------------------------
    #  Read a shape file record: 
    #---------------------------------------------------------------------------
    
    def read_record ( self ):
        rn, size = unpack( '>2i', self._fh.read( 8 ) )
        type     = unpack( '<i',  self._fh.read( 4 ) )[0]
        data     = self._fh.read( 2 * size - 4 )
        
        # Point:
        if type == 1: 
            self._data[ SOURCENAME % ( self._type, rn ) ] = unpack('<2d', data) 
           
        # MultiPoint:
        elif type == 8:
            xmin, ymin, xmax, ymax, n = unpack( '<4d i', data[0:36] )
            points = []
            index  = 36
            for i in range( n ):
                points.append( unpack( '<2d', data[ index: index + 16 ] ) )
                index += 16
            self._data[ SOURCENAME % ( self._type, rn ) ] = points
           
        # Arc/Polygon:
        elif (type == 3) or (type == 5):
            xmin, ymin, xmax, ymax, parts, n = unpack( '<4d 2i', data[0:40] )
            index   = 40 + 4 * parts
            offsets = unpack( '<%di' % parts, data[40:index] ) + ( n, )
            points  = []
            for i in range( n ):
                points.append( unpack( '<2d', data[ index: index + 16 ] ) )
                index += 16
            for i in range( parts ):
                key = '%s #%d-%d' % ( self._type, rn, (i + 1) )
                self._data[ key ] = points[ offsets[i]: offsets[i+1] ]
        else:
            print 'Unrecognized record type %d' % type
            raise ValueError
