#-------------------------------------------------------------------------------
#
#  Calculate the region between two curves
#
#  Written by: Eric Jones
#
#  Date: 07/22/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.util.scipyx import *

#-------------------------------------------------------------------------------
#  Calculate the region between two curves:
#-------------------------------------------------------------------------------

def fill_areas ( l1, l2 ):
    """ Returns a list of polygons that define the area between point arrays 
        l1 and l2
    
        This function returns a list of polygons that defines the regions where
        the y values in l1 are greater than or equal to the y values in l2.
        l1 and l2 are Nx2 arrays with a point pair (x,y) on each row.  The
        x values of both arrays should be monotonically increasing.  l1 
        and l2 do not need to have the same number of points.
        
        Only overlapping regions of l1 and l2 are filled.
                
        The returned polygons may have duplicate points in them.  Also,
        the first and last points in the polygon may or may not be the
        same.  Like l1 and l2, each polygon in the list is a Mx2 array
        of points.
    """
    # should interpolate l1 and l2 onto the same grid.
    # Results for non-overlapping lines are not well defined.
    # It only works for monotonically increasing l1 and l2.

    if alltrue(l1[:,0]==l2[:,0]):
        pass        
    else:
        # find overlap
        l1_xmin,l1_ymin = l1[0]
        l1_xmax,l1_ymax = l1[-1]
        l2_xmin,l2_ymin = l2[0]
        l2_xmax,l2_ymax = l2[-1]

        # find overlap of lines.
        lo_x = max(l1_xmin,l2_xmin)
        hi_x = min(l1_xmax,l2_xmax)    
        
        # could do soemthing similar with searchsorted faster.
        l1x = clip(l1[:,0],lo_x,hi_x)
        l2x = clip(l2[:,0],lo_x,hi_x)
        all_x = concatenate((l1x,l2x),axis=-1)
        all_x = sort(unique(all_x))
        
        # interpolate l1 onto all_x
        interp = linear1d(l1[:,0],l1[:,1],copy=0,fill_value=0)
        l1y = interp(all_x)
        l1 = concatenate([all_x[:,NewAxis],l1y[:,NewAxis]],axis=-1) 
        
        # interpolate l2 onto all_x
        interp = linear1d(l2[:,0],l2[:,1],copy=0,fill_value=0)        
        l2y = interp(all_x)
        l2 = concatenate([all_x[:,NewAxis],l2y[:,NewAxis]],axis=-1) 
    x  = l1[:,0]
    y1 = l1[:,1]
    y2 = l2[:,1]
        
    # areas where y_diff is positive will be shaded.
    y_diff = y1 - y2    
    
    # find all indices of points in the lines that lie in the shaded region.
    ind = arange(len(y_diff))
    shade_indices = compress(y_diff >=0, ind)

    if not len(shade_indices):
        polys = []
    else:
        # We really only want the "boundary" indices (first and last) of the 
        # shaded region.  That is, when we see the indices 1,2,3,7,8,and 9, 
        # we're really only care to know that 1-3 are points within a shaded
        # region and 7-9 are points within another shaded region. One way
        # to find the dividing point between shaded regions is to take the
        # derivative of the shaded_indices array.  Places where the slope=1
        # are within the same region.  Places where the slope>1 indicate
        # divisions between two regions.
        deriv_shade_indices = shade_indices[1:] - shade_indices[:-1]
        ind = arange(len(deriv_shade_indices))
        
        # This gives us the "interior" dividing points -- i.e. the index
        # into shade_indices that indicates the end of each region (except
        # for the last region).
        interior_divisions = compress(deriv_shade_indices > 1,ind)    
        
        # We also want the beginning of each region, so we'll add 1 to this
        # array to get the start of the next region.
        region_boundaries = zeros((len(interior_divisions)+1,2),Int32)
        region_boundaries[:-1,1] = take(shade_indices,interior_divisions)
        region_boundaries[1:,0] = take(shade_indices,interior_divisions+1)
        # The beginning index of the first region is first index
        region_boundaries[0,0] = shade_indices[0]
        # The end index of the last region is always the last index.
        region_boundaries[-1,1] = shade_indices[-1]
        #print 'rb:', region_boundaries
        
        polys = []
        for first,last in region_boundaries:
            pts = []
            if first == 0:
                # region starts at beginning of array
                pass
            elif y_diff[first] == 0:
                # shading starts at first data point in the region        
                pass
            else:
                xx = x[first-1:first+1]
                yy1 = y1[first-1:first+1]
                yy2 = y2[first-1:first+1]
                pt = intercept(xx,yy1,yy2)
                pts.append(pt[NewAxis,:]) 
            pts.append(l1[first:last+1])
    
            if last == len(x)-1:
                # region ends at end of array
                pass
            elif y_diff[last] == 0:
                # shading ends at last data point in the region        
                pass
            else:
                xx = x[last:last+2]
                yy1 = y1[last:last+2]
                yy2 = y2[last:last+2]
                pt = intercept(xx,yy1,yy2)
                pts.append(pt[NewAxis,:]) 
            pts.append(l2[first:last+1][::-1]) # reverse order of the points
            pts = concatenate(pts,axis=0)
            polys.append(pts)
    return polys

#----------------------------------------------------------------------------
#  Utilities:
#----------------------------------------------------------------------------

def data_bounds ( l ):
    axis = 0
    mins = minimum.reduce( l, axis )
    maxs = maximum.reduce( l, axis )
    return concatenate( ( mins, maxs ) )
    
def intercept ( x, y1, y2 ):    
    u = (y1[0]-y2[0]) / ((y2[1]-y2[0]) - (y1[1]-y1[0]))
    x = x[0] + u * (x[1]-x[0])
    y = y1[0] + u * (y1[1]-y1[0])
    return array( ( x, y ) )

#-------------------------------------------------------------------------------
#  From Tim Peters and the Python Cookbook:
#-------------------------------------------------------------------------------

def unique ( s ):
    """Returns a list of the elements in s, but without duplicates.

    For example, unique([1,2,3,1,2,3]) is some permutation of [1,2,3],
    unique("abcabc") some permutation of ["a", "b", "c"], and
    unique(([1, 2], [2, 3], [1, 2])) some permutation of
    [[2, 3], [1, 2]].

    For best speed, all sequence elements should be hashable.  Then
    unique() usually works in linear time.

    If it is not possible for the sequence elements element to be hashable,
    then they should have a total ordering. If list(s).sort() does not raise 
    TypeError, then it is assumed that the elements do have a total ordering.  
    In that case, unique() will usually work in O(N*log2(N)) time.

    If a total ordering is not possible either, the sequence elements must 
    support equality-testing.  In that case, unique() usually works in 
    quadratic time.
    """

    n = len(s)
    if n == 0:
        return []

    # Try using a dict first, as that's the fastest and will usually
    # work.  If it doesn't work, it will usually fail quickly, so it
    # usually doesn't cost much to *try* it.  It requires that all the
    # sequence elements be hashable, and support equality comparison.
    u = {}
    try:
        for x in s:
            u[x] = 1
    except TypeError:
        del u  # move on to the next method
    else:
        return u.keys()

    # We can't hash all the elements.  Second fastest is to sort,
    # which brings the equal elements together; then duplicates are
    # easy to weed out in a single pass.
    # NOTE:  Python's list.sort() was designed to be efficient in the
    # presence of many duplicate elements.  This isn't true of all
    # sort functions in all languages or libraries, so this approach
    # is more effective in Python than it may be elsewhere.
    try:
        t = list(s)
        t.sort()
    except TypeError:
        del t  # move on to the next method
    else:
        assert n > 0
        last = t[0]
        lasti = i = 1
        while i < n:
            if t[i] != last:
                t[lasti] = last = t[i]
                lasti += 1
            i += 1
        return t[:lasti]

    # Brute force is all that's left.
    u = []
    for x in s:
        if x not in u:
            u.append(x)
    return u

# The following are kludges to fix brain-deadness of 'take' and
# 'sometrue' when dealing with 0 dimensional arrays:

import enthought.util.scipyx as scipy_base

def take ( a, indices, axis = 0 ):    
    x = asarray( a ); 
    y = asarray( indices )
    if shape( x ) == (): 
       x = ravel(x)
    if shape( y ) == (): 
       y = ravel(y)
    return scipy_base.take( x, y, axis )

def sometrue ( a, axis = 0 ):    
    x = asarray( a )
    if shape( x ) == (): 
       x = ravel(x)
    return scipy_base.sometrue( x )

def reduce_sometrue ( a ):
    all = a
    while len( shape( all ) ) > 1:    
       all = sometrue( all )
    return all

#-------------------------------------------------------------------------------
#  Cut and paste from scipy.interpolate and scipy_base.shape_base:
#-------------------------------------------------------------------------------

def atleast_1d ( *arys ):
    """ Forces a sequence of arrays to each be at least 1-D.

        Description
        -----------
        Forces an array to be at least 1-D.  If an array is 0-D, the 
        array is converted to a single row of values.  Otherwise,
        the array is unaltered.
            
        Parameters
        ----------
        arys
            arrays to be converted to 1 or more dimensional array.

        Returns
        -------
        Input array converted to at least 1-D array.
    """
    res = []
    for ary in arys:
        ary = asarray(ary)
        if len(ary.shape) == 0: 
            result = scipy_base.array([ary[0]])
        else:
            result = ary
        res.append(result)
    if len(res) == 1:
        return res[0]
    else:
        return res

class linear1d:
    interp_axis = -1 # used to set which is default interpolation
                     # axis.  DO NOT CHANGE OR CODE WILL BREAK.
                     
    def __init__ ( self, x, y, axis = -1, copy = 1, bounds_error = 1, 
                         fill_value = None ):
        """Initializes a 1-D linear interpolation class.

        Description
        -----------
        x and y are arrays of values used to approximate some function f::
        
            y = f(x)

        This class returns a function whose call method uses linear
        interpolation to find the value of new points.

        Parameters
        ----------
        x
            a 1-D array of monotonically increasing real values.
            x cannot include duplicate values. (otherwise f is
            overspecified)
        y
            an n-D array of real values.  y's length along the
            interpolation axis must be equal to the length
            of x.
        axis
            specifies the axis of y along which to 
            interpolate. Interpolation defaults to the last
            axis of y.  (default: -1)
        copy
            If 1, the class makes internal copies of x and y.
            If 0, references to x and y are used. The default 
            is to copy. (default: 1)
        bounds_error
            If 1, an error is thrown any time interpolation
            is attempted on a value outside of the range
            of x (where extrapolation is necessary).
            If 0, out of bounds values are assigned the
            NaN (#INF) value.  By default, an error is
            raised, although this is prone to change.
            (default: 1)
        """      
        self.axis = axis
        self.copy = copy
        self.bounds_error = bounds_error
        if fill_value is None:
            self.fill_value = array(0.0) / array(0.0)
        else:
            self.fill_value = fill_value
            
        # Check that both x and y are at least 1 dimensional.    
        if len(shape(x)) == 0 or len(shape(y)) == 0:
            raise ValueError, "x and y arrays must have at least one dimension."                
        # make a "view" of the y array that is rotated to the
        # interpolation axis.  
        oriented_x = x
        oriented_y = swapaxes(y,self.interp_axis,axis)            
        interp_axis = self.interp_axis        
        len_x,len_y = shape(oriented_x)[interp_axis], shape(oriented_y)[interp_axis]
        if len_x != len_y:
            raise ValueError, "x and y arrays must be equal in length along "\
                              "interpolation axis."
        if len_x < 2 or len_y < 2:
            raise ValueError, "x and y arrays must have more than 1 entry"            
        self.x = array(oriented_x,copy=self.copy)
        self.y = array(oriented_y,copy=self.copy)       
        
    def __call__ ( self, x_new ):
        """ Finds linearly interpolated y_new = <name>(x_new).

			Parameters
			----------
			x_new
				New independent variables.
	
			Returns
			-------
			y_new
				Linearly interpolated values corresponding to x_new.
        """
        # 1. Handle values in x_new that are outside of x.  Throw error,
        #    or return a list of mask array indicating the outofbounds values.
        #    The behavior is set by the bounds_error variable.
        x_new = atleast_1d(x_new)
        out_of_bounds = self._check_bounds(x_new)
        # 2. Find where in the orignal data, the values to interpolate
        #    would be inserted.  
        #    Note: If x_new[n] = x[m], then m is returned by searchsorted.
        x_new_indices = searchsorted(self.x,x_new)
        # 3. Clip x_new_indices so that they are within the range of 
        #    self.x indices and at least 1.  Removes mis-interpolation
        #    of x_new[n] = x[0] 
        x_new_indices = clip(x_new_indices,1,len(self.x)-1).astype(Int)
        # 4. Calculate the slope of regions that each x_new value falls in.
        lo = x_new_indices - 1; hi = x_new_indices        
        
        # !! take() should default to the last axis (IMHO) and remove
        # !! the extra argument.
        x_lo = take(self.x,lo,axis=self.interp_axis)
        x_hi = take(self.x,hi,axis=self.interp_axis);
        y_lo = take(self.y,lo,axis=self.interp_axis)
        y_hi = take(self.y,hi,axis=self.interp_axis);
        slope = (y_hi-y_lo)/(x_hi-x_lo)
        # 5. Calculate the actual value for each entry in x_new.
        y_new = slope*(x_new-x_lo) + y_lo 
        # 6. Fill any values that were out of bounds with NaN
        # !! Need to think about how to do this efficiently for 
        # !! mutli-dimensional Cases.
        yshape = y_new.shape
        y_new = ravel(y_new)
        new_shape = list(yshape)
        new_shape[self.interp_axis] = 1
        sec_shape = [1]*len(new_shape)
        sec_shape[self.interp_axis] = len(out_of_bounds)
        out_of_bounds.shape = sec_shape
        new_out = ones(new_shape)*out_of_bounds
        putmask(y_new, ravel(new_out), self.fill_value)
        y_new.shape = yshape      
        # Rotate the values of y_new back so that they coorespond to the
        # correct x_new values.
        result = swapaxes(y_new,self.interp_axis,self.axis)
        try:
            len(x_new)
            return result
        except TypeError:
            return result[0]
        return result
    
    def _check_bounds ( self, x_new ):
        # If self.bounds_error = 1, we raise an error if any x_new values
        # fall outside the range of x.  Otherwise, we return an array indicating
        # which values are outside the boundary region.  
        # !! Needs some work for multi-dimensional x !!
        below_bounds = less(x_new,self.x[0])
        above_bounds = greater(x_new,self.x[-1])
        #  Note: sometrue has been redefined to handle length 0 arrays
        # !! Could provide more information about which values are out of bounds
        if self.bounds_error and sometrue(below_bounds):
            raise ValueError, " A value in x_new is below the"\
                              " interpolation range."
        if self.bounds_error and sometrue(above_bounds):
            raise ValueError, " A value in x_new is above the"\
                              " interpolation range."
        # !! Should we emit a warning if some values are out of bounds.
        # !! matlab does not.
        out_of_bounds = logical_or(below_bounds,above_bounds)
        return out_of_bounds
       
    def model_error(self,x_new,y_new):
        # How well do x_new,yy points fit the model?
        # Return an array of error values.
        pass

if __name__ == '__main__':    
    from enthought.util.testingx import assert_equal
    # two interior regions
    x  = array([0.00,1.00,2.00,3.00,4.00,5.00,6.00,7.00,8.00,9.00,10.00,11.00])
    y1 = array([0.00,1.00,2.00,1.33,0.66,0.00,0.00,1.50,2.00,1.33, 0.66, 0.00])
    y2 = array([1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00, 1.00, 1.00])
    
    l1 = concatenate([x[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    p1 = fill_areas(l1,l2)
    print p1
    
    # two coincident lines
    # -- should result in the 
    x  = array([0.00,1.00,2.00,3.00,4.00,5.00,6.00,7.00,8.00,9.00,10.00,11.00])
    y1 = array([0.00,1.00,2.00,1.33,0.66,0.00,0.00,1.00,2.00,1.33, 0.66, 0.00])
    y2 = array([0.00,1.00,2.00,1.33,0.66,0.00,0.00,1.00,2.00,1.33, 0.66, 0.00])
    
    l1 = concatenate([x[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    p2 = fill_areas(l1,l2)
    desired = concatenate((l1,l2[::-1]),axis=0)
    assert_equal(p2[0], desired)
    print p2

    # fill at end and beginning of areas
    x  = array([0.00,1.00,2.00,3.00,4.00,5.00,6.00,7.00,8.00,9.00,10.00,11.00])
    y1 = array([1.00,1.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00, 1.00, 1.00])
    y2 = array([0.00,0.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00, 0.00, 0.00])
    
    l1 = concatenate([x[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    p3 = fill_areas(l1,l2)
    print p3
    
    # fill entire area
    x  = array([0.00,1.00,2.00,3.00,4.00,5.00,6.00,7.00,8.00,9.00,10.00,11.00])
    y1 = array([1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00, 1.00, 1.00])
    y2 = array([0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00, 0.00, 0.00])
    
    l1 = concatenate([x[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    p3 = fill_areas(l1,l2)
    print p3
    
    p3r = fill_areas(l2,l1)
    print p3r
    # unequal lengths
    x1 = array([0.00,4.50,11.00])    
    y1 = array([0.00,2.00, 0.00])
    x2 = array([0.00,1.00,2.00,3.00,4.00,5.00,6.00,7.00,8.00,9.00,10.00,11.00])
    y2 = array([1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00, 1.00, 1.00])
    
    l1 = concatenate([x1[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x2[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    p4 = fill_areas(l1,l2)
    print p4
        
    # partial overlap
    x1  = array([0.00,1.00,2.00,3.00,4.00,5.00,6.00,7.00,8.00,9.00,10.00,11.00])
    y1  = array([1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00,1.00, 1.00, 1.00])
    x2  = x1 - 5
    y2  = array([0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00,0.00, 0.00, 0.00])
    
    l1 = concatenate([x1[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x2[:,NewAxis],y2[:,NewAxis]],axis=-1)
    print 'l2:', l2
    p5 = fill_areas(l1,l2)
    print 'overlap:', p5
    
    
    # speed test using random data.
    # This is pretty much worst case.
    from enthought.util.randomx import random
    N = 2000
    x = arange(N)
    y1 = random(N)
    y2 = random(N)
    
    l1 = concatenate([x[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    import time
    t1 = time.time()
    for x in range(10):
        p5 = fill_areas(l1,l2)
    t2 = time.time()
    print 'time -- random data (N=%d, polys=%d): %f' % (N,len(p5),t2-t1)  
    
    
    # speed test using sin/cos.
    # This is pretty much worst case.
    N = 10000
    x = arange(N)*30*pi/N
    y1 = sin(x)
    y2 = cos(x)
    
    l1 = concatenate([x[:,NewAxis],y1[:,NewAxis]],axis=-1)
    l2 = concatenate([x[:,NewAxis],y2[:,NewAxis]],axis=-1)
    
    import time
    t1 = time.time()
    for x in range(10):
        p6 = fill_areas(l1,l2)
    t2 = time.time()
    print 'time -- sin/cos (N=%d, polys=%d): %f' % (N,len(p6),t2-t1)  

    t1 = time.time()
    for x in range(10):
        p6 = fill_areas(l1,l2)
    t2 = time.time()
    print 'time -- sin/cos,same x (N=%d, polys=%d): %f' % (N,len(p6),t2-t1)  
    
    t1 = time.time()
    for x in range(10):
        assert alltrue(l1[:,0]==l2[:,0])
    t2 = time.time()
    print 'time -- alltrue (N=%d): %f' % (N,t2-t1)  
    
