"""
Defines PlotValue classes for 2D image plots.  Classes defined:
    ImageData
    AbtractImagePlotValue
    CmapImagePlotValue
    ImagePlotValue
"""

# Major package imports.
import weakref
import pdb
from math import log10

from math import floor, ceil

# Enthought library imports.
from enthought.util.numerix import arange, array, concatenate, NewAxis, swapaxes, \
                                   take, transpose, UInt8
from enthought.traits.api import Array, Delegate, Enum, Event, false, Float, \
                             HasTraits, Instance, List, Property, Range, String, \
                             Trait, TraitError, true, Tuple
from enthought.traits.ui.api import View

# Local imports.
from plot_base import plot_delegate, sequence_types, bin_search1, \
                      ArrayType, PlotItem, TOP, BOTTOM, LEFT, RIGHT
from plot_container import DefaultContainer
from plot_canvas import PlotCanvas
from plot_indexed import PlotIndexed
from plot_overlay import PlotOverlay
from plot_source import PlotSource
from plot_style import PlotStyle
from plot_traits import PlotType
from plot_views import bar_width_group, line_group, marker_group, plot_group, \
                       border_group, plot_margin_group, cursor_group, misc_group
from plot_value import PlotValue

from colormap import AbstractColormap, LinearColormap

from enthought.kiva.agg import GraphicsContextArray, Image, pix_format_string_map

class ImageData(HasTraits):
    """
    Represents a grid of data to be plotted using a Numeric 2D grid and
    two 1D arrays.  The 2D grid holds the values, and the two 1D grids
    represent index arrays along X and Y.  The grid maps from integer
    (i,j) indices into a value in data space, and the index arrays each map
    an index to an X or a Y index value in data space.

    The index for ImageData is a tuple of two arrays (Nx1, Mx1), and the
    value is represented by an array NxM.  The first array in the index tuple
    corresponds to the x-values in data space and the second array corresponds
    to the y-values.  If an index is not provided, then one is automatically
    generated from the integer grid indices into the value grid.

    The value array has dimensions NxM, but it may have more than just 2
    dimensions.  The appropriate dimensionality of the value array depends
    on the PlotValue in which the ImageData instance will be used.  A
    CmapImagePlotValue will expect a single scalar value (i.e. a simple
    2D array), whereas an ImagePlotValue expects a length-3 or length-4
    array (RGB or RGBA image).
    """

    # 'value' holds the grid data that forms the image.  It is always an NxM
    # array of floats, where N is the length of x_index and M is the length of
    # y_index.
    # Note that if ImageData was constructed with a transposed data array,
    # then internally it is still transposed (i.e. y-axis is the first axis
    # and x-axis is the second), and the 'value' array property may not be
    # contiguous.  If contiguousness is required and calling .copy() is too
    # expensive, use the raw_value attribute.
    value = Property(Array)

    # 'transposed' indicates that 'raw_value', the actual underlying image data
    # array, is transposed from 'value'. (i.e. the first axis corresponds to
    # the y direction and the second axis corresponds to the x direction.) This
    # is commonly the case with Numeric arrays returned from PIL images or from
    # GraphicsContexts. Rather than doing swapaxis/transpose on the data and
    # destroying continuity, we expose the data as both 'value' and 'raw_value'.
    transposed = false

    # 'index' is a 2-tuple of 1D arrays representing coordinates in data space
    # of the grid indices in 'value'.  The first array is the horizontal index
    # and the second array is the vertical index.  This is true even if the
    # array is transposed, i.e. len(index[0]) == len(value) and
    # len(index[1]) == len(value[0]).
    index = Property(Tuple(Array, Array))

    # x_index and y_index are just aliases for the individual arrays in 'index'.
    x_index = Property(Array)
    y_index = Property(Array)

    # A way for consumers of this class to get notified about the data changing
    data_changed = Event(bool)
    index_changed = Event(bool)

    # The actual image data array.  May be MxN or NxM, depending on the value
    # of 'transposed'.  Do not access this directly unless you take this into
    # account!
    raw_value = Array

    # shadow traits for the Property traits above
    _index = Tuple(Array, Array)


    def __init__(self, value_grid, index_arrays = (), transposed=False):
        """
        Creates a new ImageData object, optionally specifying two arrays
        with specific values to be used for the indexes, and optionally
        specifying that the X and Y axes are reversed.
        """
        self.transposed = transposed
        if transposed:
            if index_arrays:
                self._index = index_arrays[::-1]
        else:
            if index_arrays:
                self._index = index_arrays
        self.value = value_grid   # let the property setter do the work
        return

    def _get_value(self):
        if self.transposed:
            return swapaxes(self.raw_value, 0, 1)
        else:
            return self.raw_value

    def _set_value(self, new_value):
        # if the dimensions of the index arrays agree with the dimensions
        # of new_value, then don't reset the index; otherwise, automatically
        # generate a new set of integer index arrays
        self.raw_value = new_value
        if self.transposed:
            real_index_order = tuple(map(len, self._index))[::-1]
        else:
            real_index_order = tuple(map(len, self._index))

        if new_value.shape != real_index_order:
            self._index = tuple([array(range(i+1)) for i in new_value.shape[0:2]])
            if self.transposed:
                self._index = self._index[::-1]
            self.index_changed = True
        self.data_changed = True
        return

    def _get_index(self):
        return self._index

    def _set_index(self, new_index):
        self._index = new_index
        self.index_changed = True
        return

    def _get_x_index(self):
        return self._index[0]

    def _get_y_index(self):
        return self._index[1]


class AbstractImagePlotValue(PlotCanvas):
    """
    Defines the interface to which all image plot values must adhere.  For
    more information on how PlotValues interact with other Chaco classes,
    see plot_value.py.

    Also takes care of a lot of the cruft, like mapping to and from screen
    coordinates.  Child classes of this class will mostly differ in how
    they render an image data set onto a GC, but they all will have an
    underlying image dataset.

    Child classes *must* implement the _render() method!!
    """
    ### Traits specific to ImagePlotValues ###################################
    data = Instance(ImageData)

    ### Style traits. ########################################################

    # For bar charts, the width of a bar.
    bar_width      = plot_delegate

    # For bar charts, the width of a group of bars.
    bin_width      = plot_delegate

    # For line type plots, the style of the line (e.g. dash).
    line_style     = plot_delegate

    # For line type plots, the color of the line.
    line_color     = plot_delegate

    # For line type plots, the thickness of the line.
    line_weight    = plot_delegate

    # The corner of the plotting area used as the axis origin.
    origin         = plot_delegate

    # For plots with markers, the outline color of the marker.
    outline_color  = plot_delegate

    # For plots with markers, the thickness of the outline.
    outline_weight = plot_delegate

    # For plots with markers, the fill color of the marker.
    fill_color     = plot_delegate

    # For plots with markers, the fill style of the marker.
    fill_style     = plot_delegate

    # For plots with markers, the size of the marker.
    size           = plot_delegate

    # For plots with markers, the symbol (e.g. circle).
    symbol         = plot_delegate

    # The type of plot drawn.
    type           = plot_delegate

    ### Trait editors ########################################################
    traits_view = View(
        [ [ 'index_axis{Index Axis}', '_',
            'origin',                 '_',
            'type',
            '|[Plot]@' ],
          [ 'index_scale{Index Axis}',
            'value_scale{Value Axis}',
            '|[Scale Factor]' ],
          [ 'position', '_',
            'bg_color', '_',
            'visible',
            '|[Canvas]@' ],
        '|{Main}' ],
        [ [ 'weight',
            '|[Layout]' ],
          [ 'min_height{Height}',
            'min_width{Width}',
            '|[Canvas Minimum]' ],
          [ 'plot_min_height{Height}',
            'plot_min_width{Width}',
            '|[Plot Area Minimum]' ],
          bar_width_group,
          '|{Layout}' ],
        line_group,
        marker_group,
        plot_group,
        border_group,
        plot_margin_group,
        cursor_group,
        misc_group
    )

    canvas_traits_view = View(
        [ [ 'type', '_', 'visible',
            '|[Plot]@' ],
          '|{Main}' ],
        line_group,
        marker_group,
        [ [ 'menu', '|[Miscellaneous]' ], '|{Misc}' ]
    )

    indexed_traits_view = View(
        [ [ 'type', '_', 'value_scale{Value Axis Scale}', '_', 'visible',
            '|[Plot]@' ],
          '|{Main}' ],
        [ [ 'weight', '|[Layout]' ],
          [ 'plot_min_height{Height}',
            'plot_min_width{Width}',
            '|[Plot Area Minimum]' ],
          '|{Layout}' ],
        line_group,
        marker_group,
        plot_group,
        plot_margin_group,
        cursor_group,
        misc_group
    )

    #-------------------------------------------------------------------------
    # Abstract methods
    #-------------------------------------------------------------------------

    def _render(self, gc, ox, oy):
        """
        Renders the plot value onto gc with the origins (ox, oy).
        Note that dimensions are not given.  These are stored in the attribute
        self._xy, which is a 2-tuple of length-2 arrays (x,y).
        """
        raise NotImplementedError

    #-------------------------------------------------------------------------
    # Concrete methods
    #-------------------------------------------------------------------------

    def __init__(self, data=None, **traits):
        from plot_lists import PlotOverlayList, PlotTitleList, PlotValueList

        self._min          = 0.0
        self._max          = 1.0
        self._index_       = 1   # Force a 'cache miss' if '_index' called
        self._cached_index = None
        self._titles       = []
        self._overlays     = []
        self._values       = [ PlotItem( weakref.proxy( self ) ) ]
        self.titles        = PlotTitleList(   self, self._titles )
        self.overlays      = PlotOverlayList( self, self._overlays )
        self.values        = PlotValueList(   self, self._values )
        self.type          = "image"
        if data is not None:
            self.data = data
        from plottable import Plottable
        Plottable.__init__( self, **traits )

        # Indices into self._data_image that defines the zoomed-in region.
        # Expressed as the i,j of the lower left and upper right corners:
        # (i_LL, j_LL, i_UR, j_UR).
        self._data_image_bounds = []

    def default_traits_view ( self ):
        "Return the default traits view/name."
        not_overlay = not self.overlay
        if not_overlay and isinstance( self.container, PlotCanvas ):
            view = 'canvas_traits_view'
        elif not_overlay and isinstance( self.container, PlotIndexed ):
            view = 'indexed_traits_view'
        else:
            view = 'traits_view'
        return view

    def values_at ( self, x, y ):
        """
        Returns the list of PlotValue objects whose representation intersects
        a specified (x,y) point.  For now, hardcode this to return just [self].
        """
        return [self]

    def map_xy ( self, x, y ):
        """
        Maps the specified x and y coordinates from window space into data
        space.  This differs from PlotValue.map_xy() in that the "y" argument
        is not optional.
        """
        return self._plot_renderers[0]._map_xy( self, x, y )

    def map_data ( self, low_xy, high_xy = None):
        """
        map_data( low_xy, high_xy = None) -> [(i_low, j_low), (i_high, j_high)]

        Maps an (x,y) coordinate in data space to the closest range of
        integer indices into the index arrays.  If a "high" value is given, then
        the index range returned represents the bounding coordinates in integer
        index space of the rectangular region in data space between low_xy and
        high_xy.
        """

        # Use Numeric's searchsorted() to do the actual searching
        low_bin = self._map_data(low_xy)
        if high_xy != None:
            high_bin = self._map_data(high_xy)
        else:
            high_bin = low_bin + 1

        return [tuple(low_bin), tuple(high_bin)]

    def _map_data(self, datapoint):
        """
        _map_data( (x,y) ) -> (i,j) integer indices

        Returns a tuple of integer indices into the PlotValue's X and Y index
        axes corresponding to an input (x,y) point in dataspace.
        """
        pass

    def unselect ( self ):
        "Deselect any current range selection"
        self.host._select_none( self )

    def _ui_label ( self ):
        "Return the user interface label description of this plotvalue."
        return (self.label or "2D Image Plotvalue")

    def _data_changed ( self, old = None, new = None ):
        self._set_range()

    #---------------------------------------------------------------------------
    #  Set the array min/max values:
    #---------------------------------------------------------------------------

    def _set_range ( self ):
        # pzw
        # This function sets self._min and self._max to correspond to the range
        # of Values (represented by self.data).  For this to work for the rest
        # of the system, we return index.y's min and max.
        if self.data:
            if self.index_axis == 'x':
                self._min = min(self.data.y_index)
                self._max = max(self.data.y_index)
            else:
                self._min = min(self.data.x_index)
                self._max = max(self.data.x_index)
        else:
            self._min = 0.0
            self._max = 1.0
        self._dir = None

    def _can_add ( self, item ):
        "Returns whether or not a new object can be added to this plotvalue."
        return False

    def _all_values ( self ):
        "Returns a list of all contained plotvalues."
        return self._values

    def _index ( self ):
        """
        _index() -> PlotValue holding the index axis data
        """
        if self.index_axis == 'x':
            return PlotValue(self.data.x_index, axis_index=self.axis_index)
        else:
            return PlotValue(self.data.y_index, axis_index=self.axis_index)

    def _min_size_plot ( self ):
        "Calculate the minimum size of the plotting area needed"
        return ( self.min_width, self.min_height )

    #---------------------------------------------------------------------------
    #  Map from data space to screen space:
    #
    #  Note: 'index' and 'value' can be either scalar, arrays, or None.
    #  If they are None, the corresponding part of the result is also None.
    #  The 'index_offset' and 'value_offset' arguments allow an offset to be
    #  applied to the result 'after' conversion to screen space.
    #---------------------------------------------------------------------------

    def map_screen ( self, index, value, index_offset = 0.0, value_offset = 0.0,
                           xy = False ):

        # Get the x and y data values and axis information:
        if self.index_axis == 'x':
            x_axis = self._axis_info_index
            y_axis = self._axis_info
        else:
            x_axis = self._axis_info
            y_axis = self._axis_info_index

        # Do the axis 'log' scaling if necessary:
        if y_axis.log_scale and (value is not None):
            value = log10( value )

        # If the index axis is vertical, flip-flop the values:
        if not x_axis.horizontal:
            index, value               = value, index
            index_offset, value_offset = value_offset, index_offset
            x_axis, y_axis             = y_axis, x_axis

        xp = None
        if index is not None:
            # Scale and translate the x data:
            dxmin = x_axis.tick_min
            ddx   = x_axis.tick_max - dxmin

            # We shouldn't have to worry about this here (?):
            if ddx == 0.0:
                ddx    = 1.0
                dxmin -= 0.5

            # Do the actual scaling/translation of the data:
            if x_axis.origin & LEFT:
                xp = (self._plot_bounds[0] - self._plot_ox + ((index - dxmin) *
                           (float( self._plot_sxdx - 1 ) / ddx))) + index_offset
            else:
                xp = (((self._plot_bounds[0]-self._plot_ox+self._plot_sxdx-1) -
                     ((index - dxmin) * (float( self._plot_sxdx - 1 ) / ddx))) -
                     index_offset)

        # If the value is None, then we are done, return the result:
        if value is None:
            return ( xp, None )

        # Scale and translate the y data:
        dymin = y_axis.tick_min
        ddy   = y_axis.tick_max - dymin

        # We shouldn't have to worry about this here (?):
        if ddy == 0.0:
            ddy    = 1.0
            dymin -= 0.5

        # Do the actual scaling/translation of the data:
        if y_axis.origin & TOP:
            yp = (((self._plot_bounds[1]-self._plot_oy + self._plot_sydy - 1) -
                  ((value - dymin) * (float( self._plot_sydy - 1 ) / ddy))) -
                  value_offset)
        else:
            yp = (self._plot_bounds[1] - self._plot_oy + ((value - dymin) *
                           (float( self._plot_sydy - 1 ) / ddy))) + value_offset

        # Return the result:
        if xy:
            return concatenate( ( xp[:, NewAxis ], yp[:, NewAxis ] ), axis = -1)
        return ( xp, yp )

    def _draw ( self, gc ):
        PlotCanvas._draw( self, gc, False )

   #---------------------------------------------------------------------------
    #  Return the direction the data values are sorted in (if any):
    #   1 = Monotonically increasing,
    #  -1 = Monotonically decreasing
    #   0 = Not sorted
    #---------------------------------------------------------------------------

    def _get_dir ( self ):
        """
        Returns the direction the data values are sorted.  Since this is
        meaningless for Image plots, we return 1.  (This function is only used
        by map_data() in the linear PlotValue, so this should be safe.)
        """
        return 1

    def _calc_zoom_coords(self, px, py, plot_width, plot_height,
                                ix, iy, image_width, image_height):
        """
        _calc_zoom_coords() -> ((sub-image i1, j1, i2, j2),
                                (dest_gc x, y, dx, dy))
        Given the plot pixel bounds (x,y,dx,dy) and the image virtual pixel
        bounds, returns the pixels of the sub-image that needs to be extracted
        (as lower left indices and upper right indices), and the origin and
        extents that sub-image should be drawn into.

        Returns (None, None) if no zoom image extraction is necessary.

        Because of floating point limitations, it is not advisable to request a
        ludicrous level of zoom, i.e. idx or idy > 10^10.
        """
        if (image_width < 1.5*plot_width) and (image_height < 1.5*plot_height):
            return (None, None)

        # We figure out the subimage coordinates using a two-step process:
        # 1. convert the plot boundaries from screen space into pixel offsets
        #    in the virtual image
        # 2. convert the coordinates in the virtual image into indices
        #    into the image data array
        # 3. from the data array indices, compute the screen coordinates of
        #    the corners of the data array sub-indices
        # in all the cases below, x1,y1 refers to the lower-left corner, and
        # x2,y2 refers to the upper-right corner.

        # 1. screen space -> pixel offsets
        x1 = px - ix
        x2 = (px + plot_width) - ix
        y1 = py - iy
        y2 = (py + plot_height) - iy

        # 2. pixel offsets -> data array indices
        # X and Y are transposed because for image plot data, the input
        xpixels = len(self.data.x_index)
        ypixels = len(self.data.y_index)
        i1 = floor(float(x1) / image_width * xpixels)
        i2 = ceil(float(x2) / image_width * xpixels)
        j1 = floor(float(y1) / image_height * ypixels)
        j2 = ceil(float(y2) / image_height * ypixels)

        # 3. array indices -> new screen space coordinates
        x1 = float(i1)/xpixels * image_width + ix
        x2 = float(i2)/xpixels * image_width + ix
        y1 = float(j1)/ypixels * image_height + iy
        y2 = float(j2)/ypixels * image_height + iy

        # Handle really, really, subpixel cases
        subimage_index = [i1, j1, i2, j2]
        subimage_coords = [x1, y1, x2-x1, y2-y1]
        plot_dimensions = (px, py, plot_width, plot_height)
        xparams = (0, 2)
        yparams = (1, 3)
        for pos_index, size_index in (xparams, yparams):
            if subimage_index[pos_index] == subimage_index[pos_index+2]-1:
                # xcoords lie inside the same pixel, so set the subimage
                # coords to be the width of the image
                subimage_coords[pos_index] = plot_dimensions[pos_index]
                subimage_coords[size_index] = plot_dimensions[size_index]
            elif subimage_index[pos_index] == subimage_index[pos_index+2]-2:
                # coords span across a pixel boundary.  Find the scaling
                # factor of the virtual (and potentially large) subimage
                # size to the image size, and scale it down.  We can do
                # this without distortion b/c we are straddling only one
                # pixel boundary.
                #
                # If we scale down the extent to twice the screen size, we can
                # be sure that no matter what the offset, we will cover the
                # entire screen, since we are only straddling one pixel boundary.
                # The formula for calculating the new origin can be worked out
                # on paper.
                extent = subimage_coords[size_index]
                pixel_extent = extent/2   # we are indexed into two pixels
                origin = subimage_coords[pos_index]
                scale = float(2 * plot_dimensions[size_index] / extent)
                subimage_coords[size_index] *= scale
                subimage_coords[pos_index] = origin + (1-scale)*pixel_extent

        subimage_index = map(int, subimage_index)

        return [subimage_index, subimage_coords]


class CmapImagePlotValue(AbstractImagePlotValue):
    """
    A plotvalue that interprets ImageData using a colormap.
    """

    colormap = Instance(AbstractColormap)

    def __init__(self, data = None, colormap = None, **traits):
        AbstractImagePlotValue.__init__(self, data, **traits)
        if colormap is not None:
            self.colormap = colormap
        self.data.on_trait_event(self._data_changed, name="data_changed")
        self._data_image = None
        return

    def _colormap_changed(self, old_map, new_map):
        if new_map is not None:
            self._compute_data_image()
        else:
            # create a 1x1 pixel for the image
            self._data_image = GraphicsContextArray((1,1))
        return

    def _data_changed(self, dummy):
        if self.colormap is not None:
            self._compute_data_image()
        return

    def _compute_data_image(self):
        # See if we need to grab a subportion of the raw data:
        if self._data_image_bounds:
            i1,j1,i2,j2 = self._data_image_bounds
            data = self.data.value[i1:i2, j1:j2]
        else:
            data = self.data.value

        # Because graphics contexts store their internal data in row-major
        # order, i.e. gc.bmp_array.shape == (height, width, depth), we have
        # to do a transpose.  Also, since the graphics context's notion of
        # y-axis is opposite that from the dataspace notion (i.e. y-axis points
        # down), we have to reverse the order of the rows as well.  We do it
        # here on the pre-colormapped data (instead of after the colormapping)
        # in order to keep the array contiguous.  (Doing tmp_img[::-1] leads
        # to an discontiguous array; it probably has something to do with
        # the astype().)
        transposed_data = transpose(data)[::-1]

        tmp_img = (self.colormap.map_array(transposed_data) * 255).astype(UInt8)

        if self.colormap.color_depth == 'rgb':
            kiva_depth = 'rgb24'
        elif self.colormap.color_depth == 'rgba':
            kiva_depth = 'rgba32'
        else:
            raise RuntimeError, "Unknown colormap depth value '%s'!" % \
                                self.colormap.color_depth

        # Draw the data into a new kiva Image.
        self._data_image = GraphicsContextArray(tmp_img, kiva_depth,
                                                       interpolation="nearest")
        return

    def _render(self, gc, ox, oy):
        # look at our plot bounds and determine if we need to extract a
        # sub-image to render.  (this only applies when we are zoomed in.)
        ix, iy = self._xy[0]
        idx, idy = self._xy[1] - self._xy[0]
        plot_origins = array(self._plot_bounds[0:2])
        plot_extents = array(self._plot_bounds[2:])
        sub_index, sub_coords = self._calc_zoom_coords(plot_origins[0], plot_origins[1],
                                                       plot_extents[0], plot_extents[1],
                                                       ix, iy, idx, idy)

        max_bounds = [0, 0] + list(self.data.value.shape[0:2])
        if sub_index is not None:
            if self._data_image_bounds != sub_index:
                self._data_image_bounds = sub_index
                self._compute_data_image()

        elif self._data_image_bounds != max_bounds:
            # the cached data image is zoomed in, but we need to zoom out:
            self._data_image_bounds = max_bounds
            self._compute_data_image()

        elif self._data_image is None:
            self._compute_data_image()

        # if we have subimage coordinates, draw into there, otherwise
        # draw the Image into the bounds we calculated in self._layout
        if sub_coords is not None:
            rect = map(int, sub_coords)
        else:
            origin = self._xy[0]
            extents = self._xy[1] - self._xy[0]
            rect = (int(origin[0]), int(origin[1]), int(extents[0]), int(extents[1]))
        gc.draw_image(self._data_image, rect)
        return


class ImagePlotValue(AbstractImagePlotValue):
    """
    Class representing a 2D plotvalue that is loaded from an image file on
    disk. An ImagePlotValue can be constructed either with a pre-loaded
    ImageData instance or a filename.  In the latter case, it will
    automatically create an ImageData object with the appropriate settings.
    """

    # the name of the file that holds the value for this
    image_filename = Property

    # the data-space bounds of the image.
    #~ x_databounds = Property(Tuple(Float, Float))
    #~ y_databounds = Property(Tuple(Float, Float))

    # overall alpha value of the image.  0=transparent, 1=full intensity.
    alpha = Trait(1.0, Range(0.0, 1.0))

    # The pixel format of the image.  This gets set automatically by the image
    # loading process, so normally it is not necessary to set this.  However,
    # it is advisable to set this trait when constructing an ImagePlotValue
    # from a manually-constructed ImageData object.
    data_pix_format = Trait("rgb24", Enum(pix_format_string_map.keys()))

    # Cache the image of the bmp data (not the bmp data in self.data.value)
    _cached_image = Instance(GraphicsContextArray)

    # The origin indices and bounds of a rectangular region in self.data
    # represented by self._cached_image.  If the entirety of self.data has been
    # rastered into _cached_image, then this is just
    # (0,0, data.shape[0], data.shape[1]).
    _cached_image_bounds = List([])

    # shadow traits
    _image_filename = String
    #~ _x_databounds = Tuple(Float, Float)
    #~ _y_databounds = Tuple(Float, Float)

    def __init__(self, x_bounds, y_bounds, image_filename="",
                 data=None, **traits):
        AbstractImagePlotValue.__init__(self, **traits)

        if data != None:
            self.data = data
            self._update_index_arrays(x_bounds, y_bounds)
            self.data.on_trait_event(self._data_changed, name="data_changed")
        else:
            self._load_image_from_file(image_filename, x_bounds, y_bounds)

        self._compute_cached_image()
        return

    def _set_image_filename(self, filename):
        assert self.data is not None, "ImagePlotValue must always have an ImageData object!"
        xbounds = (self.data.x_index[0],self.data.x_index[-1])
        ybounds = (self.data.y_index[0],self.data.y_index[-1])
        self._load_image_from_file(filename, xbounds, ybounds)
        return

    def _load_image_from_file(self, filename, x_bounds, y_bounds):
        "Loads a file from disk into this plot value"
        self._image_filename = filename

        # regardless of self.interpolation, when we load the data, we
        # want the raw pixel data, so we use nearest-neighbor.
        img = Image(filename, interpolation="nearest")

        if self.data == None:
            self.data = ImageData(img.bmp_array, transposed=True)
            self.data.on_trait_event(self._data_changed, name="data_changed")
        else:
            self.data.value = img.bmp_array

        self.data_pix_format = img.format()

        self._update_index_arrays(x_bounds, y_bounds)
        return

    def _update_index_arrays(self, x_bounds, y_bounds):
        """
        Private function for updating the index arrays in the imagedata
        object, given the dimensions of the value array in the imagedata
        object and the x_bounds and y_bounds.
        """
        width, height = self.data.value.shape[0:2]
        index_arrays = []
        for bounds, pixsize in ((x_bounds, width),(y_bounds, height)):
            if pixsize > 1:
                step = (bounds[1] - bounds[0]) / float(pixsize-1)
                index_arrays.append(arange(bounds[0], bounds[1]+step/2.0, step))
            else:
                index_arrays.append(array(bounds[0]))
        self.data.index = tuple(index_arrays)
        return

    def _data_changed(self, dummy):
        "An event handler for self.data ImageData object"
        self._compute_cached_image()

    def _compute_cached_image(self):
        max_bounds = [0, 0] +  list(self.data.value.shape[0:2])
        if self._cached_image_bounds and \
                        self._cached_image_bounds != max_bounds:
            i1,j1,i2,j2 = self._cached_image_bounds

            # to extract the subimage, we need to touch the underlying
            # raw_value array.  Not only is this array transposed so that
            # the y-values are axis 0, but they are also reversed.
            j1 = len(self.data.y_index)+1 - j1
            j2 = len(self.data.y_index)+1 - j2
            data = self.data.raw_value[j2:j1, i1:i2]
        else:
            data = self.data.raw_value
            self._cached_image_bounds = max_bounds
        self._cached_image = GraphicsContextArray(data,
                                pix_format=self.data_pix_format)
        return

    def _render(self, gc, ox, oy):
        ix, iy = self._xy[0]
        idx, idy = self._xy[1] - self._xy[0]
        plot_origins = array(self._plot_bounds[0:2])
        plot_extents = array(self._plot_bounds[2:])
        sub_index, sub_coords = self._calc_zoom_coords(plot_origins[0], plot_origins[1],
                                                       plot_extents[0], plot_extents[1],
                                                       ix, iy, idx, idy)

        max_bounds = [0, 0] + list(self.data.value.shape[0:2])
        if sub_index is not None:
            # We need to extract a sub-image
            if self._cached_image_bounds != sub_index:
                self._cached_image_bounds = sub_index
            self._compute_cached_image()

        elif self._cached_image_bounds != max_bounds:
            # the cached data image is zoomed in, but we need to zoom out:
            self._cached_image_bounds = max_bounds
            self._compute_cached_image()

        elif self._cached_image is None:
            self._compute_cached_image()

        # if we have subimage coordinates, draw into there, otherwise
        # draw the Image into the bounds we calculated in self._layout
        if sub_coords is not None:
            rect = map(int, sub_coords)
        else:
            origin = self._xy[0]
            extents = self._xy[1] - self._xy[0]
            rect = (int(origin[0]), int(origin[1]), int(extents[0]), int(extents[1]))

        # check the pix format
        gc_fmt = gc.format()
        if gc_fmt != self._cached_image.format():
            self._cached_image.convert_pixel_format(gc_fmt, inplace=1)

        gc.set_alpha(self.alpha)
        gc.draw_image(self._cached_image, rect)
        return

#EOF
