""" A plot interaction that allows for selection of data points. """


# Major package imports.

# Enthought library imports.
from enthought.util.numerix  import transpose
from enthought.chaco.plot_interaction import PlotInteraction
from enthought.enable.drawing import DragSegment
from enthought.traits.api import Bool, Dict


class LineSelectInteraction(DragSegment, PlotInteraction):
    """ A plot interaction that allows for selection of plot values
        that cross a dragged line. """

    # The selected data points  { PlotValue : [ (x, y) ] }
    # If any segment of the PlotValue is crossed it will be in the
    # dictionary.  The values will be the endpoints of any crossed
    # segment.  The points are probably not useful, but this makes
    # this interaction consistent with other select interactions.
    selections = Dict

    # Should the selection disappear on complete.
    reset_on_complete = Bool(True)
    
    ###########################################################################
    # Trait event handlers.
    ###########################################################################
    
    def _complete_changed(self):
        """ Handle the line being completed. """

        selections = {}
        line_points = self.points
        assert len(line_points) == 2
        for value in self.canvas.values:
            index = value.index.data
            data = value.data

            # map_screen() returns a 2xN array, but to iterate over the points
            # we need Nx2.
            points = transpose(value.map_screen(index,data))

            for i in range(len(points)-1):
                segment = points[i:i+2]
                if self._intersects(segment, line_points):
                    # Put the PlotValue in the dict and get the
                    # point_list
                    point_list = selections.setdefault(value, [])
                    # Add the segment endpoints to the list
                    for (i, point) in enumerate(segment):
                        if point not in point_list:
                            point_list.append(point)
                
        self.selections = selections

        if self.reset_on_complete:
            self.reset()
        
        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _intersects(self, seg1, seg2):
        """ Determines if the 2 segments intersect.  
        
        The notation is from:
        http://softsurfer.com/Archive/algorithm_0104/algorithm_0104B.htm

        w=seg1[0]-seg2[0]
        u=seg1[1]-seg1[0]
        v=seg2[1]-seg2[2]
        """

        w = _diff(seg1[0], seg2[0])
        u = _diff(seg1[1], seg1[0])
        v = _diff(seg2[1], seg2[0])

        denom = _perp(u,v)
        if denom == 0.0:
            # In this case the segments are parallel.  They may be colinear
            # and therefore possibly coincide, but we are not going to
            # count this as an intersection.
            return False
        else:
            s = _perp(v,w)/denom
            t = _perp(u,w)/denom
            return (0.0 <= s and s <= 1.0 and 0.0 <= t and t <= 1.0)

def _diff(p1, p2):
    """
    Return the difference between the 2 points, p1-p2.
    """

    return ((p1[0]-p2[0]), (p1[1]-p2[1]))

def _perp(v1, v2):
    """
    Returns the perpendicular product of 2 2D vectors.
    """

    return v1[0]*v2[1]-v1[1]*v2[0]
    
#### EOF ######################################################################

