""" A plot interaction that allows the user to draw a polyline. """


# Enthought library imports.
from enthought.chaco.plot_interaction import PlotInteraction
from enthought.chaco.plot_value import PlotValue
from enthought.enable.drawing import PointLine
from enthought.traits.api import List, Property, Tuple


class PolylineInteraction(PointLine, PlotInteraction):
    """
    A plot interaction that allows the user to draw a polyline.
    
    The user is initially in Normal mode.  Once he single-clicks, 
    he enters Draw mode, and a vertex is placed at the point of
    mouse-down.  He then continues to add vertices by moving and
    single-clicking, until he double-clicks to place the final
    vertex.  A right-click will erase the last vertex.  Once the
    last vertex is drawn, the user is returned to Normal mode.
    
    The line stays drawn until the user starts creating a new
    polyline again by single-clicking.
    """

    # The set of screen-space points for this polyline.
    points = Property
    
    # The set of data-space points for this polyline - shadow for points.
    _points = List(Tuple)
    
    ###########################################################################
    # Trait property accessors.
    ###########################################################################

    def _get_points(self):
        """ Get the points. """

        # Map _points back to screen space coordinates.
        value = self._get_mapping_value()
        points = [ value.map_screen(point[0], point[1]) 
                   for point in self._points ] 
                   
        return points

    ###########################################################################
    # PointLine interface.
    ###########################################################################

    def add_point(self, point):
        """ Add the point. """
        
        value = self._get_mapping_value()
        self._points.append(value.map_xy(*point)[:2])

        return
        
    def get_point(self, index):
        """ Get the point at the specified index. """

        value = self._get_mapping_value()
        return value.map_screen(*self._points[index])

    def set_point(self, index, point):
        """ Set the point at the specified index to point. """
        
        value = self._get_mapping_value()
        self._points[index] = value.map_xy(*point)[:2]
        
        return
    
    def remove_point(self, index):
        """ Remove the point with the specified index. """

        del self._points[index]
        
        return
        
    ###########################################################################
    # Line interface.
    ###########################################################################

    def reset(self):
        """ Reset the polygon to the initial state. """
        
        self._points = []
        self.event_state = 'normal'
        self.updated = self
        return

    ###########################################################################
    # Component interface.
    ###########################################################################

    def complete_left_down(self, event):
        """ Called when the left button is depressed in complete state. """
        
        # This causes the polyline to reset when there is an existing
        # polyline and the user clicks on a point that is not a vertex
        # of the polyline.  (Otherwise, pass the event through to the
        # default polyline handler.)
        if self._over_point(event, self.points) is not None:
            PointLine.complete_left_down(self, event)
        else:
            self.reset()
            self.updated = self
            self.normal_left_down(event)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_mapping_value(self):
        """ Return the plot value to use to map from screen to data space. """
        
        return self.canvas.values[0]
    

#### EOF ######################################################################
