#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  Author: Scott Swarts <swarts@enthought.com>
#
#-----------------------------------------------------------------------------

"""Interaction that allows zooming the graph to a selected box.
"""

# Standard library imports.

# Major packages.

# Enthought library imports
from enthought.enable.enable_traits import Pointer
from enthought.enable.drawing import DragBox
from enthought.enable.base import transparent_color
from enthought.chaco.plot_interaction import PlotInteraction
from enthought.chaco.plot_value import PlotValue
from enthought.chaco.image_plot_value import ImagePlotValue, CmapImagePlotValue
from enthought.traits.api import Bool
from enthought.traits.api import Bool

# Application specific imports.

# Local imports.

##############################################################################
# class 'ZoomInteraction'
##############################################################################

class ZoomInteraction(DragBox, PlotInteraction):
    """Interaction that allows zooming the graph to a selected box.
    """

    ##########################################################################
    # Traits
    ##########################################################################

    ### 'DragBox' interface ##################################################
    
    # Pointer for the complete state.
    complete_pointer = Pointer('arrow')

    ### 'Box' interface ######################################################

    # Set the box color to transparent so we just get the outline
    color = transparent_color

    # Set a border size to get the box
    border_size = 1

    # Whether to reset state on complete
    reset_on_complete = Bool(False)

    ##########################################################################
    # 'DragBox' interface
    ##########################################################################

    def _complete_changed(self):
        """ Zoom to the area of the box. """
        canvas = self.canvas

        x, y, dx, dy = self.bounds
        
        if dx == dy == 0:
            self.zoom_out(canvas)
        else:
            for value in canvas.values:
                # Map the rectangle to data space
                x1, y1, ignore = value.map_xy(x, y)
                x2, y2, ignore = value.map_xy(x+dx, y+dy)
                # Put values in order
                if x1 > x2:
                    x1, x2 = x2, x1
                if y1 > y2:
                    y1, y2 = y2, y1
                # Figure out which axis is which
                if value.index_axis == 'x':
                    x_axis = value.axis_index
                    y_axis = value.axis
                else:
                    x_axis = value.axis
                    y_axis = value.axis_index
                # Save the original bounds
                self.save_bounds(x_axis)
                self.save_bounds(y_axis)
                # Set the bounds
                x_axis.bound_low = x1
                x_axis.bound_high = x2
                y_axis.bound_low = y1
                y_axis.bound_high = y2

        # Maybe reset the state
        if self.reset_on_complete:
            self.reset()
            

        # Maybe reset the state
        if self.reset_on_complete:
            self.reset()
            
        # Update the canvas
        canvas.update()
        self.event_state = 'normal'

    def reset(self):
        """
        Sets the event state back to normal so we can zoom again.
        """
        self.event_state = "normal"

    def complete_draw(self, gc):
        """ Do nothing.  The box should go away after the zoom. """
        pass

    def save_bounds(self, axis):
        """Saves the bounds on the axis if they have not already been saved."""
        if not getattr(axis, "_ZoomInteraction__bounds_saved", False):
            axis.__original_bound_low = axis.bound_low
            axis.__original_bound_high = axis.bound_high
            axis.__bounds_saved = True

    def zoom_out(plottable, visited=None):
        """
        Zooms the canvas all the way out.  If there are saved bounds in
        the axes we will restore these, otherwise we will set them
        to auto.
        """
        if visited is None:
            visited = {}
        elif id(plottable) in visited:
            return
        visited[id(plottable)] = 1
        
        if isinstance(plottable, PlotValue) or \
               isinstance(plottable, ImagePlotValue) or\
               isinstance(plottable, CmapImagePlotValue):
            for axis in (plottable.axis, plottable.axis_index):
                if getattr(axis, "_ZoomInteraction__bounds_saved", False):
                    axis.bound_low = axis.__original_bound_low
                    axis.bound_high = axis.__original_bound_high
        else:
            for value in plottable.values:
                ZoomInteraction.zoom_out(value, visited)

    zoom_out = staticmethod(zoom_out)


    #~ def normal_right_up(self, event):
        #~ """ Handle the right mouse button coming up in the 'normal' state. """

        #~ self.event_state = 'complete'
        #~ self.pointer = self.complete_pointer
        #~ self.complete = True
        #~ self.zoom_out(self.canvas)
        #~ return

    def _draw(self, *args):
        if self.event_state == "drawing":
            super(DragBox, self)._draw(*args)


#### EOF ######################################################################
