#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Copyright (c) 2002-2004 John D. Hunter
# All Rights Reserved.
#------------------------------------------------------------------------------

""" 
A colormap which has segments of linearly interpolated color steps.
"""

import enthought.util.numerix as Numeric
# Major package imports.
from enthought.util.numerix import add, arange, array, asarray, choose, clip, concatenate
from enthought.util.numerix import divide, ones, put, ravel, resize, searchsorted, shape
from enthought.util.numerix import sometrue, sort, take, where, zeros

from types import IntType, FloatType

# Enthought library imports.
from enthought.traits.api import Array, Bool, Dict, Float, Int, Property, Str, Tuple

# Local imports.
from colormap import AbstractColormap


class LinearSegmentedColormap(AbstractColormap):
    """ An AbstractColormap which has a set of linearly interpolated steps.

    The lookup transfer function is a simple linear function between defined
    intensities.  There is no limit to the number of steps that may be
    defined. As the segment intervals start containing fewer and fewer array
    locations, there will be inevitable quantization errors.

    Construction of a LinearSegmentedColormap be done through the factory
    methods 'from_palette_array' and 'from_segment_map'.  Direct calls to the
    LinearSegmentedColormap constructor are discouraged.
    """

    ###########################################################################
    # Static methods.
    ###########################################################################

    def from_palette_array(palette, **traits):
        """ Create a LinearSegmentedColormap from a palette array.

        The palette colors will be linearly interpolated across the range of
        mapped values.

        palette - A 3 by n array of intensity values, where n > 1:
          [[R0, G0, B0], ... [R(n-1), G(n-1), B(n-1)]]
        """

        n_colors = shape(palette)[0]
        if n_colors < 2:
            raise ValueError("Palette must contain at least two colors.")

        # Compute the % offset for each of the color locations.
        offsets = zeros(n_colors, Numeric.Float)
        offsets[:-1] = arange(0.0, 1.0, 1.0/(n_colors-1))
        offsets[-1] = 1.0

        # From the offsets and the color data, generate a segment map.
        segment_map = {}
        red_values = palette[::,0]
        segment_map['red'] = zip(offsets, red_values, red_values)
        green_values = palette[::,1]
        segment_map['green'] = zip(offsets, green_values, green_values)
        blue_values = palette[::,2]
        segment_map['blue'] = zip(offsets, blue_values, blue_values)

        return LinearSegmentedColormap(segment_map, **traits)

    from_palette_array = staticmethod(from_palette_array)

    def from_segment_map(segment_map, **traits):
        """ Create a LinearSegmentedColormap from a segment map.

        segment_map: a dictionary with red, green, and blue entries.  Each
        entry is a list of x, y0, y1 tuples.  See makeMappingArray for details
        on how segment maps are converted to color maps.
        """

        return LinearSegmentedColormap(segment_map, **traits)

    from_segment_map = staticmethod(from_segment_map)

    ###########################################################################
    # Traits.
    ###########################################################################

    #### Public ####

    # The color table.
    color_bands = Property(Array)

    # The maximum value for the mapped data.
    max_value = Float(1.0)

    # The minimum value for the mapped data.
    min_value = Float(0.0)

    # The total number of color steps in the map.
    steps = Int(256)
    
    

    #### Private ####

    # Flag to indicate the mapping arrays are out of date.
    _dirty = Bool(True)
    
    # The raw segment data for creating the mapping array.
    _segmentdata = Dict(Str, Tuple)


    ###########################################################################
    # Public 'object' interface.
    ###########################################################################
    
    def __init__(self, segmentdata, **traits):
        """ Constructor.

        segmentdata: a dictionary with red, green and blue entries. Each entry
        should be a list of x, y0, y1 tuples.  See makeMappingArray for details
        on how segmentdata is converted to a color map.
        """

        self._segmentdata = segmentdata
        super(AbstractColormap, self).__init__(**traits)
        return

    ###########################################################################
    # Public 'AbstractColormap' interface.
    ###########################################################################

    def map_array(self, ary):
        """ Maps an array of values to their corresponding colors. """

        if self._dirty:
            self._recalculate()

        result = self._map(
            clip((ary - self.min_value) / (self.max_value - self.min_value), 0.0, 1.0)
        )
            
        return result

    def map_array_index(self, ary):
        """ Maps an array of values to their corresponding color index. """

        if self._dirty:
            self._recalculate()

        indices = (ary - self.min_value) / (self.max_value - self.min_value) * self.steps

        return clip(indices.astype(IntType), 0, self.steps - 1)

    
    #### Property accessors. ####

    def _get_color_bands(self):
        """ Gets the color bands array. """

        if self.color_depth is 'rgba':
            alpha = ones(self.steps)
            result = zip(self._red_lut, self._green_lut, self._blue_lut, alpha)

        else:
            result = zip(self._red_lut, self._green_lut, self._blue_lut)

        return result
    
    ###########################################################################
    # Private interface.
    ###########################################################################

    def _recalculate(self):
        """ Recalculate the mapping arrays. """

        self._red_lut = self._make_mapping_array(
            self.steps, self._segmentdata['red']
        )
        self._green_lut = self._make_mapping_array(
            self.steps, self._segmentdata['green']
        )
        self._blue_lut = self._make_mapping_array(
            self.steps, self._segmentdata['blue']
        )

        self._dirty = False
        
        return

    #### matplotlib ####
    def _make_mapping_array(self, n, data):
        """Create an N-element 1-d lookup table
        
        data represented by a list of x,y0,y1 mapping correspondences.
        Each element in this list represents how a value between 0 and 1
        (inclusive) represented by x is mapped to a corresponding value
        between 0 and 1 (inclusive). The two values of y are to allow 
        for discontinuous mapping functions (say as might be found in a
        sawtooth) where y0 represents the value of y for values of x
        <= to that given, and y1 is the value to be used for x > than
        that given). The list must start with x=0, end with x=1, and 
        all values of x must be in increasing order. Values between
        the given mapping points are determined by simple linear interpolation.
        
        The function returns an array "result" where result[x*(N-1)]
        gives the closest value for values of x between 0 and 1.
        """
        
        try:
            adata = array(data)
        except:
            raise TypeError("data must be convertable to an array")
        shape = adata.shape
        if len(shape) != 2 and shape[1] != 3:
            raise ValueError("data must be nx3 format")
        
        x  = adata[:,0]
        y0 = adata[:,1]
        y1 = adata[:,2]
        
        if x[0] != 0. or x[-1] != 1.0:
            raise ValueError(
                "data mapping points must start with x=0. and end with x=1")
        if sometrue(sort(x)-x):
            raise ValueError(
                "data mapping points must have x in increasing order")
        # begin generation of lookup table
        x = x * (n-1)
        lut = zeros((n,), Numeric.Float)
        xind = arange(float(n))
        ind = searchsorted(x, xind)[1:-1]
        
        lut[1:-1] = ( divide(xind[1:-1] - take(x,ind-1),
                             take(x,ind)-take(x,ind-1) )
                      *(take(y0,ind)-take(y1,ind-1)) + take(y1,ind-1))
        lut[0] = y1[0]
        lut[-1] = y0[-1]
        
        # ensure that the lut is confined to values between 0 and 1 by clipping it
        lut = where(lut > 1., 1., lut)
        lut = where(lut < 0., 0., lut)

        return lut

    #### matplotlib ####
    def _map(self, X, alpha=1.0):
        """
        X is either a scalar or an array (of any dimension).
        If scalar, a tuple of rgba values is returned, otherwise
        an array with the new shape = oldshape+(4,).  Any values
        that are outside the 0,1 interval are clipped to that
        interval before generating rgb values.  
        Alpha must be a scalar
        """
        
        alpha = min(alpha, 1.0) # alpha must be between 0 and 1
        alpha = max(alpha, 0.0)
        if type(X) in [IntType, FloatType]:
            vtype = 'scalar'
            xa = array([X])
        else:
            vtype = 'array'
            xa = asarray(X)

        # assume the data is properly normalized
        #xa = where(xa>1.,1.,xa)
        #xa = where(xa<0.,0.,xa)

        xa = (xa *(self.steps-1)).astype(Numeric.Int)
        rgba = zeros(xa.shape+(4,), Numeric.Float)
        rgba[...,0] = take(self._red_lut, xa)
        rgba[...,1] = take(self._green_lut, xa)
        rgba[...,2] = take(self._blue_lut, xa)
        rgba[...,3] = alpha
        if vtype == 'scalar':
            rgba = tuple(rgba[0,:])
            
        return rgba

#### EOF ######################################################################
