""" Helper class to layout and draw overlays. """

from plot_base import BOTTOM, HCENTER, LEFT, RIGHT, TOP, VCENTER
from plot_overlay import PlotOverlay

# Mapping from a position mask to a list index:
position_mask_to_index = {
    (TOP + LEFT):          0,
    (TOP + HCENTER):     1,
    (TOP + RIGHT):         2,
    (VCENTER + LEFT):    4,
    (VCENTER + HCENTER): 5,
    (VCENTER + RIGHT):   6,
    (BOTTOM + LEFT):       8,
    (BOTTOM + HCENTER):  9,
    (BOTTOM + RIGHT):     10
}

class OverlayInfo:
    """ Helper class to layout and draw overlays. """
    
    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
  
    def __init__ ( self, gc, overlays = None ):
        self.dx  = self.ldx = self.rdx = self.mdx = 0
        self.dy  = self.tdy = self.bdy = self.mdy = 0
        self.dys = [ 0 ] * 12
        self.dxs = [ 0 ] * 12
        self.frame_overlays = []
        self.plot_overlays  = []
        if overlays is not None:
            self.add_overlays( gc, overlays )
 
    #---------------------------------------------------------------------------
    #  Add a group of overlays:
    #---------------------------------------------------------------------------
  
    def add_overlays ( self, gc, overlays ):
        for overlay_item in overlays:
            self.min_size( gc, overlay_item )
        dxs     = self.dxs
        dys     = self.dys
        self.dx = max( dxs[0] + dxs[1] + dxs[2],
                       dxs[4] + dxs[5] + dxs[6],
                       dxs[8] + dxs[9] + dxs[10] )
        dys[3]  = max( dys[0], dys[1], dys[2] )
        dys[7]  = max( dys[4], dys[5], dys[6] )
        dys[11] = max( dys[8], dys[9], dys[10] )
        self.dy = dys[3] + dys[7] + dys[11]

    #---------------------------------------------------------------------------
    #  Return the list of overlays that contain the specified (x,y) point:
    #---------------------------------------------------------------------------
 
    def objects_at ( self, x, y ):
        result = []
 
        # Test all of the frame overlays:
        for overlay_item in self.frame_overlays:
            objects = overlay_item.item.objects_at( x, y )
            if objects:
                result = objects
                break
 
        # Now test all of the plot overlays:
        for overlay_item in self.plot_overlays:
            overlay = overlay_item.item
            tx      = x + overlay._plot_ox
            ty      = y + overlay._plot_oy
            objects = overlay.objects_at( x, y )
            if objects:
                result.extend( objects )
            elif overlay._hline is not None:
                x1, y1, x2, y2 = overlay._hline
                lw2 = overlay.line_weight / 2
                if (tx >= x1) and (tx < x2) and (abs( ty - y1 ) <= (lw2 + 3)):
                    result.append( overlay )
            elif overlay._vline is not None:
                x1, y1, x2, y2 = overlay._vline
                lw2 = overlay.line_weight / 2
                if (ty >= y1) and (ty < y2) and (abs( tx - x1 ) <= (lw2 + 3)):
                    result.append( overlay )
 
        # Return the result:
        return result

    #---------------------------------------------------------------------------
    #  Add the minimum size information for an overlay:
    #---------------------------------------------------------------------------
        
    def min_size ( self, gc, overlay_item ):
        overlay = overlay_item.item
        dx, dy  = overlay._min_size( gc )
        overlay_item.dx = dx
        overlay_item.dy = dy
        
        from plot_value import PlotValue
        if (isinstance( overlay, PlotOverlay ) and ### EXPERIMENTAL ###
            ((overlay.value is not None) or (overlay.index is not None)) and
            isinstance( overlay.container, PlotValue )):
            self.plot_overlays.append( overlay_item )
        else:
            self.frame_overlays.append( overlay_item )
            index = position_mask_to_index[ overlay.position_ ]
            self.dxs[ index ]  = max( self.dxs[ index ], dx )
            self.dys[ index ] += dy
            
    #---------------------------------------------------------------------------
    #  Calculate the best horizontal center point for a specified layout 'band':
    #---------------------------------------------------------------------------
 
    def bestcx ( self, x, dx, band ):
        ldx   = self.dxs[ band ]
        mdx   = self.dxs[ band + 1 ]
        rdx   = self.dxs[ band + 2 ]
        mincx = x + ldx + (mdx / 2)
        maxcx = x + dx - rdx - (mdx / 2)
        cx    = x + dx / 2
        if not (mincx <= cx <= maxcx):
            if abs( mincx - cx ) < abs( maxcx - cx ):
                cx = mincx
            else:
                cx = maxcx
        return cx
           
    #---------------------------------------------------------------------------
    #  Calculate the best vertical center point for a specified layout 'band':
    #---------------------------------------------------------------------------
 
    def bestcy ( self, y, dy, band ):
        tdy   = self.dys[ band ]
        mdy   = self.dys[ band + 4 ]
        bdy   = self.dys[ band + 8 ]
        mincy = y + tdy
        maxcy = y + dy - bdy - mdy
        cy    = y + dy / 2 - (mdy / 2)
        if not (mincy <= cy <= maxcy):
            if abs( mincy - cy ) < abs( maxcy - cy ):
                cy = mincy
            else:
                cy = maxcy
        return cy
           
    #---------------------------------------------------------------------------
    #  Lay out the accumulated overlays:
    #---------------------------------------------------------------------------
 
    def _layout ( self, x, y, dx, dy ): 
       
        # Cache the bounds of the area we have been assigned:
        self._plot_bounds = ( x, y, dx, dy )
       
        # Set up the initial coordinates:
        tcx  = self.bestcx( x, dx, 0 )
        mcx  = self.bestcx( x, dx, 4 )
        bcx  = self.bestcx( x, dx, 8 )
        lcy  = self.bestcy( y, dy, 0 )
        ccy  = self.bestcy( y, dy, 3 )
        rcy  = self.bestcy( y, dy, 2 )
        rx   = x + dx
 
        lty  = cty = rty = y + dy
        lby  = cby = rby = y
        
        # Iterate over each frame overlay and lay it out based on its position:
        for overlay_item in self.frame_overlays:
            overlay  = overlay_item.item
            position = overlay.position_
            
            if position == (TOP + LEFT):
                lty -= overlay_item.dy
                overlay._layout( x, lty, overlay_item.dx, overlay_item.dy )
 
            elif position == (TOP + HCENTER):
                cty -= overlay_item.dy
                overlay._layout( tcx - overlay_item.dx / 2, cty,
                                 overlay_item.dx, overlay_item.dy )
 
            elif position == (TOP + RIGHT):
                rty -= overlay_item.dy
                overlay._layout( rx - overlay_item.dx, rty,
                                 overlay_item.dx, overlay_item.dy )
 
            elif position == (VCENTER + LEFT):
                overlay._layout( x, lcy, overlay_item.dx, overlay_item.dy )
                lcy += overlay_item.dy
 
            elif position == (VCENTER + HCENTER):
                overlay._layout( mcx - overlay_item.dx / 2, ccy,
                                 overlay_item.dx, overlay_item.dy )
                ccy += overlay_item.dy
 
            elif position == (VCENTER + RIGHT):
                overlay._layout( rx - overlay_item.dx, rcy,
                                 overlay_item.dx, overlay_item.dy )
                rcy += overlay_item.dy
                rcy -= overlay_item.dy
 
            elif position == (BOTTOM + LEFT):
                overlay._layout( x, lby, overlay_item.dx, overlay_item.dy )
                lby += overlay_item.dy
 
            elif position == (BOTTOM + HCENTER):
                overlay._layout( bcx - overlay_item.dx / 2, cby,
                                 overlay_item.dx, overlay_item.dy )
                cby += overlay_item.dy
               
            elif position == (BOTTOM + RIGHT):
                overlay._layout( rx - overlay_item.dx, rby,
                                 overlay_item.dx, overlay_item.dy )
                rby += overlay_item.dy
 
        # Iterate over each plot overlay and lay it out based on its position:
        for overlay_item in self.plot_overlays:
            overlay   = overlay_item.item
            plotvalue = overlay.container
 
            # Get the screen coordinates of the (index,value) pair:
            xs, ys = plotvalue.map_screen( overlay.index, overlay.value )

            # Here we round off to the half pixel, so that Agg will place all the
            # energy in a single pixel, rather than anti-aliasing across two.
            if xs is not None:
                xs = round( xs ) + 0.5
            if ys is not None:
                ys = round( ys ) + 0.5
 
            # Get the bounds of the underlying PlotValue:
            px, py, pdx, pdy = plotvalue._plot_bounds
 
            # Calculate the coordinates of the overlay line (if any):
            overlay._hline = overlay._vline = None
            if overlay.visible and (overlay.line_color_ is not None):
                if overlay.span:
                    spx, spy, spdx, spdy = plotvalue._span_plot_bounds
                    if xs is not None:
                        overlay._vline = ( xs, spy, xs, spy + spdy - 1 )
                    if ys is not None:
                        overlay._hline = ( spx, ys, spx + spdx - 1, ys )
                else:
                    if xs is not None:
                        overlay._vline = ( xs, py, xs, py + pdy - 1 )
                    if ys is not None:
                        overlay._hline = ( px, ys, px + pdx - 1, ys )
                  
            # Layout the frame:
            position = overlay.position_
            fdx      = overlay_item.dx
            fdy      = overlay_item.dy
            
            if position & LEFT:
                if xs is None:
                    xs = px
                else:
                    xs -= fdx
            elif position & HCENTER:
                if xs is None:
                    xs = px + ((pdx - fdx) / 2)
                else:
                    xs -= fdx / 2
            elif xs is None:
                xs = px + pdx - fdx
               
            if position & TOP:
                if ys is None:
                    ys = py + pdy - fdy
            elif position & VCENTER:
                 if ys is None:
                     ys = py + ((pdy - fdy) / 2)
                 else:
                     ys -= fdy /2
            elif ys is None:
                ys = py
            else:
                ys -= fdy
            overlay._layout( xs, ys, fdx, fdy )

    #---------------------------------------------------------------------------
    #  Draw all of the overlays:
    #---------------------------------------------------------------------------
 
    def _draw ( self, gc, ox = 0, oy = 0 ):
 
        # Draw each of the frame overlays:
        for overlay_item in self.frame_overlays:
            overlay_item.item._draw( gc )
 
    def _draw_plot ( self, gc, ox = 0, oy = 0 ):
           
        # Draw each of the plot overlays:
        for overlay_item in self.plot_overlays:
            overlay_item.item._draw_plot( gc, ox, oy )

#### EOF ######################################################################
