"""
Defines the PlotAxis class
"""

# Enthought library imports
from enthought.traits.api import Delegate, Dict, Float, HasTraits, Instance, Str, \
                             Trait, Unicode
from enthought.traits.api import Float as FloatTrait
from enthought.traits.ui.api import View

# Local imports
from axis_style     import AxisStyle
from plot_container import NestedContainer
from plot_frame     import PlotFrame
from plot_label     import PlotLabel


StyleDelegate = Delegate('style')

class PlotAxis ( NestedContainer ):
    """
    Describes the visual and layout characteristics of an axis.  It controls things
    such as the minimum and maximum plot bounds, the color and lines styles for
    axis tick marks and grid lines, label formatting, etc.
    """
    
    # Traits delegation prefix
    __prefix__ = 'axis_'

    title         = Trait( '?', PlotLabel, Str, Unicode, label = 'Title' )
    markers       = Dict(FloatTrait, Instance(PlotLabel))
    style         = Instance(AxisStyle, ())
    scale         = StyleDelegate
    position      = StyleDelegate
    menu          = StyleDelegate
    line_weight   = StyleDelegate
    line_color    = StyleDelegate
    line_style    = StyleDelegate
    label_font    = StyleDelegate
    label_color   = StyleDelegate
    label_style   = StyleDelegate
    label_offset  = StyleDelegate
    label_format  = StyleDelegate
    tick_visible  = StyleDelegate
    tick_weight   = StyleDelegate
    tick_color    = StyleDelegate
    tick_style    = StyleDelegate
    tick_in       = StyleDelegate
    tick_out      = StyleDelegate
    tick_interval = StyleDelegate
    grid_visible  = StyleDelegate
    grid_weight   = StyleDelegate
    grid_color    = StyleDelegate
    grid_style    = StyleDelegate
    margin        = StyleDelegate
    margin_height = StyleDelegate
    margin_width  = StyleDelegate
    bound_low     = StyleDelegate
    bound_high    = StyleDelegate
    zoomable      = StyleDelegate

    traits_view = View(
        [ #[ 'bound_low{Lower}', 'bound_high{Upper}', '|[Axis Bounds]' ],
          [ 'title$', 'scale{Scaling}', '_', 'position{Position}', ' ', 
            'zoomable{Is Zoomable?}', 'visible',
            '|[Axis]@' ],
            '|{Main}' ],
        [ [ 'line_color{Color}', '_', 'line_style{Style}', '_',
            'line_weight{Weight}',
            '|[Line]@' ],
            '|{Line}' ],
        [ [ 'label_color{Color}', ' ', 'label_font{Font}',  ' ',
            'label_offset{Offset}', 'label_format{Format}', '_',
            'label_style{Positions}',
            '|[Label]@' ],
            '|{Label}' ],
        [ [ 'tick_color{Color}', '_', 'tick_style{Style}', '_',
            'tick_weight{Weight}',
            '|[Tick Line]@' ],
          [ 'tick_interval{Interval}', ' ', 'tick_visible{Is Visible?}',
            '|[Tick]' ],
            '|{Tick}' ],
        [ [ 'tick_in{Inside}', '_', 'tick_out{Outside}',
            '|[Tick Length]@' ],
            '|{Tick Length}' ],
        [ [ 'grid_color{Color}', '_', 'grid_style{Style}', '_',
            'grid_weight{Weight}',
            '|[Grid Line]@' ],
          [ 'grid_visible{Is Visible?}',
            '|[Grid]' ],
            '|{Grid}' ],
        [ [ 'margin_height{Height}', 'margin_width{Width}',
            '|[Margin]' ],
            '|{Margin}' ],
        [ [ 'menu', 
            '|[Miscellaneous]' ],
            '|{Misc}' ]
    )

    
    def __init__ ( self, **traits ):
        # Initialize the 'title' trait to make sure the shadow value is set:
        traits.setdefault( 'title', '' )

        # Indicate the axis is not currently 'zoomed':
        self._bound_low = None
        
        # Initialize the label caches:
        self._available = None
        self._in_use    = []
 
        # Initialize the rest of the traits
        HasTraits.__init__( self, **traits )
        return
 
    def _title_changed ( self, old, new ):
        # If 'new' is not a PlotLabel, then it is a string:
        if not isinstance( new, PlotLabel ):
            # Convert it to a PlotLabel:
            new = PlotLabel( text = new,
                             margin = 2,
                             font=self.label_font,
                             color=self.label_color )
           
        # Save the PlotLabel in a shadow variable:
        self._title_label = new
 
        # Make sure we are the PlotLabel's container:
        new.container = self
        return
 
    def _margin_changed ( self, old, new ):
        self.margin_height = self.margin_width = new
        return

    def _label_color_changed ( self ):
        if self._title_label:
            self._title_label.color = self.label_color
        return
    
# EOF
