#-------------------------------------------------------------------------------
#
#  Define the base plot constants and types
#
#  Written by: David C. Morrill
#
#  Date: 07/10/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import exceptions
import UserList
import weakref

from enthought.kiva.constants import *

from enthought.util.scipyx import NaN, arange, array, ArrayType, floor, \
        NewAxis, shape, alltrue, isfinite, logical_or, logical_and, resize, isnan, \
        concatenate, greater_equal, nonzero, greater, less, less_equal, log10, \
        minimum, argsort, sqrt, add, amin, amax, bitwise_and, compress, PINF, take, \
        zeros, not_equal

from enthought.traits.api import Trait, HasTraits, TraitType, Delegate, TraitError,\
                             ReadOnly, Any, Undefined, Str
from types  import ListType, TupleType, FloatType, IntType, \
                   LongType, ComplexType, FunctionType, MethodType, DictType, \
                   InstanceType, ClassType
from string import lowercase, uppercase

#-------------------------------------------------------------------------------
#  Variables:
#-------------------------------------------------------------------------------

# Debug/Performance variables:
TIMING = 0

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Large floating point values:
BIG_POS_FLOAT =  1.0e308
BIG_NEG_FLOAT = -1.0e308

# Numeric 'array' type:
ArrayType = type( arange( 1.0 ) )

# An array containing just a single NaN:
nan_array = array( [ NaN ] )

# Numeric types:
numeric_types = [ IntType, LongType, FloatType, ComplexType ]

# Basic sequence types:
basic_sequence_types = [ ListType, TupleType ]

# Sequence types:
sequence_types = [ ArrayType, ListType, TupleType ]

# Positions:
TOP     = 32
VCENTER = 16
BOTTOM  =  8
LEFT    =  4
HCENTER =  2
RIGHT   =  1

# Update flags:
REDRAW         = 0
NEEDS_MIN_SIZE = 1
NEEDS_LAYOUT   = 2
RESIZE         = NEEDS_LAYOUT
UPDATE         = NEEDS_LAYOUT + NEEDS_MIN_SIZE

#-------------------------------------------------------------------------------
#  Standard Colors and Line Styles:
#-------------------------------------------------------------------------------

# Colors:
black_color       = (0.0, 0.0, 0.0, 1.0)
red_color         = (1.0, 0.0, 0.0, 1.0)
green_color       = (0.0, 1.0, 0.0, 1.0)
blue_color        = (0.0, 0.0, 1.0, 1.0)

# Dictionary of line drawing styles:
line_styles = {
    'solid':     None,
    'dot dash':  array( [ 3.0, 5.0, 9.0, 5.0 ] ),
    'dash':      array( [ 6.0, 6.0 ] ),
    'dot':       array( [ 2.0, 2.0 ] ),
    'long dash': array( [ 9.0, 5.0 ] )
}

# Line styles:
dash_line_style       = line_styles[ 'dash' ]
dot_line_style        = line_styles[ 'dot' ]
short_dash_line_style = array( [ 3.0, 3.0 ] )
solid_line_style      = line_styles[ 'solid' ]


#-------------------------------------------------------------------------------
#  Standard delegates:
#-------------------------------------------------------------------------------

delegate        = Delegate( 'container', '*' )
simple_delegate = Delegate( 'container' )
plot_delegate   = Delegate( 'container', 'plot_*' )

#-------------------------------------------------------------------------------
#  'PlotError' class:
#-------------------------------------------------------------------------------

class PlotError ( exceptions.Exception ):

    def __init__ ( self, args ):
        self.args = args

#-------------------------------------------------------------------------------
#  'PlotItem' class:
#-------------------------------------------------------------------------------

class PlotItem:

    def __init__ ( self, item ):
        self.item = item

#-------------------------------------------------------------------------------
#  'SimplePlotItemList' class:
#-------------------------------------------------------------------------------

class SimplePlotItemList ( UserList.UserList ):

    def set ( self, **traits ):
        for item in self.data:
            item.set( traits )

    def add ( self, *items ):
        for item in self.data:
            apply( item.add, items )

    def remove ( self, *items ):
        for item in self.data:
            apply( item.remove, items )

#-------------------------------------------------------------------------------
#  'PlotItemList' class:
#-------------------------------------------------------------------------------

class PlotItemList:

    def __init__ ( self, owner, list ):
        self.owner = weakref.ref( owner )
        self.list  = list

    def __len__ ( self ):
        return len( self.list )

    def __getitem__ ( self, index ):
        return self.list[ index ].item

    def __setitem__ ( self, index, item ):
        item = self.__validate__( item )
        del self.list[ index ].item.container
        item.container.remove( item )
        self.list[ index ] = PlotItem( item )
        item.container     = self.owner()

    def __delitem__ ( self, index ):
        del self.list[ index ].item.container
        del self.list[ index ]

    def __getslice__ ( self, index1, index2 ):
        return SimplePlotItemList(
                  [ x.item for x in self.list[ index1: index2 ] ] )

    def __setslice__ ( self, index1, index2, list ):
        list2 = [ PlotItem( self.__validate__( x ) ) for x in list ]
        owner = self.owner()
        for item in list:
            if item.container != owner:
                item.remove()
        for plot_item in self.list[ index1: index2 ]:
            del plot_item.item.container
        self.list[ index1: index2 ] = list2
        for item in list:
            item.container = owner

    def __delslice__ ( self, index1, index2 ):
        for plot_item in self.list[ index1: index2 ]:
            del plot_item.item.container
        del self.list[ index1: index2 ]

    def __contains__ ( self, item ):
        for plot_item in self.list:
            if item == plot_item.item:
                return True
        return False

    def __validate__ ( self, item ):
        return item

    def set ( self, **traits ):
        for value_item in self.list:
            value_item.item.set( traits )

#-------------------------------------------------------------------------------
#  'PlotCache' class:
#-------------------------------------------------------------------------------

class PlotCache:

   pass

#-------------------------------------------------------------------------------
#  'PlotUndoHandler' class:
#-------------------------------------------------------------------------------

class PlotUndoHandler:

    def add_undo ( self, command ):
        pass

#-------------------------------------------------------------------------------
#  Perform a binary search of a sorted point array looking for a specified
#  x or y value:
#-------------------------------------------------------------------------------

def bin_search2 ( pts, value, coord, ascending ):
    lo = 0
    hi = len( pts ) - 1
    while True:
        mid  = (hi + lo) / 2
        test = cmp( pts[ mid, coord ], value ) * ascending
        if test == 0:
            return mid
        if test > 0:
            hi = mid - 1
            if hi < lo:
                return lo
        else:
            lo = mid + 1
            if lo > hi:
                return hi

#-------------------------------------------------------------------------------
#  Perform a binary search of a sorted array looking for a specified value:
#-------------------------------------------------------------------------------

def bin_search1 ( values, value, ascending ):
    if ascending > 0:
        if len(values) == 0 or (value < values[0]) or (value > values[-1]):
            return -1
    else:
        if len(values) == 0 or (value < values[-1]) or (value > values[0]):
            return -1
    lo = 0
    hi = len( values )
    while True:
        mid  = (hi + lo) / 2
        test = cmp( values[ mid ], value ) * ascending
        if test == 0:
            return mid
        if test > 0:
            hi = mid
        else:
            lo = mid
        if lo >= (hi - 1):
            return lo

#-------------------------------------------------------------------------------
#  Return a 'user-friendly' name for a specified plotting object or string:
#-------------------------------------------------------------------------------

def user_name_for ( object, lower = True ):
    name = object.__class__.__name__.replace( '_', ' ' )
    if name[0:4] == 'Plot':
        name = name[4:]
    return split_name_for( name, lower )

def ui_label_for ( name ):
    return string.capwords( split_name_for( name ) )

def split_name_for ( name, lower = False ):
    result     = ''
    last_lower = False
    for c in name:
        if (c in uppercase) and last_lower:
            result += ' '
        last_lower = (c in lowercase)
        result    += c
    if lower:
        return result.lower()
    return result
