"""
Defines the PlotCanvas and Plot{Canvas, Title, Overlay, Value, Group}List
classes.
"""

import weakref

# Enthought library imports.
from enthought.enable.base import xy_in_bounds, empty_rectangle
from enthought.enable          import Component
from enthought.traits.api      import Instance, Trait
from enthought.traits.api      import Float as TraitsFloat
from enthought.traits.ui.api   import Group, View, Include

# Local imports.
from axes_info             import AxesInfo
from overlay_info          import OverlayInfo
from plottable             import Plottable
from plot_base             import PlotItemList
from plot_container        import RENDER_INFO, scrollbar_trait
from plot_data             import PlotData
from plot_frame            import PlotFrame
from plot_overlay          import PlotOverlay
from plot_source           import PlotSource
from plot_title            import PlotTitle
from plot_traits           import get_renderer_for
from plot_views            import plot_margin_group, bar_width_group, \
                                  layout_group, plot_group, marker_group, \
                                  cursor_group, line_group, border_group, \
                                  margin_group, misc_group
from title_info            import TitleInfo

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Scroll bar margin (in pixels):
scrollbar_margin = 5
    

class PlotCanvas ( Plottable ):
    """
    A PlotCanvas plots one or more PlotValue objects on a single plotting area.
    
    It can also contain PlotTitle, PlotOverlay, PlotGroup, PlotIndexed, and
    other PlotCanvas and PlotValue objects whose 'overlay' trait is True.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    index_scrollbar = Trait( scrollbar_trait, 
                             desc = 'the scrollbar to use for scrolling '
                                    'the index' )

    value_scrollbar = Trait( scrollbar_trait, 
                             desc = 'the scrollbar to use for scrolling '
                                    'the value' )

    interaction = Instance('enthought.chaco.plot_interaction.PlotInteraction')
    
    # The plot's aspect ratio (dx / dy).
    aspect_ratio = TraitsFloat

    
    def __init__ ( self, *items, **traits ):
        Plottable.__init__( self, **traits )
        self._titles   = []
        self._overlays = []
        self._values   = []
        self.titles    = PlotTitleList(   self, self._titles )
        self.overlays  = PlotOverlayList( self, self._overlays )
        self.values    = PlotCanvasList(  self, self._values )
        self.add( *items )

    #---------------------------------------------------------------------------
    # 'Plottable' interface
    #---------------------------------------------------------------------------

    def components_at ( self, x, y ):
        # FIXME THISHERE IS A TOTAL HACK!
        result = super(PlotCanvas, self).components_at(x, y)
        if self.interaction is not None:
            # If we have an interaction, we need it to be the first to receive
            # events.  Otherwise, mouse events will be swallowed up. :-(
            result.append(self.interaction)
            
        return result
    
    def values_at ( self, x, y, threshold = 5.0 ):
        """
        Return the list of PlotValue objects whose representation intersects (or 
        lies within 'threshold' of) a specified (x,y) point
        """
        result = []
        if (self.visible and (self._plot_cache is not None) and 
            self.isin_plot( x, y )):
            for renderer, values in self._plot_cache[ RENDER_INFO ]:
                for hit in renderer._hittest( values, x, y, threshold ):
                    if hit not in result:
                        result.append( hit )
        return result
 
    def _objects_at(self, x, y):
        """ Return the canvas specific objects at a location. """
        
        result = []
        if self.interaction is not None:
            result.append(self.interaction)

        return result

    def _can_add ( self, item ):
        """
        Return whether a new object can be added to a canvas:
    
        NOTE: There is no code to prevent a PlotCanvas from being added to itself
            (either directly or indirectly). Doing so will cause a recursion
            exception eventually.
        """
        return isinstance( item, PlotCanvas )
 
    def isin_plot ( self, x, y ):
        """
        Test whether a specified (x,y) point is inside the plot area of the 
        canvas
        """
        bounds = self._plot_bounds
        if bounds is None:
             return False
        px, py, pdx, pdy = bounds
        return ((px <= x < px + pdx) and (py <= y < py + pdy))
    

    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------
 
    def _all_values ( self ):
        "Return a list of all contained PlotValues"
        result = []
        for value_item in self._values:
            result += value_item.item._all_values()
        return result
 
    def _title_info ( self, gc ):
        "Get the starting title information"
        return TitleInfo( gc, self._titles )
 
    def _overlay_info ( self, gc ):
        "Get the starting overlay information"
        return OverlayInfo( gc, self._overlays )

    def _min_size ( self, gc, index_axis = None ):
        "Calculate the minimum size needed to draw the canvas"
 
        # If we are not visible, exit immediately:
        if not self.visible:
            return ( 0, 0 )
        
        # Calculate the margin sizes:
        mx, my = self._margin_size()
 
        # Calculate our title sizes:
        title_info = self._title_info( gc )
 
        # Calculate our overlay sizes:
        overlay_info = self._overlay_info( gc )
 
        # Now calculate the contained PlotValue's title, axis, and plot sizes:
        values        = self._all_values()
        render_values = self._get_values_by_renderer( values )
        process_index = (index_axis is None)
        if process_index:
            index_axis = self.index_axis
        axes_info = AxesInfo( gc, index_axis, render_values, process_index )
        pdx       = self.plot_min_width
        pdy       = self.plot_min_height
        for value_item in values:
            item = value_item.item
            if item.visible:
                title_info.add_titles( gc, item._titles )
                overlay_info.add_overlays( gc, item._overlays )
                dx, dy = item._min_size_plot()
                pdx    = max( pdx, dx )
                pdy    = max( pdy, dy )
 
        # Determine the amount of space required for scroll bars (if any):
        sbdx, sbdy = self._scrollbar_size( index_axis )
 
        # Calculate the minimum size of the required margins, titles, axes and
        # scroll bars:
        rdx = 2 * mx + title_info.dx + axes_info.dx + sbdx
        rdy = 2 * my + title_info.dy + axes_info.dy + sbdy
 
        # Calculate the minimum size of the plotting area:
        pdx = max( pdx, overlay_info.dx + 2 * self.plot_padding_width,
                        title_info.max_dx, axes_info.max_dx )
        pdy = max( pdy, overlay_info.dy + 2 * self.plot_padding_height,
                        title_info.max_dy, axes_info.max_dy )
 
        # Cache the results for use by the _layout routine:
        self._plot_cache = ( title_info, overlay_info, axes_info, render_values, 
                             rdx, rdy, pdx, pdy, mx, my, sbdx, sbdy )
 
        # Return the total minimum size needed:
        return ( max( rdx + pdx, self.min_width ),
                 max( rdy + pdy, self.min_height ) )

    def _layout ( self, x, y, dx, dy ):
        "Layout all of the contents of the canvas"
 
        # If we are not visible, exit immediately:
        if not self.visible:
            return
 
        # Restore the values cached by the _min_size method:
        ( title_info, overlay_info, axes_info, render_values, rdx, rdy, pdx, pdy, 
          mx, my, sbdx, sbdy ) = self._plot_cache
 
        # Add the x,y displacement to create the title origin:
        tx = x + mx
        ty = y + my + sbdy
 
        # Calculate the axes origin:
        ax = tx + title_info.ldx
        ay = ty + title_info.bdy
 
        # Calculate the bottom-left plot origin:
        px = ax + axes_info.ldx
        py = ay + axes_info.bdy
 
        # Remember the screen real estate we have been assigned:
        self.bounds = bounds = ( x, y, dx, dy )
            
        # Calculate the actual size assigned to the plot area:
        pdx = dx - rdx
        pdy = dy - rdy
        
        if self.aspect_ratio != 0.0:
            if pdx > pdy * self.aspect_ratio:
                adjusted_pdx = pdy * self.aspect_ratio
                px += (pdx - adjusted_pdx) / 2
                pdx = adjusted_pdx
            else:
                adjusted_pdy = pdx / self.aspect_ratio
                py += (pdy - adjusted_pdy) / 2
                pdy = adjusted_pdy
 
        # Cache the bounds of the plot area for the _draw method:
        self._plot_bounds = ( px, py, pdx, pdy )
 
        # Calculate the actual size of the axes box:
        adx = pdx + axes_info.dx
        ady = pdy + axes_info.dy
 
        # Calculate the various right and bottom side origins:
        rax = px + pdx
        bay = py + pdy
        rtx = ax + adx + title_info.rdx
        tty = ay + ady + title_info.tdy
        
        # Lay the titles out:
        title_info._layout( ax, ay, adx, ady, tx, tty, rtx, ty )
 
        # Get the x and y axes scaling factors and scrollbars:
        sx, sy, sbx, sby = self._xy_scale_cache = self._xy_scale()
 
        # Lay out the rest of the canvas:
        self._layout_canvas( bounds, px, py, pdx, pdy, sx, sy, 
                             overlay_info, axes_info, render_values )
 
        # Lay the scrollbars out:
        self._layout_scrollbars( sbdx, sbdy, sx, sy, sbx, sby, 
                                  x,  y,  dx,  dy,
                                 px, py, pdx, pdy, mx, my )

    def _layout_canvas ( self, bounds, px, py, pdx, pdy, sx, sy,
                               overlay_info, axes_info, render_values ):
        "Layout the overlays, axes, and plots for the canvas"
        
        # Lay the axes out:
        axes_info._layout( px, py, pdx, pdy, sx, sy )
               
        # Let each unique renderer lay-out its associated PlotValues:
        span_plot_bounds = self._span_plot_bounds
        for renderer, values in render_values:
            for value in values:
                value.bounds            = bounds
                value._span_plot_bounds = span_plot_bounds
            renderer._layout( values, px, py, pdx, pdy, sx, sy )
                         
        # Lay the overlays out:
        # Note: This must be done after the plot areas have been laid out, so
        #       that the _layout method can use the 'PlotValue.map_screen'
        #       method to convert data based overlays to screen coordinates.
        paddx = self.plot_padding_width
        paddy = self.plot_padding_height
        overlay_info._layout( px + paddx,      py + paddy,
                              pdx - 2 * paddx, pdy - 2 * paddy )
                           

    def _draw ( self, gc, draw_selection = True ):
        "Draw the canvas"
        
        # If we are not visible, exit immediately:
        if not self.visible:
            return
 
        # Restore the values cached by the _min_size method:
        try:
            ( title_info, overlay_info, axes_info, render_values, rdx, rdy, 
              pdx, pdy, mx, my, sbdx, sbdy ) = self._plot_cache
        except:
            # fixme: We need to catch the exception in case someone tries to 
            # render a PlotValue separately from its containing PlotCanvas:
            container            = self.container
            cpx, cpy, cpdx, cpdy = self._plot_bounds = container._plot_bounds
            self._plot_ox        = container._plot_ox
            self._plot_oy        = container._plot_oy
            ( title_info, overlay_info, axes_info, render_values, rdx, rdy, 
              pdx, pdy, mx, my, sbdx, sbdy ) = container._plot_cache
            weak_self     = weakref.proxy( self )
            render_values = [ ( renderer, [ weak_self ] )
                              for renderer, values in render_values
                              if weak_self in values ]
            axes_info = AxesInfo( gc, self.index_axis, render_values )
            sx, sy, sbx, sby = container._xy_scale_cache
            axes_info._layout( cpx, cpy, cpdx, cpdy, sx, sy )
            self._plot_cache = ( title_info, overlay_info, axes_info, 
                                 render_values, rdx, rdy, pdx, pdy, mx, my, 
                                 sbdx, sbdy )
 
        # Draw the frame:
        PlotFrame._draw( self, gc )
 
        # Draw all of the titles:
        title_info._draw( gc )
 
        # Draw the plot area background:
        self._draw_plot_frame( gc )
 
        # Get the current draw offset:
        ox = self._plot_ox
        oy = self._plot_oy
 
        # Set up the clipping region:
        clip_bounds = self._plot_bounds
        gc.save_state()
        gc.clip_to_rect(*clip_bounds)
 
        # Draw the 'background' of all of the plots using the appropriate 
        # renderer:
        for renderer, values in render_values:
            renderer._draw_canvas( gc, values, ox, oy )
 
        # Reset the clipping region:
        gc.restore_state()
 
        # Draw all of the axes:
        axes_info._draw( gc, ox, oy )
 
        # Set up the clipping region:
        gc.save_state()
        gc.clip_to_rect(*clip_bounds)
 
        # Draw all of the plots using the appropriate renderer:
        for renderer, values in render_values:
            renderer._draw( gc, values, ox, oy )
         
        # Draw all of the overlays:
        overlay_info._draw( gc, ox, oy )
        overlay_info._draw_plot( gc, ox, oy )
      
        # Draw all of the selections:
        if draw_selection:
                   
            # Widen the clipping region a little to handle the selection
            # boxes 'hanging over the edge':
            gc.restore_state()
            gc.save_state()
            px, py, pdx, pdy = self._plot_bounds
            gc.clip_to_rect(px-4, py-4, pdx+8, pdy+8)
                          
            # Now draw the selections (if any):
            for renderer, values in render_values:
                renderer._draw_selection( gc, values, ox, oy )
    
        # Reset the clipping region:
        gc.restore_state()
 
        # Draw the debug frame:
        self._draw_debug( gc )

    
    def _get_values_by_renderer ( self, values ):
        """
        Return a list of lists of visible PlotValues categorized by renderer:
      
        Note that the categorizing is further refined by the fact that all
        PlotValues must share a common index and axis.
      
        The format of the list returned is:
        [ ( r1, [ v11, v12, ..., v1n ] ), ( [ r2, v21, v22, ..., v2n ] ) ... ]
        where:
        - ri:  Renderer
        - vij: PlotValue
        """
        render_types = {}
        render_list  = []
        for value_item in values:
            value = value_item.item
            if value.visible:
                value_index = value._index()
                value_axis  = value.axis
                for renderer in value._plot_renderers:
                    render_type = '%d_%d_%d' % ( id( renderer ),
                                                 id( value_index ),
                                                 id( value_axis ) )
                    index = render_types.get( render_type, None )
                    if ((index is None) or
                        (len(render_list[index][1]) == renderer._max_values())):
                        render_types[ render_type ] = index = len( render_list )
                        render_list.append( [ renderer, [] ] )
                    render_list[ index ][1].append( value )
                   
        # If a renderer doesn't like the set of values it is given,
        # replace it by the 'invalid' renderer:
        for pair in render_list:
            if not pair[0]._is_valid( pair[1] ):
               pair[0] =  get_renderer_for( 'invalid' ) 
        
        # Return the list of [ renderer, values ] pairs:
        return render_list

    def _interaction_changed(self, old, new):
        """ Handle the interaction being changed. """
        if new is not None:
            new.canvas = self
        if old is not None:
            old.canvas = None
        return


class PlotTitleList ( PlotItemList ):

    def __validate__ ( self, item ):
        if isinstance( item, PlotTitle ):
            return item
        raise PlotError, 'Only instances of PlotTitle can be specified'


class PlotOverlayList ( PlotItemList ):

    def __validate__ ( self, item ):
        if isinstance( item, PlotOverlay ):
            return item
        raise PlotError, 'Only instances of PlotOverlay can be specified'


class PlotCanvasList ( PlotItemList ):

    def __validate__ ( self, item ):
        from plot_value import PlotValue
        if isinstance( item, PlotCanvas ):
            return item
        if type( item ) in sequence_types:
            return PlotValue( item )
        raise PlotError, ( 'Only numeric arrays or lists or instances of '
                           'PlotCanvas or PlotValue can be specified' )


class PlotGroupList ( PlotItemList ):

    def __validate__ ( self, item ):
        from plot_value import PlotValue
        if isinstance( item, Component ):
            return item
        if type( item ) in sequence_types:
            return PlotValue( item )
        raise PlotError, ( 'Only numeric arrays or lists or instances of '
                           'Component can be specified' )


class PlotValueList ( PlotItemList ):

    def __validate__ ( self, item ):
        raise PlotError, 'The list is read-only'

# EOF
