#-------------------------------------------------------------------------------
#
#  Define a PlotComponent class that allows Chaco objects to be imbedded within
#  a larger Enable window. The PlotComponent acts as a bridge or buffer between 
#  the Chaco and Enable worlds. In particular it defines a Chaco DefaultContainer
#  that all Chaco components contained within it can delegate trait values to.
#  It contains a single Chaco 'component' (such as a PlotGroup) that can be 
#  used to contain larger groups of Chaco objects.
#
#  Written by: David C. Morrill
#
#  Date: 10/02/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#  Classes defined: PlotComponent
#
#-------------------------------------------------------------------------------


# Enthought library imports.
from enthought.traits.api           import Instance, List
from enthought.enable.base_container import default_container
from enthought.enable.component import Component
from enthought.enable.scrollbar import ScrollBar

# Local imports.
from plot_base                  import NEEDS_MIN_SIZE, NEEDS_LAYOUT, RESIZE
from plot_container             import DefaultContainer
from plot_traits                import HORIZONTAL
from window_surrogate           import WindowSurrogate


class PlotComponent ( Component, DefaultContainer ):
    """ A container to allow Chaco objects to play well with Enable.

    The current Chaco model still has traits being delegated up the containment
    hierarchy. If the plot-specific traits fail if the delegation runs past the
    last Chaco component. Hence PlotComponent, which does not delegate those
    traits to Enable.
    """
    
    # The Chaco plot component contained in this component:
    component = Instance( Component )
    
    # An optional component overlaid on top of the Chaco component:
    overlay = Instance( Component )

    # A set of comoponents to be overlaid on top of the Chaco component:
    overlays = List( Component )

    # Ask for keyboard events from Enable.
    accepts_focus = True
    

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    
    def __init__ ( self, component = None, **traits ):
        """ Create a new PlotComponent.

        component: the top-level Chaco component to wrap.
        """
        
        self.host        = WindowSurrogate( self )
        self._scrollbars = {}
        if component is None:
            component = Component()
        self.component = component
        self.min_width = self.min_height = 128.0
        Component.__init__( self, **traits )

        return


    ###########################################################################
    # Public 'Component' interface.
    ###########################################################################

    def components_at ( self, x, y ):
        """ Return the components that contain a specified (x,y) point. """
        
        result = Component.components_at( self, x, y )
        if len( result ) > 0:
            result.extend( self.component.components_at( x, y ) )
            if self.overlay is not None:
                result.extend( self.overlay.components_at( x, y ) )
            for sb in self._scrollbars.values():
                result.extend( sb.components_at( x, y ) )
            for overlay in self.overlays:
                result.extend( overlay.components_at( x, y ) )

        return result

    ###########################################################################
    # Protected 'Component' interface.
    ###########################################################################
    
    def _draw (self, gc):
        host = self.host
        flags, host.update_flags = host.update_flags, 0
        component = self.component
        
        if (flags & NEEDS_MIN_SIZE) != 0:
            self.min_width, self.min_height = component._min_size( gc )
           
        scrollbars = self._scrollbars
        # fixme: The following 'True' values forces a layout each time, which
        # seems to fix some errors in Poer Pressure when a view is occuded by
        # another window and then clicked on. The resulting 'update(REDRAW)' 
        # call then fails, because the data supposedly cached by a previous 
        # layout call no longer seems to exist (DCM).
        # (BWD) This does not seem to be a problem 
        if True or ((flags & NEEDS_LAYOUT) != 0):
            for sb in scrollbars.values():
                sb._used = False
            component._layout( *self.bounds )
            for sb_id, sb in scrollbars.items():
                if not sb._used:
                    host.dispose_scrollbar( sb )
                    del scrollbars[ sb_id ]
        
        gc.save_state()
        component._draw( gc )
        gc.restore_state()
        for sb in scrollbars.values():
            sb._draw( gc )
            
        # Draw the overlay (if any):
        if self.overlay is not None:
            self.overlay.draw( gc )

        # Draw the overlays (if any):
        for overlay in self.overlays:
            overlay.draw( gc )

    def _need_layout ( self ):
        """ Called to indicate layout needs to be done. """
        
        pass

    def _scrollbar_for ( self, plottable, style, create = True ):
        """ Create a scrollbar for the specified plottable and direction. """
        
        horizontal = (style == HORIZONTAL)
        pid        = '%d_%s' % ( id( plottable ), [ 'v', 'h' ][ horizontal ] )
        sb         = self._scrollbars.get( pid )
        if (sb is None) and create:
            self._scrollbars[ pid ] = sb = ScrollBar( 
                         style = 'vh'[ horizontal ], container = self )

        return sb
        
    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait handlers. ######################################################

    #### Static ####

    def _bounds_changed ( self, old, new ):
        """ Called whenever the bound trait is changed. """
        
        Component._bounds_changed( self, old, new )
        self.component.bounds = new
        self.component.update( RESIZE )

        return

    def _component_changed ( self, component ):
        """ Called whenever the component trait is changed. """

        # The component is not allowed to be None, so change it back to an
        # empty enable component.
        if component is None:
            self.component = Component()
            return
        
        component.container = self

        return
            
    def _overlay_changed ( self, old, component ):
        """ Called whenever the overlay trait is changed. """
        
        if component is not None:
            component.container = self

        # Mark the old region as stale, so it will be repainted.
        if old is not None:
            old.redraw()

        return

    def _overlays_items_changed ( self, event ):
        """ Called whenever an item is added or removed from overlays. """
        
        for component in event.added:
            component.container = self

        for component in event.removed:
            component.redraw()
            component.container = default_container

        return

#### EOF ######################################################################
