#-------------------------------------------------------------------------------
#
#  Define the base plot container classes
#
#  Written by: David C. Morrill
#
#  Date: 07/10/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

# Import this so we can get a relative path to it.
import plot_base

import os.path
from sys                             import platform

from enthought.enable                import Component
from enthought.enable.base_container import BaseContainer as EnableBaseContainer
from enthought.enable.base_container import DefaultContainer as \
                                            EnableDefaultContainer

from enthought.traits.api import Delegate, HasTraits, Instance, Python, Range, \
                             Str, Trait, TraitPrefixList, \
                             TraitPrefixMap, false
from enthought.traits.ui.api import ImageEnumEditor

from axis_traits import *
from plot_base   import *
from plot_label_style import PlotLabelStyle
from plot_overlay_style import PlotOverlayStyle
from plot_title_style import PlotTitleStyle
from plot_style  import PlotStyle
from plot_traits import BORDER_BEVEL, BORDER_SOLID
from plot_traits import PlotRenderers
from plot_window_base import PlotTraitCommand


#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Plot styles:
PLOT_LINE      = 0
PLOT_SCATTER   = 1
PLOT_HISTOGRAM = 2
PLOT_FILLED    = 3
PLOT_STACKED   = 4
PLOT_SMALLER   = 5
PLOT_NOT_EQUAL = 6
PLOT_LARGER    = 7

# '_plot_cache' index values:
TITLE_INFO   = 0
OVERLAY_INFO = 1
AXES_INFO    = 2
RENDER_INFO  = 3

# 'Right click' styles:
ENABLED         = 1
ALL_OBJECTS     = 2
EDIT_TRAITS = 4

right_click_styles = {
    'ignore':   0,
    'top menu': ENABLED,
    'menu':     ENABLED,
    'tm':       ENABLED,
    'top edit': ENABLED + EDIT_TRAITS,
    'edit':     ENABLED + EDIT_TRAITS,
    'te':       ENABLED + EDIT_TRAITS,
    'all menu': ENABLED + ALL_OBJECTS,
    'am':       ENABLED + ALL_OBJECTS,
    'all edit': ENABLED + EDIT_TRAITS + ALL_OBJECTS,
    'ae':       ENABLED + EDIT_TRAITS + ALL_OBJECTS
}

# Trait sheet editor styles:
editor_styles = [
    'list', 'tabbed'
]

# Range selection types:
SELECT_NONE  = 0
SELECT_INDEX = 1
SELECT_VALUE = 2

select_types = {
    'none':  SELECT_NONE,
    'index': SELECT_INDEX,
    'value': SELECT_VALUE
}

index_axis_values = ['x', 'y']

#-------------------------------------------------------------------------------
#  Convert an image file name to a cached image:
#-------------------------------------------------------------------------------

# Image cache dictionary (indexed by filename):
_image_cache = {}

def image_cache ( object, name, value ):
    if value == '':
        raise TraitError
    filename = os.path.abspath( value )
    image    = _image_cache.get( filename )
    if image is None:
        if not os.path.isfile( filename ):
            raise TraitError
#       image = load_image_from_file_somehow( filename )
        _image_cache[ filename ] = image
    return image

image_cache.info = "an image file name (e.g. a '.gif', '.png' or '.jpeg' file)"

#-------------------------------------------------------------------------------
#  Convert a number into a chaco compatible color:
#-------------------------------------------------------------------------------

def value_to_color ( object, name, value ):
    try:
        if type( value ) in sequence_types:
            if ((len( value ) == 4)      and
                (0.0 <= value[0] <= 1.0) and
                (0.0 <= value[1] <= 1.0) and
                (0.0 <= value[2] <= 1.0) and
                (0.0 <= value[3] <= 1.0)):
                return value
            raise TraitError
        num = int( value )
        return ( ((num / 0x10000)   & 0xFF) / 255.0,
                 ((num / 0x100)     & 0xFF) / 255.0,
                 (num               & 0xFF) / 255.0,
                 1.0 )
    except:
        raise TraitError

value_to_color.info = ('a number, which in hex is of the form 0xrrggbb, where '
                       'rr is red, gg is green, and bb is blue, '
                       'or a sequence of 4 elements ( rr, gg, bb, aa ), where '
                       'aa is alpha, with each element a number in the range '
                       'from 0.0 to 1.0' )

#-------------------------------------------------------------------------------
#  Convert from RGB to HSV color space:
#-------------------------------------------------------------------------------

def rgb_to_hsv ( r, g, b ):
    min_rgb = min( r, g, b )
    max_rgb = max( r, g, b )
    v       = max_rgb
    delta   = max_rgb - min_rgb
    if delta == 0.0:
        return ( 0.0, 0.0, v )
    s = delta / max_rgb
    if r == max_rgb:
        h = ( g - b ) / delta         # between yellow & magenta
    elif g == max_rgb:
        h = 2.0 + ( b - r ) / delta   # between cyan & yellow
    else:
        h = 4.0 + ( r - g ) / delta   # between magenta & cyan
    h *= 60.0                         # degrees
    if h < 0.0:
        h += 360.0
    return ( h, s, v )

#-------------------------------------------------------------------------------
#  Convert from HSV to RGB color space:
#-------------------------------------------------------------------------------

def hsv_to_rgba ( h, s, v ):
    if s == 0.0:
        # Achromatic (grey):
        return ( v, v, v, 1.0 )

    h /= 60.0   # Sector 0 to 5
    i  = int( h )
    f  = h - i   # Factorial part of h
    p  = v * ( 1.0 - s )
    q  = v * ( 1.0 - s * f )
    t  = v * ( 1.0 - s * ( 1.0 - f ) )
    if i == 0:
        return ( v, t, p, 1.0 )
    if i == 1:
        return ( q, v, p, 1.0 )
    if i == 2:
        return ( p, v, t, 1.0 )
    if i == 3:
        return ( p, q, v, 1.0 )
    if i == 4:
        return ( t, p, v, 1.0 )
    return ( v, p, q, 1.0 )

#-------------------------------------------------------------------------------
#  Create a 'lighter' shade of a specified color:
#-------------------------------------------------------------------------------

def lighter_color ( color, scale = 0.4 ):
    h, s, v = rgb_to_hsv( color[0], color[1], color[2] )
    if s == 0.0:
        v += (1.0 - v) * scale
    else:
        s -= s * scale
    return hsv_to_rgba( h, s, v )

#-------------------------------------------------------------------------------
#  Draw a 'fancy' border around a specified area:
#-------------------------------------------------------------------------------

def draw_border ( gc, style, size, color, bg_color, x, y, dx, dy ):
    rx = x + dx - 1
    ty = y + dy - 1
    if style == BORDER_BEVEL:
        edge = color
        if bg_color is not None:
            edge = bg_color
            gc.set_fill_color( bg_color )
            gc.begin_path()
            gc.rect( x + size, y + size - 1, dx - 2 * size, dy - 2 * size )
            gc.fill_path()
        gc.set_stroke_color( color )
        gc.set_line_width( 1 )
        for i in range( size ):
            gc.begin_path()
            gc.move_to( rx - i, ty - i )
            gc.line_to( rx - i,  y + i )
            gc.line_to(  x + i,  y + i )
            gc.stroke_path()
        gc.set_stroke_color( lighter_color( edge, 0.75 ) )
        for i in range( size ):
            gc.begin_path()
            gc.move_to(  x + i,  y + i )
            gc.line_to(  x + i, ty - i )
            gc.line_to( rx - i, ty - i )
            gc.stroke_path()
    else:
       gc.set_stroke_color( color )
       gc.set_line_width( 1 )
       if (dx > (2 * size)) and (dy > (2 * size)):
           for i in range( size - 1 ):
               gc.begin_path()
               gc.move_to( x + size, y + i )
               gc.line_to( rx - i,   y + i )
               gc.line_to( rx - i,   ty - size )
               gc.stroke_path()
       else:
           size = 1
       if bg_color is None:
           mode = STROKE
       else:
           mode = FILL_STROKE
           gc.set_fill_color( bg_color )
       gc.begin_path()
       gc.rect( x, y + size - 1, dx - size, dy - size )
       gc.draw_path( mode )

#-------------------------------------------------------------------------------
#  Trait Editor definitions:
#-------------------------------------------------------------------------------

index_axis_editor       = ImageEnumEditor(
                              cols   = 2,
                              suffix = '_index_axis',
                              values = index_axis_values,
                              module = plot_base )

#-------------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------------

# Miscellaneous traits:
relative_weight_trait = Range( 0.0, 10.0, value = 1.0 )
index_axis_trait      = Trait( index_axis_values, editor=index_axis_editor )
menu_trait          = Str
label_trait         = Str( desc  = 'a brief description of the object',
                           label = 'Description' )
editor_style_trait  = Trait( 'tabbed', TraitPrefixList( editor_styles ) )
image_trait         = Trait( '', { '': None }, image_cache )
select_type_trait   = Trait( 'none', TraitPrefixMap( select_types ) )
select_notify_trait = Trait( None, None, FunctionType, MethodType )
# XXX: scrollbar_trait     = Trait( None, None, wx.wxScrollBar )
scrollbar_trait     = Trait( None, None )
undo_handler_trait  = Trait( PlotUndoHandler(), PlotUndoHandler )


StyleDelegate = Delegate('style')

#-------------------------------------------------------------------------------
#  'DefaultContainer' class:
#-------------------------------------------------------------------------------

class DefaultContainer ( EnableDefaultContainer ):

    # The weight of the canvas's size in relation to other canvas's in the same
    # container.
    weight = Trait( relative_weight_trait,
                    desc = "the weight of the canvas's size relative to other "
                           "canvas's in the same container",
                    label = 'Layout Weight' )

    # The index value for contained plots.
    index = Instance('enthought.chaco.plot_value.PlotValue')

    # The axis the index axis is aligned with.
    index_axis = Trait( index_axis_trait,
                        desc = 'the axis the index axis is aligned with',
                        label = 'Index Axis' )

    # The PlotAxis instance corresponding to the index for this container
    axis_index = Instance('enthought.chaco.PlotAxis', kw={'zoomable': False})

    # The PlotAxis instance corresponding to the value for this container;
    # To prevent this from being a None instance, we pass an empty tuple for ``*args``
    # to force the creation of a new instance.
    axis = Instance('enthought.chaco.PlotAxis', ())

    # The style of editor to display (list or tabbed).
    # FIXME : Verify this is still working with Traits2
    editor_style = Trait( editor_style_trait,
                          desc  = 'the trait editor display style' )

    # An object to handle undo commands.
    undo_handler = Trait( undo_handler_trait,
                          desc  = 'the object that handles undo commands' )

    # The containing window.
    host = ReadOnly

    # The graphics context into which this plot is being drawn.
    # FIXME : AFAIK, this is only used in the plt module.  Perhaps move it out?
    gc = Any

    # FIXME : What is this???
    user__ = Python

    # Draw the plot in debug mode?  In debug mode, the margin and padding areas
    # are drawn with borders.
    debug = false( desc  = 'whether the margin and padding area bounding boxes '
                           'should be drawn',
                   label = 'Show Debug Borders?' )

    # The style sheet for all of the delegated style traits (below).
    style = Instance(PlotStyle, ())

    # The base style sheet for plot labels.
    plot_label_style = Instance(PlotLabelStyle, ())

    # The base style sheet for plot overlays.
    plot_overlay_style = Instance(PlotOverlayStyle, ())

    # The base style sheet for plot titles.
    plot_title_style = Instance(PlotTitleStyle, ())

    #### Delegated traits #####################################################

    # The color of annotations/labels on data points
    plot_annotation_color = StyleDelegate

    # The relative position of a plot annotation
    plot_annotation_position = StyleDelegate

    # Number of pixels of padding between the annotation labels and the plot
    plot_annotation_padding = StyleDelegate

    # The font to use for the annotation text
    plot_annotation_font = StyleDelegate

    # The size of the annotation text
    plot_annotation_size = StyleDelegate

    # For bar charts, the width of a bar.
    plot_bar_width = StyleDelegate

    # The plot background color.
    plot_bg_color = StyleDelegate

    # For bar charts, the width of a group of bars.
    plot_bin_width = StyleDelegate

    # The border color of the plot.
    plot_border_color = StyleDelegate

    # The thickness of the plot border.
    plot_border_size = StyleDelegate

    # The style of the plot border.
    plot_border_style = StyleDelegate

    # For line type plots, the style of the line (e.g. dash).
    plot_line_style = StyleDelegate

    # For line type plots, the color of the line.
    plot_line_color = StyleDelegate

    # For line type plots, the thickness of the line.
    plot_line_weight = StyleDelegate

    # The corner of the plotting area used as the axis origin.
    plot_origin = StyleDelegate

    # For plots with markers, the outline color of the marker.
    plot_outline_color = StyleDelegate

    # For plots with markers, the thickness of the outline.
    plot_outline_weight = StyleDelegate

    # For plots with markers, the fill color of the marker.
    plot_fill_color = StyleDelegate

    # For plots with markers, the fill style of the marker.
    plot_fill_style = StyleDelegate

    # The minimum height of any contained plot.
    plot_min_height = StyleDelegate

    # The minimum width of any contained plot.
    plot_min_width = StyleDelegate

    # The padding inside any contained plot.
    plot_padding = StyleDelegate

    # The padding inside any contained plot above and below the content.
    plot_padding_height = StyleDelegate

    # The padding inside any contained plot left and right of the content.
    plot_padding_width = StyleDelegate

    # For plots with markers, the size of the marker.
    plot_size = StyleDelegate

    # For plots with markers, the symbol (e.g. circle).
    plot_symbol = StyleDelegate

    # The type of plot drawn.
    plot_type = StyleDelegate

    #### Private traits #######################################################

    # The renderers used to plot the contained plots.
    _plot_renderers = PlotRenderers


    #---------------------------------------------------------------------------
    # The following traits are really defined later, since they require
    # classes which haven't been defined yet:
    #---------------------------------------------------------------------------

#   axis                   = axis_trait
#   axis_index             = axis_trait
#   index                  = index_trait

#-------------------------------------------------------------------------------
#  'SingletonDefaultContainer' class:
#-------------------------------------------------------------------------------

class SingletonDefaultContainer ( DefaultContainer ):

    #---------------------------------------------------------------------------
    #  Trait definition:
    #---------------------------------------------------------------------------

    window = ReadOnly

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, **traits ):
        HasTraits.__init__( self, **traits )

        # Allow every object to locate our window and host:
        self.window = self.host = self

    #---------------------------------------------------------------------------
    #  Dummy method implementations for Enable:
    #---------------------------------------------------------------------------

    def redraw ( self, bounds = None, coordinates = None ):
        pass

    def _needs_redraw ( self, bounds ):
        return False

    def add ( self, *components ):
        pass

    def remove ( self, *components ):
        pass

    def _set_pointer ( self, pointer ):
        pass

    def _on_mouse_move ( self, event ):
        pass

    def _set_timer_interval ( self, component, interval ):
        pass

    #---------------------------------------------------------------------------
    #  Dummy method implementations for Chaco:
    #---------------------------------------------------------------------------

    def _add_item ( self, container, object ):
        pass

    def _remove_item ( self, container, object ):
        pass

    def update ( self, update_flags = UPDATE ):
        pass

    def _need_layout ( self ):
        pass

# Create the default container for other objects:
default_container = SingletonDefaultContainer()

#-------------------------------------------------------------------------------
#  'IsContained' class:
#-------------------------------------------------------------------------------

class IsContained ( Component ):

    #---------------------------------------------------------------------------
    #  Trait definition:
    #---------------------------------------------------------------------------

    container      = Trait( default_container, EnableBaseContainer,
                            desc = 'the object containing this object' )
    selected       = false( desc = 'whether the select markers should be '
                                   'drawn' )
    use_min_width  = false( desc = 'whether the min_width trait determines '
                            'the minimum width or it should be calculated' )
    use_min_height = false( desc = 'whether the min_height trait determines '
                            'the minimum height or it should be calculated' )
    host           = simple_delegate
    undo_handler   = simple_delegate
    editor_style   = simple_delegate
    label          = label_trait

    # Set up generic, default trait definitions:
    user__         = simple_delegate

    #---------------------------------------------------------------------------
    #  Delete an object from its container:
    #---------------------------------------------------------------------------

    def delete ( self ):
        self.container.remove( self )

    #---------------------------------------------------------------------------
    #  Allow 'edit' to be called directly from a pop-up menu:
    #---------------------------------------------------------------------------

    def menu_edit ( self, event ):
        self.edit()

    #---------------------------------------------------------------------------
    #  Allow the user to edit a trait (or traits) of an object:
    #---------------------------------------------------------------------------

#    def edit ( self, traits = None ):
#        if traits is None:
#            traits = self._editable_traits()
#        self.host._trait_sheet_for( self, traits )

    #---------------------------------------------------------------------------
    #  Return the editable traits for the object:
    #---------------------------------------------------------------------------

#    def _editable_traits ( self ):
#        try:
#            return self.__editable_traits__
#        except:
#            return self.__traits__.keys()

    #---------------------------------------------------------------------------
    #  Update the object's window:
    #---------------------------------------------------------------------------

    def update ( self, update_flags = UPDATE ):
        self.host.update( update_flags )


    #--------------------------------------------------------------------------
    #  Handle any trait being changed:
    #--------------------------------------------------------------------------

    def _anytrait_changed ( self, trait_name, old, new ):
        pass

    #---------------------------------------------------------------------------
    #  Notification that a trait sheet has modified a trait of its
    #  associated object:
    #---------------------------------------------------------------------------

    def trait_changed ( self, trait_name, old_value, new_value ):
        self.host.update()   ###KW_TODO###
        self.undo_handler.add_undo(
            PlotTraitCommand( self, trait_name, old_value, new_value ) )

    #---------------------------------------------------------------------------
    #  Draw the 'selection' for an object:
    #---------------------------------------------------------------------------

    def _draw_selection ( self, gc, xl, yb, dx, dy ):

        # Draw the outer selection rectangle:
        gc.begin_path()
        gc.set_line_width( 1 )
        gc.set_stroke_color( green_color )
        gc.rect( xl, yb, dx - 1, dy - 1 )
        gc.stroke_path()

        gc.begin_path()
        gc.set_stroke_color( black_color )
        gc.set_line_dash( short_dash_line_style )
        gc.rect( xl, yb, dx - 1, dy - 1 )
        gc.stroke_path()

        # Set up the colors and line style to use for the selection boxes:
        gc.set_fill_color( green_color )
        gc.set_line_dash( solid_line_style )

        # Calculate the selection box center coordinates:
        xc = xl + dx / 2
        yc = yb + dy / 2

        # Draw each of the eight selection boxes:
        for x in ( xl, xc, xl + dx - 1 ):
            for y in ( yb, yc, yb + dy - 1 ):
                if (x != xc) or (y != yc):
                    gc.begin_path()
                    gc.rect( x - 3, y - 3, 6, 6 )
                    gc.draw_path()

    #---------------------------------------------------------------------------
    #  Return the GUI description of the object:
    #---------------------------------------------------------------------------

    def _ui_label ( self ):
        return self.label or self.__class__.__name__[4:]

#-------------------------------------------------------------------------------
#  'NestedContainer' class:
#-------------------------------------------------------------------------------

class NestedContainer ( IsContained, EnableBaseContainer ):

    __prefix__ = 'canvas_'

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Delegate these trait classes to the container:
    canvas__   = simple_delegate
    panel__    = simple_delegate
    label__    = simple_delegate
    overlay__  = simple_delegate
    title__    = simple_delegate
    plot__     = simple_delegate
    axis__     = simple_delegate

    # Plot related traits:
    index      = simple_delegate
    index_axis = simple_delegate
    axis       = simple_delegate
    axis_index = simple_delegate

    # Minimum container size and weight traits:
    weight     = simple_delegate

    # Miscellaneous traits:
    debug      = simple_delegate

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, **traits ):
        IsContained.__init__( self, **traits )

    #---------------------------------------------------------------------------
    #  Add a new item to the container:
    #---------------------------------------------------------------------------

    def _add_item ( self, list, item ):
        item.container.remove( item )
        list.append( PlotItem( item ) )
        item.container = self
        self._need_layout()
        self.host._add_item( self, item )   ###KW_TODO###

    #---------------------------------------------------------------------------
    #  Remove an item from the container:
    #---------------------------------------------------------------------------

    def _remove_item ( self, list, item ):
        for i in range( len( list ) ):
            if item is list[i].item:
               del list[i]
               del item.container
               self._need_layout()
               self.host._remove_item( self, item )   ###KW_TODO###
               return True
        return False

    #---------------------------------------------------------------------------
    #  Indicate that the group needs to be laid out:
    #---------------------------------------------------------------------------

    def _need_layout ( self ):
        self.container._need_layout()

    #---------------------------------------------------------------------------
    #  Handle the 'axis' trait changing:
    #---------------------------------------------------------------------------

    def _axis_changed ( self, old, axis ):
        axis.container = self

    #---------------------------------------------------------------------------
    #  Handle the 'axis_index' trait changing:
    #---------------------------------------------------------------------------

    def _axis_index_changed ( self, old, axis_index ):
        if axis_index is not None:
           axis_index.container = self

    #---------------------------------------------------------------------------
    #  Handle the 'index' trait changing:
    #---------------------------------------------------------------------------

    def _index_changed ( self, old, index ):
        if index is not None:
           index.container = self

    #---------------------------------------------------------------------------
    #  Handle the 'canvas_margin' trait value changing:
    #---------------------------------------------------------------------------

    def _canvas_margin_changed ( self, canvas_margin ):
        self.canvas_margin_height = self.canvas_margin_width = canvas_margin

    #---------------------------------------------------------------------------
    #  Handle the 'canvas_padding' trait value changing:
    #---------------------------------------------------------------------------

    def _canvas_padding_changed ( self, canvas_padding ):
        self.canvas_padding_height = self.canvas_padding_width = canvas_padding

    #---------------------------------------------------------------------------
    #  Handle the 'label_margin' trait value changing:
    #---------------------------------------------------------------------------

    def _label_margin_changed ( self, label_margin ):
        self.label_margin_height = self.label_margin_width = label_margin

    #---------------------------------------------------------------------------
    #  Handle the 'label_padding' trait value changing:
    #---------------------------------------------------------------------------

    def _label_padding_changed ( self, label_padding ):
        self.label_padding_height = self.label_padding_width = label_padding

    #---------------------------------------------------------------------------
    #  Handle the 'overlay_margin' trait value changing:
    #---------------------------------------------------------------------------

    def _overlay_margin_changed ( self, overlay_margin ):
        self.overlay_margin_height = self.overlay_margin_width = overlay_margin

    #---------------------------------------------------------------------------
    #  Handle the 'overlay_padding' trait value changing:
    #---------------------------------------------------------------------------

    def _overlay_padding_changed ( self, overlay_padding ):
        self.overlay_padding_height = self.overlay_padding_width = \
             overlay_padding

    #---------------------------------------------------------------------------
    #  Handle the 'title_margin' trait value changing:
    #---------------------------------------------------------------------------

    def _title_margin_changed ( self, title_margin ):
        self.title_margin_height = self.title_margin_width = title_margin

    #---------------------------------------------------------------------------
    #  Handle the 'title_padding' trait value changing:
    #---------------------------------------------------------------------------

    def _title_padding_changed ( self, title_padding ):
        self.title_padding_height = self.title_padding_width = title_padding

    #---------------------------------------------------------------------------
    #  Handle the 'plot_padding' trait value changing:
    #---------------------------------------------------------------------------

    def _plot_padding_changed ( self, plot_padding ):
        self.plot_padding_height = self.plot_padding_width = plot_padding

    #---------------------------------------------------------------------------
    #  Draw a plot frame:
    #---------------------------------------------------------------------------

    def _draw_plot_frame ( self, gc ):
        bg_color    = self.plot_bg_color_
        border_size = self.plot_border_size

        # If the frame is invisible, do a quick exit:
        if (border_size == 0) and (bg_color is None):
            return

        # Get the device context to use:
        border_color = self.plot_border_color_
        border_style = self.plot_border_style_

        px, py, pdx, pdy = self._plot_bounds
        if (border_size > 1) and (border_style != BORDER_SOLID):
            draw_border( gc, border_style, border_size,
                             border_color, bg_color, px, py, pdx, pdy )
        else:
            # Get the pen to use:
            if border_size == 0:
                draw_method = gc.fill_path
                gc.set_fill_color( bg_color )
            else:
                if bg_color is None:
                    draw_method = gc.stroke_path
                else:
                    draw_method = gc.draw_path
                    gc.set_fill_color( bg_color )
                gc.set_stroke_color( border_color )
                gc.set_line_width( border_size )
                if border_size > 1:
                    gc.set_line_join( JOIN_MITER )

            # Now draw the rectangle:
            gc.begin_path()
            gc.rect( px, py, pdx - border_size, pdy - border_size )
            draw_method()

#### EOF ######################################################################
