""" A data source for PlotValues used as an alternative to Numeric arrays. """

#### TODO : Figure out which of the methods are really private!

#------------------------------------------------------------------------------
#
#  Define the plot data source class that can be be used as an alternative data
#  source for PlotValues (in addition to Numeric arrays)
#
#  Written by: David C. Morrill
#
#  Date: 02/14/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#  Classes defined: PlotData
#
#------------------------------------------------------------------------------


# Enthought library imports.
from enthought.traits.api import HasTraits, List

# Local imports.
from plot_source import PlotSource
from plot_window_base import PlotTraitCommand


class PlotData ( HasTraits ):
    """ A data source for PlotValues used as an alternative to Numeric arrays.

    A PlotData is a collection of one or more channels of related data. The
    channels are not required to be statically sized (as a Numeric array), so
    the PlotData is a good for plotting dynamic data.

    When added to a PlotValue, the channels of a PlotData may be added en masse
    or individually by name.
    
    An example of a PlotData is historical stock data for an individual stock.
    The channels might be the high, low, open, close, volume, etc. Each channel
    individually to a line or bar chart or the could be combined to render
    candlestick or stock bar charts.
    """

    #### 'PlotData' interface #################################################

    # The list of names of sources (channels) provided by this PlotData.
    sources = List([ 'default' ])


    ###########################################################################
    # 'PlotData' interface.
    ###########################################################################

    def data ( self, source_name ):
        """ Return the raw data for the named source (channel).

        This method must be overridden by a subclass.

        source_name : string : the name of the channel whose data to return

        returns : SequenceType : the data for the named source (channel).
        
        """
        
        raise NotImplementedError()
    
    def add_to ( self, plottable, source_name, **keywords ):
        """ Add one or more source connections to a plottable object.

        If the plottable handed in is a container, create a PlotValue to plot
        the source and add that to the container. Otherwise, add the named
        source directly to the PlotValue.

        plottable : Plottable : the Plottable to add the data source to as
                                described above.
        source_name : string : the name of the channel to add to the Plottable.
        keywords : any traits : additional traits to set on the plottable or
                                new PlotValue.

        returns : PlotValue : the PlotValue connected to the source.
        """

        # Check if this is a container.  If so, create a PlotValue for the new
        # source.
        from plot_value import PlotValue
        if not isinstance( plottable, PlotValue ):
            value = PlotValue()
            plottable.add( value )
            plottable = value
        plottable.set( **keywords )
        self.get_source( plottable, source_name or 'default' )

        return plottable

    def get_source ( self, value, source_name = 'default' ):
        """ Get a new source and connect it to the value.

        value : PlotValue : the PlotValue to connect to the source.
        source_name : string : the name of the channel used for the source.

        returns : PlotSource : the new PlotSource for the named channel.
        """
        
        return self.source( value, source_name )


    def add ( self, plottable_object, source_name = None ):
        """ Add one or more source connections to a Plottable object.

        plottable_object : Plottable : the Plottable where sources will be
                                       connected.
        source_name : string : the channel name to connect or None for all.

        returns : None
        """
        
        from plottable import Plottable
        if not isinstance( plottable_object, Plottable ):
            raise PlotError, ( 
                  'Can only add to a Plottable object from a PlotData object' )
        self.add_to( plottable_object, source_name )

        return 
     
    def update ( self ):
        """ Refresh the data of each connected source.

        The data of each connected PlotSource is updated. In turn, that results
        in updates of the view of each source.
        """
        
        for source in self._connections():
            source.update()

        return
    
    ####
    #### FIXME : Don't know what this is for... perhaps a Traits1ism for
    ####         specifying which traits are editable?
    ####
##     def _trait_names ( self ):
##         return [ x for x in HasTraits._trait_names( self ) 
##                  if (x[:1] != '_') and (x != 'source') ]
    

    ###########################################################################
    # Protected 'PlotData' interface.
    ###########################################################################

    def source ( self, value, source_name = None ):
        """ Create a PlotSource for the named channel.

        This may be overridden by subclasses if they wish to return specialized
        subclasses of PlotSource or have any additional control over how the
        source is connected.

        value : PlotValue : optionally, the PlotValue to connect to the source.
        source_name : string : the name of the channel used for the source.

        returns: PlotSource : a PlotSource for the named channel.
        """

        ### FIXME : right now this method effectively has two signatures:
        ###   source(source_name) for an unconnected source.
        ###   source(value, source_name) for a connected source.
        ### The problem is this makes the variable names here a bit wonky.
        
        if source_name is None:
            value, source_name = None, value

        return PlotSource( value, self, source_name )
            
    ## Called by PlotSource constructor...
    def add_source ( self, source ):
        """ Add a source to the set of sources to be notified on updates.

        When a PlotData has a trait change which changes the values, all
        PlotSources connected to plots need to be notified.  This method
        registers a PlotSource for notification.

        source : PlotSource : the source to register

        returns : None
        """
        
        self._connections().append( source )

        return
    
    def remove ( self, source ):
        """ Remove a source from the sources receiving updates. """
        
        try:
            self._sources.remove( source )
        except:
            pass
 
    
    ###########################################################################
    # Private 'PlotData' interface.
    ###########################################################################
        
    def _connections ( self ):
        """ Return the list of PlotSources this PlotData is connected to. """
        
        if not hasattr( self, '_sources' ):
            self._sources = []

        return self._sources


    #### Trait handlers #######################################################
    
    def _update_changed ( self, name ):
        """ Called when any 'update' traits change.

        Traits which when changed should require updates to any attached plots
        should be declared with event='update' as metadata. This will cause
        this handler to be called automatically when the trait changes. An
        example of this would be a subclass wrapping stock quotes. If the
        ticker symbol is stored in a trait called 'symbol', then changing
        the ticker should cause a change in the data -- and thus, any attached
        plots.
        """
        
        self.update()

        return
    
    
#------------------------------------------------------------------------------
#  'PlotDataSimple' class:
#------------------------------------------------------------------------------

class PlotDataSimple ( PlotData ):
    
    #--------------------------------------------------------------------------
    #  Save all constructor traits so that they can be applied to any PlotValue 
    #  the object is bound to:
    #--------------------------------------------------------------------------
 
    def __init__ ( self, **traits ):
        PlotData.__init__( self )
        self._traits = traits
    
    #--------------------------------------------------------------------------
    #  Get a new source:
    #--------------------------------------------------------------------------
    
    def get_source ( self, value, source_name = 'default' ):
        value.set( **self._traits )
        return self.source( value, source_name )
  
#### EOF ######################################################################
