"""
Defines PlotFrame base class
"""

# Enthought library imports.
from enthought.traits.api import Delegate, HasTraits, Instance, Trait
from enthought.kiva.constants import JOIN_MITER


# Local imports.
from plot_container import blue_color, dash_line_style, dot_line_style, \
                           draw_border, image_trait, red_color, \
                           solid_line_style
from plot_traits import BORDER_SOLID
from plot_frame_style import PlotFrameStyle


StyleDelegate = Delegate('style')


class PlotFrame ( HasTraits ):
    """
    PlotFrame is the base class for labels, panels, and all plottable objects.
    """
    
    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    bg_image       = Trait( image_trait, desc = 'the background image' )

    style    = Instance(PlotFrameStyle, ())
    
    bg_color       = StyleDelegate
    border_color   = StyleDelegate
    border_size    = StyleDelegate
    border_style   = StyleDelegate
    margin         = StyleDelegate
    margin_height  = StyleDelegate
    margin_width   = StyleDelegate
    padding        = StyleDelegate
    padding_height = StyleDelegate
    padding_width  = StyleDelegate
    menu           = StyleDelegate


    #---------------------------------------------------------------------------
    # Trait event handlers
    #---------------------------------------------------------------------------
 
    def _margin_changed ( self, margin ):
        self.margin_height = self.margin_width = margin
 
 
    def _padding_changed ( self, padding ):
        self.padding_height = self.padding_width = padding


    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------
 
    def _isin ( self, x, y, cache = None ):
        """
        Test whether a specified (x,y) point is inside of the frame
        """
        if not self.visible:
            return False
       
        mx               = self.margin_width
        my               = self.margin_height
        fx, fy, fdx, fdy = (cache or self).bounds
        return (((fx + mx) <= x < (fx + fdx - mx)) and
                ((fy + my) <= y < (fy + fdy - my)))
 
 
    def objects_at ( self, x, y ):
        """
        Return the list of objects contained in the frame that contain a
        specified (x,y) point.
        
        TODO: Thus method just returns 'self' if (x,y) is inside the frame.
        """
        if self._isin( x, y ):
           return [ self ]
        return []
    
 
    def _draw ( self, gc, ox = 0, oy = 0, cache = None ):
        bg_color = self.bg_color_
        bs       = self.border_size
 
        # If the frame is invisible, do a quick exit:
        bitmap = self.bg_image_
        if (bs == 0) and (bg_color[3] == 0) and (bitmap is None):
           return
       
        mx               = self.margin_width
        my               = self.margin_height
        border_color     = self.border_color_
        border_style     = self.border_style_
        fx, fy, fdx, fdy = (cache or self).bounds
 
        if (bs > 1) and (border_style != BORDER_SOLID):
            draw_border( gc, border_style, bs,
                             border_color, bg_color,
                             fx + mx - ox, fy + my - oy,
                             fdx - 2 * mx, fdy - 2 * my )
        else:
             if bs > 1:
                 gc.set_line_join( JOIN_MITER )
             gc.set_line_width( bs )
             if bg_color is None:
                 gc.set_fill_color( ( 0.0, 0.0, 0.0, 0.0 ) ) # transparent
             else:
                 gc.set_fill_color( bg_color )
             gc.set_stroke_color( border_color )
         
             # Now draw the rectangle:           
             # Note: chaco does not seem to calculate the x and y line weight
             # offsets the same, hence the slightly different offset values in
             # the following call:
             gc.begin_path()   
             gc.rect( fx + mx + bs / 2.0 - ox, fy + my + bs / 2.0 - oy,
                      fdx - 2 * mx - bs,       fdy - 2 * my - bs )
             gc.draw_path()

 
    def _draw_debug ( self, gc, ox = 0, oy = 0, cache = None ):
        # Only draw the debug frame if we are in 'debug' mode:
        if not (self.debug or self.selected):
            return
 
        bs = self.border_size
        mx = self.margin_width
        my = self.margin_height
        px = self.padding_width
        py = self.padding_height
        fx, fy, fdx, fdy = (cache or self).bounds
 
        if self.debug:
            # Draw the inner padding bounding box:
            gc.set_stroke_color( blue_color )
            gc.set_line_width( 1 )
            gc.set_line_dash( dash_line_style ) 
         
            gc.begin_path()
            gc.rect( fx  + mx + bs + px - ox,      fy  + my + bs + py - oy,
                     fdx - 2 * (mx + bs + px) - 1, fdy - 2 * (my + bs + py) - 1)
            gc.stroke_path()
         
            # Draw the outer margin bounding box:
            gc.set_stroke_color( red_color )
            gc.set_line_dash( dot_line_style ) 
         
            gc.begin_path()
            gc.rect( fx - ox, fy - oy, fdx - 1, fdy - 1 )
            gc.stroke_path()
         
            # Restore normal 'solid' line dash style:
            gc.set_line_dash( solid_line_style ) 
           
        # Draw the selection marker (if requested):
        if self.selected:
            self._draw_selection( gc, fx - ox, fy - oy, fdx, fdy )

 
    def _margin_size ( self ):
        """
        Calculate the horizontal and vertical margin sizes
        """
        border_size = self.border_size
        return ( self.margin_width  + border_size + self.padding_width,
                 self.margin_height + border_size + self.padding_height )

# EOF
