

""" A group of plots arranged in a grid. """

# Enthought library imports.
from enthought.enable.base import empty_rectangle
from enthought.traits.api import Int

# Local imports.
from axes_info import AxesInfo
from overlay_info import OverlayInfo
from plottable import Plottable
from plot_canvas import PlotCanvas
from plot_container import RENDER_INFO
from plot_frame import PlotFrame
from plot_lists import PlotCanvasList, PlotOverlayList, PlotTitleList
from title_info import TitleInfo

class PlotGrid(Plottable):
   """ A group of plots arranged in a grid.

   The defining characteristic of a PlotGrid is that for any given row all of
   the horizontal bounds of the plot areas are aligned.  Likewise, for any
   given column the vertical bounds of the plot areas are aligned.
   """

   # The number of rows in the grid.
   rows = Int(2)

   # The number of columns in the grid.
   columns = Int(2)

   ##########################################################################
   # 'object' interface.
   ##########################################################################

   def __init__(self, **traits):
       """ Creates a new plot grid. """

       # Base class constructor.
       super(Plottable, self).__init__(**traits)

       # Initialize the collection of titles.
       self._titles   = []
       self.titles    = PlotTitleList(   self, self._titles )

       # Initialize the collection of overlays.
       self._overlays = []
       self.overlays  = PlotOverlayList( self, self._overlays )

       # Initialize the collection of plot values.
       self._values   = []
       self.values    = PlotCanvasList(  self, self._values )

       return

   ##########################################################################
   # Protected 'Plottable' interface.
   ##########################################################################

   def _can_add(self, item):
       """ Test if the item can be added to the plot grid.

       A PlotGrid only supports PlotCanvas objects.
       """

       ### FIXME Might also want to check that we don't have too many items
       ###       to plot.
       return isinstance(item, PlotCanvas)

   def _objects_at(self, x, y):
       """ Return the list of values at the point (x,y). """

       result = []

       for value_item in self._values:
           result = value_item.item.objects_at(x,y)
           if result:
               break

       return result

   def _min_size(self, gc):
       """ Calculate the minimum size needed to draw the plot grid.

       As a side effect, cache some of the information so that it can be
       used again in the _layout and _draw routines.
       """

       # Calculate the title sizes.
       title_info = TitleInfo(gc, self._titles)

       # Calculate the overlay sizes.
       overlay_info = OverlayInfo(gc, self._overlays)

       mdx = mdy = 0

       # The maximum space needed by any of the children to plot their outside
       # regions (titles, margins, and scrollbars) on the top and the bottom.
       top_dy = bottom_dy = 0

       # The maximum space needed by any of the children to plot their outside
       # regions (titles, margins, and scrollbars) on the left and right.
       left_dx = right_dx = 0

       # The minimum height that will allow any of the children to plot itself.
       child_inside_height = 0
       child_top_height = 0
       child_bottom_height = 0
       child_margin_height = 0

       # The minimum width that will allow any of the children to plot itself.
       child_inside_width = 0
       child_left_width = 0
       child_right_width = 0
       child_margin_width = 0

       # Iterate over each of the children and roll up the sizing information.
       for value_item in self._values:
           plot_item = value_item.item

           # Ask the child what it thinks its sizings are.
           #   The interesting data points are:
           #     pdx: The minimum width the child needs to plot its inside.
           #     pdy: The minimum height the child needs to plot its inside.
           #     mx:  The horizontal margin of the child.
           #     my:  The vertical margin of the child.
           #     sbdx:  The minimum width the child needs for scroll bars.
           #     sbdy:  The minimum height the child needs for scroll bars.
           plot_item._min_size(gc)
           ( titles, overlays, axes, render_values, trdx, trdy, pdx, pdy,
             mx, my, sbdx, sbdy ) = plot_item._plot_cache

           # Roll up the heights.
           child_inside_height = max(child_inside_height, pdy)
           child_margin_height = max(child_margin_height, my)
           child_top_height = max(child_top_height, titles.tdy + axes.tdy)
           child_bottom_height = max(child_bottom_height, titles.bdy + axes.bdy)

           # Roll up the widths.
           child_inside_width = max(child_inside_width, pdx)
           child_margin_width = max(child_margin_width, mx)
           child_left_width = max(child_left_width, titles.ldx + axes.ldx)
           child_right_width = max(child_right_width, titles.rdx + axes.rdx)


       # The total width required to plot a single child.
       child_width = child_inside_width + child_left_width + \
                     child_right_width + 2 * child_margin_width

       # The total width required to plot a row of children.
       plot_width = self.columns * child_width

       # The total height required to plot a single child.
       child_height = child_inside_height + child_top_height + \
                      child_bottom_height + 2 * child_margin_height

       # The total height required to plot a column of children.
       plot_height = self.rows * child_height

       # Calculate the margin sizes.
       margin_x, margin_y = self._margin_size()

       # Get the scroll bar size.
##        scrollbar_sizes = self._scrollbar_size(index_axis)
##        scrollbar_dx = scrollbar_sizes[0]
##        scrollbar_dy = scrollbar_sizes[1]

       scrollbar_dx = scrollbar_dy = 0

       # The minimum size required by margins, titles, and scroll bars.
       outside_dx = 2 * margin_x + title_info.dx + scrollbar_dx
       outside_dy = 2 * margin_y + title_info.dy + scrollbar_dy

       # The minimum size required to plot the inside region.
       inside_dx = max(plot_width, overlay_info.dx, title_info.max_dx)
       inside_dy = max(plot_height, overlay_info.dy, title_info.max_dy)

       # Cache the results for use by the _layout routine:
       self._plot_cache = (title_info, overlay_info, None,
                           margin_x, margin_y, scrollbar_dx, scrollbar_dy,
                           outside_dx, outside_dy, inside_dx, inside_dy,
                           child_inside_height, child_margin_height,
                           child_top_height, child_bottom_height,
                           child_inside_width, child_margin_width,
                           child_left_width, child_right_width)

       return (max(inside_dx + outside_dx, self.min_width),
               max(inside_dy + outside_dy, self.min_height))
   

   def _layout(self, x, y, dx, dy):
       """ Layout the contents of the plot grid.

       x and y locate the origin of the plot region.
       dx and dy are the width and height of the plot region respectively.
       """

       # Restore the information cached by the _min_size method.
       (title_info, overlay_info, axes_info,
        margin_x, margin_y, scrollbar_dx, scrollbar_dy,
        outside_dx, outside_dy, inside_dx, inside_dy,
        child_inside_height, child_margin_height,
        child_top_height, child_bottom_height,
        child_inside_width, child_margin_width,
        child_left_width, child_right_width) = self._plot_cache


       # Add the x,y displacement to create the title origin.
       title_left = x + margin_x
       title_right = x + dx - margin_x
       title_top = y + dy - margin_y
       title_bottom = y + margin_y + scrollbar_dy
       
       # Calculate the inner canvas area origin.
       inner_canvas_x = title_left + title_info.ldx
       inner_canvas_y = title_bottom + title_info.bdy
       inner_canvas_width = dx - 2 * margin_x - title_info.dx - scrollbar_dx
       inner_canvas_height = dy - 2 * margin_y - title_info.dy - scrollbar_dy

       # Remember the screen space that we have been given.
       self.bounds = ( x, y, dx, dy )

       # Calculate the unallocated space...
       extra_x = dx - outside_dx - inside_dx
       extra_y = dy - outside_dy - inside_dy

       # ... and allocate it to the children
       child_extra_width = extra_x / self.columns
       child_extra_height = extra_y / self.rows
       child_inside_width += child_extra_width
       child_inside_height += child_extra_height
       
       # Lay out the titles.
       title_info._layout(inner_canvas_x, inner_canvas_y,
                          inner_canvas_width, inner_canvas_height,
                          title_left, title_top, title_right, title_bottom)

       # Lay out the overlays.
       overlay_info._layout(inner_canvas_x, inner_canvas_y,
                            inner_canvas_width, inner_canvas_height)

       # Lay out each child canvas, starting in the lower left of the grid
       # and working left to right, bottom to top.  If we need to change
       # this, we can.
       row = column = 0

       child_total_width = 2 * child_margin_width + child_inside_width + \
                           child_left_width + child_right_width
       child_total_height = 2 * child_margin_height + child_inside_height + \
                            child_top_height + child_bottom_height
       
       for value_item in self._values:
           child = value_item.item
           if child.visible:

               # Restore the layout values cached by the child's _min_size.
               child_title_info = child._plot_cache[0]
               child_overlay_info = child._plot_cache[1]
               child_axes_info = child._plot_cache[2]
               child_render_values = child._plot_cache[3]
               
               # Let's figure out the bounds of the current child.
               child_origin_x = inner_canvas_x + column * child_total_width
               child_origin_y = inner_canvas_y + (self.rows - 1 - row) * child_total_height

               # The outermost bounds of the current child.
               child.bounds = (child_origin_x, child_origin_y,
                               child_total_width, child_total_height)

               child_inside_origin_x = child_origin_x + \
                                       child_margin_width + \
                                       child_left_width

               child_inside_origin_y = child_origin_y + \
                                       child_margin_height + \
                                       child_bottom_height

               child._plot_bounds = (child_inside_origin_x,
                                     child_inside_origin_y,
                                     child_inside_width,
                                     child_inside_height)

               child._span_plot_bounds = self._plot_bounds

               child_title_left = child_origin_x + \
                                  child_margin_width

               child_title_right = child_origin_x + \
                                   child_total_width - \
                                   child_margin_width

               child_title_bottom = child_origin_y + \
                                    child_margin_height

               child_title_top = child_origin_y + \
                                 child_total_height - \
                                 child_margin_height

               child_title_width = child_total_width - \
                                   2 * child_margin_width - \
                                   child_title_info.dx

               child_title_height = child_total_height - \
                                    2 * child_margin_height - \
                                    child_title_info.dy

               # Lay out the child's titles.
               child_title_info._layout(child_title_left + child_title_info.ldx,
                                        child_title_bottom + child_title_info.bdy,
                                        child_title_width, child_title_height,
                                        child_title_left,
                                        child_origin_y + child_total_height - child_margin_height,
                                        child_origin_x + child_total_width - child_margin_width,
                                        child_title_bottom
                                        )

               # Lay out the child's canvas.
               child._layout_canvas(child.bounds,
                                    child_inside_origin_x,
                                    child_inside_origin_y,
                                    child_inside_width,
                                    child_inside_height,
                                    1, # No scaling.
                                    1, # No scaling.
                                    child_overlay_info,
                                    child_axes_info,
                                    child_render_values)
               
               # Finish laying the scrollbars out:
               child._layout_scrollbars( 0,
                                         0,
                                         1, # No scaling.
                                         1, # No scaling.
                                         None,
                                         None, 
                                         child_origin_x,
                                         child_origin_y,
                                         child_total_width,
                                         child_total_height,
                                         child_inside_origin_x,
                                         child_inside_origin_y,
                                         child_inside_width,
                                         child_inside_height,
                                         child_margin_width,
                                         child_margin_height )

           # Get the coordinates of the next cell.
           row, column = self._next_cell(row, column)

           # If there's no more room to plot, break out.
           if row < 0:
               break
           
       return
      

   def _draw ( self, gc ):
 
       # If we are not visible, return immediately:
       if not self.visible:
           return
       
       # Restore the information cached by the _min_size method:
       title_info = self._plot_cache[0]
       overlay_info = self._plot_cache[1]
       
       # Draw the frame:
       PlotFrame._draw( self, gc )
       
       # Draw all of the titles:
       title_info._draw( gc )
 
       # Draw all of the plot's:
       for value_item in self._values:
           value_item.item._draw( gc )
           
       # Draw all of the overlays:
       overlay_info._draw( gc )
 
       # Draw the debug frame:
       self._draw_debug( gc )


   ##########################################################################
   # 'PlotGrid' private interface. 
   ##########################################################################
   
   def _next_cell(self, row, column):
       """ Return the tuple of the next row and column. """

       # Move to the next cell.
       column += 1
       
       # If we've run out of columns, move to the next row.
       if column == self.columns:
           row += 1
           column = 0
           
           # If we've run out of rows, stop plotting!
           if row == self.rows:
               return (-1, -1)

       return (row, column)
               
                   
#### EOF ######################################################################
