""" A group of plots plotted together. """

# Enthought library imports.
from enthought.enable import Component
from enthought.traits.api import Trait
from enthought.traits.ui.api import Group, View

# Local imports
from overlay_info import OverlayInfo
from plottable import Plottable
from plot_canvas import PlotGroupList, PlotOverlayList, PlotTitleList
from plot_frame import PlotFrame
from plot_traits import Orientation, HORIZONTAL
from plot_views import plot_margin_group, bar_width_group, \
                       layout_group, plot_group, marker_group, \
                       cursor_group, line_group, group_main_group, \
                       border_group, margin_group, misc_group
from title_info import TitleInfo


class PlotGroup ( Plottable ):
    """
    A group of plots plotted together.
    """


    orientation = Orientation(desc  = 'the layout direction for contained objects',
                              label = 'Orientation')

    traits_view = View(
        [ group_main_group,
          [ 'orientation', '_',
            'bg_color',    '_',
            'visible',
            '|[Group]@' ],
          '|{Main}' ],
        layout_group,
        border_group,
        plot_group,
        margin_group,
        cursor_group,
        misc_group
    )
    
    overlay_traits_view = View(
        [ group_main_group,
          [ 'orientation', '_',
            'position',    '_',
            'bg_color',    '_',
            'visible',
            '|[Group]@' ],
          '|{Main}' ],
        layout_group,
        border_group,
        plot_group,
        margin_group,
        cursor_group,
        misc_group
    )

    
    #---------------------------------------------------------------------------
    # Public methods
    #---------------------------------------------------------------------------
    
    def __init__ ( self, *items, **traits ):
        Plottable.__init__( self, **traits )
        self._titles   = []
        self._overlays = []
        self._values   = []
        self.titles    = PlotTitleList(   self, self._titles )
        self.overlays  = PlotOverlayList( self, self._overlays )
        self.values    = PlotGroupList(   self, self._values )
        self.add( *items )

    def clear( self ):
        for title in self.titles:
            self.remove(title)
        for overlay in self.overlays:
            self.remove(overlay)
        for value in self.values:
            self.remove(value)
        return

    #---------------------------------------------------------------------------
    # 'Plottable' interface
    #---------------------------------------------------------------------------
 
    def _can_add ( self, item ):
        "Return whether a new object can be added to the group"
        return isinstance( item, Component )

    #---------------------------------------------------------------------------
    #  Return the list of value objects contained in the PlotGroup that contain
    #  a specified (x,y) point:  ###KW_TODO###
    #---------------------------------------------------------------------------
 
    def _objects_at ( self, x, y ):
        result = []
        for value_item in self._values:
            result = value_item.item.objects_at( x, y )
            if result:
                break
        return result

    #---------------------------------------------------------------------------
    #  Calculate the minimum size needed to draw the group:
    #---------------------------------------------------------------------------
 
    def _min_size ( self, gc ):
 
        # If we are not visible, return immediately:
        if not self.visible:
            return ( 0, 0 )
        
        # Get the margin information:
        mx, my = self._margin_size()
 
        # Calculate the minimum area needed for the plots:
        dir = self.orientation_
        wdx = wdy = pdx = pdy = 0
        total_weight = 0.0
        weighted     = 0
        for value_item in self._values:
            item                 = value_item.item
            tdx, tdy             = item._min_size( gc )
            value_item.mdxy      = [ tdx, tdy ]
            value_item.ldxy      = [ tdx, tdy ]
            value_item.mweighted = False
            weight               = item.weight
            value_item.stretch   = (weight == 'stretch')
            if (not value_item.stretch) and (weight > 0.0) and item.visible:
                value_item.mweighted = True
                total_weight        += weight
                weighted            += 1
                wdx                 += tdx
                wdy                 += tdy
            value_item.lweighted = value_item.mweighted
            if dir == HORIZONTAL:
                pdx += tdx
                pdy  = max( pdy, tdy )
            else:
                pdx  = max( pdx, tdx )
                pdy += tdy
 
        # Get the size of the titles:
        title_info = TitleInfo( gc, self._titles )
 
        # Get the size of the overlays:
        overlay_info = OverlayInfo( gc, self._overlays )
 
        # Calculate the required minimum size:
        rdx = title_info.dx + 2 * mx
        tdx = round( rdx + max( pdx, overlay_info.dx, title_info.max_dx ) )
        rdy = title_info.dy + 2 * my
        tdy = round( rdy + max( pdy, overlay_info.dy, title_info.max_dy ) )
 
        # Cache the results for use by the _layout routine:
        self._plot_cache = ( title_info, overlay_info, None, rdx, rdy,
                             pdx, pdy, wdx, wdy, mx, my, weighted,
                             total_weight, dir )
        
        # Return the total minimum size needed:
        self.min_width  = tdx
        self.min_height = tdy
        return ( tdx, tdy )

    #---------------------------------------------------------------------------
    #  Layout all of the contents of the group:
    #---------------------------------------------------------------------------
 
    def _layout ( self, x, y, dx, dy ):
 
        # If we are not visible, return immediately:
        if not self.visible:
            return
 
        # Restore the information cached by the _min_size method:
        ( title_info, overlay_info, axes_info, rdx, rdy, pdx, pdy, wdx, wdy,
          mx, my, weighted, total_weight, dir ) = self._plot_cache
        
        # Add the x,y displacement to create the title origin:
        tx = x + mx
        ty = y + my
 
        # Calculate the plot origin:
        
        px = tx + title_info.ldx
        py = ty + title_info.bdy
 
        # Remember the screen real estate we have been assigned:
        self.bounds = ( x, y, dx, dy )
            
        # Calculate any extra space remaining:
        extra_x = dx - rdx - pdx
        extra_y = dy - rdy - pdy
 
        # Adjust the plot area to use all available space:
        pdx += extra_x
        pdy += extra_y
 
        # If any plots are weighted, allocate the extra space amongst them:
        if total_weight > 0.0:
 
            # Handle the case of a horizontal layout:
            if dir == HORIZONTAL:
                self._equal_weighted( 1, pdy )
                if extra_x > 0:
                    self._layout_weighted( 0, wdx + extra_x, weighted, 
                                           total_weight ) 
                      
            # Handle the case of a vertical layout:
            else:
                self._equal_weighted( 0, pdx )
                if extra_y > 0:
                    self._layout_weighted( 1, wdy + extra_y, weighted,
                                           total_weight ) 
 
        # Lay the titles out:
        title_info._layout( px, py, pdx, pdy, tx, py + pdy + title_info.tdy,
                            px + pdx + title_info.rdx, ty )
        
        # Lay the overlays out:
        overlay_info._layout( px, py, pdx, pdy )
        
        # Lay the subcomponents out:
        if dir == HORIZONTAL:
            for value_item in self._values:
                value_item.item._layout( px, py, value_item.ldxy[0],
                                                 value_item.ldxy[1] )
                px += value_item.ldxy[0]
                value_item.sx  = px
                value_item.sy  = py
                value_item.sdx = 0
                value_item.sdy = pdy
        else:
            py += pdy
            for value_item in self._values:
                py -= value_item.ldxy[1]
                value_item.item._layout( px, py, value_item.ldxy[0],
                                                 value_item.ldxy[1] )
                value_item.sx  = px
                value_item.sy  = py
                value_item.sdx = pdx
                value_item.sdy = 0

    #---------------------------------------------------------------------------
    #  Set all weighted plots to the same size along a specified axis:
    #---------------------------------------------------------------------------
 
    def _equal_weighted ( self, index, new_size ):
        for value_item in self._values:
            if value_item.mweighted or value_item.stretch:
                value_item.ldxy[ index ] = new_size
 
    #---------------------------------------------------------------------------
    #  Allocate all extra space along a specified axis to the weighted plots:
    #---------------------------------------------------------------------------
 
    def _layout_weighted ( self, index, twd, weighted, total_weight ):
        values = self._values
        for value_item in values:
            value_item.lweighted = value_item.mweighted
        while True:
            used = i = 0
            redo = False
            for value_item in values:
                if value_item.lweighted:
                    weight = value_item.item.weight
                    i     += 1
                    if i == weighted:
                        nd = twd - used
                    else:
                        nd = int( round( (weight / total_weight) * twd ) )
                    used += nd
                    cd    = value_item.mdxy[ index ]
                    if nd < cd:
                        twd          -= cd
                        weighted     -= 1
                        total_weight -= weight
                        value_item.lweighted = False
                        redo = True
                        break
                    else:
                        value_item.nd = nd
            if not redo:
                for value_item in values:
                    if value_item.lweighted:
                        value_item.ldxy[ index ] = value_item.nd
                break
            if weighted <= 0:
                break

 
    def _draw ( self, gc ):
 
        # If we are not visible, return immediately:
        if not self.visible:
            return
 
        # Restore the information cached by the _min_size method:
        ( title_info, overlay_info, axes_info, rdx, rdy, pdx, pdy, wdx, wdy,
          mx, my, weighted, total_weight, dir ) = self._plot_cache
 
        # Draw the frame:
        PlotFrame._draw( self, gc )
 
        # Draw all of the titles:
        title_info._draw( gc )
 
        # Draw all of the plot's:
        for value_item in self._values:
            value_item.item._draw( gc )
 
        # Draw all of the overlays:
        overlay_info._draw( gc )
 
        # Draw the debug frame:
        self._draw_debug( gc )


# EOF
