""" A group of plots sharing a common index axis. """

# Enthought library imports.
from enthought.enable.base import empty_rectangle
from enthought.traits.api import Trait

# Local imports.
from axes_info import AxesInfo
from overlay_info import OverlayInfo
from plottable import Plottable
from plot_canvas import PlotCanvas
from plot_canvas import PlotCanvasList, PlotOverlayList, PlotTitleList
from plot_container import scrollbar_trait, RENDER_INFO
from plot_frame import PlotFrame
from title_info import TitleInfo

class PlotIndexed ( Plottable ):
    """ A group of plots sharing a common index axis. """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    index_scrollbar = Trait( scrollbar_trait, 
                             desc = 'the scrollbar to use for scrolling '
                                    'the index' )
    value_scrollbar = Trait( scrollbar_trait, 
                             desc = 'the scrollbar to use for scrolling '
                                    'the value' )

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
    
    def __init__ ( self, *items, **traits ):
        Plottable.__init__( self, **traits )
        self._titles   = []
        self._overlays = []
        self._values   = []
        self.titles    = PlotTitleList(   self, self._titles )
        self.overlays  = PlotOverlayList( self, self._overlays )
        self.values    = PlotCanvasList(  self, self._values )
        self.add( *items )
 
    #---------------------------------------------------------------------------
    #  Return whether a new object can be added to the indexed canvas:
    #---------------------------------------------------------------------------
 
    def _can_add ( self, item ):
        return isinstance( item, PlotCanvas )
       
    #---------------------------------------------------------------------------
    #  Return the list of value objects contained in the PlotIndexed that  
    #  contain a specified (x,y) point:
    #---------------------------------------------------------------------------
 
    def _objects_at ( self, x, y ):
        result = []
        for value_item in self._values:
            result = value_item.item.objects_at( x, y )
            if result:
               break
        return result

    #---------------------------------------------------------------------------
    #  Calculate the minimum size needed to draw the indexed canvas:
    #---------------------------------------------------------------------------
 
    def _min_size ( self, gc ):
 
        # If we are not visible, exit immediately:
        if not self.visible:
            return ( 0, 0 )
 
        # Calculate our title sizes:
        title_info = TitleInfo( gc, self._titles )
 
        # Calculate our overlay sizes:
        overlay_info = OverlayInfo( gc, self._overlays )
 
        # Get the layout direction:
        index_axis = self.index_axis
        horizontal = (index_axis == 'y')
 
        # Gather title, overlay, axes and plot size for all contained canvas's:
        axes_info = AxesInfo( gc, index_axis )
        mdx = ldx = rdx = mdy = tdy = bdy = 0
        total_weight    = 0.0
        weighted = wdxy = 0
        for value_item in self._values:
            item = value_item.item
            if item.visible:
                item._min_size( gc, index_axis )
                ( titles, overlays, axes, render_values, trdx, trdy, pdx, pdy, 
                  mx, my, sbdx, sbdy ) = item._plot_cache
                axes_info.factor_index_axes( axes )
                if horizontal:
                    dxy  = pdx + 2 * mx + titles.dx + axes.dx
                    mdx += dxy
                    mdy  = max( mdy, pdy )
                    tdy  = max( tdy, my + titles.tdy + axes.tdy )
                    bdy  = max( bdy, my + titles.bdy + axes.bdy + sbdy )
                    value_item.mdxy = value_item.ldxy = dxy
                else:
                    dxy  = pdy + 2 * my + titles.dy + axes.dy
                    mdy += dxy
                    mdx  = max( mdx, pdx )
                    ldx  = max( ldx, mx + titles.ldx + axes.ldx )
                    rdx  = max( rdx, mx + titles.rdx + axes.rdx + sbdx )
                    value_item.mdxy = value_item.ldxy = dxy
                value_item.mweighted = False
                weight               = item.weight
                if weight > 0.0:
                    value_item.mweighted = True
                    total_weight        += weight
                    weighted            += 1
                    wdxy                += dxy
 
        # Finalize/integrate all of the axes information we've gathered:
        axes_info.process_axes( gc )
        if horizontal:
            dx  = mdx + axes_info.dx
            tdy = max( tdy, axes_info.tdy )
            bdy = max( bdy, axes_info.bdy )
            dy  = tdy + mdy + bdy 
        else:
            dy  = mdy + axes_info.dy
            ldx = max( ldx, axes_info.ldx )
            rdx = max( rdx, axes_info.rdx )
            dx  = ldx + mdx + rdx
        
        # Calculate the margin sizes:
        mx, my = self._margin_size()
 
        # Determine the amount of space required for scroll bars (if any):
        sbdx, sbdy = self._scrollbar_size( index_axis )
        if horizontal:
            sbdy = 0
        else:
            sbdx = 0
 
        # Calculate the minimum size of the required margins, titles and
        # scroll bars:
        ndx = 2 * mx + title_info.dx + sbdx
        ndy = 2 * my + title_info.dy + sbdy
 
        # Calculate the minimum size of the inner plotting area:
        dx = max( dx, overlay_info.dx, title_info.max_dx )
        dy = max( dy, overlay_info.dy, title_info.max_dy )
 
        # Cache the results for use by the _layout routine:
        self._plot_cache = ( title_info, overlay_info, axes_info,
                             mx, my, sbdx, sbdy, ldx, rdx, tdy, bdy, 
                             horizontal, mdx, mdy, wdxy, weighted, total_weight)
 
        # Return the total minimum size needed:
        ndx = round( ndx + dx )
        ndy = round( ndy + dy )
        self.min_width  = ndx
        self.min_height = ndy
        return ( ndx, ndy )

    #---------------------------------------------------------------------------
    #  Layout all of the contents of the indexed canvas:
    #---------------------------------------------------------------------------
 
    def _layout ( self, x, y, dx, dy ):
 
        # If we are not visible, exit immediately:
        if not self.visible:
            return
 
        # Remember the screen real estate we have been assigned:
        self.bounds = ( x, y, dx, dy )
 
        # Restore the values cached by the _min_size method:
        ( title_info, overlay_info, axes_info, mx, my, sbdx, sbdy,
          ldx, rdx, tdy, bdy, horizontal, mdx, mdy,
          wdxy, weighted, total_weight ) = self._plot_cache
 
        # Add the x,y displacement to create the title origin:
        tx = x + mx
        ty = y + my + sbdy
 
        # Calculate the inner canvas area origin:
        icx  = tx + title_info.ldx
        icy  = ty + title_info.bdy
        icdx = dx - 2 * mx - title_info.dx - sbdx
        #### FIXME I THINK THIS IS A CUT AND PASTE BUG!!! SHOULD BE 2 * my
        icdy = dy - 2 * mx - title_info.dy - sbdy
        
        # Lay the titles out:
        title_info._layout( icx, icy, icdx, icdy,
                            tx, y + dy - my, x + dx - mx - sbdx, ty )
        
        # Lay the overlays out:
        overlay_info._layout( icx, icy, icdx, icdy )
 
        # Calculate the bottom-left plot origin:
        if horizontal:
            py    = icy  + bdy
            pdy   = icdy - tdy - bdy
            icx  += axes_info.ldx
            px    = icx
            pdx   = icdx - axes_info.dx
            extra = pdx - mdx
        else: 
            px    = icx  + ldx
            pdx   = icdx - ldx - rdx
            icy  += axes_info.bdy
            py    = icy
            pdy   = icdy - axes_info.dy
            extra = pdy - mdy
 
        # Cache the bounds of the plotting area:
        self._plot_bounds = span_plot_bounds = ( px, py, pdx, pdy )
 
        # Get the x and y axes scaling factors:
        sx, sy, sbx, sby = self._xy_scale()
        if horizontal:
            sx  = 1.0
            sbx = None
        else:
            sy  = 1.0
            sby = None
 
        # Lay the axes out:
        axes_info._layout( px, py, pdx, pdy, sx, sy )
 
        # Lay the scrollbars out:
        self._layout_scrollbars( sbdx, sbdy, sx, sy, sbx, sby,
                                  x,  y,  dx,  dy,
                                 px, py, pdx, pdy, mx, my )
 
        # If any plots are weighted, allocate the extra space amongst them:
        if (total_weight > 0.0) and (extra > 0):
            self._layout_weighted( wdxy + extra, weighted, total_weight ) 
 
        # Layout each of the individual canvas's:
        index_axis = self.index_axis
        if not horizontal:
            icy += pdy
        for value_item in self._values:
            item = value_item.item
            if item.visible:
                
                # Restore the values cached by the _min_size method:
                ( title_info, overlay_info, axes_info, render_values, rdx, rdy, 
                  cpdx, cpdy, mx, my, sbdx, sbdy ) = item._plot_cache
             
                # Determine the bounds of the current canvas:
                x  = icx
                dx = icdx
                dy = icdy
                if horizontal:
                    y    = icy
                    dx   = value_item.ldxy
                    icx += dx
                    px   = x + mx + title_info.ldx + axes_info.ldx
                    pdx  = dx - 2 * mx - title_info.dx - axes_info.dx
                else:
                    dy   = value_item.ldxy
                    icy -= dy
                    y    = icy
                    py   = y + my + title_info.bdy + axes_info.bdy
                    pdy  = dy - 2 * my - title_info.dy - axes_info.dy
                 
                # Remember the screen real estate we have been assigned:
                item.bounds = bounds = ( x, y, dx, dy )
             
                # Cache the bounds of the plot area for the _draw method:
                item._plot_bounds      = ( px, py, pdx, pdy )
                item._span_plot_bounds = span_plot_bounds
             
                # Set up the scrollbar related info:
                isx, isy, isbx, isby = item._xy_scale_cache = \
                     item._xy_scale( index_axis )
                if horizontal:
                    sbdx = 0
                    isby = None
                    isy  = sy
                else:
                    sbdy = 0
                    isbx = None
                    isx  = sx
             
                # Add the x,y displacement to create the title origin:
                tx = x + mx
                ty = y + my + sbdy
             
                # Lay the titles out:
                title_info._layout( tx + title_info.ldx, ty + title_info.bdy,
                                    dx - 2 * mx - title_info.dx,
                                    dy - 2 * my - title_info.dy,
                                    tx, ty, x + dx - mx, y + dy - my )
             
                # Lay the rest of the canvas out:
                item._layout_canvas( bounds, px, py, pdx, pdy, isx, isy, 
                                     overlay_info, axes_info, render_values )
             
                # Finish laying the scrollbars out:
                item._layout_scrollbars( sbdx, sbdy, isx, isy, isbx, isby, 
                                          x,  y,  dx,  dy,
                                         px, py, pdx, pdy, mx, my )

    #---------------------------------------------------------------------------
    #  Allocate all extra space to the weighted plots:
    #---------------------------------------------------------------------------
 
    def _layout_weighted ( self, twd, weighted, total_weight ):
        values = self._values
        for value_item in values:
            value_item.lweighted = value_item.mweighted
        while True:
            used = i = 0
            redo = False
            for value_item in values:
                if value_item.lweighted:
                    weight = value_item.item.weight
                    i     += 1
                    if i == weighted:
                        nd = twd - used
                    else:
                        nd = int( round( (weight / total_weight) * twd ) )
                    used += nd
                    cd    = value_item.mdxy
                    if nd < cd:
                        twd          -= cd
                        weighted     -= 1
                        total_weight -= weight
                        value_item.lweighted = False
                        redo = True
                        break
                    else:
                        value_item.nd = nd
            if not redo:
                for value_item in values:
                    if value_item.lweighted:
                        value_item.ldxy = value_item.nd
                break
            if weighted <= 0:
                break

    #---------------------------------------------------------------------------
    #  Draw the indexed canvas:
    #---------------------------------------------------------------------------
 
    def _draw ( self, gc ):
 
        # If we are not visible, exit immediately:
        if not self.visible:
            return
 
        # Restore the values cached by the _min_size method:
        ( title_info, overlay_info, axes_info, mx, my, sbdx, sbdy,
          ldx, rdx, tdy, bdy, horizontal, mdx, mdy,
          wdxy, weighted, total_weight ) = self._plot_cache
 
        # Draw our frame:
        PlotFrame._draw( self, gc )
 
        # Draw all of our titles:
        title_info._draw( gc )
 
        # Get the current draw offsets:
        ox = self._plot_ox
        oy = self._plot_oy
        
        # Get our clipping region:
        ipx, ipy, ipdx, ipdy = self._plot_bounds
 
        # Draw all of the inner canvas object backgrounds:
        for value_item in self._values:
            item = value_item.item
            if item.visible:
                gc.save_state()
                gc.clip_to_rect(*item.bounds)
              
                # Draw the frame:
                PlotFrame._draw( item, gc )
                
                # Draw the plot area background:
                item._draw_plot_frame( gc )
                
                # Set up the drawing offset:
                if horizontal:
                    ox = item._plot_ox
                else:
                    oy = item._plot_oy
                
                # Set up the clipping region:
                pbx, pby, pbdx, pbdy = item._plot_bounds
                gc.clip_to_rect(*item._plot_bounds)
                
                # Draw the canvas specific plot information:
                for renderer, values in item._plot_cache[ RENDER_INFO ]:
                    renderer._draw_canvas( gc, values, ox, oy )
                
                # Reset the clipping region:
                gc.restore_state()
 
        # Draw all of the index axes:
        axes_info._draw( gc, self._plot_ox, self._plot_oy )
 
        # Draw all of the inner canvas objects:
        for value_item in self._values:
            item = value_item.item
            if item.visible:
    
                # Restore the values cached by the _min_size method:
                ( titles, overlays, axes, render_values, rdx, rdy, pdx, pdy, 
                  mx, my, sbdx, sbdy ) = item._plot_cache
                  
                gc.save_state()
                gc.clip_to_rect(*item.bounds)
              
                # Draw all of the titles:
                titles._draw( gc )
             
                # Set up the drawing offset:
                if horizontal:
                    ox = item._plot_ox
                else:
                    oy = item._plot_oy
             
                # Draw all of the axes:
                axes._draw( gc, ox, oy )
             
                # Set up the clipping region:
                gc.clip_to_rect(*item._plot_bounds)
             
                # Draw all of the plots:
                for renderer, values in render_values:
                    renderer._draw( gc, values, ox, oy )
              
                # Draw all of the overlays:
                overlays._draw( gc, ox, oy )
                   
                # Reset the clipping region:
                gc.restore_state()
                   
                # Allow the plot overlays to use our clipping region so that
                # 'span' overlays will draw correctly:
                gc.save_state()
                gc.clip_to_rect(*self._plot_bounds)
                overlays._draw_plot( gc, ox, oy )
             
                # Draw all of the selections:
                from plot_value import PlotValue # fixme
                if not isinstance( item, PlotValue ):
                    
                    # Widen the clipping region a little to handle the selection
                    # boxes 'hanging over the edge':
                    gc.restore_state()
                    gc.save_state()
                    px, py, pdx, pdy = item._plot_bounds
                    gc.clip_to_rect(px-4, py-4, pdx+8, pdy+8)

                    # Now draw the selections (if any):
                    for renderer, values in render_values:
                        renderer._draw_selection( gc, values, ox, oy )

                # Reset the clipping region:
                gc.restore_state()
              
                # Draw the debug frame:
                item._draw_debug( gc )
 
        # Draw all of our overlays:
        overlay_info._draw( gc, self._plot_ox, self._plot_oy )
 
        # Draw the debug frame:
        self._draw_debug( gc )

#### EOF ######################################################################
