""" Abstract base class for plot iteractions. """

# Enthought library imports.
from enthought.enable import Component
from enthought.traits.api import Delegate, Float, Instance

# Local imports.
from plot_canvas import PlotCanvas
from window_surrogate import WindowSurrogate

class PlotInteraction(Component):
    """ Abstract base class for plot iteractions. """


    # The canvas providing the interactions.
    canvas = Instance(PlotCanvas, copy='shallow')

    ##########################################################################
    # Static trait handlers.
    ##########################################################################

    def _canvas_changed(self, old, new):
        """ Hook up relevant trait handlers to the new canvas. """

        if old is not None:
            old.on_trait_change(self.plot_bounds_changed, '_plot_bounds',
                                remove=True)
            old.on_trait_change(self.plot_container_changed, 'container',
                                remove=True)
            self._remove_overlay_from(old.container)

        if new is not None:
            new.on_trait_change(self.plot_bounds_changed, '_plot_bounds')
            new.on_trait_change(self.plot_container_changed, 'container')
            self._add_overlay_to(new.container)

            if new._plot_bounds is not None:
                self.bounds = new._plot_bounds

        return

    ##########################################################################
    # Dynamic trait handlers. 
    ##########################################################################

    def plot_container_changed(self, source, trait, old, new):
        """ Handle the canvas's container being changed. """

        if new is not None:
            self._add_overlay_to(new)

        if old is not None:
            self._remove_overlay_from(old)

        return
    
    def plot_bounds_changed(self, source, trait, old, new):
        """ Handle the canvas's bounds being changed.

        Set the bounds of this component to the bounds of the canvas.
        """

        self.bounds = new

        return

    ###########################################################################
    # 'Component' interface. 
    ###########################################################################

    def _draw(self, gc):
        pass

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _add_overlay_to(self, container):
        """ Attach an overlay to the specified container object. """

        # This is a bit gorpy because the container can either be a
        # PlotComponent, a Plottable,  or the SingletonDefaultContainer. The
        # SingletonDefaultContainer does not have overlays and it acts as its
        # own host and owner, but it does not have overlays.  This is an issue
        # that needs to be resolved.
        if isinstance(container.host, WindowSurrogate):
            # FIXME : This is in direct violation of the Law of Demeter.
            container.host.owner.overlays.append(self)

        return

    def _remove_overlay_from(self, container):
        """ Remove the overlay from the specified container object. """

        # See discussion in _add_overlay_to...
        if isinstance(container.host, WindowSurrogate):
            if self in container.host.owner.overlays:
                container.host.owner.overlays.remove(self)

        return

#### EOF ######################################################################
