"""
Defines the PlotLabel class.
"""

from enthought.util.numerix import fromstring, NewAxis, pi, repeat, resize, UInt8, zeros

# Enthought library imports.
from enthought.traits.api import Delegate, Instance, Str, Trait, Unicode, Int
from enthought.traits.ui.api import View

# Local imports.
from plot_base import LEFT, RIGHT
from plot_base import black_color
from plot_container import IsContained
from plot_frame import PlotFrame
from plot_label_style import PlotLabelStyle
from plot_views import border_group, legend_group, margin_group, misc_group

# Imports to support graphical plot labels (instead of textual ones)
from enthought.kiva import CompiledPath
from enthought.kiva.backend_image import GraphicsContext as GraphicsContextArray
from enthought.traits.api import Tuple, Float


try:
    #if matplotlib isn't available, the text just won't be rendered correctly
    from matplotlib.mathtext import math_parse_s_ft2font as parse_mathtext
except ImportError:
    parse_mathtext = None



# Constants defining the minimum size of a plot label, the padding around
# it, and the total space it requires
LEGEND_DXY   = 10
LEGEND_SPACE = 6
LEGEND_TOTAL = LEGEND_DXY + LEGEND_SPACE


class PlotLabel ( PlotFrame, IsContained ):
    """
    PlotLabel represents anything that decorates a plot, inside or outside
    of the actual plotting area.  In general it is a textual label (e.g.
    axis labels, plot titles) or an overlay or legend inside a plot.
    """

    #---------------------------------------------------------------------------
    # Public traits
    #---------------------------------------------------------------------------

    # The text of the label.
    text  = Trait("", Str(desc='the label text', label='Text'), Unicode)

    # Path defining symbol label.
    path         = Instance(CompiledPath)

    # Padding between path and text if we have both.
    path_padding = Int(3)
    _path_offset = Float(0)

    # X and Y scaling factors for the symbol defined by 'path'.
    path_size    = Tuple(Float, Float)


    #---------------------------------------------------------------------------
    # Style- and style delegation-related traits
    #---------------------------------------------------------------------------

    # Delegation prefix.
    __prefix__ = 'plot_label_'

    # The stylesheet for this plot label.  By default, delegate up the
    # container hierarchy for 'plot_label_container'.
    style = Delegate('container', '*')

    # The orientation of the text.  Set to 0 for horizontal; 90 for vertical.
    alignment = Delegate('style')

    # The font color for text.
    color = Delegate('style')

    # The font to use for text.
    font = Delegate('style')

    # If this label is a legend, the background color for the legend box.
    legend_color = Delegate('style')

    # The orientation of the label.
    rotate = Delegate('style')

    # Is this plot drawn in debug mode?
    debug = Delegate('container')

    traits_view = View(
        [ [ 'text$', 'color', 'bg_color{Background Color}', 'font',
            'alignment', 'rotate', 'visible$',
            '|[Main]<@' ],
          '|{Main}' ],
        border_group,
        margin_group,
        legend_group,
        misc_group
    )


    def __init__ ( self, text = None, **traits ):
        if text is not None:
           self.text = text
        IsContained.__init__( self, **traits )
        return


    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------

    def _text_changed ( self, text ):
        "Trait event handler that gets called when the 'text' trait is changed."
        if self._is_mathtext():
            dx, dy, fonts = parse_mathtext(text, 72, 18)
            bitmap = None
            for i, font in enumerate(fonts):
               w, h, bits = font.image_as_str()
               if bitmap is None:
                   bitmap = zeros(h*w, UInt8)
               bitmap += fromstring(bits, UInt8)
            ary = resize(fromstring(bitmap, UInt8), (h,w))

            self._mathtext_dx = dx
            self._mathtext_dy = dy
            self._mathtext_mask = ary   # 2D array that can be used for alphas
            # extrude the mask into 4 bitplanes
            tmp_array = repeat(self._mathtext_mask[:, :, NewAxis], 4, axis=-1)
            tmp_array[:,:,0] = int(self.color_[0] * 255)
            tmp_array[:,:,1] = int(self.color_[1] * 255)
            tmp_array[:,:,2] = int(self.color_[2] * 255)
            img = GraphicsContextArray(tmp_array, "rgba32")

            gc = GraphicsContextArray((w, h), "rgba32")
            gc.set_fill_color(self.bg_color_)
            gc.rect(0, 0, w, h)
            gc.fill_path()
            gc.draw_image(img)

            self._mathtext_img = gc
            self._mathtext_color = self.color_
            self._mathtext_bgcolor = self.bg_color_
        else:
            self._text_list = text.strip().split( '\n' )
        return

    def _min_size ( self, gc ):
        """
        Calculate the minimum size needed to draw the label.  Returns a tuple
        (min_width, min_height).
        """

        # If the label is not visible, then it has no size:
        if not self.visible:
            return ( 0, 0 )

        # If the text string is empty, then we don't have to calculate its size:
        py      = self.padding_height
        min_tdx = min_tdy = 0

        if self._is_mathtext():
            min_tdx = self._mathtext_dx
            min_tdy = self._mathtext_dy
            self._frame_tdx = min_tdx
            self._frame_tdy = min_tdy
            self._frame_descent = 0
        elif (self._text_list is not None) and (self._text_list[0] != ''):
            text = self._text_list
            font   = self.font
            rotate = self.rotate

            gc.set_font( font )

            # Get the 'worst case' size of the text:
            n = len( text )
            self._frame_tdx = tdx = [ 0 ] * n
            self._frame_tdy = tdy = [ 0 ] * n
            i      = 0
            multiline = len(text) > 1
            for line in text:
                ( min_dx,  min_dy,
                  descent, leading ) = gc.get_full_text_extent( line )
                min_dy  += py
                min_tdy += min_dy
                min_tdx  = max( min_tdx, min_dx )
                tdx[i]   = min_dx
                tdy[i]   = min_dy
                if multiline:
                    tdy[i] += 5
                i       += 1
            min_tdy -= py
            self._frame_descent = descent

            # Check to see if a legend is to be displayed:
            if self.legend_color_[3] != 0:
                min_tdx += LEGEND_TOTAL
                min_tdy  = max( min_tdy, LEGEND_DXY )

            # See if it is rotated and adjust accordingly:
            if self.rotate == 90:
                min_tdx, min_tdy = min_tdy, min_tdx

        elif self.legend_color_[3] != 0:
            min_tdx = min_tdy = LEGEND_DXY

        if self.path is not None:
            path_tdx, path_tdy = self.path_size
            if min_tdy > 0:
                # We have some text, so add some padding
                min_tdy += self.path_padding

            self._path_offset = path_tdy + self.path_padding + \
                                self.margin_height + self.border_size + py

            min_tdy += path_tdy
            min_tdx = max(path_tdx, min_tdx)

        # Include the bitmap background size (if necessary):
        pad = True
        # TODO: temporarily commented out until bitmap support is working
        #if min_tdx == 0:
           #bitmap = self.bg_image_
           #if bitmap is not None:
           #   min_tdx = bitmap.GetWidth()
           #   min_tdy = bitmap.GetHeight()
           #   pad     = False

        # Cache the various sizes for the drawing and layout routines:
        self._frame_min_tdx = min_tdx
        self._frame_min_tdy = min_tdy

        # Include our margins:
        border_size    = self.border_size
        self._frame_mx = mx = (self.margin_width  + border_size +
                                    pad * self.padding_width)
        self._frame_my = my = (self.margin_height + border_size +
                                    pad * py)

        # Cache and return the final result:
        if self.use_min_width:
            dx = self.min_width
        else:
            dx = min_tdx + 2 * mx
        if self.use_min_height:
            dy = self.min_height
        else:
            dy = min_tdy + 2 * my
        self._frame_min_dx = dx
        self._frame_min_dy = dy

        return ( dx, dy )

    def _is_mathtext(self):
        "returns True if the contained text is a MathText string."
        if self.text.startswith("$") and self.text.endswith("$") and parse_mathtext is not None:
            return True
        else:
            return False

    def _layout ( self, x, y, dx, dy, cache = None ):
        "Lay the label out into the assigned real estate."

        # Don't do anything if we are not visible:
        if not self.visible:
            return

        # Cache the new bounds:
        cache        = cache or self
        cache.bounds = ( x, y, dx, dy )

        mx           = self._frame_mx
        my           = self._frame_my
        py           = self.padding_height
        lx           = LEGEND_TOTAL * (self.legend_color_[3] != 0)

        # If the text is the empty string, then we are done:
        text = self._text_list
        if (text is None) or (text[0] == ''):
            self._frame_lx = x + mx
            self._frame_ly = y + my
            return

        n               = len( text )
        cache._frame_tx = tx = [ 0 ] * n
        cache._frame_ty = ty = [ 0 ] * n
        tdx             = self._frame_tdx
        tdy             = self._frame_tdy
        descent         = self._frame_descent
        alignment       = self.alignment_

        if self.rotate == 90:
            # Calculate the position of the legend:
            self._frame_lx = x + (dx - LEGEND_DXY) / 2
            if alignment & RIGHT:
                self._frame_ly = y + dy - my - LEGEND_DXY
            else:
                self._frame_ly = y + my

            # Calculate the rotated position:
            cx = x + mx + descent - py
            for i in range( n ):
                cx   += tdy[i]
                tx[i] = cx

                # Calculate the y coordinate based on the specified text
                # alignment:
                if alignment & LEFT:
                    ty[i] = y + my + lx
                elif alignment & RIGHT:
                    ty[i] = (y + dy - lx - my - tdx[i])
                else:
                    ty[i] = y + ((dy + lx - tdx[i]) / 2)
            
            adjustment = sum(tdy) / 2 - tdy[0] / 2
            if tx[0] - adjustment > self._frame_lx:
                for i in range(n):
                    tx[i] -= adjustment
        else:
            # Calculate the position of the legend:
            self._frame_ly = y + (dy - LEGEND_DXY) / 2
            if alignment & RIGHT:
                self._frame_lx = x + dx - mx - LEGEND_DXY
            else:
                self._frame_lx = x + mx

            # Calculate the normal position:
            cy = y + dy - my - descent + py
            for i in range( n ):
                cy   -= tdy[i]
                ty[i] = cy

                # Calculate the x coordinate based on the specified text
                # alignment:
                if alignment & LEFT:
                    tx[i] = x + mx + lx
                elif alignment & RIGHT:
                    tx[i] = x + dx - lx - tdx[i] - mx
                else:
                    tx[i] = x + ((dx + lx - tdx[i]) / 2)

        return

    def _draw ( self, gc, ox = 0, oy = 0, cache = None ):
        """ Draw the label. """

        # Don't draw anything if we are not visible:
        if not self.visible:
           return

        # Draw the frame:
        PlotFrame._draw( self, gc, ox, oy, cache )

        # Draw the legend:
        legend_color = self.legend_color_
        if legend_color[3] != 0:
            x = self._frame_lx
            y = self._frame_ly
            gc.set_stroke_color( black_color )
            gc.set_fill_color( legend_color )
            gc.set_line_width( 1 )
            gc.begin_path()
            gc.move_to( x, y )
            gc.line_to( x + LEGEND_DXY, y )
            gc.line_to( x + LEGEND_DXY, y + LEGEND_DXY )
            gc.line_to( x, y + LEGEND_DXY )
            gc.close_path()
            gc.draw_path()

        # If there is text to draw, draw it.
        offset = 0
        if self.path:
            offset = self._path_offset
        if self._is_mathtext():
            self._draw_text(gc, self.text, ox, oy + offset, cache)
        elif (self._text_list is not None) and (self._text_list[0] != ''):
            self._draw_text(gc, self._text_list, ox, oy + offset, cache)

        # Otherwise, if there is a marker to draw, draw it.
        if self.path is not None:
            min_x, min_y = self._min_size(gc)
            path_offset = min_y - self._path_offset
            self._draw_marker(gc, self.path, ox, oy + path_offset, cache)

    def _draw_marker( self, gc, path, ox, oy, cache ):
        # gc.move_to(cache.bounds[0], cache.bounds[1])
        gc.save_state()
        gc.set_stroke_color(self.color_)
        gc.set_line_width(1)

        gc.translate_ctm(cache.bounds[0]+ox, cache.bounds[1]+oy)
        gc.begin_path()
        gc.add_path(self.path)
        gc.close_path()
        gc.stroke_path()
        gc.restore_state()

    def _draw_text( self, gc, text, ox, oy, cache ):
        if self._is_mathtext():
            self._draw_math_text(gc, text, ox, oy, cache)
        else:
            self._draw_normal_text(gc, text, ox, oy, cache)

    def _draw_math_text( self, gc, text, ox, oy, cache ):
        # checks if the color of the cached mathtext image is the same as the
        # current color.
        if (self._mathtext_color != self.color_) or \
                                (self._mathtext_bgcolor != self.bg_color_):
            self._text_changed(self.text)

        # Access the saved cache (x,y) data:
        cache = cache or self
        if hasattr(cache, "_frame_mx"):
            mx = cache._frame_mx
            my = cache._frame_my
        else:
            mx = my = 0
        gc.save_state()
        gc.translate_ctm(cache.bounds[0]+mx, cache.bounds[1]+my)
        gc.draw_image(self._mathtext_img)
        gc.restore_state()
        return


    def _draw_normal_text( self, gc, text, ox, oy, cache):
        gc.set_fill_color( self.color_ )
        gc.set_font( self.font )

        # Access the saved cache (x,y) data:
        cache = cache or self
        tx = cache._frame_tx
        ty = cache._frame_ty
        if self.rotate == 90:
            # Draw the text rotated 90 degrees:
            for i in range( len( text ) ):
                # FIXME: This is a workaround for a bug where kiva's save_state
                # and restore state does not reset the text position.  If it
                # did, we could move this outside the loop.
                gc.set_text_position( 0, 0 )

                gc.save_state()
                gc.translate_ctm( round( tx[i] - ox ), round( ty[i] - oy ) )
                gc.rotate_ctm( pi / 2.0 )
                gc.show_text( text[i] )
                gc.restore_state()
        else:
            # Draw the text normally:
            for i in range( len( text ) ):
                gc.set_text_position( round( tx[i] - ox ), round( ty[i] - oy ) )
                gc.show_text( text[i] )

        # Draw the debug frame:
        self._draw_debug( gc, ox, oy, cache )

        return

# EOF
