"""
Defines the PlotOverlay class
"""


# Enthought library imports.
from enthought.traits.api import Delegate, Float, Trait
from enthought.traits.ui.api import View

# Local imports.
from plot_label import PlotLabel
from plot_overlay_style import PlotOverlayStyle
from plot_views import border_group, legend_group, margin_group, misc_group


class PlotOverlay ( PlotLabel ):
    """
    A PlotOverlay is drawn inside the plotting area of a PlotCanvas, instead
    of outside like PlotTitle.  It can be positioned relative to the plotting
    area or relative to a particular PlotValue's index or data.
    """

    #---------------------------------------------------------------------------
    # Traits.
    #---------------------------------------------------------------------------

    # The location along the parent's value axis to place the overlay.
    value = Trait(None, None, Float)

    # The location along the parent's index axis to place the overlay.
    index = Trait(None, None, Float)

    
    #---------------------------------------------------------------------------
    # Style-related traits
    #---------------------------------------------------------------------------
    
    # Delegation prefix.
    __prefix__ = 'plot_overlay_'

    # The style sheet which controls the visual style of this overlay.
    style = Delegate('container', '*')

    # The relative position of the overlay.
    position = Delegate('style')

    # Should the overlay span more than one canvas?
    span = Delegate('style')

    # The line color of the location cross hairs.
    line_color = Delegate('style')

    # The line style of the location cross hairs. 
    line_style = Delegate('style')

    # The line weight of the location cross hairs.
    line_weight = Delegate('style')

    traits_view = View(
        [ [ 'text{Label}$', '_', 'color', ' ', 'font', '_',
            'alignment',   ' ', 'rotate{Angle}',
            '|[Text]@' ],
          [ 'bg_color{Color}', '_', 'position', ' ', 'visible',
            '|[Overlay]@' ],
          '|{Main}' ],
        [ [ 'index', 'value',
            '|[Value For]' ],
          [ 'line_color{Color}', '_', 'line_style{Style}', '_',
            'line_weight{Weight}', 'span{Span?}',
            '|[Line]@' ],
          '|{Data}' ],
        border_group,
        margin_group,
        legend_group,
        misc_group
    )

    def remove ( self ):
        """ Remove this overlay from its container. """
        
        self.container.remove( self )
        return


    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------

    def _draw_plot ( self, gc, ox, oy ):
        """ Draw an overlay that is relative to the underlying plot area.
        
        When a PlotOverlay has its value and index traits set and is overlaid
        on a PlotValue object, it is drawn at a location near the screen 
        location defined by (index, value) in the data space of the containing
        plot value. This method is invoked to draw such overlays.
        
        @see OverlayInfo._draw_plot
        """

        # Save the current drawing offset in case we do a 'hittest' later:
        self._plot_ox = ox
        self._plot_oy = oy
        
        # Don't draw anything if we are not visible:
        if not self.visible:
            return
        
        # Draw the overlay (if any):
        if (self._hline is not None) or (self._vline is not None):
            gc.save_state()
            gc.set_stroke_color( self.line_color_ )
            gc.set_line_width( self.line_weight )
            gc.set_line_dash( self.line_style_ )
            if self._hline is not None:
                x1, y1, x2, y2 = self._hline
                gc.begin_path()
                gc.move_to( x1 - ox, y1 - oy )
                gc.line_to( x2 - ox, y2 - oy )
                gc.stroke_path()
            if self._vline is not None:
                x1, y1, x2, y2 = self._vline
                gc.begin_path()
                gc.move_to( x1 - ox, y1 - oy )
                gc.line_to( x2 - ox, y2 - oy )
                gc.stroke_path()
            gc.restore_state()
           
        # Draw the frame:
        self._draw( gc, ox, oy )

        return
    
#### EOF ######################################################################
