""" A plot data source. """

# Enthought library imports.
from enthought.traits.api import Delegate, HasTraits, Instance, Trait
from enthought.traits.ui.api import EnumEditor, Item, View

# Local imports.
from plot_window_base import PlotTraitCommand


def validate_source_name(object, name, value):
    """ A validator function for the source_name trait. """

    # If the data trait has been set, validate that value is a valid source
    # in data.
    if object._data is not None:
        if value not in object._data.sources and value is not 'default':
            raise ValueError()

    return value

class PlotSource ( HasTraits ):
    """ A plot data source. """

    #### Public Traits ####

    # The name of the channel.
    source_name = Trait('default', type=str, validator=validate_source_name)

    # The plot that is rendering this data source.
    value = Instance('enthought.chaco.PlotCanvas')

    #### Private Traits ####

    # The PlotData of which this PlotSource is a single channel.
    _data = Instance('enthought.chaco.PlotData')

    # The set of valid sources (channel names) to reconnect to this source.
    # This trait is private, even though it's name suggests it is public. The
    # confusion with the name is to make delegation work correctly.
    sources = Delegate('_data')

    #### Trait Editor ####

    traits_view = View(
        Item('source_name', editor=EnumEditor(name='sources'))
    )

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__ ( self, value, data, source_name='default', **traits):
        """ Construct a new PlotSource. """

        # Invoke the super class (HasTraits) constructor with the traits.
        super(PlotSource, self).__init__(**traits)

        # Set the required traits.
        self.source_name=source_name
        self.value  = value
        self._data  = data

        # Register with the PlotData to receive changes to the data set.
        data.add_source( self )

        # From this point forward, if any trait changes, we need to refresh.
        self.on_trait_change(self._on_anytrait_changed)

        # Ask the PlotValue to refresh, since we have effectively changed its
        # data.
        self.refresh()

        return

    ###########################################################################
    # Public 'PlotSource' interface.
    ###########################################################################

    def data ( self ):
        """ Return the current data values for this source. """

        return self._data.data( self.source_name )

    def refresh ( self ):
        """ Refresh the PlotValue with the current data.

        This method returns True if the data has indeed changed and False
        otherwise. The return code can be used to make decisions about what
        else may need updating. The 'update' method uses this method and
        forces an update of the PlotValue if this method returns True.

        Subclasses may override this if they implement a mechanism for
        detecting stale data.
        """

        if self.value is not None:
            self.value.data = self
            return True

        return False

    def update ( self ):
        """ Update the PlotValue with the current data.

        The distinction between this method and refresh is that this method
        will request an update from the underlying PlotValue -- forcing the
        redraw.
        """

        rc = self.refresh()
        if rc:
            self.value.update()

        return rc

    def remove ( self ):
        """ Disconnect this source.

        When this method is completed, it is an orphan. It no longer provides
        data (or even preserves a reference) to the PlotValue and it no longer
        receives notifications from its PlotData when the data has changed.
        """

        self._data.remove( self )
        self._data = None
        self.value = None

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def add_undo ( self, command ):
        """ Add an undo command to the PlotValue's undo handler. """

        # FIXME: I'm not sure that the undo mechanism is actually working.
        if self.value is not None:
            self.value.undo_handler.add_undo( command )

        return

    #### Trait event handlers #################################################

    #### Dynamic ####

    def _on_anytrait_changed ( self, obj, name, old_value, new_value ):
        """ Called when any trait changes. """

        # Request a refresh of the PlotValue, since the data has likely changed
        # as a result of this trait change (i.e., it's probably a different
        # channel).
        if self.update():

            # If the plot was updated, push an undo command onto the undo
            # stack.
            undo = PlotTraitCommand( self, name, old_value, new_value )
            self.add_undo(undo)

        return

#### EOF ######################################################################
