""" Style traits for plots. """
from traceback import print_exc

from enthought.util.numerix import array
from types import ClassType, ListType

import plot_base

from enthought.chaco.renderers.renderer   import Renderer
from enthought.traits.api import Range, Trait, TraitPrefixList, TraitPrefixMap,\
                             Enum, TraitError
from enthought.traits.ui.api import CheckListEditor, EnumEditor, ImageEnumEditor

from plot_base import BOTTOM, HCENTER, LEFT, RIGHT, TOP, VCENTER
from plot_base import basic_sequence_types

###############################################################################
# Alignment
###############################################################################

_alignment_positions = {
    'top':    TOP + LEFT,
    'left':   TOP + LEFT,
    'bottom': BOTTOM + RIGHT,
    'right':  BOTTOM + RIGHT,
    'center': VCENTER + HCENTER
}

_edit_alignment_positions = {
    TOP + LEFT:        '1:left',
    VCENTER + HCENTER: '2:center',
    BOTTOM + RIGHT:    '3:right'
}

_alignment_editor = ImageEnumEditor(
    cols   = 3,
    suffix = '_align',
    values = _edit_alignment_positions,
    module = plot_base
)

Alignment = Trait(
    'center',
    TraitPrefixMap(_alignment_positions),
    editor = _alignment_editor
)

###############################################################################
# Border size.
###############################################################################

_border_size_editor = ImageEnumEditor(
    cols   = 4,
    suffix = '_weight',
    values = range(0, 16),
    module = plot_base
)

BorderSize = Range(0, 15, value=0, editor=_border_size_editor)

###############################################################################
# Border size.
###############################################################################

BORDER_SOLID = 0
BORDER_BEVEL = 1
BORDER_DROP  = 2

_border_styles = {
    'solid':       BORDER_SOLID,
    'bevel':       BORDER_BEVEL,
    'drop shadow': BORDER_DROP
}

_border_style_editor = ImageEnumEditor(
    cols   = 3,
    suffix = '_border',
    values = _border_styles.keys(),
    module = plot_base
)

BorderStyle = Trait(
    'solid',
    TraitPrefixMap(_border_styles),
    editor = _border_style_editor
)

###############################################################################
# Edge.
###############################################################################

_edge_positions = {
    'top':    TOP,
    'bottom': BOTTOM,
    'left':   LEFT,
    'right':  RIGHT
}

_edge_editor = ImageEnumEditor(
    cols   = 2,
    suffix = '_position',
    values = _edge_positions.keys(),
    module = plot_base
)

Edge = Trait('top', TraitPrefixMap(_edge_positions), editor=_edge_editor)


###############################################################################
# Fill style.
###############################################################################

_fill_styles = {
    'solid':         None,   ### NOT IMPLEMENTED YET ###
    'stipple':       None,   ### NOT IMPLEMENTED YET ###
    'back hatch':    None,   ### NOT IMPLEMENTED YET ###
    'diag hatch':    None,   ### NOT IMPLEMENTED YET ###
    'forward hatch': None,   ### NOT IMPLEMENTED YET ###
    'cross hatch':   None,   ### NOT IMPLEMENTED YET ###
    'horz hatch':    None,   ### NOT IMPLEMENTED YET ###
    'vert hatch':    None,   ### NOT IMPLEMENTED YET ###
    'transparent':   None    ### NOT IMPLEMENTED YET ###
}


FillStyle = Trait(
    'solid',
    TraitPrefixMap(_fill_styles),
    cols = 3
)

###############################################################################
# Line style.
###############################################################################

_line_styles = {
    'solid':     None,
    'dot dash':  array( [ 3.0, 5.0, 9.0, 5.0 ] ),
    'dash':      array( [ 6.0, 6.0 ] ),
    'dot':       array( [ 2.0, 2.0 ] ),
    'long dash': array( [ 9.0, 5.0 ] )
}

_line_style_editor = ImageEnumEditor(
    cols   = 3,
    suffix = '_line',
    values = _line_styles.keys(),
    module = plot_base )

LineStyle = Trait(
    'solid',
    TraitPrefixMap(_line_styles),
    editor=_line_style_editor
)

###############################################################################
# Line weight.
###############################################################################

_line_weight_editor = ImageEnumEditor(
    cols   = 3,
    suffix = '_weight',
    values = range(1, 10),
    module = plot_base
)

LineWeight = Range(1, 9, value=1, editor=_line_weight_editor)

###############################################################################
# Margin
###############################################################################

Margin = Range(0, 31, cols=6)

###############################################################################
# Orientation.
###############################################################################

VERTICAL   = 0
HORIZONTAL = 1

_orientation_types = {
    'vertical':   VERTICAL,
    'horizontal': HORIZONTAL
}

_orientation_editor = ImageEnumEditor(
    cols   = 2,
    suffix = '_orientation',
    values = _orientation_types.keys(),
    module = plot_base
)

Orientation = Trait(
    'horizontal',
    TraitPrefixMap(_orientation_types),
    editor = _orientation_editor
)


###############################################################################
# Origin.
###############################################################################

_origin_positions = {
    'top left':      TOP + LEFT,
    'left top':      TOP + LEFT,
    'top right':     TOP + RIGHT,
    'right top':     TOP + RIGHT,
    'bottom left':   BOTTOM + LEFT,
    'left bottom':   BOTTOM + LEFT,
    'bottom right':  BOTTOM + RIGHT,
    'right bottom':  BOTTOM + RIGHT,
    'tl':            TOP + LEFT,
    'lt':            TOP + LEFT,
    'tr':            TOP + RIGHT,
    'rt':            TOP + RIGHT,
    'bl':            BOTTOM + LEFT,
    'lb':            BOTTOM + LEFT,
    'br':            BOTTOM + RIGHT,
    'rb':            BOTTOM + RIGHT
}

_edit_origin_positions = [
    'top left',
    'top right',
    'bottom left',
    'bottom right'
]

_plot_origin_editor = ImageEnumEditor(
    cols   = 2,
    suffix = '_origin',
    values = _edit_origin_positions,
    module = plot_base
)

Origin = Trait(
    'bottom left',
    _origin_positions,
    editor = _plot_origin_editor
)

###############################################################################
# Outline weight.
###############################################################################

_outline_weight_editor = ImageEnumEditor(
    cols   = 3,
    suffix = '_weight',
    values = range(0, 9),
    module = plot_base
)

OutlineWeight = Range(0, 9, value=1, editor=_outline_weight_editor)

###############################################################################
# Plot rederers.
###############################################################################

renderer_map = {}

def get_renderer_for ( name ):
    if isinstance( name, Renderer ):
        return name
    if type( name ) is ClassType:
        if not issubclass( name, Renderer ):
           raise TraitError
        return name()
    name     = name.strip().lower()
    renderer = renderer_map.get( name, None )
    if renderer is None:
        module   = __import__( 'enthought.chaco.renderers.%s' % name,
                               globals(), locals(), [ '' ] )
        cls_name = '%srenderer' % name
        cls      = [ x for x in dir( module ) if (
                       (x.lower() == cls_name) and
                       #(type( getattr( module, x ) ) is ClassType) and
                       issubclass( getattr( module, x ), Renderer )) ]
        if len( cls ) != 1:
            raise TraitError
        renderer_map[ name ] = renderer = getattr( module, cls[0] )()
    return renderer

PlotRenderers = Trait([get_renderer_for('line')], ListType)

###############################################################################
# Plot type.
###############################################################################

def _plot_type_map ( object, name, value ):
    type_value = type( value )
    if type_value in basic_sequence_types:
        values = value
    elif isinstance(type_value, basestring):
        values = value.split( ',' )
    else:
        values = [ value ]

    object._plot_renderers = [ get_renderer_for( x ) for x in values ]
    return value

_plot_type_map.info = ( "a comma separated plot type string (e.g. 'line,"
                       "scatter') or plot type list (e.g. ['line','scatter'])" )

_plot_type_editor = CheckListEditor( cols = 3, values = [
    [ 'line',        'Line (1)' ],
    [ 'bar',         'Bar (1+)'  ],
    [ 'range',       'Range (2)' ],
    [ 'scatter',     'Scatter (1)' ],
    [ 'polygon',     'Polygon (1)' ],
    [ 'stackedline', 'Stacked Line (1+)' ],
    [ 'stackedbar',  'Stacked Bar (1+)' ],
    [ 'rangebar',    'Range Bar (2)' ],
    [ 'hloc',        'H/L/O/C (3,4)' ],
    [ 'segment',     'Segment (1)' ],
    [ 'filllt',      'Fill < (2)' ],
    [ 'fillne',      'Fill <> (2)' ],
    [ 'fillgt',      'Fill > (2)' ],
    [ 'candle',      'Candlestick (4)' ],
    [ 'image',       'Image (1)' ] ] )

PlotType = Trait('line', _plot_type_map, editor=_plot_type_editor)

###############################################################################
# Position.
###############################################################################

_all_positions =  {
    'top left':      TOP + LEFT,
    'left top':      TOP + LEFT,
    'top right':     TOP + RIGHT,
    'right top':     TOP + RIGHT,
    'top center':    TOP + HCENTER,
    'center top':    TOP + HCENTER,
    'center left':   VCENTER + LEFT,
    'left center':   VCENTER + LEFT,
    'center right':  VCENTER + RIGHT,
    'right center':  VCENTER + RIGHT,
    'bottom left':   BOTTOM + LEFT,
    'left bottom':   BOTTOM + LEFT,
    'bottom right':  BOTTOM + RIGHT,
    'right bottom':  BOTTOM +RIGHT,
    'bottom center': BOTTOM + HCENTER,
    'center bottom': BOTTOM + HCENTER,
    'center':        VCENTER + HCENTER,
    'c':             VCENTER + HCENTER,
    'tl':            TOP + LEFT,
    'lt':            TOP + LEFT,
    'tr':            TOP + RIGHT,
    'rt':            TOP + RIGHT,
    'bl':            BOTTOM + LEFT,
    'lb':            BOTTOM + LEFT,
    'br':            BOTTOM + RIGHT,
    'rb':            BOTTOM + RIGHT,
    'top':           TOP + HCENTER,
    't':             TOP + HCENTER,
    'bottom':        BOTTOM + HCENTER,
    'b':             BOTTOM + HCENTER,
    'left':          VCENTER + LEFT,
    'l':             VCENTER + LEFT,
    'right':         VCENTER + RIGHT,
    'r':             VCENTER + RIGHT
}

# Trait editor version of 'all_positions':
_edit_all_positions = [
    'top left',
    'top center',
    'top right',
    'center left',
    'center',
    'center right',
    'bottom left',
    'bottom center',
    'bottom right'
]

_position_editor = ImageEnumEditor(
    cols = 3,
    suffix = '_position',
    values = _edit_all_positions,
    module = plot_base
)

Position = Trait('top left', _all_positions, editor=_position_editor)

###############################################################################
# Relative Size.
###############################################################################

_relative_sizes = {
    'tiny': 1,
    'small': 2,
    'medium': 4,
    'large': 6
}

_edit_relative_sizes = ['tiny', 'small', 'medium', 'large']

_relative_size_editor = EnumEditor(
    cols = 4,
    values = _edit_relative_sizes
)

# Trait definition for symbol size.
RelativeSize = Trait(
    'medium',
    TraitPrefixMap(_relative_sizes),
    editor=_relative_size_editor
)

###############################################################################
# Rotation.
###############################################################################

# FIXME : Seems like we want a better editor.
Rotation = Trait(0, [0, 90])

###############################################################################
# Symbol.
###############################################################################

symbol_styles = [
    'circle',
    'square',
    'diamond',
    'triangle',
    'inverted triangle',
    'cross',
    'plus',
    'dot',
    'pixel',
    'custom'
]

Symbol = Trait(
    'square',
    TraitPrefixList(symbol_styles),
    cols = 3
)

AnnotationTypeTrait = Enum('string')

#### EOF ######################################################################
