#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

# Major package imports:

import weakref

from enthought.util.scipyx \
    import *

# Enthought library imports:

from enthought.traits.api \
    import Any, Array, Delegate, Dict, Instance, Trait, TraitError, Bool

from enthought.traits.ui.api \
    import View

from enthought.kiva \
    import CompiledPath

# Local imports:

from plot_base \
    import plot_delegate, sequence_types, bin_search1, ArrayType, PlotItem, \
           TOP, BOTTOM, LEFT, RIGHT

from plot_container \
    import DefaultContainer

from plot_canvas \
    import PlotCanvas

from plot_data \
    import PlotData

from plot_indexed \
    import PlotIndexed

from plot_overlay \
    import PlotOverlay

from plot_source \
    import PlotSource

from plot_style \
    import PlotStyle

from plot_traits \
    import AnnotationTypeTrait

from plot_views \
    import bar_width_group, line_group, marker_group, plot_group, \
           border_group, plot_margin_group, cursor_group, misc_group

from backbuffer_plotter \
    import BackBuffer

#-------------------------------------------------------------------------------
#  Make sure a PlotValue's 'data' is a valid Numeric scalar array or a
#  PlotSource data source:
#-------------------------------------------------------------------------------

def data_or_array ( object, name, value ):
    if isinstance( value, PlotData ):
        value = value.get_source( object )
    if isinstance( value, PlotSource ):
        source = object._source
        if (source is not None) and (source != value):
            source.remove()
        object._source = value
        value.value    = object
        value          = value.data()
    type_value = type( value )
    if isinstance(value, ArrayType):
        shape = value.shape
        if len( shape ) == 1:
            return value
        elif len( shape ) == 2:
            if shape[1] == 2:
                axis = 1
            elif shape[0] == 2:
                axis = 0
            else:
                raise TraitError
            object.index = PlotValue( ravel( take( value, ( 0, ), axis ) ) )
            return ravel( take( value, ( 1, ), axis ) )
    elif type_value in sequence_types:
        try:
            if len( value ) > 0:
                value_type = type( value[0] )
                if value_type in sequence_types or isinstance(value, ArrayType):
                    if ((len( value ) == 2)  and
                        (len( value[0] ) == len( value[1] ))):
                        object.index = PlotValue( value[0][:] )
                        value        = value[1][:]
                    elif isinstance(value[0][0], basestring):
                        object.index = PlotValue( [ x[0] for x in value ] )
                        value        = [ x[1] for x in value ]
                elif isinstance(value_type, basestring):
                    object._labels = value[:]
                    value = range( len( value ) )
            return data_or_array( object, name, array( value, Float ) )
        except:
            pass
    raise TraitError

data_or_array.info = ('a sequence or tuple of numbers, a PlotSource object, '
                      'or an array whose shape is either (n), (n,2) '
                      'or (2,n)')

class PlotValue ( PlotCanvas ):
    """ A collection of data to be plotted.

    A PlotValue object represents an array of related, plottable data. It can
    either be plotted by itself, or added to a PlotCanvas object where it can
    be plotted along with other PlotValue objects on a single plotting area,
    or canvas.

    PlotValue objects can also be added to PlotGrid, PlotGroup, and PlotIndexed
    objects, although in these cases the plot will not overlap any other
    plotting area.

    A PlotValue can also be added to a PlotValue, PlotCanvas, PlotGroup or
    PlotIndexed object with its overlay trait set to true. In this case, the
    PlotValue will be drawn as a separated plot overlaying the object it is
    added to.

    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # The data to be plotted.
    data           = Trait( array( [ 0.0, 1.0 ] ), data_or_array,
                            rich_compare = False,
                            desc         = 'the data to be plotted' )

    style          = Instance(PlotStyle, ())

    ### Style traits. #########################################################

    # For bar charts, the width of a bar.
    bar_width      = Delegate('style', 'plot_*')

    # For bar charts, the width of a group of bars.
    bin_width      = Delegate('style', 'plot_*')

    # For line type plots, the style of the line (e.g. dash).
    line_style     = Delegate('style', 'plot_*')

    # For line type plots, the color of the line.
    line_color     = Delegate('style', 'plot_*')

    # For line type plots, the thickness of the line.
    line_weight    = Delegate('style', 'plot_*')

    antialias      = Bool(True)

    # The corner of the plotting area used as the axis origin.
    origin         = plot_delegate
    # FIXME : For now, don't delegate this to the style, since doing so screws
    # up track plots.

    # For plots with markers, the outline color of the marker.
    outline_color  = Delegate('style', 'plot_*')

    # For plots with markers, the thickness of the outline.
    outline_weight = Delegate('style', 'plot_*')

    # For plots with markers, the fill color of the marker.
    fill_color     = Delegate('style', 'plot_*')

    # For plots with markers, the fill style of the marker.
    fill_style     = Delegate('style', 'plot_*')

    # For plots with markers, the size of the marker.
    size           = Delegate('style', 'plot_*')

    # For plots with markers, the symbol (e.g. circle).
    symbol         = Delegate('style', 'plot_*')

    # The type of plot drawn.
    type           = Delegate('style', 'plot_*')

    custom_symbol  = Instance(CompiledPath)

    # The type of the annotation instances in self.annotations
    annotation_type = AnnotationTypeTrait('string')

    # The actual annotation dictionary mapping data indices to annotation
    # instances (of the type specified in annotation_type)
    annotations    = Dict

    # The color of annotations/labels on data points
    annotation_color = Delegate('style', 'plot_*')

    # The relative position of a plot annotation
    annotation_position = Delegate('style', 'plot_*')

    # Number of pixels of padding between the annotation labels and the plot
    annotation_padding = Delegate('style', 'plot_*')

    # The font to use for the annotation text
    annotation_font = Delegate('style', 'plot_*')

    # The size of the annotation text
    annotation_size = Delegate('style', 'plot_*')

    ### Backbuffer ############################################################

    backbuffer = Trait(None, Instance(BackBuffer))

    ### Interaction traits. ###################################################

##     # The type of range the user can drag select.
##     select_type    = delegate

##     # The routine to be called when drag selection is complete.
##     select_notify  = delegate

    #---------------------------------------------------------------------------
    #  Trait editor definition:
    #---------------------------------------------------------------------------

    traits_view = View(
        [ [ 'index_axis{Index Axis}', '_',
            'origin',                 '_',
            'type',
            '|[Plot]@' ],
          [ 'index_scale{Index Axis}',
            'value_scale{Value Axis}',
            '|[Scale Factor]' ],
          [ 'position', '_',
            'bg_color', '_',
            'visible',
            '|[Canvas]@' ],
        '|{Main}' ],
        [ [ 'weight',
            '|[Layout]' ],
          [ 'min_height{Height}',
            'min_width{Width}',
            '|[Canvas Minimum]' ],
          [ 'plot_min_height{Height}',
            'plot_min_width{Width}',
            '|[Plot Area Minimum]' ],
          bar_width_group,
          '|{Layout}' ],
        line_group,
        marker_group,
        plot_group,
        border_group,
        plot_margin_group,
        cursor_group,
        misc_group
    )

    canvas_traits_view = View(
        [ [ 'type', '_', 'visible',
            '|[Plot]@' ],
          '|{Main}' ],
        line_group,
        marker_group,
        [ [ 'menu', '|[Miscellaneous]' ], '|{Misc}' ]
    )

    indexed_traits_view = View(
        [ [ 'type', '_', 'value_scale{Value Axis Scale}', '_', 'visible',
            '|[Plot]@' ],
          '|{Main}' ],
        [ [ 'weight', '|[Layout]' ],
          [ 'plot_min_height{Height}',
            'plot_min_width{Width}',
            '|[Plot Area Minimum]' ],
          '|{Layout}' ],
        line_group,
        marker_group,
        plot_group,
        plot_margin_group,
        cursor_group,
        misc_group
    )

    def _style_changed(self, old, new):

        if new is not None:
            new.on_trait_change(self._style_trait_changed)

        if old is not None:
            old.on_trait_change(self._style_trait_changed, remove=True)

        return

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, data = None, **traits ):
        from plot_lists import PlotOverlayList, PlotTitleList, PlotValueList

        self._min          = 0.0
        self._max          = 1.0
        self._index_       = 1   # Force a 'cache miss' if '_index' called
        self._cached_index = None
        self._source       = None
        self._titles       = []
        self._overlays     = []
        self._values       = [ PlotItem( weakref.proxy( self ) ) ]
        self.titles        = PlotTitleList(   self, self._titles )
        self.overlays      = PlotOverlayList( self, self._overlays )
        self.values        = PlotValueList(   self, self._values )
        if data is not None:
            self.data = data

        from plottable import Plottable
        Plottable.__init__( self, **traits )


    #---------------------------------------------------------------------------
    #  Return the default traits view/name:
    #---------------------------------------------------------------------------

    def default_traits_view ( self ):
        """ Return the default traits view/name.
        """
        not_overlay = not self.overlay
        if not_overlay and isinstance( self.container, PlotCanvas ):
            view = 'canvas_traits_view'
        elif not_overlay and isinstance( self.container, PlotIndexed ):
            view = 'indexed_traits_view'
        else:
            view = 'traits_view'
#        source = self._source
#        if source is not None:
#            traits = traits[:]
#            traits.extend( source._editable_traits() )
#        return traits
        return view

    #---------------------------------------------------------------------------
    #  Return the list of PlotValue objects whose representation intersects (or
    #  is within a small delta of) a specified (x,y) point:
    #---------------------------------------------------------------------------

    def values_at ( self, x, y ):
        return []

    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in the PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------

    def map_xy ( self, x, y = None ):
        if y is None:
            x, y = self.host.event_xy( x, y )
        return self._plot_renderers[0]._map_xy( self, x, y )

    #---------------------------------------------------------------------------
    #  Map an index data value (or range of index values) to the closest range
    #  of indices in the index data.
    #---------------------------------------------------------------------------

    def map_data ( self, low, high = None ):

        # We can only calculate the closest index if the index data is sorted:
        index = self._index()
        dir   = index._get_dir()
        if dir == 0:
            return ( None, None )

        # If no high value, map it to the low value:
        if high is None:
            high = low

        # If they are in the wrong order, interchange them:
        elif low > high:
            low, high = high, low

        # Find the index into the index data that brackets 'low/high':
        if dir > 0:
            first = 0;  last = -1
        else:
            first = -1; last = 0

        data = index.data
        low  = min( max( low,  data[ first ] ), data[ last ] )
        high = min( max( high, data[ first ] ), data[ last ] )
        i    = bin_search1( data, low,  dir )
        j    = bin_search1( data, high, dir )

        # Now round the indices to the closest matching index:
        last = len( data ) - 1

        if i < last:
            delta = float( data[ i + 1 ] - data[ i ] )
            if delta != 0.0:
                i += int( round( float( low - data[ i ] ) / delta ) )

        if j < last:
            delta = float( data[ j + 1 ] - data[ j ] )
            if delta != 0.0:
                j += int( round( float( high - data[ j ] ) / delta ) )

        # Return the final range:
        return ( min( i, j ), max( i, j ) + 1 )

    #---------------------------------------------------------------------------
    #  Unselect any current range selection:
    #---------------------------------------------------------------------------

    def unselect ( self ):
        self.host._select_none( self )

    #---------------------------------------------------------------------------
    #  Return the user interface description of the object:
    #---------------------------------------------------------------------------

    def _ui_label ( self ):
        return (self.label or
                (self._source and self._source._ui_label()) or
                self.__class__.__name__[4:])

    #---------------------------------------------------------------------------
    #  Handle the 'data' trait changing value:
    #---------------------------------------------------------------------------

    def _data_changed ( self, old = None, new = None ):
        self._set_range()

    #---------------------------------------------------------------------------
    #  Set the array min/max values:
    #---------------------------------------------------------------------------

    def _set_range ( self ):
        data = compress( isfinite( self.data ), self.data )
        if len( data ) > 0:
            self._min = amin( data )
            self._max = amax( data )
        else:
            self._min = 0.0
            self._max = 1.0
        self._dir       = None
        self._xy_ranges = None

    #---------------------------------------------------------------------------
    #  Return whether a new object can be added to the value:
    #---------------------------------------------------------------------------

    def _can_add ( self, item ):
        return False

    #---------------------------------------------------------------------------
    #  Return a list of all contained PlotValues:
    #---------------------------------------------------------------------------

    def _all_values ( self ):
        return self._values

    #---------------------------------------------------------------------------
    #  Get the index object associated with the value:
    #---------------------------------------------------------------------------

    def _index ( self ):
        index = self.index
        if ((index == self._index_) and
            (len( self.data ) == len( self._cached_index.data ))):
            return self._cached_index
        self._index_ = index
        if (index is None) or (len( self.data ) != len( index.data )):
            if index is not None:
                self.add( PlotOverlay(
                    'len(index) != len(value)',
                    bg_color    = 'red',
                    color       = 'white',
                    position    = 'center',
                    border_size = 1 ) )
            index = PlotValue( arange( float( len( self.data ) ) ) )
            index.container = self
        self._cached_index = index
        return index

    #---------------------------------------------------------------------------
    #  Calculate the minimum size of the plotting area needed:
    #---------------------------------------------------------------------------

    def _min_size_plot ( self ):
        return ( self.min_width, self.min_height )

    #---------------------------------------------------------------------------
    #  Map from data space to screen space:
    #
    #  Note: 'index' and 'value' can be either scalar, arrays, or None.
    #  If they are None, the corresponding part of the result is also None.
    #  The 'index_offset' and 'value_offset' arguments allow an offset to be
    #  applied to the result 'after' conversion to screen space.
    #---------------------------------------------------------------------------

    def map_screen ( self, index, value, index_offset = 0.0, value_offset = 0.0,
                           xy = False ):

        # Get the x and y data values and axis information:
        x_axis = self._axis_info_index
        y_axis = self._axis_info

        # Do the axis 'log' scaling if necessary:
        if y_axis.log_scale and (value is not None):
            value = log10( value )

        # If the index axis is vertical, flip-flop the values:
        if not x_axis.horizontal:
            index, value               = value, index
            index_offset, value_offset = value_offset, index_offset
            x_axis, y_axis             = y_axis, x_axis

        xp = None
        if index is not None:
            # Scale and translate the x data:
            dxmin = x_axis.tick_min
            ddx   = x_axis.tick_max - dxmin

            # We shouldn't have to worry about this here (?):
            if ddx == 0.0:
                ddx    = 1.0
                dxmin -= 0.5

            # Do the actual scaling/translation of the data:
            if x_axis.origin & LEFT:
                xp = (self._plot_bounds[0] - self._plot_ox + ((index - dxmin) *
                           (float( self._plot_sxdx - 1 ) / ddx))) + index_offset
            else:
                xp = (((self._plot_bounds[0]-self._plot_ox+self._plot_sxdx-1) -
                     ((index - dxmin) * (float( self._plot_sxdx - 1 ) / ddx))) -
                     index_offset)

        # If the value is None, then we are done, return the result:
        if value is None:
            return ( xp, None )

        # Scale and translate the y data:
        dymin = y_axis.tick_min
        ddy   = y_axis.tick_max - dymin

        # We shouldn't have to worry about this here (?):
        if ddy == 0.0:
            ddy    = 1.0
            dymin -= 0.5

        # Do the actual scaling/translation of the data:
        if y_axis.origin & TOP:
            yp = (((self._plot_bounds[1]-self._plot_oy + self._plot_sydy - 1) -
                  ((value - dymin) * (float( self._plot_sydy - 1 ) / ddy))) -
                  value_offset)
        else:
            yp = (self._plot_bounds[1] - self._plot_oy + ((value - dymin) *
                           (float( self._plot_sydy - 1 ) / ddy))) + value_offset

        # Return the result:
        if xy:
            return concatenate( ( xp[:, NewAxis ], yp[:, NewAxis ] ), axis = -1)
        return ( xp, yp )

    #---------------------------------------------------------------------------
    #  Draw the plot value:
    #---------------------------------------------------------------------------

    def _draw ( self, gc ):
        PlotCanvas._draw( self, gc, False )

    #---------------------------------------------------------------------------
    #  Return the direction the data values are sorted in (if any):
    #   1 = Monotonically increasing,
    #  -1 = Monotonically decreasing
    #   0 = Not sorted
    #---------------------------------------------------------------------------

    def _get_dir ( self ):
        if self._dir is None:
            try:
                diff      = self.data[1:] - self.data[:-1]
                self._dir = logical_and.reduce( greater_equal( diff, 0.0 ) )
                if not self._dir:
                    self._dir = -logical_and.reduce( less_equal( diff, 0.0 ) )
            except:
                self._dir = 0
        return self._dir

    def _style_trait_changed(self):
        """ Called when any trait on the style sheet is changed. """

        self.update()

        return

    #---------------------------------------------------------------------------
    #  Methods to make a PlotValue behave as an alias for its 'data' array:
    #---------------------------------------------------------------------------

    def __len__ ( self ):
        return len( self.data )

    def __getitem__ ( self, index ):
        return self.data[ index ]

    def __setitem__ ( self, index, value ):
        if self.data[ index ] != value:
            self.data[ index ] = value
            self._set_range()

    def __delitem__ ( self, index ):
        temp = list( self.data )
        del temp[ index ]
        if len( temp ) == 0:
            temp = [ 0.0 ]
        self.data = temp

    def __getslice__ ( self, index1, index2 ):
        return self.data[ index1: index2 ]

    def __setslice__ ( self, index1, index2, values ):
        temp = list( self.data )
        temp[ index1: index2 ] = list( values )
        if len( temp ) == 0:
            temp = [ 0.0 ]
        self.data = temp

    def __delslice__ ( self, index1, index2 ):
        temp = list( self.data )
        del temp[ index1: index2 ]
        if len( temp ) == 0:
            temp = [ 0.0 ]
        self.data = temp

    def __contains__ ( self, value ):
        return value in self.data

class SymmetricErrorbarPlotValue ( PlotValue ):
    """
    Represents a PlotValue that has a certain amount of error at each data
    point.  The amount of error is the same above and below the data point,
    and this amount is stored in the .error_value array, which is the same
    length as the .data array.
    """

    error_value = Array


class ErrorbarPlotValue ( PlotValue ):
    """
    Represents a PlotValue that has a certain amount of error at each data
    point.  At each datapoint, the amount of error above and below may
    differ.  The error values forming the upper bound are stored in the
    .high_value array, and the lower bound values are stored in .low_value.
    Both of these error value arrays are the same length as .data.
    """

    low_value = Array
    high_value = Array


