"""
Defines ColorMappedPlotValue, used for representing colormapped scatter-plot
data.

See renderers/colormappedscatter.py for the Renderer that uses this plotvalue.
See demo/scatterplot_colormap.py for a demonstration for a usage example.
"""

# We need weakref in order to properly implement the PlotValue __init__().
import weakref


# Enthought library imports
from enthought.util.numerix import arange, argsort, concatenate, nonzero, searchsorted, take
from enthought.traits.api import Array, Dict, Enum, false, Instance, List, Property

# local imports
from plot_value import PlotValue
from datamapper import left_shift, right_shift
from colormap import AbstractColormap
from plot_base import PlotItem


class ColorMappedPlotValue(PlotValue):
    """
    This subclass of PlotValue represents a set of (X,Y,V) values, where
    XY should be plotted using a normal scatter plot, and V should be mapped
    to color via a colormap.
    """

    #---------------------------------------------------------------------------
    # Public traits
    #---------------------------------------------------------------------------
    
    # The Nx3 array of (x,y,v) points to be plotted.  Because of the specific
    # nature of the data that a colormapped scatter plot requires, this
    # data attribute is not flexible like the parent PlotValue's data attribute.
    # It also lacks the automagical 'massaging' of data.
    data = Property
    
    # A colormap to apply to the values of the plotvalue.  For this to
    # have any effect, a colormapping renderer must be used.
    colormap = Instance(AbstractColormap)

    #---------------------------------------------------------------------------
    # Private traits
    #---------------------------------------------------------------------------
    
    # Shadow trait for the (x,y) data series
    _data = Array
    
    # Shadow trait for the Z axis (that gets mapped to color)
    _zdata = Array
    
    # Because of the cost of the colormapping operation, we cache a number
    # of the results.  The _cache_valid flag refers to the validity
    # of all of them.
    _cache_valid = false

    # The _render_method tells the renderer how it should render this
    # plotvalue:
    #   bruteforce: set the stroke color before drawing each marker
    #   bands:      draw the points colorband by colorband, thus reducing
    #               the number of set_stroke_color() calls
    #
    # TODO: Based on preliminary results, "branded" isn't significantly
    #       more expensive than "bruteforce" for small datasets (<1000),
    #       so perhaps bruteforce should be removed.
    _render_method = Enum("bruteforce", "banded")

    # A dict mapping colormap indices to arrays of indices into self.data.
    # This is used for the "banded" render method.
    # This is only valid if _cache_valid is True.
    _index_bands = Dict
    
    # An array of the color index corresponding to each data point.
    # This is used by the "bruteforce" render method.
    _color_indices = Array
    
    def __init__ ( self, data = None, **traits ):
        from plot_lists import PlotOverlayList, PlotTitleList, PlotValueList

        self._min          = 0.0
        self._max          = 1.0
        self._index_       = 1   # Force a 'cache miss' if '_index' called
        self._cached_index = None
        self._source       = None
        self._titles       = []
        self._overlays     = []
        self._values       = [ PlotItem( weakref.proxy( self ) ) ] 
        self.titles        = PlotTitleList(   self, self._titles )
        self.overlays      = PlotOverlayList( self, self._overlays )
        self.values        = PlotValueList(   self, self._values )
        if data is not None:
            self.data = data

        from plottable import Plottable
        Plottable.__init__( self, **traits )


    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------
    
    def _calc_render_method(self, numpoints):
        if numpoints > 300:
            return 'banded'
        else:
            return "bruteforce"

    def _compute_bands(self, smartmode=False):
        """
        Sorts self.data into a list of arrays of data points by color,
        filling in self._index_bands.  If 'smartmode' is True, then first
        calls _calc_render_method() to see which rendering method is
        optimal for the number of points and the distribution of 
        color indices; if the rendering method is 'bruteforce', then
        _compute_bands() will short-circuit and return without doing
        anything.
        """
        if self.colormap is None:
            return
            
        # map the V values in the (x,y,v) self.data array
        color_indices = self.colormap.map_array_index(self._zdata)
        
        self._render_method = self._calc_render_method(len(self._zdata)) 
        if smartmode and self._render_method == 'bruteforce':
             pass
        else:
            # shuffle_indices indicates how to sort the points in self.data
            # so that their color_indices are in order.  We don't really care
            # about the sorting so much as the fact that once they are sorted,
            # points of the same color are grouped together into "bands".
            shuffle_indices = argsort(color_indices)
            
            # This pulls values from the color_indices array into 
            # sorted_color_indices, using the results of the sort we just did.
            sorted_color_indices = take(color_indices, shuffle_indices)
            
            # Now we want to determine where the continuous bands are.  We do
            # this by right-shifting the sorted_color_indices array, subtracting
            # it from the original, and looking for all the nonzero points.
            shifted = right_shift(sorted_color_indices, sorted_color_indices[0])
            start_indices = concatenate([[0], nonzero(sorted_color_indices - shifted)])
            end_indices = left_shift(start_indices, len(sorted_color_indices))

            # Store the shuffled indices in self._index_bands.  We don't store the
            # actual data points because we need to allow the renderer to index into
            # the mapped XY screen positions.
            for (start, end) in zip(start_indices, end_indices):
                color_index = sorted_color_indices[start]
                self._index_bands[color_index] = shuffle_indices[start:end]
        
        self._color_indices = color_indices
        self._cache_valid = True

    def _set_data(self, newval):
        self._cache_valid = False
        self.index = PlotValue(newval[:, 0])
        self._data = newval[:, 1]
        self._zdata = newval[:, 2]
        self._set_range()
        self._compute_bands(True)

    def _get_data(self):
        return self._data


#EOF
