#-------------------------------------------------------------------------------
#
#  Define non-trait based items of use to Chaco applications using a GUI
#  interface
#
#  Written by: David C. Morrill
#
#  Date: 01/02/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#  Classes defined: PlotTraitCommand
#                   PlotAddCommand
#                   PlotRemoveCommand
#                   PlotRemoveAxisCommand
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from plot_base import user_name_for

#-------------------------------------------------------------------------------
#  'PlotTraitCommand' class:
#-------------------------------------------------------------------------------

class PlotTraitCommand:
    
    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
    
    def __init__ ( self, object, trait_name, original_value, new_value ):
        self.object         = object
        self.trait_name     = trait_name
        self.original_value = original_value
        self.new_value      = new_value
        self.desc           = 'setting %s %s' % ( 
                                  user_name_for( object ), 
                                  trait_name.replace( '_', ' ' ) )
        
    #---------------------------------------------------------------------------
    #  'Undo' the command:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        setattr( self.object, self.trait_name, self.original_value )
        self.object.update()
    
    #---------------------------------------------------------------------------
    #  'Redo' the command:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        setattr( self.object, self.trait_name, self.new_value )
        self.object.update()

#-------------------------------------------------------------------------------
#  'PlotAddCommand' class
#-------------------------------------------------------------------------------

class PlotAddCommand:
    
    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
    
    def __init__ ( self, object, container ):
        self.object    = object
        self.container = container
        self.host      = container.host
        self.desc      = 'add %s' % user_name_for( object )
        
    #---------------------------------------------------------------------------
    #  Undo the command:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        self.container.remove( self.object )
        self.host.update()
    
    #---------------------------------------------------------------------------
    #  Redo the command:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        self.container.add( self.object )
        self.host.update()

#-------------------------------------------------------------------------------
#  'PlotRemoveCommand' class
#-------------------------------------------------------------------------------

class PlotRemoveCommand:
    
    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
    
    def __init__ ( self, object, container ):
        import plot_frame   
        self.object    = object
        self.container = container
        self.host      = object.host
        self.desc      = 'remove %s' % user_name_for( object )
        if isinstance( object, plot_title.PlotTitle ):
            self.values = container.titles[:]
        elif isinstance( object, plot_overlay.PlotOverlay ):
            self.values = container.overlays[:]
        else:
            self.values = container.values[:]
        
    #---------------------------------------------------------------------------
    #  Undo the command:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        import plot_frame   
        if isinstance( self.object, plot_title.PlotTitle ):
            self.container.titles[:] = self.values
        elif isinstance( self.object, plot_overlay.PlotOverlay ):
            self.container.overlays[:] = self.values
        else:
            self.container.values[:] = self.values 
        self.host._add_item( self.container, self.object )
        self.host.update()
    
    #---------------------------------------------------------------------------
    #  Redo the command:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        self.object.remove()
        self.host.update()

#-------------------------------------------------------------------------------
#  'PlotRemoveAxisCommand' class
#-------------------------------------------------------------------------------

class PlotRemoveAxisCommand:
    
    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
    
    def __init__ ( self, object ):
        self.object = object
        self.axis   = object.axis
        self.host   = object.host
        self.desc   = 'remove %s axis' % user_name_for( object )
        
    #---------------------------------------------------------------------------
    #  Undo the command:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        self.object.axis = self.axis
        self.host.update()
    
    #---------------------------------------------------------------------------
    #  Redo the command:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        del self.object.axis
        self.host.update()
