"""Common base class for plot elements."""

# Enthought library imports.
from enthought.traits.api import Any, Delegate, Instance, ReadOnly
from enthought.traits.ui.api import Group, View

# Local imports.
from plottable_style import PlottableStyle
from plot_base import simple_delegate, ArrayType, InstanceType, \
                      sequence_types, \
                      PlotError, PlotItemList, \
                      NEEDS_MIN_SIZE
from plot_container import AXES_INFO, TITLE_INFO, OVERLAY_INFO,  NestedContainer
from plot_data import PlotData
from plot_frame import PlotFrame
from plot_overlay import PlotOverlay
from plot_title import PlotTitle
from plot_traits import HORIZONTAL, VERTICAL
from plot_views import layout_group, border_group, plot_group, margin_group, \
                       cursor_group, misc_group


StyleDelegate = Delegate('plottable_style')


class Plottable ( PlotFrame, NestedContainer ):
    """ Common base class for plot elements. """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    titles          = ReadOnly
    overlays        = ReadOnly
    values          = ReadOnly

    plottable_style = Instance(PlottableStyle, ())

    overlay         = StyleDelegate
    position        = StyleDelegate

    index_scale     = StyleDelegate
    value_scale     = StyleDelegate

    _plot_renderers = simple_delegate

    _plot_cache     = Any


    #---------------------------------------------------------------------------
    # Traits view definitions - We define several different ones, and the
    # default_traits_view() method determines which one we return.
    #---------------------------------------------------------------------------

    # 'Plottable' main page common traits:
    plottable_main_group1 = Group(
        'index_axis',          '_',
        'plot_origin{Origin}', '_',
        'plot_type{Type}',
        '|[Plot]@'
    )

    plottable_main_group2 = Group(
        'index_scale{Index Axis}',
        'value_scale{Value Axis}',
        '|[Scale Factor]'
    )

    traits_view = View(
        [ plottable_main_group1,
          plottable_main_group2,
          [ 'bg_color{Background Color}', '_', 'visible',
            '|[Canvas]@' ],
          '|{Main}' ],
        layout_group,
        border_group,
        plot_group,
        margin_group,
        cursor_group,
        misc_group
    )

    overlay_traits_view = View(
        [ plottable_main_group1,
          plottable_main_group2,
          [ 'position', '_', 'bg_color{Background Color}', '_', 'visible',
            '|[Canvas]@' ],
          '|{Main}' ],
        layout_group,
        border_group,
        plot_group,
        margin_group,
        cursor_group,
        misc_group
    )


    #---------------------------------------------------------------------------
    # Public methods
    #---------------------------------------------------------------------------

    def default_traits_view ( self ):
        "Return the name of the default traits view"
        if self.overlay:
            return 'overlay_traits_view'
        else:
            return 'traits_view'


    def add ( self, *items ):
        "Add new objects to this object"
        from plot_value import PlotValue
        for item in items:
            type_item = type( item )
            if isinstance(type_item, basestring):
                item = PlotTitle( item )
            elif type_item == ArrayType:
                item = PlotValue( item )
            elif type_item in sequence_types:
                if len( item ) == 0:
                    continue
                type_item = type( item[0] )
                if ((type_item in numeric_types) or
                    (type_item in sequence_types)):
                    item = PlotValue( item )
                else:
                    for element in item:
                        self.add( element )
                    continue
            if isinstance( item, PlotItemList ):
                for element in item[:]:
                    self.add( element )
            elif isinstance( item, PlotTitle ):
                self._add_item( self._titles, item )
            elif isinstance( item, PlotOverlay ):
                self._add_item( self._overlays, item )
            elif isinstance( item, Plottable ) and item.overlay:
                self._add_item( self._overlays, item )
            elif isinstance( item, PlotData ):
                item.add( self )
            elif self._can_add( item ):
                self._add_item( self._values, item )
            else:
                try:
                    item.add_to_plot( self )
                except:
                    if type( item ) == InstanceType:
                        kind = 'a %s' % item.__class__.__name__
                    else:
                        kind = repr( item )
                    raise PlotError, 'Cannot add %s to a %s' % (
                                     kind, self.__class__.__name__ )

            # Indicate that any cached values are no longer valid:
            self._plot_cache = None


    def remove ( self, *items ):
        "Remove some objects from the plottable"
        # If no items were specified, remove ourself from our own container:
        if not items:
            self.container.remove( self )
            return

        # Otherwise, remove the items specified from ourself:
        for item in items:
            if type( item ) in sequence_types:
                for sub_item in item:
                    self.remove( sub_item )
                rc = True
            elif isinstance( item, PlotTitle ):
                rc = self._remove_item( self._titles, item )
            elif isinstance( item, PlotOverlay ):
                rc = self._remove_item( self._overlays, item )
            elif isinstance( item, Plottable ):
                if item.overlay:
                    rc = (self._remove_item( self._overlays, item ) or
                          self._remove_item( self._values,   item ))
                else:
                    rc = (self._remove_item( self._values,   item ) or
                          self._remove_item( self._overlays, item ))
            else:
                rc = self._remove_item( self._values, item )
                if not rc:
                    try:
                        item.remove_from_plot( self )
                        rc = True
                    except:
                        pass
            if not rc:
                if type( item ) == InstanceType:
                    kind = 'the %s' % item.__class__.__name__
                else:
                    kind = repr( item )
                raise PlotError, 'Could not remove %s from the %s' % (
                                 kind, self.__class__.__name__ )

            # Indicate that any cached values are no longer valid:
            self._plot_cache = None


    def render ( self, gc, x = 0, y = 0, dx = 640, dy = 480 ):
        "Render the object into the associated graphics context"
        self._min_size( gc )
        self._layout( x, y, dx, dy )
        self._draw( gc )


    def components_at ( self, x, y ):
        "Generate the set of components that contain a specified (x,y) point"
        if self.visible and self._drawable:
            result = self.values_at( x, y ) + self.objects_at( x, y )
            result.reverse()
            return result
        return []


    def objects_at ( self, x, y ):
        """
        Return the list of objects contained in the Plottable that contain a
        specified (x,y) point.

        Subclasses should not override this method; instead they should
        implement _objects_at().
        """
        if not self._isin( x, y ):
            return []
        plot_cache = self._plot_cache
        if plot_cache is None:
            return []
        result = plot_cache[ TITLE_INFO ].objects_at( x, y )
        axes_info = plot_cache[ AXES_INFO ]
        if axes_info is not None:
            result += [ info.axis for info in axes_info.objects_at( x, y ) ]
        result += self._objects_at( x, y )
        result = plot_cache[ OVERLAY_INFO ].objects_at( x, y ) + result
        result.append( self )
        return result


    def values_at ( self, x, y ):
        """
        Return the list of PlotValue objects whose representation intersects (or
        is within a small delta of) a specified (x,y) point
        """
        result = []
        if self.visible:
            for value_item in self._values:
                result = value_item.item.values_at( x, y )
                if result:
                    break
        return result


    #---------------------------------------------------------------------------
    # Abstract methods that subclasses should or must implement
    #---------------------------------------------------------------------------

    def _can_add ( self, item ):
        raise NotImplementedError


    def _objects_at ( self, x, y ):
        """
        Return the list of value objects contained in the Plottable that contain
        a specified (x,y) point
        """
        return []


    #---------------------------------------------------------------------------
    # Trait event handlers
    #---------------------------------------------------------------------------

    def _overlay_changed ( self, old = None, new = None ):
        self.container.add( self )
        return


    def __plot_cache_changed ( self, old, new ):
        """
        If the cache has been set to None, then we need to notify our window
        to call _min_size before layout.  This is required because _min_size
        has the side effect of initializing _plot_cache.
        """
        if new is None:
            self.update( NEEDS_MIN_SIZE )
        return


    #---------------------------------------------------------------------------
    # Private methods
    #---------------------------------------------------------------------------

    def _scrollbar_size ( self, index_axis ):
        "Determine the amount of space needed for scrollbars"
        ###KW_TODO###

        sbdx = sbdy = 0

        if (self.index_scale > 1.0) and (self.index_scrollbar is None):
            if index_axis == 'x':
                sbdy = self.host.scrollbar_dy + self.padding_height
            else:
                sbdx = self.host.scrollbar_dx + self.padding_width

        if (self.value_scale > 1.0) and (self.value_scrollbar is None):
            if index_axis == 'x':
                sbdx = self.host.scrollbar_dx + self.padding_width
            else:
                sbdy = self.host.scrollbar_dy + self.padding_height

        return ( sbdx, sbdy )


    def _layout_scrollbars ( self, sbdx, sbdy, sx, sy, sbx, sby,
                                      x,  y,  dx,  dy,
                                     px, py, pdx, pdy, mx, my ):
        "Layout the scrollbars needed by the canvas"
        ###KW_TODO###

        self._plot_ox = self._plot_oy = 0
        if sbdx > 0:
            self.host._scrollbar_for( self, VERTICAL, sy, pdy,
                                    x + dx - mx - sbdx + self.padding_width, py,
                                    self.host.scrollbar_dx, pdy )
        elif sby is not None:
            self.host._init_scrollbar( sby, self, VERTICAL, sy, pdy )
        if sbdy > 0:
            self.host._scrollbar_for( self, HORIZONTAL, sx, pdx,
                                      px, y + my, pdx, self.host.scrollbar_dy )
        elif sbx is not None:
            self.host._init_scrollbar( sbx, self, HORIZONTAL, sx, pdx )


    def _xy_scale ( self, index_axis = None ):
        "Return the x and y axes scaling factors"
        if index_axis is None:
            index_axis = self.index_axis
        if index_axis == 'x':
            return ( self.index_scale,     self.value_scale,
                     self.index_scrollbar, self.value_scrollbar )
        return ( self.value_scale,     self.index_scale,
                 self.value_scrollbar, self.index_scrollbar  )


    def _cursor_bounds ( self ):
        "Returns the bounds that the component's cursor should be clipped against"
        return self._plot_bounds


    def _pre_mouse_wheel_changed ( self, event ):
        "Vertically scrolls the plot when the mouse scroll wheel is spun"
        sb_for = self.host.owner._scrollbar_for
        sb     = sb_for( self, VERTICAL, False )
        if sb is None:
            sb = sb_for( self, HORIZONTAL, False )
        if sb is not None:
            event.handled = True
            sb.position  += (event.mouse_wheel * sb.page_size) / 20
        return


    #---------------------------------------------------------------------------
    #  Methods to make object behave as an alias for its 'values' array:
    #---------------------------------------------------------------------------

    def __len__ ( self ):
        return len( self.values )

    def __getitem__ ( self, index ):
        return self.values[ index ]

    def __setitem__ ( self, index, value ):
        self.values[ index ] = value

    def __delitem__ ( self, index ):
        del self.values[ index ]

    def __getslice__ ( self, index1, index2 ):
        return self.values[ index1: index2 ]

    def __setslice__ ( self, index1, index2, values ):
        self.values[ index1: index2 ] = values

    def __delslice__ ( self, index1, index2 ):
        del self.values[ index1: index2 ]

    def __contains__ ( self, value ):
        return value in self.values


# EOF
