#-------------------------------------------------------------------------------
#
#  Base code for Chaco interactive plotting classes
#
#  Written by: David C. Morrill
#
#  Date: 01/02/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import re
import os.path
from types import NoneType

from enthought.util.scipyx.fastumath import *
from enthought.util.scipyx import limits, nan_to_num
from enthought.util.numerix import asarray, reshape, transpose, clip, size, ravel

from enthought.traits.api     import Any
from enthought.traits.standard import flexible_false_trait

from enthought.kiva.agg   import Image

from plot_axis            import PlotAxis
from plot_base            import *
from plot_canvas          import PlotCanvas
from plot_container       import default_container
from plot_grid            import PlotGrid
from plot_group           import PlotGroup
from plot_overlay         import PlotOverlay
from plot_title           import PlotTitle
from plot_value           import PlotValue
from plot_value_cmap      import ColorMappedPlotValue
from plot_window_base     import PlotTraitCommand,  PlotAddCommand, \
                                 PlotRemoveCommand, PlotRemoveAxisCommand
from image_plot_value     import AbstractImagePlotValue, CmapImagePlotValue, \
                                 ImageData, ImagePlotValue
from default_colormaps    import gray, prism

from alphawrap_colormap   import AlphaWrapColormap

from interaction import PanZoomInteraction

from enthought.chaco.contour.contour_data import ContourData

from enthought.util.numerix import around

from enthought.chaco.default_colormaps import *


#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

DEBUG = True

# Application title:
app_title = 'Chaco Plot Window: %s'

# Numerical constants:
BIG   = 1.0e20
SMALL = 1.0e-20

# Exceptions:
SizeMismatch   = 'SizeMismatch'
NotImplemented = 'NotImplemented'

# Color schemes:
color_schemes = [
    ( 'blue',    'solid' ),
    ( 'red',     'solid' ),
    ( 'black',   'solid' ),
    ( 'magenta', 'solid' ),
    ( 'blue',    'dot'   ),
    ( 'red',     'dot'   ),
    ( 'black',   'dot'   ),
    ( 'magenta', 'dot'   ),
    ( 'blue',    'dash'  ),
    ( 'red',     'dash'  ),
    ( 'black',   'dash'  ),
    ( 'magenta', 'dash'  )
]

# Standard images sizes:
image_sizes = (
    (  320,  240 ),
    (  640,  480 ),
    (  800,  600 ),
    ( 1024,  768 ),
    ( 1280, 1024 ),
    ( 1600, 1200 )
)

# Valid save file types:
file_types = [ '.pdf', '.png', '.jpg', '.pcx', '.tif' ]

simple_types = ( str, float, int )

#-------------------------------------------------------------------------------
#  Main Menu Description:
#-------------------------------------------------------------------------------

plot_frame_menu = """
    &File
        Save as
            Single page...:         self.create_file(None,0)
            One canvas per page...: self.create_file(None,1)
            One value per page...:  self.create_file(None,2)
        ---
        Close | Ctrl-Q:             self.on_exit()
    &Edit
        Clear:                      self.clear_canvas()
    &Options
        Image size
            320 x 240  : self.set_image_size(0)
            640 x 480  : self.set_image_size(1)
            800 x 600  : self.set_image_size(2)
           1024 x 768  : self.set_image_size(3)
           1280 x 1024 : self.set_image_size(4)
           1600 x 1200 : self.set_image_size(5)
"""

#-------------------------------------------------------------------------------
#  Object menus:
#-------------------------------------------------------------------------------

# PlotCanvas menu:
canvas_menu = """
    Add title:                  self.user_handler.add_title(self,1)
    Add overlay:                self.user_handler.add_overlay(self,1)
    Save as
        Single page...:         self.user_handler.create_file(self,0)
        One value per page...:  self.user_handler.create_file(self,2)
    Edit...:                    self.edit()
"""

# PlotValue menu:
value_menu = """
    Add axis:             self.user_handler.add_axis(self,1)
    Remove axis:          self.user_handler.remove_axis(self)
    Add title:            self.user_handler.add_title(self,1)
    Add overlay:          self.user_handler.add_overlay(self,1)
    Save as
        Single page...:   self.user_handler.create_file(self,0)
    Remove:               self.user_handler.remove_object(self)
    Edit...:              self.edit()
"""

# PlotOverlay, PlotTitle and PlotLabel menu:
default_menu = """
    Remove:  self.user_handler.remove_object(self)
    Edit...: self.edit()
"""

# PlotAxis menu:
axis_menu = """
    Edit...: self.edit()
"""
#-------------------------------------------------------------------------------
#  Regular expressions:
#-------------------------------------------------------------------------------

color_re    = re.compile( '[ymcrgbwk]' )
color_trans = {
    'y': 'yellow',
    'm': 'magenta',
    'c': 'cyan',
    'r': 'red',
    'g': 'green',
    'b': 'blue',
    'w': 'white',
    'k': 'black'
}

# This one isn't quite right:
marker_re    = re.compile( '[ox+s^v]|(?:[^-])[.]' )
marker_trans = {
    '.': 'dot',
    'o': 'circle',
    'x': 'cross',
    '+': 'plus',
    's': 'square',
    '^': 'triangle',
    'v': 'down triangle'
}

line_re    = re.compile( '--|-\.|[-:]' )
line_trans = {
    '-':  'solid',
    ':':  'dot',
    '-.': 'dot dash',
    '--': 'dash'
}

_colormap = None

#-------------------------------------------------------------------------------
#  Data:
#-------------------------------------------------------------------------------

# List of current plot windows:
_figure = []

# Currently active plot window:
_active = None

# Currently selected object:
object = None

#-------------------------------------------------------------------------------
#  Define/Select a plot window:
#-------------------------------------------------------------------------------

def figure ( which_one = None ):
    global _figure, _active

    if which_one is None:
        title = 'Figure %d' % ( len( _figure ) + 1 )
        _figure.append( ShellPlotFrame_factory( title = title ) )
        _active = _figure[-1]

    elif type( which_one ) in ( IntType, FloatType ):
        try:
            _active = _figure[ int( which_one ) - 1 ]
        except IndexError:
            msg = "There are currently only %d active figures" % len( _figure )
            raise IndexError, msg

    elif which_one in _figure:
        _active = which_one

    else:
        try:
            if which_one.__type_hack__ == "plot_canvas":
                _active = which_one
                _figure.append( _active )
            else:
                raise ValueError, \
                      "The specified figure or index is not not known"
        except AtrributeError:
            pass

    _active.raise_window()
    return current()

#-------------------------------------------------------------------------------
#  Make sure that there is an active window:
#-------------------------------------------------------------------------------

def _validate_active ( ):
    global _active
    if _active is None:
        figure()
    try:
        if not _active.is_alive():
            _active = None
            figure()
    except:
        pass
    return _active

#-------------------------------------------------------------------------------
#  Return the currently active plot window:
#-------------------------------------------------------------------------------

def current ( ):
    return _active

#-------------------------------------------------------------------------------
#  Make sure the currently active plot window is updated:
#-------------------------------------------------------------------------------

def redraw ( ):
    _validate_active().update_window()

draw = redraw  # MatLab equivalence

#-------------------------------------------------------------------------------
#  Close a specified plot window:
#-------------------------------------------------------------------------------

def close ( which_one = None ):
    global _figure, _active

    if which_one is None:
        try:
            _active.close_window()
            _figure.remove( _active )
        except ValueError:
            pass

        try:
             # Should make sure the new plot window really exists:
             _set_new_active()
        except IndexError:
             _active = None

    elif which_one == 'all':
        for fig in _figure[:]:
            fig.close_window()
        _figure = []
        _active = None

    else:
        raise NotImplementedError, ( "Currently close only works with the"
                                     "active window or 'all'" )

#-------------------------------------------------------------------------------
#  Handle the user closing a plot window:
#-------------------------------------------------------------------------------

def _figure_closed ( which_one ):
    _figure.remove( which_one )
    if which_one == _active:
        _set_new_active()

#-------------------------------------------------------------------------------
#  Set a new active plot window:
#-------------------------------------------------------------------------------

def _set_new_active ( ):
    global _figure, _active

    # Should validate new active here:
    try:
        _active = _figure[-1]
        _active.raise_window()
    except IndexError:
        _active = None

#-------------------------------------------------------------------------------
#  Enable auto-scaling for the currently active plot window:
#-------------------------------------------------------------------------------

def autoscale ( ):
    _validate_active().auto_scale()

#-------------------------------------------------------------------------------
#  Set the range for a specified axis for the currently active plot window:
#-------------------------------------------------------------------------------

def _an_axis ( axis, arange ):
    if (len( arange ) == 0) or (arange[0] in [ 'normal', 'auto' ]):
        axis.set( bound_low  = 'auto',
                  bound_high = 'auto' )
    elif arange[0] in [ 'tight', 'fit' ]:
        axis.set( bound_low     = 'fit',
                  bound_high    = 'fit',
                  tick_interval = 'auto' )
###     elif arange[0] == 'freeze':
###        interval = axis.ticks[1] - axis.ticks[0]
###        axis.bounds = [axes[0],axes[1]]
###        axis.tick_interval = interval
    else:
        axis.set( bound_low  = arange[0],
                  bound_high = arange[1] )
        if len( arange ) > 2:
            axis.tick_interval = arange[2]
    redraw()

#-------------------------------------------------------------------------------
#  Set the currently active plot window's x-axis range:
#-------------------------------------------------------------------------------

def xaxis ( *arange ):
    """Correct usage is: ``xaxis(low|'normal'|'auto'|'tight'|'fit',
                               [high[,tick_interval]])``
    """
    _an_axis( _validate_active()._x_axis, arange )

#-------------------------------------------------------------------------------
#  Set the currently active plot window's y-axis range:
#-------------------------------------------------------------------------------

def yaxis ( *arange ):
    """Correct usage is: ``yaxis(low|'normal'|'auto'|'tight'|'fit',
                               [high[,tick_interval]])``
    """
    _an_axis( _validate_active()._y_axis, arange )

#-------------------------------------------------------------------------------
#  Set the currently active plot window's title:
#-------------------------------------------------------------------------------

def title ( text ):
    """Correct usage is: ``title(text)``"""
    _validate_active()._title.text = text
    redraw()

#-------------------------------------------------------------------------------
#  Set the currently active plot window's x-axis label:
#-------------------------------------------------------------------------------

def xtitle ( text ):
    """Correct usage is: ``xtitle(text)``"""
    _validate_active()
    if not (hasattr( _active.canvas, '_x_title' ) and
            _active.canvas._x_title is not None):
        _active.canvas._x_title = PlotTitle( position = 'bottom' )
        _active.canvas.add( _active.canvas._x_title )
    _active.canvas._x_title.text = text
    _active.update_window()

#-------------------------------------------------------------------------------
#  Set the currently active plot window's y-axis label:
#-------------------------------------------------------------------------------

def ytitle ( text ):
    """Correct usage is: ``ytitle(text)``"""
    _validate_active()
    if not (hasattr( _active.canvas, '_y_title' ) and
            _active.canvas._y_title is not None):
        _active.canvas._y_title = PlotTitle( position = 'left', rotate = 90 )
        _active.canvas.add( _active.canvas._y_title )
    _active.canvas._y_title.text = text
    _active.update_window()

#-------------------------------------------------------------------------------
#  Set the grid visibility for the active plot window:
#-------------------------------------------------------------------------------

def grid ( state = None ):
    """Correct usage is: ``grid([True|False])``"""
    _validate_active()
    if state is None:
        state = not _active._x_axis.grid_visible
    _active._x_axis.grid_visible = state
    _active._y_axis.grid_visible = state
    redraw()

#-------------------------------------------------------------------------------
#  Set the plot hold state for the currently active plot window:
#-------------------------------------------------------------------------------

def hold ( state = None ):
    """Correct usage is: ``hold([True|False])``"""
    _validate_active()
    if state is None:
        state = not _active.hold_
    _active.hold = state

#-------------------------------------------------------------------------------
#  Apply the specified axis settings to the currently active plot window:
#-------------------------------------------------------------------------------

def axis ( *setting ):
    """Correct usage is: ``axis('fit'|'auto'|x_lo,x_hi,y_lo,y_hi)``"""
    _validate_active()
###    x_ticks = _active._x_axis.ticks
###    x_interval = x_ticks[1]- x_ticks[0]
###    y_ticks = _active._y_axis.ticks
###    x_interval = x_ticks[1]- y_ticks[0]
###    axes = array((x_ticks[0],x_ticks[-1],y_ticks[0],y_ticks[-1]),Float)
###    # had to use client below cause of __setattr__ troubles in plot_frame
###    if setting == 'normal':
###        _active.client.aspect_ratio = setting
###        auto_scale()
###    elif setting == 'equal':
###        _active.client.aspect_ratio = setting
###    elif setting == 'freeze':
###        _active._x_axis.bounds = [axes[0],axes[1]]
###        _active._y_axis.bounds = [axes[2],axes[3]]
###        _active._x_axis.tick_interval = x_interval
###        _active._x_axis.tick_interval = y_interval
    if len( setting ) == 0:
        return ( ( _active._x_axis.bound_low, _active._x_axis.bound_high ),
                 ( _active._y_axis.bound_low, _active._y_axis.bound_high ) )
    if len( setting ) == 1:
        mode = setting[0]
        if mode in [ 'fit', 'auto' ]:
            _active._x_axis.set( bound_low     = mode,
                                 bound_high    = mode,
                                 tick_interval = 'auto' )
            _active._y_axis.set( bound_low     = mode,
                                 bound_high    = mode,
                                 tick_interval = 'auto' )
        else:
            raise ValueError
    elif len( setting ) == 4:
         _active._x_axis.set( bound_low  = setting[0],
                              bound_high = setting[1] )
         _active._y_axis.set( bound_low  = setting[2],
                              bound_high = setting[3] )
    else:
        raise ValueError, \
              "Correct usage is: axis('fit'|'auto'|x_lo,x_hi,y_lo,y_hi)"
    redraw()

def colormap ( palette, min_value=None, max_value=None ):
    """ Set the colormap on the active figure. """

    global _colormap

    # If the palette is a predefined generator function, create the palette.
    palette_type = type(palette)
    if palette_type is FunctionType:
        if not min_value:
            min_value=0.0
        if not max_value:
            max_value=1.0
        _colormap = palette(min_value, max_value)

    elif palette_type is ArrayType:
        min_value = 0
        max_value = palette.shape[0]-1
        from enthought.chaco.array_colormap import ArrayColormap
        _colormap = ArrayColormap(
            palette
        )

    return

#-------------------------------------------------------------------------------
#  Save the currently active window's plotting area to an image file:
#-------------------------------------------------------------------------------

def save ( file_name, format = None ):
    _validate_active().save( file_name, format )

#-------------------------------------------------------------------------------
#  Plotting machinery:
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Array utilities:
#-------------------------------------------------------------------------------

def is1D ( a ):
    as = shape( a )
    return ((len( as ) == 1) or (as[0] == 1) or (as[1] == 1))

def is2D ( a ):
    return (len( shape( a ) ) == 2)

def row ( a ):
    return reshape( asarray( a ), [1,-1] )

def col ( a ):
    return reshape( asarray( a ), [-1,1] )

def as_vector ( a ):
    a = remove_bad_vals( asarray( a ) )
    if is1D( a ):
        return reshape( a, ( size( a ), ) )
    raise ValueError

def as_matrix ( a ):
    a = remove_bad_vals( asarray( a ) )
    if not is2D( a ):
        if not is1D( a ):
            raise ValueError
        a = col( a )
    return a

#-------------------------------------------------------------------------------
#  Plot group parsing functions:
#-------------------------------------------------------------------------------

def plot_groups ( data ):
    remains = data
    groups  = []
    while len( remains ) > 0:
        group, remains = get_plot_group( remains )
        groups.append( group )
    return groups

def get_plot_group ( data ):
    group    = ()
    remains  = data
    state    = 0
    finished = False
    while (len(remains) > 0) and (not finished):
        el = remains[0]
        if state == 0:
            el    = asarray( el )
            state = 1
        elif state == 1:
            if isinstance(el, basestring):
                finished = True
            else:
                el = asarray( el )
            state = 2
        elif state == 2:
            finished = True
            if not isinstance(el, basestring):
                break
        try:
            if typecode(el) == 'D':
                print 'Warning: plotting magnitude of complex values'
                el = abs( el )
        except:
            pass
        group   = group + ( el, )
        remains = remains[1:]
    return ( group, remains )

def hstack ( tup ):
    # Horizontal stack (column wise):
    return concatenate( tup, 1 )

#-------------------------------------------------------------------------------
#  Create PlotValues from a specified data group:
#-------------------------------------------------------------------------------

def values_from_group ( group ):
    values   = []
    plotinfo = ''
    x        = group[0]
    ar_num   = 1
    if len( group ) > 1:
        if isinstance(group[1], basestring):
            plotinfo = group[1]
        else:
            ar_num = 2
            y      = group[1]
    if len( group ) == 3:
         plotinfo = group[2]

    # Force 1D arrays to 2D columns:
    if is1D( x ):
         x = col( x )
    if (ar_num == 2) and is1D( y ):
         y = col( y )

    xs = shape( x )
    if ar_num == 2:
        ys = shape(y)

    # Test that x and y have compatible shapes:
    if ar_num == 2:
        # Check that each array has the same number of rows:
        if xs[0] != ys[0]:
            raise SizeMismatch, ( 'rows', xs, ys )

        # Check that x.cols = y.cols:
        # (No error if x has 1 column)
        if (xs[1] > 1) and (xs[1] != ys[1]):
             raise SizeMismatch, ( 'cols', xs, ys )

    # Plot x against index:
    if ar_num == 1:
        for y_data in transpose( x ):
            pts = hstack( ( col( arange( len( y_data ) ) ),
                            col( y_data ) ) )
            values.append( PlotValue( remove_bad_vals( pts ) ) )

    # Plot x vs y:
    elif ar_num == 2:

         # x is effectively 1D:
         if xs[1] == 1:
             for y_data in transpose( y ):
                 pts = hstack( ( col( x ), col( y_data ) ) )
                 values.append( PlotValue( remove_bad_vals( pts ) ) )
         else:
             # x is 2D:
             x = transpose( x )
             y = transpose( y )
             for i in range( len( x ) ):
                 pts = hstack( ( col( x[i] ), col( y[i] ) ) )
                 values.append( PlotValue( remove_bad_vals( pts ) ) )

    color, line_style, marker, marker_color = process_format( plotinfo )
    _color, _line_style = color, line_style

    _validate_active()
    if not _active.hold_:
        _active.reset_line_style()

    for value in values:
        if (color == 'auto') and (line_style == 'auto'):
            _color, _line_style = _active.next_line_style()
        elif color == 'auto':
            _color, ignore = _active.next_line_style()
        elif line_style == 'auto':
            ignore, _line_style = _active.next_line_style()

        value.line_color = _color
        renderers = []
        if marker is not None:
            value.symbol     = marker
            value.fill_color = value.outline_color = marker_color or _color
            renderers.append( 'scatter' )
        if line_style is not None:
            value.line_style = _line_style
            renderers.append( 'line' )
        value.type = renderers

    return values

#-------------------------------------------------------------------------------
#  Convert a format string into a ( color, line, marker, marker_color ) tuple:
#-------------------------------------------------------------------------------

def process_format ( format ):

    if format == '':
        return ( 'auto', 'auto', None, None )
    color, line, marker, marker_color = 'auto', None, None, None
    m = color_re.findall( format )
    if len( m ) > 0:
        color = marker_color = color_trans[ m[0] ]
        if len( m ) > 1:
            marker_color = color_trans[ m[1] ]
    m = marker_re.findall( format )
    # The -1 takes care of 'r.', etc:
    if len( m ) > 0:
        marker = marker_trans[ m[0][-1] ]
    m = line_re.findall( format )
    if len( m ):
        line = line_trans[ m[0] ]
    return ( color, line, marker, marker_color )

#-------------------------------------------------------------------------------
#  Clip out illegal values from an array:
#-------------------------------------------------------------------------------

def remove_bad_vals ( x ):
    # !! Fix axis order when interface changed.
    # mapping:
    #    NaN -> 0
    #    Inf -> limits.double_max
    #   -Inf -> limits.double_min
    y     = nan_to_num( x )
    big   = limits.double_max / 10
    small = limits.double_min / 10
    y     = clip( y, small, big )
    return y

#-------------------------------------------------------------------------------
#  Plot 'stem' data:
#-------------------------------------------------------------------------------

def stem ( *data ):
    if len( data ) == 1:
        n     = arange( len( data[0] ) )
        x     = data[0]
        ltype = [ 'b-', 'mo' ]
    elif len( data ) == 2:
        if isinstance(data[1], basestring):
            ltype = [ data[1],'mo']
            n     = arange(len(data[0]))
            x     = data[0]
        else:
            n     = data[0]
            x     = data[1]
            ltype = [ 'b-', 'mo' ]
    elif len( data ) > 2:
        n     = data[0]
        x     = data[1]
        ltype = data[2]
        if isinstance(ltype, basestring):
            ltype = [ ltype, 'mo' ]
    else:
        raise ValueError, "Invalid input arguments."

    if len( n ) != len( x ):
         raise SizeMismatch, ( 'lengths', len( n ), len( x ) )

    # Line at zero:
    newdata = []
    newdata.extend( [ [ n[0], n[-1] ], [0,0], ltype[0] ] )

    # Stems:
    for k in range( len( x ) ):
        newdata.extend( [ [ n[k], n[k] ], [ 0, x[k] ], ltype[0] ] )

    # Circles:
    newdata.extend( [ n, x, ltype[1] ] )
    keywds = { 'fill_style': 'transparent' }
    plot( *newdata, **keywds )
    ###return plot( *newdata, **keywds )

#-------------------------------------------------------------------------------
#  Plot some data:
#-------------------------------------------------------------------------------

def plot ( *data, **kwtraits):
    values = []
    for group in plot_groups( data ):
        values.extend( values_from_group( group ) )

    # Check for hold:
    if kwtraits.has_key( 'hold' ):
        hold( kwtraits[ 'hold' ] )
        del kwtraits[ 'hold' ]

    # Assign all remaining traits to the values:
    for value in values:
        value.set( **kwtraits )

    # Add all of the values to the plot window:
    _validate_active().add_values( *values )

    ###return _active

#-------------------------------------------------------------------------------
#  Scatter plot some data:
#-------------------------------------------------------------------------------

def scatter ( X, Y, C='b', marker='o', **traits ):
    """Correct usage is: ``scatter(X,Y[,C][,markertype]``"""

    global _colormap

    # fixme: Currently, 'filled' is ignored (i.e. always filled)
    filled, color, line = ( False, 'auto', 'auto' )

    try:
        x, y = as_vector( X ), as_vector( Y )
        if shape( x ) != shape( y ): raise ValueError

        if isinstance(C, basestring):
            color = process_format(C)[0]
            is_mapped = False

        else:
            z = as_vector(C)
            if shape(x) != shape(z):
                raise ValueError('C must be same shape as X')
            plottype = 'colormappedscatter'
            cmap = _colormap
            is_mapped = True

    except Exception, ve:
        raise ValueError, \
              "Correct usage is: scatter(X,Y[,C],[,markertype][,'filled'])"


    ignore_color, ignore_line, marker, ignore_marker_color = process_format( marker )

    _validate_active()
    if not _active.hold_:
        _active.reset_line_style()
    marker = marker or 'circle'

    if not is_mapped:
        if color == 'auto':
            color, line_style = _active.next_line_style()
        value  = PlotValue( y, index         = PlotValue( x ),
                               type          = 'scatter',
                               symbol        = marker,
                               fill_color    = color,
                               outline_color = color,
                               **traits)
    else:
        value  = ColorMappedPlotValue(transpose(array([x,y,z])),
                                      colormap = cmap,
                                      type     = 'colormappedscatter',
                                      symbol   = marker,
                                      **traits)

    value.set( **traits )

    # Add the value to the plot window:
    _validate_active().add_values( value )

#-------------------------------------------------------------------------------
#  Color bar for the active color map.
#-------------------------------------------------------------------------------


def colorbar(title='', weight=0.1):
    """ Add a colorbar for the active figure to the current figure.

    title
        a title for the colorbar
    weight
        a relative weight to size the colorbar
    """

    global _colormap

    from enthought.chaco.colormap_legend import ColormapLegend
    value = ColormapLegend(_colormap)
    value.weight = weight
    value.bg_color =  (.75, .75, .75, 1.0)
    value.plot_min_width = 40
    if title is not None and len(title) > 0:
        value.add(PlotTitle(text=title))
    _validate_active().add_colorbar( value )

    return


def bwd():
    return _validate_active()


#-------------------------------------------------------------------------------
#  Horizontal/Vertical bar charts:
#-------------------------------------------------------------------------------

def bar ( *data, **traits ):
    """Correct usage is: ``bar([x,]Y[,width][,'grouped'|'stacked'][,bar_color])``"""
    kind       = 'bar'
    bin_width  = 100.0
    x          = None
    color      = 'auto'
    index_axis = traits.get( 'index_axis', 'x' )
    try:
        y = remove_bad_vals( asarray( data[0] ) )
        if len( data ) > 1:
            extra = 1
            if type( data[1] ) not in simple_types:
                extra = 2
                x, y  = y, remove_bad_vals( asarray( data[1] ) )
            for arg in data[extra:]:
                if isinstance(arg, basestring):
                    if   arg == 'grouped': kind = 'bar'
                    elif arg == 'stacked': kind = 'stackedbar'
                    else:
                        color = process_format( arg )[0]
                else:
                    bin_width = 100.0 * arg
        if (x is not None) and (not is1D( x )): raise ValueError
        y = as_matrix( y )
    except:
        raise ValueError, ("Correct usage is: bar([x,]Y[,width]"
                           "[,'grouped'|'stacked'][,bar_color])")

    _validate_active()
    if not _active.hold_:
        _active.reset_line_style()

    if x is None: x = arange( 1.0, shape( y )[0] + 0.1 )
    elif x[0] > x[-1]:
        x = x[::-1]
        y = y[::-1,:]
    index  = PlotValue( x )
    values = []
    for i in range( shape( y )[1] ):
        fill_color = color
        if fill_color == 'auto':
            fill_color = _active.next_line_style()[0]
        value = PlotValue( y[:,i], index      = index,
                                   type       = kind,
                                   fill_color = fill_color,
                                   bar_width  = bin_width,
                                   bin_width  = bin_width,
                                   index_axis = index_axis )
        value.set( **traits )
        values.append( value )

    # Add the value to the plot window:
    _validate_active().add_values( *values )
    value.container.index_axis = index_axis

def barh ( *data, **traits ):
    """Correct usage is: ``barh([x,]Y[,width][,'grouped'|'stacked'][,bar_color])``
    """
    try:
        bar( index_axis = 'y', *data, **traits )
    except:
        raise ValueError, ("Correct usage is: barh([x,]Y[,width]"
                           "[,'grouped'|'stacked'][,bar_color])")

#-------------------------------------------------------------------------------
#  Histogram chart:
#-------------------------------------------------------------------------------

def hist ( *data, **traits ):
    """Correct usage is: ``hist(Y[,x|n][,color])``"""
    from scipy.stats import histogram

    x     = outline_color = None
    bins  = 10
    color = 'auto'
    try:
        y = remove_bad_vals( asarray( data[0] ) )
        if len( data ) > 1:
            extra = 1
            if type( data[1] ) not in simple_types:
                extra = 2
                x = remove_bad_vals( asarray( data[1] ) )
            for arg in data[extra:]:
                if isinstance(arg, basestring):
                    color, ignore, ignore, outline_color = process_format( arg )
                elif x is None:
                    bins = int( arg )
                else:
                    raise ValueError
        if (x is not None) and (not is1D( x )): raise ValueError
        y = as_matrix( y )
    except:
        raise ValueError, "Correct usage is: hist(Y[,x|n][,color])"

    if x is not None:
        if x[0] > x[-1]:
            x = x[::-1]
            y = y[::-1,:]
        low, high, n = min( x ), max( x ) * 1.000001, len( x )
        dx           = (high - low) / (2 * (n - 1))
        low, high    = low - dx, high + dx
    else:
        low, high, n = min( y ), max( y ) * 1.000001, bins
        dx           = (high - low) / n
        x            = arange( low + dx / 2.0, low + bins * dx, dx )

    _validate_active()
    if not _active.hold_:
        _active.reset_line_style()

    index  = PlotValue( x )
    values = []
    if (outline_color is None) or (outline_color == color):
        outline_color = 'black'
    bounds        = ( low, high )
    for i in range( shape( y )[1] ):
        fill_color = color
        if fill_color == 'auto':
            fill_color = _active.next_line_style()[0]
        data  = histogram( y[:,i], n, bounds )[0]
        value = PlotValue( data, index         = index,
                                 type          = 'stackedbar',
                                 fill_color    = fill_color,
                                 outline_color = outline_color,
                                 bar_width     = 100.0 )
        value.set( **traits )
        values.append( value )

    # Add the value to the plot window:
    _validate_active().add_values( *values )

#-------------------------------------------------------------------------------
#  Error Bar charts:
#-------------------------------------------------------------------------------

def errorbar ( *data, **traits ):
    """Correct usage is: ``errorbar([X,]Y,E[,U][,color])``"""
    x     = u = bar_color = None
    color = 'auto'
    try:
        y = as_vector( data[0] )
        e = as_vector( data[1] )
        extra = 2
        if (len( data ) > 2) and (type( data[2] ) not in simple_types):
            extra += 1
            x, y, e = y, e, as_vector( data[2] )
            if (len( data ) > 3) and (type( data[3] ) not in simple_types):
                extra += 1
                u = as_vector( data[3] )
        for arg in data[extra:]:
            color, ignore, ignore, bar_color = process_format( arg )
        if ((len( y ) != len( e )) or
            ((x is not None) and (len( x ) != len( y ))) or
            ((u is not None) and (len( u ) != len( y )))):
            raise ValueError
    except:
        raise ValueError, "Correct usage is: errorbar([X,]Y,E[,U][,color])"

    _validate_active()
    if not _active.hold_:
        _active.reset_line_style()

    if color == 'auto': color = _active.next_line_style()[0]
    if bar_color is None: bar_color = color
    if u is None: u = e
    if x is None: x = arange( 0.0, len( y ) )

    index  = PlotValue( x )
    values = [ PlotValue( y, index         = index,
                             type          = 'errorbar',
                             line_color    = color,
                             outline_color = bar_color,
                             bar_width     = 50.0 ),
               PlotValue( e, index         = index,
                             type          = 'errorbar' ),
               PlotValue( u, index         = index,
                             type          = 'errorbar' ) ]

    # Add the value to the plot window:
    _validate_active().add_values( *values )

#-------------------------------------------------------------------------------
#  Draw a horizontal/vertical line on the plot:
#-------------------------------------------------------------------------------

def hline ( *data, **traits ):
    """Correct usage is: ``hline(y[,'=label'][,color])``"""
    horizontal = True
    if traits.get( 'vertical' ):
        horizontal = False
        del traits[ 'vertical' ]
    text        = ''
    color       = 'red'
    line_style  = bg_color = None
    border_size = 0
    try:
        y = float( data[0] )
        for arg in data[1:]:
            if type( arg ) in ( int, float ):
                text        = str( arg )
                bg_color    = bg_color or 'cyan'
                border_size = 1
            elif arg[:1] == '=':
                text        = arg[1:]
                bg_color    = bg_color or 'cyan'
                border_size = 1
            else:
                color, line_style, ignore, bg_color = process_format( arg )
    except:
        raise ValueError, "Correct usage is: hline(y[,'=label'][,color])"

    if color == 'auto': color = 'red'
    value = PlotOverlay( text        = text,
                         line_color  = color,
                         line_style  = line_style or 'solid',
                         bg_color    = bg_color,
                         border_size = border_size )
    pv = _validate_active().canvas.values[0]
    if horizontal:
        value.value = y
        x = min( pv.index.data )
    else:
        value.index = x = y
        y = min( pv.data )
    _add_overlay( x, y, value, **traits )

def vline ( *data, **traits ):
    """Correct usage is: ``vline(y[,'=label'][,color])``"""
    try:
        hline( vertical = True, *data, **traits )
    except:
        raise ValueError, "Correct usage is: vline(y[,'=label'][,color])"

def text ( *data, **traits ):
    """Correct usage is: ``text(x,y,text[,color])``"""
    bg_color   = 'auto'
    line_color = 'black'
    line_style = None
    try:
        x, y = float( data[0] ), float( data[1] )
        text = str( data[2] )
        for arg in data[3:]:
            bg_color, line_style, ignore, line_color = process_format( arg )
    except:
        raise ValueError, "Correct usage is: text(x,y,text[,color])"

    if bg_color == 'auto': bg_color = 'cyan'
    value = PlotOverlay( text        = text,
                         line_color  = line_color,
                         line_style  = line_style or 'solid',
                         bg_color    = bg_color,
                         index       = x,
                         value       = y )
    _add_overlay( x, y, value, **traits )

def _add_overlay ( x, y, overlay, **traits ):
    overlay.set( **traits )
    fake = PlotValue( [ y, y ], index = PlotValue( [ x, x ] ),
                                type  = 'scatter',
                                size  = 'tiny' )
    fake.add( overlay )
    _validate_active().canvas.add( fake )
    _validate_active().update_window()

#-------------------------------------------------------------------------------
#  Divide the current figure up into multiple subplots:
#-------------------------------------------------------------------------------

def subplot ( *data, **traits ):
    """Correct usage is: ``subplot(rows,columns,index)`` or
                         ``subplot(rci)``"""
    try:
        value = int( data[0] )
        if len( data ) == 1:
            rows  = value / 100
            cols  = (value / 10) % 10
            index = (value - 1) % 10
        else:
            rows  = value
            cols  = int( data[1] )
            index = int( data[2] ) - 1
        if ((not (1 <= rows <= 5)) or
            (not (1 <= cols <= 8)) or
            (not (0 <= index < (rows * cols)))):
            raise ValueError
    except:
        raise ValueError, "Correct usage is: subplot(rows,columns,index)"

    _validate_active().subplot( rows, cols, index )
    _validate_active().update_window()

#-------------------------------------------------------------------------------
#  ???
#-------------------------------------------------------------------------------

def markers ( visible = None ):
    pass

#-------------------------------------------------------------------#
#--------------------------- image ---------------------------------#
#-------------------------------------------------------------------#

def imread(filename):
    """Read an image file into an array.  Accepts a file name and will
    automatically handle most formats (as provided for by PIL)
    """
    return Image(filename).bmp_array

def image(data,x=None,y=None,colormap=None,scale=False,_contourhelper=False,
          origin='bottom left', discrete_colormap=None, alpha=None,  **kwtraits):
    """Plot an image.  Data is either a 2 dimensional array, in which
    case we treat it as a colormap plot.  You may specifiy a colormap
    object, otherwise a standard grayscale colormap is used.  If the data
    is a 3 dimensional array, it should be a width x height x 3 array, with
    the final dimension being Red, Green, and Blue values.  These can be
    floats in the range [0,1], or integers in the range [0,255].  If
    scale is True on a colormapped plot, image will auto-scale your data
    to linearly fit in the [0,1] range.  You may also specify scale=(min,max),
    in which case, that range will be taken as the proper range of pixels
    in the colormap.  You may optionally specify x and y, which control
    the x and y range into which the image will be plotted.  By default
    the image will be plotted into the coordinates of the matrix, i.e.
    1 pixel in the matrix = 1 unit on the scale.
    """
    # Ideally colormap would default to the current colormap
    # Check for hold:
    width, height = data.shape[0:2]
    if x:
        xbounds = (x[0], x[-1])
    else:
        xbounds = (-.5,width-.5)
    if y:
        ybounds = (y[0], y[-1])
    else:
        ybounds = (-.5,height-.5)

    _validate_active()

    if kwtraits.has_key( 'hold' ):
        hold( kwtraits[ 'hold' ] )
        del kwtraits[ 'hold' ]
    if _active.hold_:
        xaxis = _active._x_axis
        yaxis = _active._y_axis
    else:
        xaxis = PlotAxis(tick_visible=True, grid_visible=False, \
                     bound_low = xbounds[0], bound_high = xbounds[1])
        yaxis = PlotAxis(tick_visible=True, grid_visible=False, \
                     bound_low = ybounds[0], bound_high = ybounds[1])
        _active._x_axis = xaxis
        _active._y_axis = yaxis

    index_arrays = []
    for bounds, pixsize in ((xbounds, width),(ybounds, height)):
        if pixsize > 1:
            step = (bounds[1] - bounds[0]) / float(pixsize-1)
            index_arrays.append(arange(bounds[0], bounds[1]+step/2.0, step))
        else:
            index_arrays.append(array(bounds[0]))

    if len(data.shape) == 3: #This is a true color plot, not colormap
        # We need the [::-1] to flip the data here.  Unfortunately, CmapImagePlotValue
        # and ImagePlotValue have different ideas of what the y direction of the data
        # is supposed to be.
        if typecode(data) in ('d', 'f'): # If floats, multiply and convert
            converted_image_data = (data*255)[::-1].astype(UnsignedInt8)
        else:
            converted_image_data = data[::-1].astype(UnsignedInt8)
        image_data = ImageData(converted_image_data)
        plot_value = ImagePlotValue(data=image_data, axis=xaxis, axis_index=yaxis,
                                    origin=origin,
                                    x_bounds=xbounds, y_bounds=ybounds, **kwtraits)
        image_data.index = (index_arrays[0], index_arrays[1])
    else: # This is a colormapped plot
        # It would be nice if we accepted colormaps in matlab format
        # and converted them to our format.
        global _colormap
        colormap_obj = colormap or _colormap or gray(0.0,1.0)
        if alpha:
            colormap_obj = AlphaWrapColormap(colormap_obj, alpha, mode='multiply')
        if scale:
            if scale == True: # Auto-scale
                maxdata = float(maximum.reduce(ravel(data)))
                mindata = float(minimum.reduce(ravel(data)))
                colormapmin = colormap_obj.min_value
                colormapmax = colormap_obj.max_value
                data = (data-mindata)/(maxdata-mindata)* \
                             (colormapmax-colormapmin)+colormapmin

            else: # They specified an explicit scale
                colormapmin = colormap_obj.min_value
                colormapmax = colormap_obj.max_value
                data = (data-scale[0])/(scale[1]-scale[0])* \
                       (colormapmax-colormapmin) + colormapmin

        image_data = ImageData(data, transposed=True)
        plot_value = CmapImagePlotValue(data=image_data, colormap=colormap_obj,
                                         axis_index=yaxis, axis=xaxis,
                                         origin=origin, type='image', **kwtraits)
        image_data.index = (index_arrays[1], index_arrays[0])


    if _contourhelper:
        return (plot_value, image_data)
    else:
        _active.add_values(plot_value)
        return _active

def imagesc(img,x=None,y=None,colormap=None, scale=True, _contourhelper=False,
            origin='bottom left', alpha=None, **kwtraits):
    """imagesc is just like image with auto-scaling on by default."""
    image(img,x,y,colormap,scale=scale, _contourhelper=_contourhelper, origin=origin,
          alpha=alpha, **kwtraits)

def contour(data, x=None, y=None, numlevels=10, filled=False, line_color='blue', colormap=None,
            line_weight=1, alpha=None, origin='bottom left'):
    """Contour plot -- takes a matrix of heights and number of contour levels,
    and displays a contour plot of the data, optionally backed with an scaled colormap
    of the data behind it."""
    global _colormap
    colormap_obj = colormap or _colormap or gray(0.0,numlevels)
    if alpha:
        colormap_obj = AlphaWrapColormap(colormap_obj, alpha, 'multiply')

    # Generate an ImagePlotValue and ImageData for the matrix.
    (ipv, pd) = image(data, x=x, y=y, scale=True, colormap=colormap_obj, _contourhelper=True)
    #if filled:
        #Disable the axes for the image-- we'll be using the axes generated by the contour
    #    ipv.axis.visible = False
    #    ipv.axis_index.visible = False
    #    canvas = PlotCanvas(ipv, plot_origin='top left')
    #else:
        #If we're not filling, just generate an empty canvas instead of one with an
        #ImagePlotValue in it.
    canvas = PlotCanvas(plot_origin=origin)

    cont = ContourData(pd, color=line_color)
    cont.auto_level(numlevels)
    if filled:
        cont.colormap=colormap_obj
        cont.add_to(canvas, source_name='regions' )
        if not (alpha and alpha<0.5):
            cont.add_to(canvas)
#        cont.add_to(_active.canvas, source_name='regions' )
#        cont.add_to(_active.canvas)

    else:
        cont.colormap=colormap_obj
        cont.add_to(canvas, line_weight=line_weight)
    _active.add_values(canvas)

    # We don't want the plot auto-sizing to where the contours are, so
    # We explicitly set the bounds to be the bounds of the matrix.
    if _active.hold_:
        _active.canvas.axis = _active._x_axis
        _active.canvas.axis_index = _active._y_axis
    else:
        if y:
            _active.canvas.axis.bound_low=y[0]
            _active.canvas.axis.bound_high=y[-1]
        else:
            _active.canvas.axis.bound_low=0
            _active.canvas.axis.bound_high=data.shape[0]-1
        if x:
            _active.canvas.axis_index.bound_low=x[0]
            _active.canvas.axis_index.bound_high=x[-1]
        else:
            _active.canvas.axis_index.bound_low=0
            _active.canvas.axis_index.bound_high=data.shape[1]-1

    _active.update_window()

def contourf(data, x=None, y=None, numlevels=10, line_color='black', colormap=None, alpha=None):
    contour(data, x=x, y=y, numlevels=numlevels, filled=True, line_color=line_color,
             colormap=colormap, alpha=alpha)

#-------------------------------------------------------------------------------
#  Matlab equivalence:
#-------------------------------------------------------------------------------

xlabel = xtitle
ylabel = ytitle

#-------------------------------------------------------------------------------
#  'ShellPlotFrame' class:
#-------------------------------------------------------------------------------

class ShellPlotFrame ( HasTraits, PlotUndoHandler ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    hold = Trait( flexible_false_trait,
                  desc = 'whether previous plots should be kept' )

    # PDF support is available flag:
    _has_pdf = True


    # Bogus trait definitions to make this compatible with HasPrivateTraits
    dic = Any
    canvas = Any
    container = Any
    scheme_index = Any

    #---------------------------------------------------------------------------
    #  Initialize object:
    #---------------------------------------------------------------------------

    def __init__ ( self, title ):

        # Get the saved state dictionary (if any):
        self.dic = self.get_state()

        # Set up the current sub-plot information:
        self._rows  = self._cols = 0
        self._index = -1
        self.canvas = None

        ## Create the menu bar (if any):
        #self.make_menu( plot_frame_menu )

        # Create the plotting canvas:
        self._value = self._overlay = self._title = self._axis = 0

        # Create the canvas container:
        self.container = self.make_plot_container(
            PlotGroup(bg_color = (.75, .75, .75, 1.0)),
            right_click  = 'all menu',
            user_handler = self,
            undo_handler = self,

            # Defaults:
            canvas_border_size    = 0,
            canvas_margin_width   = 0,
            canvas_margin_height  = 0,
            canvas_padding_width  = 4,
            canvas_padding_height = 4
        )

        # Set up the default subplot:
        self.subplot( 1, 1, 0 )

        # Add a title:
        self._title.text = title

        # Restore user preference items:
        self.restore_state()

        # Initialize the 'undo' history:
        self.reset_undo()

        # Make sure the plotting window (if any) is updated:
        self.update_window()

    #---------------------------------------------------------------------------
    #  Set up a specified sub-plot:
    #---------------------------------------------------------------------------

    def subplot ( self, rows, cols, index ):
        if (rows != self._rows) or (cols != self._cols):
            self.container.component.clear()

            self._rows  = rows
            self._cols  = cols
            self._index = -1
            groups      = []
            grid = PlotGrid( columns = cols, rows = rows, weight=1.0, bg_color=(.75,.75,.75,1.0 ))
            for r in range( rows ):
                for c in range( cols ):
                    # Create an x-axis and y-axis:
                    x_axis = PlotAxis( menu = axis_menu, label = 'x axis' )
                    y_axis = PlotAxis( menu = axis_menu, label = 'y axis' )

                    # Create a plotting canvas:
                    canvas = PlotCanvas( menu          = canvas_menu,
                                         label         = 'Canvas',
                                         bg_color      = (.75,.75,.75,1.0),
                                         padding       = 6,
                                         axis          = y_axis,
                                         axis_index    = x_axis,
                                         _scheme_index = -1,
                                         _hold         = 'no',
                                         interaction = PanZoomInteraction(color="blue"))
                    canvas.add( PlotValue( [ 0, 0 ],
                                           index = PlotValue( [ 0, 0 ] ),
                                           type  = 'scatter',
                                           size  = 'tiny' ),
                                PlotTitle( text = '', menu = default_menu ) )
                    grid.add( canvas )
            self.container.component.add(grid)
            self.container.component._grid = grid

        if index != self._index:
            if self.canvas is not None:
                self.canvas._scheme_index = self.scheme_index
                self._hold                = self.hold
            self._index       = index
            row               = index / self._cols
            col               = index % self._cols
            root              = self.container.component._grid
            self.canvas       = canvas = root.values[ index ]
            self._x_axis      = canvas.axis_index
            self._y_axis      = canvas.axis
            self._title       = canvas.titles[0]
            self.scheme_index = canvas._scheme_index
            self.hold         = canvas._hold


    #---------------------------------------------------------------------------
    #  Clear the canvas of any current plot information:
    #---------------------------------------------------------------------------

    def clear_canvas ( self ):
        canvas = self.canvas
        for value in canvas.values:      # fixme: Garbage disposal hack
            self.dispose_value( value )  # fixme: Garbage disposal hack
        del canvas.values[:]
        del canvas.overlays[:]
        del canvas.titles[1:]
        self._x_axis = PlotAxis( menu = axis_menu )
        self._y_axis = PlotAxis( menu = axis_menu )
        canvas.set( axis = self._y_axis, axis_index = self._x_axis )
        self.reset_undo()
        self.update_window()

    ##---------------------------------------------------------------------------
    ##  Reset the 'undo' history:
    ##---------------------------------------------------------------------------
    #
    #def reset_undo ( self ):
    #    self.undo_history = []
    #    self.undo_index   = 0
    #    self.check_undo()
    def reset_undo(self):
        pass

    #---------------------------------------------------------------------------
    #  Add a set of values to the canvas:
    #---------------------------------------------------------------------------

    def add_values ( self, *values ):
        canvas = self.canvas
        if not self.hold_:
            for value in canvas.values[:]:   # fixme: Garbage disposal hack
                self.dispose_value( value )  # fixme: Garbage disposal hack
            del canvas.values[:]


        for value in values:
            value.set( menu  = value_menu,
                       label = 'Value %d' % self._value )
            # The way we treat the X axis vs. the Y axis is strange,
            # so we have to special-case images.
            if hasattr(value.index, 'axis_index'):
                value.index.axis_index = value.axis_index = self._x_axis
            canvas.add( value )
            self._value += 1

        self.update_window()

    #---------------------------------------------------------------------------
    #  Add one object to another:
    #---------------------------------------------------------------------------

    def add_object ( self, object, container, update ):
        container.add( object )
        if update:
            self.add_undo( PlotAddCommand( object, container ) )

    #---------------------------------------------------------------------------
    #  Remove an object:
    #---------------------------------------------------------------------------

    def remove_object ( self, object ):
        self.add_undo( PlotRemoveCommand( object, object.container ) )
        object.remove()

    #---------------------------------------------------------------------------
    #  Add a new PlotOverlay to a specified object:
    #---------------------------------------------------------------------------

    def add_overlay ( self, object, update = False ):
        self._overlay += 1
        overlay        = PlotOverlay( menu  = default_menu,
                                      text  = 'Overlay %d' % self._overlay,
                                      label = 'Overlay %d' % self._overlay )
        self.add_object( overlay, object, update )
        if update:
            self.update_window()
        return overlay

    #---------------------------------------------------------------------------
    #  Add a new PlotTitle to a specified object:
    #---------------------------------------------------------------------------

    def add_title ( self, object, update = False ):
        self._title += 1
        title        = PlotTitle( menu  = default_menu,
                                  text  = 'Title %d' % self._title,
                                  label = 'Title %d' % self._title )
        self.add_object( title, object, update )
        if update:
            self.update_window()
        return title

    #---------------------------------------------------------------------------
    #  Add a new PlotAxis to a specified object:
    #---------------------------------------------------------------------------

    def add_axis ( self, object, update = False ):
        self._axis  += 1
        axis         = PlotAxis( menu  = axis_menu,
                                 title = 'Axis %d' % self._axis,
                                 label = 'Axis %d' % self._axis )
        self.add_undo(
            PlotTraitCommand( object, 'axis', object.axis, axis ) )
        object.axis = axis
        if update:
            self.update_window()
        return axis

    #---------------------------------------------------------------------------
    #  Remove the axis for a specified object:
    #---------------------------------------------------------------------------

    def remove_axis ( self, object ):
        self.add_undo( PlotRemoveAxisCommand( object ) )
        del object.axis
        self.update_window()

    #---------------------------------------------------------------------------
    #  Add a colorbar to this window.
    #---------------------------------------------------------------------------

    def add_colorbar( self, colorbar ):

        if hasattr(self, '_colorbar') and self._colorbar is not None:
            self.container.component.remove(self._colorbar)

        self._colorbar = colorbar
        self.container.component.add(colorbar)

        self.update_window()
        return

    #---------------------------------------------------------------------------
    #  fixme: The next three methods are a hack to force proper garbage
    #  collection:
    #---------------------------------------------------------------------------

    def dispose_value ( self, value, axis = False ):
        if axis:
            value.axis.__dict__.clear()
            value.axis_index.__dict__.clear()
        value.values.__dict__.clear()
        value.titles.__dict__.clear()
        value.overlays.__dict__.clear()
        value.__dict__.clear()

    def dispose_canvas ( self, canvas ):
        for value in canvas.values:
            self.dispose_value( value, True )
        canvas.axis.__dict__.clear()
        if not isinstance(canvas.axis_index, NoneType):
            canvas.axis_index.__dict__.clear()
        canvas.values.__dict__.clear()
        canvas.titles.__dict__.clear()
        canvas.overlays.__dict__.clear()
        canvas.__dict__.clear()

    def dispose_group ( self, group ):
        for value in group.values:
            if isinstance( value, PlotGroup ):
                self.dispose_group( value )
            else:
                self.dispose_canvas( value )
        group.values.__dict__.clear()
        group.titles.__dict__.clear()
        group.overlays.__dict__.clear()
        group.__dict__.clear()

    ##---------------------------------------------------------------------------
    ##  Check the settings of the undo/redo menu options:
    ##---------------------------------------------------------------------------
    #
    #def check_undo ( self ):
    #    if self.has_ui():
    #        enabled = (self.undo_index > 0)
    #        self.menu_undo.enabled( enabled )
    #        if enabled:
    #            self.menu_undo.label(
    #                 'Undo %s\tCtrl-Z' %
    #                 self.undo_history[ self.undo_index - 1 ].desc )
    #        else:
    #            self.menu_undo.label( "Can't undo" )
    #        enabled = (self.undo_index < len( self.undo_history ))
    #        self.menu_redo.enabled( enabled )
    #        if enabled:
    #            self.menu_redo.label( 'Redo %s\tCtrl-Y' %
    #                                 self.undo_history[ self.undo_index ].desc )
    #        else:
    #            self.menu_redo.label( "Can't redo" )
    #
    ##---------------------------------------------------------------------------
    ##  Add a new command to the 'undo' history:
    ##---------------------------------------------------------------------------
    #
    #def add_undo ( self, command ):
    #    self.undo_history[ self.undo_index: ] = [ command ]
    #    self.undo_index = len( self.undo_history )
    #    self.check_undo()
    #
    ##---------------------------------------------------------------------------
    ##  Undo the last command:
    ##---------------------------------------------------------------------------
    #
    #def undo ( self ):
    #    self.undo_index -= 1
    #    self.undo_history[ self.undo_index ].undo()
    #    self.check_undo()
    #
    ##---------------------------------------------------------------------------
    ##  Redo the last undone command:
    ##---------------------------------------------------------------------------
    #
    #def redo ( self ):
    #    self.undo_history[ self.undo_index ].redo()
    #    self.undo_index += 1
    #    self.check_undo()

    #---------------------------------------------------------------------------
    #  Handle an update request:
    #---------------------------------------------------------------------------

    def update_window ( self ):
        self.container.component.update()

    #---------------------------------------------------------------------------
    #  Set the default axes to auto-scale their data:
    #---------------------------------------------------------------------------

    def auto_scale ( self ):
        self._x_axis.set( tick_interval = 'auto',
                          bound_low     = 'fit',
                          bound_high    = 'fit' )
        self._y_axis.set( tick_interval = 'auto',
                          bound_low     = 'auto',
                          bound_high    = 'auto' )
        self.update_window()

    #---------------------------------------------------------------------------
    #  Reset the 'next style' counter:
    #---------------------------------------------------------------------------

    def reset_line_style ( self ):
        self.scheme_index = -1

    #---------------------------------------------------------------------------
    #  Return the next available color and line style scheme:
    #---------------------------------------------------------------------------

    def next_line_style ( self ):
        self.scheme_index += 1
        return color_schemes[ self.scheme_index % len( color_schemes ) ]

    #---------------------------------------------------------------------------
    #  Set the current image size:
    #---------------------------------------------------------------------------

    def set_image_size ( self, index ):
        self.image_index     = index
        self.bmdx, self.bmdy = image_sizes[ index ]
        #for i in range( len( image_sizes ) ):
        #    getattr( self, 'menu_image_%d' % i ).checked( i == index )

    #---------------------------------------------------------------------------
    #  Save the plot canvas to a file in a specified format:
    #---------------------------------------------------------------------------

    def save ( self, filename, ext ):
        if ext is None:
            ext = os.path.splitext( filename )[1]
        if ext[0:1] != '.':
            ext = '.' + format
        ext = ext.lower()
        if ext not in file_types:
            raise ValueError, (
                  'Unknown file format. The valid formats are: %s' %
                  ', '.join( file_types ) )
        self.save_file( filename, ext )

    #---------------------------------------------------------------------------
    #  Create an output file for the specified object:
    #---------------------------------------------------------------------------

    def create_file ( self, object, mode ):
        filename = self.get_save_filename()
        if filename is None:
            return

        ext = os.path.splitext( filename )[1].lower()
        if ext not in file_types:
            self.show_error( "Invalid file extension: '%s'.\n\n"
                             "Select an extension from the specified list." %
                             ext )
            return

        self.save_file( filename, ext, object, mode )

    #---------------------------------------------------------------------------
    #  Save the specified object in the specified output file:
    #---------------------------------------------------------------------------

    def save_file ( self, filename, ext, object = None, mode = 0 ):
        if object is None:
            object = self.canvas

        list = []
        self.reset_scaling( object, list )

        if ext == '.pdf':
            try:
                import enthought.kiva.backend_pdf
            except:
                self._has_pdf = False
                self.show_error(
                     "Unable to create PDF files.\n\n"
                     "Please make sure that you have ReportLab installed.\n\n"
                     "You can obtain ReportLab at:\n"
                     "    http://www.reportlab.org/downloads.html" )
                return
            canvas       = enthought.kiva.backend_pdf.Canvas( filename )
            cdx, cdy     = canvas.size()
            cdx         -= 16
            cdy         += 50
            self.container.gc = gc = canvas.gc
            self.draw_page( object, mode, 20, 20, cdx, cdy, self.pdf_end_page )
            canvas.save()

        else:
            pass
#            import enthought.kiva.wxcore2d
#            self.kind       = kind
#            self.filename   = filename
#            self.file_index = 0
#            self.bitmap     = wx.wxEmptyBitmap( self.bmdx, self.bmdy )
#            self.dc = dc    = wx.wxMemoryDC()
#            dc.SelectObject( self.bitmap )
#            dc.SetDeviceOrigin( 0, self.bmdy - 1 )
#            dc.SetAxisOrientation( 1, 1 )
#            self.erase_dc()
#            self.container.gc = enthought.kiva.wxcore2d.GraphicsContext( dc )
#            self.draw_page( object, mode, 0, 0, self.bmdx, self.bmdy,
#                            self.image_end_page )
#            del self.kind
#            del self.filename
#            del self.file_index
#            del self.dc
#            del self.bitmap

        self.restore_scaling( list )
        self.container.gc = None
        self.update_window()

    #---------------------------------------------------------------------------
    #  Draw a single PlotCanvas or PlotValue object:
    #---------------------------------------------------------------------------

    def draw_page ( self, object, mode, x, y, dx, dy, end_page_handler ):
        if ( (mode == 0) or
            ((mode == 1) and isinstance( object, PlotCanvas )) or
            ((mode == 2) and isinstance( object, PlotValue ))):
            gc = self.container.gc
            gc.save_state()
            gc.scale_ctm( 1.0, 1.0 )
            object.render( gc, x, y, dx, dy )
            gc.restore_state()
            end_page_handler()
            return

        for subobject in object.values:
            self.draw_page( subobject, mode, x, y, dx, dy, end_page_handler )

    #---------------------------------------------------------------------------
    #  Handle the end of a PDF page:
    #---------------------------------------------------------------------------

    def pdf_end_page ( self ):
        self.container.gc.end_page()

    #---------------------------------------------------------------------------
    #  Handle the end of an image 'page':
    #---------------------------------------------------------------------------

#    def image_end_page ( self ):
#        image    = wx.wxImageFromBitmap( self.bitmap )
#        filename = self.filename
#        self.file_index += 1
#        if self.file_index > 1:
#            name, ext = os.path.splitext( filename )
#            filename  = '%s_%d%s' % ( name, self.file_index, ext )
#        image.SaveFile( filename, self.kind )
#        self.erase_dc()

    #---------------------------------------------------------------------------
    #  Clear the current image background:
    #---------------------------------------------------------------------------

#    def erase_dc ( self ):
#        dc       = self.dc
#        color    = self.container.bg_color_
#        bg_color = wx.wxColour( int( color[0] * 255 ),
#                                int( color[1] * 255 ),
#                                int( color[2] * 255 ) )
#        dc.SetPen( wx.wxTRANSPARENT_PEN )
#        dc.SetBrush( wx.wxBrush( bg_color ) )
#        dc.DrawRectangle( 0, -1, self.bmdx, self.bmdy )

    #---------------------------------------------------------------------------
    #  Reset any index/value scaling temporarilly:
    #---------------------------------------------------------------------------

    def reset_scaling ( self, object, list ):
        for attr in ( 'canvas_value_scale',
                      'canvas_index_scale',
                      'value_scale',
                      'index_scale' ):
            value = getattr( object, attr )
            if value != 1.0:
                list.append( ( object, attr, value ) )
                setattr( object, attr, 1.0 )
        if not isinstance( object, PlotValue ):
            for subobject in object.values:
                self.reset_scaling( subobject, list )

    #---------------------------------------------------------------------------
    #  Restore previously saved object scaling values:
    #---------------------------------------------------------------------------

    def restore_scaling ( self, list ):
        for object, attr, value in list:
            setattr( object, attr, value )

    #---------------------------------------------------------------------------
    #  Handle the user exiting the plot window:
    #---------------------------------------------------------------------------

    def on_exit ( self ):
        self.on_close()

    #---------------------------------------------------------------------------
    #  Handle the user closing the plot window:
    #---------------------------------------------------------------------------

    def on_close ( self, event = None ):
        self.save_state()
        _figure_closed( self )
        self.destroy_window()

    #---------------------------------------------------------------------------
    #  Open the file used to save the user preference 'pickles' in:
    #---------------------------------------------------------------------------

    def open_state ( self, mode ):
        try:
            path = os.environ[ 'HOME' ]
        except:
            path = ''
        return open( os.path.join( path, self.state_file ), mode )

    #---------------------------------------------------------------------------
    #  Pickle all of the user preference items:
    #---------------------------------------------------------------------------

    def save_state ( self ):

        # Try to save the session state:
        fd = None
        try:
            fd  = self.open_state( 'w' )
            p   = cPickle.Pickler( fd )
            dic = self.dic

            # Save the window size and position:
            dic[ 'window' ]      = self.get_window_geometry()
            dic[ 'image_index' ] = self.image_index

            # Save the dictionary:
            p.dump( dic )
        except:
            pass

        # Close the file (if necessary):
        if fd:
            fd.close()

    #---------------------------------------------------------------------------
    #  Return the dictionary of user preference items:
    #---------------------------------------------------------------------------

    def get_state ( self ):
        # Only get the state if there is a user interface:
        if not self.has_ui():
            return None

        # Assume that the restore will fail:
        dic = {}
        fd  = None

        # Try to restore the previous session state:
        try:
            fd = self.open_state( 'r' )

            # Restore the dictionary of saved values:
            dic = cPickle.Unpickler( fd ).load()
        except:
            pass

        # Close the file (if necessary):
        if fd:
            fd.close()

        # Return the saved state dictionary:
        return dic

    #---------------------------------------------------------------------------
    #  Restore the user preference items:
    #---------------------------------------------------------------------------

    def restore_state ( self ):
        dic = self.dic
        if dic is not None:

            # Restore the window size and position:
            self.set_window_geometry( dic.get( 'window' ) )

            # Initialize the default image size:
            self.set_image_size( dic.get( 'image_index', 2 ) )

    #---------------------------------------------------------------------------
    #  Indicate the object has a UI:
    #---------------------------------------------------------------------------

    def has_ui ( self ):
        return True

#-------------------------------------------------------------------------------
#  'ShellPlot' class:
#-------------------------------------------------------------------------------

class ShellPlot ( ShellPlotFrame ):

    #---------------------------------------------------------------------------
    #  Indicate the object has no UI:
    #---------------------------------------------------------------------------

    def has_ui ( self ):
        return False

    #---------------------------------------------------------------------------
    #  Bind the canvas to the container:
    #---------------------------------------------------------------------------

    def set_canvas ( self ):
        self.canvas.container = self.container

    #---------------------------------------------------------------------------
    #  Define the menu for the window:
    #---------------------------------------------------------------------------

    def make_menu ( self, menu_desc ):
        pass

    #---------------------------------------------------------------------------
    #  Create a container object for the plot canvas:
    #---------------------------------------------------------------------------

    def make_plot_container ( self, canvas, **traits ):
        return default_container

    #---------------------------------------------------------------------------
    #  Handle an update request:
    #---------------------------------------------------------------------------

    def update_window ( self ):
        pass

    #---------------------------------------------------------------------------
    #  Make the window topmost:
    #---------------------------------------------------------------------------

    def raise_window ( self ):
        pass

#-------------------------------------------------------------------------------
#  Initialize the factory for creating a top-level window of the right GUI
#  toolkit type:
#-------------------------------------------------------------------------------

ShellPlotFrame_factory = ShellPlot
