#-------------------------------------------------------------------------------
#
#  Define the basic bar (i.e. histogram) chart plot rendering class
#
#  Written by: David C. Morrill
#
#  Date: 12/12/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: BarRenderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from renderer                  import Renderer
from enthought.util.scipyx import Float

#-------------------------------------------------------------------------------
#  'BarRenderer' class:
#-------------------------------------------------------------------------------

class BarRenderer ( Renderer ):
    
    # Select points along the value = 0.0 line:
    select_zero = True
    
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        # Ensure that the index values are monotonically increasing:
        return self._is_index_monotonic( values[0] )
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_value ( self, values ): 
        return ( min( 0.0, min( [ x._min for x in values ] ) ),
                 max( 0.0, max( [ x._max for x in values ] ) ) )
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        value    = values[0]
        percent2 = value.bin_width / 200.0
        data     = value._cached_index.data
        return ( ((1.0 + percent2) * data[0])  - (percent2 * data[1]), 
                 ((1.0 + percent2) * data[-1]) - (percent2 * data[-2]) ) 
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        
        # Get the baseline (i.e. 0.0) and index values:
        value0  = values[0]
        indices = value0._cached_index.data
        
        # Generate the index coordinate for the other corners of the bars:
        binw2   = self._index_dx( indices, value0.bin_width / 200.0 )
        barpct  = value0.bar_width / 100.0
        bart    = (2.0 / (len( values ) + barpct - 1.0)) * binw2
        barw    = barpct * bart
        indices = indices - binw2
        
        for value in values:
            # Cache the bounds of the plot area and the scaling factors:
            self._bounds( value, x, y, dx, dy, sx, sy )
  
            # Map the 'index' and 'value' arrays to the plot area:
            value._xyb0 = value.map_screen( indices, 
                               zeros( ( len( value0.data ), ), Float ), 
                               xy = True )
            value._xyb1 = value.map_screen( indices + barw, value.data, 
                               xy = True )
 
            # Advance the indices to the next value's start point:
            indices += bart     
                         
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _do_draw_canvas ( self, gc, values, ox, oy ):
        gc.set_line_dash( None ) 
        for value in values:
            # Save the drawing offset in case we have to do a 'hittest' later:
            value._plot_ox = ox
            value._plot_oy = oy
            gc.save_state()
            gc.translate_ctm(ox,oy)
            weight = value.outline_weight
            gc.set_fill_color(   value.fill_color_ )
            gc.set_stroke_color( value.outline_color_ )
            gc.set_line_width(   weight )
            gc.begin_path()
            xy0 = value._xyb0
            xy1 = value._xyb1
            for i in range( len( xy0 ) ):
                x0, y0 = xy0[i]
                x1, y1 = xy1[i]
                gc.rect( x0, y0, x1 - x0, y1 - y0 )
            gc.draw_path()
            gc.restore_state()     
 
    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------
 
    def _map_xy ( self, value, x, y ):
 
        # Modify x and y to be their fractional offset into the plot area:
        x = (float( x - value._plot_bounds[0] + value._plot_ox ) / 
             (value._plot_sxdx - 1))
        y = (float( y - value._plot_bounds[1] + value._plot_oy ) / 
             (value._plot_sydy - 1))
  
        # Get the x and y axis information:
        x_axis = value._axis_info_index
        y_axis = value._axis_info
 
        # If the index axis is vertical, flip-flop the values:
        use_y = (not x_axis.horizontal)
        if use_y:
            x_axis, y_axis = y_axis, x_axis
 
        # Scale and translate the x data:
        dxmin = x_axis.tick_min
        dymin = y_axis.tick_min
 
        # Get the plot origin:
        # Note:  Once upon a time, this used plot_origin_ instead of origin_.
        origin = value.origin_
 
        # Invert the x and y values if necessary:
        if (origin & LEFT) == 0:
            x = 1.0 - x
        if origin & TOP:
            y = 1.0 - y
 
        # Calculate (x',y'):
        xp = dxmin + x * (x_axis.tick_max - dxmin)
        yp = dymin + y * (y_axis.tick_max - dymin)
 
        # Get the right x' value for calculating v':
        vi = xp
        if use_y:
            vi = yp
 
        # We can only calculate v' if the index data is sorted:
        vp    = None
        index = value._index()
        dir   = index._get_dir()
        if dir:
            # Find the index into the index data that brackets x':
            index_data = index.data
            i = bin_search1( index_data, vi, dir )
         
            # Make sure x' lies within the range of the index data (it should!):
            if i >= 0:
                low = index_data[i]
             
                if vi == low:
                    # Exact match of an index point:
                    vp = value.data[i]
                else:
                   # Otherwise interpolate between two index/data points:
                    vp = (value.data[i] +
                          ((float( vi - low ) / (index_data[i+1] - low)) *
                           (value.data[i+1] - value.data[i])))
             
                # If the result is 'NaN', map it to 'undefined':
                if isnan( vp ):
                    vp = None
         
        # Return the result:
        return ( xp, yp, vp )
        
    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold = 0.0 ):
        for value in values:
            if self._hittest_rect( x, y, value._xyb0, value._xyb1, threshold):
                return [ value ]
        return []
