#-------------------------------------------------------------------------------
#
#  Define an indexed color bar chart plot rendering class. That is, the values 
#  are used to compute indices into a color LUT that are then used to draw bars 
#  that extend across the entire plotting area, orthogonal to the index axis.
#
#  Written by: David C. Morrill
#
#  Date: 12/15/2005
#
#  (c) Copyright 2005 by Enthought, Inc.
#
#  Classes defined: BarIndexRenderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
    
from enthought.chaco.subdivision_cells import arg_find_runs
    
from bar import BarRenderer

from enthought.util.scipyx import Float

#-------------------------------------------------------------------------------
#  'BarIndexRenderer' class:
#-------------------------------------------------------------------------------

class BarIndexRenderer ( BarRenderer ):
     
    #---------------------------------------------------------------------------
    #  Return the maximum number of values the renderer can process:
    #
    #  Note: A value of '0' means the renderer can handle an unlimited number
    #        of values.
    #---------------------------------------------------------------------------
    
    def _max_values ( self ):
        return 1
    
    #---------------------------------------------------------------------------
    #  Returns the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_value ( self, values ):
        return ( 0.0, 1.0 )
    
    #---------------------------------------------------------------------------
    #  Returns the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        index = values[0]._cached_index
        return ( index._min, index._max )
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        # Get the baseline (i.e. 0.0) and index values:
        value0 = values[0]
        
        # Cache the bounds of the plot area and the scaling factors:
        self._bounds( value0, x, y, dx, dy, sx, sy )
        
        # Generate the index coordinate for the other corners of the bars:
        indices = value0._cached_index.data
        barw    = self._index_dx( indices )
        indices = indices - (barw / 2.0)
        
        # Map the 'index' and 'value' arrays to the plot area:
        n            = len( value0.data )
        value0._xyb0 = value0.map_screen( indices, zeros( ( n, ), Float ), 
                                          xy = True )
        value0._xyb1 = value0.map_screen( indices + barw, ones( ( n, ), Float ), 
                                          xy = True )
                         
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _do_draw_canvas ( self, gc, values, ox, oy ):
        gc.set_line_dash( None ) 
        for value in values:
            # Save the drawing offset in case we have to do a 'hittest' later:
            value._plot_ox = ox
            value._plot_oy = oy
            gc.save_state()
            gc.translate_ctm( ox, oy )
            xy0    = value._xyb0
            xy1    = value._xyb1
            lut    = value._lut
            n      = len( lut )
            dc     = (value._max - value._min + (0.001 * n)) / float( n )
            data   = ((value.data - value._min) / dc).astype( Int )
            ranges = arg_find_runs( data, 'flat' )
            horizontal   = (value.index_axis == 'x')
            x, y, dx, dy = value._plot_bounds
            dx -= 2
            dy -= 2
            for i in range( n ):
                color = lut[i]
                if color is not None:
                    gc.set_fill_color( color )
                    gc.begin_path()
                    for lo, hi in ranges:
                        if data[ lo ] == i:
                            x0, y0 = xy0[ lo ]
                            x1, y1 = xy1[ hi - 1 ]
                            if horizontal:
                                gc.rect( x0, y, x1 - x0, dy )
                            else:
                                gc.rect( x, y0, dx, y1 - y0 )
                    gc.fill_path()
            gc.restore_state()     
        
    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold = 0.0 ):
        # fixme: Implement this...
        return []            
