#--------------------------------------------------------------------------------
#
#  Define the candlestick chart plot rendering class
#
#  Written by: David C. Morrill
#
#  Date: 12/14/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: CandleRenderer
#
#--------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from hloc                      import HLOCRenderer

#-------------------------------------------------------------------------------
#  'CandleRenderer' class:
#-------------------------------------------------------------------------------

class CandleRenderer ( HLOC ):
   
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        # Ensure that 4 values are specified and that the 3rd and 4th values lie 
        # between the 1st and 2nd:
        return (
            (len( values ) == 4)                                           and
            logical_and.reduce( (values[0].data - values[2].data) >= 0.0 ) and 
            logical_and.reduce( (values[2].data - values[1].data) >= 0.0 ) and
            logical_and.reduce( (values[0].data - values[3].data) >= 0.0 ) and 
            logical_and.reduce( (values[3].data - values[1].data) >= 0.0 )
        )
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        
        # Cache the bounds of the plot area and the scaling factors:
        for value in values:
            self._bounds( value, x, y, dx, dy, sx, sy )
 
        # Get the index values:
        value0  = values[0]
        indices = value0._cached_index.data
  
        # Map the 'index' and 'value' arrays to the plot area:
        index_dx     = self._index_dx( indices, value0.bar_width / 200.0 )
        value0._xyh  = value0.map_screen( indices, value0.data,    xy = True )
        value0._xyl  = value0.map_screen( indices, values[1].data, xy = True )
        value0._xyol = value0.map_screen( indices - index_dx, values[2].data, 
                                          xy = True )
        value0._xycr = value0.map_screen( indices + index_dx, values[3].data, 
                                          xy = True )
    
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _do_draw_canvas ( self, gc, values, ox, oy ):
        value0          = values[0]
        value0._plot_ox = ox
        value0._plot_oy = oy
        gc.set_line_dash( None )
        
        # Draw the high-low 'shadow' lines:
        gc.set_line_width( value0.line_weight )
        gc.set_stroke_color( value0.line_color_ )
        self._draw_segments( gc, value0._xyh, value0._xyl, ox, oy ) 
        gc.stroke_path()
 
        # Draw the candlestick bodies:
        choose = ((values[2].data - values[3].data) < 0.0)
        xyol   = value0._xyol
        xycr   = value0._xycr
        colors = ( ( value0.fill_color_,    value0.outline_color_   ), 
                   ( values[1].fill_color_, values[1].outline_color_ ) )
        cur_fill_color, cur_outline_color = colors[0]
        gc.set_line_dash( None )
        gc.set_line_width(   value0.outline_weight )
        gc.set_stroke_color( cur_outline_color )
        gc.set_fill_color(   cur_fill_color )
        gc.begin_path()
        for i in range( len( xyol ) ):
            x0, y0 = xyol[i]
            x1, y1 = xycr[i]
            fill_color, outline_color = colors[ choose[i] ]
            if ((fill_color    != cur_fill_color) or 
                (outline_color != cur_outline_color)):
                gc.draw_path()
                gc.set_fill_color(   fill_color )
                gc.set_stroke_color( outline_color )
                cur_fill_color    = fill_color
                cur_outline_color = outline_color
            gc.move_to( x0 - ox, y0 - oy )
            gc.line_to( x1 - ox, y0 - oy )
            gc.line_to( x1 - ox, y1 - oy )
            gc.line_to( x0 - ox, y1 - oy )
            gc.close_path()
        gc.draw_path()
    
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'foreground' part of the plot (i.e. the part that should
    #  overlay the background part of itself and all of its siblings in the 
    #  same plot canvas).
    #
    #  It should also draw the selection marks for the PlotValue if 
    #  'draw_selection' is true and the PlotValue is selected.
    #---------------------------------------------------------------------------
    
    def _do_draw ( self, gc, values, ox, oy ):
        pass
 
    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------
 
    def _map_xy ( self, value, x, y ):
        raise NotImplementedError    
        
    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold=0.0 ):
        value0 = values[0]
        if self._hittest_rect( x, y, value0._xyol, value0._xycr, threshold ):
            return [ values[2], values[3] ]
        dx, dy = self._delta( value0 )
        if self._hittest_rect( x, y, value0._xyh - dx, value0._xyl + dx, threshold ):
            return [ value0, values[1] ]
        return []
