
from enthought.chaco.plot_base import *
from enthought.chaco.plot_value_cmap import ColorMappedPlotValue
from line import LineRenderer
from scatter_markers import SquareMarker, CircleMarker, TriangleMarker, \
                            Inverted_TriangleMarker, PlusMarker, CrossMarker, \
                            DiamondMarker, PixelMarker, DotMarker, CustomMarker

import time

class ColormappedScatterRenderer ( LineRenderer ):
    """
    Renderer for ColorMappedPlotValue instances.

    Some of the code is lifted from ScatterRenderer.
    """

    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        LineRenderer._layout( self, values, x, y, dx, dy, sx, sy )
        for value in values:
            # Eliminate any (x,y) pairs with 'NaN' or 'Inf' values:
            value._xy = compress( bitwise_and.reduce( isfinite( value._xy ),
                                                      axis = 1 ), value._xy, 0 )

    def _do_draw ( self, gc, value, ox, oy ):
        if isinstance(value, ColorMappedPlotValue):
            if not value._cache_valid:
                value._compute_bands(True)

            # If the GC doesn't have draw_*_at_points, then use bruteforce
            if hasattr(gc, 'draw_marker_at_points') or hasattr(gc, 'draw_path_at_points'):
                batch_capable = True
            else:
                batch_capable = False

            if value._render_method == 'bruteforce' or (not batch_capable):
                self._draw_bruteforce(gc, value, ox, oy)
            elif value._render_method == 'banded':
                self._draw_banded(gc, value, ox, oy)

    def _get_marker_and_size( self, gc, value ):
        avail_markers = [eval(cls) for cls in globals() if cls.endswith('Marker')]
        marker_name  = value.symbol.replace( ' ', '_' ).lower()
        marker_cls = [c for c in avail_markers
                      if c.__name__.lower().startswith(marker_name)]
        if len(marker_cls) != 1:
            return
        else:
            marker_cls = marker_cls[0]
            marker = marker_cls()
        size = value.size_
        return marker, size

    def _set_draw_info(self, gc, mode, color, outline_color, outline_weight):
        if mode == STROKE:
            gc.set_stroke_color( color )
        else:
            gc.set_stroke_color( outline_color )
            gc.set_fill_color( color )
        gc.set_line_width(outline_weight)
        return

    def _draw_banded ( self, gc, value, ox, oy ):
        """
        Requires the input colormapped plotvalue to have already been
        sorted into groups (or "bands") of values sharing the same colors.

        This allows us to use the very fast "draw_marker_at_points" and
        "draw_path_at_points" instead of drawing each point one by one.
        """
        index1, index2, xy = self._get_plot_data( value, ox, oy )
        n = len( xy )

        # Grab the XY values corresponding to each color band of points
        point_bands = {}
        for color_index, indices in value._index_bands.items():
            point_bands[color_index] = take(xy, indices)

        marker, size = self._get_marker_and_size(gc, value)

        # Set up the GC for drawing
        gc.set_line_dash( None )
        if marker.draw_mode == STROKE:
            gc.set_line_width( value.line_weight )
        else:
            gc.set_line_width( value.outline_weight )
        gc.begin_path()

        cmap = value.colormap

        if (hasattr( gc, 'draw_marker_at_points' ) and value.symbol != 'custom'):
            # This is the fastest method: we use one of the built-in markers.
            for color_index, points in point_bands.items():
                self._set_draw_info(gc, marker.draw_mode, cmap.color_bands[color_index], value.outline_color_, value.outline_weight)
                gc.draw_marker_at_points( points, size, marker.kiva_marker )
        elif hasattr( gc, 'draw_path_at_points' ):
            # We have to construct the path for the marker.
            if value.symbol != 'custom':
                path = gc.get_empty_path()
                marker.add_to_path(path, size)
                mode = marker.draw_mode
            else:
                path = value.custom_symbol
                mode = STROKE

            for color_index, points in point_bands.items():
                self._set_draw_info(gc, mode, cmap.color_bands[color_index], value.outline_color_, value.outline_weight)
                gc.draw_path_at_points( points, path, mode )
        else:
            raise RuntimeError, "Batch drawing requested on non-batch-capable GC."
        return

    def _draw_bruteforce ( self, gc, value, ox, oy ):
        """
        Draw all of the datapoints in one go, changing the pen color at each
        one.
        """
        index1, index2, xy = self._get_plot_data( value, ox, oy )
        n = len( xy )

        marker, size = self._get_marker_and_size(gc, value)

        # Set up the GC
        gc.set_line_dash( None )
        if marker.draw_mode == STROKE:
            gc.set_line_width( value.line_weight )
        else:
            gc.set_line_width( value.outline_weight )
        gc.begin_path()

        # Set up the path and draw mode
        if value.symbol != 'custom':
            path = gc.get_empty_path()
            marker.add_to_path(path, size)
            mode = marker.draw_mode
        else:
            path = value.custom_symbol
            mode = STROKE

        cmap = value.colormap

        for low, high in value._xy_ranges:
            high -= index1
            if high > 0:
                low = max( 0, low - index1 )
                if low >= n:
                    break

            for i in range( low, min( n, high ) ):
                line_color = cmap.color_bands[value._color_indices[i]]
                self._set_draw_info(gc, mode, line_color, value.outline_color_, value.outline_weight)
                x,y = xy[i]
                gc.translate_ctm(x, y)
                gc.add_path(path)
                gc.draw_path(mode)
                gc.translate_ctm(-x,-y)
        return

    def _distance_to ( self, value, x, y ):
        """
        Calculate the distance from a given point (x,y) to the closest point
        in the data
        """
        xy  = self._get_xydata( value, x, y )
        # If there's no data, return an infinite distance (just like the line plot)
        if xy is None:
            return PINF
        dpt = xy - array( [ x, y ] )
        result = sqrt( amin( add.reduce( dpt * dpt, 1 ) ) )
        return result

#EOF
