#--------------------------------------------------------------------------------
#
#  Define the 'filled against constant' line plot base class
#
#  Written by: David C. Morrill
#
#  Date: 11/11/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#  Classes defined: FillCRendererBase
#
#--------------------------------------------------------------------------------

#--------------------------------------------------------------------------------
#  Imports:
#--------------------------------------------------------------------------------

from enthought.util.scipyx          import *

from enthought.chaco.plot_base      import bin_search1
from enthought.chaco.plot_container import TOP, LEFT
from line                           import LineRenderer

#-------------------------------------------------------------------------------
#  'FillCRendererBase' class:
#
#  This is the concrete rendering base class for filled against constants line 
#  plots.
#-------------------------------------------------------------------------------

class FillCRendererBase ( LineRenderer ):
    
    #---------------------------------------------------------------------------
    #  Return the 'fill type' to use:
    #     -1: Less than
    #      0: Not equal to
    #     +1: Greater than
    #---------------------------------------------------------------------------
    
    def _plot_type ( self ):
        return 0
     
    #---------------------------------------------------------------------------
    #  Return the maximum number of values the renderer can process:
    #
    #  Note: A value of '0' means the renderer can handle an unlimited number
    #        of values.
    #---------------------------------------------------------------------------
    
    def _max_values ( self ):
        return 0
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        LineRenderer._layout( self, values, x, y, dx, dy, sx, sy )
 
        # Lay the plots out:
        for value in values:
            origin = value.plot_origin_
            flipxy = 1 - 2 * (value.index_axis == 'y')
            constant           = value._constant or 0.0
            value._xy_constant = value.map_screen( None, constant )[flipxy == 1]
            value._fill_info   = fill_areas( value._xy[ ::, ::flipxy ],
                                             value._xy_constant )
    
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _do_draw_canvas ( self, gc, values, ox, oy ):
        plot_type = self._plot_type()
        larger    = (plot_type >= 0)
        smaller   = (plot_type <= 0)
        for value in values:
            index0, index1, xy = self._get_plot_data( value, ox, oy )
            if larger:
                self.draw_areas( gc, value, ox, oy, index0, index1, xy,
                                 value.fill_color_, 1.0 )
            if smaller:
                self.draw_areas( gc, value, ox, oy, index0, index1, xy,
                                 value.outline_color_, 255.0 )
                  
    #---------------------------------------------------------------------------
    #  Fills a specified region type with a specified color:  
    #---------------------------------------------------------------------------
    
    def draw_areas ( self, gc, value, ox, oy, index0, index1, xy, color, type ):
        fn, fx, ft = value._fill_info
        if len( fn ) == 0:
            return
        xy      = value._xy
        y       = value._xy_constant
        flip_xy = (value.index_axis == 'y')
        first   = True
        for i in xrange( max( 0, bin_search1( fn, index0, 1 ) ), len(fn) - 1 ):
            n0 = fn[i]
            if n0 >= index1:
                break
            if ft[i] == type:
                n1 = fn[ i + 1 ]
                x0 = fx[i]
                x1 = fx[ i + 1 ]
                if first:
                    first = False
                    gc.set_fill_color( color )
                    gc.begin_path()
                y0 = y1 = y
                if flip_xy:
                    x0, y0, x1, y1 = y, x0, y, x1
                if ox or oy:
                    slice = xy[ n0: n1] - ( ox, oy )
                else:
                    slice = xy[ n0: n1 ]
                gc.lines( concatenate( ( array( [ [ x0 - ox, y0 - oy ] ] ),
                                         slice,
                                         array( [ [ x1 - ox, y1 - oy ] ] ) ) ) )
                gc.close_path()
        if not first:  
            gc.fill_path()

#-------------------------------------------------------------------------------
#  Calculate the region between a line and a constant:
#-------------------------------------------------------------------------------

def fill_areas ( xy, c ):
    xy2      = concatenate( ( array( [ [ xy[0][0],  c ] ] ), 
                              xy, 
                              array( [ [ xy[-1][0], c ] ] ) ) )
    dxy2     = xy2[1:] - xy2[:-1]
    ix       = xy2[:-1,0] + ((dxy2[:,0] * (c - xy2[:-1,1])) / dxy2[:,1])
    gtc      = (xy2[:,1] > c)
    dgtc     = gtc[1:] - gtc[:-1]
    dgtc[0]  =  cmp( xy[0][1],  c )
    dgtc[-1] = -cmp( xy[-1][1], c )
    return ( nonzero(  dgtc ), compress( dgtc, ix ), compress( dgtc, dgtc ) )
            
