#-------------------------------------------------------------------------------
#
#  Define the High/Low/Open/Close plot rendering class
#
#  Written by: David C. Morrill
#
#  Date: 12/14/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: HLOCRenderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from renderer                  import Renderer

#-------------------------------------------------------------------------------
#  'HLOCRenderer' class:
#-------------------------------------------------------------------------------

class HLOCRenderer ( Renderer ):
    
    #---------------------------------------------------------------------------
    #  Return the maximum number of values the renderer can process:
    #---------------------------------------------------------------------------
    
    def _max_values ( self ):
        return 4
    
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        # Ensure that either 3 or 4 values are specified and that the 3rd (and
        # 4th) values lie between the 1st and 2nd, and that the 1st >= 2nd:
        n = len( values )
        return (
            (3 <= n <= 4) and
            logical_and.reduce( (values[0].data - values[2].data) >= 0.0 ) and 
            logical_and.reduce( (values[2].data - values[1].data) >= 0.0 ) and
            ((n == 3) or
            (logical_and.reduce( (values[0].data - values[3].data) >= 0.0 ) and 
             logical_and.reduce( (values[3].data - values[1].data) >= 0.0 )))
        )
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_value ( self, values ): 
        return ( values[1]._min, values[0]._max )
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        value    = values[0]
        percent2 = value.bar_width / 200.0
        percent1 = percent2 * (len( values ) == 4)
        data     = value._cached_index.data
        return ( ((1.0 + percent1) * data[0])  - (percent1 * data[1]), 
                 ((1.0 + percent2) * data[-1]) - (percent2 * data[-2]) ) 
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        
        # Cache the bounds of the plot area and the scaling factors:
        for value in values:
            self._bounds( value, x, y, dx, dy, sx, sy )
 
        # Get the index values:
        value0  = values[0]
        indices = value0._cached_index.data
  
        # Map the 'index' and 'value' arrays to the plot area:
        index_dx = self._index_dx( indices, value0.bar_width / 200.0 )
        value0._xyh = value0.map_screen( indices, value0.data,    xy = True )
        value0._xyl = value0.map_screen( indices, values[1].data, xy = True )
        valuec = values[2]
        if len( values ) == 4:
            value0._xyol = value0.map_screen( indices - index_dx, valuec.data, 
                                              xy = True )
            value0._xyor = value0.map_screen( indices,  valuec.data, xy = True )
            valuec       = values[3]
        value0._xycl = value0.map_screen( indices, valuec.data, xy = True )
        value0._xycr = value0.map_screen( indices + index_dx, valuec.data, 
                                          xy = True )
    
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'foreground' part of the plot (i.e. the part that should
    #  overlay the background part of itself and all of its siblings in the 
    #  same plot canvas).
    #
    #  It should also draw the selection marks for the PlotValue if 
    #  'draw_selection' is true and the PlotValue is selected.
    #---------------------------------------------------------------------------
    
    def _draw ( self, gc, values, ox, oy ):
        value0          = values[0]
        value0._plot_ox = ox
        value0._plot_oy = oy
        gc.set_line_dash( None )
        
        # Draw the high-low lines:
        gc.set_line_width( value0.line_weight )
        gc.set_stroke_color( value0.line_color_ )
        self._draw_segments( gc, value0._xyh, value0._xyl, ox, oy ) 
        gc.stroke_path()
 
        valuec = values[2]
        if len( values ) == 4:
            # Draw the 'open' bars:
            gc.set_line_width( valuec.line_weight )
            gc.set_stroke_color( valuec.line_color_ )
            self._draw_segments( gc, value0._xyol, value0._xyor, ox, oy ) 
            gc.stroke_path()
            valuec = values[3]
           
        # Draw the 'close' bars:
        gc.set_line_width( valuec.line_weight )
        gc.set_stroke_color( valuec.line_color_ )
        self._draw_segments( gc, value0._xycl, value0._xycr, ox, oy ) 
        gc.stroke_path()
 
    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------
 
    def _map_xy ( self, value, x, y ):
        raise NotImplementedError    
        
    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold ):
        result = []
        value0 = values[0]
        dx, dy = self._delta( value0 )
        if self._hittest_rect( x, y, value0._xyh - dx, value0._xyl + dx, threshold ):
            result = [ value0, values[1] ]
        valuec = values[2]
        if len( values ) == 4:
            dx, dy = self._delta( valuec )
            if self._hittest_rect( x, y, value0._xyol - dy, value0._xyor + dy, threshold ):
                result.append( valuec )
            valuec = values[3]
        dx, dy = self._delta( valuec )
        if self._hittest_rect( x, y, value0._xycl - dy, value0._xycr + dy, threshold ):
            result.append( valuec )
        return result
