"""
Renderer for line plots.  The LineRenderer also serves as the base class
for several other renderers.  (This will probably be fixed in the future.)
"""

from enthought.util.scipyx import add, amin, array, isnan, PINF, putmask, \
                                  transpose, sqrt

from enthought.kiva.constants import JOIN_BEVEL

from enthought.chaco.plot_base import bin_search1, bin_search2, LEFT, TOP

from renderer import Renderer


class LineRenderer ( Renderer ):
    """
    Renderer for line plots.
    """

    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        """
        Layout a list of PlotValues
        """
        for value in values:
            # Cache the bounds of the plot area and the scaling factors:
            translate = self._bounds( value, x, y, dx, dy, sx, sy )
            if ((translate is None) or (translate == ( 0, 0 )) or
                (value._xy is None)):
                # Map the 'index' and 'value' arrays to the plot area:
                value._xy = value.map_screen( value._cached_index.data, 
                                             value.data, xy = True )
            else:
                value._xy += translate


    def _do_draw ( self, gc, value, ox, oy ):
        """
        Draw the plotting area
        """

        color = value.line_color_
        if color is None:
            return
        index1, index2, xy = self._get_plot_data( value, ox, oy )
        n                  = len( xy )
        
        gc.set_stroke_color( color )
        gc.set_line_width( value.line_weight )
        gc.set_line_dash( value.line_style_ ) 
        gc.set_line_join( JOIN_BEVEL ) 
        gc.set_antialias(value.antialias)
        
        gc.begin_path()
        for low, high in value._xy_ranges:
            high -= index1
            if high > 0:
                low = max( 0, low - index1 )
                if low >= n:
                    break
                gc.lines( xy[ low: high ] )
        gc.stroke_path()  
        return


    def _map_xy ( self, value, x, y ):
        """
        Map a specified (x,y) point in window space to the corresponding point
        (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
        of a line orthogonal to the index axis that passes through the point
        (x',y'):
        """

        # Modify x and y to be their fractional offset into the plot area:
        x = (float( x - value._plot_bounds[0] + value._plot_ox ) / 
            (value._plot_sxdx - 1))
        y = (float( y - value._plot_bounds[1] + value._plot_oy ) / 
            (value._plot_sydy - 1))

        # Get the x and y axis information:
        x_axis = value._axis_info_index
        y_axis = value._axis_info

        # If the index axis is vertical, flip-flop the values:
        use_y = (not x_axis.horizontal)
        if use_y:
            x_axis, y_axis = y_axis, x_axis
        
        # Scale and translate the x data:
        dxmin = x_axis.tick_min
        dymin = y_axis.tick_min

        # Get the plot origin:
        origin = value.origin_
##         origin = value.plot_origin_

        # Invert the x and y values if necessary:
        if (origin & LEFT) == 0:
            x = 1.0 - x
        if origin & TOP:
            y = 1.0 - y

        # Calculate (x',y'):
        xp = dxmin + x * (x_axis.tick_max - dxmin)
        if x_axis.log_scale:
            xp = pow(10, xp)

        yp = dymin + y * (y_axis.tick_max - dymin)
        if y_axis.log_scale:
            yp = pow(10, yp)
            
        # Get the right x' value for calculating v':
        vi = xp
        if use_y:
            vi = yp
        
        # We can only calculate v' if the index data is sorted:
        vp    = None
        index = value._index()
        dir   = index._get_dir()
        if dir:
            # Find the index into the index data that brackets x':
            index_data = index.data
            i = bin_search1( index_data, vi, dir )
        
            # Make sure x' lies within the range of the index data (it should!):
            if i >= 0:
                low = index_data[i]
         
                if vi == low:
                    # Exact match of an index point:
                    vp = value.data[i]
                else:
                    # Otherwise interpolate between two index/data points:
                    vp = (value.data[i] +
                            ((float( vi - low ) / (index_data[i+1] - low)) *
                            (value.data[i+1] - value.data[i])))
         
                # If the result is 'NaN', map it to 'undefined':
                if isnan( vp ):
                    vp = None
        
        # Return the result:
        return ( xp, yp, vp )
       

    def _hittest ( self, values, x, y, threshold = 5.0 ):
        """
        Determine if any PlotValue's representation intersects (or lies within a
        small delta of) a specified (x,y) point:
        """
        result = []
        for value in values:
            if self._distance_to( value, x, y ) <= threshold:
                result.append( value )
        return result

    def _line_hit_test(self, values, x, y, threshold ):
        """
        Returns true or false indicating whether or not the point (x,y) lines
        within 'threshold' of the connected line described by 'values'.
        """
        nearby_pts = self._hittest(values, x, y, threshold)
        if len(nearby_pts) == 0:
            return false
        

    def _distance_to ( self, value, x, y ):
        """
        Calculate the distance from a given point (x,y) to the closest point
        on the line.
        """

        # If 'fast' is true, and the data points are sorted in either
        # the x or y direction, then we will restrict the computation to a
        # subset of the data 'near' the (x,y) point (this will give a
        # result which may not be totally accurate, but which should be good
        # enough for 'hit test' applications):
 
        # NOTE: As it stands now, this algorithm should work well in most
        # cases, even for very large plots. The only exceptions are:
        # - A large line plot where the points are not sorted in either the
        #   x or y direction.
        # - A scatter plot or ordered line plot where a very large number of
        #   points are clustered within a few pixels of the specified (x,y)
        #   point.
        xy = self._get_xydata( value, x, y )
        if xy is None:
            return PINF
        dpt = xy - array( [ x, y ] )
#       if (not line.is_visible()) or (len( xy ) <= 1):
        if len( xy ) <= 1:
            return sqrt( amin( add.reduce( dpt * dpt, 1 ) ) )
        d10  = dpt[:-1]
        d20  = dpt[1:]
        d21  = xy[1:] - xy[:-1]
        t    = -(add.reduce( d21 * d10, 1 ) / add.reduce( d21 * d21, 1 ))
        dpt2 = add.reduce( dpt * dpt, 1 )
        temp = transpose( array( [ t, t ] ) ) * d21 + d10
        dt   = add.reduce( temp * temp, 1 )
        putmask( dt, t < 0.0, dpt2[:-1] )
        putmask( dt, t > 1.0, dpt2[1:] )
        try:
            result = sqrt( amin( dt ) )
        except:
            # Probably tried to take sqrt( +INF ):
            return PINF

        # If the result is 'NaN', map it to positive infinity:
        if isnan( result ):
            result = PINF
        return result


    def _get_xydata ( self, value, x, y ):
        """
        Get the subset of data values that apply to a distance calculation
        """
        xy = value._xy
        
        # If this is empty, bin_search2 will complain, and the correct 
        # answer is an empty list anyway
        if len(xy) == 0:
            return None
    
        # Adjust the data by the current drawing offset (if necessary):
        if value._plot_ox or value._plot_oy:
            xy = xy - ( value._plot_ox, value._plot_oy )
           
        dir = self._get_xydir( value )
        if dir != 0:
            pt    = x
            coord = (abs( dir ) == 2)
            if coord:
                pt = y
            index1 = bin_search2( xy, pt - 10.0, coord, dir )
            index2 = bin_search2( xy, pt + 10.0, coord, dir )
            if index1 > index2:
                index1, index2 = index2, index1
            index1 = max( 0, index1 - 1 )
            index2 = min( len( xy ), index2 + 2 )
            xy     = xy[ index1: index2 ]
              
        return xy

# EOF
