#-------------------------------------------------------------------------------
#
#  Define the 'polygon' line plot class
#
#  Written by: David C. Morrill
#
#  Date: 2/21/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#  Classes defined: PolygonRenderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.util.numerix import around, alltrue

from enthought.chaco.plot_base import *
from line                      import LineRenderer

#-------------------------------------------------------------------------------
#  'PolygonRenderer' class:
#-------------------------------------------------------------------------------

class PolygonRenderer ( LineRenderer ):

    def _compute_bounding_box(self, value):
        xmin = min(value.index.data)
        xmax = max(value.index.data)
        ymin = min(value.data)
        ymax = max(value.data)
        return (xmin, ymin, xmax, ymax)


    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        LineRenderer._layout(self, values, x, y, dx, dy, sx, sy)
        
        # We need to compute the new transformed_bounding_box coordinates;
        # Do this after the LineRenderer's ._layout() so that things like
        # plot_bounds are set up properly when we call map_screen().
        for value in values:
            if not hasattr(value, "_bounding_box"):
                pass
            
            if value._bounding_box is None:
                value._bounding_box = self._compute_bounding_box(value)
            
            xmin, ymin, xmax, ymax = value._bounding_box
            ll, ur = value.map_screen(array([xmin, xmax]), array([ymin, ymax]), xy = True)
            value._bounding_box_transformed = (ll[0], ll[1], ur[0], ur[1])

            # Here we also clear out the previously cached distance_to result,
            # since it will not be relevant anymore.
            value._distance_to_cache = None

    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _draw_canvas ( self, gc, values, ox, oy ):
        for value in values:
            # This is somewhat of a hack, but since we 'know' this data is
            # a 'polygon', it's not sorted.  So for a performance boost, we
            # give this hint to the line renderer.
            value._xydir = 0

            
            fill_color = value.fill_color_
            if fill_color is None:
                continue
        
            # Save the drawing offset in case we have to do a 'hittest' later:
            value._plot_ox = ox
            value._plot_oy = oy
            
            if value._xy_ranges is None:
                self._find_nans( value )
     
            xy = value._xy
            if ox or oy:
                xy = xy - ( ox, oy )
               
            # fixme: Do a trivial intersection test here to avoid drawing if
            # the polygon does not intersect the draw region of the plot...
            
            gc.set_fill_color( fill_color )
            gc.begin_path()
            for low, high in value._xy_ranges:
                if high > 0:
                    # HACK: The Agg blending algorithm allows the background
                    # color to leak through at the polygon edges if we do not
                    # round to whole pixels -- hence, the around.
                    gc.lines( around(xy[ low: high ]))
            gc.fill_path()

            return
        
    #---------------------------------------------------------------------------
    #  Determine whether the specified point is inside the polygon:
    #---------------------------------------------------------------------------
 
    def _distance_to ( self, value, x, y ):

        # The first thing to try is if this event is in the same place as the
        # prior event, we can get the previous result from the cache.
        if hasattr(value, "_distance_to_cache") and \
               value._distance_to_cache is not None:
            if value._distance_to_cache[0] == (x,y):
                return value._distance_to_cache[1]
            
        # Since distance_to is expect to be used in hittesting, and its return
        # value will be compared against an unknown threshold, we define a very
        # large value to return if x,y is outside of the polygon and expect
        # it to be larger than any reasonable threshold that might be used
        OUTSIDE_POLYGON_THRESHOLD = 1.0e10
        
        if hasattr(value, "_bounding_box_transformed") and \
                value._bounding_box_transformed is not None:
            # quick reject if x,y is outside the bounding box
            xmin, ymin, xmax, ymax = value._bounding_box_transformed
            if (x < xmin) or (x > xmax) or (y < ymin) or (y > ymax):
                #~ print "xmin:", xmin, "\tx:", x, "\txmax:", xmax
                #~ print "ymin:", ymin, "\ty:", y, "\tymax:", ymax

                value._distance_to_cache = ((x,y), OUTSIDE_POLYGON_THRESHOLD)
                return OUTSIDE_POLYGON_THRESHOLD
        
        xy = value._xy
 
        # Adjust the data by the current drawing offset (if necessary):
        if value._plot_ox or value._plot_oy:
            xy = xy - ( value._plot_ox, value._plot_oy )
           
        # If the polygon is not already closed, then close it:
        if not alltrue(xy[0] == xy[-1]):
            xy = concatenate( ( xy, array( [ xy[0] ] ) ) )
 
        # The following is a Numeric implementation of a 'point in polygon'
        # algorthm by William Randolph Franklin found at:
        #
        #    http://www.ecse.rpi.edu/Homepages/wrf/misc_notes/pnpoly.html
        #
        # and originally written in C as:
        #    
        #    int pnpoly( int npol, float *xp, float *yp, float x, float y ) {
        #       int i, j, c = 0;
        #       for ( i = 0, j = npol-1; i < npol; j = i++ ) {
        #           if ((((yp[i] <= y) && (y < yp[j])) ||
        #               ((yp[j] <= y) && (y < yp[i]))) &&
        #               (x < (xp[j] - xp[i]) * (y - yp[i]) / 
        #               (yp[j] - yp[i]) + xp[i]))
        #              c = !c;
        #       }
        #       return c;
        #    }
            
        dxy   = xy[1:] - xy[:-1]
        dpt   = xy - array( [ x, y ] )
        dpt_y = dpt[:-1,1]
        le0   = less_equal( dpt[1:,1] * dpt_y, 0.0 )
        dxy_y = dxy[:,1]
        not0  = not_equal(  dxy_y, 0.0 )
        lt0   = less( ((dxy[:,0] * dpt_y) / dxy_y) - dpt[:-1,0], 0.0 ) 
        land  = logical_and( logical_and( le0, not0 ), lt0 )
        
        result = ((add.reduce( land ) + 1) & 1) * OUTSIDE_POLYGON_THRESHOLD
        value._distance_to_cache = ((x,y),result)
        
        return result
