#--------------------------------------------------------------------------------
#
#  Define the range plot rendering class
#
#  Written by: David C. Morrill
#
#  Date: 12/12/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: RangeRenderer
#
#--------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from renderer                  import Renderer

#-------------------------------------------------------------------------------
#  'RangeRenderer' class:
#-------------------------------------------------------------------------------

class RangeRenderer ( Renderer ):
    
    #---------------------------------------------------------------------------
    #  Return the maximum number of values the renderer can process:
    #---------------------------------------------------------------------------
    
    def _max_values ( self ):
        return 2
    
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        # Ensure that exactly two values have been specified:
        return (len( values ) == 2)
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        value    = values[0]
        percent2 = value.bar_width / 200.0
        data     = value._cached_index.data
        return ( ((1.0 + percent2) * data[0])  - (percent2 * data[1]), 
                 ((1.0 + percent2) * data[-1]) - (percent2 * data[-2]) ) 
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
 
        # Get the index values:
        value0  = values[0]
        indices = value0._cached_index.data
        
        # Generate the index coordinates for the other edge of the range:
        index_dx = self._index_dx( indices, value0.bar_width / 200.0 )
        indices2 = indices + index_dx
        indices  = indices - index_dx
        
        for value in values:
            # Cache the bounds of the plot area and the scaling factors:
            self._bounds( value, x, y, dx, dy, sx, sy )
 
            # Combine the scaled x,y values and cache the result for one corner:
            value._xyr0 = value.map_screen( indices,  value.data, xy = True )
            value._xyr1 = value.map_screen( indices2, value.data, xy = True )
    
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'foreground' part of the plot (i.e. the part that should
    #  overlay the background part of itself and all of its siblings in the 
    #  same plot canvas).
    #
    #  It should also draw the selection marks for the PlotValue if 
    #  'draw_selection' is true and the PlotValue is selected.
    #---------------------------------------------------------------------------
    
    def _draw ( self, gc, values, ox, oy ):
        value0, value1  = values
        value0._plot_ox = value1._plot_ox = ox
        value0._plot_oy = value1._plot_oy = oy
        xy00      = value0._xyr0
        xy01      = value0._xyr1
        xy10      = value1._xyr0
        xy11      = value1._xyr1
        choose    = ((value0.data - value1.data) < 0.0)
        colors    = ( value0.line_color_, value1.line_color_ )
        cur_color = colors[0]
        gc.set_line_dash( None )
        gc.set_line_width( value0.line_weight )
        gc.set_stroke_color( cur_color )
        for i in range( len( xy00 ) ):
            x00, y00 = xy00[i]
            x01, y01 = xy01[i]
            x10, y10 = xy10[i]
            x11, y11 = xy11[i]
            color    = colors[ choose[i] ]
            if color != cur_color:
                gc.stroke_path()
                gc.set_stroke_color( color )
                cur_color = color
            gc.move_to( x00 - ox, y00 - oy )
            gc.line_to( x01 - ox, y01 - oy )
            gc.move_to( x10 - ox, y10 - oy )
            gc.line_to( x11 - ox, y11 - oy )
            gc.move_to( (x00 + x01) / 2.0 - ox, (y00 + y01) / 2.0 - oy )
            gc.line_to( (x10 + x11) / 2.0 - ox, (y10 + y11) / 2.0 - oy )
        gc.stroke_path()
 
    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------
 
    def _map_xy ( self, value, x, y ):
        raise NotImplementedError    
        
    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold ):
        value0, value1 = values
        dx, dy = self._delta( value0 )
        result = []
        if self._hittest_rect( x, y, value0._xyr0 - dy, value0._xyr1 + dy, threshold ): 
            result = [ value0 ]
        if self._hittest_rect( x, y, value1._xyr0 - dy, value1._xyr1 + dy, threshold ): 
            result.append( value1 )
        if ((not result) and 
            self._hittest_rect( x, y,
                ((value0._xyr0 + value0._xyr1) / 2.0) - dx,
                ((value1._xyr0 + value1._xyr1) / 2.0) + dx ), threshold ): 
            result = [ value0, value1 ]
        return result
