#-------------------------------------------------------------------------------
#
#  Define the base plot rendering classes for drawing plots
#
#  Written by: David C. Morrill
#
#  Date: 12/02/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: Renderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.enable.base import xy_in_bounds
from enthought.chaco.plot_base import *

from enthought.traits.api import HasTraits

#-------------------------------------------------------------------------------
#  'Renderer' class:
#
#  This is the abstract base class for all plot rendering classes
#-------------------------------------------------------------------------------

class Renderer(HasTraits):

    # By default, don't select points along the value = 0.0 line:
    select_zero = False

    #---------------------------------------------------------------------------
    #  Return the maximum number of values the renderer can process:
    #
    #  Note: A value of '0' means the renderer can handle an unlimited number
    #        of values.
    #---------------------------------------------------------------------------

    def _max_values ( self ):
        return 0

    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------

    def _is_valid ( self, values ):
        return True

    #---------------------------------------------------------------------------
    #  Return the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------

    def _get_bounds_value ( self, values ):
        return ( min( [ x._min for x in values ] ),
                 max( [ x._max for x in values ] ) )

    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------

    def _get_bounds_index ( self, values ):
        index = values[0]._cached_index
        return ( index._min, index._max )

    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:
    #---------------------------------------------------------------------------

    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        raise NotImplementedError

    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the
    # same plot canvas).
    #---------------------------------------------------------------------------

    def _draw_canvas ( self, gc, values, ox, oy ):
        if len(values) == 0:
            return
        else:
            # We determine whether or not to backbuffer etc. based on the first
            # plotvalue in the list of values.  If you don't want this behavior,
            # then don't be passing in a list of plotvalues.
            primary_value = values[0]

        if hasattr(primary_value, "backbuffer") and \
                                    primary_value.backbuffer is not None:
            mode = primary_value.backbuffer.mode
        else:
            mode = "draw_front"

        if mode in ("draw_back", "draw_both"):
            if ((not primary_value.backbuffer.draw_once) or
                (not primary_value.backbuffer.is_drawn)):
                back_gc = primary_value.backbuffer.gc
                old_ctm = back_gc.get_ctm()
                back_gc.translate_ctm(-primary_value._plot_bounds[0],
                                      -primary_value._plot_bounds[1])
                self._do_draw_canvas(back_gc, values, 0, 0)
                back_gc.set_ctm(old_ctm)

                # take care of the draw_once setting
                if primary_value.backbuffer.draw_once:
                    primary_value.backbuffer.is_drawn = True

        if mode in ("draw_front", "draw_both"):
            self._do_draw_canvas( gc, values, ox, oy )

    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'foreground' part of the plot (i.e. the part that should
    #  overlay the background part of itself and all of its siblings in the
    # same plot canvas).
    #---------------------------------------------------------------------------

    def _draw ( self, gc, values, ox, oy ):
        for value in values:
            if hasattr(value, "backbuffer") and value.backbuffer is not None:
                mode = value.backbuffer.mode
            else:
                mode = "draw_front"
            if mode in ("draw_back", "draw_both"):
                if ((not value.backbuffer.draw_once) or
                    (not value.backbuffer.is_drawn)):
                    # draw onto the back buffer
                    back_gc = value.backbuffer.gc
                    old_ctm = back_gc.get_ctm()
                    back_gc.translate_ctm(-value._plot_bounds[0],
                                          -value._plot_bounds[1])
                    self._do_draw(back_gc, value, 0, 0)
                    back_gc.set_ctm(old_ctm)

                    # take care of the draw_once setting
                    if value.backbuffer.draw_once:
                        value.backbuffer.is_drawn = True

            if mode in ("draw_front", "draw_both"):
                self._do_draw( gc, value, ox, oy )

    #---------------------------------------------------------------------------
    #  Functions that subclasses should override:
    #---------------------------------------------------------------------------

    def _do_draw ( self, gc, value, ox, oy ):
        pass

    def _do_draw_canvas ( self, gc, values, ox, oy ):
        pass

    #---------------------------------------------------------------------------
    #  Draw the selection markers for any selected values:
    #---------------------------------------------------------------------------

    def _draw_selection ( self, gc, values, ox, oy ):
        for value in values:
            if (value.selected and
                (not self._draw_selection_line( gc, value, ox, oy ))):
                self._draw_selection_box( gc, value, ox, oy )

    #---------------------------------------------------------------------------
    #  Draw the selection markers for the specified value:
    #---------------------------------------------------------------------------

    def _draw_selection_line ( self, gc, value, ox, oy ):

        # Get the index info:
        index = value._index()

        # If the index data is not ordered, we can't draw the selection boxes
        # correctly, so do the default selection drawing:
        if index._get_dir() == 0:
            return False

        # Set up the colors and line style to use for the selection boxes:
        gc.set_line_width( 1 )
        gc.set_stroke_color( black_color )
        gc.set_fill_color( green_color )
        gc.set_line_dash( solid_line_style )

        # Get the scaling factyor so we can scale up the number of selection boxes:
        if value.index_axis == 'x':
            scale = value._plot_sx
        else:
            scale = value._plot_sy

        # Get an evenly spaced (along the index) list of (index,value) pairs:
        n        = int(12 * scale)
        delta    = (index._max - index._min) / (2.0 * (n - 1))
        idata    = index.data
        vdata    = value.data
        map_data = value.map_data
        ns       = map( lambda x: map_data( x )[0],
                        arange( index._min, index._max + delta, 2.0 * delta ) )

        # Draw each of the selection boxes:
        indices = array( map( lambda x: idata[ x ], ns ) )
        for x, y in value.map_screen( indices,
                       array( map( lambda x: vdata[ x ], ns ) ), xy = True ):
            # Only plot the marker if the point itself is within the plot bounds;
            # otherwise, the marker will be clipped incorrectly.
            if xy_in_bounds(x, y, value._plot_bounds):
                gc.begin_path()
                gc.rect( x - 3, y - 3, 6, 6 )
                gc.draw_path()

        if self.select_zero:
            for x, y in value.map_screen( indices, array( [ 0.0 ] * n ),
                                          xy = True ):
                # Only plot the marker if the point itself is within the plot bounds;
                # otherwise, the marker will be clipped incorrectly.
                if xy_in_bounds(x, y, value._plot_bounds):
                    gc.begin_path()
                    gc.rect( x - 3, y - 3, 6, 6 )
                    gc.draw_path()

        # Indicate that the selection markers were drawn:
        return True

    #---------------------------------------------------------------------------
    #  Draw a bounding box selection marker for the specified value:
    #---------------------------------------------------------------------------

    def _draw_selection_box ( self, gc, value, ox, oy ):

        # Get the bounding box and center point info:
        index  = value._index()
        xs, ys = value.map_screen(
                    array( [ index._min, (index._min + index._max) / 2.0,
                             index._max ] ),
                    array( [ value._min, (value._min + value._max) / 2.0,
                             value._max ] ) )

        # Draw the outer selection rectangle:
        gc.set_line_width( 1 )
        gc.set_stroke_color( green_color )

        gc.begin_path()
        gc.move_to( xs[0], ys[0] )
        gc.line_to( xs[0], ys[2] )
        gc.line_to( xs[2], ys[2] )
        gc.line_to( xs[2], ys[0] )
        gc.close_path()
        gc.stroke_path()

        gc.set_stroke_color( black_color )
        gc.set_line_dash( short_dash_line_style )
        gc.begin_path()
        gc.move_to( xs[0], ys[0] )
        gc.line_to( xs[0], ys[2] )
        gc.line_to( xs[2], ys[2] )
        gc.line_to( xs[2], ys[0] )
        gc.close_path()
        gc.stroke_path()

        gc.set_fill_color( green_color )
        gc.set_line_dash( solid_line_style )

        # Draw each of the eight selection boxes:
        for x in xs:
            for y in ys:
                if (x != xs[1]) or (y != ys[1]):
                   gc.begin_path()
                   gc.rect( x - 3, y - 3, 6, 6 )
                   gc.draw_path()

    #---------------------------------------------------------------------------
    #  Draw a set of high/low selection boxes:
    #---------------------------------------------------------------------------

    def _draw_selection_maxmin ( self, gc, value, maxs, mins ):

        n = 12
        if len( mins ) > n:
            tmins = []
            tmaxs = []
            flen  = float( len( mins ) - 1 )
            for jf in arange( 0.0, flen + 0.001, flen / (n - 1) ):
                j = int( jf + 0.5 )
                tmins.append( mins[ j ] )
                tmaxs.append( maxs[ j ] )
            mins = array( tmins )
            maxs = array( tmaxs )

        # Set up the colors and line style to use for the selection boxes:
        gc.set_line_width( 1 )
        gc.set_stroke_color( black_color )
        gc.set_fill_color( green_color )
        gc.set_line_dash( solid_line_style )

        # Draw the two sets of selection boxes:
        for x, y in mins:
            gc.begin_path()
            gc.rect( x - 3, y - 3, 6, 6 )
            gc.draw_path()

        for x, y in maxs:
            gc.begin_path()
            gc.rect( x - 3, y - 3, 6, 6 )
            gc.draw_path()

    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------

    def _map_xy ( self, value, x, y ):
        raise NotImplementedError

    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's representation intersects (or lies within
    #  'threshold' of) a specified (x,y) point:
    #---------------------------------------------------------------------------

    def _hittest ( self, values, x, y, threshold = 5.0):
        raise NotImplementedError

#-------------------------------------------------------------------------------
#  Helper methods:
#-------------------------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Cache the bounds of the plot area and the scaling factors:
    #---------------------------------------------------------------------------

    def _bounds ( self, value, x, y, dx, dy, sx, sy ):
        sxdx   = dx * sx
        sydy   = dy * sy
        pb     = value._plot_bounds
        result = None
        if ((pb is not None) and (dx == pb[2]) and (dy == pb[3]) and
            (sxdx == value._plot_sxdx) and (sxdx == value._plot_sxdx)):
             result = ( x - pb[0], y - pb[1] )
        value._plot_bounds = ( x, y, dx, dy )
        value._plot_sxdx   = sxdx
        value._plot_sydy   = sydy
        value._plot_sx     = sx
        value._plot_sy     = sy
        value._plot_ox     = value._plot_oy = 0
        value._xydir       = None
        return result

    #---------------------------------------------------------------------------
    #  Generate a set of 'width' values for a list of index values using a
    #  specified fraction of the available width between index values:
    #---------------------------------------------------------------------------

    def _index_dx ( self, indices, percentage = 1.0 ):
        if len( indices ) == 0:
            return array([])
        if len( indices ) == 1:
             return array( [ percentage ] )
        indices2     = resize( indices[1:], ( len( indices ), ) )
        indices2[-1] = 2.0 * indices[-1] - indices[-2]
        if percentage == 1.0:
            return indices2 - indices
        return percentage * (indices2 - indices)

    #---------------------------------------------------------------------------
    #  Draw a set of disconnected line segments:
    #---------------------------------------------------------------------------

    def _draw_segments ( self, gc, xy0, xy1, ox, oy ):
        if (ox != 0) or (oy != 0):
            oxy = array( (ox, oy ) )
            xy0 = xy0 - oxy
            xy1 = xy1 - oxy
        for i in range( len( xy0 ) ):
            x0, y0 = xy0[i]
            x1, y1 = xy1[i]
            gc.move_to( x0, y0 )
            gc.line_to( x1, y1 )

    #---------------------------------------------------------------------------
    #  Perform a hit test on a point against a list of rectangles whose
    #  diametrically opposite corners are specified:
    #---------------------------------------------------------------------------

    def _hittest_rect ( self, x, y, xy0, xy1, threshold=0.0 ):
        xy = array( ( x-threshold/2, y-threshold/2 ) )
        return logical_or.reduce(
               logical_and.reduce( ((xy - xy0-threshold) * (xy - xy1-threshold)) <= 0.0, axis = -1 ))

    #---------------------------------------------------------------------------
    #  Compute a set of displacement values for doing a rectangle hittest on
    #  line segments drawn using the specified value's line weight:
    #---------------------------------------------------------------------------

    def _delta ( self, value ):
        dxy = value.line_weight/2 + 3
        dx  = array( ( dxy, 0 ) )
        dy  = array( ( 0, dxy ) )
        if value.index_axis == 'x':
            return dx, dy
        return dy, dx

    #---------------------------------------------------------------------------
    #  Return whether a given value's index is monotonically increasing:
    #---------------------------------------------------------------------------

    def _is_index_monotonic ( self, value ):
        indices = value._cached_index.data
        return logical_and.reduce( (indices[1:] - indices[:-1]) > 0.0 )

    #---------------------------------------------------------------------------
    #  Return the direction the (x,y) points are sorted in (if any):
    #  1 = +x sorted, -1 = -x sorted,
    #  2 = +y sorted, -2 = -y sorted,
    #  0 = Not sorted
    #---------------------------------------------------------------------------

    def _get_xydir ( self, value ):
        if value._xydir is None:
            try:
                diff = value._xy[1:] - value._xy[:-1]
                xdir, ydir = logical_and.reduce( greater_equal( diff, 0.0 ), 0 )
                if xdir:
                    value._xydir = 1
                elif ydir:
                    value._xydir = 2
                else:
                    xdir, ydir = logical_and.reduce( less_equal( diff, 0.0 ), 0)
                    if xdir:
                        value._xydir = -1
                    else:
                        value._xydir = -2 * ydir
            except:
                value._xydir = 0

        return value._xydir

    #---------------------------------------------------------------------------
    #  Get the set of data points inside the plotting bounds if one or both of
    #  the index/value data sets is monotonically increasing or decreasing.
    #---------------------------------------------------------------------------

    def _get_plot_data ( self, value, ox, oy ):

        # Save the drawing offset in case we have to do a 'hittest' later:
        value._plot_ox = ox
        value._plot_oy = oy

        if value._xy_ranges is None:
            self._find_nans( value )

        xy = value._xy

        # If we don't have any data, the rest of those is kind of pointless,
        # so, jump out here
        if len(xy) == 0:
            return (0, -1, xy)

        # Try to perform a binary search to find the endpoints of the region
        # we need to draw. For large, scaled data sets, this can result in a
        # significant reduction in the number of lines that need to be drawn:
        dir = self._get_xydir( value )
        if dir != 0:
            coord = (abs( dir ) == 2)
            px, py, pdx, pdy = value._plot_bounds
            if coord:
                pt1 = py + oy
                pt2 = pt1 + pdy
            else:
                pt1 = px + ox
                pt2 = pt1 + pdx
            index1 = bin_search2( xy, pt1, coord, dir )
            index2 = bin_search2( xy, pt2, coord, dir )
            if index1 > index2:
                index1, index2 = index2, index1
            index1 = max( 0, index1 - 1 )
            index2 = min( len( xy ), index2 + 2 )
            xy     = xy[ index1: index2 ]
        else:
            index1 = 0
            index2 = len( xy )

        if ox or oy:
            xy = xy - ( ox, oy )

        return ( index1, index2, xy )

    #---------------------------------------------------------------------------
    #  Divide the data into ranges that do not contain any 'NaN's:
    #---------------------------------------------------------------------------

    def _find_nans ( self, value ):

        # Short circuit if we have no NaN's.
        if alltrue(isfinite(value.data)):
            value._xy_ranges = array([[0, len(value.data)]])
            return

        nans = isnan( concatenate( ( nan_array, value.data, nan_array ) ) )
        s1   = nans[:-1]
        s2   = nans[1:]
        value._xy_ranges = concatenate(
                             ( nonzero( greater( s1, s2 ) )[:,NewAxis],
                               nonzero( less(    s1, s2 ) )[:,NewAxis] ), -1 )
