"""
Renderer for scatter plots.
"""

from enthought.chaco.plot_base import *
from enthought.util.numerix import asarray
from line                      import LineRenderer
from scatter_markers import SquareMarker, CircleMarker, TriangleMarker, \
                            Inverted_TriangleMarker, PlusMarker, CrossMarker, \
                            DiamondMarker, DotMarker, PixelMarker, CustomMarker


class ScatterRenderer ( LineRenderer ):
    """
    Renderer for scatter plots.
    """

    # This class currently subclasses from LineRenderer, and all 3 of its
    # methods override/shadow methods on LineRenderer.  In the future the
    # shared bits will be refactored out into a cleaner base class.

    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        LineRenderer._layout( self, values, x, y, dx, dy, sx, sy )
        for value in values:
            # Eliminate any (x,y) pairs with 'NaN' or 'Inf' values:
            value._xy = compress( bitwise_and.reduce( isfinite( value._xy ),
                                                      axis = 1 ), value._xy, 0 )


    def _do_draw ( self, gc, value, ox, oy ):
        index1, index2, xy = self._get_plot_data( value, ox, oy )
        n                  = len( xy )

        avail_markers = [eval(cls) for cls in globals() if cls.endswith('Marker')]
        marker_name  = value.symbol.replace( ' ', '_' ).lower()
        marker_cls = [c for c in avail_markers
                      if c.__name__.lower().startswith(marker_name)]
        if len(marker_cls) != 1:
            return
        else:
            marker_cls = marker_cls[0]
            marker = marker_cls()

        gc.set_line_dash( None )
        if marker_cls.draw_mode == STROKE:
            gc.set_stroke_color( value.line_color_ )
            gc.set_line_width( value.line_weight )
        else:
            gc.set_stroke_color( value.outline_color_ )
            gc.set_line_width( value.outline_weight )
            gc.set_fill_color( value.fill_color_ )

        gc.begin_path()
        size = value.size_

        for low, high in value._xy_ranges:
            high -= index1
            if high > 0:
                low = max( 0, low - index1 )
                if low >= n:
                    break
            # This is the fastest method - we use one of the built-in
            # markers.  If available, try it first:
            if (hasattr( gc, 'draw_marker_at_points' ) and
                        value.symbol != 'custom'):
                if gc.draw_marker_at_points( xy, size, marker.kiva_marker ) != 0:
                    continue
                else:
                    # the draw_marker_at_points() call was unsuccessful, so we
                    # pass through to the next drawing method.
                    pass


            # The 2nd fastest is to draw the path once into a compiled path,
            # and then draw that path at various points.
            if hasattr( gc, 'draw_path_at_points' ):
                if value.symbol != 'custom':
                    path = gc.get_empty_path()
                    marker.add_to_path(path, size)
                    mode = marker.draw_mode
                else:
                    path = value.custom_symbol
                    mode = STROKE
                xy         = asarray( xy[ low: min( n, high ) ] )
                gc.draw_path_at_points( xy, path, mode )
                continue

            # Neither of the special case functions were (a) available,
            # or (b) worked.  Use the standard drawing methods:
            if value.symbol != 'custom':
                for i in range( low, min( n, high ) ):
                    gc.save_state()
                    x, y = xy[i]
                    gc.translate_ctm( x, y )
                    # Kiva GCs have a path-drawing interface
                    marker.add_to_path(gc, size)
                    gc.draw_path( marker.draw_mode )
                    gc.restore_state()
            else:
                path = value.custom_symbol
                for i in range( low, min( n, high ) ):
                    gc.save_state()
                    x, y = xy[i]
                    gc.translate_ctm(x, y)
                    gc.add_path(path)
                    gc.draw_path(STROKE)
                    gc.restore_state()


    def _distance_to ( self, value, x, y ):
        """
        Calculate the distance from a given point (x,y) to the closest point
        in the data:
        """
        xy  = self._get_xydata( value, x, y )
        # If there's no data, return an infinite distance (just like the line plot)
        if xy is None:
            return PINF
        dpt = xy - array( [ x, y ] )
        result = sqrt( amin( add.reduce( dpt * dpt, 1 ) ) )
        return result

# EOF
