#-------------------------------------------------------------------------------
#
#  Define the segmented line plot rendering class for drawing plots
#
#  Written by: David C. Morrill
#
#  Date: 5/06/2003
#
#  (c) Copyright 2003 by Enthought, Inc.
#
#  Classes defined: SegmentRenderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from line                      import LineRenderer
from enthought.util.numerix    import transpose

#-------------------------------------------------------------------------------
#  'SegmentRenderer' class:
#
#  This is the concrete rendering class for segmented line plots.
#-------------------------------------------------------------------------------

class SegmentRenderer ( LineRenderer ):
   
    #---------------------------------------------------------------------------
    #  Draw the plotting area:
    #---------------------------------------------------------------------------
 
    def _do_draw ( self, gc, value, ox, oy ):
        
        color = value.line_color_
        if color is None:
            return
        
        index1, index2, xy = self._get_plot_data( value, ox, oy )
        n                  = len( xy )
 
        gc.set_stroke_color( color )
        gc.set_line_width( value.line_weight )
        gc.set_line_dash( value.line_style_ ) 
        
        gc.begin_path()
        for low, high in value._xy_ranges:
            high -= index1
            if high > 0:
                low = max( 0, low - index1 )
                if low >= n:
                    break
                for i in range( low, high - 1, 2 ):
                    x0, y0 = xy[ i ]
                    gc.move_to( x0, y0 )
                    x1, y1 = xy[ i + 1 ]
                    gc.line_to( x1, y1 )
        gc.stroke_path()  

    #---------------------------------------------------------------------------
    #  Calculate the distance from a given point (x,y) to the closest point
    #  on the line:
    #
    #  Note: This method is very similar to the method it overrides. The two
    #  main differences are:
    #  - The range of values tested is not reduced to the neighborhood of the
    #    point of interest.
    #  - Only every other computed distance is used in producing the result.
    #
    #  The reason for these changes is that the segment renderer 'skips' drawing
    #  the line between every other pair of points, and so the distances to the
    #  'missing' line segments should not affect the result (hence the reason
    #  for only using every other distance). Reducing the range of values tested
    #  can also cause the points tested to be 'misaligned', resulting in the
    #  'missing' line segments being used instead of the 'real' line segments.
    #  Hence the full range of points is used in the computation to prevent
    #  misalignment.
    #---------------------------------------------------------------------------
 
    def _distance_to ( self, value, x, y ):
        xy = value._xy
 
        # Adjust the data by the current drawing offset (if necessary):
        if value._plot_ox or value._plot_oy:
            xy = xy - ( value._plot_ox, value._plot_oy )
 
        dpt = xy - array( [ x, y ] )
        if len( xy ) <= 1:
            return sqrt( amin( add.reduce( dpt * dpt, 1 ) ) )
        d10  = dpt[:-1]
        d20  = dpt[1:]
        d21  = xy[1:] - xy[:-1]
        t    = -(add.reduce( d21 * d10, 1 ) / add.reduce( d21 * d21, 1 ))
        dpt2 = add.reduce( dpt * dpt, 1 )
        temp = transpose( array( [ t, t ] ) ) * d21 + d10
        dt   = add.reduce( temp * temp, 1 )
        putmask( dt, t <  0.0, dpt2[:-1] )
        putmask( dt, t > 1.0, dpt2[1:] )
        result = sqrt( amin( dt[::2] ) )
 
        # If the result is 'NaN', map it to positive infinity:
        if isnan( result ):
            result = PINF
 
        return result
