#--------------------------------------------------------------------------------
#
#  Define the stacked bar chart plot rendering class
#
#  Written by: David C. Morrill
#
#  Date: 12/05/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: StackedBarRenderer
#
#--------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from renderer                  import Renderer
from enthought.util.scipyx import Float

#-------------------------------------------------------------------------------
#  'StackedBarRenderer' class:
#-------------------------------------------------------------------------------

class StackedBarRenderer ( Renderer ):
   
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        # Ensure that the index values are monotonically increasing and that no
        # data value < 0.0:
        if not self._is_index_monotonic( values[0] ):
            return False
        for value in values:
            if amin( value.data ) < 0.0:
                return False
        return True
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_value ( self, values ): 
        totals = zeros( ( len( values[0].data ), ), Float )
        for value in values:
            totals = totals + value.data
        return ( 0.0, amax( totals ) )
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        value    = values[0]
        percent2 = value.bar_width / 200.0
        data     = value._cached_index.data
        if len( data ) < 2:
            return ( data[0] - 0.5, data[0] + 0.5 )
        return ( ((1.0 + percent2) * data[0])  - (percent2 * data[1]), 
                 ((1.0 + percent2) * data[-1]) - (percent2 * data[-2]) ) 
    
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        
        # Get the baseline (i.e. 0.0) and index values:
        value0  = values[0]
        totals  = baseline = zeros( ( len( value0.data ), ), Float )
        indices = value0._cached_index.data
        
        # Generate the index coordinate for the other corners of the bars:
        index_dx = self._index_dx( indices, value0.bar_width / 200.0 )
        indices2 = indices + index_dx
        indices  = indices - index_dx
        
        for value in values:
            # Cache the bounds of the plot area and the scaling factors:
            self._bounds( value, x, y, dx, dy, sx, sy )
  
            # Map the 'index' and 'value' arrays to the plot area:
            totals       = totals + value.data
            value._xysb  = value.map_screen( indices,  totals, xy = True )
            value._xysb2 = value.map_screen( indices2, totals, xy = True )
            
        # Generate the baseline coordinates and cache them in the first value:
        value0._xysb0  = value0.map_screen( indices,  baseline, xy = True )
        value0._xysb02 = value0.map_screen( indices2, baseline, xy = True )
 
        # !!This should be incorporated into (replace) the above, but I was 
        # fuzzy on how to do that.
        mins = value0._xysb0
        for value in values:
            value._rects = concatenate( ( mins, value._xysb2 - mins ), 1 )
            mins         = value._xysb
       
    #---------------------------------------------------------------------------
    #  Draw the selection markers for any selected values: 
    #---------------------------------------------------------------------------
    
    def _draw_selection ( self, gc, values, ox, oy ):
        i = 0
        for value in values:
            if value.selected:
                if i == 0:
                    mins = (value._xysb0 + value._xysb02) / 2.0
                else:
                    temp = values[ i - 1 ]
                    mins = (temp._xysb + temp._xysb2) / 2.0
                maxs = (value._xysb + value._xysb2) / 2.0
                self._draw_selection_maxmin( gc, value, maxs, mins )
 
            # Move on to the next value:
            i += 1
   
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _do_draw_canvas ( self, gc, values, ox, oy ):
        gc.set_line_dash( None ) 
        for value in values:
            # Save the drawing offset in case we have to do a 'hittest' later:
            value._plot_ox = ox
            value._plot_oy = oy
 
            # Draw the rectangles using the appropriate drawing attributes:
            gc.save_state()
            gc.translate_ctm( -ox, -oy )
            gc.set_fill_color( value.fill_color_ )
            gc.set_stroke_color( value.outline_color_ )
            gc.set_line_width( value.outline_weight )
            gc.rects( value._rects )
            gc.draw_path()  
            gc.restore_state()
       
    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------
 
    def _map_xy ( self, value, x, y ):
 
        # Modify x and y to be their fractional offset into the plot area:
        x = (float( x - value._plot_bounds[0] + value._plot_ox ) / 
             (value._plot_sxdx - 1))
        y = (float( y - value._plot_bounds[1] + value._plot_oy ) / 
             (value._plot_sydy - 1))
  
        # Get the x and y axis information:
        x_axis = value._axis_info_index
        y_axis = value._axis_info
 
        # If the index axis is vertical, flip-flop the values:
        use_y = (not x_axis.horizontal)
        if use_y:
            x_axis, y_axis = y_axis, x_axis
 
        # Scale and translate the x data:
        dxmin = x_axis.tick_min
        dymin = y_axis.tick_min
 
        # Get the plot origin:
        # Note: Once upon a time this was value.plot_origin_
        origin = value.origin_
 
        # Invert the x and y values if necessary:
        if (origin & LEFT) == 0:
            x = 1.0 - x
        if origin & TOP:
            y = 1.0 - y
 
        # Calculate (x',y'):
        xp = dxmin + x * (x_axis.tick_max - dxmin)
        yp = dymin + y * (y_axis.tick_max - dymin)
 
        # Get the right x' value for calculating v':
        vi = xp
        if use_y:
            vi = yp
 
        # We can only calculate v' if the index data is sorted:
        vp    = None
        index = value._index()
        dir   = index._get_dir()
        if dir:
            # Find the index into the index data that brackets x':
            index_data = index.data
            i = bin_search1( index_data, vi, dir )
          
            # Make sure x' lies within the range of the index data (it should!):
            if i >= 0:
                low = index_data[i]
              
                if vi == low:
                    # Exact match of an index point:
                    vp = value.data[i]
                else:
                    # Otherwise interpolate between two index/data points:
                    vp = (value.data[i] +
                          ((float( vi - low ) / (index_data[i+1] - low)) *
                           (value.data[i+1] - value.data[i])))
              
                # If the result is 'NaN', map it to 'undefined':
                if isnan( vp ):
                    vp = None
 
        # Return the result:
        return ( xp, yp, vp )
       
    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's stacked bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold ):
        mins = values[0]._xysb02
        for value in values:
            if self._hittest_rect( x, y, mins, value._xysb, threshold ):
                return [ value ]
            mins = value._xysb2
        return []
