#--------------------------------------------------------------------------------
#
#  Define the stacked line plot rendering class
#
#  Written by: David C. Morrill
#
#  Date: 12/11/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: StackedLineRenderer
#
#--------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from line                      import LineRenderer
from enthought.util.scipyx import Float

#-------------------------------------------------------------------------------
#  'StackedLineRenderer' class:
#-------------------------------------------------------------------------------

class StackedLineRenderer ( LineRenderer ):
    
    # Select points along the value = 0.0 line:
    select_zero = True
   
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        # Ensure that the index values are monotonically increasing and that no
        # data value < 0.0:
        if not self._is_index_monotonic( values[0] ):
            return False
        for value in values:
            if amin( value.data ) < 0.0:
                return False
        return True
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_value ( self, values ): 
        totals = zeros( ( len( values[0].data ), ), Float )
        for value in values:
            totals += value.data
        return ( 0.0, amax( totals ) )
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        data = values[0]._cached_index.data
        return ( data[0], data[-1] )
    
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
 
        # Get the baseline (i.e. 0.0) and index values:
        value0  = values[0]
        totals  = baseline = zeros( ( len( value0.data ), ), Float )
        indices = value0._cached_index.data
        
        for value in values:
            # Cache the bounds of the plot area and the scaling factors:
            self._bounds( value, x, y, dx, dy, sx, sy )
  
            # Map the 'index' and 'value' arrays to the plot area:
            totals   += value.data
            value._xy = value.map_screen( indices, totals, xy = True )
 
        value0._xy0 = value0.map_screen( array( [ indices[0], indices[-1] ] ), 
                                         array( [ 0.0, 0.0 ] ), xy = True )
   
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'background' part of the plot (i.e. the part that should
    #  underlie the foreground part of itself and all of its siblings in the 
    #  same plot canvas).
    #---------------------------------------------------------------------------
    
    def _do_draw_canvas ( self, gc, values, ox, oy ):
        mins = values[0]._xy0
        for value in values:
            maxs = value._xy
            gc.set_fill_color( value.fill_color_ )
            gc.begin_path()
            if ox or oy:
                gc.lines( concatenate( ( maxs, mins[::-1] ), axis=0 ) - 
                          ( ox, oy ) )
            else:
                gc.lines( concatenate( ( maxs, mins[::-1] ), axis=0 ) ) 
            gc.close_path()
            gc.fill_path()
            mins = maxs
