#-------------------------------------------------------------------------------
#
#  Define the Error Bar plot rendering class for SymmetricErrorbarPlotValues.
#
#  Written by: Peter Wang (based on errorbar.py)
#
#  Date: 7/6/2005
#
#  (c) Copyright 2005 by Enthought, Inc.
#
#  Classes defined: SymmetricErrorBarRenderer
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from enthought.chaco.plot_base import *
from enthought.chaco.plot_value import SymmetricErrorbarPlotValue
from line                      import LineRenderer

#-------------------------------------------------------------------------------
#  'ErrorBarRenderer' class:
#-------------------------------------------------------------------------------

class SymmetricErrorBarRenderer ( LineRenderer ):
    
    #---------------------------------------------------------------------------
    #  Return the maximum number of values the renderer can process:
    #---------------------------------------------------------------------------
    
    def _max_values ( self ):
        return 0
    
    #---------------------------------------------------------------------------
    #  Determine if the renderer can correctly plot a set of values:
    #---------------------------------------------------------------------------
    
    def _is_valid ( self, values ):
        for value in values:
            if not isinstance(value, SymmetricErrorbarPlotValue):
                return False
        else:
            return True
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the value part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_value ( self, values ):
        cur_min = None
        cur_max = None
        for value in values:
            if len(value.data) == 0:
                continue
            else:
                tmp_min = min(value.data - value.error_value)
                tmp_max = max(value.data + value.error_value)
            if cur_min is None or tmp_min < cur_min:
                cur_min = tmp_min
            if cur_max is None or tmp_max > cur_max:
                cur_max = tmp_max
        if cur_min is None or cur_max is None:
            return (0.0, 1.0)
        else:
            return (cur_min, cur_max)
    
    #---------------------------------------------------------------------------
    #  Return the min/max range for the index part of a set of PlotValues:
    #---------------------------------------------------------------------------
    
    def _get_bounds_index ( self, values ):
        cur_min = None
        cur_max = None
        for value in values:
            percent = value.bar_width / 200.0
            data    = value._cached_index.data
            if len(data) >= 2:
                tmp_min = ((1.0 + percent) * data[0])  - (percent * data[1]) 
                tmp_max = ((1.0 + percent) * data[-1]) - (percent * data[-2])
                if cur_min is None or tmp_min < cur_min:
                    cur_min = tmp_min
                if cur_max is None or tmp_max > cur_max:
                    cur_max = tmp_max
        if cur_min is None or cur_max is None:
            return (0.0, 1.0)
        else:
            return (cur_min, cur_max)
     
    #---------------------------------------------------------------------------
    #  Layout a list of PlotValues:  
    #---------------------------------------------------------------------------
    
    def _layout ( self, values, x, y, dx, dy, sx, sy ):
        
        # Cache the bounds of the plot area and the scaling factors:
        for value in values:
            self._bounds( value, x, y, dx, dy, sx, sy )

            # Get the index values:
            indices = value._cached_index.data
            
            valueh = valuel = value.error_value
      
            # Map the 'index' and 'value' arrays to the plot area:
            index_dx     = self._index_dx( indices, value.bar_width / 200.0 )
            yh           = value.data + valueh
            yl           = value.data - valuel
            xl           = indices - index_dx
            xr           = indices + index_dx
            value._xy   = value.map_screen( indices, value.data, xy = True )
            value._xyh  = value.map_screen( indices, yh,          xy = True )
            value._xyl  = value.map_screen( indices, yl,          xy = True )
            value._xyhl = value.map_screen( xl, yh,               xy = True )
            value._xyhr = value.map_screen( xr, yh,               xy = True )
            value._xyll = value.map_screen( xl, yl,               xy = True )
            value._xylr = value.map_screen( xr, yl,               xy = True )
    
    #---------------------------------------------------------------------------
    #  Draw a list of PlotValues:
    #
    #  This draws the 'foreground' part of the plot (i.e. the part that should
    #  overlay the background part of itself and all of its siblings in the 
    #  same plot canvas).
    #
    #  It should also draw the selection marks for the PlotValue if 
    #  'draw_selection' is true and the PlotValue is selected.
    #---------------------------------------------------------------------------
    
    def _draw ( self, gc, values, ox, oy ):
        for value in values:
            self._do_draw( gc, value, ox, oy )  # call LineRenderer _do_draw
            value._plot_ox = ox
            value._plot_oy = oy
            gc.set_line_dash( None )
        
            # Draw the error bars:
            gc.set_line_width( value.outline_weight )
            gc.set_stroke_color( value.outline_color_ )
            self._draw_segments( gc, value._xyh, value._xyl,   ox, oy ) 
            self._draw_segments( gc, value._xyhl, value._xyhr, ox, oy ) 
            self._draw_segments( gc, value._xyll, value._xylr, ox, oy ) 
            gc.stroke_path()
 
    #---------------------------------------------------------------------------
    #  Map a specified (x,y) point in window space to the corresponding point
    #  (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
    #  of a line orthogonal to the index axis that passes through the point
    #  (x',y'):
    #
    #  NOTE: This routine currently does not take into account 'log scaling':
    #---------------------------------------------------------------------------
 
    def _map_xy ( self, value, x, y ):
        """
        Map a specified (x,y) point in window space to the corresponding point
        (x',y',v') in a PlotValue's data space (v' is the PlotValue intercept
        of a line orthogonal to the index axis that passes through the point
        (x',y'):
        """
        # FIXME: It seems like _map_xy is always the same of the different
        # renderers.  This is just a copy of the one from line.  These
        # should probably just be put in the base class.

        # Modify x and y to be their fractional offset into the plot area:
        x = (float( x - value._plot_bounds[0] + value._plot_ox ) / 
            (value._plot_sxdx - 1))
        y = (float( y - value._plot_bounds[1] + value._plot_oy ) / 
            (value._plot_sydy - 1))

        # Get the x and y axis information:
        x_axis = value._axis_info_index
        y_axis = value._axis_info

        # If the index axis is vertical, flip-flop the values:
        use_y = (not x_axis.horizontal)
        if use_y:
            x_axis, y_axis = y_axis, x_axis
        
        # Scale and translate the x data:
        dxmin = x_axis.tick_min
        dymin = y_axis.tick_min

        # Get the plot origin:
        origin = value.origin_
##         origin = value.plot_origin_

        # Invert the x and y values if necessary:
        if (origin & LEFT) == 0:
            x = 1.0 - x
        if origin & TOP:
            y = 1.0 - y

        # Calculate (x',y'):
        xp = dxmin + x * (x_axis.tick_max - dxmin)
        if x_axis.log_scale:
            xp = pow(10, xp)

        yp = dymin + y * (y_axis.tick_max - dymin)
        if y_axis.log_scale:
            yp = pow(10, yp)
            
        # Get the right x' value for calculating v':
        vi = xp
        if use_y:
            vi = yp
        
        # We can only calculate v' if the index data is sorted:
        vp    = None
        index = value._index()
        dir   = index._get_dir()
        if dir:
            # Find the index into the index data that brackets x':
            index_data = index.data
            i = bin_search1( index_data, vi, dir )
        
            # Make sure x' lies within the range of the index data (it should!):
            if i >= 0:
                low = index_data[i]
         
                if vi == low:
                    # Exact match of an index point:
                    vp = value.data[i]
                else:
                    # Otherwise interpolate between two index/data points:
                    vp = (value.data[i] +
                            ((float( vi - low ) / (index_data[i+1] - low)) *
                            (value.data[i+1] - value.data[i])))
         
                # If the result is 'NaN', map it to 'undefined':
                if isnan( vp ):
                    vp = None
        
        # Return the result:
        return ( xp, yp, vp )
       

    #---------------------------------------------------------------------------
    #  Determine if any PlotValue's bar representation intersects a
    #  specified point:
    #---------------------------------------------------------------------------
 
    def _hittest ( self, values, x, y, threshold ):
        result = []
        ###value0 = values[0]
        ###dx, dy = self._delta( value0 )
        ###if self._hittest_rect( x, y, value0._xyh - dx, value0._xyl + dx):
        ###    result = [ value0, values[1] ]
        ###valuec = values[2]
        ###if len( values ) == 4:
        ###    dx, dy = self._delta( valuec )
        ###    if self._hittest_rect( x, y, value0._xyol - dy, value0._xyor + dy ):
        ###        result.append( valuec )
        ###    valuec = values[3]
        ###dx, dy = self._delta( valuec )
        ###if self._hittest_rect( x, y, value0._xycl - dy, value0._xycr + dy ):
        ###    result.append( valuec )
        return result
