import unittest
import pdb

from enthought.util.numerix import *
from enthought.util.testingx import *

from enthought.chaco.datamapper import BruteForceDataMapper
from enthought.chaco.subdivision_mapper import SubdivisionDataMapper, \
                                               SubdivisionLineDataMapper
from enthought.chaco.subdivision_cells import find_runs


data0 = array([])
data1 = array([[1,4], [5,2], [2,3], [4,8]])
data1_ascending = array([[1,4], [2,3], [4,8], [5,2]])
data1_descending = data1_ascending[::-1]
data2 = transpose(array([[5,0,1,9,8,2.5,5.1,8.7,3,2.2], [8,0,3,7,0,9,6,9,5,0]]))

# grid data is a (2,) array in a 8x8 grid
gridsize = 4
griddata1 = array([(i,j) for i in range(gridsize) for j in range(gridsize)], \
                    Float)

class UtilityTestCase(unittest.TestCase):
    def check_find_runs_middle(self):
        x = array([0,8,7,8,9,2,3,4,10])
        assert_equal(find_runs(x) , [[0], [8], [7,8,9], [2,3,4], [10]])

    def check_find_runs_start(self):
        x = array([3,4,5,12,9,17])
        assert_equal(find_runs(x) , [[3,4,5],[12],[9],[17]])

    def check_find_runs_end(self):
        x = array([18,23,24,25])
        assert_equal(find_runs(x) , [[18],[23,24,25]])

    def check_find_runs_offset(self):
        # because of the nature of the find_runs algorithm, there may be
        # fencepost errors with runs that start at x[1] or x[-2]
        x = array([10,12,13,14,28,16])
        assert_equal(find_runs(x) , [[10],[12,13,14],[28],[16]])
        x = array([10,15,16,17,34])
        assert_equal(find_runs(x) , [[10],[15,16,17],[34]])

    def check_find_runs_none(self):
        x = array([])
        assert_equal(find_runs(x) , [])
        x = array([12,15,27])
        assert_equal(find_runs(x) , [[12],[15],[27]])

    def check_find_runs_descending(self):
        x = array([30,41,40,39,38,37,12])
        assert_equal(find_runs(x, order='descending') , \
                            [[30], [41,40,39,38,37], [12]])


class DataMapperTestCase(unittest.TestCase):
    """
    This exercises the interfaces and confirms that the data mapper
    adheres to them.
    """

    # child classes must override this with a valid DataMapper class
    DataMapper = None

    def check_transpose_handling(self):
        dm = self.DataMapper(transpose(data1))
        assert_equal(transpose(data1) , dm.get_data())
        dm.sort_order = 'ascending'
        assert_equal(transpose(data1_ascending) , dm.get_data())

    def check_sort_on_construction(self):
        # test sorting on construction
        dm = self.DataMapper(data1, sort_order = 'ascending')
        assert_equal(dm.get_data() , data1_ascending)
        dm = self.DataMapper(data1, sort_order = 'descending')
        assert_equal(dm.get_data() , data1_descending)

    def check_sort_in_situ(self):
        # test sorting with an existing data set
        dm = self.DataMapper(data1, data_sorting='none', sort_order = 'descending')
        assert_equal(dm.get_data() , data1_descending)
        dm.sort_order = 'ascending'
        assert_equal(dm.get_data() , data1_ascending)

    def check_already_sorted(self):
        # test sorting of an already-sorted data set
        dm = self.DataMapper(sort_order='descending')
        dm.set_data(data1_ascending, 'ascending')
        assert_equal(dm.get_data() , data1_descending)

    def check_empty_data(self):
        dm = self.DataMapper(sort_order='ascending')
        dm.set_data(data0)
        dm.set_data(data1)
        dm.set_data(data0)
        dm.set_data(data1)


class BruteForceTestCase(DataMapperTestCase):
    DataMapper = BruteForceDataMapper

    def check_get_near(self):
        dm = self.DataMapper(griddata1, sort_order = 'ascending')

        # check for non-intersecting point
        all_data = dm.get_data()
        ret = dm.get_points_near([(1.5, 1.5)])
        assert_equal(ret , all_data)

        # check for intersecting point
        ret = dm.get_points_near([(1, 1)])
        assert_equal(ret , all_data)


class SubdivisionTestCase(DataMapperTestCase):
    DataMapper = SubdivisionDataMapper

    def check_get_near_points1(self):
        # insert a grid of sorted data
        dm = self.DataMapper(sort_order='ascending')
        dm._points_per_cell = 1
        dm.set_data(griddata1, new_data_sorting='ascending')
        assert_equal(dm.get_points_near(array([[1,1]])) , array([1.0,1.0]))

    def check_get_near_points2(self):
        # insert a list of unordered, non-uniform data
        dm = self.DataMapper(sort_order='ascending')
        dm._points_per_cell = 1
        dm.set_data(data2, new_data_sorting='none')
        assert_equal(dm.get_points_near(array([[2.1,0.1]])) , array([[0.0,0.0],[1.0,3.0]]))

    def check_get_near_points3(self):
        pass

    def check_get_last_region(self):
        pass


class SubdivisionLineTestCase(DataMapperTestCase):
    DataMapper = SubdivisionLineDataMapper

    def check_get_near_points(self):
        # insert a big long run of data
        x = arange(0, pi, pi/1000)
        d = transpose(array([x,sin(x)]))
        dm = self.DataMapper(sort_order='ascending')
        dm._points_per_cell = 50
        dm.set_data(d, 'ascending')
        pt = d[50]
        assert_equal(dm.get_points_near(resize(pt, (1,2))), pt)

def test_suite(level=1):
    suites = []
    if level > 0:
        suites.append(unittest.makeSuite(UtilityTestCase, 'check_'))
        suites.append(unittest.makeSuite(BruteForceTestCase, 'check_'))
        suites.append(unittest.makeSuite(SubdivisionTestCase,'check_'))
        suites.append(unittest.makeSuite(SubdivisionLineTestCase,'check_'))
    total_suite = unittest.TestSuite(suites)
    return total_suite

def test(level=10):
    all_tests = test_suite(level)
    runner = unittest.TextTestRunner()
    runner.run(all_tests)
    return runner

if __name__ == "__main__":
    test()

#EOF
