#-------------------------------------------------------------------------------
#
#  Define the top-level PlotWindow class (which is an actual Tkinter subclass)
#
#  Written by: David C. Morrill
#
#  Date: 10/21/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#  Classes defined: PlotWindow
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import time        ### TEMPORARY ###
import os.path
import plot_base

import Tkinter   as tk
import OpenGL.GL as ogl
import Pmw
import tktree

from enthought.traits.tkmenu import MakeMenu
from enthought.kiva.glcore2d import GraphicsContext, CAP_BUTT
from tkplot_trait_sheet      import TraitSheet, trait_sheet_handler
from plot_base               import *
from plot_container          import HORIZONTAL
from plot_canvas             import PlotValue, Plottable
from plot_window             import PlotWindowAbstract

#-------------------------------------------------------------------------------
#  'PlotWindow' class:
#-------------------------------------------------------------------------------

class PlotWindow ( tk.Widget, PlotWindowAbstract ):

    # Screen scroll increment amount:
    scroll_incr = 20
 
    # Width/Height of standard scrollbars:
    scrollbar_dx = 24  # TBD: Find out how to get scrollbar size
    scrollbar_dy = 24  # TBD: Find out how to get scrollbar size

    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
     
    def __init__ ( self, parent, **traits ):
        tk.Widget.__init__( self, parent, 'togl', { 'double': True } )
        PlotWindowAbstract.__init__( self, **traits )
               
        # Set up the 'paint' and 'resize' event handlers:
        self.bind('<Map>',       self._on_paint )
        self.bind('<Expose>',    self._on_paint )
        self.bind('<Configure>', self._on_paint )
        
        # Set up mouse event handlers:
        self.bind( '<Double-Button-1>', self._on_left_dclick )
        self.bind( '<ButtonPress-1>',   self._on_left_down )
        self.bind( '<ButtonRelease-1>', self._on_left_up )
        self.bind( '<ButtonRelease-3>', self._on_right_click )
        self.bind( '<Motion>',          self._on_mouse_move )
        self.bind( '<Leave>',           self._on_leave_window )

    #---------------------------------------------------------------------------
    #  'Togl OpenGL' helper methods:
    #---------------------------------------------------------------------------
     
    def showoverlay ( self ):
        self.tk.call( self._w, 'showoverlay' )
 
    def hideoverlay ( self ):
        self.tk.call( self._w, 'hideoverlay' )
 
    def existsoverlay ( self ):
        return self.tk.getboolean( self.tk.call( self._w, 'existsoverlay' ) )
 
    def getoverlaytransparentvalue ( self ):
        return self.tk.getint( 
                    self.tk.call( self._w, 'getoverlaytransparentvalue' ) )
 
    def ismappedoverlay ( self ):
        return self.tk.getboolean( self.tk.call( self._w, 'ismappedoverlay' ) )
 
    def alloccoloroverlay ( self, red, green, blue ):
        return self.tk.getint( 
               self.tk.call( self._w, 'alloccoloroverlay', red, green, blue ) )
 
    def freecoloroverlay ( self, index ):
        self.tk.call( self._w, 'freecoloroverlay', index )
    
    #---------------------------------------------------------------------------
    #  Capture the mouse:
    #---------------------------------------------------------------------------
            
    def capture_mouse ( self ):
        self.grab_set()    
   
    #---------------------------------------------------------------------------
    #  Release the mouse capture:
    #---------------------------------------------------------------------------
            
    def release_mouse ( self ):
        self.grab_release()
    
    #---------------------------------------------------------------------------
    #  Return the position (x,y) of the mouse relative to the window:
    #---------------------------------------------------------------------------
    
    def position_mouse ( self ):
        return ( self.winfo_pointerx() - self.winfo_rootx(),
                 self.winfo_pointery() - self.winfo_rooty() )
           
    #---------------------------------------------------------------------------
    #  Return a translated version of an event's (x,y) location:
    #---------------------------------------------------------------------------
 
    def event_xy ( self, x, y = None ):
        if y is None:
            y = x.y
            x = x.x
        # TBD: dx, dy = self.GetViewStart()
        dx, dy = 0, 0
        return ( x + dx * self.scroll_incr,
                 self._flip_y( y + dy * self.scroll_incr ) )

    #---------------------------------------------------------------------------
    #  Force a complete update:
    #---------------------------------------------------------------------------
 
    def update ( self ):
        PlotWindowAbstract.update( self )
              
    #---------------------------------------------------------------------------
    #  Force the contents of the window to be redrawn;
    #---------------------------------------------------------------------------
 
    def force_update ( self ):
        self._on_paint()
 
    #---------------------------------------------------------------------------
    #  Create a scrollbar that overlays the window:
    #---------------------------------------------------------------------------
       
    def make_scrollbar ( self, style ):
        sb = tk.Scrollbar( self, 
               orient  = ( tk.VERTICAL, tk.HORIZONTAL ) [ style == HORIZONTAL ] )
        TKScrollHelper( sb )
        return sb
 
    #---------------------------------------------------------------------------
    #  Destroy a scrollbar that overlays the window:
    #---------------------------------------------------------------------------
       
    def destroy_scrollbar ( self, sb ):
        sb.configure( command = None )
        sb.destroy()
 
    #---------------------------------------------------------------------------
    #  Initialize a user provided scrollbar for a specified 'plottable' object:
    #---------------------------------------------------------------------------
 
    def init_scrollbar ( self, sb ):
        TKScrollHelper( sb )
 
    #---------------------------------------------------------------------------
    #  Set the thumbbar position and size and the scrollbar range and page size:
    #---------------------------------------------------------------------------
    
    def set_scrollbar ( self, sb, position, thumb_size, range, page_size ):
        sb._range      = range
        sb._page_size  = page_size
        sb._thumb_size = thumb_size
        sb.set( float( position ) / range,
                float( position + thumb_size ) / range )
        
    #---------------------------------------------------------------------------
    #  Get the current position of a specified scrollbar:
    #---------------------------------------------------------------------------
       
    def positionof_scrollbar ( self, sb, range ):
        return int( sb.get()[0] * range )
        
    #---------------------------------------------------------------------------
    #  Update scrollbar related information when the size of the canvas changes:
    #---------------------------------------------------------------------------
    
    def update_scrollbars ( self, dx, dy ):
        pass
        # TBD: scroll_incr = self.scroll_incr
        # TBD: self.SetScrollbars( scroll_incr, scroll_incr,
        # TBD:                     (dx + scroll_incr - 1) / scroll_incr,
        # TBD:                     (dy + scroll_incr - 1) / scroll_incr )
    
    #---------------------------------------------------------------------------
    #  Start the timer running at a specified interval:
    #---------------------------------------------------------------------------
 
    def start_timer ( self, interval ):
        self._timer = self.after( interval, function = self.on_timer )
   
    #---------------------------------------------------------------------------
    #  Cancel any pending timer:
    #---------------------------------------------------------------------------
    
    def cancel_timer ( self ):
        if self._timer is not None:
            self.after_cancel( self._timer )
            self._timer = None
    
    #---------------------------------------------------------------------------
    #  Create and display a pop-up menu:
    #---------------------------------------------------------------------------
    
    def popup_menu ( self, menu, object, event ):
        x, y = self._screen_xy( event )
        MakeMenu( menu, object, True, self ).menu.tk_popup( x - 10, y - 10 )
    
    #---------------------------------------------------------------------------
    #  Create and display a pop-up menu to allow editing a list of objects:
    #---------------------------------------------------------------------------
    
    def popup_edit_menu ( self, objects, event ):
        popup = tk.Menu( tearoff = False )
        for object in objects:
            popup.add_command( label   = 'Edit %s' % self._get_label( object ),
                               command = object.menu_edit )
        x, y = self._screen_xy( event )
        popup.tk_popup( x - 10, y - 10 )
   
    #---------------------------------------------------------------------------
    #  Create and display a pop-up menu that merges the menus for a list of
    #  objects:
    #---------------------------------------------------------------------------
    
    def popup_merged_menu ( self, objects, event ):
        submenus = 0
        for object in objects:
            if object.menu != '':
                menu_object = object
                submenus   += 1
        
        if submenus == 0:
            return
 
        if submenus == 1:
            self.popup_menu( menu_object.menu, menu_object, event )
            return
       
        # Build a menu with submenus determined by each object's menu:
        popup = tk.Menu( tearoff = False )
        for object in objects:
            if object.menu != '':
                popup.add_cascade( label = self._get_label( object ), 
                                   menu  = MakeMenu( object.menu, object, True, 
                                                     popup ).menu )
        x, y = self._screen_xy( event )
        popup.tk_popup( x - 10, y - 10 )
       
    #---------------------------------------------------------------------------
    #  Set the cursor to either the standard pointer or a + (cross) shape:
    #---------------------------------------------------------------------------
    
    def set_cursor ( self, is_cross ):
        self.configure( cursor = ( 'arrow', 'crosshair' )[ is_cross ] )
    
    #---------------------------------------------------------------------------
    #  Set up the clipping region for a specified device context:
    #---------------------------------------------------------------------------
    
    def set_clipping_region ( self, gc, x, y, dx, dy ):
        gc.device_set_clipping_path( x, self._flip_y( y + dy - 1 ), dx, dy )
        
    #---------------------------------------------------------------------------
    #  Set up the drawing pen for a specified device context:
    #---------------------------------------------------------------------------
    
    def set_pen ( self, gc, color, weight, style ):
        gc.set_stroke_color( color )
        gc.set_line_width( weight )
        gc.set_line_dash( style ) 
  
    #---------------------------------------------------------------------------
    #  Draw a line using the current pen in the specified device context:
    #---------------------------------------------------------------------------
          
    def draw_line ( self, gc, x1, y1, x2, y2 ):
        self._pre_draw()
        gc.begin_path()
        gc.lines( [ ( x1 - 1, self._flip_y( y1 ) ), 
                    ( x2 - 1, self._flip_y( y2 ) ) ] )
        gc.stroke_path()
        self._post_draw()
    
    #---------------------------------------------------------------------------
    #  Raise a window to the foreground:
    #---------------------------------------------------------------------------
    
    def raise_window ( self, window ):
        window.tkraise()
 
    #---------------------------------------------------------------------------
    #  Set the position and size of a specified window:
    #---------------------------------------------------------------------------
    
    def position_window ( self, window, x, y, dx, dy ):
        window.place( x = x, y = y, width = dx, height = dy )
   
    #---------------------------------------------------------------------------
    #  Create a trait editor dialog for a specified object:
    #---------------------------------------------------------------------------
    
    def trait_dialog ( self, object ):
        return PlotTraitSheet( self, object )
    
    #---------------------------------------------------------------------------
    #  Create a trait sheet for a specified object:
    #---------------------------------------------------------------------------
    
    def trait_sheet ( self, parent, object ):
        return TraitSheet( parent, object )
   
    #---------------------------------------------------------------------------
    #  Determine whether a specified rectangle. specified in Chaco coordinates,
    #  is completely on screen or not:
    #---------------------------------------------------------------------------
 
    def on_screen ( self, x, y, dx, dy ):
        return True
        # TBD: sx, sy   = self.GetViewStart()
        # TBD: sx      *= self.scroll_incr
        # TBD: sy      *= self.scroll_incr
        # TBD: sdx, sdy = self.GetClientSizeTuple()
        # TBD: y        = self._flip_y( y )
        # TBD: return ((x >= sx) and ((x + dx) <= (sx + sdx)) and
        # TBD:         ((y - dy) > sy) and (y < (sy + sdy)))
 
    #---------------------------------------------------------------------------
    #  Get a graphics context to use for drawing:
    #---------------------------------------------------------------------------
 
    def get_gc ( self, dc = None, dy = 100 ):
        if self._layout_size is not None:
            gc = GraphicsContext( size = self._layout_size )
        else:
            gc = GraphicsContext()
        gc.set_line_cap( CAP_BUTT )
        return gc
 
    #---------------------------------------------------------------------------
    #  Get a device context to use for drawing:
    #---------------------------------------------------------------------------
 
    def get_dc ( self ):
        return self.get_gc()
    
    #---------------------------------------------------------------------------
    #  Redraw the screen with an 'inverted' rectangle:
    #---------------------------------------------------------------------------
 
    def _draw_inverted_rect ( self, x, y, dx, dy ):
        # HACK: I don't know why the x coordinate is off by 2, but this
        # adjustment is necessary for the OpenGL version of the code:
        x = max( 0, x - 2 )
        ogl.glEnable( ogl.GL_COLOR_LOGIC_OP )
        ogl.glLogicOp( ogl.GL_XOR )
        ogl.glPolygonMode( ogl.GL_FRONT_AND_BACK, ogl.GL_FILL )
        ogl.glBegin( ogl.GL_POLYGON )
        ogl.glColor( 1.0, 1.0, 1.0, 1.0 )
        ogl.glVertex( x, y, 0.0 )
        ogl.glVertex( x, y + dy, 0.0 )
        ogl.glVertex( x + dx, y + dy, 0.0 )
        ogl.glVertex( x + dx, y, 0.0 )
        ogl.glEnd()
        ogl.glDisable( ogl.GL_COLOR_LOGIC_OP )
     
    #---------------------------------------------------------------------------
    #  Update a drag rectangle:
    #---------------------------------------------------------------------------
 
    def update_drag_rect ( self, dc, rx, ry, rdx, rdy, x0, xc, xn, use_x ):
        self.draw_drag_rect( dc, rx, ry, rdx, rdy, x0, xc, xn, use_x )
     
    #---------------------------------------------------------------------------
    #  Draw a drag rectangle:
    #---------------------------------------------------------------------------
 
    def draw_drag_rect ( self, dc, rx, ry, rdx, rdy, x0, xc, xn, use_x ):
        self._pre_draw()
        if xc is None:
            xc = x0
            if xc > xn:
                xc, xn = xn, xc
        else:
            if xc > xn:
                xc, xn = xn, xc
            if x0 <= xc:
                xc += 1
            elif x0 >= xn:
                xn -= 1
        if use_x:
            self._draw_inverted_rect( xc, ry, xn - xc + 1, rdy )
        else:
            self._draw_inverted_rect( rx, xc, rdx, xn - xc + 1 )
        self._post_draw()
   
    #---------------------------------------------------------------------------
    #  Repaint a section of the window from the backing image:
    #---------------------------------------------------------------------------
    
    def repaint_rect ( self, dc, x, y, dx, dy ):
        if dx is None:
            dx, dy = self._layout_size
 
        # HACK: I don't know why the x coordinate is off by 2, but this
        # adjustment is necessary for the OpenGL version of the code:
        x = max( 0, x - 2 )
        y = self._flip_y( y + dy - 1 )
        self._pre_draw()
        ogl.glRasterPos( x, y )
        ogl.glCopyPixels( x, y, dx, dy, ogl.GL_COLOR )
        self._post_draw()
    
    #---------------------------------------------------------------------------
    #  Invert a specified section of the window:
    #---------------------------------------------------------------------------
    
    def invert_rect ( self, dc, x, y, dx, dy ):
        self._pre_draw()
        self._draw_inverted_rect( x, self._flip_y( y + dy - 1 ), dx, dy )
        self._post_draw()

    #---------------------------------------------------------------------------
    #  Handle a scroll event for a scaled plottable:
    #---------------------------------------------------------------------------
 
    def _on_scroll ( self, sb, action, value, units ):
        plottable = sb._plottable
        range     = sb._range
        delta     = sb._page_size
        maxpos    = range - delta
        if action == 'moveto':
            position = float( value ) * range
        else:
            if units == 'units':
                delta = max( 1, delta / 20 )
            position = max( 0, min( maxpos, 
                            sb.get()[0] * range + delta * int( value ) ) )
        if sb.cget( 'orient' ) == tk.HORIZONTAL:
            if position == plottable._plot_ox:
                return
            plottable._plot_ox = position
        else:
            newpos = maxpos - position
            if newpos == plottable._plot_oy:
                return
            plottable._plot_oy = newpos
        sb.set( float( position ) / range,
                float( position + sb._thumb_size ) / range )
        self.draw()
   
    #---------------------------------------------------------------------------
    #  Handle a timer event:
    #---------------------------------------------------------------------------
 
    def on_timer ( self, event ):
        pass
        # TBD: x, y   = wx.wxGetMousePosition()
        # TBD: x, y   = self.ScreenToClientXY( x, y )
        # TBD: dx, dy = self.GetClientSizeTuple()
        # TBD: vx, vy = self.GetViewStart()
        # TBD: use_x  = self._select_info[6]
        # TBD: if use_x:
        # TBD:    if x < 0:
        # TBD:       self.Scroll( vx - 1, -1 )
        # TBD:    elif x >= dx:
        # TBD:       self.Scroll( vx + 1, -1 )
        # TBD: else:
        # TBD:    if y < 0:
        # TBD:       self.Scroll( -1, vy - 1 )
        # TBD:    elif y >= dy:
        # TBD:       self.Scroll( -1, vy + 1 )
 
    #---------------------------------------------------------------------------
    #  Paint event handler:
    #---------------------------------------------------------------------------
 
    def _on_paint ( self, event = None ):
        begin    = time.time()
        test_msg = self._min_size()
        pdx      = max( self.winfo_width(),  self.mdx )
        pdy      = max( self.winfo_height(), self.mdy )
 
        self.tk.call( self._w, 'makecurrent' )
        mode = ogl.glGetDoublev( ogl.GL_MATRIX_MODE )
        ogl.glMatrixMode( ogl.GL_PROJECTION )
        ogl.glPushMatrix()
        ogl.glDrawBuffer( ogl.GL_BACK )
        self._layout_needed |= (( pdx, pdy ) != self._layout_size)
        if self._draw_needed or self._layout_needed:
            ogl.glClearColor( *self.bg_color_ )
            ogl.glClear( ogl.GL_COLOR_BUFFER_BIT )
            test_msg += self._layout_draw( None, 0, 0, pdx, pdy )
 
        self.tk.call( self._w, 'swapbuffers' )
          
        self._draw_done()
        
        ogl.glFlush()
        ogl.glMatrixMode( ogl.GL_PROJECTION )
        ogl.glPopMatrix()
        ogl.glMatrixMode( mode )
           
        if plot_base.TIMING:
            print '%sTotal:    %.3f seconds\n' % ( test_msg, 
                                                   time.time() - begin )

    #---------------------------------------------------------------------------
    #  Do first part of a screen 'overdraw' operation:
    #---------------------------------------------------------------------------
    
    def _pre_draw ( self ):
        self.tk.call( self._w, 'makecurrent' )
        self._mode = ogl.glGetDoublev( ogl.GL_MATRIX_MODE )
        ogl.glMatrixMode( ogl.GL_PROJECTION )
        ogl.glPushMatrix()
        ogl.glDrawBuffer( ogl.GL_FRONT )
        ogl.glReadBuffer( ogl.GL_BACK )
 
    #---------------------------------------------------------------------------
    #  Do rest of a screen 'overdraw' operation:
    #---------------------------------------------------------------------------
           
    def _post_draw ( self ):
        ogl.glFlush()
        ogl.glMatrixMode( ogl.GL_PROJECTION )
        ogl.glPopMatrix()
        ogl.glMatrixMode( self._mode )
        
    #---------------------------------------------------------------------------
    #  Return the absolute screen coordinates of a specified event:
    #---------------------------------------------------------------------------
     
    def _screen_xy ( self, event ):
        return ( event.x + self.winfo_rootx(), 
                 event.y + self.winfo_rooty() ) 

#-------------------------------------------------------------------------------
#  'TKScrollHelper' class:
#-------------------------------------------------------------------------------

class TKScrollHelper:
    
    #---------------------------------------------------------------------------
    #  Initialize the object:
    #---------------------------------------------------------------------------
    
    def __init__ ( self, sb ):
        self.sb = sb
        sb.configure( command = self._on_scroll )
        
    #---------------------------------------------------------------------------
    #  Handle a scroll event:
    #---------------------------------------------------------------------------
    
    def _on_scroll ( self, action, value, units = None ):
        self.sb._plottable.window._on_scroll( self.sb, action, value, units )
       
#-------------------------------------------------------------------------------
#  'PlotTraitSheet' class:
#-------------------------------------------------------------------------------
          
class PlotTraitSheet ( tk.Toplevel ):

    #---------------------------------------------------------------------------
    #  Initialize object: 
    #---------------------------------------------------------------------------
 
    def __init__ ( self, owner, object = None ):
        tk.Toplevel.__init__( self, owner ) 
        self.title( 'Traits' )
        self.owner  = owner
        self.root   = root = owner.canvas
        self.ignore = False
        
        # Set inital width of tree view pane:
        self.tree_size = 200
        
        # Indicate no trait sheet has been loaded yet:
        self.trait_sheet = None
        
        # Initialize the path to the icon images:
        import tkplot_window
        self.image_path = os.path.join( 
                             os.path.dirname( tkplot_window.__file__ ),
                             'images' )
 
        # Initialize tree item icon info:
        self.image_map = {}
        
        # Initialize the contents of the trait sheet:
        self.pane = pane = Pmw.PanedWidget( self,
                                            orient  = tk.HORIZONTAL,
                                            command = self.on_resize )
        pane.add( 'tree', size = self.tree_size )
        pane.add( 'traits' )
        pane.grid( row = 0, col = 0, sticky = 'nesw' )
        self.columnconfigure( 0, weight = 1 )
        self.rowconfigure(    0, weight = 1 )
    
        # Initialize window contents:
        self.tree = tree = pane.pane( 'tree' )
        self.traits  = pane.pane( 'traits' )
        
        # Create the tree hierarchy control:
        self.next_id = 0
        self.objects = objects = tktree.Tree( 
                          tree,
                          root_id     = '0',
                          root_label  = '',
                          bg          = '#FFFFFF',
                          dist_x      = 28, 
                          dist_y      = 33,
                          text_offset = 20,
                          get_contents_callback = self.expand_node,
                          select_callback       = self.on_select )
        objects.grid( row = 0, column = 0, sticky = 'nesw' )
        tree.rowconfigure(    0, weight = 1 )
        tree.columnconfigure( 0, weight = 1 )
 
        # Add scrollbars:
        sb = tk.Scrollbar( tree, orient = tk.VERTICAL )
        sb.grid( row = 0, column = 1, sticky= 'ns' )
        objects.configure( yscrollcommand = sb.set )
        sb.configure( command = objects.yview )
 
        sb = tk.Scrollbar( tree, orient = tk.HORIZONTAL )
        sb.grid( row = 1, column = 0, sticky = 'ew' )
        objects.configure( xscrollcommand = sb.set )
        sb.configure( command = objects.xview )
        
        # Initialize event handlers:
        self.bind( '<Destroy>', self.on_close )
        
        # Construct the object hierarchy:
        self.refresh()
        
        # Select the specified object (or root object if no object specified):
        self.object = None
        self.select( object or root )
        
        # Make sure the window cannot be resized by the user:
        self.resizable( width = False, height = FALSE )
        
    #---------------------------------------------------------------------------
    #  Handle a tree node being expanded:
    #---------------------------------------------------------------------------
    
    def expand_node ( self, node ):
        object = node.object
        try:
            if not isinstance( object, PlotValue ):
                for value in object.values:
                    self.add_object( node, value )
            for title in object.titles:
                self.add_object( node, title )
            for overlay in object.overlays:
                self.add_object( node, overlay )
            container = object.container
            axis      = object.axis
            if axis is not container.axis:
                self.add_object( node, axis )
            axis = object.axis_index
            if axis is not container.axis_index:
                self.add_object( node, axis )
        except:
           pass
 
    #---------------------------------------------------------------------------
    #  Handle the user closing the editor window:
    #---------------------------------------------------------------------------
 
    def on_close ( self, event ):
        if not self.ignore:
            self.ignore = True
            self.owner._trait_sheet_for()
       
    #---------------------------------------------------------------------------
    #  Handle the user selecting an object in the hierarchy:
    #---------------------------------------------------------------------------
    
    def on_select ( self, objects, node ):
        self.select( node.object )
 
    #---------------------------------------------------------------------------
    #  Handle the user moving the splitter bar:
    #---------------------------------------------------------------------------
    
    def on_resize ( self, sizes ):
        tree_size = sizes[0]
        if tree_size != self.tree_size:
            self.tree_size = tree_size
            self.set_size()
           
    #---------------------------------------------------------------------------
    #  Set the correct size for the window:
    #---------------------------------------------------------------------------
    
    def set_size ( self ):
        self.geometry( '=%dx%d' % ( 
             self.traits.winfo_reqwidth() + self.tree_size, 
             self.traits.winfo_reqheight() ) )
       
    #---------------------------------------------------------------------------
    #  Refresh the contents of the object hierarchy view:
    #---------------------------------------------------------------------------
    
    def refresh ( self ):
        root = self.owner.canvas
        if (root is not None) and (self.root is not root):
            self.root   = root
            self.object = None
        root_node = self.objects.root
        for node in root_node.children():
            node.delete()
        self.add_object( root_node, root ).expand()
        self.add_object( root_node, root.axis )
        self.add_object( root_node, root.axis_index )
        
    #---------------------------------------------------------------------------
    #  Select a new object in the object hierarchy view:
    #---------------------------------------------------------------------------
    
    def select ( self, object ):
        cur_object = self.object
        if object is cur_object:
            return
        if cur_object is not None:
            cur_object.selected = False
        self.object = object
        if self.trait_sheet is not None:
            self.ignore = True
            self.trait_sheet.destroy()
            self.trait_sheet = None
            self.ignore = False
        if object is not None:
            frame            = self.traits
            self.trait_sheet = ps = TraitSheet( frame, object, 
                                                object._editable_traits(),
                                                trait_sheet_handler )
            ps.grid( row = 0, column = 0, sticky = 'nesw' )
            frame.rowconfigure(    0, weight = 1 )
            frame.columnconfigure( 0, weight = 1 )
            stack = []
            item  = object
            top   = self.root.container
            while item is not top:
                stack[0:0] = [ item ]
                item       = item.container
            node = self.objects.root
            for item in stack:
                for cur_node in node.children():
                    if cur_node.object is item:
                        node = cur_node
                        if item is not object:
                            node.expand()
                        break
            self.update()
            self.set_size()
            self.objects.move_cursor( node )
        self.owner.select( object )
    
    #---------------------------------------------------------------------------
    #  Handle an object being added to an object in the hierarchy view:
    #---------------------------------------------------------------------------
    
    def add_item ( self, container, object ):
        self.add_object( container._tree_node, object )
        self.select( object )
    
    #---------------------------------------------------------------------------
    #  Handle an object being removed from an object in the hierarchy view:
    #---------------------------------------------------------------------------
    
    def remove_item ( self, container, object ):
        if object is self.object:
            self.select( None )
        object._tree_node.delete()
        object._tree_node = None
        
    #---------------------------------------------------------------------------
    #  Add a new object to the tree:
    #---------------------------------------------------------------------------
 
    def add_object ( self, node, object ):
        if object is None:
            return None
    
        image_name = object.__class__.__name__
        try:
            icon = self.image_map[ image_name ]
        except:
            icon = self.load_image( image_name )
        
        name          = object._ui_label()
        self.next_id += 1
        id            = str( self.next_id )
        new_node = node.append_children(
                      self.objects.add_list(
                           name           = name, 
                           id             = id, 
                           flag           = isinstance( object, Plottable ),
                           collapsed_icon = icon,
                           expanded_icon  = icon ) )[0]
        new_node.object   = object
        object._tree_node = new_node
        
        return new_node
 
    #---------------------------------------------------------------------------
    #  Load a requested image into the image list:
    #---------------------------------------------------------------------------
 
    def load_image ( self, image_name ):
        filename = os.path.join( self.image_path, 'icon_%s.gif' % image_name )
        self.image_map[ image_name ] = image = tk.PhotoImage( file = filename )
        return image

