#-------------------------------------------------------------------------------
#
#  Interactive Chaco plotting classes for use with Tkinter
#
#  Written by: David C. Morrill
#
#  Date: 09/11/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import Tkinter      as tk
import OpenGL.Tk    as tkgl
import tkFileDialog as tkfd
import tkMessageBox as tkmb

from tkplot                  import *
from plt                     import *
from enthought.traits.tkmenu import MakeMenu

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Screen size:
screen_dx = tkgl._default_root.winfo_screenwidth()
screen_dy = tkgl._default_root.winfo_screenheight()

#-------------------------------------------------------------------------------
#  'tkShellPlotFrame' class:
#-------------------------------------------------------------------------------
     
class tkShellPlotFrame ( tk.Toplevel, ShellPlotFrame ):

    # Name of file to save session state into:
    state_file = '.tkplt.save'
 
    #---------------------------------------------------------------------------
    #  Initialize object: 
    #---------------------------------------------------------------------------
 
    def __init__ ( self, title ):
        # Initialize the Tkinter part of the window:
        tk.Toplevel.__init__( self )
        
        # Now initialize the Chaco part of the window:
        ShellPlotFrame.__init__( self, title )
 
        # Set up the window title:
        self.title( app_title % title )
 
        # Initialize event handlers and layout:
        container = self.container
        container.grid( row = 1, col = 0, sticky = 'nesw' )
        self.rowconfigure(    1, weight = 1 )
        self.columnconfigure( 0, weight = 1 )
        container.bind( '<Control-ButtonRelease-1>', 
                        container.on_control_left_up )
        container.bind( '<Shift-ButtonRelease-1>',   
                        container.on_control_left_up )
        container.bind( '<Alt-ButtonRelease-1>',     
                        container.on_control_left_up )
        self.bind( '<Destroy>', self.on_close )
        
        # Make sure we are visible:
        self.tkraise()
     
    #---------------------------------------------------------------------------
    #  Get the current window geometry:
    #---------------------------------------------------------------------------
    
    def get_window_geometry ( self ):
        return ( self.winfo_width(), self.winfo_height(),
                 self.winfo_x(),     self.winfo_y() )
    
    #---------------------------------------------------------------------------
    #  Set the window geometry:
    #---------------------------------------------------------------------------
    
    def set_window_geometry ( self, geometry ):
        if geometry is None:
           geometry = ( screen_dx / 4, screen_dy / 4, 
                        screen_dx / 2, screen_dy / 2 )
        self.geometry( '=%dx%d+%d+%d' % geometry ) 
    
    #---------------------------------------------------------------------------
    #  Return whether or not the window is still 'alive':
    #---------------------------------------------------------------------------
    
    def is_alive ( self ):
        return True
 
    #---------------------------------------------------------------------------
    #  Make this window the topmost visible window:
    #---------------------------------------------------------------------------
    
    def raise_window ( self ):
        self.tkraise()
    
    #---------------------------------------------------------------------------
    #  Close the window:
    #---------------------------------------------------------------------------
    
    def close_window ( self ):
        self.destroy()
 
    #---------------------------------------------------------------------------
    #  Destroy the window:
    #---------------------------------------------------------------------------
    
    def destroy_window ( self ):
        self.destroy()
       
    #---------------------------------------------------------------------------
    #  Show an error message:
    #---------------------------------------------------------------------------
    
    def show_error ( self, msg ):
        tkmb.showerror( app_title + ' Error', msg ) 
    
    #---------------------------------------------------------------------------
    #  Create a specified menu bar for the window:
    #---------------------------------------------------------------------------
    
    def make_menu ( self, menu_desc ):
        MakeMenu( menu_desc, self ).menubar.grid( row = 0, sticky = 'new' )
        
    #---------------------------------------------------------------------------
    #  Make a Chaco PlotWindow of the appropriate sub-class type:
    #---------------------------------------------------------------------------
    
    def make_plot_container ( self, **traits ):
        window = ShellPlotWindow( self, **traits )
        window.set( canvas_menu  = default_menu,
                    overlay_menu = default_menu,
                    title_menu   = default_menu )
        return window
        
    #---------------------------------------------------------------------------
    #  Get the name of a file to save the window contents to:
    #---------------------------------------------------------------------------
    
    def get_save_filename ( self ):
        filetypes = [ ( "PNG file",  "png" ), 
                      ( "JPEG file"  "jpg" ),
                      ( "PCX file",  "pcx" ),
                      ( "TIFF file", "tif" ) ]
        if self._has_pdf:
            filetypes.append( ( "PDF file", "pdf" ) )
        return tkfd.askopenfilename( filetypes = filetypes )

#-------------------------------------------------------------------------------
#  'ShellPlotWindow' class:
#-------------------------------------------------------------------------------

class ShellPlotWindow ( PlotWindow ):

    #---------------------------------------------------------------------------
    #  Handle a left mouse down event:
    #---------------------------------------------------------------------------
 
    def on_left_down ( self, event ):
        self._x, self._y = self.event_xy( event )
        self._is_control = False

    #---------------------------------------------------------------------------
    #  Handle a left mouse up event by editing the topmost object under the
    #  mouse pointer:
    #---------------------------------------------------------------------------
 
    def on_left_up ( self, event ):
        global selected_item
 
        if self._x is None:
            return
           
        if self._is_control:
            selected_item = self._x = None
            self.select( None )
            return
           
        objects = self.objects_at( event )
        stack   = []
        if (len( objects ) > 0) and isinstance( objects[0], PlotOverlay ):
            stack   = [ objects[0] ]
            objects = objects[1:]
        stack.extend( self.values_at( event ) )
        stack.extend( objects )
        self._previous_selected = None
        i = 0
        if stack[0] == self._last_top:
            try:
                i = stack.index( self._selected ) + 1
                self._previous_selected = self._selected
            except:
                pass
        try:
            self._last_top = stack[0]
            selected_item  = stack[i]
        except:
            self._last_top = selected_item = None
        self.select( selected_item )

    #---------------------------------------------------------------------------
    #  Handle a double-click event by editing the currently selected object:
    #---------------------------------------------------------------------------
        
    def on_left_dclick ( self, event ):
        object = self._previous_selected or self._selected
        x, y   = self.event_xy( event )
        if ((object is not None) and
            ((abs( self._x - x ) + abs( self._y - y )) <= 4)):
            try:
                self._previous_selected = None
                object.edit()
            except:
                pass
        self._x = None
        
    def on_control_left_up ( self, event ):
        self._is_control = True
        self._on_left_up( event )
       
#-------------------------------------------------------------------------------
#  Make sure the factory for the main 'plt' module uses our class:
#-------------------------------------------------------------------------------

import plt
plt.ShellPlotFrame_factory = tkShellPlotFrame   
