#-------------------------------------------------------------------------------
#
#  Interactive Chaco plotting classes for use with wxPython
#
#  Written by: David C. Morrill
#
#  Date: 09/11/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

__all__ = ["redraw", "autoscale", "xaxis", "yaxis", "title", "xtitle",
           "ytitle", "grid", "hold", "axis", "save", "stem", "plot",
           "markers", "figure" ]

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import os.path 
import wx

from enthought.chaco.plt         import *
from enthought.enable            import Box
from enthought.enable.wx         import Window
from enthought.pyface.api            import SystemMetrics
from enthought.traits.api            import Trait
from enthought.traits.ui.wx.menu import MakeMenu

# Local imports.
from plot import *

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

system_metrics = SystemMetrics()

#-------------------------------------------------------------------------------
#  'wxProxyFrame' class:
#-------------------------------------------------------------------------------

class wxProxyFrame ( wx.Frame ):
    """
    This is a thin wrapper around wx.Frame whose sole purpose is to re-route
    menu event handlers from wx back up to wxShellPlotFrame.  Since
    wxShellPlotFrame cannot inherit from both ShellPlotFrame (which defines
    various menu function handlers) and wx.Frame (which is what MakeMenu()
    requires to attach its menus to), this class serves as a go-between.
    
    The only methods that need to be proxied here are ones that menu item
    handlers are likely to call.
    """
    
    proxyAttribList = ("create_file", "save_file", "set_image_size", "on_exit",
                       "on_close", "open_state", "save_state", "get_state",
                       "restore_state", "reset_scaling", "restore_scaling",
                       "pdf_end_page", "draw_page", "save", "next_line_style",
                       "reset_line_style", "auto_scale", "update_window",
                       "clear_canvas", "reset_undo")
    
    def __init__ ( self, parentPlotFrame, parentWxFrame, wid, title, pos, 
                         size ):
        wx.Frame.__init__( self, parentWxFrame, wid, title, pos, size )
        self.parent = parentPlotFrame
        wx.EVT_CLOSE( self, self.on_exit )
    
#    def __getattr__(self, name):
#        if name in self.proxyAttribList:
#            return self.parent.__getattr__(name)
    
    def create_file ( self, object, mode ):
        return self.parent.create_file(object, mode)
        
    def save_file ( self, *args, **kwargs ):
        return self.parent.save_file( *args, **kwargs )

    def set_image_size ( self, *args, **kwargs ):
        return self.parent.set_image_size ( *args, **kwargs )

    def on_exit ( self, event ):
        return self.parent.on_exit()

#-------------------------------------------------------------------------------
#  'wxShellPlotFrame' class:
#-------------------------------------------------------------------------------
     
class wxShellPlotFrame ( ShellPlotFrame ):

    # Name of file to save session state into:
    state_file = '.wxplt.save'
    delegate   = Trait( None, wxProxyFrame )
 
    #---------------------------------------------------------------------------
    #  Initialize object: 
    #---------------------------------------------------------------------------
 
    def __init__ ( self, title ):
        # Initialize the wxPython part of the window:
        self.delegate = wxProxyFrame( self, None, -1, app_title % title, 
                                      pos = (-1,-1), size = (-1,-1) )
        ShellPlotFrame.__init__( self, title )
        
        # Initialize event handlers:
        wx.EVT_CLOSE( self.delegate, self.on_close )
        
        # Make sure we are visible:
        self.delegate.Show()

    #---------------------------------------------------------------------------
    #  Get the current window geometry:
    #---------------------------------------------------------------------------
    
    def get_window_geometry ( self ):
        return self.delegate.GetPositionTuple() + self.delegate.GetSizeTuple()
    
    #---------------------------------------------------------------------------
    #  Set the window geometry:
    #---------------------------------------------------------------------------
    
    def set_window_geometry ( self, geometry ):
        screen_dx = system_metrics.screen_width
        screen_dy = system_metrics.screen_height
        if geometry is None:
            x, y, dx, dy = ( screen_dx / 4, screen_dy / 4, 
                             screen_dx / 2, screen_dy / 2 )
        else:
            x, y, dx, dy = geometry
        self.delegate.SetDimensions( x, y, dx, dy )
    
    #---------------------------------------------------------------------------
    #  Return whether or not the window is still 'alive':
    #---------------------------------------------------------------------------
    
    def is_alive ( self ):
        return self.proxy_object_alive

    #---------------------------------------------------------------------------
    #  Make this window the topmost visible window:
    #---------------------------------------------------------------------------
    
    def raise_window ( self ):
        self.delegate.Raise()
    
    #---------------------------------------------------------------------------
    #  Close the window:
    #---------------------------------------------------------------------------
    
    def close_window ( self ):
        self.delegate.Close()
 
    #---------------------------------------------------------------------------
    #  Destroy the window:
    #---------------------------------------------------------------------------
    
    def destroy_window ( self ):
        self.delegate.DestroyChildren()
        self.delegate.Destroy()
        
        # fixme: The following lines are a hack to force garbage collection:
        self.dispose_group( self.container.component )
        self.container.__dict__.clear()
        self._window.__dict__.clear() 
        self._window = self.container = self.delegate = None
        
    #---------------------------------------------------------------------------
    #  Show an error message:
    #---------------------------------------------------------------------------
    
    def show_error ( self, msg ):
        dlg = wx.MessageDialog( self.delegate, str( msg ), app_title + ' Error', 
                                wx.OK )
        dlg.ShowModal()
        dlg.Destroy()
    
    #---------------------------------------------------------------------------
    #  Create a specified menu bar for the window:
    #---------------------------------------------------------------------------
    
    def make_menu ( self, menu_desc ):
        if self.delegate != None:
            MakeMenu( menu_desc, self.delegate )
       
    #---------------------------------------------------------------------------
    #  Make a Chaco PlotWindow of the appropriate sub-class type:
    #---------------------------------------------------------------------------
    
    def make_plot_container ( self, canvas, **traits ):
        self._window = ShellPlotWindow( self.delegate, 
                                        component = PlotComponent( canvas ),
                                        **traits )
        container = self._window.component
        #container.set( canvas_menu  = default_menu,
        #               overlay_menu = default_menu,
        #               title_menu   = default_menu )
        # For zoom functionality with ShellPlotWindow
        #container.on_trait_change( window.on_left_down,   'pre_left_down' )
        #container.on_trait_change( window.on_left_up,     'pre_left_up' )
        #container.on_trait_change( window.on_left_dclick, 'pre_left_dclick' )
        #container.on_trait_change( window.on_right_up,    'pre_right_up' )
        return container
        
    #---------------------------------------------------------------------------
    #  Get the name of a file to save the window contents to:
    #---------------------------------------------------------------------------
    
    def get_save_filename ( self ):
        dlg = wx.FileDialog( self.delegate, 
                               message = 'Specify a file and type to save to',
                               style   = wx.SAVE | wx.OVERWRITE_PROMPT  )
        filters = ( "PNG file (*.png)|*.png|" 
                    "JPEG file (*.jpg)|*.jpg|"
                    "PCX file (*.pcx)|*.pcx|" 
                    "TIFF file (*.tif)|*.tif" )
        if self._has_pdf:
            filters = "PDF file (*.pdf)|*.pdf|" + filters 
        dlg.SetWildcard( filters )
        filename = None
        if dlg.ShowModal() == wx.ID_OK:
            filename = os.path.abspath( dlg.GetPath() )
        dlg.Destroy()
        return filename

    # Stub out missing functionality that ShellPlotFrame expects
    def menu_undo(self):
        pass

#-------------------------------------------------------------------------------
#  'ShellPlotWindow' class:
#-------------------------------------------------------------------------------

class ShellPlotWindow ( Window ):
    """ShellPlotWindow
    
       A zoom capability is now supported on the window.  Selecting an area on
       the plot with the "Shift" key down will result in a zoom to that screen
       area.  Repeating this action will zoom further into the plot.  The stack
       of zooms that have occured is stored.  Right clicking in the plot will
       pop up a menu that allows you to return to the previous zoom or reset to
       the original zoom.
       
       !! There appear to be some redraw issues that are occuring.  
       !!   1. When you zoom in several times, sometimes only a portion of the
       !!      data in the screen is drawn.  I don't know if this is a clipping
       !!      issue or a redraw issue. 
       !!   2. Sometimes -- especially when the window is revealed, the window
       !!      as an L shaped area that isn't redrawn.  
       !!
       !! I haven't figured out how to make this repeatable yet.
    """
    def __init__ ( self, parent, wid = -1, pos = wx.DefaultPosition,
                                 size = wx.DefaultSize, **traits ):
         Window.__init__( self, parent, wid, pos, size, **traits )
         self._zoom_stack = []
         
    #---------------------------------------------------------------------------
    #  Handle a left mouse down event:
    #---------------------------------------------------------------------------
 
    def on_left_down ( self, event ):
        if event.shift_down:
             # Set up drag:
             zoom_rect = Box( color = ( 0.5, 0.5, .5, 0.09 ),
                              border_color = ( 0.5, 0.62, .73, 1.0 ),
                              border_size = 1 ) 
          
             zoom_rect.bounds = ( event.x, event.y, 1, 1 )
             zoom_rect.on_trait_change( self.drag_zoom, 'resized' )
             drag_constraints = self.component._plot_bounds
             self.component.window.drag_resize( zoom_rect, drag_constraints, 
                                                event, overlay = False )
             self.component.overlay = zoom_rect
        else:
             # prepare for selection.
             self._x, self._y = event.x, event.y
               
    def drag_zoom ( self, obj, old, event ):
        """ Handle zooming to a new plot region selected by a zoom operation.
        
            This callback is attached to the zoom_rect in on_left_down() and
            is called when the resizing operation for the zoom_rect finishes.
        """    
        # Get rid of the overlay, and unhook the zoom rectangles events.
        # !! Not sure whether the redraw() or overlay=None causes the
        # !! rectangle to disappear...
        zoom_rect = self.component.overlay
        zoom_rect.on_trait_change(self.drag_zoom, 'resized', remove=True)
        zoom_rect.redraw()
        self.component.overlay = None        
 
        # Map the window coordinates of the zoom_rect through to calculate
        # the coordinates on the graph.
        # !! It seems a little strange to map through a PlotValue.
        x, y, width, height = zoom_rect.bounds
        graph_space = self.component.component.values[0]
        low_x, low_y, intercept = graph_space.map_xy(x,y)
        high_x, high_y, intercept = graph_space.map_xy(x+width,y+height)
 
        # update to the new zoom on the plot.  Keep a stack of previous 
        # zooms to go back to.
        self.save_zoom()              
        self.set_zoom((low_x, high_x, low_y, high_y))
 
    def on_right_up ( self, obj, old, event ):
        """ Handles a request to display the canvas context menu.
        """
        if len( self._zoom_stack ) > 0:
            zoom_menu = """
                        Zoom Previous: self.restore_zoom()
                        Zoom Reset: self.reset_zoom()
                        """
        else:
            zoom_menu = "Zoom Reset: self.reset_zoom()"
        
        event.handled = True
        menu = self.window.create_menu(zoom_menu, self)
        self.window.popup_menu(menu, event.x - 10, event.y + 10)

    #----------------------------------------------------------------------------
    #  Zoom management routines.
    #----------------------------------------------------------------------------
    
    def save_zoom ( self ):
        """ Save the current axis zoom settings on the zoom stack.
        """
        plot_canvas = self.component.component
        current_axis_bounds = (plot_canvas.axis_index.bound_low,
                               plot_canvas.axis_index.bound_high,
                               plot_canvas.axis.bound_low,
                               plot_canvas.axis.bound_high)
        self._zoom_stack.append( current_axis_bounds )
                                   
    def restore_zoom ( self ):
        """ Restore the previous zoom settings to the plot.
        """
        try:
            values = self._zoom_stack.pop()
            self.set_zoom(values)
        except IndexError:
            pass
                
    def set_zoom ( self, axis_values ):
        """ Set the plot axis end points to the given values.
        
            axis_values is a 4-sequence of (low, high, low, high) for
            the index and value axes respectively.
        """
        ia_low, ia_high, a_low, a_high = axis_values
        plot_canvas = self.component.component
 
        plot_canvas.axis_index.bound_low = ia_low
        plot_canvas.axis_index.bound_high = ia_high
        plot_canvas.axis.bound_low = a_low
        plot_canvas.axis.bound_high = a_high
        plot_canvas.update()
    
    def reset_zoom ( self ):
        """ Reset the plot zoom to its original settings and clear the 
            _zoom_stack.
        """
        try:
            values = self._zoom_stack[0]
            self.set_zoom(values)
            self._zoom_stack = []
        except IndexError:
            pass
                
    #---------------------------------------------------------------------------
    #  Handle a left mouse up event by selecting the topmost object under the
    #  mouse pointer:
    #---------------------------------------------------------------------------
 
    def on_left_up ( self, event ):
        global selected_item
 
        if self._x is None:
            return
 
        if event.control_down:
            selected_item = self._x = None
            self.select( None )
            return
           
        objects = self.objects_at( event )
        stack   = []
        if (len( objects ) > 0) and isinstance( objects[0], PlotOverlay ):
            stack   = [ objects[0] ]
            objects = objects[1:]
        stack.extend( self.values_at( event ) )
        stack.extend( objects )
        self._previous_selected = None
        i = 0
        if stack[0] == self._last_top:
            try:
                i = stack.index( self._selected ) + 1
                self._previous_selected = self._selected
            except:
               pass
        try:
            self._last_top = stack[0]
            selected_item  = stack[i]
        except:
            self._last_top = selected_item = None
        if event.shift_down or event.alt_down:
            self._x = None
            return
        self.select( selected_item )

    #---------------------------------------------------------------------------
    #  Handle a double-click event by editing the currently selected object:
    #---------------------------------------------------------------------------
        
    def on_left_dclick ( self, event ):
        pass
        #if self._x is not None:
        #    object = self._previous_selected or self._selected
        #    x, y   = event.x, event.y
        #    if ((object is not None) and
        #        ((abs( self._x - x ) + abs( self._y - y )) <= 4)):
        #        try:
        #            self._previous_selected = None
        #            if self._handler is None:
        #                self._handler = PltTraitSheetHandler( self.component )
        #            object.edit_traits( handler = self._handler )
        #        except:
        #            pass
        #    self._x = None
               
    #---------------------------------------------------------------------------
    #  Return the objects at a specified point:
    #---------------------------------------------------------------------------
           
    def objects_at ( self, event ):
        return self.component.component.objects_at( event.x, event.y )
           
    def values_at ( self, event ):
        return self.component.component.values_at( event.x, event.y )
              
    #---------------------------------------------------------------------------
    #  Select an object: 
    #---------------------------------------------------------------------------
    
    def select ( self, object ):
        cur_object = self._selected
        if object is cur_object:
            return
        if cur_object is not None:
            cur_object.selected = False
        self._selected  = object
        if object is not None:
            object.selected = True
        self.component.redraw()

# ui_hack
##-------------------------------------------------------------------------------
##  'PltTraitSheetHandler' class:
##-------------------------------------------------------------------------------
#
#class PltTraitSheetHandler ( TraitSheetHandler ):
#    
#    #---------------------------------------------------------------------------
#    #  Initialize the object: 
#    #---------------------------------------------------------------------------
#     
#    def __init__ ( self, component ):
#        self.component = component
#    
#    #---------------------------------------------------------------------------
#    #  Notification that a trait sheet has modified a trait of its
#    #  associated object: 
#    #---------------------------------------------------------------------------
#    
#    def changed ( self, object, trait_name, new_value, old_value, is_set ):
#        try:
#            object.update()
#        except:
#            pass
   
#-------------------------------------------------------------------------------
#  Initialize plotting classes:
#-------------------------------------------------------------------------------

wx.InitAllImageHandlers()

# Make sure the factory for the main 'plt' module uses our class:
import enthought.chaco.plt as chacoplt
chacoplt.ShellPlotFrame_factory = wxShellPlotFrame   

