#!/usr/bin/env python
"""
Demonstration of getting plotvalues to render onto a secondary graphics context.
"""

import wx

from enthought.traits.api import Any, Instance
from enthought.enable.wx import Window
from enthought.pyface.api import ApplicationWindow, GUI
from enthought.util.numerix import arange, array, Float, pi, sin, transpose, zeros
import math

from enthought.chaco import PlotAxis, PlotCanvas, PlotValue, PlotGroup, PlotValue
from enthought.chaco.image_plot_value import ImageData, CmapImagePlotValue, ImagePlotValue
from enthought.chaco.backbuffer_plotter import gc_plot
from enthought.chaco.default_colormaps import hsv
from enthought.chaco.colormap_legend import ColormapLegend
from enthought.chaco.renderers.scatter import ScatterRenderer

from demo_base import PlotApplicationWindow


def make_scatter(axis_index, axis):
    # create a scatter plot
    x = arange(-15.0, 15.0, 0.3)
    y = sin(x*8*pi/30.0) * 15.0 + 15.0
    scatter_data = transpose(array([x,y]))
    scatter_pv = PlotValue(scatter_data, axis_index=axis_index, axis=axis,
                           fill_color = (0.0,1.0,0.0,0.3),
                           type="line")
    return scatter_pv

def compute_function(ary):
    "Fills in ary with the sin(r)/r function"
    
    width, height = ary.shape
    for i in range(width):
        for j in range(height):
            x = i - width / 2.0
            x = x / (width/2.0) * 15
            y = j - height / 2.0
            y = y / (height/2.0) * 15
            
            radius = pow(x*x + y*y, 0.2)
            if radius == 0.0:
                ary[i,j] = 1
            else:
                ary[i,j] = math.sin(radius) / radius
    return


class ImagePlotApplicationWindow( PlotApplicationWindow ):

    ###########################################################################
    # PlotApplicationWindow interface.
    ###########################################################################
    
    def _create_plot( self ):
        """
        Creates the plot to be dispayed.  Although we create a plot with a
        scatter and image plot, we only save out the scatter plotvalue to
        the backbuffer.
        """

        # Create the image data and the index values
        value_grid = zeros((50,100), Float)
        compute_function(value_grid)
        index_vals = (arange(value_grid.shape[0]), arange(value_grid.shape[1]))
        data = ImageData(value_grid, index_vals)
        
        # Create the index axes
        xaxis = PlotAxis(tick_visible=True, grid_visible=False, \
                         bound_low = index_vals[0][0], bound_high = index_vals[0][-1])
        yaxis = PlotAxis(tick_visible=True, grid_visible=False, \
                     bound_low = index_vals[1][0], bound_high = index_vals[1][-1])
        
        # Create the scatter plot
        scatter = make_scatter(xaxis, yaxis)
    
        # Create the value axis (i.e. colormap)
        cmap = hsv(-0.5, 1.2, steps=15)
    
        # Create the Image PlotValue
        image_pv = CmapImagePlotValue(data, cmap, axis_index = xaxis, axis = yaxis,
                                     type='image')
        
        canvas = PlotCanvas(image_pv, scatter, axis_index=xaxis, axis=yaxis)
        
        # Generate a GC that just contains the image plot and the scatter plot.
        # Since image plots render during the draw_canvas phase, image_pv does
        # not hide the scatter plot behind it, even though it is supplied second
        # in the list.
        gc = gc_plot([scatter, image_pv], canvas, (1024, 768), pix_format="rgb24")
        
        # Save out the backbuffer so we can inspect it (for debugging purposes).
        gc.save("backbuffer.png")
        
        # Grab the data out of the backbuffer GC and create a new ImagePlotValue
        # out of it.  Re-use the axes from above; this isn't necessary to make
        # the process work, but we might as well.
        data = ImageData(gc.bmp_array)
        image = ImagePlotValue((xaxis.bound_low, xaxis.bound_high),
                                  (yaxis.bound_low, yaxis.bound_high), data=data,
                                  data_pix_format = "rgb24",
                                  axis_index=xaxis, axis=yaxis, type="image")
        
        canvas = PlotCanvas(image, axis_index = xaxis, axis = yaxis)
        pg = PlotGroup(canvas, ColormapLegend(cmap, weight=0.1))
        
        return pg

def main():

    # Create the GUI (this does NOT start the GUI event loop).
    gui = GUI()

    # Create and open the main window.
    window = ImagePlotApplicationWindow( title = "Plot" )
    window.size = (1024, 768)
    window.open()
    clientsize = window.control.GetClientSizeTuple()
    
    # Start the GUI event loop.
    gui.start_event_loop()
    

#=============================================================================== 
#  Program start-up:
#=============================================================================== 

if __name__ == '__main__':
    main()

#EOF
