#!/usr/bin/env python
"""
Demonstration of drawing a 2D image plot with a custom LinearColormap
"""

#--------------------------------------------------------------------------------
#  Imports:
#--------------------------------------------------------------------------------

import wx

from enthought.traits.api import Any, Instance
from enthought.enable.wx import Window
from enthought.enable.base import blue_color, transparent_color
from enthought.pyface.api import ApplicationWindow, GUI


from enthought.util.numerix import arange, array, ceil, concatenate, floor, \
                                   Float, ones, pi, ravel, resize, transpose, zeros
from math import sqrt, sin

from enthought.chaco import PlotAxis, PlotCanvas, PlotComponent, PlotGroup, \
                            PlotTitle, ImageData, CmapImagePlotValue, \
                            LinearColormap, ColormapLegend

from enthought.chaco.interaction import ZoomInteraction

from demo_base import PlotApplicationWindow


class ImagePlotApplicationWindow( PlotApplicationWindow ):

    ###########################################################################
    # PlotApplicationWindow interface.
    ###########################################################################
    
    def _create_plot( self ):
        """ Create the plot to be displayed. """
        
        # Create the image data and the index values
        value_grid = zeros((50,100), Float)
        self._compute_function(value_grid)
        index_vals = (arange(value_grid.shape[0]), arange(value_grid.shape[1]))
        data = ImageData(value_grid, index_vals)
        
        # Create the index axes
        xaxis = PlotAxis(tick_visible=False, grid_visible=False, \
                         bound_low = index_vals[0][0], bound_high = index_vals[0][-1])
        yaxis = PlotAxis(tick_visible=False, grid_visible=False, \
                     bound_low = index_vals[1][0], bound_high = index_vals[1][-1])
        xaxis.visible = False
        yaxis.visible = False
        
        # Create the value axis (i.e. colormap)
        cmap = LinearColormap(color_depth='rgba')
        cmap.min_color = (0.0, 0.0, 0.0, 1.0)
        cmap.max_color = (1.0, 0.0, 0.0, 1.0)
        cmap.min_value = floor(min(ravel(value_grid)))
        cmap.max_value = ceil(max(ravel(value_grid)))
        cmap.steps = 50
        
        zoom = ZoomInteraction()
        zoom.color = (0.439216, 0.858824, 0.576471, 0.3)    # transparent aquamarine
        zoom.border_color = blue_color
        zoom.bounds = (0,0,0,0)
        zoom.on_trait_change(self._do_complete, 'complete')
        
        # Create the Image PlotValue
        image = CmapImagePlotValue(data, cmap, axis_index = xaxis, axis = yaxis,
                                     type='image')
        image.weight = 10

        canvas = PlotCanvas(image,
                            PlotTitle('Dark at upper left. Bright at lower right.'),
                            PlotTitle('Brighter lower left than upper right.'),
                            plot_origin='top left',
                            interaction=zoom,
                            margin_width=10, margin_height=10)
        
        cmap_legend = ColormapLegend(cmap, plot_width=30)
        legend_canvas = PlotCanvas(cmap_legend, PlotTitle('Color', plot_min_width=10),
                                   PlotTitle('Key', plot_min_width=10),plot_min_width=10,
                                   margin_width=20, margin_height=20)
        legend_canvas.weight = 0.03
        
        group = PlotGroup(legend_canvas, canvas, orientation='horizontal')

        return group

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _do_complete(self, source, trait, event):
        source.event_state = 'normal'

    def _compute_function(self, ary):
        "Fills in ary with the sin(r)/r function"
        
        width, height = ary.shape
        for i in range(width):
            for j in range(height):
                x = i - width / 2.0
                x = x / (width/2.0) * 15
                y = j - height / 2.0
                y = y / (height/2.0) * 15
                
                radius = sqrt(x*x + y*y)
                if radius == 0.0:
                    ary[i,j] = 1
                else:
                    ary[i,j] = sin(radius) / radius + y/15 + x/30
                    
        return
    
def main():

    # Create the GUI (this does NOT start the GUI event loop).
    gui = GUI()

    # Screen size:
    screen_width = gui.system_metrics.screen_width or 1024
    screen_height = gui.system_metrics.screen_height or 768

    # Create and open the main window.
    window = ImagePlotApplicationWindow( title = "Plot" )
    window.size = ( 2 * screen_width / 3, 2 * screen_height / 3 )
##    window.size = (500,500)
    window.open()
    
    # Start the GUI event loop.
    gui.start_event_loop()


#=============================================================================== 
#  Program start-up:
#=============================================================================== 

if __name__ == '__main__':
    main()
