#!/usr/bin/env python
"""
Demonstration of drawing a 2D image plot with a custom LinearColormap
"""

#--------------------------------------------------------------------------------
#  Imports:
#--------------------------------------------------------------------------------

import wx

from enthought.traits.api import Any, Instance
from enthought.enable.wx import Window
from enthought.enable.base import blue_color, transparent_color
from enthought.pyface.api import ApplicationWindow, GUI

from enthought.util.numerix import arange, array, ceil, concatenate, floor,\
     Float, ones, pi, ravel, resize, transpose, zeros

from enthought.util.scipyx import ogrid, mgrid, sin, cos, sqrt # all symbols are available in numpy
from scipy.special import sinc                 # sinc is available in numpy

from enthought.chaco import PlotAxis, PlotCanvas, PlotComponent, PlotGroup, \
                            PlotTitle, ImageData, CmapImagePlotValue, \
                            LinearColormap, ColormapLegend

from enthought.chaco.interaction import PanZoomInteraction
from enthought.chaco.contour.contour_data import ContourData
from demo_base import PlotApplicationWindow


class ImagePlotApplicationWindow( PlotApplicationWindow ):

    def _get_data(self):

        value_grid = zeros((500,1000), Float)
        self._compute_function(value_grid)

        return value_grid
    
        x, y = mgrid[:500,:500] * 2* pi / 250.0
        z = sin(x) + cos(y)

        return z

    ###########################################################################
    # PlotApplicationWindow interface.
    ###########################################################################
    
    def _create_plot( self ):
        """ Create the plot to be displayed. """
        
        # Create the image data and the index values
        value_grid = self._get_data()
        index_vals = (arange(value_grid.shape[0]), arange(value_grid.shape[1]))
        data = ImageData(value_grid, index_vals)
        
        # Create the index axes
        xaxis = PlotAxis(tick_visible=False, grid_visible=False, \
                         bound_low = index_vals[0][0], bound_high = index_vals[0][-1])
        yaxis = PlotAxis(tick_visible=False, grid_visible=False, \
                     bound_low = index_vals[1][0], bound_high = index_vals[1][-1])
        xaxis.visible = False
        yaxis.visible = False
        
        # Create the value axis (i.e. colormap)
        cmap = LinearColormap(color_depth='rgba')
        cmap.min_color = (0.0, 0.0, 0.0, 1.0)
        cmap.max_color = (1.0, 0.0, 0.0, 1.0)
        
        cmap.min_value = floor(min(ravel(value_grid)))
        cmap.max_value = ceil(max(ravel(value_grid)))
        cmap.steps = 50
        
        # Create the Image PlotValue
        image = CmapImagePlotValue(data, cmap, axis_index = xaxis, axis = yaxis,
                                     type='image')
        image.weight = 10

        pan_zoom = PanZoomInteraction()
        canvas = PlotCanvas(image,
                            #PlotTitle('Dark at upper left. Bright at lower right.'),
                            #PlotTitle('Brighter lower left than upper right.'),
                            plot_origin='top left',
                            interaction=pan_zoom)

        contour = ContourData(data, 
                              color='black')
        contour.auto_level(10)
        contour.add_to(canvas)
        
        group = PlotGroup(canvas, orientation='horizontal')

        return group

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _do_complete(self, source, trait, event):
        source.event_state = 'normal'

    def _compute_function(self, ary):
        "Fills in ary with the sin(r)/r function"
                
        width, height = ary.shape
        
        x,y = ogrid[-1:1:width*1j,-1:1:height*1j]                
        radius = sqrt(x**2 + y**2) * 5.0
        
        # fix me: Should we really fill the array this way?  Seems like a 
        # return value would be better...        
        ary[:,:] = sinc(radius) + y + x/2

        return

def main():

    # Create the GUI (this does NOT start the GUI event loop).
    gui = GUI()

    # Screen size:
    screen_width = gui.system_metrics.screen_width or 1024
    screen_height = gui.system_metrics.screen_height or 768

    # Create and open the main window.
    window = ImagePlotApplicationWindow( title = "Plot" )
    window.size = ( 2 * screen_width / 3, 2 * screen_height / 3 )

    window.open()
    
    # Start the GUI event loop.
    gui.start_event_loop()


#=============================================================================== 
#  Program start-up:
#=============================================================================== 

if __name__ == '__main__':
    main()
