

# Other Enthought library imports.
from enthought.enable.wx import Window
from enthought.pyface.api import ApplicationWindow, GUI
from enthought.traits.api import List

# Chaco library imports.
from enthought.chaco.plot_canvas import PlotCanvas
from enthought.chaco.plot_component import PlotComponent
from enthought.chaco.plot_overlay import PlotOverlay
from enthought.chaco.plot_title import PlotTitle

from enthought.chaco.interaction import ClickInteraction

from gis_parser import GISParser
from state import State
from state_data import StateData

class GISApplicationWindow(ApplicationWindow):
    """ A PyFace application window for displaying the Chaco GIS plot.

    This example builds on Example 1 by adding a click interaction to the plot.
    The click interaction observes mouse click events on the plot and sets
    its 'clicked_value' trait when a PlotValue has been clicked. We use Traits'
    notification mechanism to listen for these clicks and update a basic label
    with the name of the state.
    """

    states = List(State)

    ###########################################################################
    # ApplicationWindow interface.
    ###########################################################################

    def _create_contents(self, parent):
        """ Create the toolkit-specific content for this window. """

        canvas = PlotCanvas()

        for state in self.states:
            state_value = StateData(state=state)
            canvas.add(state_value)
            
        canvas.add(PlotTitle('Continental United States', font='modern 14'))

        click = ClickInteraction()
        click.on_trait_change(self._value_clicked, 'clicked_value')
        canvas.interaction = click

        self._overlay = overlay = PlotOverlay(
            text="Click a State", position='bottom', bg_color=(0.8,0.8,0.8,1.0))
        canvas.add(overlay)

        plot_component = PlotComponent(component=canvas)

        window = Window(parent, component=plot_component)

        return window.control

    ###########################################################################
    # Private interface.
    ###########################################################################
    
    def _value_clicked(self, source, old, new):
        """ Called when a value is clicked on. """

        # Update the overlay with the State's name.
        if source.clicked_value is not None:
            state_data = source.clicked_value._source._data
            self._overlay.text = state_data.state.name
            self._overlay.bg_color = state_data.color 
            self._overlay.update()
                
        return

###############################################################################
# Program start-up.
###############################################################################

def main():
    """ Application main routine. """

    # Parse all of the state data into a 'State' objects.
    parser = GISParser()
    states = parser.parse('../data/states/statesp020')

    # Exclude states not in Continental U.S.
    exclude = ['Alaska', 'Hawaii', 'U.S. Virgin Islands', 'Puerto Rico']
    continental_states = [ state for state in states.values()
                           if state.name not in exclude ]
                           
    for state in continental_states:
        state.downsample(1200, 600, 65, 25)

    # Create the GUI object (this does NOT start the event loop).
    gui = GUI()

    # Create the main window and open it.
    window = GISApplicationWindow(size=(600, 400), states=continental_states)
    window.open()
    
    # Start the GUI event loop.
    gui.event_loop()

    return

if __name__ == '__main__':
    main()

#### EOF ######################################################################
