

# Other Enthought library imports.
from enthought.enable.wx import Window
from enthought.pyface.api import ApplicationWindow, GUI
from enthought.traits.api import List

# Chaco library imports.
from enthought.chaco.plot_canvas import PlotCanvas
from enthought.chaco.plot_component import PlotComponent
from enthought.chaco.plot_group import PlotGroup
from enthought.chaco.plot_overlay import PlotOverlay
from enthought.chaco.plot_title import PlotTitle

from enthought.chaco.interaction import ClickInteraction

from gis_parser import GISParser
from state import State
from state_data import StateData
from unemployment_parser import UnemploymentParser

class GISApplicationWindow(ApplicationWindow):
    """ A PyFace application window for displaying the Chaco GIS plot.

    This example builds on Example 1 by adding a click interaction to the plot.
    The click interaction observes mouse click events on the plot and sets
    its 'clicked_value' trait when a PlotValue has been clicked. We use Traits'
    notification mechanism to listen for these clicks and update a basic label
    with the name of the state.
    """

    # The list of all states.
    states = List(State)

    # The list of selected states.
    selected_states = List

    ###########################################################################
    # ApplicationWindow interface.
    ###########################################################################

    def _create_contents(self, parent):
        """ Create the toolkit-specific content for this window. """

        canvas = PlotCanvas()

        for state in self.states:
            state_value = StateData(state=state)
            canvas.add(state_value)
            
        canvas.add(PlotTitle('Continental United States', font='modern 14'))

        click = ClickInteraction()
        click.on_trait_change(self._value_clicked, 'clicked_value')
        canvas.interaction = click

        self._overlay = overlay = PlotOverlay(
            text="Click a State", position='bottom', bg_color=(0.8,0.8,0.8,1.0))
        canvas.add(overlay)

        self._detail = PlotCanvas(
            PlotTitle('Historical Unemployment Rate', font='modern 14')
        )
        self._group = group = PlotGroup(canvas, self._detail)

        plot_component = PlotComponent(component=group)

        window = Window(parent, component=plot_component)

        return window.control

    ###########################################################################
    # Private interface.
    ###########################################################################

    ### Trait Listeners #######################################################

    ### Static ###

    def _selected_states_items_changed(self, event):
        """ Called when the list of selected states changes. """

        # Add any/all states added.
        for state_data in event.added:
            state_data.add_to(
                self._detail, 'unemployment_Seasonally Adjusted_rate'
            )

        # Remove any/all states removed.
        for state_data in event.removed:
            for value in self._detail.values:
                if value._source._data is state_data:
                    self._detail.remove(value)

        self._detail.update()

        return

    def _selected_states_changed(self, old, new):

        # Remove all the previously added states.
        for value in self._detail.values[:]:
            self._detail.remove(value)

        if new is not None:
            for state_data in new:
                state_data.add_to(
                    self._detail, 'unemployment_Seasonally Adjusted_rate'
                )

        self._detail.update()

        return

    ### Dynamic ###
    
    def _value_clicked(self, source, old, new):
        """ Called when a value is clicked on. """

        if source.clicked_value is not None:
            state_data = source.clicked_value._source._data
            
            # If the control key was down, toggle the selection of the state.
            if source.event.control_down:
                if state_data in self.selected_states:
                    self.selected_states.remove(state_data)
                else:
                    self.selected_states.append(state_data)

            # Otherwise, set the selection to exactly the state clicked.
            else:
                self.selected_states = [state_data]

        return

###############################################################################
# Program start-up.
###############################################################################

def _get_state_data():
    
    parser = GISParser()
    states = parser.parse('../data/states/statesp020')
    
    parser = UnemploymentParser()
    unemployment = parser.parse('../data/states/unemployment.csv')
    
    # Exclude states not in Continental U.S.
    exclude = ['Alaska', 'Hawaii', 'U.S. Virgin Islands', 'Puerto Rico']
    continental_states = [ state for state in states.values()
                           if state.name not in exclude ]

    for state in continental_states:
        state.downsample(1200, 600, 65, 25)

    for state in continental_states:
        if unemployment.has_key(state.name):
            state.statistics['unemployment'] = unemployment[state.name]
            
    return continental_states

def main():
    """ Application main routine. """

    # Parse all of the state data into a 'State' objects.
    continental_states = _get_state_data()

    # Create the GUI object (this does NOT start the event loop).
    gui = GUI()

    # Create the main window and open it.
    window = GISApplicationWindow(size=(1200, 600), states=continental_states)
    window.open()

    # Start the GUI event loop.
    gui.event_loop()

    return

if __name__ == '__main__':
    main()

#### EOF ######################################################################
