

# Enthought library imports.
from enthought.util.numerix import array
from enthought.traits.api import Dict, HasTraits, String, List


class State(HasTraits):
    """ A U.S. state and some associated data. """

    # The name of the state.
    name = String
    
    # The state boundaries - A list of arrays of coordinates.
    boundaries = List

    # A dictionary of miscellaneous statistics about the state.
    statistics = Dict

    # A backup of the non-downsampled boundary data in this object
    _full_boundaries = List([])

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, name, **traits):
        """ Constructor. """

        # Super-class constructor.
        super(State, self).__init__(**traits)

        # Explicitly set required traits.
        self.name = name
        
        return

    
    def downsample(self, longitude_pix, latitude_pix, longitude_span=None, 
                   latitude_span=None):
        """
        Downsamples all the boundaries in this object given the number of
        pixels that will be used to express the logitude and latitude.
        
        If longitude span is given, then that represents the delta of longitude
        that will be represented by in logitude_pix number of pixels.  If it is
        None, then the longitude range will be drawn from each boundary.
        
        The same applies for latitude_span and latitude.
        """
        # Create the backup of the current data if there isn't one.
        if self._full_boundaries == []:
            self._full_boundaries = self.boundaries
        
        old_len = len(self.boundaries)
        self.boundaries = []
        for i, boundary in enumerate(self._full_boundaries):
            old_inner_len = len(boundary)
            inner_boundary_list = []
            for j, sub_boundary in enumerate(boundary):
                tmp = self._downsample(sub_boundary, longitude_pix, latitude_pix,
                                       longitude_span, latitude_span)
                if tmp:
                    inner_boundary_list.append(tmp)
            self.boundaries.append(inner_boundary_list)
        return

    def _downsample(self, boundary, longitude_pix, latitude_pix, longitude_span,
                    latitude_span):
        
        if len(boundary) <= 2:
            return boundary
        
        # create an Nx2 array out of the list of 2-tuples
        boundary = array(boundary)
        
        if longitude_span is None:
            # This is very naive and doesn't handle the prime meridian or any
            # coordinate wraparound nonsense.
            longitude_span = max(boundary[:,0]) - min(boundary[:,0])
        
        if latitude_span is None:
            latitude_span = max(boundary[:,1]) - min(boundary[:,1])
        
        hres = float(longitude_span) / longitude_pix
        vres = float(latitude_span) / latitude_pix
        
        new_points = [boundary[0]]
        last_pt = boundary[0]
        for pt in boundary:
            if (abs(pt[0]-last_pt[0]) < hres) and (abs(pt[1]-last_pt[1]) < vres):
                continue
            else:
                new_points.append(pt)
                last_pt = pt
        
        new_points.extend(boundary[-2:])
        if len(new_points) < 5:
            return []
        else:
            #print "downsampled %d points to %d" % (len(boundary), len(new_points))
            return new_points


#### EOF ######################################################################
