#-------------------------------------------------------------------------------
# 
#  Sample program to illustrate rendering Chaco plot objects into a PDF file
#  
#  Written by: David C. Morrill
#
#  Date: 12/30/2002
#
#  (c) Copyright 2002 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

from enthought.chaco import PlotAxis, PlotCanvas, PlotFrame, PlotGroup, \
                            PlotIndexed, PlotLabel, PlotOverlay, PlotTitle, \
                            PlotValue, PlotTitle

from trait_object                import TraitObject, TraitObjectError
from enthought.kiva.pdfcore2d    import Canvas as PDF_Canvas
from glob                        import glob
from os.path                     import join, dirname
from enthought.util.numerix      import *
from enthought.util.randomx      import *
    
import enthought.chaco.plot_canvas
       
#-------------------------------------------------------------------------------
#  Handle the user opening a new chaco demo file:  
#-------------------------------------------------------------------------------
    
def render_scripts ( ):
    # Set up a PDF graphics context to render into:
    pdf    = PDF_Canvas( 'plot.pdf' )
    dx, dy = pdf.size()
    dx    -= 16   # HACK: Seems necessary to get correct page size
    dy    += 50   # HACK: Seems necessary to get correct page size
    gc     = pdf.gc
    groups = []
    
    for file in glob( join( dirname( enthought.chaco.plot_canvas.__file__ ), 
                            'demo', 'scripts', '*.co' ) ):
        try:     
            # Construct a Plottable object from the contents of the script file:                   
            group = PlotGroup( orientation = 'vertical',
                               index       = PlotValue( arange( 31.0 ), 
                                                        axis = PlotAxis() ) )
            group.add( PlotTitle( text           = 'File: %s' % file,
                                  bg_color       = 'cyan',
                                  border_size    = 1,
                                  margin_height  = 4,
                                  padding_height = 4 ) )
            build_plot_object( group, TraitObject( file ) )
         
            # Render a full-page plot:
            render_at( group, gc, dx, dy, 0.0, 0.0, 1.0 )
            gc.end_page()
            
            # If it worked, add it to the collection:
            groups.append( group )
        except:
            print 'Error occurred rendering script file: %s' % file

    # Render each plot rotated 90 degrees:
    for group in groups:
        render_at_90(  group, gc, dx, dy, 0.0, 0.0, 1.0 )
        gc.end_page()
        
    # Render 4 smaller plots per page:
    for i in range( 0, len( groups ), 4 ):
        try:
            render_at( groups[i],   gc, dx, dy, 0.0, 0.0 )
            render_at( groups[i+1], gc, dx, dy, 0.5, 0.0 )
            render_at( groups[i+2], gc, dx, dy, 0.0, 0.5 )
            render_at( groups[i+3], gc, dx, dy, 0.5, 0.5 )
        except:
            pass
        gc.end_page()

    # Render 4 smaller plots per page rotated 90 degrees:
    for i in range( 0, len( groups ), 4 ):
        try:
            render_at_90( groups[i],   gc, dx, dy, 0.0, 0.0 )
            render_at_90( groups[i+1], gc, dx, dy, 0.5, 0.0 )
            render_at_90( groups[i+2], gc, dx, dy, 0.0, 0.5 )
            render_at_90( groups[i+3], gc, dx, dy, 0.5, 0.5 )
        except:
            pass
        gc.end_page()
    
    # Save the entire set of pages to a file:
    pdf.save()
    
#-------------------------------------------------------------------------------
#  Render a Plottable object at a specified position and scale factor:
#-------------------------------------------------------------------------------
    
def render_at ( canvas, gc, dx, dy, x, y, scale = 0.5 ):
    gc.save_state()
    gc.translate_ctm( x * dx + ((dx * scale) / 40.0), 
                      y * dy + ((dy * scale) / 40.0) )
    gc.scale_ctm( 0.95 * scale, 0.95 * scale )
    canvas.render( gc, 0, 0, dx, dy )
    gc.restore_state()
    
#-------------------------------------------------------------------------------
#  Render a Plottable object at a specified position and scale factor 
#  rotated 90 degrees:
#-------------------------------------------------------------------------------
    
def render_at_90 ( canvas, gc, dx, dy, x, y, scale = 0.5 ):
    gc.save_state()
    gc.rotate_ctm( 3.14159625 / 2.0 )
    gc.translate_ctm( y * dy + ((dy * scale) / 40.0), 
                      x * dx + ((dx * scale) / 40.0) - dx )
    gc.scale_ctm( 0.95 * scale, 0.95 * scale )
    canvas.render( gc, 0, 0, dy, dx )
    gc.restore_state()

#-------------------------------------------------------------------------------
#  Construct a plot object hierarchy from a TraitObject:
#-------------------------------------------------------------------------------
   
def build_plot_object ( object, trait_object ):
    traits = trait_object.traits
    try:
        del traits[ 'index_scale' ]
    except:
        pass
    try:
        del traits[ 'value_scale' ]
    except:
        pass
    object.set( traits )
    for po in trait_object.contents:
        handler = { 'group':   add_group,
                    'indexed': add_indexed,
                    'canvas':  add_canvas,
                    'value':   add_value,
                    'overlay': add_overlay,
                    'title':   add_title,
                    'axis':    add_axis }.get( po.name.lower() )
        if handler is None:
            raise TraitObjectError( "Unknown chaco object type: '%s'" % 
                                    po.name )
        build_plot_object(
            handler( object, overlay = po.traits.get( 'overlay', False ) ), po )
   
#-------------------------------------------------------------------------------
#  Add a new PlotGroup to a specified object:
#-------------------------------------------------------------------------------

def add_group ( object, orientation = 'h', overlay = False ):
    group = PlotGroup( orientation = orientation, overlay = overlay )
    object.add( group )
    return group

#-------------------------------------------------------------------------------
#  Add a new PlotIndexed to a specified object:
#-------------------------------------------------------------------------------

def add_indexed ( object, overlay = False ):
    indexed = PlotIndexed( overlay = overlay )
    object.add( indexed )
    return indexed

#-------------------------------------------------------------------------------
#  Add a new PlotCanvas to a specified object:
#-------------------------------------------------------------------------------

def add_canvas ( object, overlay = False ):
    canvas = PlotCanvas( overlay = overlay )
    object.add( canvas )
    return canvas

#-------------------------------------------------------------------------------
#  Add a new PlotValue to a specified object:
#-------------------------------------------------------------------------------

def add_value ( object, bias = 1.5, overlay = False ):
    value = rpv( 30, bias, overlay )
    object.add( value )
    return value

#-------------------------------------------------------------------------------
#  Add a new PlotOverlay to a specified object:
#-------------------------------------------------------------------------------

def add_overlay ( object, overlay = False ):
    overlay = PlotOverlay( text = 'Overlay' )
    object.add( overlay )
    return overlay

#-------------------------------------------------------------------------------
#  Add a new PlotTitle to a specified object:
#-------------------------------------------------------------------------------

def add_title ( object, overlay = False ):
    title = PlotTitle( text = 'Title' )
    object.add( title )
    return title

#-------------------------------------------------------------------------------
#  Add a new PlotAxis to a specified object:
#-------------------------------------------------------------------------------

def add_axis ( object, overlay = False ):
    object.axis = axis = PlotAxis( title = 'Axis' )
    return axis

#-------------------------------------------------------------------------------
#  Generate a set of pseudo-random noise values to plot:
#-------------------------------------------------------------------------------

def rpv ( n = 30, bias = 1.5, overlay = False ):
    delta = (2 * pi) / float( n )
    xx    = arange( 0.0, 2 * pi + delta / 2.0,
                    delta ) * ((random() * 2.0) + 0.5) + (2 * pi * random())
    xx2   = 0.4 * random( ( n + 1, ) )
    return PlotValue( sin( xx ) + xx2 + bias, overlay = overlay )

#-------------------------------------------------------------------------------
#  Main program:
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    render_scripts()
