""" The package explorer view. """


# Major package imports.
import wx

# Enthought library imports.
from enthought.charm.module import Module
from enthought.charm.ui.package_tree import PackageTree
from enthought.envisage.workbench.api import View
from enthought.io.api import File
from enthought.traits.api import HasTraits, Str

# Local imports.
from charm_view import CharmView


class PackageExplorer(CharmView):
    """ The package explorer view. """

    #### 'PackageExplorer' interface ##########################################
    
    # The root of the tree (currently the name of a package).
    #
    # e.g. 'enthought.envisage.core'
    root = Str('enthought.charm')

    ###########################################################################
    # 'View' interface.
    ###########################################################################

    def create_control(self, parent):
        """ Creates the toolkit-specific control that represents the view.

        'parent' is the toolkit-specific control that is the view's parent.

        """

        control = self._create_package_explorer(parent)

        # When we are parsing a module, put a message on the status bar.
        def listener(event):
            self.status = 'parsing module %s' % event
            
        self.charm.on_trait_change(listener, 'parsing_module')

        # Clear the message bar.
        self.status = ''
        
        return control
    
    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_package_explorer(self, parent):
        """ Creates the package explorer.

        This is the combination of the package chooser and the package tree.

        """

        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)
        
        # Create the chooser.
        package_chooser = self._create_package_chooser(panel)
        sizer.Add(package_chooser, 0, wx.EXPAND)

        # Create the tree.
        package_tree = self._create_package_tree(panel)
        sizer.Add(package_tree, 1, wx.EXPAND)

        # Resize the panel to match the sizer's minimum size.
        sizer.Fit(panel)
        
        return panel

    def _create_package_chooser(self, parent):
        """ Creates the package chooser. """

        class PackageChooser(HasTraits):
            package_name = Str(self.root)

        chooser = PackageChooser()
        chooser.on_trait_change(self._on_package_name_changed, 'package_name')
        
        # Create the chooser.
        ui = chooser.edit_traits(parent=parent, kind='subpanel')
        
        return ui.control
    
    def _create_package_tree(self, parent):
        """ Creates the package tree. """

        self._package_tree = tree = PackageTree(parent)
        tree.root = self.charm.code_browser.read_package(self.root)
        tree.on_trait_change(self._on_node_activated, 'node_activated')
        tree.on_trait_change(self._on_selection_changed, 'selection')

        return tree.control

    #### Trait change handlers ################################################

    def _on_node_activated(self, tree, trait_name, old, new):
        """ Dynamic trait change handler. """

        if isinstance(new, Module):
            # Set Charm's current module.
            self.charm.module = new

            # Edit the module.
            workbench = self.window.workbench
            workbench.edit(File(new.filename))
            
        return

    def _on_selection_changed(self, tree, trait_name, old, new):
        """ Dynamic trait change handler. """

        if len(new) == 1:
            if isinstance(new[0], Module):
                self.charm.module = new[0]
            
        return
    
    def _on_package_name_changed(self, new):
        """ Dynamic trait change handler. """
        
        self.busy = True
        try:
            self._package_tree.root = self.charm.code_browser.read_package(new)

        except:
            pass
        self.busy = False

        self.status = ''

        return

#### EOF ######################################################################
