#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A simple code browser. """


# Major package imports.
import logging
import wx

# Enthought library imports.
from enthought.charm.api import Charm, Klass, Module
from enthought.pyface.api import GUI, PythonEditor, PythonShell
from enthought.pyface.api import SplitApplicationWindow, SplitPanel
from enthought.traits.api import HasTraits, Instance, Str

# Local imports.
from outline_tree import OutlineTree
from package_tree import PackageTree


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class MainWindow(SplitApplicationWindow):
    """ The main application window. """

    #### 'Window' interface ###################################################

    title = 'Code Browser'

    #### 'SplitApplicationWindow' interface ###################################

    # The direction in which the panel is split.
    direction = 'vertical'

    # The ratio of the size of the left/top pane to the right/bottom pane.
    ratio = 0.25

    #### 'MainWindow' interface ###############################################

    # The Charm that we are the UI for.
    charm = Instance(Charm)

    # The root of the tree (currently the name of a package).
    #
    # e.g. 'enthought.envisage.core'
    root = Str

    ###########################################################################
    # Protected 'SplitApplicationWindow' interface.
    ###########################################################################

    def _create_lhs(self, parent):
        """ Creates the left hand side or top depending on the style. """

        panel = SplitPanel(
            parent    = parent,
            direction = 'horizontal',
            ratio     = 0.5,
            lhs       = self._create_package_explorer,
            rhs       = self._create_outline_tree
        )

        return panel.control

    def _create_rhs(self, parent):
        """ Creates the panel containing the selected preference page. """

        panel = SplitPanel(
            parent    = parent,
            direction = 'horizontal',
            ratio     = 0.8,
            lhs       = self._create_python_editor,
            rhs       = self._create_python_shell
        )

        return panel.control

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_package_explorer(self, parent):
        """ Creates the package explorer.

        This is the combination of the package chooser and the package tree.

        """

        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # Create the chooser.
        package_chooser = self._create_package_chooser(panel)
        sizer.Add(package_chooser, 0, wx.EXPAND)

        # Create the tree.
        package_tree = self._create_package_tree(panel)
        sizer.Add(package_tree, 1, wx.EXPAND)

        # Resize the panel to match the sizer's minimum size.
        sizer.Fit(panel)

        return panel

    def _create_package_chooser(self, parent):
        """ Creates the package chooser. """

        class PackageChooser(HasTraits):
            package_name = Str(self.root)

        package_chooser = PackageChooser()
        package_chooser.on_trait_change(
            self._on_package_name_changed, 'package_name'
        )

        # Create the chooser.
        ui = package_chooser.edit_traits(parent=parent, kind='subpanel')

        return ui.control

    def _create_package_tree(self, parent):
        """ Creates the package tree. """

        self._package_tree = tree = PackageTree(parent)
        tree.on_trait_change(self._on_node_activated, 'node_activated')
        tree.root = self.charm.code_browser.read_package(self.root)

        return tree.control

    def _create_python_editor(self, parent):
        """ Creates the Python editor. """

        self._python_editor = python_editor = PythonEditor(parent)
        python_editor.on_trait_change(self._on_key_pressed, 'key_pressed')

        return python_editor.control

    def _create_outline_tree(self, parent):
        """ Creates the outline tree. """

        self._outline_tree = tree = OutlineTree(parent, show_lines=False)
        tree.on_trait_change(self._on_selection_changed, 'selection')

        return tree.control

    def _create_python_shell(self, parent):
        """ Creates the Python shell. """

        self._python_shell = python_shell = PythonShell(parent)

        return python_shell.control

    #### Trait event handlers #################################################

    def _on_node_activated(self, tree, trait_name, old, new):
        """ Called when a node in the package tree is activated. """

        # fixme: If we don't import this here, then when module is used in this
        # function it resolves to a CTrait object?!?! How did that happen?
        from enthought.charm.module import Module

        # A klass, method, trait or attribute...
        if hasattr(new, 'lineno'):

            namespace = new.namespace
            while not hasattr(namespace, 'filename'):
                namespace = namespace.namespace

            self._python_editor.path = namespace.filename
            # fixme: It seems we need '-1' here otherwise the STC shows the
            # line after the class definition (even thought the line number
            # is definitely the line number that the class keyword appears
            # on!).
            self._python_editor.select_line(new.lineno - 1)

            # Center the selected line.
            stc = self._python_editor.control
            lines_on_screen = stc.LinesOnScreen()
            stc.ScrollToLine(new.lineno - (lines_on_screen / 2))

            if isinstance(new, Klass):
                self._outline_tree.root = new
                self._outline_tree.expand_all()

        # A module...
        elif isinstance(new, Module):
            self._python_editor.path = new.filename
            self._outline_tree.root = new
            self._outline_tree.expand_all()

            self.title = 'Code Browser (%s)' % new.filename

        return

    def _on_selection_changed(self, tree, trait_name, old, new):
        """ Called when a node in the outline tree is selected. """

        if len(new) == 0:
            return

        new = new[0]

        # A klass, method, trait or attribute...
        if hasattr(new, 'lineno'):

            namespace = new.namespace
            while not hasattr(namespace, 'filename'):
                namespace = namespace.namespace

            self._python_editor.path = namespace.filename

            stc = self._python_editor.control
            lines_on_screen = stc.LinesOnScreen()
            current_lineno = stc.LineFromPosition(stc.GetCurrentPos())

            # fixme: It seems we need '-1' here otherwise the STC shows the
            # line after the class definition (even thought the line number
            # is definitely the line number that the class keyword appears
            # on!).
            self._python_editor.select_line(new.lineno - 1)

            if abs(new.lineno - current_lineno) > (lines_on_screen / 3):
                # Center the selected line.
                stc.ScrollToLine(new.lineno - (lines_on_screen / 2))

        return

    def _on_package_name_changed(self, new):
        """ Called when the package name has changed. """

        try:
            self._package_tree.root = self.charm.code_browser.read_package(new)

        except:
            pass

        return

    def _on_key_pressed(self, event):
        """ Called when a key is pressed. """

        # fixme: Errr, we might want to loosen up the coupling between key
        # bindings and actions!
        #
        # Ctrl-S saves the file!
        if event.control_down and event.key_code == 19:
            self._python_editor.save()
            logger.debug('saved file %s' % self._python_editor.path)

            m = self.charm.code_browser.read_file(self._python_editor.path)
            self._outline_tree.root = m# refresh(self._outline.root)
            self._outline_tree.expand_all()

            # fixme: Ok, we've exhautsed the static model we currently have...
            # Time to make it dynamic!
            n = self._package_tree.selection[0]
            n.attributes = m.attributes
            n.functions = m.functions
            n.klasses = m.klasses

            self._package_tree.refresh(n)

        # Ctrl-R saves AND runs the file!
        elif event.control_down and event.key_code == 18:
            self._python_editor.save()

            # Execute the script in the python shell.
            path = self._python_editor.path
            self._python_shell.execute_command('execfile(r"%s")' % path)

        return


# Application entry point.
def main():
    """ Do it! """

    # Create a new Charm!
    charm = Charm()

    # Create the GUI (this does NOT start the GUI event loop).
    gui = GUI()

    # Create and open the main window.
    window = MainWindow(
        charm    = charm,
        root     = 'enthought.charm',
        size     = (1200, 800),
        position = (125, 25)
    )
    window.open()

    # Start the GUI event loop.
    gui.start_event_loop()

    # Save the code database.
    charm.save()

    return

if __name__ == '__main__':
    main()

##### EOF #####################################################################
