""" Base-class for all callable components (functions, scripts) etc. """


# Standard library imports.
import logging
import wx

# Enthought library imports.
from enthought.traits.api import Trait, TraitList, Property

# Local imports.
from enthought.component.component import Component
from enthought.component.input import input_list_trait
from enthought.component.output import output_list_trait
from enthought.units.unit_manager import IncompatibleUnitFamilies


# Setup a logger for this module
logger = logging.getLogger(__name__)


class CallableComponent(Component):
    """ Base-class for all callable components (functions, scripts) etc. """

    inputs = input_list_trait

    outputs = output_list_trait

    ###########################################################################
    # 'CallableComponent' interface.
    ###########################################################################

    #### Properties ###########################################################
    def get_input_names(self):
        return self._get_input_names()

    def _get_input_names(self):
        """ Returns the names of the component's inputs. """

        return [input.name for input in self.inputs]


    input_names = Property(_get_input_names)

    def get_output_names(self):
        return self._get_output_names()

    def _get_output_names(self):
        """ Returns the names of the component's outputs. """

        return [output.name for output in self.outputs]

    output_names = Property(_get_output_names)

    #### Methods ##############################################################

    def invoke(self, namespace):
        """ Invoke (run) the component in the specified namespace. """

        raise NotImplementedError

    ###########################################################################
    # Protected 'CallableComponent' interface.
    ###########################################################################

    def _prepare_argument(self, input, arg, __index__=None):
        """ Do any unit conversion/type coercion on an argument. """

        # The coercer does, errrr, type coercion.
        #
        # fixme: We need a better way to do this... it needs to be extensible
        # so that future plugins can add to the coercion rules.
        #
        # double fixme: An import from CP!
        from cp.layer.coercer import Coercer
        coercer = Coercer()

        # Can we convert this type of object?
        if hasattr(arg, 'change_unit_system'):
            arg = self._change_unit_system(input, arg)

        # Do any required type coercion for the input type.
        arg = coercer.coerce(input, arg, index=__index__)

        return arg

    def _change_unit_system(self, input, value):
        """ Converts a value to the unit system required by the input. """

        # fixme: CP import!
        from cp.log import LogSuite
        if isinstance(value, LogSuite):
            return value.change_unit_system(input.unit_system)

        # Algorithms define their inputs to be in a certain unit system
        # such as 'kgs' and here is where we convert from the units that
        # the input is in into the units expected by the algorithm.
        # Compare un-aliased family names

        # todo - clean this up - move it to Log?

        to_system = input.unit_system
        to_family = to_system.get_family_name(input.unit_family)

        if to_family != 'unknown':
            try:
                value_family = to_system.get_family_name(value.family_name)
                if to_family != value_family :
                    raise IncompatibleUnitFamilies(value_family, to_family)

                # change_unit_system will raise if fundamental units can't be
                # converted...
                value = value.change_unit_system(to_system)

            except Exception, ex:
                logger.exception(
                    "While constructing namespace for callable component %s \n"
                    "  failed to convert %s\n"
                    "  to the %s system\n"
                    "  with the family name of %s\n"
                    "  value is %s \n"
                    "  with the family name of %s\n"
                    "  and the class is %s"
                    % (
                    self.name, input, to_system, to_family, value,
                    value_family, value.__class__
                    ))
                raise ex

        return value

#### EOF ######################################################################

