""" A table model used for listing components in the registry. """


# Enthought library imports.
from enthought.component.registry import registry
from enthought.pyface.table_model import TableModel
from enthought.traits.api import HasTraits, Trait, TraitProperty


class ComponentRegistryTableModel(TableModel):
    """ A table model used for listing components in the registry. """

    __traits__ = {
        # The query used to match components in the registry.
        'query' : Trait('', None, str),
    }

    def __init__(self, query=None):
        """ Create a new table model. """

        # Base-class constructor.
        HasTraits.__init__(self, query=query)

        # The columns in the model.
        self._columns = ['Name', 'Description']
        
        return

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    #### Static ###############################################################
    
    def query_changed(self, old, new):
        """ Called when the query is changed. """

        if new is None:
            # Get ALL components in the registry.
            self._components = registry.components

        else:
            # Find all components in the registry that match the query.
            self._components = registry.find(new)

        return

    ###########################################################################
    # 'ComponentRegistryTableModel' interface.
    ###########################################################################

    def get_component_at(self, row):
        """ Return the component at the specified row. """

        return self._components[row]

    def get_index_of(self, component):
        """ Return the index of the specified component.

        Returns -1 if the component is not in the model.

        """

        for index, c in enumerate(self._components):
            if c.name == component.name:
                break

        else:
            index = -1
            
##         try:
##             index = self._components.index(component)
            
##         except ValueError:
##             index = -1
            
        return index
    
    ###########################################################################
    # 'TableModel' interface.
    ###########################################################################

    def _get_row_count(self):
        """ Return the number of rows in the model. """
        return len(self._components)

    __traits__['row_count'] = TraitProperty(_get_row_count)

    def _get_column_count(self):
        """ Return the number of columns in the model. """
        
        return len(self._columns)

    __traits__['column_count'] = TraitProperty(_get_column_count)
    
    def get_value_at(self, row, column):
        """ Return the value for the cell at (row, column). """

        component = self._components[row]

        if column == 0:
            return component.name

        elif column == 1:
            # Get the first line of the function's doc string as the
            # description.
            doc = component.description.split('\n')[0]
            return doc

        else:
            raise 'Unknown column!!!!'
        
        return

    def get_column_name(self, index):
        """ Return the name of the column at the specified index. """

        return self._columns[index]

#### EOF ######################################################################
