""" A function and its input and output bindings. """


# Standard library imports.
import logging
import new
import types

# Enthought library imports.
from enthought.traits.api import Bool, CBool, CStr, HasTraits, Instance, List, \
                             Property, Str

# Local imports.
from callable_component import CallableComponent


# Setup a logger for this module
logger = logging.getLogger(__name__)


class Binding(HasTraits):
    """ A string->string binding. """

    name = Str()
    value = CStr()
    description = Str()
    visible = CBool(True)

    ###########################################################################
    # 'HasTraitsPlus' interface.
    ###########################################################################

    def clone(self):
        """ Returns a clone of the object. """

        return Binding(name = self.name,
                       value = self.value,
                       description = self.description,
                       visible = self.visible)


class FunctionBinding(HasTraits):
    """ A function and its input and output bindings. """

    # The function that we are binding.
    function = Property

    # Bindings from formal to actual inputs.
    input_bindings = List(Instance(Binding))

    # Bindings from formal to actual outputs.
    output_bindings = List(Instance(Binding))

    # fixme: I don't think this should be here. Some higher-level object
    # should aggregate a function call and the enabled/disabled notion.
    #
    # Is the function call enabled?
    enabled = Bool(True)

    ###########################################################################
    # protected interface.
    ###########################################################################

    # The name of the function we're bound to
    _function_name = Str

    # The actual object we're bound to. This is lazy loaded to allow
    # scripts in a project's local directory to be referenced.
    _function = Instance(CallableComponent, allow_none = True)


    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Creates a new function binding. """

        # Base-class constructor.
        super(FunctionBinding, self).__init__(**traits)

        return


    def _function_changed(self):

        # initialize bindings to make the defaults work.
        # fixme: shouldn't the default behavior be pushed down into some
        #        kind of get_value function on an input?

        # 1/11/06 jsugg: build the input and output bindings separately,
        # then set them to the trait, rather than repeatedly calling append
        # on the trait list itself. performance enhancement.
        input_bindings = []
        for input in self.function.inputs:
            binding = Binding(name = input.name,
                              value = self._get_default_value(input),
                              description = input.description,
                              visible = input.visible)
            input_bindings.append(binding)

        self.input_bindings = input_bindings

        output_bindings = []
        for output in self.function.outputs:
            binding = Binding(name = output.name,
                              value = self._get_default_value(output),
                              description = output.description,
                              visible = output.visible)
            output_bindings.append(binding)

        self.output_bindings = output_bindings

        return

    def _get_default_value(self, io):
        """ The default value of an input.
        """
        if io.default_value != '' or io.default_value is None:
            value = io.default_value
        else:
            value = io.name

        return value

    def __str__(self):
        """ Returns an 'informal' string representation of the object. """

        s = self._function_name

        if not self.enabled:
            s = s + '  (disabled)'

        return s

    ###########################################################################
    # 'HasTraitsPlus' interface.
    ###########################################################################

    def clone(self):
        """ Returns a clone of the object. """

        input_bindings = [
            binding.clone() for binding in self.input_bindings
        ]

        output_bindings = [
            binding.clone() for binding in self.output_bindings
        ]

        clone = FunctionBinding(function = self.function,
                                input_bindings = input_bindings,
                                output_bindings = output_bindings,
                                enabled = self.enabled)

        return clone

    ###########################################################################
    # 'FunctionBinding' interface.
    ###########################################################################

    def bind_input(self, name, value):
        """ Binds the formal input named 'name' to 'value. """

        for binding in self.input_bindings:
            if binding.name == name:
                binding.value = value
                break

        else:
            raise ValueError("No input named %s" % name)

        return

    def bind_output(self, name, value):
        """ Binds the formal output named 'name' to 'value. """

        for binding in self.output_bindings:
            if binding.name == name:
                binding.value = value
                break

        else:
            raise ValueError("No output named %s" % name)

        return


    def run(self, namespace):
        """ Runs the function in the specified namespace. """

        # Extract the input arguments from the namespace.
        args = []
        for binding in self.input_bindings:
            input_name = binding.value

            # Attempt to lookup the name in the namespace.
            try:
                arg = namespace[input_name]

            # If the lookup fails then attempt to evaluate the name in the
            # namespace. This allows the input name to be either a literal
            # Python value (eg. 7, "foo", {'z' : 26} etc) or a reference to
            # an attribute of some instance in the namespace (eg.
            # 'foo.bar', or 'foo.bar.baz' etc).
            except KeyError:
                arg = eval(input_name, namespace, namespace)
                # Eval puts __builtins__ in the namespace, which we don't need.
                if '__builtins__' in namespace:
                    del namespace['__builtins__']

            args.append(arg)

        # Run it!
        #print 'FunctionBinding.run running: ', self.function.name
        result = self.function(*args)

        # Put any outputs back into the namespace.
        output_bindings = self.output_bindings

        #print 'FunctionBinding.run output_bindings: ', output_bindings
        if len(output_bindings) == 0:
            pass

        elif len(output_bindings) == 1:
            self._set_output(namespace, output_bindings[0].value, result)

        else:
            for index, binding in enumerate(output_bindings):
                self._set_output(namespace, binding.value, result[index])

        return result


    ###########################################################################
    # Private interface.
    ###########################################################################

    def _set_output(self, namespace, output_name, value):
        """ Sets a value in the output namespace. """

        # Is the variable name actually a reference to an attribute of
        # some instance in the namespace (eg. 'foo.bar', or 'foo.bar.baz'
        # etc).
        components = output_name.split('.')
        if len(components) > 1:
            # Find the object (using all but the last component).
            obj = eval('.'.join(components[:-1]), namespace)

            # Set the attribute.
            setattr(obj, components[-1], value)

        # Otherwise, the variable name is a simple one (eg. 'foo', or
        # 'bar', etc), so we make a new binding in the namespace.
        else:
            namespace[output_name] = value

        return

    def _get_function(self):

        return self._real_get_function()

    def _set_function(self, func):

        old = self._real_get_function()

        self._function = func
        self._function_name = func.name

        # notify any listeners
        self.trait_property_changed('function', old, func)

        return

    def _real_get_function(self):

        if self._function_name == '':
            if self._function is None:
                return None
            else:
                self._function_name = self._function.name

        if self._function is None:
            from enthought.component.registry import registry
            try:
                self._function = registry[self._function_name]
            except Exception, ex:
                self.enabled = False
                logger.error("Error loading script '%s', disabling in workflow" % self._function_name)
                return None


        return self._function

    ###########################################################################
    # 'pickle' interface.
    ###########################################################################
    def __getstate__(self):

        state = self.__dict__.copy()

        # don't pickle the actual component as this will keep projects from
        # being relocatable.
        if state.has_key('_function'):
            del state['_function']

        return state

    def __setstate__(self, state):

        self.__dict__.update(state)

        # we need to update here to make sure we pick up any new inputs
        # or outputs defined on the original script, but we want to
        # be sure we retain our original bindings
        if self.function is not None:
            inputs = self.input_bindings[:]
            outputs = self.output_bindings[:]
            self._function_changed()

            for input in inputs:
                try:
                    self.bind_input(input.name, input.value)
                except ValueError:
                    # if the input no longer exists then discard the
                    # value and continue
                    pass

            for output in outputs:
                try:
                    self.bind_output(output.name, output.value)
                except ValueError:
                    # if the output no longer exists then discard the
                    # value and continue
                    pass

        return

#### EOF ######################################################################
