""" A trait sheet for a function binding. """


# Major package imports.
import wx

# Enthought library imports.
from enthought.traits.trait_base import user_name_for
from enthought.traits.ui.api import Group, Item, View
from enthought.traits.ui.wx.editor import Editor
from enthought.traits.ui.wx.basic_editor_factory import BasicEditorFactory

class FunctionBindingTraitEditor(Editor):

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """

        self.control = wxFunctionBindingEditor(parent, self.value, self)
        return

    def dispose ( self ):
        """ Disposes of the contents of an editor.
        """
        self.control.dispose()

class wxFunctionBindingEditor(wx.Panel):
    """ A trait sheet for a function binding. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, parent, function_binding, editor, **kw):
        """ Create a new editor. """

        self._editors = []
        self._function_binding = function_binding
        self._editor = editor

        # Base-class constructor.
        wx.Panel.__init__(self, parent, -1, **kw)

        # Create the widget.
        self._create_widget(function_binding)

        return

    def GetValue(self):
        return self._function_binding

    def SetValue(self, new_value):
        self._function_binding = new_value
        
    ###########################################################################
    # Private interface.
    ###########################################################################
    
    def _create_widget(self, function_binding):
        """ Create the widget! """

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.SetAutoLayout(True)
        
        # Inputs.
        if len(function_binding.input_bindings) > 0:
            self._input_panel = panel = self._binding_list(
                self, function_binding.input_bindings, 'Inputs'
            )
            sizer.Add(panel, 0, wx.EXPAND)

        # Outputs.
        if len(function_binding.output_bindings) > 0:
            self._output_panel = panel = self._binding_list(
                self, function_binding.output_bindings, 'Outputs'
            )
            sizer.Add(panel, 1, wx.EXPAND)
        
        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(self)

        return

    def _binding_list(self, parent, binding_list, title):
        """ Creates editors for a list of bindings. """
         
        sizer = wx.FlexGridSizer(0, 2, 5, 5)
        sizer.AddGrowableCol(1)

        panel = wx.Panel(parent, -1)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        for binding in binding_list:
            if binding.visible:
                # Label.
                #
                # Get a 'user-friendly' version of the binding name (ie. nicely
                # capitalized, under_scores replaced with spaces etc.).
                name = user_name_for(binding.name)
                label = wx.StaticText(panel, -1, "%s:" % name) 
                sizer.Add(label, 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL)
            
                # Tooltip
                if binding.description == "":
                    msg = "No description available"
                else:
                    msg = binding.description
                    
                binding_view = View(Group(Item(name = 'value',
                                               tooltip = msg),
                                          show_labels = False))
        
                # Editor.
                editor = binding.edit_traits(kind = 'subpanel',
                                             parent = panel,
                                             view = binding_view)
                # we have to make sure the history objects are the same so
                # Cancel will work.
                editor.history = self._editor.ui.history
                self._editors.append(editor)
                sizer.Add(editor.control, 1, wx.EXPAND, wx.ALIGN_CENTER)

        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(panel)

        # Wrap a box around it.
        box = wx.StaticBox(parent, -1, title)
        box_sizer = wx.StaticBoxSizer(box, wx.HORIZONTAL)
        box_sizer.Add(panel, 1, wx.EXPAND | wx.ALL, 5)
        
        return box_sizer

    def dispose(self):

        if hasattr(self, '_editors') and self._editors is not None:
            for item in self._editors:
                item.dispose()

        return

class FunctionBindingEditor ( BasicEditorFactory ):

    klass = FunctionBindingTraitEditor
    
#### EOF ######################################################################
