""" A browser for function components. """

# Standard library imports.
import inspect

# Major package imports.
import wx

# Enthought library imports.
from enthought.component.registry import registry
from enthought.pyface.python_editor import PythonEditor
from enthought.pyface.grid.api import Grid, TraitGridColumn, TraitGridModel, \
                                  TraitGridSelection
from enthought.traits.api import Any, Color, HasTraits, Str
from enthought.traits.ui.wx.editor import Editor
from enthought.traits.ui.wx.basic_editor_factory import BasicEditorFactory

# Local imports.
#from component_registry_table_model import ComponentRegistryTableModel

class BrowserGridModel(TraitGridModel):

    alternate_color = Color('light blue')

    def get_cell_bg_color(self, row, col):
        """ Alternate colors of rows. """

        color = None
        if (row % 2) == 1:
            color = self.alternate_color

        return color

# Background color for selected grid rows.
SelectionBgColor = 0xFBD391

component_columns = [TraitGridColumn(name = 'name',
                                     label = 'Name',
                                     size = 150
                                     ),
                     TraitGridColumn(name = 'description',
                                     label = 'Description',
                                     size = 400
                                     ),
                     TraitGridColumn(name = 'id',
                                     label = 'Path',
                                     size = 575
                                     ),
                                 ]

io_columns =  [TraitGridColumn(name = 'name',
                               label = 'Name'),
               TraitGridColumn(name = 'description',
                               label = 'Description',
                               size = 150),
               TraitGridColumn(name = 'default_value',
                               label = 'Default Value'),
               TraitGridColumn(name = 'type',
                               label = 'Type'),
               TraitGridColumn(name = 'unit_family',
                               label = 'Measure Of'),
               TraitGridColumn(name = 'unit_system',
                               label = 'Unit System'),
               TraitGridColumn(name = 'visible',
                               label = 'Visible'),
               ]


class FunctionSelection(HasTraits):

    function = Any
    activated = Any

class FunctionBrowser(wx.Panel):
    """ A browser for function components. """


    # The first clause of the query used to return function components in
    # the component registry.
    ALL_FUNCTIONS = "isinstance(component, CallableComponent)"


    def __init__(self, parent, id, default=None, query=None, selection = None,
                 **kw):
        """ Create a new function browser. """

        # Base-class constructor.
        wx.Panel.__init__(self, parent, id, **kw)

        self.default = default
        self.selection = selection

        if query is not None:
            self.query = query
        else:
            self.query = ''

        # Create the widget!
        self._create_widget()

        return

    def select(self, function):
        """ Set the browser selection to 'function'."""

        self._function_table.set_selection([TraitGridSelection(obj =
                                                               function)])

        return

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    def _on_row_selected(self, row):
        """ Called when a row is selected. """

        # Get the selected function component.
        selection = self._function_table.get_selection()

        if len(selection) == 0:
            return

        function = selection[0].obj

        # fixme: This entire method is a quick hack for the demo!
        # Get the source code for the layer's algorithm.
        if hasattr(function, 'function'):
            text = inspect.getsource(function.function)

        else:
            text = function.code

        self._stc.SetReadOnly(False)
        self._stc.SetText(text)
        self._stc.SetReadOnly(True)

        # Display the inputs for the selected function.
        self._input_table.model.data = function.inputs
        self._input_table.set_selection([])

        # Display the outputs for the selected function.
        self._output_table.model.data = function.outputs
        self._output_table.set_selection([])

        # Trait event notification.
        self.selection.function = function

        return

    def _on_function_activated(self, pos):
        """ Called when a row is activated. """

        selection = self._function_table.get_selection()

        if len(selection) == 0:
            return

        function = selection[0].obj
        # Trait event notification.
        # fixme: this assumes the activated object is selected -- is
        #        that valid?
        self.selection.activated = function

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_widget(self):
        """ Create the widget! """

        self.sizer = sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.SetAutoLayout(True)

        # Query dropdown - display it only if a query was specified
        if self.query != '':
            label = wx.StaticText(self,  -1,  "Show")
            font = label.GetFont()
            font.SetWeight(wx.BOLD)
            label.SetFont(font)
            sizer.Add(label, 0, wx.LEFT, 7)
            self._query_choice = wx.Choice(self, -1)
            self._query_choice.Append("Queried")
            self._query_choice.Append("All")
            self._query_choice.SetSelection(0)
            wx.EVT_CHOICE(self, self._query_choice.GetId(),
                          self._on_query_choice_changed)
            sizer.Add(self._query_choice, 0, wx.ALL, 5)

        # Available functions.
        sizer.Add(self._create_function_table(), 1, wx.EXPAND | wx.ALL, 5)

        # Inputs/Outputs.
        sizer.Add(self._inputs_outputs(), 1, wx.EXPAND)

        # Code viewer.
        sizer.Add(self._code_viewer(), 2, wx.EXPAND | wx.ALL, 5)

        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(self)

        # If a default was specified (and it actually exists in the table!)
        # then select it.
        selection = None
        if self.default is not None:
            selection = [TraitGridSelection(obj = self.default)]
            self._function_table.set_selection(selection)
        elif len(self._components) > 0:
            # set the selection to the first component in the list
            selection = [TraitGridSelection(obj = self._components[0])]
            self._function_table.set_selection(selection)

        # make sure our selected value is REALLY selected
        if selection is not None:
            self.selection.function = selection[0].obj

        return

    def _update_function_table(self, query, model):
        # Table.
        if not query:
            real_query = self.ALL_FUNCTIONS

        else:
            real_query = self.ALL_FUNCTIONS + " and (%s)" % query

        # grab the components to display in this view
        self._components = components = registry.find(real_query)
        model.data = components
        return

    def _create_function_table(self):
        """ Create the available functions table. """

        sizer = wx.BoxSizer(wx.VERTICAL)

        # Label.
        label = wx.StaticText(self,  -1,  "Available Script Components")
        font = label.GetFont()
        font.SetWeight(wx.BOLD)
        label.SetFont(font)
        sizer.Add(label, 0, wx.LEFT, 2)

        # set up the available functions table
        self.model = BrowserGridModel(columns = component_columns)
        self._update_function_table(self.query, self.model)

        self._function_table = table = Grid(self,
                                            model = self.model,
                                            selection_mode = 'rows',
                                            selection_bg_color = \
                                                 SelectionBgColor,
                                            selection_text_color = 'black',
                                            read_only = True,
                                            show_row_headers = False,
                                            grid_line_color = 'grey',
                                            autosize = False)

        # Eliminate cell highlighting in wx grid widget.  Not exposed in
        # Grid interface!
        table._grid.SetCellHighlightPenWidth(0)

        sizer.Add(table.control, 1, wx.EXPAND | wx.TOP, 5)

        # Wire it up.
        table.on_trait_change(self._on_row_selected, 'selection_changed')
        table.on_trait_change(self._on_function_activated, 'cell_activated')

        return sizer

    def _inputs_outputs(self):
        """ Create the input/output tables. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(self._input_table(), 1, wx.EXPAND | wx.ALL, 5)
        sizer.Add(self._output_table(), 1, wx.EXPAND | wx.ALL, 5)

        return sizer

    def _input_table(self):
        """ Create the input table. """

        sizer = wx.BoxSizer(wx.VERTICAL)

        # Label.
        label = wx.StaticText(self,  -1,  "Inputs")
        font = label.GetFont()
        font.SetWeight(wx.BOLD)
        label.SetFont(font)
        sizer.Add(label, 0, wx.LEFT, 2)

        # Table.
        model = BrowserGridModel(columns = io_columns,
                                 data = [])

        self._input_table = table = Grid(self,
                                         model = model,
                                         selection_mode = 'rows',
                                         selection_bg_color = SelectionBgColor,
                                         selection_text_color = 'black',
                                         read_only = True,
                                         show_row_headers = False,
                                         grid_line_color = 'grey',
                                         autosize = False)

        # Eliminate cell highlighting in wx grid widget.  Not exposed in
        # Grid interface!
        table._grid.SetCellHighlightPenWidth(0)

        sizer.Add(table.control, 1, wx.EXPAND | wx.TOP, 5)

        return sizer

    def _output_table(self):
        """ Create the output table. """

        sizer = wx.BoxSizer(wx.VERTICAL)

        # Label.
        label = wx.StaticText(self,  -1,  "Outputs")
        font = label.GetFont()
        font.SetWeight(wx.BOLD)
        label.SetFont(font)
        sizer.Add(label, 0, wx.LEFT, 2)

        # Table.
        model = BrowserGridModel(columns = io_columns,
                                 data = [])

        self._output_table = table = Grid(self,
                                         model = model,
                                         selection_mode = 'rows',
                                         selection_bg_color = SelectionBgColor,
                                         selection_text_color = 'black',
                                         read_only = True,
                                         show_row_headers = False,
                                         grid_line_color = 'grey',
                                         autosize = False)

        # Eliminate cell highlighting in wx grid widget.  Not exposed in
        # Grid interface!
        table._grid.SetCellHighlightPenWidth(0)

        sizer.Add(table.control, 1, wx.EXPAND | wx.TOP, 5)

        return sizer

    def _code_viewer(self):
        """ Create the code viewer. """

        sizer = wx.BoxSizer(wx.VERTICAL)

        # Label.
        label = wx.StaticText(self,  -1,  "Code:")
        font = label.GetFont()
        font.SetWeight(wx.BOLD)
        label.SetFont(font)
        sizer.Add(label, 0, wx.LEFT, 2)

        self._stc = stc = PythonEditor(self).control
        stc.SetReadOnly(True)
        sizer.Add(stc, 1, wx.EXPAND | wx.TOP, 5)

        return sizer

    def _on_query_choice_changed(self, event):

        if self._query_choice.GetSelection() == 0:
            self._update_function_table(self.query, self.model)
        else:
            self._update_function_table('', self.model)

        return


class FunctionTraitEditor(Editor):

    scrollable = True

    def init(self, parent):

        self.control = FunctionBrowser(parent, -1, query = self.factory.query)

        return

    #--------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #--------------------------------------------------------------------------
    def update_editor ( self ):
        """ Updates the editor when the object trait changes external to the
            editor.
        """
        if self.value is not None:
            self.control.select(self.value)
        return

class FunctionEditor(BasicEditorFactory):

    klass = FunctionTraitEditor

    query = Str()

#### EOF ######################################################################
