""" A dialog containing a function browser. """


# Major package imports.
import wx

# Enthought Library imports
from enthought.envisage import get_application

# Local imports.
from function_browser import FunctionBrowser, FunctionSelection


class FunctionBrowserDialog(wx.Dialog):
    """ A dialog containing a function browser. """

    STYLE = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
    
    def __init__(self, parent, id, style=STYLE, default=None,
                 query=None, **kw):
        """ Create a new dialog. """

        # Base-class constructor.
        wx.Dialog.__init__(
            self, parent, id, "Component Library Browser", style=style, **kw
        )

        # The default function to display in the browser.
        self.default = default
        self._function = None
        
        # The query used to select function components (if no query is
        # specified, ALL function components will be returned).
        self.query = query
        
        # Set help ID
        self._help_id = "enlib|HID_Function_Browser_Dlg"
            
        # Create the widget!
        self._create_widget()
        
        return

    def get_function(self):
        return self._function

    ###########################################################################
    # Trait event handlers.
    ###########################################################################

    def _on_function_activated(self, function):
        """ Called when a function is activated (ie. double-clicked). """

        self._function = function
        self.EndModal(wx.ID_OK)
        
        return

    def _on_function_selected(self, function):
        """ Called when a function is selected. """

        self._function = function

        # Once a function has been selected, enable the 'OK' button and make
        # it the default.
        if hasattr(self, '_ok'):
            self._ok.Enable(True)
            self._ok.SetDefault()
        
        return
    
    ###########################################################################
    # Private interface.
    ###########################################################################
    
    def _create_widget(self):
        """ Create the widget! """

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.SetAutoLayout(True)

        self.selection = FunctionSelection(function = None)
        
        # Wire it up!
        self.selection.on_trait_change(self._on_function_activated,
                                       'activated')
        self.selection.on_trait_change(self._on_function_selected, 'function')

        # The function browser.
        self._fb = fb = FunctionBrowser(self, -1,
                                        default = self.default,
                                        query = self.query,
                                        selection = self.selection)
        sizer.Add(fb, 1, wx.EXPAND)

        # The buttons.
        buttons = self._create_buttons(self)
        sizer.AddSizer(buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        # Resize the dialog to match the sizer's minimal size.
        sizer.Fit(self)
        
        return

    def _create_buttons(self, parent):
        """ Create the buttons. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # Until a function has been selected, the 'OK' button is disabled, and
        # the 'Cancel' button is the default.
        self._ok = ok = wx.Button(self, wx.ID_OK, "OK")
        #ok.Disable()
        sizer.Add(ok, 0, wx.ALL, 5)

        cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")
        cancel.SetDefault()
        sizer.Add(cancel, 0, wx.ALL, 5)
        
        helpbtn = wx.Button(self, wx.ID_HELP, "Help")
        sizer.Add(helpbtn, 0, wx.ALL, 5)
        # Wire it up.
        wx.EVT_BUTTON(self, wx.ID_HELP, self._on_help)

        return sizer
        
    def _on_help(self, event):
        """Called when the 'Help' button is pressed. """
        hp = get_application().get_service('enthought.help.IHelp')
        hp.library.show_topic(self._help_id)
        return

#### EOF ######################################################################
