""" A function component. """


# Enthought library imports.
from enthought.traits.api import Any, Property

# Local imports.
from enthought.component.callable_component import CallableComponent
from enthought.component.registry import registry


class FunctionComponent(CallableComponent):
    """ A function component. """

    function = Any
    
    ###########################################################################
    # 'object' interface.
    ###########################################################################
    
    def __init__(self, **traits):
        """ Creates a new function component. """

        # Default values for trait properties.
        self._id = ''
        self._name = ''
        self._description = ''

        # Base-class constructor.
        super(FunctionComponent, self).__init__(**traits)

        # Add the component to the registry.
        self.name = registry.create_unique_name(self.name)
        registry.add_component(self.id, self)
        
        return

    # fixme: We only pass POSITIONAL arguments to the component, the keyword
    # arguments are used to control the type coercion process (this is not as
    # bad as it sounds, as there is currently no way for a component to specify
    # that it takes keyword arguments!).
    def __call__(self, *args, **kw):
        """ Calls the component like a regular Python function. """

        values = []
        for input, value in zip(self.inputs, args):
            # Do any required unit conversion/type coercion.
            values.append(self._prepare_argument(input, value, **kw))

        return self.function(*values)

    ###########################################################################
    # 'Component' interface.
    ###########################################################################

    def _get_id(self):
        """ Returns the component's unique identifier. """

        return self.name

    def _set_id(self, value):
        """ Sets the component's unique identifier. """

        old = self.name
        self.name = value

        self.trait_property_changed('id', old, value)

        return

    id = Property(_get_id, _set_id)

    def _get_name(self):
        """ Returns the component's name. """

        if len(self._name) > 0:
            name = self._name

        elif self.function is not None:
            name = self.function.__name__

        else:
            name = ''

        return name

    def _set_name(self, name):
        """ Sets the component's name. """

        old = self._name
        self._name = name

        self.trait_property_changed('name', old, name)

        return

    name = Property(_get_name, _set_name)

    def _get_description(self):
        """ Returns the component's description. """

        if len(self._description) > 0:
            description = self._description

        elif self.function is not None and self.function.__doc__ is not None:
            description = self.function.__doc__

        else:
            description = ''

        return description

    def _set_description(self, description):
        """ Sets the component's description. """

        old = self._description
        self._description = description

        self.trait_property_changed('description', old, description)

        return

    description = Property(_get_description, _set_description)

    ###########################################################################
    # 'CallableComponent' interface.
    ###########################################################################

    def invoke(self, namespace):
        """ Invokes (runs) the component in the specified namespace. """

        # Extract the input arguments from the namespace.
        args = []
        for name in self.input_names:
            args.append(namespace[name])

        # Run it!
        result = self.function(*args)

        # Put any outputs back into the namespace.
        output_names = self.output_names

        if len(output_names) == 0:
            pass
        
        elif len(output_names) == 1:
            namespace[self.output_names[0]] = result

        else:
            for index, name in enumerate(self.output_names):
                namespace[name] = result[index]
 
        # Return the namespace to allow for nested component invocations.
        #
        # e.g. foo.invoke(bar.invoke(baz.invoke(namespace)))
        return namespace
    
#### EOF ######################################################################
