""" Inputs to callable components (ie. functions, scripts etc). """

# Enthought library imports.
from enthought.traits.api import CBool, CStr, HasTraits, Instance, List, \
                             Property, Str
from enthought.traits.ui.api import Group, InstanceEditor, Item, View
from enthought.units.unit_traits import unit_family_trait, unit_system_trait

# Local imports.
from enthought.component.common_type_defs import any_type
from enthought.component.type_def import TypeDef
from enthought.component.type_def_manager import type_def_manager
from enthought.units.unit_manager import unit_manager

class Input(HasTraits):
    """ Inputs to callable components (ie. functions, scripts etc). """

    # The name of the input.
    name = Str()

    # A description of the input.
    description = Str()

    # N.B. The 'default_value' trait is generated dynamically as a property
    default_value = CStr()
        
    # The type of the input.
    type = Instance(TypeDef,
                    editor = \
                    InstanceEditor(name = '_available_types',
                                   editable = False))
    
    # The unit family name.
    unit_family = unit_family_trait

    # The unit system.
    unit_system = unit_system_trait
        
    # String used to determine what type of GUI editor to use.
    meta_type = Str()
        
    # Allow the script writer to hide this input from the end user
    visible = CBool(True)

    view = View(Group(Item(name = 'name'),
                      Item(name = 'description'),
                      Item(name = 'default_value'),
                      Item(name = 'type'),
                      Item(name = 'unit_family', label='Measure of'),
                      Item(name = 'unit_system'),
                      Item(name = 'visible')),
                scrollable = True)

    # Protected interface
    _available_types = Property()

    ###########################################################################
    # 'object' interface.
    ###########################################################################
    
    def __init__(self, name = '', type = None, **kw_args):
        self.name = name
        self.type = type
        super(Input, self).__init__(**kw_args)
        return
        
    def __getstate__(self):
        """ Get the state of the object for pickling. """
        
        state = {}
        state['name'] = self.name
        state['description'] = self.description
        state['default_value'] = self.default_value
        if self.type is not None:
            state['type'] = self.type.id
        else:
            state['type'] = None
        state['meta_type'] = self.meta_type
        state['unit_system'] = str(self.unit_system)
        state['unit_family'] = self.unit_family
        state['visible'] = str(self.visible)
            
        return state

    def __setstate__(self, state):
        """ Set the state of the object for unpickling. """

        self.__dict__['name'] = state['name']
        self.__dict__['description'] = state['description']     
        self.__dict__['meta_type'] = state['meta_type']

        # Handles older script components when we used 'type_id' instead
        # of 'type' in the XML.
        if state.has_key('type_id'):
            type_id = state['type_id']
        else:
            type_id = state['type']
            
        self.__dict__['type'] = type_def_manager.lookup(type_id)
        
        # Handles the case of legacy objects missing a trait.
        if state.has_key('unit_system'):
            unit_system = unit_manager.lookup_system(state['unit_system'])
            self.__dict__['unit_system'] = unit_system
            self.__dict__['unit_family'] = state['unit_family']
        
        # Handles the case of legacy objects missing a trait.
        if state.has_key('visible'):
            self.__dict__['visible'] = eval(state['visible'])
        else:
            self.__dict__['visible'] = True
            
        # Handles the case of legacy objects missing a trait.
        if state.has_key('default_value'):
            self.__dict__['default_value'] = state['default_value']   
        else:
            self.__dict__['default_value'] = state['name']   
        
        return
        
    def __str__(self):
        return  'Input: name=%s desc=%s default=%s type=%s system=%s family=%s' \
        % (self.name, self.description, self.default_value, self.type, self.unit_system, self.unit_family)

    def _get__available_types(self):

        from type_def_manager import type_def_manager

        return type_def_manager.type_defs

# A list of inputs.
input_list_trait = List(Input)
        
#### EOF ######################################################################
