""" The component registry. """

# Standard library imports.
import logging
import string

# Enthought library imports.
from enthought.traits.api import HasTraits, Property


# Setup a logger for this module
logger = logging.getLogger(__name__)


class Registry(HasTraits):
    """ The component registry. """

    # components = Property(_get_components)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self):
        """ Creates a new component registry. """

        # All of the components in the registry.
        self._components = {}

        # A dictionary containing the classes of all components added to the
        # registry.
        #
        # This dictionary is added to the namespace for query evaluation.
        self._classes = {} # { string class_name : class klass }

        return

    ###########################################################################
    # 'dict' interface.
    ###########################################################################

    # fixme: Long term I'm not so sure that this is a good idea, but for now
    # this is shorthand to lookup a component by NAME (not its unique key).
    def __getitem__(self, name):
        """ Returns the component with the specified name. """

        results = self.find("component.name == '%s'" % name)
        if len(results) == 0:
            raise KeyError("No component found with name %s" % name)

        elif len(results) > 1:
            raise KeyError("Duplicate components found for %s" % name)

        return results[0]

    ###########################################################################
    # 'Registry' interface.
    ###########################################################################

    def create_unique_name(self, name):
        """ Find a unique name based on "name" that isn't already in the
            registry.
        """

        # Find the base name for the component.  If the string is empty, put
        # a space in it so base_name[-1] will work.
        if name == "":
            base_name = " "
        else:
            base_name = name

        # strip off any trailing characters
        while base_name[-1] in string.digits:
            base_name = base_name[:-1]

        base_name = base_name.strip()

        results = self.find("component.name == '%s'" % name)
        count = 0
        while results:
            count += 1
            name = base_name + str(count)
            results = self.find("component.name == '%s'" % name)

        return name


    def _get_components(self):
        """ Returns all components in the registry. """

        return self._components.values()[:]

    components = Property(_get_components)

    def add_component(self, key, component):
        """ Adds a component to the registry. """

        if component.name != self.create_unique_name(component.name):
            msg = "'%s' is not a unique name in the registry." \
                  " call registry.create_unique_name(name)" \
                  "first." % component.name
            raise ValueError, msg

        self._components[key] = component

        # Recursively add the components classes and base classes to our
        # class dictionary.
        self._add_class(component.__class__)

        return

    def remove_component(self, key):
        """ Removes a component from the registry. """

        try:
            del self._components[key]

        except KeyError:
            pass

        return

    def find(self, query):
        """ Finds all components in the registry that match the query. """

        if query is None:
            matches = self.components

        else:
            glob = globals()

            matches = []
            for component in self._components.values():
                namespace = {}
                namespace.update(self._classes)
                namespace.update(component.metadata)
                namespace.update(component.__dict__)
                # note that we have to do this because the names on
                # function components are properties, which don't appear
                # in the __dict__
                namespace['name'] = component.name
                namespace['component'] = component
                try:
                    if eval(query, glob, namespace):
                        matches.append(component)

                except:
                    #import traceback; traceback.print_exc()
                    #print component, namespace
                    pass

        # Sort the results by name.
        matches.sort(lambda x, y: cmp(x.name, y.name))

        return matches

    def find_script_component(self, script):
        """ Returns the script component associated with this script. """

        query = 'path == r"%s"' % script.path

        result = self.find(query)
        if len(result) > 0:
            script_component = result[0]
        else:
            # Proava2 still works ok in spite of this error so bob removed it
            #logger.error('No match found for ' + query)
            script_component = None

        if script_component is None:
            from enthought.component.script_component_factory \
                 import ScriptComponentFactory
            factory = ScriptComponentFactory()
            script_component = factory.from_file(script.path)

        return script_component

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _add_class(self, klass):
        """ Recursively add klass and its bases to the class dictionary. """

        self._classes[klass.__name__] = klass

        for base in klass.__bases__:
            self._add_class(base)

        return

# The (as yet unenforced) singleton instance.
registry = Registry()

#### EOF ######################################################################
