""" A script component. """


# Standard/built-in imports.
import logging
import os
import sys

# Enthought library imports.
from enthought.traits.api import Button, Dict, HasTraits, Instance, List, \
                             Property, Str, Trait
from enthought.traits.ui.api import CodeEditor, Group, InstanceEditor, Item, \
                                TableEditor, View, Handler
from enthought.traits.ui.key_bindings import KeyBinding, KeyBindings
from enthought.traits.ui.menu import Action, CancelButton
from enthought.traits.ui.table_column import ObjectColumn
from enthought.traits.ui.extras.checkbox_column import CheckboxColumn
from enthought.traits.ui.help import on_help_call

# Local imports.
from callable_component import CallableComponent
from input import Input
from output import Output
from registry import registry
from exception import AbortSerialWorkflow, InputNotInNamespace


# Setup a logger for this module
logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
#  Define a KeyBindings object:
#------------------------------------------------------------------------------

key_bindings = KeyBindings(

    KeyBinding( binding1    = 'Ctrl-s',
                description = 'Save to a file',
                method_name = '_save_clicked' ),

    KeyBinding( binding1    = 'Ctrl-r',
                description = 'Run script',
                method_name = '_save_and_run_clicked' ),
)

input_columns = [ObjectColumn(name = 'name'),
                 ObjectColumn(name = 'description'),
                 ObjectColumn(name = 'default_value'),
                 ObjectColumn(name = 'type',
                              editor = \
                                  InstanceEditor(name = '_available_types',
                                                 editable = False)),
                 ObjectColumn(name = 'unit_family', label='Measure Of'),
                 ObjectColumn(name = 'unit_system'),
                 CheckboxColumn(name = 'visible'),
                 ]

output_columns = [ObjectColumn(name = 'name'),
                  ObjectColumn(name = 'description'),
                  ObjectColumn(name = 'type',
                               editor = \
                                  InstanceEditor(name = '_available_types',
                                                 editable = False)),
                  ObjectColumn(name = 'unit_family', label='Measure Of'),
                  ObjectColumn(name = 'unit_system'),
                  CheckboxColumn(name = 'visible'),
                  ]

input_view = View(Group(Item(name = 'name'),
                        Item(name = 'description'),
                        Item(name = 'default_value'),
                        Item(name = 'type'),
                        Item(name = 'unit_family', label='Measure of'),
                        Item(name = 'unit_system'),
                        Item(name = 'visible')))

output_view = View(Group(Item(name = 'name'),
                         Item(name = 'description'),
                         Item(name = 'type'),
                         Item(name = 'unit_family', label='Measure of'),
                         Item(name = 'unit_system'),
                         Item(name = 'visible')))

input_editor = TableEditor(columns = input_columns,
                           row_factory = Input,
                           deletable = True,
                           reorderable = True,
                           auto_size = False,
                           edit_view = input_view,
                           selection_bg_color = 0xFBD391,
                           selection_color = 'black')

output_editor = TableEditor(columns = output_columns,
                            row_factory = Output,
                            deletable = True,
                            reorderable = True,
                            auto_size = False,
                            edit_view = output_view,
                            selection_bg_color = 0xFBD391,
                            selection_color = 'black')


class ScriptComponentHandler(Handler):

    def object_path_changed(self, info):

        info.ui.title = 'Edit \'' + info.object.path + '\''
        return

    def _save_clicked(self, info):
        info.object._save_changed()
        return

    def _refresh_clicked(self, info):
        info.object.refresh()
        return

    def _save_and_run_clicked(self, info):
        info.object._save_and_run_changed()
        return

    def _cancel_clicked(self, info):
        info.object.refresh()
        info.ui.dispose(result = False)
        return

    def _help_clicked(self, info):
        hh = on_help_call()
        hh(info, None)

    def close(self, info, is_ok):
        info.object.refresh()
        return True

class ScriptComponent(CallableComponent):
    """ A script component. """

    # The text that marks the start of the XML component description.
    START_MAGIC = "#### <enthought version='1.0'> ####\n"
    END_MAGIC   = "#### </enthought> ####\n"


    # The path to the script that we are componentizing.
    path = Str()

    # The component id.
    id = Property

    # The component's code.
    code = Property

    # A list version of our metadata
    _metadata_list = Property

    # Define the View to use:
    traits_view = \
       View(Group(
              Group(
                Group(
                  Item('code',
                       editor = CodeEditor( key_bindings = key_bindings ),
                       width = 720,
                       has_focus = True),
                  label       = 'Code',
                  export      = 'DockShellWindow',
                  show_labels = False,
                  ),
                Group(
                  Group(
                    Item('inputs',
                         style = 'custom',
                         id    = 'inputs_element',
                         editor = input_editor,
                         height = 200
                         ),
                    show_labels = False,
                    export      = 'DockShellWindow',
                    orientation = 'vertical',
                    label = 'Inputs',
                    ),
                  Group(
                    Item('outputs',
                         style = 'custom',
                         id    = 'outputs_element',
                         editor = output_editor,
                         height = 200,
                         ),
                    show_labels = False,
                    label = 'Outputs',
                    export      = 'DockShellWindow',
                    orientation = 'vertical',
                    ),
                  Group(
                    Item('name'),
                    Item('description'),
                    Group(
                      Item('_metadata_list',
                           editor = InstanceEditor(),
                           style = 'custom'),
                      show_labels = False,
                      label = 'Metadata',
                      show_border = True,
                      ),
                    label       = 'General',
                    export      = 'DockShellWindow',
                    ),
                  layout = 'tabbed'
                  ),
                id          = 'splitter1',
                orientation = 'vertical',
                layout      = 'split'
            ),
            ),
        id        = 'enthought.component.script_component.editor',
        kind      = 'live',
        dock      = 'tab',
        buttons   = [Action(name = 'Save'),
                     Action(name = 'Save and Run',
                            action = '_save_and_run_clicked'),
                     Action(name = 'Refresh'),
                     Action(name = 'Help  ', action = '_help_clicked'),
                     Action(name = 'Close ', action = '_cancel_clicked'),
                     ],
        resizable = True,
        handler   = ScriptComponentHandler(),
        help_id   = 'enlib|HID_Script_Editing_Wnd',
    )


    # Protected interface
    _available_types = Property()

    # fixme: We only pass POSITIONAL arguments to the component, the keyword
    # arguments are used to control the type coercion process (this is not as
    # bad as it sounds, as there is currently no way for a component to specify
    # that it takes keyword arguments!).
    def __call__(self, *args, **kw):
        """ Calls the component like a regular Python function. """

        # Create a namespace for the script to execute in using the positional
        # arguments.
        execution_namespace = {}

        try:
            # Add the positional arguments...
            for i, value in enumerate(args):
                input = self.inputs[i]

                # Do any required unit conversion/type coercion.
                value = self._prepare_argument(input, value, **kw)

                # Add it to the namespace.
                execution_namespace[input.name] = value

                # Run the script in the namespace.

            exec self.code in execution_namespace
            # Get any outputs from the namespace and return them.
            output_names = self.output_names

            if len(output_names) == 0:
                result = None

            elif len(output_names) == 1:
                result = execution_namespace[output_names[0]]

            else:
                result = []
                for name in output_names:
                    result.append(execution_namespace[name])

                result = tuple(result)

            execution_namespace.clear()

        # intercept a component legitimately cancelling the workflow execution
        except AbortSerialWorkflow, msg:
            raise AbortSerialWorkflow, msg

        except Exception, msg:
            logger.error('Received exception while executing script ' + \
                         self.path)
            type, value, traceback = sys.exc_info()
            logger.exception(msg)
            logger.exception(traceback)
            execution_namespace.clear()
            raise

        return result

    ###########################################################################
    # 'Component' interface.
    ###########################################################################

    def _get_id(self):
        """ Returns the component's unique identifier. """

        return self.path

    def _set_id(self, value):
        """ Sets the component's unique identifier. """

        old = self.path
        self.path = value

        self.trait_property_changed('id', old, value)

        return

    ###########################################################################
    # 'CallableComponent' interface.
    ###########################################################################

    def invoke(self, namespace):
        """ Invokes (runs) the component in the specified namespace. """

        # Extract the inputs (arguments) from the namespace.
        execution_namespace = {}
        for name in self.input_names:
            if not namespace.has_key(name):
                raise InputNotInNamespace(name)
            execution_namespace[name] = namespace[name]

        # Run it!
        exec self.code in execution_namespace

        # Put any outputs back into the namespace.
        for name in self.output_names:
            namespace[name] = execution_namespace[name]

        # Return the namespace to allow for nested component invocations.
        #
        # e.g. foo.invoke(bar.invoke(baz.invoke(namespace)))
        return namespace

    ###########################################################################
    # 'ScriptComponent' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_code(self):
        """ Returns the contents of the script. """

        # this is stupid, but we have to do it because of the way trait
        # properties work
        return self._really_get_code()

    def _set_code(self, code):
        """ Sets the contents of the script. """

        old_value = self._really_get_code()
        self._code = code

        self.trait_property_changed('code', old_value, code)

        return

    def _get__metadata_list(self):
        return NameValueList(component = self)

    def _really_get_code(self):

        if hasattr(self, '_code') and self._code is not None:
            code = self._code
        else:
            code = self._load_code()
        return code

    #### Methods ##############################################################

    def unregister(self):
        """ Removes this component from the registry. """

        registry.remove_component(self.id)

        return

    def refresh(self):
        """ Refreshes the code of this component from the file system. """
        # we'll force the code to refresh next time it's loaded
        old_code = self.code
        self._code = None
        self.trait_property_changed('code', old_code, self.code)
        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _load_code(self):
        if os.path.exists(self.path):

            # 'rU' means open file for read w/ universal newline support
            # for handling cross-platform line endings
            code = file(self.path, 'rU').read()

            end_magic = ScriptComponent.END_MAGIC
            end = code.find(end_magic)
            if end != -1:
                code = code[end + len(end_magic):]

        else:
            code = ''

        return code

    def _save(self, code):
        """ Writes the specified code to the script. """

        from script_component_factory import ScriptComponentFactory
        ScriptComponentFactory.to_file(self, code)

        logger.debug("script %s saved" % self.path)

        return

    #### Static trait handlers ################################################

    def _path_changed(self, old, new):
        """ Called when the path of the script component is changed. """

        self.id = new

        return

    def _id_changed(self, old, new):
        """ Called when the Id of the script component is changed. """

        # When our Id is changed we have to update the component registry.
        if old is not None and len(old) > 0:
            registry.remove_component(old)

        if new is not None and len(new) > 0:
            self.name = registry.create_unique_name(self.name)
            registry.add_component(new, self)

        return

    def _save_changed(self):

        self._save(self.code)
        return

    def _save_and_run_changed(self):

        self._save(self.code)

        from enthought.envisage.ui.python_shell import PythonShellPlugin
        plugin = PythonShellPlugin.instance
        # if the shell plugin is available we run the script in the shell
        if plugin is not None:
            plugin.execute_command('execfile(r"%s")' % self.path)
        else:
            # otherwise we directly invoke the script and let it do its thing
            self()

        return

    def _get__available_types(self):

        from type_def_manager import type_def_manager

        return type_def_manager.type_defs


class NameValue(HasTraits):

    name = Str()
    value = Str()

def nvcmp(a, b):
    return cmp(a.name, b.name)

metadata_columns = [ObjectColumn(name = 'name'),
                    ObjectColumn(name = 'value'),
                   ]

metadata_editor = TableEditor(columns = metadata_columns,
                              row_factory = NameValue,
                              deletable = True,
                              selection_bg_color = 0xFBD391,
                              selection_color = 'black')

# NameValueList is a model class used to edit the metadata dictionary.
# It makes the dictionary look like a list, and manages the changes
# in the editor and responds to external changes.
class NameValueList(HasTraits):

    nvlist = List(NameValue)

    component = Instance(ScriptComponent)

    view = View(Group(Item('nvlist',
                           editor = metadata_editor),
                      show_labels = False
                      )
                )

    def __init__(self, **traits):

        # base-class constructor
        super(NameValueList, self).__init__(**traits)

        self._updating = False
        self._initalizing = False
        # we need to build our list from the specified dictionary
        self._initializing = True
        self.nvlist = self._build_list_from_dict(self.component.metadata)
        self._initializing = False

        return

    def _build_list_from_dict(self, dict):

        nvlist = []
        for item in self.component.metadata.items():
            nv = NameValue(name = item[0],
                           value = item[1])
            nv.on_trait_change(self._name_changed, 'name')
            nv.on_trait_change(self._value_changed, 'value')
            nvlist.append(nv)

        nvlist.sort(nvcmp)
        return nvlist

    def _nvlist_items_changed(self, event):

        if self._initializing:
            return

        self._updating = True

        for pair in event.added:
            pair.on_trait_change(self._name_changed, 'name')
            pair.on_trait_change(self._value_changed, 'value')
            self.component.metadata[pair.name] = pair.value

        for pair in event.removed:
            pair.on_trait_change(self._name_changed, 'name', remove = True)
            pair.on_trait_change(self._value_changed, 'value',
                               remove = True)
            if pair.name in self.component.metadata:
                del self.component.metadata[pair.name]

        self._updating = False

        return

    def _dict_changed(self):
        if not self._updating:
            # make sure we remove the listeners on the old nv objects
            # before rebuilding the list
            for nv in self.nvlist:
                nv.on_trait_change(self._name_changed, 'name', remove = True)
                nv.on_trait_change(self._value_changed, 'value',
                                   remove = True)

            self.nvlist = self._build_list_from_dict(self.component.metadata)
        return

    def _value_changed(self, obj, trait, old, new):

        self.component.metadata[obj.name] = obj.value
        return

    def _name_changed(self, obj, trait, old, new):

        if old in self.component.metadata:
            del self.component.metadata[old]

        self.component.metadata[obj.name] = obj.value
        return


#### EOF ######################################################################
