""" A function component editor. """


# Major package imports.
import wx

# Enthought library imports.
from enthought.util.wx.image import get_bitmap
   
    
class ScriptComponentEditor(wx.Panel):
    """ A script component editor. """

    
    # Default style.
    STYLE = wx.CLIP_CHILDREN


    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, parent, wxid, component, handler=None,style=STYLE,**kw):
        """ Creates a new editor. """

        # The script component that we are editing.
        self.component = component
        
        # Images used on the show/hide button (they should be the same width!).
        self.HIDE_IMAGE = get_bitmap(self, 'image/less.png')
        self.SHOW_IMAGE = get_bitmap(self, 'image/more.png')

        # Base-class constructor.
        wx.Panel.__init__(self, parent, wxid, style=style, **kw)

        # Create the widget!
        self._create_widget(component, handler)

        return

    ###########################################################################
    # 'EditorPanel' interface.
    ###########################################################################

    def _create_widget(self, component, handler):
        """ Creates the widget! """

        self.sizer = sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self.SetAutoLayout(True)

        # Splitter.
        self._splitter = splitter = wx.SplitterWindow(
            self, -1, style=wx.SP_NOBORDER | wx.SP_3DSASH | wx.CLIP_CHILDREN
        )
        sizer.Add(splitter, 1, wx.EXPAND)
        
        # Code editor.
        code = self._code(splitter, component)

        # Details notebook.
        details = self._details(splitter, component)
        
        # We set the splitter position when the window is rendered (ie. on
        # a size event).
        splitter.SplitVertically(code, details, 0)
        wx.EVT_SIZE(splitter, self._on_size)

        # The 'Save' and 'Save And Run' button.
        buttons = self._buttons(self)
        sizer.Add(buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        
        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(self)

        return

    ###########################################################################
    # wx event handlers.
    ###########################################################################

    def _on_size(self, event):
        """ Called when the splitter window is sized/re-sized. """

        # If the detail panel is hidden then we don't want the panel containing
        # the show/hide button to expand.
        if self._hidden:
            width, height = self._splitter.GetSize()

            # fixme: How can we calculate this?!?
            fudge = 7
            button_width = self.SHOW_IMAGE.GetWidth() + fudge
            
            self._splitter.SetSashPosition(width - button_width)

        # Allow the default event handler to do its thing.
        event.Skip()

        return

    def _on_button(self, event):
        """ Called when the show/hide button is pressed. """

        if self._hidden:
            self._show()
            
        else:
            self._hide()

        return

    def _on_save(self, event):
        """ Called when the 'Save' button is pressed. """

        # fixme: Calling a private method here!
        self._code_editor._save()

        return

    def _on_save_and_run(self, event):
        """ Called when the 'Save and Run' button is pressed. """

        # fixme: Calling a private method here!
        self._code_editor._save_and_run()

        return
    
    ###########################################################################
    # Private interface.
    ###########################################################################

    def _code(self, parent, component):
        """ Creates the code editor. """

        panel = wx.Panel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)
        
        # Code editor.
        self._code_editor = component.simple_trait_editor(panel, 'code')
        # fixme: Hack used inside charm plugin!
        self._stc = self._code_editor._stc
        sizer.Add(self._code_editor, 1, wx.EXPAND)

        from enthought.envisage.plugins.charm.editor_file import EditorFile
        self._code_editor._stc.SetFile(EditorFile(path=component.path))
        
        return panel
    
    def _details(self, parent, component):
        """ Creates the details notebook. """

        # Is the detail panel hidden?
        self._hidden = True

        panel = wx.Panel(parent, -1)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # The show/hide button.
        self._button = button = wx.BitmapButton(
            panel, -1, self.SHOW_IMAGE, style=0
        )
        wx.EVT_BUTTON(button, button.GetId(), self._on_button)
        sizer.Add(button, 0, wx.ALIGN_CENTER_VERTICAL)

        # The details notebook.
        self._notebook = notebook = self._notebook(panel, component)
        sizer.Add(notebook, 1, wx.EXPAND)
        self._notebook.Show(False)

        # Default body width for show/hide.
        #
        # fixme: Can we calculate this?
        self._width = 250
        
        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(panel)

        return panel

    def _hide(self):
        """ Hide the detail panel. """

        # Remember the width of the detail panel.
        self._width, height = self._notebook.GetSize()
        self._notebook.Show(False)

        # Move the sash to take up the space created by hiding the detail
        # panel.
        sash_position = self._splitter.GetSashPosition()
        self._splitter.SetSashPosition(sash_position + self._width)

        # Display the 'show' image on the show/hide button.
        self._button.SetBitmapLabel(self.SHOW_IMAGE)

        # The detail panel is hidden!
        self._hidden = True

        return

    def _show(self):
        """ Show the detail panel. """
        
        # Show the detail panel.
        self._notebook.Show(True)

        # Move the sash to make the detail panel fully visible.
        sash_position = self._splitter.GetSashPosition()
        self._splitter.SetSashPosition(sash_position - self._width)
            
        # Display the 'hide' image on the show/hide button.
        self._button.SetBitmapLabel(self.HIDE_IMAGE)

        # The detail panel is NOT hidden!
        self._hidden = False

        return
    
    def _notebook(self, parent, component):
        """ Creates the details notebook. """

        notebook = wx.Notebook(parent, -1, style=wx.CLIP_CHILDREN)

        # Create the inputs/outputs page.
        inputs_outputs = self._inputs_outputs(notebook, component)
        notebook.AddPage(inputs_outputs, "Inputs/Outputs")

        # Create the general page.
        general = self._general(notebook, component)
        notebook.AddPage(general, "General")
        
        return notebook
    
    def _inputs_outputs(self, parent, component):
        """ Creates the input/output tables. """

        panel = wx.Panel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # Inputs.
        label = wx.StaticText(panel,  -1,  "Inputs")
        sizer.Add(label, 0, wx.LEFT, 5)
        
        sizer.Add(
            component.simple_trait_editor(panel, 'inputs'),
            1,
            wx.EXPAND | wx.LEFT,
            5
        )

        # Outputs.
        label = wx.StaticText(panel,  -1,  "Outputs")
        sizer.Add(label, 0, wx.TOP | wx.LEFT, 5)

        sizer.Add(
            component.simple_trait_editor(panel, 'outputs'),
            1,
            wx.EXPAND | wx.LEFT,
            5
        )

        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(panel)

        return panel

    def _general(self, parent, component):
        """ Creates the general section (Name etc). """

        panel = wx.Panel(parent, -1)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # General fields (name etc).
        sizer.Add(
            self._fields(panel, component), 0, wx.EXPAND | wx.LEFT, 5
        )

        # Metadata.
        label = wx.StaticText(panel,  -1,  "Metadata")
        sizer.Add(label, 0, wx.TOP | wx.LEFT, 5)
        
        sizer.Add(
            component.simple_trait_editor(panel, 'metadata'),
            1,
            wx.EXPAND | wx.LEFT,
            5
        )

        # Resize the panel to match the sizer's minimal size.
        sizer.Fit(panel)

        return panel
    
    def _fields(self, parent, component):
        """ Creates the fields section (Name etc). """

        sizer = wx.FlexGridSizer(0, 2, 5, 5)
        sizer.AddGrowableCol(1)

        # Name.
        label = wx.StaticText(parent,  -1,  "Name")
        sizer.Add(label, 0)

        text = component.text_trait_editor(parent, 'name')
        sizer.Add(text, 1, wx.EXPAND | wx.LEFT, 5)

        # Description.
        label = wx.StaticText(parent,  -1,  "Description")
        sizer.Add(label, 0)

        text = component.text_trait_editor(parent, 'description')
        sizer.Add(text, 1, wx.EXPAND | wx.LEFT, 5)
        
        return sizer

    def _buttons(self, parent):
        """ Creates the buttons. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Save' button.
        save = wx.Button(parent, -1, "Save")
        save.SetDefault()
        wx.EVT_BUTTON(parent, save.GetId(), self._on_save)
        sizer.Add(save)

        # 'Save and Run' button.
        save_and_run = wx.Button(parent, -1, "Save and Run")
        wx.EVT_BUTTON(parent, save_and_run.GetId(), self._on_save_and_run)
        sizer.Add(save_and_run, 0, wx.LEFT, 10)

        return sizer
    
#### EOF ######################################################################
