""" Components. """


# Standard/built-in imports.
import new, os
from xml.dom.minidom import parseString

# Local imports.
from enthought.component.script_component import ScriptComponent
from enthought.component.input import Input
from enthought.component.output import Output


def object_to_xml(f, indent, obj):
    """ Writes the XML representation of an object to a file. """
    
    element_name = obj.__class__.__name__.lower()
    f.write('%s<%s ' % (indent, element_name))

    state = obj.__getstate__()
    for name, value in state.items():
        f.write("%s='%s' " % (name, value))

    f.write('/>\n')
        
    return
    
def object_from_xml(node, klass):
    """ Writes the XML representation of an object to a file. """
    
    state = {}

    for name in node.attributes.keys():
        state[name] = str(node.attributes[name].value)
        
    obj = klass.__new__(klass)
    obj.__setstate__(state)
    
    return obj


class ScriptComponentFactory:
    """ A factory for script components. """

    def from_file(self, path):
        """ Create a script component from the file at the specified path. """
        
        magic = ScriptComponent.START_MAGIC
        magic_len = len(magic)
        end_magic = ScriptComponent.END_MAGIC

        try:
            # 'rU' means open file for read w/ universal newline support
            # for handling cross-platform line endings
            text = file(path, 'rU').read()
            if len(text) >= magic_len and text[:magic_len] == magic:
                end = text.find(end_magic)
                xml_decl = text[magic_len+1:end]
                start = xml_decl.find('<')
                end = xml_decl.rfind('>')

                xml_decl = xml_decl[start:end+1]
                script = ScriptComponentFactory._from_xml(xml_decl)
                script.path = path

            else:
                name = os.path.splitext(os.path.basename(path))[0]
                script = ScriptComponent(path=path, name=name)

        except:
            import traceback; traceback.print_exc()
            name = os.path.splitext(os.path.basename(path))[0]
            script = ScriptComponent(path=path, name=name)

        return script
    
    def to_file(self, script_component, code):
        """ Write the script component to disk. """

        f = file(script_component.path, 'w')
        
        f.write(ScriptComponent.START_MAGIC)
        f.write('"""\n')
        
        f.write("<?xml version='1.0' encoding='utf-8'?>\n\n")
        f.write("<script name='%s' description='%s'>\n" %
                (script_component.name, script_component.description)
        )

        if len(script_component.inputs) > 0:
            f.write('  <inputs>\n')
            for input in script_component.inputs:
                object_to_xml(f, '    ', input)
            f.write("  </inputs>\n")

        if len(script_component.outputs) > 0:
            f.write("  <outputs>\n")
            for output in script_component.outputs:
                object_to_xml(f, '    ', output)
            f.write("  </outputs>\n")

        if len(script_component.metadata) > 0:
            f.write("  <metadata>\n")
            for name, value in script_component.metadata.items():
                f.write("    <nv name='%s' value='%s'/>\n" % (name, value))
            f.write("  </metadata>\n")

        f.write("</script>\n")

        f.write('"""\n')
        f.write(ScriptComponent.END_MAGIC)

        f.write(code)
        f.close()
        
        return

    to_file = classmethod(to_file)

    ###########################################################################
    # Stuff only used from within this rather crappy module!
    ###########################################################################
    

    def _from_xml(self, string):

        doc = parseString(string)

        return ScriptComponentFactory._from_xml_doc(doc)

    _from_xml = classmethod(_from_xml)

    
    # fixme: We should also have an interface that takes a DOM node.
    def _from_xml_doc(self, doc):
        """ Create a new instance from its XML representation. """
        
        for child in doc.childNodes:
            if child.nodeType is not child.ELEMENT_NODE:
                continue

            if child.tagName == 'script':
                name = str(child.attributes['name'].value)
                description = str(child.attributes['description'].value)
                
                inputs = []
                outputs = []
                metadata = {}
                
                for c in child.childNodes:
                    if c.nodeType is not c.ELEMENT_NODE:
                        continue

                    if c.tagName == 'inputs':
                        for node in c.childNodes:
                            if node.nodeType is not node.ELEMENT_NODE:
                                continue

                            input = object_from_xml(node, Input)
                            inputs.append(input)

                    if c.tagName == 'outputs':
                        for node in c.childNodes:
                            if node.nodeType is not node.ELEMENT_NODE:
                                continue

                            output = object_from_xml(node, Output)
                            outputs.append(output)

                    if c.tagName == 'metadata':
                        for node in c.childNodes:
                            if node.nodeType is not node.ELEMENT_NODE:
                                continue

                            n = str(node.attributes['name'].value)
                            value = str(node.attributes['value'].value)
                            metadata[n] = value

        component = ScriptComponent(
            name=name,
            description=description,
            inputs=inputs,
            outputs=outputs,
            **metadata
        )

        return component
        
    _from_xml_doc = classmethod(_from_xml_doc)
    
#### EOF ######################################################################
