from enthought.logger import logger

from enthought.traits.api import HasTraits, Property
from executable import Executable
from ioport import IOPort
from exceptions import InsufficientInputError

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class Actor(HasTraits, Executable) :
    """ An Actor is an executable entity. This interface defines the 
    	common functionality in AtomicActor and CompositeActor.
    """

    def __init__( self, *args, **kw ) :
        HasTraits.__init__( self, *args, **kw )
        return
        
    #### Trait Properties #####################################################
    
    # Property: executive_director isa Director
    
    def _get_executive_director(self) :
        """ Return the executive director, if there is one, or None.
        The executive director governs the execution within the container
        of the Actor.
        """
        if (not hasattr( self, 'container' )) or self.container is None :
            director = None

        else :
            director = self.container.director
            
        return director
    
    executive_director = Property( _get_executive_director )

    # Property: director isa Director
    
    def _get_director(self) :
        """ Return the local director, if there is one, or the executive
        director, if there is one, or None.
        The director governs the execution within this actor.
        """
        return self.executive_director
        
    director = Property( _get_director )

    #### Executable Methods ##################################################
    def iterate( self, count ) :
        """ Iterate count times.
        An iteration is: prefire, fire, postfire
        If prefire or postfire return False, stop iterating
        """
        for i in xrange( count ) :
            if not self.prefire() :
                raise InsufficientInputError, ("Actor '%s'" % self.name)
            
            self.fire()
            
            if not self.postfire() :
                return False
        
        return True
    
    def wrapup( self ):
        for port in self.ports:
            port.wrapup()
        
        return
        
    #### Methods ##############################################################
        
    def input_port_list(self) :
        """ Return a list of this actor's input ports
        """
        ports = [ port for port in self.ports if port.is_input ]

        return ports
        
    def output_port_list(self) :
        """ Return a list of this actor's output ports
        """
        ports = [ port for port in self.ports if not port.is_input ]

        return ports
        
    

    def new_receiver(self, port) :
        """ Return a new receiver that is compatible with the director above
        this actor.
        """
        director = self.executive_director
        
        if director is None :
            raise RuntimeError, "Can't create receiver without executive director"
        
        receiver = director.new_receiver( port )
        
        return receiver

    def new_inside_receiver(self, port) :
        """ Return a new receiver that is compatible with the composite's
        director.
        """
        director = self.director
        if director is None :
            raise RuntimeError, "Can't create receiver without director"
        
        return director.new_receiver( port )


    def dependents(self) :
        """ Return a list of actors dependent on my outputs.
        """
        actors = []
        
        for port in self.output_port_list() :
            actors += port.dependent_actors()
                
        return actors
        
    #### Node Methods ########################################################
    # I don't really like this.  Actor is assuming that it will be a mixin
    # with Node.  That is clearly my intent on usage, but the language
    # syntax does not allow me to express what I want or I haven't learnt how.

    def new_port(self, name, is_input ) :
        """ Create a port compatible with this node.
        """
        self._check_port_name( name )
        port = IOPort( name, is_input )
        port.container = self
        return port


### EOF ######################################################################        
