import os


from enthought.traits.api \
    import Instance, Button, Any, Code, List, Str, Bool

from enthought.traits.ui.api \
    import View, Item, Group, Handler,  message, CodeEditor

from enthought.traits.ui.menu \
    import NoButtons
    
from enthought.traits.ui.key_bindings \
    import KeyBinding, KeyBindings
    
from enthought.envisage import get_application

from enthought.envisage.project.project_plugin import ProjectPlugin

from enthought.graph.function.metadata_elements import FunctionElement

from enthought.graph.function.xml_element_names import K_ARG_TAG

from enthought.graph.vpl.chip_manager import ChipManager

from enthought.logger import logger
    
from xml_script_util import XmlScriptUtil 

from function_argument_information import FunctionArgumentInformation   
    
#-------------------------------------------------------------------------------
#  Define a KeyBindings object:  
#-------------------------------------------------------------------------------
        
key_bindings = KeyBindings(

    KeyBinding( binding1    = 'Ctrl-s',
                description = 'Save to a file',
                method_name = 'object_save_changed' ),
                
    KeyBinding( binding1    = 'Ctrl-r',
                description = 'Run script',
                method_name = 'object_run_changed' ),
                                
    KeyBinding( binding1    = 'Ctrl-g',
                description = 'Generate chip',
                method_name = 'object_generate_chip_changed' ),
                
    KeyBinding( binding1    = 'Ctrl-p',
                description = 'Auto populate',
                method_name = 'object_auto_populate_changed' )
)                
           
class PortEditor ( Handler ):
    """ Graphical editor for ports.
    """
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
    
    # Code    
    code = Code    
     
    # Actor being edited:
    function_element = Instance( FunctionElement )
    
    # Save button
    save = Button
        
    # Run button
    run = Button
    
    # Auto populate
    auto_populate = Button
    
    # Help button
    help = Button
        
    # Overwrite existing values
    overwrite_values = Bool(True)
    
    # Display overwrite dialog
    display_dialog = Bool(True)
    
    # The resource being edited.
    resource = Any
          
    # Dirty indicator
    dirty = Bool(False)
         
    #---------------------------------------------------------------------------
    #  Trait editor views:  
    #---------------------------------------------------------------------------
    
    # Define the View to use:    
    traits_view = View(
        Group( 
            Group( 
                Group(
                    Item( 'code',
                          editor = CodeEditor( key_bindings = key_bindings ) ),
                    label       = 'Code',
                    export      = 'DockShellWindow',
                    show_labels = False
                ),
                Group( 
                    Item( 'function_element', 
                          style = 'custom',
                          id    = 'function_element' ), 
                    label       = 'Parameters',
                    export      = 'DockShellWindow',
                    show_labels = False
                ),
                id          = 'splitter1',
                orientation = 'vertical',
                layout      = 'split'
            ),
            Group(
                Item( 'object.save' ),
                Item( 'object.run' ),
                Item( 'object.auto_populate', label='Populate' ),
                Item( 'object.help' ), 
                orientation = 'horizontal',
                layout      = 'normal',
                show_labels = False
            ),
        ),
        id        = 'enthought.graph.scripts.editor.port_editor',
        kind      = 'live',
        dock      = 'tab',
        buttons   = NoButtons,
        resizable = True,
        width     = 1.0,
        height    = 1.0,
        help_id   = 'enlib|HID_Script_wnd',
    )    
                        
                        
    def __init__(self, **traits):
        """ construct new PortEditor """
        
        super(Handler, self).__init__(**traits)
        
        scriptUtil = XmlScriptUtil()
        
        script_text = self._load()
        
        self.function_element, self.code = scriptUtil.get_function_element(script_text)
        
        self.dirty = False
   
        return
                                 
    def object_save_changed(self, info):
        """ Save the modifications in the editor
        """
        
        if info.initialized:
            self._save()
            
        return
        
    def object_run_changed(self, info):
        """ run the script
        """
        
        if info.initialized:
            self._save()
            self._run()
            
        return
        
    def object_auto_populate_changed(self, info):
        """ Auto populate fields.
        """
        
        if info.initialized:
 
            scriptUtil = XmlScriptUtil()

            arg_info = FunctionArgumentInformation(self.code)
            
            # set name and function name
            if arg_info.function_name != '':
                # set the function name
                self.function_element.function_name = arg_info.function_name
                
                # set the name
                self.function_element.name = arg_info.suggested_chip_name
                                            
            self._remove_missing_arguments(arg_info.input_arguments, self.function_element.input_args)
            self._remove_missing_arguments(arg_info.output_arguments, self.function_element.output_args)
            
            self._update_existing_arguments(arg_info.input_defaults, self.function_element.input_args)
            self._update_existing_arguments(arg_info.input_defaults, self.function_element.output_args)
            
            self._add_new_arguments(arg_info.input_arguments, arg_info.input_defaults, self.function_element.input_args)
            self._add_new_arguments(arg_info.output_arguments, arg_info.input_defaults, self.function_element.output_args)
            
            self._reorder_arguments(arg_info.input_arguments, self.function_element.input_args)
            self._reorder_arguments(arg_info.output_arguments, self.function_element.output_args)
            
            self.overwrite_values = True
            self.display_dialog = True
                
        return
            
    def object_help_changed(self, info):
        """Help button was clicked.
        """
        if info.initialized:
            hp = get_application().get_service('enthought.help.IHelp')
            hp.library.show_topic(info.ui.view.help_id)
        return
    
    def object_code_changed(self, info):
        """ code has changed, set dirty flag
        """
       
        if info.initialized:
            self.dirty = True
                                     
        return
        
    def _remove_missing_arguments(self, arguments, list):
        """ Remove arguments that no longer exist in the list.
        """
        
        if not self.overwrite_values:
            return 
        
        list_copy = list[:]
        for argument_element in list_copy:
            
            if argument_element.arg_name not in arguments:
                
                if not self._show_overwrite_dialog():
                    break
                
                list.remove(argument_element)
            
        return
        
    def _add_new_arguments(self, arguments, input_arguments, list):
        """ Add new arguments 
        """
        
        if not self.overwrite_values:
            return 
        
        default_service = get_application().service_registry.get_service('enthought.graph.function.MetadataDefaults')
        
        for name in arguments:
            
            if self._find_argument_element_by_arg_name(name, list) is None:
                default = default_service.lookup(K_ARG_TAG, name)
                
                if default is not None:
                    default.arg_name = name
                    
                    if input_arguments.has_key(name):
                        default.default = input_arguments[name]
                    
                    list.append(default)
                    
        return
            
    def _update_existing_arguments(self, input_defaults, list):
        """ Update the existing arguments
        """
        
        if not self.overwrite_values:
            return 
        
        default_service = get_application().service_registry.get_service('enthought.graph.function.MetadataDefaults')
        
        for argument_element in list:
            arg_name = argument_element.arg_name
            
            default = default_service.lookup(K_ARG_TAG, arg_name)
              
            if default is not None:
                if self._check_values_equal(argument_element.name, default.name):
                    argument_element.name = default.name
                
                if self._check_values_equal(argument_element.modifier, default.modifier):    
                    argument_element.modifier = default.modifier
                    
                if self._check_values_equal(argument_element.bundle, default.bundle):
                    argument_element.bundle = default.bundle
                    
                if self._check_values_equal(argument_element.units, default.units):
                    argument_element.units = default.units
                    
                if self._check_values_equal(argument_element.family_name, default.family_name):
                    argument_element.family_name = default.family_name
                    
                if self._check_values_equal(argument_element.units_system_name, default.units_system_name):
                    argument_element.units_system_name = default.units_system_name
                
                if self._check_values_equal(argument_element.description, default.description):
                    argument_element.description = default.description
                    
                if self._check_values_equal(argument_element.external_type, default.external_type):
                    argument_element.external_type = default.external_type
                    
                if self._check_values_equal(argument_element.internal_type, default.internal_type):
                    argument_element.internal_type = default.internal_type
                    
                if self._check_values_equal(argument_element.is_index, default.is_index):
                    argument_element.is_index = default.is_index
                    
                if self._check_values_equal(argument_element.is_optional, default.is_optional):
                    argument_element.is_optional = default.is_optional
                    
                #argument_element.arg_name = default.arg_name
                
                if self._check_values_equal(argument_element.arg_name, arg_name):
                    argument_element.arg_name = arg_name
                
                #if self._check_values_equal(argument_element.position, default.position):
                    argument_element.position = default.position
                    
#                if self._check_values_equal(argument_element.composite_name, default.composite_name):
#                    argument_element.composite_name = default.composite_name
                    
                if input_defaults.has_key(arg_name):
                    if self._check_values_equal(argument_element.default, input_defaults[arg_name]):
                        argument_element.default = input_defaults[arg_name]
                elif self._check_values_equal(argument_element.default, default.default):
                    argument_element.default = default.default
                    
                if self._check_values_equal(argument_element.repeat_group, default.repeat_group):
                    argument_element.repeat_group = default.repeat_group
            
        return
        
    def _reorder_arguments(self, arguments, list):
        """ put the arguments back in the correct order
        """
        
        # if arguments do not match up, don't reorder.
        if not self._check_arguments_exist(arguments, list):
            return
         
        for index in range(len(arguments)):
            arg_name = arguments[index]
            
            element_index = self._find_argument_element_index_by_name(arg_name, list)
            
            # Element isn't found, for some reason the arguments in the script must not match
            # the arguments in the port editor.  Don't continue to reorder.
            if element_index == -1:
                break
            
            if index != element_index:
                argument_element = list[element_index]
                
                # put the argument element back in the correct order
                list.remove(argument_element)
                list.insert(index, argument_element)
 
        return
            
    def _find_argument_element_by_arg_name(self, name, list):
        """ find argument element by arg_name
        """
        
        for argument_element in list:
            if argument_element.arg_name == name:
                return argument_element
                
        return None
    
    def _find_argument_element_index_by_name(self, arg_name, list):
        """ find the index of the argument element with the given arg_name
        """
        
        for index in range(len(list)):
            argument_element = list[index]
                
            if argument_element.arg_name == arg_name:
                return index
            
        return -1
        
    def _show_overwrite_dialog(self):
        """ prompt to determine if the user would like to overwrite existing values
        """
        
        if self.display_dialog:
            result = message(message = 'Some existing values will be overwritten.  Do you want to continue?', 
                             title = 'Warning', buttons = [ 'OK', 'Cancel' ])
        
            self.display_dialog = False
            self.overwrite_values = result
        
        return self.overwrite_values
        
    def _check_values_equal(self, value1, value2):
        """ check to see if the values are equal to each other.
        
            if not display a dialog to warn of overwriting values.
        """
        
        if value1 != value2: 
            self._show_overwrite_dialog()
                        
        return self.overwrite_values
        
    def _check_arguments_exist(self, arguments, list):
        """ check that all arguments exist in the list
        """
        
        if len(arguments) != len(list):
            return False
            
        for argument in arguments:
            element = self._find_argument_element_by_arg_name(argument, list)
            
            if element is None:
                return False
                
        return True
        
    def _save(self):
        """ Perform activities necessary to save the file.
        """
        
        scriptUtil = XmlScriptUtil()
        
        arg_info = FunctionArgumentInformation(self.code)
        
        # reorder arguments before save.
        self._reorder_arguments(arg_info.input_arguments, self.function_element.input_args)
        self._reorder_arguments(arg_info.output_arguments, self.function_element.output_args)
        
        function_element = self.function_element 
        originalText = self.code
                  
        # Get full function text including metadata doc string.
        function_string = scriptUtil.put_xml_info_from_element(originalText, function_element)
        
        f = file(self.resource.path, 'w')
        f.write(function_string)
        f.close()
        
        self.dirty = False

        # generate chip
        self._generate()

        return
        
    def _generate(self):
        """ Generate chip.
        """
        
        chipManager = ChipManager()
        scriptUtil = XmlScriptUtil()
        
        if not scriptUtil.check_empty_function_element(self.function_element):
            
            error = False
            exception_value = None
            # generate and add actor to the manager.
            try: 
                chipManager.generate(self.resource.path)
            except Exception, value:
                error = True
                exception_value = value                
                logger.exception('trouble generating chip')
              
            if error:
                error_message = 'Chip generation failed.'
                
                if exception_value is not None:
                    error_message = error_message + '\n\n' + str(exception_value)
                
                message(message = error_message, title = 'Generate Chip', buttons = [ 'OK' ]) 
            else:
                message(message = 'Chip generation successful.', title = 'Generate Chip', buttons = [ 'OK' ])
                                       
        return
        
    def _run(self):
        """ Run the script being edited in the python shell. """

        # fixme: Nice hack!!!!
        from enthought.envisage.ui.python_shell import PythonShellPlugin
        view = PythonShellPlugin.instance.views[0]
        
        view.shell.execute_command('execfile(r"%s")' % self.resource.path)
        
        return
                
    def _load(self, path=None):
        """ Loads the contents of the editor. """
        
        if path is None:
            path = self.resource.path

        # We will have no path for a new script.
        if len(path) > 0:
            f = open(path, 'r')
            text = f.read()
            f.close()

        else:
            text = ''
        
        return text
        
        
        
#### EOF ######################################################################
