import os

import copy

from enthought.naming.api import PyFSContext

from enthought.naming.pyfs_context import ENVIRONMENT

from enthought.io.file import File

from enthought.envisage.ui import View
from enthought.envisage.ui import UIPlugin
from enthought.envisage import get_application
        
from enthought.envisage.resource import ResourceManager
from enthought.envisage.resource import FolderResourceType
from enthought.envisage.resource.ui import ResourceTree
from enthought.envisage.resource.ui.resource_sorter import FoldersBeforeFiles

from enthought.envisage.project.project_plugin import ProjectPlugin
from enthought.envisage.project import CookieManager
from enthought.envisage.project.action.open_cookie import OpenCookie

from enthought.traits.api import Instance

from enthought.pyface.action.api import Action, Group, MenuManager, Separator

from enthought.naming.api import Binding, Context

from enthought.graph.chips.resource_type.chip_resource_type import ChipResourceType

from enthought.graph.vpl.chip_manager import ChipManager

class ChipView(View):
    
    # singleton instance
    instance = None
    
    def add_paths(self, paths):
        """ Add a list of paths to be displayed in the Chips tree.
        """
        
        context = self.root.obj
        
        # TODO is this the right way to add additional filtering?
        environment = copy.copy(ENVIRONMENT)
        environment[PyFSContext.FILTERS].append('*.pyc')
        environment[PyFSContext.FILTERS].append('__init__.py')
        environment[PyFSContext.FILTERS].append('tests')
        
        for path in paths:
            path = path.replace('\\', '/')
            name = path.split('/')[-1]
            pyfs_context = PyFSContext(path=path, environment = environment)  
            context.bind(name, pyfs_context) 
            
        self._bind_my_chips_to_context(context) 
            
        return
    
    ###########################################################################
    # View interface.
    ###########################################################################

    def _create_contents(self, parent):
        """ Create the toolkie-specific control that represents the view. """

        return self._create_chip_tree(parent)
        
    def _create_chip_tree(self, parent):
        """ Create the chip tree. """
        
        workspace = ProjectPlugin.instance.workspace
                 
        resource_manager = self._create_resource_manager() 
        
        context = self._create_root_context() 
 
        self.root = Binding(name='Chips', obj=context)
        
        tree = ResourceTree(
            parent,
            resource_manager = resource_manager,
            root = self.root,
            sorter = FoldersBeforeFiles(),
# FIXME: binding a new context to the root fails with a wxAssertFailure "can't retrieve virtual root item" 
# if show_root is equal to False
#            show_root = False
            show_root = True
        )
        
        # FIXME: I need this callback to be performed anytime a new file is created.
        # However, the the _label_edit_callback trait is private and generally 
        # only set when calling edit_label on tree.  There needs to be a better way to set a
        # callback that will be used every time.
        tree._label_edit_callback = self._on_label_edited
        
        ChipView.instance = self
        
        from enthought.graph.chips.chip_resource_plugin import ChipResourcePlugin
        self.add_paths(ChipResourcePlugin.instance.action_paths)
        
        return tree.control
        
    def _create_resource_manager(self):
        """ Creates a resource manager. """
        
        resource_manager = ResourceManager()
        
        # Folders!
        folder_resource_type = FolderResourceType()
        folder_node_type = folder_resource_type.node_type
        folder_node_type.context_menu = MenuManager(
            Group(NewFolderAction(view = self)),
            Group(NewChipAction(view = self)),
            Group(
#                PasteAction(resource_manager=resource_manager),
                DeleteFolderAction()
            )
        )
        resource_manager.add_resource_type(folder_resource_type)
        
        # Python files!
        python_resource_type = ChipResourceType()
        python_node_type = python_resource_type.node_type
        python_node_type.context_menu = MenuManager(
            Group(
                EditChipAction(),
#                CopyAction(resource_manager=resource_manager),
                DeleteChipAction()
            )
        )
        resource_manager.add_resource_type(python_resource_type)
        
        return resource_manager
        
    def _on_label_edited(self, tree, node, label):
        """ A label has been edited. """
        
        if isinstance(node.obj, File):
            cookieManager = CookieManager()
            cookie = cookieManager.get_cookie(OpenCookie, node.obj)
            
            workbench_window = UIPlugin.instance.active_window
            
            cookie.open(workbench_window, node)
        
        return
        
    def _create_root_context(self):
        """ Create a root context for the tree.
        """
        # create the root context  
        context = Context()

        return context
        
    def _bind_my_chips_to_context(self, context):
        """ Create my chips directory if needed and bind it to the specified context
        """
        
        from enthought.graph.vpl.graph_editor_plugin import GraphEditorPlugin
        
#        path = GraphEditorPlugin.instance.state_location + '/user_action_path/my_actions'
        path = GraphEditorPlugin.instance.my_actions_path
        
        context.bind('my_actions', PyFSContext(path = path))
        
        return
        
class NewFolderAction(Action):
    """ Creates a new folder. """

    name = "New Folder"
    
    view = Instance(ChipView)

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        parent_folder = event.node
        
        # The context to create the folder in.
        context = parent_folder.obj

        # Generate a name that is unique within the context.
        default_name = self._get_default_name('New Folder', context)

        # Create the folder.
        sub_context = context.create_subcontext(default_name)
        
        # create empty __init__.py file
        init_file = os.path.abspath(sub_context.path + '/__init__.py')
        file = open(init_file, 'w')
        file.close()
        
        # Lookup the binding.
        binding = parent_folder.obj.lookup_binding(default_name)

        # Start a label edit.
        event.widget.invoke_later(event.widget.edit_label, *(binding, self.view._on_label_edited))
            
        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_default_name(self, prefix, context):
        """ Returns a name that is not already bound in a context. """

        names = context.list_names('')
        name  = prefix
        
        id = 2
        while name in names:
            name = '%s (%d)' % (prefix, id)
            id += 1
            
        return name
        
class DeleteFolderAction(Action):
    """ Deletes a node. """

    name = "Delete"

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        # The selected node.
        target_node = event.node
        
        context = target_node.obj
        
        self._traverse_context(context)
        
        # Nuke it!
        target_node.context.unbind(target_node.name)
            
        return
        
    def _traverse_context(self, context):
        """ traverse the given context removing all registered chips.
        """
        
        names = context.list_names()
        
        for name in names:
            obj = context.lookup(name)
               
            if isinstance(obj, Context):
                self._traverse_context(obj)
            elif isinstance(obj, File):
                # this is probably a registered chip, try to remove it.
                ChipManager().remove_with_path(obj.path)
                
        return
         
class DeleteChipAction(Action):
    """ Deletes a node. """

    name = "Delete"

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        # The selected node.
        target_node = event.node
        
        ChipManager().remove_with_path(target_node.obj.path)
    
        # Nuke it!
        target_node.context.unbind(target_node.name)
            
        return
        
class CopyAction(Action):
    """ Copy a node. """

    name = "Copy"
    
    resource_manager = Instance(ResourceManager)

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        # The selected node.
        target_node = event.node
    
        # Get the resource type of the node that we are copying.
        resource_type = self.resource_manager.get_type_of(target_node.obj)
        copy_value = resource_type.node_type.get_copy_value(target_node)

        # fixme: Maybe a more formal clipboard!
        get_application().clipboard = [copy_value]
        
        return

# Error messages.
DST_SAME_AS_SRC = "The destination folder is the same as the source folder"
DST_SUB_OF_SRC  = "The destination folder is a sub-folder of the source folder"

class PasteAction(Action):
    """ Paste a node. """

    name = "Paste"
    
    resource_manager = Instance(ResourceManager)

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Performs the action. """

        # The selected node.
        target_node = event.node
        ctx = self._get_target_context(target_node)

        # fixme: Maybe a more formal clipboard!
        for data in get_application().clipboard:
            # Make sure that we are not trying to paste a folder onto itself
            # or a sub-folder onto one of its ancestors.
            if isinstance(data.obj, Context):
                if ctx.namespace_name == data.obj.namespace_name:
                    raise ValueError(DST_SAME_AS_SRC)
                
                elif ctx.namespace_name.startswith(data.obj.namespace_name):
                    raise ValueError(DST_SUB_OF_SRC)

            # Get the resource type of the object being pasted.
            resource_type = self.resource_manager.get_type_of(data.obj)

            # Generate a unique name for the object being pasted (unique within
            # the context that it is being pasted into).
            name = self._get_unique_name(data.name, ctx)
            
            # Ask the node type for the actual value to paste.
            value = resource_type.node_type.get_paste_value(data)

            # Make sure that the value that we are pasting has the appropriate
            # name.
            #
            # fixme: What if the paste value is of a different resource type?!
            resource_type.set_name(value, name)

            # fixme: Hmmmmmmm!
            resource_type.copy(ctx, name, value)

        return

    def _get_target_context(self, node):
        """ Returns the target context for the paste action. """
        
        # Note that paste actions can ONLY be performed on either context
        # objects or objects that have a context adapter.
        ctx = self.resource_manager.type_manager.object_as(node.obj, Context)
        if ctx is None:
            raise ValueError('Can only paste into contexts')

        return ctx

    def _get_unique_name(self, prefix, context):
        """ Returns a name that is not already bound in a context. """

        names = context.list_names('')
        name  = prefix
        
        id = 2
        while name in names:
            name = '%s (%d)' % (prefix, id)
            id += 1
            
        return name
        
class EditChipAction(Action):
    
    name = "Edit Chip"
    
    def perform(self, event):
        """ Performs the action. """
         
        # The selected node.
        target_node = event.node
        
        cookieManager = CookieManager()
        cookie = cookieManager.get_cookie(OpenCookie, target_node.obj)
        
        workbench_window = UIPlugin.instance.active_window
        
        cookie.open(workbench_window, target_node)
        
        return
        
class NewChipAction(Action):
    
    name = "New Chip"
    
    view = Instance(ChipView)
    
    def perform(self, event):
        """ Performs the action. """
         
        parent_folder = event.node
        
        # The context to create the folder in.
        context = parent_folder.obj

        # Generate a name that is unique within the context.
        default_name = self._get_default_name('New Chip', '.py', context)
        
        # Create the file.
        file = File(os.path.join(context.path, default_name))
        
        # Bind the file.
        context.bind(default_name, file)
        
        # Lookup the binding.
        binding = parent_folder.obj.lookup_binding(default_name)

        # Start a label edit.
        event.widget.invoke_later(event.widget.edit_label, *(binding, self.view._on_label_edited))
        
        return
        
    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_default_name(self, prefix, suffix, context):
        """ Returns a name that is not already bound in a context. """

        names = context.list_names('')
        name  = prefix + suffix
        
        id = 2
        while name in names:
            name = '%s (%d)%s' % (prefix, id, suffix)
            id += 1
            
        return name
        
#### EOF ######################################################################
        
        