
from enthought.logger import logger
from enthought.traits.api import Trait, Property

from composite_node import CompositeNode
from actor import Actor
from director import Director
from iorelation import IORelation

class CompositeActor( Actor, CompositeNode ) :
    """ An executable CompositeNode
    """
    
    def __init__(self, *args, **kw) :
        """ Creates a new CompositeActor node.
        """
        CompositeNode.__init__(self, *args, **kw)
        self._director = None
        
    #### Trait Properties #####################################################
    
    #### Trait Properties #####################################################
    
    # Property: is_opaque isa bool

    # The internal topology of opaque nodes is hidden from the outside
    # if the composite has a Director.
    def _get_is_opaque( self ) :
        return self._director is not None
        
    is_opaque = Property( _get_is_opaque )


    # Property: director isa Director
    
    def _get_director(self) :
        """ Return the local director, if there is one, or the executive
        director, if there is one, or None.
        The director governs the execution within this actor.
        """
        if self._director is None :
            director = self.executive_director
        
        else :
            director = self._director
            
        return director
        
    def _set_director( self, director ) :
        """ Set the local director for execution of this CompositeActor.
		This should not be called be directly.  Instead, set the container
		on the director.
     	"""
        self._director = director
        
        return
        
    director = Property( _get_director, _set_director )
    
    
    #### Methods ##############################################################
        


    #### Executable methods ###################################################

    def preinitialize( self ) :
        #logger.debug( "CompositeActor.preinitializing: %s" % self.full_name )
        if not self.is_opaque :
            raise RuntimeError, \
                "Cannot preinitialize: no director, non-opaque actor."

        self._create_receivers()
        director = self.director
        director.preinitialize()
        return
        
    def initialize( self ) :
        #logger.debug( "CompositeActor.initialize: %s" % self.full_name )
        if not self.is_opaque :
            raise RuntimeError, \
                "Cannot initialize: no director, non-opaque actor."

        # clear my input port receivers
        for port in self.input_port_list() :
            port.clear()
                    
        # clear all my output port inside rececivers
        for port in self.output_port_list() :
            port.clear_inside()
            

        # have director initialize actors contained within this composite
        director = self.director
        director.initialize()
        return
        
    def prefire( self ) :
        """ If this actor is opaque, invoke the prefire method of the local
		director. This method returns true if the actor is ready to fire
		(determined by the prefire() method of the director).
        """
        #logger.debug( "CompositeActor.prefire: %s" % self.full_name )
        director = self.director
        if not self.is_opaque :
            raise RuntimeError, \
                "Cannot prefire: no director, non-opaque actor."
        
        return director.prefire()
        
        
    def fire( self ) :
        """ Execute the actor.
        
        This may be invoked several times between invocations of prefire()
        and postfire(). Output data may (and normally will) be produced.
        Typically, the fire() method performs the computation associated
        with an actor. This method is not required to have bounded execution.
        However, after stopFire() is called, this method should
        return in bounded time.        
        """
        #logger.debug( "CompositeActor.fire: %s is_opaque: %s" 
        # % (self.full_name, self.is_opaque) )
        if not self.is_opaque :
            raise RuntimeError, "Cannot fire a non-opaque actor."

        director = self.director
        
        # transfer inputs
        for port in self.input_port_list() :
            director.transfer_inputs( port )
            
        # fire - doit
        self._fire()
        
        # transfer outputs
        for port in self.output_port_list() :
            director.transfer_outputs( port )
            
        return True

    def postfire( self ) :
        """ If this actor is opaque, invoke the postfire method of the local
        director.
        """
        #logger.debug( "CompositeActor.postfire: %s" % self.full_name )
        director = self.director
        if not self.is_opaque :
            raise RuntimeError, \
                "Cannot postfire: no director or non-opaque actor."
        
        return director.postfire()
        
    def wrapup( self ) :
        """ If this actor is opaque, invoke the wrapup method of the local
        director.
        """
        #logger.debug( "CompositeActor.wrapup: %s" % self.full_name )
        director = self.director
        if not self.is_opaque :
            raise RuntimeError, \
                "Cannot wrapup: no director or non-opaque actor."
        
        return director.wrapup()

    ###########################################################################
    # Actor methods
    ###########################################################################
        

    ###########################################################################
    # Protected methods
    ###########################################################################
    
    def _fire( self ) :
        """ Perform actor specific fire action.  Inputs have been transfered
        and outputs will be transfered by caller of _fire.
        Default implementation calls the local director's fire method.
        Transparent composite actors are not fired.
        """
        self.director.fire()

        return
    
    def _create_receivers( self ) :
        """ Create receivers for all reachable ports.
        """
        if len( self.ports ) == 0 :
            return
        
        for port in self.ports :
            port.receivers()
            
        return
        
    def _create_relation( self, name ) :
        """ Factory for relations compatible with this Node """
        return IORelation(name)
        
### EOF ######################################################################    
