from enthought.logger import logger
from enthought.traits.api import HasTraits, Property
from enthought.traits.api import Str, List, Bool

from node import Node
from relation import Relation
from composite_port import CompositePort

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class CompositeNode( Node ) :
    """ A cluster of nodes and relations in a hierarchical graph.
    	A composite has ports through which the cluster is connected to a
    	the composite up one level in the hierarchy.
    	To add a node or relation to this composite, call its setContainer method
    	with this composite as an argument.
    	Ports on a transparent (i.e. opaque == False) CompositeNode allow
    	traversals to see through the composite directly to the ports on the
    	contained nodes as if the composite did not exist.  
    """
# fixme: when traits handles co-dependent classes with inheritance
    nodes = List         # should be List(Node)
    relations = List     # should be List(Relation)
    
    def __init__(self, name, *args, **kw) :
        """ Creates a new CompositeNode. """
        Node.__init__(self, name, *args, **kw)
        self.atomic = False
        
        return

    #### Trait Properties #####################################################
    
    # Property: is_opaque isa bool

    # The internal topology of opaque nodes is hidden from the outside
    # Composite nodes are by definition opaque.
    def _get_is_opaque( self ) :
        return False
        
    is_opaque = Property( _get_is_opaque )

    #### Methods ##############################################################
    def new_port(self, name) :
        """ Create a port compatible with this node. 
        	except: ValueError if node name conflict exists.
        """
        self._check_port_name(name)
        port = CompositePort( name )
        port.container = self
        return port

    def new_node(self, name) :
        """ Create a node compatible with this composite. 
        	except: ValueError if node name conflict exists.
		"""
        self._check_node_name(name)
        node = Node(name)
        node.container = self
        return node

    def get_node(self, name) :
        """ Return the node contained by this entity that has the specified name.
            If there is no such node, return None.
        """
        #ToDo: add compound name handling e.g. foo.bar.baz
        #ToDo: use the for else form in similar methods, more pythonic
        for node in self.nodes :
            if node.name == name :
                break
                
        else:
             node = None
        
        return node
        
    def node_list(self) :
        """ Return a copy of the list of contained Nodes.
        """
        return self.nodes[:]

    def reachable_node_list(self) :
        """ Return the list of opaque nodes reachable from the inside of
        this composite.
        """
        reachable_nodes = []
        
        if self.nodes is not None :
            
            for node in self.nodes :
                if node.is_opaque :
                    reachable_nodes.append( node )
                else :
                    reachable_nodes.extend( node.reachable_node_list() )
        
        return reachable_nodes
        
    def new_relation( self, name ) :
        """ Create a Relation compatible with this CompositeNode.
        """
        self._check_relation_name(name)
        relation = self._create_relation(name)
        relation.container = self
        return relation
            
    def get_relation( self, name ) :
        """ Return a contained relation with the given name or None.
        """
        for relation in self.relations :
            if relation.name == name :
                return relation
        
        return None
        
    def relations_list(self) :
        """ Return a copy of the list of contained Relations.
        """
        return self.relations[:]

    def connect(self, port1, port2, name=None ) :
        """ Create a new relation with the specified name and use it to connect
	        two ports. A reference to the newly created relation is returned.
	        When name is None, the relation name is the same as port2's name.
        """
        if port1 is None :
            raise ValueError, "port1 is None"
            
        if port2 is None :
            raise ValueError, "port2 is None"
        
        if port1 == port2 :
            raise ValueError, "port1 == port2"
            
        if name is None :
            name = port2.name
        
        self._check_relation_name(name)
        relation = self.new_relation( name )
        port1.link(relation)
        port2.link(relation)
        
        return relation
        
    ###########################################################################
    # Protected methods
    ###########################################################################

    def _add_port(self, port) :
        """ Used by Port to add itself to this Node
            excecpt: Node already contains a different Port with the same name
        """
        if not isinstance(port, CompositePort) :
            raise TypeError, "port must be a CompositePort"
        Node._add_port(self, port)

        return
        
    def _add_node(self, node) :
        """ Used by Node to add itself to this CompositeNode
        	except: ValueError if node name conflict exists.
        """
        self._check_node_name( node.name )
        self.nodes.append( node )

        return

    def _remove_node(self, node) :
        """ Node will remove itself when its setContainer method is called
        """
        if node in self.nodes :
            self.nodes.remove(node)

        return

    def _add_relation(self, relation) :
        """ Relation will add itself when its setContainer method is called
        """
        self._check_relation_name( relation.name )
        self.relations.append(relation)

        return
        
    def _remove_relation(self, relation) :
        """ Relation will remove itself when its setContainer method is called
        """
        if relation in self.relations :
            self.relations.remove(relation)

        return

    def _check_node_name(self, name) :
        """ except: ValueError if there is a conflict with an existing name """
        if name in [ n.name for n in self.nodes ] : 
            raise ValueError, 'Conflict with existing Node name: %s' % name

        return
            
    def _check_relation_name(self, name) :
        """ except: ValueError if there is a conflict with an existing name """
        if name in [ relation.name for relation in self.relations ] : 
            raise ValueError, 'Conflict with existing Relation name: %s' % name

        return

    def _create_relation( self, name ) :
        """ Factory for relations compatible with this Node """
        return Relation(name)
        
        
#### EOF ######################################################################
