from enthought.logger import logger
from enthought.traits.api import HasTraits
from enthought.traits.api import Trait, Str, List, Any, Property

from port import Port

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class CompositePort( Port ) :
    """ A Port supporting hierarchy.
    	CompositeNodes have ports that are instances of CompositePort.
    	CompositePorts have inside and outside links; non-composite ports
    	only have outside links.
    	Inside links connect a CompositePort to Relations contained by the 
    	CompositeNode that contains the CompositePort.
    	Outside links connect a CompositePort to Relations contained by the
    	container of the container of the CompositePort.
    	A CompositePort is transparent or opaque.  If transparent, deep topology
    	access will traverse through the port thus crossing levels of the
    	graph hierarchy.
    """
    
    inside_relations = List        #fixme should really be List(Relation)

    def __init__(self, *args, **kw) :
        """ Creates a new Port. """
        Port.__init__(self, *args, **kw)

    #### Trait Properties ####################################################
    
    # Property: is_opaque Bool, ReadOnly
    def _is_opaque(self) :
        """ return False if there is no container or the container's opacity
        """
        is_opaque = False
        
        if self.container is not None :
            is_opaque = self.container.is_opaque
        
        return is_opaque
    
    is_opaque = Property( _is_opaque )
    
    #### Methods #############################################################
    
    def link(self, relation) :
        """ Link this port with a Relation """
        logger.debug("CompositePort.link p: %s r: %s" 
            % (self.full_name, relation.full_name ))
        self._check_link(relation)
        
        if relation is None :
            self.relations.append( relation )
    
        elif self._is_inside_linkable( relation.container ) :
            self.inside_relations.append( relation )

        else :
            self.relations.append( relation )

        relation._link(self)

    def is_inside_linked(self, relation) :
        """ Return true if the given Relation is insided linked to this port.
        """
        return relation in self.inside_relations

    def unlink(self, relation) :
        """ Unlink the Relation (connected or inside) from this port. """
        if relation in self.inside_relations :
            relation._unlink(self)
            self.inside_relations.remove(relation)

        else :
            Port.unlink(self, relation)
        
        return

    def get_inside_relation( self, name ) :
        """ Return a connected relation with the given name or None.
        """
        for relation in self.inside_relations :
            if relation.name == name :
                return relation
        
        return None

    def inside_relations_list(self) :
        """ Return a copy of the list of inside Relations.
        """
        return self.inside_relations[:]

    def inside_connected_ports(self) :
        """ List the ports this port is connected to via inside linked relations.
            Note that a port may be listed more than once if more 
            than one connection to it has been established.
         """
        ports = []
        for relation in self.inside_relations :
            ports += relation.linked_ports( exclude=self )
        
        return ports

            
        
    ###########################################################################
    # Protected methods
    ###########################################################################
    
    def _check_link(self, relation) :
        """ Raise ValueError if port and relation are incompatible.
            Rules: Port rules plus relation can't cross hierarchy layers
            Subclasses should override to apply more specific rules.
        """
        Port._check_link(self, relation)
        
        if relation is None :
            return
        
        # CompositePort is the means to move between hierarchy levels
        # The relation must be at the same level (same container)
        # or one level above.
        container = self.container
        if ( container != relation.container
            and container.container != relation.container ) :
                raise ValueError, "Link crosses levels of the hierarchy"
        
    def _is_inside_linkable(self, node) :
        """ Return true if this port is either a port of the specified node,
            or a port of a node that (deeply) contains the specified node.
        """
        # strictly speaking, the algorithm used here is looser than the
        # one used in _check_link.  This one allows layer crossing. However
        # this method is really used to determine direction (inside vs. outside)
        # _check_link is used to enforce depth constraints
        port_container = self.container
        while node is not None :
            if port_container == node :
                return True
            node = node.container
        return False
        
#### EOF ######################################################################
