
# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

from enthought.traits.api import HasTraits, Property

from executable import Executable

class Director( HasTraits, Executable ) :
    """
	A Director governs the execution within a CompositeActor.  A composite actor
	that contains a director is said to be opaque, and the execution model
	within the composite actor is determined by the contained director. 
	
	A composite actor without a director is under the control of the director
	of the composite actor it is contained within.  The top most composite actor
	is required to have a direct for a graph to be executable.
	
	The manner in which a Director implements the Executable methods defines
	the semantics of the execution model.
    """

    def __init__( self, *args, **kw ) :
        # Property defaults
        self._container = None
        HasTraits.__init__(self, *args, **kw)
        return
        
    #### Trait Properties ####################################################
    
    # Property: container isa CompositeNode
    
    def _set_container(self, container) :
        """ Specify the container for this director.
            Director is removed from existing container if contained.
        """
        # no-op case
        if self._container is not None :
            if self._container == container :
                return

        # link into the new container
        if container is not None :
            container.director = self
        
        previous_container = self._container
        self._container = container
        
        # unlink from previous container
        if previous_container is not None :
            previous_container.director = None
        
        return
                
    def _get_container(self) :
        """ Return the container of this Director
        """
        return self._container
    
    container = Property( _get_container, _set_container )
    
    #### Methods ##############################################################    
    
    def new_receiver( self, port ) :
        """ Return a new receiver that is compatible with the director.
        """
        raise NotImplementedError
    
    def transfer_inputs( self, port ) :
        """ Transfer at most one token from given input port of the container
        to the ports it is connected to on the inside.
        Returns True if a token was transferred.
        Raises ValueError if port is not both input and opaque
        """
        if not (port.is_input and port.is_opaque) :
            raise ValueError, \
                "Attempt to transfer inputs on non-input or non-opaque port."

        token_transferred = False
        
        if port.has_tokens() :
            token = port.get()
            port.put(token)
            token_transferred = True
            
        return token_transferred
    
    def transfer_outputs( self, port ) :
        """ Transfer at most one token from given output port of the container
        to the ports it is connected to on the outside.
        Returns True if a token was transferred.
        Raises ValueError if port is input or transparent
        """
        if port.is_input or not port.is_opaque :
            raise ValueError, \
                "Attempt to transfer outputs on input or non-opaque port."

        token_transferred = False
        
        if port.has_tokens() :
            token = port.get()
            port.put(token)
            token_transferred = True
            
        return token_transferred
    
    #### Executable Methods ###################################################
    def preinitialize(self) :
        """ This method is invoked exactly once per execution of a model,
            before any of these other methods are invoked. For actors, this may
            trigger changes in the topology.
        """
        container = self.container
        
        for actor in container.reachable_node_list() :
            actor.preinitialize()

        return
        
    def initialize(self) :
        """ This method is invoked exactly once per execution of a model,
            before any of these other methods are invoked. For actors, this may
            trigger changes in the topology.
        """
        container = self.container
        
        for actor in container.reachable_node_list() :
            actor.initialize()

        return
        
    def prefire( self ) :
        """ Return true if the director is ready to fire. This method is
		called by the container of this director to determine whether the
		director is ready to execute. It does not
		call prefire() on the contained actors.
        """
        return True

        
    def postfire( self ) :
        """ Return True if exeuction should continue.
        """
        return True
        
    def wrapup(self) :
        """ Invoke the wrapup() method of all the actors contained in the
		director's container.   In this base class wrapup() is called on the
		associated actors in the order of their creation.  
        """
        container = self.container
        
        for actor in container.reachable_node_list() :
            actor.wrapup()

        return
        
#### EOF ######################################################################    
