from enthought.logger import logger

from enthought.traits.api import HasTraits
from enthought.traits.api import List, Int

from fifo_history_receiver import FifoHistoryReceiver

class FifoAutoHistoryReceiver( FifoHistoryReceiver ) :
    """ Implments a Receiver with first-in-first-out semantics that maintains
    a history of tokens. Operations automatically fall back to the history list
    if needed.  Methods never combine both unused and history tokens. 

    Thus has_tokens(n) will return True if there n or more tokens
    in the unused fifo or n or more tokens in the history fifo.  has_tokens(n)
    returns False if the unused fifo has p tokens and the history has q tokens
    when p < n and p + q >= n.  Likewise for get.
    
    The depth of the history is configurable.  get moves tokens to the history
    buffer.  get will retrieve tokens from the history if needed. """
    
    # The tokens currently stored in this receiver
    history_fifo = List
    
    # Maximum number of tokens in the history buffer
    depth = Int(1)

    #### Receiver Methods ####################################################


    def available_unused_tokens( self ) :
        """ Return the number of unused tokens available for get to return."""
        return FifoHistoryReceiver.available_tokens( self )

    def has_unused_tokens( self, n=1 ) :
        """ Return True if the receiver has n unused tokens for get to return.
        """
        return FifoHistoryReceiver.has_tokens( self, n )

    def get_unused( self, n=1 ) :
        """ Return a token or list of tokens, when n > 1, from this receiver.
        Tokens are removed from the unused fifo and put into the history fifo.
        """
        return FifoHistoryReceiver.get( self, n )
        
       
    def available_tokens( self ) :
        """ Return the number of tokens available for get to return.
        The returned count is from available_history_tokens() when
        available_unused_tokens() is zero.
        """
        size = self.available_unused_tokens()
        if size == 0 :
            size = self.available_history_tokens()
        
        return size

    def has_tokens( self, n=1 ) :
        """ Return True if the receiver has n tokens for get to return. """
        if self.has_unused_tokens(1) :
            hastokens = self.has_unused_tokens( n )
        else :
            hastokens = self.has_history_tokens( n )
        
        return hastokens

    def get( self, n=1 ) :
        """ Return the n oldest unused tokens or n newest history tokens.
        Never returns a mixed list of unused and history.  History tokens
        are only returned if there are zero unused tokens.  Returned unused
        tokens are become the newest history tokens.
        """
        if self.has_unused_tokens(1) :
            tokens = self.get_unused( n )
        else :
            tokens = self.get_history( n )
        
        return tokens

    
#### EOF ######################################################################    
