from enthought.logger import logger

from enthought.traits.api import HasTraits
from enthought.traits.api import List, Int

from fifo_receiver import FifoReceiver

class FifoHistoryReceiver( FifoReceiver ) :
    """ Implments a Receiver with first-in-first-out semantics that maintains
    a history of tokens. 
    
    The depth of the history is configurable.  get moves tokens to the history
    buffer.  get will retrieve tokens from the history if needed. """
    
    # The tokens currently stored in this receiver
    history_fifo = List
    
    # Maximum number of tokens in the history buffer
    depth = Int(1)

    #### Receiver Methods ####################################################

    def clear( self ) :
        """ Clear this receiver of any contained tokens. """
        self.history_fifo = []
        FifoReceiver.clear( self )
        
    def available_history_tokens( self ) :
        """ Return the number of tokens available for get to return."""
        return len(self.history_fifo)

    def has_history_tokens( self, n=1 ) :
        """ Return True if the receiver has n tokens for get to return. """
        return len(self.history_fifo) >= n
    
    def get_history( self, n=1 ) :
        """ Return the n most recent tokens.
        """
        if n > len(self.history_fifo) :
            raise ValueError, "Insufficient history data in receiver"
        
        # return the n most recent tokens
        tokens = self.history_fifo[-n:]

        if n == 1 :
            tokens = tokens[0]
            
        return tokens

        
    def get( self, n=1 ) :
        """ Return a token or list of tokens, when n > 1, from this receiver.
        Tokens are removed from the unused fifo and put into the history fifo.
        """
        if n > len(self.fifo) :
            raise ValueError, "Insufficient data in receiver"
        
        tokens = self.fifo[:n]
        del self.fifo[:n]

        # put those tokens in the history and drop really old tokens if needed
        self.history_fifo += tokens
        if len(self.history_fifo) > self.depth :
            del self.history_fifo[ : len(self.history_fifo) - self.depth ]

        # don't return a list if only one token was requested
        if n == 1 :
            tokens = tokens[0]
            
        return tokens

#### EOF ######################################################################    
