""" Factory for ConvergeActor prototypes based on function metadata. """
import sys
import os 
import time
import re
from types  import FunctionType
from string import capwords

from enthought.traits.api import HasTraits, Dict, List

from enthought.graph.vpl.converge_actor \
    import ConvergeActor, ConvergeInputBundle, ConvergeOutputBundle, \
           ConvergeInputPort, ConvergeOutputPort, ConvergeRepeatBundle, \
           bundle_suffix_map

from metadata_parser import MetadataParser, ParseException
from metadata_elements import BundleElement


class ActorPrototypeFactory( HasTraits ):
    """ Factory for ConvergeActor prototypes based on function metadata. """
    
    ###########################################################################
    # Public interface.
    ###########################################################################
    
    def from_filename(self, path):
        """ Returns an actor instance from the sepecified file, or None.
        
        Path is the path to the file on the file system. The file should contain
        a valid function to wrap with a ConvergeActor.
        """
        
        try:
            file_text = self._read_file(path)
        except:
            logger.exception('Error reading file "' + path + '"')
            return None

        return self.from_string(file_text)
    

    def from_string( self, text ):
        """ Returns an actor instance from the sepecified text, or None. 
        
        Limitation: Text can only contain one function with actor metadata.
        """
        
        parser = MetadataParser()
        elements = parser.elements_from_string( 'function', text )
        
        # If source text contains no fucntions with metadata, bail out.
        if len(elements) == 0:
            return None
            
        # Current implementation only supports one function, complain if needed.
        if len(elements) > 1:
            raise ValueError, "Input text contains more than one actor function."

        function_element = elements[0]
        
        function = self._get_callable(text, function_element.function_name)
        
        # get the function from the text
        return self.from_function_element( function_element, function )
        
    def from_function( self, xml_metadata, function ):
        """ Returns an actor instance from the specified function and the given xml metadata, or None. 
        """
        
        parser = MetadataParser()
        elements = parser.elements_from_string( 'function', xml_metadata )
        
        # If xml_metadata contains no fucntions with metadata, bail out.
        if len(elements) == 0:
            return None
            
        # Current implementation only supports one function, complain if needed.
        if len(elements) > 1:
            raise ValueError, "Input text contains more than one actor function."

        function_element = elements[0]
        
        return self.from_function_element( function_element, function )

    def from_function_element( self, function_element, function ):
        """ Returns an actor instance from the metadata function_element and
            the given function.
        """
        name = function_element.name
        help = function_element.help.text
        history = function_element.history.text
        
        input_bundles  = self._build_bundles( function_element.input_bundles,
                                    ConvergeInputBundle, ConvergeInputPort )
        output_bundles = self._build_bundles( function_element.output_bundles,
                                    ConvergeOutputBundle, ConvergeOutputPort )

        actor = ConvergeActor(  name=name, chip_name=name, 
                                description=function_element.description,
                                help=help,
                                history=history,
                                is_hidden = function_element.is_hidden,
                                class_input_bundles=input_bundles, 
                                class_output_bundles=output_bundles,
                                function=(function,))
        
        return actor
        
    ###########################################################################
    # Private interface.
    ###########################################################################

    def _build_bundles( self, meta_bundles, bundleClass, portClass ):
        """ return a list of ConvergeBundle from the BundleElement list.
        """
        
        bundles = []
        for meta_bundle in meta_bundles:
            ports = []
            for meta_arg in meta_bundle.args:
                port = self._port_from_arg( meta_arg, portClass )
                ports.append( port )
            
            name = meta_bundle.name
            # Hackish 
            if meta_bundle.args[0].repeat_group is not None:
                rg = meta_bundle.args[0].repeat_group
                bundle = ConvergeRepeatBundle( name=name,
                    names=rg.names, min_num=rg.min_num, max_num=rg.max_num,
                    ports=ports)
                    
            else:
                bundle = bundleClass( name=name, ports=ports )
                
            bundles.append( bundle)
        
        return bundles
    
    def _port_from_arg( self, arg, klass):
        """ Return port instance of klass configured from arg """
        return klass( name         = arg.composite_name,
                    basename       = arg.name,
                    modifier       = arg.modifier,
                    bundle         = arg.bundle,
                    description    = arg.description,
                    units          = arg.units,
                    family_name    = arg.family_name,
                    units_system_name = arg.units_system_name,
                    is_index       = arg.is_index,
                    is_optional    = arg.is_optional,
                    arg_name       = arg.arg_name,
                    index          = arg.position,
                    value          = 1.0,
                    external_type  = arg.external_type,
                    internal_type  = arg.internal_type )
        
        
    def _read_file(self, file_name):
        """ read in the contents of the file specified by fileName """
        
        file = open(file_name, "r")
        
        try:
            file_string = file.read()
        finally:
            file.close()
            
        return file_string
        
    
    def _get_callable( self, text, name ):
        """ Return the callable function named 'name' defined in 'text'. 
        """
        
        text = text.replace('\r\n', '\n')
        text = text + '\n'    
                
        exec text in globals()
                
        function = eval('%s' % name)
        
        return function
        
        
#### EOF ######################################################################
    