import os

from enthought.envisage import get_application

from enthought.traits.ui.api import Handler, message

from enthought.graph.function.xml_element_names import K_ARG_TAG

from enthought.logger import logger


class MetadataDefaultsHandler(Handler):
    """ Handler for adding a new default to the metadata.
    """
    
    def add_default(self, editor):
        """ Add a new default to the metadata
        
            Get the argument selected in the editor and create register new default
        """
        
        try:  
            # Get the argument element from the editor
            argument_element = editor.row_object.clone_traits()
            
            
            # Set repeat group to none.  Don't want this being written to the xml.
            argument_element.repeat_group = None
            # TODO: need to put this back in at some point.  Defaults are not
            # currently supported by the app and it causes problems when it is written out.
            argument_element.default = None
            
            argument_element.position = 0
            
            if self._argument_exists(argument_element):
                if self._display_overwrite_dialog():
                    # Write the default to the defaults.xml file
                    self._write_argument(argument_element)
                    # Bind the default
                    self._bind_argument(argument_element)
            else:
                # Write the default to the defaults.xml file
                self._write_argument(argument_element)
                # Bind the default
                self._bind_argument(argument_element)
                
                self._display_ok_dialog()

        except:
            # Catching and logging here as something further
            # up the chain seems to lose the exceptions otherwise.
            logger.exception('Error in metadata defaults handler.') 
    
        return
          
    def _argument_exists(self, argument_element):
        """ Check if argument already exists in xml metadata.
        """     
        
        default_service = get_application().service_registry.get_service('enthought.graph.function.MetadataDefaults')
        
        default = default_service.lookup(K_ARG_TAG, argument_element.arg_name, exact_match = True)
                
        if default is not None:
            return True
            
        return False
        
    def _display_overwrite_dialog(self):
        """ Present a dialog checking if it is ok to overwrite existing default.
        """
        
        result = message(message = 'Metadata default already exists and will be overwritten.  Do you want to continue?', 
                             title = 'Warning', buttons = [ 'OK', 'Cancel' ])
                         
        return result
                                   
    def _display_ok_dialog(self):
        """ Show dialog indicating metadata update successful
        """
        
        message(message = 'Metadata default added successfully.', title = 'Add Successful',
                buttons = ['OK'])
                
        return
                
    def _write_argument(self, argument_element):
        """ detemine defaults file  location and write defaults out.
        """
        
        # Find the defaults.xml file location
        from enthought.graph.vpl.graph_editor_plugin import GraphEditorPlugin
        graph_editor_plugin = GraphEditorPlugin.instance
        default_file_location = os.path.abspath(graph_editor_plugin.state_location + '/defaults.xml')
        
        # load the default element from file
        default_element = self._load_from_filename(default_file_location)
        
        # add newly created argument to the elements.
        self._add_argument_to_elements(argument_element, default_element)
                
        # write the defaults to file
        self._write_to_file(default_element, default_file_location)
         
        return
        
    def _load_from_filename(self, filename):
        """ from the specified filename load defaults """
        
        from metadata_parser import MetadataParser
        parser = MetadataParser()
        
        # load from file if it exists.
        if os.path.exists(filename):
            elements = parser.elements_from_file('default', filename )
        else:
            # create an empty default xml.
            elements = parser.elements_from_string('default', '<default>\n</default>')
            
        return elements[0]    
        
    def _add_argument_to_elements(self, argument_element, default_element):
        """ Add new default argument to element.
        """
        
        # Remove the argument if it already exists.
        self._remove_argument_from_elements(argument_element, default_element)
                
        # Add the argument to the end.
        default_element.default_args.append(argument_element)
        
        return
        
    def _remove_argument_from_elements(self, argument_element, default_element):
        """ Remove argument from elements if it already exists.
        
            We don't want to write it out twice.
        """
               
        for arg in default_element.default_args:
            if argument_element.name == arg.name:
                default_element.default_args.remove(arg)
                break
        
        return
        
    def _write_to_file(self, default_element, filename):
        """ Open file and write out xml.
        """
        
        file = open(filename, 'w')
        
        try:
            file.write(default_element.as_xml())
        finally:
            file.close()
        
        return
        
    def _bind_argument(self, argument_element):
        """ Bind new argument to the default service.
        """
        
        default_service = get_application().service_registry.get_service('enthought.graph.function.MetadataDefaults') 
        default_service.bind(argument_element)
        
        return
        
#### EOF ######################################################################
