""" Element definitions for function xml metadata actor definitions. """

import sys

from enthought.traits.api import Enum, HasTraits, Instance, CInt, List, Any, \
    Property, Str
    
from enthought.traits.ui.api import View, Group, Item, TableEditor
from enthought.traits.ui.menu import Menu, Action

from enthought.graph.vpl.converge_actor import bundle_suffix_map

from enthought.traits.ui.table_column \
    import ObjectColumn

from metadata_defaults_handler import MetadataDefaultsHandler

from traits import Bool
from xml_element_names import *

class NodeElement( HasTraits ):
    """ Traits class that represents a DOM node. """
    
    # XML Element tag name
    tag_name = "node"
    
    # Reference to the Element containing this element.
    parent = Instance( 'NodeElement' )
    
    def set_from_node( self, node ):
        """ Set traits from XML node attribute name value pairs. """
        traits = {}
        for name, value in node.attrib.items():
            traits[ str(name) ] = str(value).strip()
        
        self.set( **traits )
        
        return self
    
    def set_from_object(self, obj):    
        """ Set element attributes from obj if defined on the object """
        names = []
        for attr_name in self.trait_names():
            if hasattr( obj, attr_name ):
                names.append( attr_name )

        self.set( **obj.get( names ) )
        
    def equals(self, obj, names=None):    
        """ Set element attributes from obj if defined on the object """
        if obj is self:
            return True
        
        if names == None:
            names = self.trait_names()
        
        for attr_name in names:
            if hasattr( obj, attr_name ):
                if self.get(attr_name)[attr_name] != obj.get(attr_name)[attr_name]:
                    return False
            else:
                return False

        return True
    
    def as_xml(self, attr_names=None):
        """ Returns XML string representation of the element.
        Specify a list of atribute names to include as a list.  If none
        is specified, all traits are included.
        """
        xml = '<' + self.tag_name + ' '
        xml += self.format_xml_attributes(attr_names=attr_names)
        xml += '/>\n'
        
        return xml

    def format_xml_attributes(self, attr_names=None):
        """ Returns a string with the attributes formated for proper XML. """
        xml = ''
        
        if attr_names == None:
            attr_names = self.trait_names()
# not a trait           attr_names.remove('tag_name')
        
        for attr_name in attr_names:
            if hasattr( self, attr_name ) and self.get(attr_name)[attr_name] is not None:
                attr_str = "%s='%s'" % (attr_name, self.get(attr_name)[attr_name])
                attr_str = '%-25s ' % attr_str
                xml += attr_str
        
        return xml
        
        
class AbstractLeafElement( NodeElement ):
    """ A simple metadata element that contains a single string """

    # XML Element tag name
    tag_name = "abstract_leaf"

    # The content of the element.
    text = Str

    def set_from_node( self, node ):
        """ Set traits from node attribute name value pairs. """
        super( AbstractLeafElement, self ).set_from_node( node )
        
        traits = {}
        traits['text'] = node.text.strip()
        
        self.set( **traits )
        
        return self
        

class RepeatGroupElement( NodeElement ):
    """ Metadata that describes a repeating group of inputs or outputs """
    
    # XML Element tag name
    tag_name = K_REPEAT_TAG
    
    # Names sequentially used for naming the instances of the repeat group.
    names = List(Str)
    
    # Minimum number of occurrances of the repeat group.
    min_num = CInt(1)
    
    # Maximum number of occurrances of the repeat group.
    max_num = CInt(sys.maxint)

    def set_from_node( self, node ):
        """ Set traits from node attribute name value pairs. """

        # The DOM attribute is a string of the form 'name1 name2 ...'
        # The names trait won't like that, take it out and handle separately
        if 'names' in node.attrib:
            names_attr = node.attrib['names']
            del node.attrib['names']
        else:
            names_attr = None
            
        super( RepeatGroupElement, self ).set_from_node( node )
        
        if names_attr is not None:
            node.attrib['names'] = names_attr
            self.names = str(names_attr).split()
            
        if self.max_num < 0:
            self.max_num = sys.maxint
            
        return self
    
class ArgumentElement( NodeElement ):
    """ Metadata that describes a single function input argument or output
    and corresponding actor port.
    """
    
    # XML Element tag name
    tag_name = K_ARG_TAG
    
    # Basename: e.g. vp, k, rhob, the 'k' in k_wat_ss.
    name = Str
    
    # Name modifier: e.g. 'wat' in k_wat_ss.
    modifier = Str
    
    # Bundle name: e.g. 'ss' in k_wat_ss.
    bundle = Str
    
    # Units: e.g. km/s.
    units = Str
    
    # Units family name: e.g. pvelocity.
    family_name = Str
    
    # Units system name: e.g. 'KGS'
    units_system_name = Str('KGS')
    
    # Short description of argument purpose.
    description = Str
    
    # Type of the value expected (input) or produced (output) by the actor port.
    external_type = Str
    
    # Type of the value expected (input) or produced (input) by the function.
    internal_type = Str
    
    # Is the argument an index?  
    # FIXME: Probably don't need this: external_type == LogIndex
    is_index = Bool()
    
    # Is the argument optional?
    is_optional = Bool()
    
    # Name of the function input argument
    arg_name = Str
    
    # Argument position in function call or return list.
    position = CInt
    
    # Composite name: i.e. name_modifier_bundle.
    composite_name = Property
    
    # Default value for the argument.
    default = Any
    
    # Reference to the repeat group info.
    repeat_group = Instance( RepeatGroupElement, None )
    
    # This object contained in a list.
    list_representation = Property

    def set_from_node(self, node, repeat_group):
        """ Set traits from node attribute name value pairs. """
        super( ArgumentElement, self ).set_from_node( node )
        self.repeat_group = repeat_group
        
        if self.bundle == '':
            if self.modifier == '':
                self.bundle = self.name
            else:
                self.bundle = self.modifier
                self.modifier = ''
                
        self.units = self.units.lower()
        return self

    def as_xml(self, attr_names=None):
        """ Returns XML string representation of the element.
        """

        if attr_names == None:
            # default order is random use explicit list for now...
            # attr_names = self.trait_names()
            # not a trait           attr_names.remove('tag_name')
            # attr_names.remove('parent')
            # attr_names.remove('composite_name')
            # attr_names.remove('trait_added')
            # attr_names.remove('repeat_group')
            # attr_names.remove('list_representation')
            attr_names = [
                'arg_name', 'name', 'modifier', 'bundle',
                'position', 
                'external_type', 
                'internal_type', 
                'description',
                'units', 
                'family_name', 
                'units_system_name',
                'default', 
                'is_optional', 
                'repeat_group', 
                'is_index', 
                ]
            
        return super( ArgumentElement, self ).as_xml( attr_names=attr_names )
            
    def _get_composite_name( self ):
        """ Returns the full composite name: name_modifier_bundle """
        name = self.name
        if self.modifier is not None and len(self.modifier) > 0:
            name += '_' + self.modifier
        if (self.bundle is not None and len(self.bundle) > 0
            and self.bundle != self.name):
            name += '_' + self.bundle
            
        return name

    def _set_composite_name(self, name):
        """ Set name, modifier, and bundle names by splitting the provided name.
        """
        split_name = name.split('_')
        
        # only name part specified, duplicate for bundle name
        if len(split_name) == 1:
            split_name += ['', split_name[0]]
        
        # name and bundle specified, insert an empty modifier
        elif len(split_name) == 2:
            split_name.insert(1, '')
              
        self.name, self.modifier, self.bundle = split_name
            
    def _get_list_representation( self ):
        """ Return this object in a list """
        
        list = []
        list.append(self)
        
        return list
    
    
class HistoryElement( AbstractLeafElement ):
    """ Change history component for FunctionElement """
    # XML Element tag name
    tag_name = K_HISTORY_TAG
    
class HelpElement( AbstractLeafElement ):
    """ Usage description component for FunctionElement """
    # XML Element tag name
    tag_name = K_HELP_TAG
    
class BundleElement( NodeElement ):
    """ Element for Bundle definition """
    # XML Element tag name
    tag_name = K_ARG_BUNDLE
    
    # Name: e.g. 'sand'
    name = Str
    
    # A list of Arguments within the bundle.
    args = List( ArgumentElement )  
    
class FunctionElement( NodeElement ):
    """ Metadata that describes a function that can be converted to an Actor """

    # XML Element tag name
    tag_name = K_FCN_TAG
    
    # Name of the function as presented to the user.
    name = Str
    
    # Short description of function's purpose.
    description = Str

    # Function usage description.
    help = Instance( HelpElement )
    
    # Modification history.
    history = Instance( HistoryElement )
    
    # Should the function normally be present in user interfaces?
    is_hidden = Bool()
    
    # Name of the associated python function.
    function_name = Str
    
    # Function inputs.
    input_args = List( ArgumentElement )
    
    # Funtion inputs grouped by bundle
    input_bundles = List( BundleElement )
    
    # Function outputs.
    output_args = List( ArgumentElement )
    
    # Funtion outputs grouped by bundle
    output_bundles = List( BundleElement )
    
    # All bundle argument elements
    all_arguments = Property
    
    # Only bundle input arguments
    inputs = Property
    
    # Only bundle output arguments
    outputs = Property
    
    add_default_menu = Menu(
        Action( name         = 'Add Default',
                action       = 'handler.add_default(editor)'),
        name = 'popup'
    )
    
    argument_columns = [
        ObjectColumn( name      = 'arg_name', 
                      label     = 'Arg Name', 
                      editable  = False,
                      menu      = add_default_menu 
                      ),
        ObjectColumn( name      = 'default',
                      label     = 'Default',
                      editable  = True,
                      menu      = add_default_menu),
        ObjectColumn( name      = 'description', 
                      label     = 'Description', 
                      editable  = True,
                      menu      = add_default_menu ),     
        ObjectColumn( name      = 'name', 
                      label     = 'Port Name', 
                      editable  = True,
                      menu      = add_default_menu),
        ObjectColumn( name      = 'modifier', 
                      label     = 'Modifier', 
                      editable  = False,
                      menu      = add_default_menu ),
        ObjectColumn( name      = 'external_type', 
                      label     = 'External Type', 
                      editable  = True,
                      menu      = add_default_menu ),
        ObjectColumn( name      = 'internal_type', 
                      label     = 'Internal Type', 
                      editable  = True,
                      menu      = add_default_menu),
        ObjectColumn( name      = 'units', 
                      label     = 'Units', 
                      editable  = True,
                      menu      = add_default_menu ), 
        ObjectColumn( name      = 'family_name', 
                      label     = 'Measure Of', 
                      editable  = True,
                      menu      = add_default_menu ),
        ObjectColumn( name      = 'units_system_name', 
                      label     = 'Units System', 
                      editable  = True,
                      menu      = add_default_menu ),
        ObjectColumn( name      = 'bundle',
                      label     = 'Bundle',
                      editable  = False,
                      menu      = add_default_menu )
    ]    
    
    argument_editor = TableEditor(
        columns       = argument_columns,
        editable      = True,
        deletable     = False,
        orientation   = 'vertical',
        sortable      = False,
    )
     
    traits_view = View( 
        Group( 
#            Group( 
#                Item( 'name' ),
#            ),
            Group( 
                Group(
                    Item( 'input_args', 
                          editor = argument_editor ), 
                    label       = 'Inputs', 
                    export      = 'DockShellWindow',
                    show_labels = False
                ),
                Group(
                    Item( 'output_args', 
                          editor = argument_editor ), 
                    label       = 'Outputs', 
                    export      = 'DockShellWindow',
                    show_labels = False
                ),
                id          = 'splitter2',
                orientation = 'horizontal',
                layout      = 'split',
                show_labels = False,
                padding     = 10
            )
        ),
        id   = 'enthought.graph.function.metadata_elements',
        dock = 'tab',
        handler = MetadataDefaultsHandler()
    )
                                   
    def __init__( self, **traits ):
        self.help = HelpElement( text='Help wanted.' )
        self.history = HistoryElement( text='History unknown.' )
        super( FunctionElement, self ).__init__( **traits )
        return
        
    def build_bundle_elements( self ):
        """ Build bundle elements for input and output args.
        """
        self.input_bundles  = self._build_bundle_elements( self.input_args )
        self.output_bundles = self._build_bundle_elements( self.output_args )
        return
        
    def as_xml(self, attr_names=None):
        """ Returns XML string representation of the element.
        """
        if attr_names == None:
            attr_names = self.trait_names()
# not a trait           attr_names.remove('tag_name')
            attr_names.remove('help')
            attr_names.remove('history')
            attr_names.remove('inputs')
            attr_names.remove('input_args')
            attr_names.remove('input_bundles')
            attr_names.remove('outputs')
            attr_names.remove('output_args')
            attr_names.remove('output_bundles')
            attr_names.remove('all_arguments')

        xml = '<' + self.tag_name + ' '
        xml += self.format_xml_attributes(attr_names=attr_names)
        xml += ' >\n'
        
        xml += '<inputs>\n'
        for arg in self.inputs:
            xml += '    ' + arg.as_xml()
        xml += '<inputs />\n'
        
        xml += '<outputs>\n'
        for arg in self.inputs:
            xml += '    ' + arg.as_xml()
        xml += '<outputs />\n'
        
        xml += '<' + self.tag_name + ' />\n'
        return xml

    def _build_bundle_elements( self, args ):
        """ Returns a list of bundles containing the args with a common bundle
        name.
        """
        bundle_dict = {}
        
        for arg in args:
            name = arg.bundle
            if bundle_suffix_map.has_key(name):
                name = bundle_suffix_map[name]
            bundle = bundle_dict.setdefault( name, BundleElement( name=name ) )
            bundle.args.append( arg )

        keys = bundle_dict.keys()
        keys.sort()
        bundles = [ bundle_dict[key] for key in keys ]
        return bundles
        
    def _get_all_arguments( self ):
        """ Returns a list of all argument elements wrapped in bundles
        """
        
        args = []
        
        for bundle in self.input_bundles:
            if bundle.args is not None and len(bundle.args) > 0:
                args.extend(bundle.args)
                
        for bundle in self.output_bundles:
            if bundle.args is not None and len(bundle.args) > 0:
                args.extend(bundle.args)
                
        return args
        
    def _get_inputs( self ):
        """ Returns a list of all input argument elements wrapped in bundles
        """
        
        args = []
        
        for bundle in self.input_bundles:
            if bundle.args is not None and len(bundle.args) > 0:
                args.extend(bundle.args)
                
        return args
        
    def _get_outputs( self ):
        """ Returns a list of all output argument elements wrapped in bundles
        """
        
        args = []
        
        for bundle in self.output_bundles:
            if bundle.args is not None and len(bundle.args) > 0:
                args.extend(bundle.args)
                
        return args


class DefaultElement( NodeElement ):
    """ Top level DEFAULT tag for encoding default ARGUMENTS in XML. """

    # XML Element tag name
    tag_name = K_DEFAULT_TAG

    # Argument default definitions.
    default_args = List( ArgumentElement )

    def as_xml(self, attr_names=None):
        """ Returns XML string representation of the element.
        """
        if attr_names == None:
            attr_names = self.trait_names()
            attr_names.remove('default_args')
# not a trait           attr_names.remove('tag_name')

        xml = '<' + self.tag_name
        
        attribs = self.format_xml_attributes(attr_names=attr_names)
        if attribs is not None and len(attribs) > 0:
            xml += ' '
            xml += attribs
        
        xml += '>\n'
        
        for arg in self.default_args:
            xml += '    ' + arg.as_xml()

        xml += '</' + self.tag_name + '>\n'
        return xml
    

#### EOF ######################################################################
