""" MetadataParser - parser for function xml metadata actor definitions """

import sys
import re

import cElementTree as ElementTree

from xml_element_names import *

from traits import Bool
from metadata_elements import \
    NodeElement, AbstractLeafElement, RepeatGroupElement, ArgumentElement, \
    HistoryElement, HelpElement, FunctionElement, DefaultElement


class ParseException( Exception ):
    def __init__( self, ex, text, msg=None ):
        self.ex = ex
        self.text = text
        self.msg = msg
        return
    
    def __str__( self ):
        if isinstance( self.ex, SAXParseException ):
            line = self.ex.getLineNumber()
            text = ' line: %s column: %s\noffending line: %s\n%s' \
            % (line, self.sax_exception.getColumnNumber(), \
                self.text.split('\n')[line-1], self.ex.getMessage())
        else:
            text = self.msg 
            if self.ex:
                text += ';'.join(self.ex.args)
            text += "\n\"" + self.text + "\""

        return text
            
class MetadataParser:
    """ Parser for xml encoded function metadata. """
    
    def extract_from_file( self, outer_tag_name, filename ):
        """ returns list of xml strings bounded by <function> tags from file.
        """
        # open file containing function(s) with xml metadata
        file = open(filename)
        text = file.read()
        file.close()
        
        return self.extract_from_string( text, outer_tag_name )
        
    def extract_from_string( self, outer_tag_name, text ):
        """ returns list of xml strings bounded by <tag_name> tags from s.
        """

        # set up regular expression to find <function> elements in file
        s = r'<%s(.*?)</%s>' % (outer_tag_name, outer_tag_name)
        func_pat = re.compile( s, re.DOTALL)
    
        # now apply the regular expression to the function text
        matches = func_pat.findall(text)
        
        xml = [ ''.join(["<", outer_tag_name, match, "</", outer_tag_name, ">"])
                for match in matches ]
                
        return xml
        
    def elements_from_file(self, outer_tag_name, filename) :
        """ Returns a list of FunctionElements from the xml stored in filename.
        """
        # open file containing function(s) with xml metadata
        file = open(filename)
        try:
            text = file.read()
        finally:
            file.close()
        
        return self.elements_from_string( outer_tag_name, text )
        
    def elements_from_string(self, outer_tag_name, text) :
        """ Returns a FunctionElement from the xml encoded in str.
        """
        xmls = self.extract_from_string( outer_tag_name, text )
        
        elements = []
        
        for xml in xmls:
            try:
                root = ElementTree.XML( xml )
                doc_elements = self.elements_from_root( root )
            except SyntaxError, e:
                raise ParseException( e, xml )
            except ParseException, e:
                e.text = text
                raise
                
            elements.extend( doc_elements )

        return elements
        
    def elements_from_root(self, root):
        """ Walk the elements beginning at root and return the FunctionElement.
        """
       
        elements = []
        
        root_tag = root.tag.lower()
        if root_tag == K_FCN_TAG:
            function_element = self._process_function_element(root)
            elements.append(function_element)
        elif root_tag == K_DEFAULT_TAG:
            default_element = self._process_default_element(root)
            elements.append(default_element)
        else:
            raise ParseException(None, None,
            "unrecognized element: top element must be <function> or <default>.")
        
        return elements

    def _process_function_element(self, root):
        """ Process a <function> root element.
        """
        function_element = FunctionElement().set_from_node(root)
        
       
        for node in root:
            node_tag = node.tag.lower()
            if node_tag == K_HISTORY_TAG :
                function_element.history = HistoryElement().set_from_node(node) 

            elif node_tag == K_HELP_TAG :
                function_element.help = HelpElement().set_from_node(node)

            elif node_tag == K_INPUTS_TAG :
                self._process_args(node, function_element.input_args)
                
            elif node_tag == K_OUTPUTS_TAG :
                self._process_args(node, function_element.output_args)

            else:
                break
        
        function_element.build_bundle_elements()
        
        return function_element


    def _process_default_element(self, root):
        """ Process a <default> root element.
        """
        default_element = DefaultElement().set_from_node(root)
        
        self._process_args(root, default_element.default_args,
            allow_repeats=False)
        return default_element    
        

    def _process_args(self, node, elements, position=0, allow_repeats=True,
                repeat_group=None):
        """ Extract input or output elements from walker and append to elements.
        """
        
        for child in node:
            child_tag = child.tag.lower()
            if child_tag == K_REPEAT_TAG and allow_repeats:
                repeat_group = RepeatGroupElement().set_from_node(child)
                self._process_args(child, elements, position=position,
                    allow_repeats=False,
                    repeat_group=repeat_group)

            elif child_tag == K_ARG_TAG :
                arg_info = ArgumentElement( position=position )
                position += 1
                arg_info.set_from_node(child, repeat_group)
                
                elements.append( arg_info )

            else:
                break;
                
        return 
        


        
#### EOF ######################################################################
    