from enthought.util.scipyx \
        import amax, amin, arange, array, zeros, ones, Float, matrixmultiply, \
               where, sqrt, compress

from enthought.util.numeric \
        import safe_len

import types
import copy


def backus_vertical (*in_args) :
    """
    <function function_name='backus_vertical'
              name='backus_vertical' 
              category='Elasticity'
              description='Compute vertical velocity Backus average of
                           n lithologies'
              is_hidden='true' >

        <history>
            2003/03/10  rjb  Changed parameter names to family names
            2003/03/10  rjb  Changed fcn name to match file name
            2003/08/07  rjb  Changed REPEATS to REPEAT
            2005/04/06  wjd  Convert to Python
        </history>

        <inputs>
            <repeat_group names='1 2 3'
                          min_num='1'
                          max_num='-1' >
                <arg   name='x'
                       modifier='lith'
                       bundle=''
                       units='V/V'
                       description='Lamination fraction of lithology n'
                       family_name='volume_frac'
                       external_type='Log'
                       internal_type='Array'
                       is_index='false'
                       is_optional='false'
                       arg_name='x_lith'/>
                <arg   name='vp'
                       modifier='lith'
                       bundle=''
                       units='KM/S'
                       description='P-velocity of lithology n'
                       family_name='pvelocity'
                       external_type='Log'
                       internal_type='Array'
                       is_index='false'
                       is_optional='false'
                       arg_name='vp_lith'/>
                <arg   name='vs'
                       modifier='lith'
                       bundle=''
                       units='KM/S'
                       description='S-velocity of lithology n'
                       family_name='svelocity'
                       external_type='Log'
                       internal_type='Array'
                       is_index='false'
                       is_optional='false'
                       arg_name='vs_lith'/>
                <arg   name='rhob'
                       modifier='lith'
                       bundle=''
                       units='G/CC'
                       description='Bulk density of lithology n'
                       family_name='density'
                       external_type='Log'
                       internal_type='Array'
                       is_index='false'
                       is_optional='false'
                       arg_name='rhob_lith'/>
            </repeat_group>
        </inputs>

        <outputs>
            <arg   name='vp'
                   modifier=''
                   bundle='bck'
                   units='KM/S'
                   description='Vertical P-velocity of wet rock'
                   family_name='pvelocity'
                   external_type='Log'
                   internal_type='Array'
                   is_index='false'
                   is_optional='false'
                   arg_name='vp_bck'/>
            <arg   name='vs'
                   modifier=''
                   bundle='bck'
                   units='KM/S'
                   description='Vertical S-velocity of wet rock'
                   family_name='svelocity'
                   external_type='Log'
                   internal_type='Array'
                   is_index='false'
                   is_optional='false'
                   arg_name='vs_bck'/>
            <arg   name='rhob'
                   modifier=''
                   bundle='bck'
                   units='G/CC'
                   description='Bulk density laminated solid'
                   family_name='density'
                   external_type='Log'
                   internal_type='Array'
                   is_index='false'
                   is_optional='false'
                   arg_name='rhob_bck'/>
        </outputs>

        <help>
            backus_vertical:
            Compute the vertical velocity Backus average of n lithologies
        </help>

    </function>
    """

    in_parms_per_group = 4
    num_in_groups = len(in_args)//in_parms_per_group
    # check to make sure there is at least one group of input parameters
    if num_in_groups == 0 :
        raise TypeError, 'Must provide minimum of one group of inputs'

    # check to make sure input parameters are in groups of in_parms_per_group
    if len(in_args) % in_parms_per_group != 0 :
        raise TypeError, 'Must provide input parameters in groups of %s' % \
            in_parms_per_group

    # accumulate m "compliance"
    cbm = zeros(safe_len(in_args[1]), Float)
    # accumulate shear "compliance"
    cbu = zeros(safe_len(in_args[1]), Float)
    # accumulate bulk density
    crhob = zeros(safe_len(in_args[1]), Float)
    xtot = zeros(safe_len(in_args[1]), Float)
    
    for k in range(0, len(in_args), in_parms_per_group) :
        x_lith    = in_args[k]
        vp_lith   = in_args[k+1]
        vs_lith   = in_args[k+2]
        rhob_lith = in_args[k+3]

        m = rhob_lith * vp_lith * vp_lith
        u = rhob_lith * vs_lith * vs_lith

        # accumulate sum of reciprocals for moduli
        xtot = xtot + x_lith
        cbm = cbm + x_lith/m
        cbu = cbu + x_lith/u
        crhob = crhob + x_lith * rhob_lith

    # prevent division by zero
    if safe_len(xtot) > 1 :
        iz = compress(xtot == 0.0, arange(len(xtot)))
        for i1 in range(len(iz)) :
            xtot[iz[i1]] = 1.
        iz = compress(cbm == 0.0, arange(len(cbm)))
        for i1 in range(len(iz)) :
            cbm[iz[i1]] = 1.
        iz = compress(cbu == 0.0, arange(len(cbu)))
        for i1 in range(len(iz)) :
            cbu[iz[i1]] = 1.
    else :
        if xtot == 0.0 :
            xtot = array(( 1. ))
        if cbm == 0.0 :
            cbm = array(( 1. ))
        if cbu == 0.0 :
            cbu = array(( 1. ))

    mavg = xtot/cbm            # backus average m modulus
    uavg = xtot/cbu            # backus average u modulus
    rhob_bck = crhob/xtot      # average density

    if safe_len(rhob_bck) > 1 :
        i0 = compress(rhob_bck == 0.0, arange(len(rhob_bck)))
        for i1 in range(len(i0)) :
            rhob_bck[i0[i1]] = 1.
    else :
        set_rhob_to_1 = False
        if rhob_bck == 0.0 :
            rhob_bck = array(( 1. ))
            set_rhob_to_1 = True

    vp_bck = sqrt(mavg/rhob_bck)
    vs_bck = sqrt(uavg/rhob_bck)

    if safe_len(rhob_bck) > 1 :
        for i1 in range(len(i0)) :
            vp_bck[i0[i1]] = 0.
            vs_bck[i0[i1]] = 0.
            rhob_bck[i0[i1]] = 0.
    else :
        if set_rhob_to_1 == True :
            vp_bck = array(( 0. ))
            vs_bck = array(( 0. ))
            rhob_bck = array(( 0. ))

    return (vp_bck, vs_bck, rhob_bck)

