from enthought.util.scipyx \
        import amax, amin, arange, array, zeros, ones, Float, matrixmultiply, \
               where, sqrt, compress

from enthought.util.numeric \
        import safe_len

import types
import copy


def hill_average(*in_args):
    """
    <function function_name='hill_average'
              name='Hill Average' 
              category='Elasticity'
              description='Compute the Hill (mean of Voigt and Reuss) Average given
                           a set of properties and volume fractions'
              is_hidden='true' >

        <history>
            2003/05/16 rjb Changed input name to match family names
            2004/11/15 mlr Convert to python
            2005/03/02 wjd Add check of input parameter count
        </history>

        <inputs>
            <repeat_group names='1 2 3'
                          min_num='1'
                          max_num='-1' >
                <arg   name='x'
                       modifier=''
                       bundle=''
                       units='V/V'
                       description='volume fraction'
                       family_name='volume_frac'
                       external_type='Log'
                       internal_type='Array'
                       is_index='false'
                       is_optional='false'
                       arg_name='x'/>
                <arg   name='m'
                       modifier=''
                       bundle=''
                       units='GPa'
                       description='property to be averaged'
                       family_name='modulus'
                       external_type='Log'
                       internal_type='Array'
                       is_index='false'
                       is_optional='false'
                       arg_name='m'/>
            </repeat_group>
        </inputs>

        <outputs>
            <arg   name='m'
                   modifier='avg'
                   bundle=''
                   units='GPa'
                   description='Hill average of property'
                   family_name='modulus'
                   external_type='Log'
                   internal_type='Array'
                   is_index='false'
                   is_optional='false'
                   arg_name='m_avg'/>
        </outputs>

        <help>
            Hill Modulus Average
            Constructs the Hill average of a property based on volume fractions.
            Hill is the mean of the Voigt and Reuss averages.
 
            Note: This function does not ensure that mineral fractions total
                  to one.
        </help>

    </function>
    """

    inparms_per_group = 2
    n_groups_in = len(in_args)//inparms_per_group
    # check to make sure there is at least one group of input parameters
    if n_groups_in == 0 :
        raise TypeError, 'Must provide minimum of one group of inputs'

    # check to make sure input parameters are in groups of inparms_per_group
    if len(in_args) % inparms_per_group != 0 :
        raise TypeError, 'Must provide input parameters in groups of %s' % \
            inparms_per_group

    mreuss = zeros(safe_len(in_args[1]), Float)
    mvoigt = zeros(safe_len(in_args[1]), Float)

    for i1 in range(0, len(in_args), inparms_per_group):
        x = in_args[i1]
        m = in_args[i1+1]
        mreuss = mreuss + x/m
        mvoigt = mvoigt + x*m
    
    # If we are only mixing two phases at a single measurement point we 
    # don't need to take the performance hit of instantiating the arrays 
    # and we can unroll the loop over the phases ...
    if len(in_args) == 4 and type(in_args[0]) == types.FloatType:    
        frac1, mod1, frac2, mod2 = in_args
        mreuss = frac1/mod1 + frac2/mod2
        mvoigt = frac1*mod1 + frac2*mod2
    
    # ... otherwise we are mixing more than two phases and/or computing 
    # for an array of measurements
    else:     
        mreuss = zeros(safe_len(in_args[1]), Float)
        mvoigt = zeros(safe_len(in_args[1]), Float)
    
        for i1 in range(0, len(in_args), 2):
            x = in_args[i1]
            m = in_args[i1+1]
            mreuss = mreuss + x/m
            mvoigt = mvoigt + x*m
        
    mreuss = 1./mreuss
    m_avg = (mreuss + mvoigt)/2.

    return m_avg

