import unittest

from os.path import abspath, dirname, join

from enthought.traits.api import HasTraits, Str

from enthought.graph.function.metadata_elements import ArgumentElement


class MetadataDefaultsTestCase( unittest.TestCase ) :

    def setUp( self ):
        try:
            from enthought.graph.function import MetadataDefaults
            self.md = MetadataDefaults()
        except ImportError, e:
            import testoob.testing
            testoob.testing.skip("import error: " + str(e))
    
    def tearDown( self ):
        return
        

    def test_bind_new_tag( self ):
        """ test adding and retrieving the first metadata default for a tag. """
        arg = ArgumentElement( name='k', modifier='wat', bundle='sh',
                units='m', family_name='something')
        self.md.bind( arg )

        def_arg = self.md.lookup( arg.tag_name, 'k_wat_sh')
        
        self.assertNotEqual( def_arg, None )
        self.assertEqual( def_arg.name, 'k')
        self.assertEqual( def_arg.modifier, 'wat')
        self.assertEqual( def_arg.bundle, 'sh')
        self.assertEqual( def_arg.units, 'm')
        self.assertEqual( def_arg.family_name, 'something')
        
        return
    
    def test_bind_reuse_tag( self ):
        """ test adding and retrieving two defaults for a single tag. """
        arg = ArgumentElement( name='k', modifier='wat', bundle='sh',
                units='m', family_name='something')
        self.md.bind( arg )

        def_arg = self.md.lookup( arg.tag_name, 'k_wat_sh')
        self.assertNotEqual( def_arg, None )
        self.assertEqual( def_arg.name, 'k')
        self.assertEqual( def_arg.modifier, 'wat')
        self.assertEqual( def_arg.bundle, 'sh')
        self.assertEqual( def_arg.units, 'm')
        self.assertEqual( def_arg.family_name, 'something')
        
        arg = ArgumentElement( name='rhob', bundle='sh',
                units='m', family_name='something')
        self.md.bind( arg )

        def_arg = self.md.lookup( arg.tag_name, 'rhob_sh')
        self.assertNotEqual( def_arg, None )
        self.assertEqual( def_arg.name, 'rhob')
        self.assertEqual( def_arg.modifier, '')
        self.assertEqual( def_arg.bundle, 'sh')
        self.assertEqual( def_arg.units, 'm')
        self.assertEqual( def_arg.family_name, 'something')
        return

    def test_lookup_short_name(self):
        """ test retrieving a default defined with a name prefix. """
        arg = ArgumentElement( name='k',
                units='h', family_name='time')
        self.md.bind( arg )

        def_arg = self.md.lookup( arg.tag_name, 'k_wat_sh')
        
        self.assertNotEqual( def_arg, None )
        self.assertEqual( def_arg.name, 'k')
        self.assertEqual( def_arg.modifier, 'wat')
        self.assertEqual( def_arg.bundle, 'sh')
        self.assertEqual( def_arg.units, 'h')
        self.assertEqual( def_arg.family_name, 'time')
        
        return
        
    def test_lookup_no_name(self):
        """ test retrieving a default defined with no name. """
        arg = ArgumentElement( name='',
                units='f', family_name='otherthing')
        self.md.bind( arg )

        def_arg = self.md.lookup( arg.tag_name, 'k_wat_sh')
        
        self.assertNotEqual( def_arg, None )
        self.assertEqual( def_arg.name, 'k')
        self.assertEqual( def_arg.modifier, 'wat')
        self.assertEqual( def_arg.bundle, 'sh')
        self.assertEqual( def_arg.units, 'f')
        self.assertEqual( def_arg.family_name, 'otherthing')
        
        return

    def test_lookup_none_no_tag_bindings(self):
        """ test retrieving when no defaults are defined for the specified tag.
        """

        def_arg = self.md.lookup( 'blargle', 'k_wat_sh')
        self.assertEqual( def_arg, None )
        return
        
    def test_lookup_none(self):
        """ test retrieving a default when there are no matching defaults.
        """
        from enthought.graph.function import FunctionElement
        
        arg = FunctionElement( name='foo',
                category='cat', description='otherthing')
        
        self.md.bind( arg )
        def_arg = self.md.lookup( 'arg', 'foo')
        self.assertEqual( def_arg, None )
        return

    def test_unbind(self):
        """ Test removing a default """
        from enthought.graph.function import ArgumentElement
        
        arg = ArgumentElement( name='k', modifier='wat', bundle='sh',
                units='m', family_name='something')
        self.md.bind( arg )

        def_arg = self.md.lookup( arg.tag_name, 'k_wat_sh')
        self.assertNotEqual( def_arg, None )
        
        self.md.unbind( arg.tag_name, 'k_wat_sh' )
        
        def_arg = self.md.lookup( arg.tag_name, 'k_wat_sh')
        self.assertEqual( def_arg, None )
        
        return

    def test_parsing_from_file(self):
        """ Test parsing defaults from xml file. """
        from enthought.graph.function import MetadataParser
        filename = join( abspath(dirname(__file__)), "defaults.xml")
        
        parser = MetadataParser()
        elements = parser.elements_from_file( 'default', filename )
        
        self.assertEqual( 1, len(elements) )
        
        default_element = elements[0]
        self.assertEqual( default_element.tag_name, 'default' )
        self.assertEqual( len(default_element.default_args), 9)
        
        return

    def test_binding_from_file(self):
        """ Test loading and binding defaults from xml file. """
        
        filename = join( abspath(dirname(__file__)), "defaults.xml")
        
        self.md.load_from_filename( filename )
        def_arg = self.md.lookup( 'arg', 'watdth')
        self.assertNotEqual( def_arg, None )
        
        return
        
    def test_set_from_object(self):
        """ Test ArgumentElement.set_from_object """
        obj = ExampleObject( name='example', modifyer='mod', bundle='bund',
                            units='m/s', family_name='velocity')
        
        ae = ArgumentElement()
        ae.set_from_object( obj )
        self.assertEqual( ae.name, obj.name )
        self.assertEqual( ae.modifier, obj.modifier )
        self.assertEqual( ae.bundle, obj.bundle )
        self.assertEqual( ae.units, obj.units )
        self.assertEqual( ae.family_name, obj.family_name )
        
        return
        
    def test_equality(self):
        """ Test metadata element equals """
        from enthought.graph.function import ArgumentElement
        ae = ArgumentElement( name='example', modifyer='mod', bundle='bund',
                            units='m/s', family_name='velocity')

        ae2 = ArgumentElement( name='example', modifyer='mod', bundle='bund',
                            units='m/s', family_name='velocity')
        
        names = ['is_index', 'units_system_name', 'internal_type', 
                 'description', 'family_name', 'default', 
                 'bundle', 'external_type', 'units', 
                 'is_optional', 'position', 'modifier',
                 'repeat_group', 'arg_name', 'name']
        self.assert_( ae.equals(ae2, names=names) )

        ae2.name = 'different'
        self.assert_( not ae.equals(ae2, names=names) )
        
        return
        
class ExampleObject(HasTraits):
    """ Class for test_set_from_object """
    name = Str
    
    # Name modifier: e.g. 'wat' in k_wat_ss.
    modifier = Str
    
    # Bundle name: e.g. 'ss' in k_wat_ss.
    bundle = Str
    
    # Units: e.g. km/s.
    units = Str
    
    # Units family name: e.g. pvelocity.
    family_name = Str
    


### EOF #######################################################################
