from enthought.graph.function.xml_element_names import *
from enthought.graph.function.metadata_parser import MetadataParser

class XmlMetadataWriter:
    """ Write the XML metadata, from an object model description of the XML hierarchy.
    """ 
    
    ###########################################################################
    # Public interface.
    ###########################################################################
    
    def write(self, function):
        """ Write the XML metadata, from an object model description of the XML hierarchy.
        """ 
        
        self.xml = ''
        
        self._write_function(function)
        
        return self.xml
        
    ###########################################################################
    # Private interface.
    ###########################################################################
        
    def _write_function(self, function):
        """ Write the function element for the XML metadata.
        """ 

        if function is not None:
            indent_level = 1
            
            self.xml += self._indent(indent_level) + '<' + K_FCN_TAG
            
            self._write_xml_attrib(indent_level, K_FCN_FUNC_NAME, function.function_name)
            self._write_xml_attrib(indent_level, K_FCN_NAME, function.name)
            self._write_xml_attrib(indent_level, K_FCN_CATEGORY, function.category)
            self._write_xml_attrib(indent_level, K_FCN_DESCRIPTION, function.description)
            self._write_xml_attrib(indent_level, K_FCN_HIDDEN, function.is_hidden)
            
            self.xml += '>\n'
            
            self._write_history(indent_level, function.history)
            self._write_input(indent_level, function.input_args)
            self._write_output(indent_level, function.output_args)
            self._write_help(indent_level, function.help)
            
            self.xml += '\n' + self._indent(indent_level) + '</' + K_FCN_TAG + '>'
        
        return
        
    def _write_history(self, indent_level, history):
        """ Write the history element for the XML metadata.
        """
        
        if history is not None:
            indent_level = indent_level + 1
            
            self.xml += '\n' + self._indent(indent_level) + '<' + K_HISTORY_TAG + '>'
            self.xml += '\n' + self._indent(indent_level + 1) + self._escape_chars(history.text)
            self.xml += '\n' + self._indent(indent_level) + '</' + K_HISTORY_TAG + '>'
            
        return
        
    def _write_help(self, indent_level, help):
        """ Write the help element for the XML metadata.
        """ 
         
        if help is not None:
            indent_level = indent_level + 1
            
            self.xml += '\n' + self._indent(indent_level) + '<' + K_HELP_TAG + '>'
            self.xml += '\n' + self._indent(indent_level + 1) + self._escape_chars(help.text)
            self.xml += '\n' + self._indent(indent_level) + '</' + K_HELP_TAG + '>'   
    
        return
        
    def _write_input(self, indent_level, input_args):
        """ Write the input element for the XML metadata.
        """
        
        if input_args is not None:
            indent_level = indent_level + 1
            
            self.xml += '\n' + self._indent(indent_level) + '<' + K_INPUTS_TAG + '>' + '\n'
            
            for argument in input_args:
                 self._write_argument(indent_level, argument)
                     
#            for bundle in input_bundles:
#                self._write_bundle(indent_level, bundle)
            
            self.xml += '\n' + self._indent(indent_level) + '</' + K_INPUTS_TAG + '>'
            
        return
    
    def _write_output(self, indent_level, output_args):
        """ Write the output element for the XML metadata.
        """
        
        if output_args is not None:
            indent_level = indent_level + 1
            
            self.xml += '\n' + self._indent(indent_level) + '<' + K_OUTPUTS_TAG + '>' + '\n' 
        
            for argument in output_args:
                self._write_argument(indent_level, argument)
        
#        for bundle in output_bundles:
#            self._write_bundle(indent_level, bundle)
            
            self.xml += '\n' + self._indent(indent_level) + '</' + K_OUTPUTS_TAG + '>'   
        
        return
        
#    def _write_bundle(self, indent_level, bundle):
#        """ Write out bundle arguments
#        """
#        
#        if bundle is not None:
#            for argument in bundle.args:
#                self._write_argument(indent_level, argument, bundle.name)
#            
#        return
         
    def _write_argument(self, indent_level, argument):
        """ Write an argument element for the XML metadata.
        """
        
        if argument is not None:
            indent_level = indent_level + 1
            
            self.xml += self._indent(indent_level) + '<' + K_ARG_TAG
            
            self._write_xml_attrib(indent_level, K_ARG_NAME, argument.name)
            self._write_xml_attrib(indent_level, K_ARG_MODIFIER, argument.modifier)
            self._write_xml_attrib(indent_level, K_ARG_BUNDLE, argument.bundle)
            self._write_xml_attrib(indent_level, K_ARG_UNITS, argument.units)
            self._write_xml_attrib(indent_level, K_ARG_DESCRIPTION, argument.description)
            self._write_xml_attrib(indent_level, K_ARG_FAMILY_NAME, argument.family_name)
            self._write_xml_attrib(indent_level, K_ARG_UNITS_SYSTEM_NAME, argument.units_system_name)
            self._write_xml_attrib(indent_level, K_ARG_E_TYPE, argument.external_type)
            self._write_xml_attrib(indent_level, K_ARG_I_TYPE, argument.internal_type)
            self._write_xml_attrib(indent_level, K_ARG_ISINDEX, str(argument.is_index))
            self._write_xml_attrib(indent_level, K_ARG_OPTIONAL, str(argument.is_optional))
            self._write_xml_attrib(indent_level, K_ARG_ARG_NAME, argument.arg_name)
            
            self.xml += '\n' + self._indent(indent_level) + '/>\n'
        
        return 
                
    def _write_xml_attrib(self, indent_level, arg_name, arg_value):
        """ Write an XML level attribute.
        """
        
        indent_level = indent_level + 1
        
        name_value = '\n' + self._indent(indent_level) + arg_name + "='%s'" % self._escape_chars(str(arg_value))
        
        self.xml += name_value
        
        return
        
    def _indent(self, indent_level):
        """ Return a number of tabs equal to the number of indent levels specified.
        """
        
        indent = '    ' * indent_level
        
        self.xml += indent
        
        return indent
        
    def _escape_chars(self, text):
        """ Escape necessary characters.
        """
        
        result = []
        
        for char in text:    
            if char == "&":
                result.append("&amp;")
            elif char == "<":
                result.append("&lt;")
            elif char == ">":
                result.append("&gt;")
            elif char == "'":
                result.append("&apos;")
            elif char == '"':
                result.append("&quot;")  
            else:
                result.append(char)
        
        return ''.join(result)
   
#### EOF ######################################################################
        