from sets import Set

from enthought.logger import logger
from enthought.traits.api import HasTraits, Trait
from enthought.traits.api import Bool, Property

from composite_port import CompositePort
from receiver import Receiver
#from atomic_actor import AtomicActor

class IOPort(CompositePort) :
    """ A port that has direction and can participate in data transfers.
    """
    
    is_input = Bool(False)
    is_optional = Property(Bool)
    
    # transparent ports do not have a receiver
    # Opaque ports have receivers if it is an input port, the links
    # that provide data are from the outside.  For output ports the links
    # that provide data are from the inside (and thus the container is a
    # composite node.
    _local_receiver = Trait(None, Receiver)
    
    def __init__(self, name, is_input=False, *args, **kw) :
        """ Creates a new Port. """
        self._is_optional = False
        CompositePort.__init__(self, name, *args, **kw)

        self.is_input = is_input
    
        return

    #### Trait Properties ####################################################
    
    # Property: is_opaque Bool, ReadOnly
    def _is_opaque(self) :
        """ return False for output ports on AtomicActors; call the super class
        method for all other cases.
        """
        return CompositePort._is_opaque(self)        
    
    is_opaque = Property( _is_opaque )

    # Property: is_optional Bool
    def _get_is_optional(self) :
        return self._is_optional
    
    def _set_is_optional(self, optional):
        self._is_optional = optional
        return

    # Property: full_name isa Str
    
    def _get_full_name( self ) :
        full_name = ""
        if self.container is not None :
            full_name += self.container.full_name + "."
        
        if self.is_input:
            full_name += 'inputs.'
        else:
            full_name += 'outputs.'
            
        full_name += self.name
        return full_name

    full_name = Property( _get_full_name )
        
    #### Methods #############################################################



    def receivers( self ) :
        """ Return a list of the receivers that receive data for this port.
        
        If the port is opaque there is one receiver that is contained
        by the port.
     
     	An input port receives data from ports linked via
     	outside relations. A transparent input port will return the
     	receivers of the ports connected via relations linked on the inside.
        
		Output ports of CompositeActor instances receive data from ports
		linked via inside relations. A transparent output port returns the
		receivers for the ports connected on the outside.

		For an opaque port, this method creates the receiver by calling
		_newReceiver() if there are no receivers.
        """
        receivers = []
        
        #logger.debug( "IOPort.receivers..." )
        if self.is_opaque :
            receivers = [ self._get_local_receiver() ]
        
        else :
            if self.is_input :
                receivers = self.deep_receivers()
            else :
                receivers = self.remote_receivers()
            
        return receivers
        
        
    def deep_receivers( self ) :
        """ If the port is an input port, return the list of receivers linked
        to the inside of this port.
        
        A port can only be inside linked to relations where the relation's
        container is the same as the port's container.  The relation can only
        be linked to other ports with the same container and ports where
        the container of the port's container is the same as this port's
        container.  I.e. this port's container is the parent container of
        ports connected to the same releation.
        
        The deep receivers are the receivers for input ports where the parent
        container is the same is this port's container and output ports with
        the same container as this port (ie. inside connected output ports.)
        """
        receivers = []
        
        if self.is_input :
            container = self.container

            for port in self.inside_connected_ports() :
                parent_container = port.container.container
                if parent_container == container :
                    if port.is_input :
                        receivers += port.receivers()
                elif not port.is_input :
                        receivers += port.receivers()
            
        return receivers

    def remote_receivers( self ) :
        """ If the port is an output port, return the list of receivers linked
        to the outside to this port.

        A port can only be outside connected to a relation where the
        relation's container is the same as the container of the port's
        container (i.e. parent container).  That relation can only be 
        connected to ports with the same parent container as this port
        or ports whose container is the parent container.

        The remote receivers are the receivers for input ports with the same
        parent container as this port and output ports on the parent container.
        """
        receivers = []
        
        if not self.is_input :
            
            container = self.container
            parent_container = container.container
          
            for port in self.connected_ports() :
                if port.container == parent_container : 
                    if not port.is_input :
                        receivers += port.receivers()
                elif port.is_input :
                        receivers += port.receivers()
                    
        #logger.debug("IOPort.remote_receivers returns %s" % receivers )
        return receivers


    def dependent_actors( self ) :
        """ return a list of actors that receive input from this actor """
        return list( self._remote_dependent_actors() )
        
    def _deep_dependent_actors( self ) :
        """ If the port is an input port, return the set of destination
        actors linked to the inside to this port.
        """
        actors = Set()
        
        if self.is_input :
            for port in self.inside_connected_destination_ports() :
                if port.is_opaque :
                    if port.is_input:
                        # outside connected to input of AtomicActor or Opaque CompositeActor
                        actors.add( port.container )
                else :
                    # transparent, thus composite, go recursive
                    if port.is_input :
                        actors.union_update(port._deep_dependent_actors())
                    else :
                        actors.union_update(port._remote_dependent_actors())

        #logger.debug("deep_dependent_actors returns %s" % actors )
        return actors

    def _remote_dependent_actors( self ) :
        """ If the port is an output port, return the set of destination
        actors linked to the outside to this port.
        """
        actors = Set()
        
        if not self.is_input :
            for port in self.connected_destination_ports() :
                if port.is_opaque :
                    if port.is_input:
                        # outside connected to input of AtomicActor or Opaque CompositeActor
                        actors.add( port.container )
                else :
                    # transparent, thus composite, go recursive
                    if port.is_input :
                        actors.union_update( port._deep_dependent_actors() )
                    else :
                        actors.union_update( port._remote_dependent_actors() )

        #logger.debug("remote_dependent_actors returns %s" % actors )
        return actors


    def available_tokens( self ) :
        """ Return the maximum number of tokens available for get to return.
        """
        if not self.is_opaque :
            raise RuntimeError, "Attempt to get on a transparent port"

        tokens = 0

        for receiver in self.receivers() :
            r_tokens = receiver.available_tokens()
            if r_tokens > tokens :
                tokens = r_tokens

        return tokens

    def has_tokens( self, n=1 ) :
        """ Return True if the port has n tokens for get to return. 
        At least one recevier must have n tokens.
        """
        if not self.is_opaque :
            raise RuntimeError, "Attempt to get on a transparent port"

        has_tokens = False

        for receiver in self.receivers() :
            if receiver.has_tokens( n ) :
                has_tokens = True
                break

        return has_tokens

    def get( self, n=1 ) :
        """ Return a token or list of tokens, when n > 1.
        
        An opaque port will get the tokens from its local receiver.
        
        Raises RuntimeError when port is transparent 
        or when no receiver has n tokens
        """
        if not self.is_opaque :
            raise RuntimeError, "Attempt to get on a transparent port"
            
        for receiver in self.receivers() :
            if receiver.has_tokens( n ) :
                token = receiver.get(n)
                #logger.debug("IOPort.get = %s, from receiver on port '%s'" 
                #       % (token, receiver.container.full_name) )
                return token

        raise RuntimeError, \
            "'%s' get called without sufficient data available." % self.full_name


    def has_room( self, n=1 ) :
        """ Return True if the port can accept n inputs via put. 
        All receviers must have room for n inputs.
        """
        if not self.is_opaque :
            raise RuntimeError, "Attempt to get on a transparent port"

        has_room = False
        
        if self.is_input :
            receivers = self.deep_receivers()
        else :
            receivers = self.remote_receivers()
        
        if len(receivers) > 0 :
            has_room = True
        
            for receiver in receivers :
                has_room = has_room and receiver.has_room(n)
        
        return has_room

    def put( self, token ) :
        """ Put a token into the receviers of ports
        connected to this port.  If this port is an input port, the token
        is sent to the deep_receivers. If this port is an output port,
        the token is sent to the remote_receivers.
        
        Raises RuntimeError when port is transparent 
        or when no receiver has n tokens
        """
        if not self.is_opaque :
            raise RuntimeError, "Attempt to get on a transparent port"

        if self.is_input :
            receivers = self.deep_receivers()
        else :
            receivers = self.remote_receivers()
    
        for receiver in receivers :
            #logger.debug("IOPort put, %s, into receiver on port %s" 
            #    % (token, receiver.container.full_name) )
            receiver.put( token )

        return

     
       
    def clear( self ) :
        """ Clear this port's receivers of any contained tokens. """
        #logger.debug( "IOPort.clear %s" % self.full_name )
        for receiver in self.receivers() :
            receiver.clear()

        return
    
    def clear_inside( self ) :
        """ Clear this port's receivers of any contained tokens. """
        for receiver in self.deep_receivers() :
            receiver.clear()

        return

    def wrapup( self ):
        """ At the end of the exectuion cycle wrapup is called to free execution
        resources.
        """
        self._local_receiver = None
        return
        
    def connected_destination_ports(self) :
        """ return a list of outside linked ports where my output goes
        """
        ports = []

        if not self.is_input :
            for relation in self.relations :
                ports.extend( relation.linked_destination_ports( exclude=self ))

        return ports
        
    def inside_connected_destination_ports(self) :
        """ return a list of inside linked ports where my output goes
        """
        ports = []

        if self.is_input :
            for relation in self.inside_relations :
                ports.extend( relation.linked_destination_ports( exclude=self ))

        return ports
        
    ###########################################################################
    # Protected methods
    ###########################################################################

    def _get_local_receiver( self ) :
        """ return the local receiver, create it if necessary """
        if not self.is_opaque :
            raise RuntimeError, "Attempt to create receiver on transparent port"
        
        if self._local_receiver is None :
            receiver = self._create_local_receiver()
            receiver.container = self
            self._local_receiver = receiver
        
        return self._local_receiver
            
    def _create_local_receiver( self ) :
        """ Create and return a local receiver """
        container = self.container
        if container is None :
            raise RuntimeError, "Cannot create receiver without a container."
        
        #logger.debug( "IOPort._create_local_receiver %s ..." % self.full_name )
        receiver = None
        
        if self.is_input :
            # create receiver compatible with the executive director
            receiver = container.new_receiver( self )

        else : 
            # create receiver compatible with the local director
            receiver = container.new_inside_receiver( self )

#        logger.debug("IOPort: container._create_local_receiver"
#            + " container: '%s' port '%s' returns: %s"
#            % (container.name, self.name, receiver) )
            
        return receiver

        
#### EOF ######################################################################    
