from enthought.logger import logger
from enthought.traits.api import HasTraits, Trait
from enthought.traits.api import Property, Str, List, Bool

from port import Port

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class Node( HasTraits ) :
    """ A vertex in a program graph. An aggregation of ports.
		The ports can be linked to relations. The relations thus represent
		connections between ports, and hence, connections between nodes.
		To add a port to a node, set its container to the node.  
		To remove it, set its container to None, or to some other node.
    """
    
    # All non-composite nodes are atomic
    # that is, no further decomposition is possible
    atomic = Bool(True)
    

    # fixme: when traits handles co-dependent classes with inheritance
    ports = List                          # should really be List(Ports)
    
    def __init__(self, name, *args, **kw) :
        """ Creates a new Node.
        """
        # Trait Property defaults
        self._container = None
        self._name = 'unknown'

        HasTraits.__init__(self, *args, **kw)

        self.name = name
        return
        
    #### Trait Properties ####################################################
    
    def _get_name( self ):
        """ return name trait """
        return self._name
    
    def _set_name( self, name ):
        """ set the name trait, must be unique within the container.
        Raises ValueError if name is invalid or a duplicate. """
        if name == self._name:
            return
        
        if self.container is not None:
            self.container._check_node_name( name )
            
        old = self._name
        self._name = name
        
        self.trait_property_changed('name', old, self._name)
        
        return
        
    name = Property( _get_name, _set_name )
    
    # Property: full_name isa Str
    
    def _get_full_name( self ) :
        full_name = ""
        if self.container is not None :
            full_name += self.container.full_name + "."
        
        full_name += self.name
        return full_name

    full_name = Property( _get_full_name )
    
    
    # Property: is_opaque isa bool

    # The internal topology of opaque nodes is hidden from the outside
    # Atomic nodes are by definition opaque.
    def _get_is_opaque( self ) :
        return True
        
    is_opaque = Property( _get_is_opaque )



    # Property: container isa CompositeNode
    
    def _set_container(self, container) :
        """ Specify the container for this node.
        	Node is removed from existing container if contained.
        	If container is None, the node's ports are unlinked.
        """
        # no-op case
        if self._container is not None :
            if self._container == container :
                return

        # link into the new container
        if container is not None :
            container._add_node( self )
        
        previous_container = self._container
        self._container = container
        
        # unlink from previous container
        if previous_container is not None :
            previous_container._remove_node( self )
        
        # unlink the ports if Node no longer contained
        if container is None :
            for port in self.ports :
                port.unlink_all()
    
        return
                
    def _get_container(self) :
        """ Return the container of this Port
        """
        return self._container
    
    container = Property( _get_container, _set_container )
    
    #### Methods ##############################################################

    def new_port(self, name) :
        """ Create a port compatible with this node.
        """
        self._check_port_name( name )
        port = Port( name )
        port.container = self
        return port

    def get_port(self, name ) :
        """ Return the port contained by this entity that has the specified name.
            If there is no such port, return None.
            name - The name of the desired port.
        """
        for port in self.ports :
            if port.name == name :
                return port
        
        return None

    def port_list(self) :
        """ Return a copy of the list of Ports belonging to this Node.
        """
        return self.ports[:]

    def connected_ports(self) :
        """ Return a list of the ports that are connected to contained ports.
            Ports on this node are not included unless there is a loopback,
            meaning that two distinct ports of this node are linked to the 
            same relation. The connected nodes can be obtained from the ports
            using container.  Note that a port may be listed more than
            once if there is more than one connection to it.
            The connected ports will all have the same container as this node.
        """
        ports = []
        for port in self.ports :
            ports.extend( port.connected_ports() )
        return ports

    def linked_relations(self) :
        """ Get all relations that are linked to ports contained by this entity.
       		Note that a relation may be listed more once.
        """
        relations = []
        for port in self.ports :
            relations.extend( port.relations )
       
        return relations


       
    ###########################################################################
    # Protected methods
    ###########################################################################
    def _add_port(self, port) :
        """ Used by Port to add itself to this Node
        	except: Node already contains a different Port with the same name
        """
        self._check_port_name( port.name )
        self.ports.append( port )

        return
        
    def _remove_port( self, port ) :
        """ Used by Port to remove itself to this Node
            except: Node already contains a different Port with the same name
        """
        if port in self.ports :
            self.ports.remove(port)

        return
        
    def _check_port_name(self, name) :
        """ except: ValueError if there is a conflict with an existing name
        """
        if name in [ p.name for p in self.ports ] : 
            raise ValueError, 'Conflict existing Port name: %s' % name
            
        return


#### EOF ######################################################################
