from enthought.logger import logger
from enthought.traits.api import HasTraits
from enthought.traits.api import Trait, Str, List, Any, Property

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class Port( HasTraits ) :
    """ A Port is the interface of a Node to any number of Relations.
        Normally, a Port is contained by an Node, although a port
        may exist with no container.  The role of a port is to aggregate
        a set of links to relations.  Thus, for example, to represent
        a directed graph, nodes can be created with two ports, one for
        incoming arcs and one for outgoing arcs.  More generally, the arcs
        to an node may be divided into any number of subsets, with one port
        representing each subset.
        
        Use the newPort factory method on the Node object.
    """
    
    name = Str('unknown')
    relations = List        #fixme should really be List(Relation)
    
    def __init__(self, name, *args, **kw) :
        """ Creates a new Port. """
        #defaults for Property Traits
        self._container = None

        HasTraits.__init__(self, *args, **kw)
        self.name = name
        
        return
        
    #### Trait Properties #####################################################

    # Property: full_name isa Str
    
    def _get_full_name( self ) :
        full_name = ""
        if self.container is not None :
            full_name += self.container.full_name + "."
        
        full_name += self.name
        return full_name

    full_name = Property( _get_full_name )
    
    # Property: container isa Node
    
    def _set_container(self, container) :
        """ Specify the container for this port. Container must be a Node. """
        if self._container is not None :
            if self._container == container :
                return
        
            else :
                self._container._remove_port(self)
                self._container = None
        
        if container is not None :
            container._add_port( self )
            self._container = container
        
        return
                
    def _get_container(self) :
        """ Return the container of this Port """
        return self._container
    
    container = Property( _get_container, _set_container )
    
    #### Methods ##############################################################
    
    def link(self, relation) :
        """ Link this port with a Relation """
        self._check_link(relation)
        self.relations.append( relation )
        relation._link(self)

    def is_linked(self, relation) :
        """ Return true if the given Relation is linked to this port.
        """
        return relation in self.relations

    def unlink(self, relation) :
        """ Unlink the Relation from this port. """
        if relation in self.relations :
            relation._unlink(self)
            self.relations.remove(relation)
        
    def unlink_all(self) :
        """ Break links to all relations """
        for relation in self.relations[:] :
            relation._unlink(self)

    def get_relation( self, name ) :
        """ Return a connected relation with the given name or None.
        """
        for relation in self.relations :
            if relation.name == name :
                return relation
        
        return None

    def relations_list(self) :
        """ Return a copy of the list of contained Relations.
        """
        return self.relations[:]

    def connected_ports(self) :
        """ List the ports this port is connected to via linked relations.
        	Note that a port may be listed more than once if more 
        	than one connection to it has been established.
        	Outside links only.
     	"""
        ports = []
        for relation in self.relations :
            ports.extend( relation.linked_ports( exclude=self ))
        
        return ports
        

    ###########################################################################
    # Protected methods
    ###########################################################################
    
    def _check_link(self, relation) :
        """ Raise ValueError if port and relation are incompatible.
        	Rules: port must have a container; Relation._check_port rules.
        	Subclasses should override to apply more specific rules.
        """
        if self.container is None :
            raise ValueError, "Port must be contained before linked"
        if relation is not None :
            relation._check_port( self )
        

       
#### EOF ######################################################################
