from enthought.logger import logger
from enthought.traits.api import HasTraits, Trait, Property, Any

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class Receiver( HasTraits ) :
    """ Interface for objects that can hold tokens.
    
	An implementation of this interface has two key methods: put() and get().
	The put() method deposits a token into the receiver. 
	The get() method retrieves a token that has been put. The order of
    the retrieved tokens depends on specific implementations, and does not
    necessarily match the order in which tokens have been put.
   
    All implementations of this interface must follow these rules, regardless
    of the number of threads that are accessing the receiver:
   
    - If hasToken() returns true, then the next call to get() must not
      result in a NoTokenException being thrown.
    - If hasRoom() returns true, then the next call to put() must not
      result in a NoRoomException being thrown.
   
    In general, this means that multithreaded domains must provide
    synchronization for receivers. Note that both NoTokenException
    and NoRoomException are runtime exceptions, so they need not be
    declared explicitly.
   
    Objects that implement this interface can only be contained
    by an instance of IOPort.
    """
    
    def __init__(self, container=None, *args, **kw) :
        """ Create a new IOPort. """
        #Property defaults
        self._container = None
        
        HasTraits.__init__(self, *args, **kw)
        self.container = container

    #### Trait Properties ####################################################
    
    # Property: container isa IOPort
    def _get_container( self ) :
        """ Return the container, IOPort, of this receiver """
        return self._container

    def _set_container( self, container ) :
        """ Set the container, an IOPort, of this receiver """
        self._container = container
    
    container = Property( _get_container, _set_container )
    
    #### Methods ##############################################################
    
    def clear( self ) :
        """ Clear this receiver of any contained tokens. """
        raise NotImplementedError

    def available_tokens( self ) :
        """ Return the number of tokens available for get to return."""
        raise NotImplementedError

    def has_tokens( self, n=1 ) :
        """ Return True if the receiver has n tokens for get to return. """
        raise NotImplementedError
    
    def get( self, n=1 ) :
        """ Return a token or list of tokens, when n > 1, from this receiver.
        Raises ValueError if n tokens are not available.
        """
        raise NotImplementedError

        
    def has_room( self, n=1 ) :
        """ Return True if the receiver can accept n inputs. """
        raise NotImplementedError

    def put( self, token ) :
        """ Put a token or list of tokens into this recevier.
        
        Token is either a single data object or a list of tokens.
		To put a list object as a single token, wrap it with another list.
        
        Raises ValueError if room for n tokens is not available.
        """
        raise NotImplementedError
        

    
#### EOF ######################################################################    
