from enthought.logger import logger
from enthought.traits.api import HasTraits
from enthought.traits.api import Property, Str, List, Any

#from port import Port

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

class Relation( HasTraits ) :
    """ A Relation links ports, and therefore the nodes that contain them.
        To link a port to a relation, use the link() method in the Port class.
        To remove a link, use the unlink() method in the Port class.
    """
    
    name = Str('unknown')
    ports = List                 # should be List(Port)
    
    def __init__(self, name, *args, **kw) :
        """ Create a new Relation. """
        # defaults for Trait Properties
        self._container = None
        
        HasTraits.__init__(self, *args, **kw)
        self.name = name
        
        return

    #### Trait Properties #####################################################
    
    # Property: container isa CompositeNode
    
    def _set_container(self, container) :
        """ Specify the container for this relation.
            Relation is removed from existing container if contained.
            If container is None, the relations's ports are unlinked.
        """
        if container is not None :
            # no-op case
            if self._container == container :
                return

# enforce proper hook-up has circular references between classes....
#
#            if not isinstance(container, CompositeNode ) :
#                raise (TypeError,
#                    "type mismatch between Relation (%s) and container (%s)" %
#                    (self.__class__.__name__, container.__class__.__name__) )
            
            # tell the new container about me
            container._add_relation( self )

        # link to the new container
        previous_container = self._container
        self._container = container
        
        # unlink from previous container
        if previous_container is not None :
            previous_container._remove_relation( self )
        
        # unlink the ports if Node no longer contained
        if container is None :
            for port in self.ports[:] :
                port.unlink(self)
    
        return
                
    def _get_container(self) :
        """ Return the container of this Relation """
        return self._container

    container = Property( _get_container, _set_container )
    
    
    # Property: full_name isa Str
    
    def _get_full_name( self ) :
        full_name = ""
        if self.container is not None :
            full_name += self.container.full_name + "."
        
        full_name += self.name
        return full_name

    full_name = Property( _get_full_name )

    ### Methods ###############################################################
    
    
    def linked_ports( self, exclude=None ) :
        """ return the list of ports linked to this relation 
        	exclude the specified port.
        """
        if exclude is not None :
            return [ p for p in self.ports if p != exclude ]
        else :
            return self.ports

    def unlink_all() :
        """ Break links to all ports """
        for port in self.ports[:] :
            port.unlink(self)
        
        return
        
    ###########################################################################
    # Protected methods
    ###########################################################################
   
    def _check_port( self, port ) :
        """ Raise ValueError if port and relation are incompatible.
        	Rules: Relation must have a container.
        	Subclasses should override to enforce more detailed rules.
        """
        if self.container is None :
            raise ValueError, "Relation must be contained before linked"
        
        return
        
    def _link(self, port) :
        """ Link this port with a Relation. """
        self.ports.append( port )
        return

    def _unlink(self, port) :
        """ Unlink the specified port from this relation """
        if port in self.ports :
            self.ports.remove(port)
        return
        
# Self-referential/circular traits.
# fixme: when traits handles co-dependent classes with inheritance
#from port import Port
#Relation.add_class_trait( 'ports', List( Port ))      

#from composite_node import CompositeNode
#Relation.add_class_trait( 'container', CompositeNode )
  
#### EOF ######################################################################
