
import traceback

# Enthought library

from enthought.logger import logger
from enthought.traits.api import Class

# inspired by Ptolemy http://ptolemy.eecs.berkeley.edu/ptolemyII

from director import Director
from fifo_receiver import FifoReceiver
from scheduler import Scheduler
from schedule import Schedule
from topology_scheduler import TopologyScheduler
from exceptions import FireError


def DefaultSchedulerFactory(director):
    return TopologyScheduler( container=director.container )


class SdfDirector( Director ) :
    """
    A Director that implements Synchronous DataFlow semantics.
    """
    
    def __init__(self, *args, **kw) :
        self._scheduler = None
        Director.__init__( self, *args, **kw )
        
        if kw.has_key('receiver_class') :
            self.receiver_class = kw['receiver_class']

        else:
            self.receiver_class = FifoReceiver
            
        # A factory function that gets called to create new instances of our
        # scheduler.  The only argument for this function is a reference to self.
        # TODO: This should be removed when we stop blowing away the existing
        #       scheduler in preinitialize().
        self.scheduler_factory = DefaultSchedulerFactory
        
        return
        
    def new_receiver( self, port ) :
        """ Return a new receiver that is compatible with the director.
        """
        receiver = self.receiver_class()
        #logger.debug("SdfDirector.new_receiver returns: %s" % receiver )
        return receiver
    
    def preinitialize( self ) :
        """ Preinitialize the contained actors and compute the schedule.
        """
        Director.preinitialize( self )
        
        # TODO: Not listening for changes in topology, so force a new scheduler
        self._scheduler = None
        self._get_scheduler()
        
        return
        
    def initialize(self) :
        """ Initialize the actors associated with this director.
        The order in which the actors are initialized is arbitrary.
        In addition, if actors connected directly to output ports
        have initial production, then copy that initial production
        to the outside of the composite actor.        
        """
        Director.initialize(self)
        
        # TODO: if output ports should have initial tokens produced do so now.
        
        return

    def prefire( self ) :
        """ Check the input ports of the container composite actor (if there
		are any) to see whether they have enough tokens, and return true
		if they do.  If there are no input ports, then also return true.
		Otherwise, return false.  Note that this does not call prefire()
		on the contained actors.
        """
        Director.prefire( self )
        
        schedule = self._get_scheduler().schedule
        rate = schedule.iterations
        
        for port in self.container.input_port_list() :
            if not port.is_optional and not port.has_tokens( rate ) : 
                return False
        
        return True
        
        
    def fire( self ) :
        """ Calculate the current schedule, if necessary, and iterate the
		contained actors in the order given by the schedule. 
		
		Iterating an actor involves calling the actor's iterate() method,
		which is equivalent to calling the actor's prefire(), fire() and
		postfire() methods in succession.  
        """
        schedule = self._get_scheduler().schedule
        self._run_schedule( schedule )

        return


        
    def postfire( self ) :
        """ 
        """
        return Director.postfire( self )
        
    #### Protected Methods ####################################################
    
    def _get_scheduler( self ) :
        """ return the Scheduler for node execution.
        Subclasses are expected to override this method to change the default
        class of the Scheduler.
        """
        if self._scheduler is None :
            self._scheduler = self.scheduler_factory(self)
        
        return self._scheduler


    def _set_scheduler( self, newscheduler ) :
        """ sets the Scheduler for node execution.
        This can be used to change the type of scheduling at runtime, without
        subclassing _get_scheduler.
        """
        self._scheduler = newscheduler
        return
        

    def _run_schedule( self, schedule ) :
        """ Iterate through the schedule items.  If an item is an Actor, call
        item.iterate passing schedule.iterations as iteration count argument.
        If item is a schedule, call this method recusively.
        """
        try:
            for item in schedule.items :
                if type(item) is Schedule :
                    for i in xrange( schedule.iterations ) :
                        self._run_schedule( item )
                else :
                    item.iterate( schedule.iterations )
        
        except Exception, ex:
            raise FireError( item, ex )
        
        return
        
        
#### EOF ######################################################################    
