"""Defines utility functions for extracting a subgraph from a graph of actors."""


from sets import Set


class ExtractionState:
    """
    Encapsulates references to objects that are "global" to the recursive
    function compute_is_output().
    """
    def __init__(self, output_actors):
        self.output_actors = output_actors
        self.ancestor_map = {}
        self.is_output_cache = {}


def compute_is_output(actor, state):
    """Returns whether or a not actor is an "output" chip
        
        actor: instance of graph.Actor
        state: instance of ExtractionState
    
    Recursively checks to see if any of its downstream actors are output
    chips and if the input_actor is in output_actors.  Returns True or False.
    """
    
    # short-circuit by checking the memoizing dict
    if state.is_output_cache.has_key(actor):
        return state.is_output_cache[actor]
    
    is_output = False   # whether input_actor is in the output node
    
    # recursively check each downstream ("child") actor to see if any of them
    # are outputs.  If they are, then this actor is an output as well.
    dependents = actor.dependents()
    for child_actor in dependents:
        if compute_is_output(child_actor, state):
            # if we have a child that is an output, then we are an output
            is_output = True
            
            # create the ancestor map entry for this child actor
            if state.ancestor_map.has_key(child_actor):
                state.ancestor_map[child_actor].add(actor)
            else:
                state.ancestor_map[child_actor] = Set([actor])
    
    # check the actor itself
    if (actor in state.output_actors) or (state.output_actors is None):
        is_output = True
    
    # add the actor to the memo dict
    state.is_output_cache[actor] = is_output
    return is_output


def create_ancestor_map(input_actors, output_actors=None):
    """create_ancestor_map(input_actors, output_actors=None) -> ancestor_map
        
        input_actors: list of Actors
        output_actors: list of Actors
    
    Returns the "ancestor map", a dict that maps each actor to its list of
    ancestor actors.  The only items on the ancestor map will be actors
    lying in the graph between input_actors and output_actors.
    
    If output_actors is None, then all leaf nodes are considered output
    nodes.
    """

    state = ExtractionState(output_actors)

    for actor in input_actors:
        # input actors are root nodes, and thus have empty ancestor sets
        state.ancestor_map[actor] = Set()
        compute_is_output(actor, state)

    return state.ancestor_map


# EOF
