from enthought.logger import logger

from enthought.graph import AtomicActor
from enthought.graph import CompositeActor

        
class HelloWorldActor( AtomicActor ) :
    """ Actor that prints 'Hello World!!!!!!' to the logger 
    No input or output ports.
    """
    def __init__(self, *args, **kw) :
        AtomicActor.__init__(self, *args, **kw)
        
    def fire(self) :
        logger.debug( "Hello World!!!!!!" )
        return
        
class OutputActor( AtomicActor ) :
    """ Actor with one output port, 'out'.
    if value_list is not None, the first item is removed from the list and
    assigned to value before value is output.
    When fired, self.value is output.
    """
    def __init__(self, *args, **kw) :
        AtomicActor.__init__(self, *args, **kw)
        if kw.has_key('value') :
            self.value = kw['value']
        
        self.value_list = None
        self._create_ports()
        return
        
    def fire(self) :
        p = self.get_port("out")
        if self.value_list is not None and len(self.value_list) > 0:
            self.value = self.value_list[0]
            self.value_list = self.value_list[1:]
            
        #logger.debug( "%s.fire: put(%s)" % (self.name, self.value) )
        p.put(self.value)
        return

    def _create_ports(self) :
        self.new_port("out", False )
        return

class InputActor( AtomicActor ) :
    """ Actor with one input port, 'in'.
    When fired, the first token in the input port queue is read
    and saved in self.value.
    """
    def __init__(self, *args, **kw) :
        AtomicActor.__init__(self, *args, **kw)
        self._create_ports()
        self.value = None
        return
        
    def fire(self) :
        p = self.get_port("in")
        b = p.has_tokens(1)
        self.value = p.get()
        #logger.debug( "%s.fire: get() = %s" % (self.name, self.value) )
        return

    def _create_ports(self) :
        self.new_port("in", True )
        return

    def _available_tokens(self) :
        p = self.get_port("in")
        n = p.available_tokens()
        return n
        
        
class TransparentCompositeActor( CompositeActor ) :
    """ 
    A CompositeActor with one input port, 'cin', and one output port, 'cout'.
    Because the node is transparent, it is not fired.  Data flows through
    cin to ports attached to cin, in this case 'cout'. Data flows through
    cout to ports attached to cout.  Thus, data flows through this transparent
    composite without stopping on the way.
    """
    def __init__(self, *args, **kw) :
        CompositeActor.__init__(self, *args, **kw)
        self.opaque=False
        self._create_ports()
        return
        
    def fire(self) :
        
        # some test cases will add a Director to this actor 
        # making it not transparent
        if not self.is_opaque :
            raise RuntimeError, \
                "Transparent composite actor fire has been called."
        
        logger.debug( "******* TransparentCompositeActor '%s' fire" % self.name)
        return CompositeActor.fire(self)

    def _create_ports(self) :
        cin = self.new_port("cin", True )
        cout = self.new_port("cout", False )
        self.connect( cin, cout )
        return

    def _available_tokens(self, port_name) :
        p = self.get_port(port_name)
        n = p.available_tokens()
        return n

        
class NoopActor( AtomicActor ) :
    """ Actor with one input port, 'in', and one output port, 'out'.
    When fired, one token is copied from in to out and saved in self.value.
    """
    def __init__(self, *args, **kw) :
        AtomicActor.__init__(self, *args, **kw)
        self._create_ports()
        return
        
    def fire(self) :
        pin = self.get_port("in")
        self.value = pin.get()
        #logger.debug( "%s fire inputs get %s" % (self.name, self.value) )
        pout = self.get_port("out")
        pout.put( self.value )
        return

    def _create_ports(self) :
        self.new_port("in", True )
        self.new_port("out", False )
        return

    def _available_tokens(self) :
        p = self.get_port("in")
        n = p.available_tokens()
        return n


class FanOutActor( AtomicActor ) :
    """ Actor with one input, 'in', and two outputs 'out1' and 'out2.
    When fired, one input token is copied to both outputs.
    """
    def __init__(self, *args, **kw) :
        AtomicActor.__init__(self, *args, **kw)
        self._create_ports()
        return

    def fire(self) :
        pin = self.get_port("in")
        self.value = pin.get()
        logger.debug( "%s fire inputs get %s" % (self.name, self.value) )
        for port in self.output_port_list() :
            port.put( self.value )
        return

    def _create_ports(self) :
        self.new_port("in", True )
        self.new_port("out1", False )
        self.new_port("out2", False )
        return

    def _available_tokens(self) :
        p = self.get_port("in")
        n = p.available_tokens()
        return n
    
class FanInActor( AtomicActor ) :
    """ Actor with two inputs, 'in1' and 'in2', and one output 'out'.
	When fired a token is read from both inputs, in order, and put to 'out'.
    """
    def __init__(self, *args, **kw) :
        AtomicActor.__init__(self, *args, **kw)
        self._create_ports()
        return

    def fire(self) :
        self.value = [ port.get() for port in self.input_port_list() ]
        
        logger.debug( "%s fire inputs get %s" % (self.name, self.value) )
        
        for port in self.output_port_list() :
            port.put( self.value )
        
        return

    def _create_ports(self) :
        self.new_port("in1", True )
        self.new_port("in2", True )
        self.new_port("out", False )
        return

    def _available_tokens(self) :
        """ return a list of counts of avaliable tokens on input ports """
        l = []
        
        for port in self.input_port_list() :
            l.append( port.available_tokens() )

        return l

class ListSinkActor( InputActor ) :
    """ Actor with one input, 'in'.
    All inputs are appended to a list available as self.values
    """
    def __init__( self, *args, **kw ) :
        InputActor.__init__(self, *args, **kw)
        self.values = []
        return
    
    def fire( self ) :
        InputActor.fire(self)
        self.values.append(self.value)
        return
        
        
    
### EOF ######################################################################
