import unittest

from enthought.logger import logger

from enthought.graph import Node
from enthought.graph import CompositeNode
from enthought.graph import Port
from enthought.graph import CompositePort
from enthought.graph import Relation
from enthought.graph.fifo_receiver import FifoReceiver
from enthought.graph import IORelation
from enthought.graph import IOPort
from enthought.graph import AtomicActor
from enthought.graph import CompositeActor
from enthought.graph import SdfDirector

        

class ComplexTestCase( unittest.TestCase ) :
    #before every test
    def setUp(self) :
        #  --F----------------------------------------------------
        #  |                                                      |         
        #  |            -----A-----------------------             |
        #  |            |                           |             |
        #  | |-D--|     |     |-B--|     |-C--|     |     |-E--|  |
        #  | |    m--v--n--w--o    p--x--q    r--y--s--z--t    |  |
        #  | |----|     |    /|----|     |----|     |     |----|  |
        #  |           nn-ww                        |             |
        #  |             ---------------------------              |
        #  |                                                      |
        #  -------------------------------------------------------
        #
        #
        self.f = CompositeNode('F')
        
        self.a = CompositeNode('A')
        self.a.container = self.f
        
        self.b = Node('B')
        self.b.container = self.a
        
        self.c = Node('C')
        self.c.container = self.a
        
        self.d = Node('D')
        self.d.container = self.f
        
        self.e = Node('E')
        self.e.container = self.f
        
        
        self.m = self.d.new_port('m')
        self.m.container = self.d
        
        self.n = self.a.new_port('n')
        self.n.container = self.a
        
        self.nn = self.a.new_port('nn')
        self.nn.container = self.a
        
        self.o = self.b.new_port('o')
        self.o.container = self.b

        self.p = self.b.new_port('p')
        self.p.container = self.b
        
        self.q = self.c.new_port('q')
        self.q.container = self.c
        
        self.r = self.c.new_port('r')
        self.r.container = self.c
        
        self.s = self.a.new_port('s')
        self.s.container = self.a
        
        self.t = self.e.new_port('t')
        self.t.container = self.e


        self.v = self.f.new_relation('v')
        self.v.container = self.f
        self.m.link(self.v)
        self.n.link(self.v)

        self.w = self.a.new_relation('w')
        self.w.container = self.a
        self.n.link(self.w)
        self.o.link(self.w)

        self.ww = self.a.new_relation('ww')
        self.ww.container = self.a
        self.nn.link(self.ww)
        self.o.link(self.ww)

        self.x = self.a.new_relation('x')
        self.x.container = self.a
        self.p.link(self.x)
        self.q.link(self.x)
        
        self.y = self.a.new_relation('y')
        self.y.container = self.a
        self.r.link(self.y)
        self.s.link(self.y)
        
        self.z = self.f.new_relation('z')
        self.z.container = self.f
        self.s.link(self.z)
        self.t.link(self.z)
        return

    def test_node_linked_relations(self) :
        node_relations = self.a.linked_relations()
        self.assertEqual( [self.v, self.z], node_relations, 
            "a.linked_relations returned incorrect list: %s" % node_relations)

        node_relations = self.b.linked_relations()
        self.assertEqual( [self.w, self.ww, self.x], node_relations, 
            "b.linked_relations returned incorrect list: %s" % node_relations)
        
        node_relations = self.c.linked_relations()
        self.assertEqual( [self.x, self.y], node_relations, 
            "c.linked_relations returned incorrect list: %s" % node_relations)
        return
        
    def test_port_relations_list(self) :
        relations_list = self.o.relations_list()
        self.assertEqual( [self.w, self.ww], relations_list, 
            "o.relations_list returned incorrect list: %s" % relations_list)
        return
        
    def test_port_get_relation_by_name(self) :
        relation = self.o.get_relation('w')
        self.assertEqual( self.w, relation, 
            "self.o.get_relation('w') incorrect" )
        relation = self.o.get_relation('ww')
        self.assertEqual( self.ww, relation, 
            "self.o.get_relation('ww') incorrect" )
        return
        
    def test_relation_linked_ports(self) :
        linked_ports = self.x.linked_ports()
        self.assertEqual( [self.p, self.q], linked_ports, 
            "x.linked_ports returned incorrect list: %s" % linked_ports)

        linked_ports = self.y.linked_ports()
        self.assertEqual( [self.r, self.s], linked_ports, 
            "y.linked_ports returned incorrect list: %s" % linked_ports)
        return

    def test_port_connected_ports(self) :
        connected_ports = self.p.connected_ports()
        self.assertEqual( [self.q], connected_ports, 
            "p.connected_ports returned incorrect list: %s" % connected_ports)

        connected_ports = self.q.connected_ports()
        self.assertEqual( [self.p], connected_ports, 
            "q.connected_ports returned incorrect list: %s" % connected_ports)

        connected_ports = self.r.connected_ports()
        self.assertEqual( [self.s], connected_ports, 
            "r.connected_ports returned incorrect list: %s" % connected_ports)

        connected_ports = self.o.connected_ports()
        self.assertEqual( [self.n, self.nn], connected_ports, 
            "o.connected_ports returned incorrect list: %s" % connected_ports)
        return

    def test_port_inside_outside(self) :
        connected_ports = self.n.connected_ports()
        self.assertEqual( [self.m], connected_ports, 
            "n.connected_ports returned incorrect list: %s" % connected_ports)
            
        connected_ports = self.n.inside_connected_ports()
        self.assertEqual( [self.o], connected_ports, 
            "n.inside_connected_ports returned incorrect list: %s" 
            % connected_ports)
            
        connected_ports = self.s.connected_ports()
        self.assertEqual( [self.t], connected_ports, 
            "s.connected_ports returned incorrect list: %s" % connected_ports)
        
        connected_ports = self.s.inside_connected_ports()
        self.assertEqual( [self.r], connected_ports, 
            "s.inside_connected_ports returned incorrect list: %s" 
            % connected_ports)
        return
        
    def test_unlink_relations(self) :
        self.p.unlink(self.x)
        node_relations = self.b.linked_relations()
        self.assertEqual( [self.w, self.ww], node_relations, 
            "node.connected_relations returned incorrect list after unlink" )
        return

    def test_node_list(self) :
        node_list = self.f.node_list()
        self.assertEqual( [self.a, self.d, self.e], node_list, 
            "CompositeNode.node_list returned incorrect" )
        
        node_list = self.a.node_list()
        self.assertEqual( [self.b, self.c], node_list, 
            "CompositeNode.node_list returned incorrect" )
        return
        
    def test_node_by_name(self) :
        node = self.f.get_node('A')
        self.assertEqual( self.a, node, 
            "CompositeNode.get_node by name returned incorrect" )
        return
        
    def test_port_by_name(self) :
        node = self.f.get_port('g')
        self.assertEqual( None, node, 
            "Node.get_port by name returned incorrect" )
        
        node = self.a.get_port('n')
        self.assertEqual( self.n, node, 
            "Node.get_port by name returned incorrect" )
        
        node = self.e.get_port('t')
        self.assertEqual( self.t, node, 
            "Node.get_port by name returned incorrect" )

    def test_composite_node_relation_by_name(self) :
        relation = self.a.get_relation('w')
        self.assertEqual( self.w, relation, 
            "compositeNode.get_relation by name returned incorrect" )
        
        relation = self.a.get_relation('ww')
        self.assertEqual( self.ww, relation, 
            "compositeNode.get_relation by name returned incorrect" )
        
        relation = self.a.get_relation('none')
        self.assertEqual( None, relation, 
            "compositeNode.get_relation by name returned incorrect" )
        return
        
    def test_opacity(self) :
        self.assertEqual( False, self.a.is_opaque, "self.a.is_opaque is True" )
        self.assertEqual( False, self.f.is_opaque, "self.f.is_opaque is True" )
        self.assertEqual( True, self.b.is_opaque, "self.b.is_opaque is False" )
        self.assertEqual( True, self.c.is_opaque, "self.c.is_opaque is False" )
        self.assertEqual( True, self.d.is_opaque, "self.d.is_opaque is False" )
        self.assertEqual( True, self.e.is_opaque, "self.e.is_opaque is False" )
        return
        
    def test_atomic(self) :
        self.assertEqual( False, self.a.atomic, "self.a.atomic is True" )
        self.assertEqual( False, self.f.atomic, "self.f.atomic is True" )
        self.assertEqual( True, self.b.atomic, "self.b.atomic is False" )
        self.assertEqual( True, self.c.atomic, "self.c.atomic is False" )
        self.assertEqual( True, self.d.atomic, "self.d.atomic is False" )
        self.assertEqual( True, self.e.atomic, "self.e.atomic is False" )
        return
        
    def test_composite_node_relations_list(self) :
        relations_list = self.f.relations_list()
        self.assertEqual( [self.v, self.z], relations_list, 
            "self.f.relations_list incorrect" )
        
        relations_list = self.a.relations_list()
        self.assertEqual( [self.w, self.ww, self.x, self.y], relations_list, 
            "self.a.relations_list incorrect" )
        return
        
    
### EOF ######################################################################
