"""
This unittest shows how to code the dataflow graph structure
of a converge problem using the graph classes.
"""
import unittest

from enthought.graph import Node
from enthought.graph import CompositeNode
from enthought.graph import Port
from enthought.graph import CompositePort
from enthought.graph import Relation

class BaseTestCase( unittest.TestCase ):
    #before every test
    def setUp(self) :
        pass
    
    #after every test
    def tearDown(self) :
        pass
        
# TSD_DEMO_3_4_1_GROUP_IN_GROUP Converge function group mfile
#
#%% <INPUTS>
#% <watdth_>...Water depth  <units=F>
#% <oftemp_>...Temperature at ocean floor  <units=degF>
#% <tgrad_>....Temperature gradient  <units=degF/100F>
#% <pgrad_>....Pressure gradient (normal = 0.465 psi/ft)  <units=PSI/FT>
#% <tvdss_>....   depth index <units=F>
#% <salw_>.....Brine salinity  <units=PPM>
#% <salw_sh_>..Brine salinity  <units=PPM>
#% <vp_ss_>....P-Velocity of sand end-member  <units=KM/S>
#% <vp_sh_>....P-Velocity of shale end-member  <units=KM/S>
#% <kg_ss_>....Grain bulk modulus of sand end-member  <units=GPa>
#% <rhog_ss_>..Grain density of sand  <units=G/CC>
#% <rhob_ss_>..Bulk density of sand end-member  <units=G/CC>
#% <kg_sh_>....Grain bulk modulus of shale end-member  <units=GPa>
#% <rhog_sh_>..Grain density of shale end-member  <units=G/CC>
#% <rhob_sh_>..Bulk density of shale end-member  <units=G/CC>
#% <gg_ss_>....Grain shear modulus of sand end-member  <units=GPa>
#% <phit_>.....Total porosity  <units=V/V>
#% <xsh_>......Total volume fraction of shale  <units=V/V>
#%% </INPUTS>
#%
#%% <OUTPUTS>
#% <vs_ss_>....Shear Velocity of sand end-member  <units=KM/S>
#% <vs_sh_>....S-Velocity of shale end-member  <units=KM/S>
#% <biot_ss_>..biot coefficient  <units=dV/dV>
#% <biotmax_ss>.upper bound biot coefficient  <units=dV/dV>
#% <biot_sh_>..biot coefficient  <units=dV/dV>
#% <biotmax_sh>.upper bound biot coefficient  <units=dV/dV>
#% <vp_wet_>...P-velocity  <units=KM/S>
#% <vs_wet_>...S-velocity  <units=KM/S>
#% <rhob_wet_>.Bulk Density  <units=G/CC>
#% <vp_sm_>....P-Velocity of the Sandy member  <units=KM/S>
#% <vs_sm_>....S-Velocity of the Sandy member  <units=KM/S>
#% <phi_sm_>...Porosity fraction of the sandy member  <units=V/V>
#% <rhob_sm_>..Density of the sandy member  <units=G/CC>
#% <phie_sm_>..Effective porosity fraction of the sandy member  <units=V/V>
#% <rhog_sm_>..Effective grain density of sandy member  <units=G/CC>
#% <kg_sm_>....Grain bulk modulus of sandy member  <units=GPa>
#% <vp_hs_>....P-Velocity of hard shale  <units=KM/S>
#% <vs_hs_>....S-Velocity of hard shale  <units=KM/S>
#% <phi_hs_>...Porosity fraction of the Hard Shale  <units=V/V>
#% <xlam_>.....Laminated Shale fraction  <units=V/V>
#% <xdis_>.....Dispersed Shale fraction  <units=V/V>
#% <phi_ss_>...Porosity of Sand end member  <units=V/V>
#% <phi_sh_>...Porosity of Shale end member  <units=V/V>
#% <k_wat_>....Brine bulk modulus  <units=GPa>
#% <k_watsh_>..Brine bulk modulus  <units=GPa>
#% <ppore_>....Pore fluid pressure at each depth sample  <units=MPa>
#% <rho_wat_>..Brine density  <units=G/CC>
#% <rho_watsh_>.Brine density  <units=G/CC>
#% <tform_>....Temperature at each depth sample  <units=degC>
#%% </OUTPUTS>
#%
#%

# This is the group function for...
#[vs_ss, vs_sh, biot_ss, biotmax_ss, biot_sh, biotmax_sh, vp_wet, vs_wet,
# rhob_wet, vp_sm, vs_sm, phi_sm, rhob_sm, phie_sm, rhog_sm, kg_sm, vp_hs, vs_hs,
# phi_hs, xlam, xdis, phi_ss, phi_sh, k_wat, k_watsh, ppore, 
# rho_wat, rho_watsh, tform]
#  = tsd_demo_3_4_1_group_in_group(watdth, oftemp, tgrad, pgrad, depth, 
#                                      salw, salw_sh, vp_ss, vp_sh, kg_ss, 
#                                      rhog_ss, rhob_ss, kg_sh, rhog_sh, rhob_sh, 
#                                      gg_ss, phit, xsh
#                                  );
#
#content of the group....
#[tform, ppore] = temp_press_linear(watdth, oftemp, tgrad, pgrad, depth);;
#
#[rho_wat, k_wat] = brine_properties(ppore, tform, salw);;
#
#[rho_watsh, k_watsh] = brine_properties(ppore, tform, salw_sh);;
#
#[vs_ss, vs_sh, biot_ss, biotmax_ss, biot_sh, biotmax_sh, vp_wet, vs_wet, 
# rhob_wet, vp_sm, vs_sm, phi_sm, rhob_sm, phie_sm, rhog_sm, kg_sm, vp_hs, 
# vs_hs, phi_hs, xlam, xdis, phi_ss, phi_sh] 
# = wet_trends_and_rhob(vp_ss, vp_sh, k_wat, rho_wat, kg_ss, rhog_ss, rhob_ss, 
#            k_watsh, rho_watsh, kg_sh, rhog_sh, rhob_sh, gg_ss, phit, xsh);;
#]]>
#         
class ConvergeExample1TestCase( BaseTestCase ) :

    def create_tsd_demo_group(self) :
        """ Create a TSD Demo Group node with all necessary ports """
        group = CompositeNode('tsd_demo_3_4_1_group_in_group')

        input_names = [
            'watdth', 'oftemp', 'tgrad', 'pgrad', 'depth', 
            'salw', 'salw_sh', 'vp_ss', 'vp_sh', 'kg_ss', 
            'rhog_ss', 'rhob_ss', 'kg_sh', 'rhog_sh', 'rhob_sh', 
            'gg_ss', 'phit', 'xsh'        
            ]
        inputs = [group.new_port(name) for name in input_names]
        
        output_names = [
            'vs_ss', 'vs_sh', 'biot_ss', 'biotmax_ss', 'biot_sh', 'biotmax_sh',
            'vp_wet', 'vs_wet', 'rhob_wet', 'vp_sm', 'vs_sm', 'phi_sm',
            'rhob_sm', 'phie_sm', 'rhog_sm', 'kg_sm', 'vp_hs', 'vs_hs',
            'phi_hs', 'xlam', 'xdis', 'phi_ss', 'phi_sh', 'k_wat', 'k_watsh',
            'ppore', 'rho_wat', 'rho_watsh', 'tform'
            ]
        outputs = [group.new_port(name) for name in output_names]

        return group, outputs, inputs
   
 
    def create_sandy_brine_properties(self) :
        """ Create a sandy_brine_properties node with all necessary ports """
        sandy_brine_properties = Node('sand_brine_properties')
        
        inputs = [
            sandy_brine_properties.new_port('ppore'),
            sandy_brine_properties.new_port('tform'),
            sandy_brine_properties.new_port('salw')
            ]
        
        outputs = [
            sandy_brine_properties.new_port('rho_wat'),
            sandy_brine_properties.new_port('k_wat')
            ]
        
        return sandy_brine_properties, outputs, inputs


    def create_shale_brine_properties(self) :
        """ Create a shale_brine_properties node with all necessary ports """
        shale_brine_properties = Node('shale_brine_properties')

        inputs = [
            shale_brine_properties.new_port('ppore'),
            shale_brine_properties.new_port('tform'),
            shale_brine_properties.new_port('salw_sh')
            ]

        outputs = [
            shale_brine_properties.new_port('rho_watsh'),
            shale_brine_properties.new_port('k_watsh')
            ]

        return shale_brine_properties, outputs, inputs


    def create_wet_trends_and_rhob( self ):
        """ Create a wet_trends_and_rhob node with all necessary ports """
        wet_trends_and_rhob = Node('wet_trends_and_rhob')

        input_names = [
            'vp_ss', 'vp_sh', 'k_wat', 'rho_wat', 'kg_ss', 'rhog_ss', 'rhob_ss',
            'k_watsh', 'rho_watsh', 'kg_sh', 'rhog_sh', 'rhob_sh', 'gg_ss', 
            'phit', 'xsh'
            ]
        inputs = [wet_trends_and_rhob.new_port(name) 
                            for name in input_names]
        
        output_names = [
            'vs_ss', 'vs_sh', 'biot_ss', 'biotmax_ss', 'biot_sh', 'biotmax_sh',
            'vp_wet', 'vs_wet', 'rhob_wet', 'vp_sm', 'vs_sm', 'phi_sm', 
            'rhob_sm', 'phie_sm', 'rhog_sm', 'kg_sm', 'vp_hs', 'vs_hs', 
            'phi_hs', 'xlam', 'xdis', 'phi_ss', 'phi_sh'
            ]
        outputs = [wet_trends_and_rhob.new_port(name) 
                            for name in output_names]
        
        return wet_trends_and_rhob, outputs, inputs

    def create_temp_press_linear(self) :
        """ Create a temp_press_linear node with all necessary ports """
        temp_press_linear = Node('temp_press_linear')

        input_portNames = ['watdth', 'oftemp', 'tgrad', 'pgrad', 'depth']
        inputs = [temp_press_linear.new_port(name) 
                            for name in input_portNames]
        
        outputs = [
            temp_press_linear.new_port('tform'),
            temp_press_linear.new_port('ppore')
            ]
        
        return temp_press_linear, outputs, inputs

        
    def test_tsd_demo(self ) :
        # create the Group and the Nodes to be grouped
        group, group_output_ports, group_input_ports \
            = self.create_tsd_demo_group()

        temp_press_linear, temp_press_linear_outputs, temp_press_linear_inputs \
             = self.create_temp_press_linear()
             
        temp_press_linear.container = group

        sandy_brine_properties, sandy_brine_outputs, sandy_brine_inputs \
            = self.create_sandy_brine_properties()
        
        sandy_brine_properties.container = group
        
        shale_brine_properties, shale_brine_outputs, shale_brine_inputs \
            = self.create_shale_brine_properties()
        
        shale_brine_properties.container = group
        
        (wet_trends_and_rhob, wet_trends_and_rhob_outputs,
            wet_trends_and_rhob_inputs ) = self.create_wet_trends_and_rhob()

        wet_trends_and_rhob.container = group

        # hook 'em up
        #inputs
        group.connect( group.get_port('watdth'), 
                        temp_press_linear.get_port('watdth') )
        group.connect( group.get_port('oftemp'), 
                        temp_press_linear.get_port('oftemp') )
        group.connect( group.get_port('tgrad'), 
                        temp_press_linear.get_port('tgrad') )
        group.connect( group.get_port('pgrad'), 
                        temp_press_linear.get_port('pgrad') )

        # run it
        
        #outputs
        ppore_relation = group.connect( temp_press_linear.get_port('ppore'), 
                                        group.get_port('ppore'))
        tform_relation = group.connect( temp_press_linear.get_port('tform'),
                                        group.get_port('tform'))

        
        #inputs

        # relations already exists, so link to them, don't create a new ones
        sandy_brine_properties.get_port('ppore').link(ppore_relation)
        sandy_brine_properties.get_port('tform').link(tform_relation)
        group.connect( group.get_port('salw'), 
                        sandy_brine_properties.get_port('salw') )

        #run it
        
        #output
        rho_wat_relation = group.connect(
                                    sandy_brine_properties.get_port('rho_wat'),
                                    group.get_port('rho_wat'))
        k_wat_relation = group.connect(
                                    sandy_brine_properties.get_port('k_wat'), 
                                    group.get_port('k_wat'))
        
        #inputs
        shale_brine_properties.get_port('ppore').link(ppore_relation)
        shale_brine_properties.get_port('tform').link(tform_relation)
        group.connect( group.get_port('salw_sh'), 
                        shale_brine_properties.get_port('salw_sh') )

        #run it
        
        #output
        rho_watsh_relation = group.connect( 
                                shale_brine_properties.get_port('rho_watsh'),
                                group.get_port('rho_watsh'))
        k_watsh_relation = group.connect( 
                                shale_brine_properties.get_port('k_watsh'), 
                                group.get_port('k_watsh'))
        
        #inputs
        for name in [ 'vp_ss', 'vp_sh', 'kg_ss', 'rhog_ss', 'rhob_ss', 'kg_sh',
                     'rhog_sh', 'rhob_sh', 'gg_ss', 'phit', 'xsh'
                    ] :
            group.connect( group.get_port(name), 
                            wet_trends_and_rhob.get_port(name) )
        
        wet_trends_and_rhob.get_port('k_wat').link(k_wat_relation)
        wet_trends_and_rhob.get_port('rho_wat').link(rho_wat_relation)
        wet_trends_and_rhob.get_port('k_watsh').link(k_watsh_relation)
        wet_trends_and_rhob.get_port('rho_watsh').link(rho_watsh_relation)

        #run it
        
        #output
        for name in [ 'vs_ss', 'vs_sh', 'biot_ss', 'biotmax_ss', 'biot_sh',
                    'biotmax_sh', 'vp_wet', 'vs_wet', 'rhob_wet', 'vp_sm',
                    'vs_sm', 'phi_sm', 'rhob_sm', 'phie_sm', 'rhog_sm',
                    'kg_sm', 'vp_hs', 'vs_hs', 'phi_hs', 'xlam', 'xdis',
                    'phi_ss', 'phi_sh'
                    ] :
            group.connect( wet_trends_and_rhob.get_port(name),
                            group.get_port(name) ) 

#  Test suites:
#-------------------------------------------------------------------------------

def create_test_suite ( level = 1 ):
    suites = []
    if level > 0:
        suites.extend( [ unittest.makeSuite( x, 'test_' ) 
                            for x in [
                                    ConvergeExample1TestCase,                        
                            ] ] )
    testSuite = unittest.TestSuite( suites )
    return testSuite
    
def test ( level = 10 ):
    suite = create_test_suite( level )
    runner = unittest.TextTestRunner()
    runner.run( suite )
    return runner

#-------------------------------------------------------------------------------
#  Run tests:
#-------------------------------------------------------------------------------

if __name__ == "__main__":
    test()
    
    
### EOF ######################################################################
