import unittest

from enthought.logger import logger

from enthought.graph import SdfDirector

from actors import *

class DependentActorsTestCase( unittest.TestCase ) :

    def setUp(self) :
        self.top = CompositeActor('composite')

        self.output_actor = OutputActor('output_actor', value='a')
        self.output_actor.container = self.top

        self.fanout_actor = FanOutActor('fanout_actor')
        self.fanout_actor.container = self.top

        self.noop1_actor = NoopActor('noop1_actor')
        self.noop1_actor.container = self.top

        self.input1_actor = InputActor('input1_actor')
        self.input1_actor.container = self.top

        self.input2_actor = ListSinkActor('input2_actor')
        self.input2_actor.container = self.top

        self.fanin_actor = FanInActor('fanin_actor')
        self.fanin_actor.container = self.top

        self.top.connect( self.output_actor.get_port('out'), 
                        self.fanout_actor.get_port('in'))

        self.top.connect( self.fanout_actor.get_port('out1'), 
                        self.noop1_actor.get_port('in'), 'noop1.in')

        self.top.connect( self.fanout_actor.get_port('out2'), 
                        self.input2_actor.get_port('in'), 'in2')

        nr1 = self.top.connect( self.noop1_actor.get_port('out'),
                        self.input1_actor.get_port('in'), 'nr1')
        
        self.fanin_actor.get_port('in1').link(nr1)
                       
        self.ca = TransparentCompositeActor('transparent_composite')
        self.ca.container = self.top
        self.ca.get_port('cin').link(nr1)
        
        self.top.connect( self.ca.get_port('cout'),  
                       self.fanin_actor.get_port('in2'), 'fin2')
        
        
        self.director = SdfDirector()
        self.director.container = self.top

        logger.debug("test execution begins...")
        
        self.top.preinitialize()
        self.top.initialize()
        return

    def test_single_dependent( self ) :
        self.assertEqual( [self.fanout_actor], self.output_actor.dependents(),
            "Incorrect dependent list for output_actor, got: %s"
            % self.output_actor.dependents() )
        return
        
    def test_fanout_dependent( self ) :
        self.assertEqual( [self.noop1_actor, self.input2_actor], 
            self.fanout_actor.dependents(),
            "Incorrect dependent list for fanout_actor, got: %s"
            % self.fanout_actor.dependents() )
        return

    def test_no_dependent( self ) :
        self.assertEqual( [], self.input2_actor.dependents(),
            "Incorrect dependent list for input2_actor, got: %s"
            % self.input2_actor.dependents() )
            
        self.assertEqual( [], self.input1_actor.dependents(),
            "Incorrect dependent list for input1_actor, got: %s"
            % self.input1_actor.dependents() )

        self.assertEqual( [], self.fanin_actor.dependents(),
            "Incorrect dependent list for fanin_actor, got: %s"
            % self.fanin_actor.dependents() )
        return

    def test_relation_fanout_dependent( self ) :
        self.assertEqual( [self.input1_actor, self.fanin_actor ],
            self.noop1_actor.dependents(),
            "Incorrect dependent list for noop1_actor, got: %s"
            % self.noop1_actor.dependents() )
        return

    def test_node_in_transparent_composite( self ) :
        self.ca_noop_actor = NoopActor('noop_actor')
        self.ca_noop_actor.container = self.ca
        
        #kill the existing link from ca.cin to ca.cout
        relation = self.ca.get_relation("cout")
        relation.container = None
        
        
        self.ca.connect( self.ca.get_port( "cin"), 
                        self.ca_noop_actor.get_port("in" ))
        self.ca.connect( self.ca_noop_actor.get_port("out" ),
                        self.ca.get_port( "cout"))
        
        self.assertEqual( 
            [self.input1_actor, self.fanin_actor, self.ca_noop_actor ],
            self.noop1_actor.dependents(),
            "Incorrect dependent list for noop1_actor, got: %s"
            % self.noop1_actor.dependents() )
        
        self.assertEqual( 
            [self.fanin_actor],
            self.ca_noop_actor.dependents(),
            "Incorrect dependent list for ca_noop_actor, got: %s"
            % self.ca_noop_actor.dependents() )

        return
        

    def test_node_in_opaque_composite( self ) :
        self.ca_noop_actor = NoopActor('noop_actor')
        self.ca_noop_actor.container = self.ca
        
        #kill the existing link from ca.cin to ca.cout
        relation = self.ca.get_relation("cout")
        relation.container = None
        
        
        self.ca.connect( self.ca.get_port( "cin"), 
                        self.ca_noop_actor.get_port("in" ))
        self.ca.connect( self.ca_noop_actor.get_port("out" ),
                        self.ca.get_port( "cout"))
        
        director = SdfDirector()
        director.container = self.ca

        self.assertEqual( 
            [self.input1_actor, self.fanin_actor, self.ca ],
            self.noop1_actor.dependents(),
            "Incorrect dependent list for noop1_actor, got: %s"
            % self.noop1_actor.dependents() )
        
        self.assertEqual( 
            [],
            self.ca_noop_actor.dependents(),
            "Incorrect dependent list for ca_noop_actor, got: %s"
            % self.ca_noop_actor.dependents() )

        self.assertEqual( 
            [self.fanin_actor],
            self.ca.dependents(),
            "Incorrect dependent list for ca, got: %s"
            % self.ca.dependents() )

        return



### EOF ######################################################################
