import unittest

from enthought.logger import logger

from enthought.graph import SdfDirector

from actors import *

        
class FanInActorTestCase( unittest.TestCase ) :

    def setUp(self) :
        logger.debug("-------------------------------------------------------")
        self.top = CompositeActor('composite')

        self.output1_actor = OutputActor('output1_actor')
        self.output1_actor.container = self.top

        self.output2_actor = OutputActor('output2_actor')
        self.output2_actor.container = self.top

        self.fanin_actor = FanInActor('fanin_actor')
        self.fanin_actor.container = self.top

        self.input_actor = InputActor('input_actor')
        self.input_actor.container = self.top
        
        self.top.connect( self.output1_actor.get_port('out'), 
                        self.fanin_actor.get_port('in1'))

        self.top.connect( self.output2_actor.get_port('out'), 
                        self.fanin_actor.get_port('in2'))

        self.top.connect( self.fanin_actor.get_port('out'), 
                        self.input_actor.get_port('in'), 'serialized')
                        
        self.director = SdfDirector()
        self.director.container = self.top

        logger.debug("test execution begins...")
        
        self.top.preinitialize()
        self.top.initialize()
        return

    def tearDown( self ) :
        logger.debug("test execution ends.")
        self.top.wrapup()
        return
        
    def test_simple_token( self ) :
        self.output1_actor.value = 1
        self.output2_actor.value = 2
        self.top.iterate(1)
        self.assertEqual( [1, 2], self.input_actor.value, \
            "input_actor has unexpected value: %s" % self.input_actor.value )    
        return

    def test_list_token( self ) :
        self.output1_actor.value = [1,3,5]
        self.output2_actor.value = [2,4,6]
        self.top.iterate(1)
        self.assertEqual( [[1,3,5], [2,4,6]], self.input_actor.value, \
            "input_actor has unexpected value: %s" % self.input_actor.value )    
        self.assertEqual( 0, self.input_actor._available_tokens(), \
            "input_actor has incorrect number of available tokens")
        self.assertEqual( [0,0], self.fanin_actor._available_tokens(), \
            "input_actor has incorrect number of available tokens")
        return

    def test_list_of_tokens( self ) :
        tokens = range(3)
        self.output1_actor.value_list = range( 1, 7, 2 )
        self.output2_actor.value_list = range( 0, 6, 2 )
        for value in [[1,0], [3,2], [5,4]] :
            self.top.iterate(1)
            self.assertEqual( value, self.input_actor.value,
                "input_actor has unexpected value: %s expected: %s"
                 % (self.input_actor.value, value) )    
        return

        
class FanInRelationTestCase( FanInActorTestCase ) :

    def setUp(self) :
        logger.debug("-------------------------------------------------------")
        self.top = CompositeActor('composite')

        self.output1_actor = OutputActor('output1_actor')
        self.output1_actor.container = self.top

        self.output2_actor = OutputActor('output2_actor')
        self.output2_actor.container = self.top

        self.input_actor = InputActor('input_actor')
        self.input_actor.container = self.top
        
        fanin_relation = self.top.connect(  self.output1_actor.get_port('out'), 
                                            self.input_actor.get_port('in'),
                                            'fanin_relation' )

        self.output2_actor.get_port('out').link(fanin_relation)
        
        self.director = SdfDirector()
        self.director.container = self.top

        logger.debug("test execution begins...")
        
        self.top.preinitialize()
        self.top.initialize()
        return

    def tearDown( self ) :
        logger.debug("test execution ends.")
        self.top.wrapup()
        return
        
    def test_simple_token( self ) :
        self.output1_actor.value = 1
        self.output2_actor.value = 2
        self.top.iterate(1)
        self.assertEqual( 1, self.input_actor.value, \
            "input_actor has unexpected value: %s" % self.input_actor.value )    
        self.assertEqual( 1, self.input_actor._available_tokens(), \
            "input_actor has incorrect number of available tokens")

        # scheduling does not account for different firing rates
        # relation fan-in creates the need to fire receivers multiple times
        self.input_actor.iterate(1)
        self.assertEqual( 2, self.input_actor.value, \
            "input_actor has unexpected value: %s" % self.input_actor.value )    
        self.assertEqual( 0, self.input_actor._available_tokens(), \
            "input_actor has incorrect number of available tokens")
        return

    def test_list_token( self ) :
        self.output1_actor.value = [1,3,5]
        self.output2_actor.value = [2,4,6]
        self.top.iterate(1)
        self.assertEqual( [1,3,5], self.input_actor.value, \
            "input_actor has unexpected value: %s" % self.input_actor.value )    
        self.assertEqual( 1, self.input_actor._available_tokens(), \
            "input_actor has incorrect number of available tokens")

        self.input_actor.iterate(1)
        self.assertEqual( [2,4,6], self.input_actor.value, \
            "input_actor has unexpected value: %s" % self.input_actor.value )    
        self.assertEqual( 0, self.input_actor._available_tokens(), \
            "input_actor has incorrect number of available tokens")
        return

    def test_list_of_tokens( self ) :
        tokens = range(3)
        self.output1_actor.value_list = range( 1, 7, 2 )
        self.output2_actor.value_list = range( 0, 6, 2 )
        for first, second in [(1,0), (3,2), (5,4)] :
            self.top.iterate(1)
            self.assertEqual( first, self.input_actor.value,
                "input_actor has unexpected first value: %s expected: %s"
                 % (self.input_actor.value, first) )    
            self.input_actor.iterate(1)
            self.assertEqual( second, self.input_actor.value,
                "input_actor has unexpected second value: %s expected: %s"
                 % (self.input_actor.value, second) )    
        return

        
    
### EOF ######################################################################
