""" Unittest TestCase for performance testing ValuePort against IOPort 

use ``-t perf_`` with the test runner tool.

"""
import unittest
import time

from enthought.logger import logger

from enthought.traits.api import Any

from enthought.graph import CompositeActor, SdfDirector, \
    SimpleSdfDirector, ValueActor
    
from enthought.graph import InsufficientInputError

from actors import OutputActor, InputActor, NoopActor
        
class ValueSinkActor( ValueActor ):
    """ An actor with an input but no output.  Most recent input in 'value'. """
    
    value = Any
    
    def __init__(self, *args, **kw) :
        super( ValueSinkActor, self).__init__(*args, **kw)
        self._create_ports()
        self.value = None
        return
        
    def fire(self) :
        p = self.get_port("in")
        self.value = p.get()
        return

    def _create_ports(self) :
        self.new_port("in", True )
        return


class ValueSourceActor( ValueActor ):
    """ An actor with attribute defined output using value ports.
    The actor will put the value in the 'value' trait.
    """
    
    value = Any

    def __init__(self, *args, **kw) :
        super( ValueSourceActor, self).__init__(*args, **kw)
        self._create_ports()
        self.value = "ValueSourceActor default"
        return
        
    def fire(self) :
        p = self.get_port("out")
        p.put( self.value )
        return

    def _create_ports(self) :
        self.new_port("out", False )
        return
        
class ValueNoopActor( ValueActor ):
    """ passes input to output unchanged """
    def __init__(self, *args, **kw) :
        super( ValueNoopActor, self).__init__(*args, **kw)
        self._create_ports()
        self.value = "ValueNoopActor default"
        
        self.function = [ self.my_function ]
        return
        
    def xfire(self) :
        pin = self.get_port("in")
        pout = self.get_port("out")
        
        value = pin.get()
        pout.put( value )
        return

    def fire ( self ):
        """ basically the same as ConvergeActor.fire so that tests reflect what
        users should experience.
        """
        inputs = self._inputs[:]
        ports  = self._input_ports
        for i, port in enumerate( ports ):
            inputs[i] = port.get()

        results = self.function[0]( *inputs )
        
        for i, port in enumerate( self._output_ports ):
            port.put( results[i] )
    
    def _create_ports(self) :
        pin = self.new_port("in", True )
        pout = self.new_port("out", False )
        
        # something like what ConvergeActor does....
        self._input_ports = [ pin ]
        self._output_ports = [ pout ]
        self._inputs = range(1)
        return
    
    def my_function( self, x ):
        return (x,)
    
class PerfValueReceiversTestCase( unittest.TestCase ) :

    def setUp(self) :
        self.num_noops = 100
        return
        
    
    def _add_noop_actor( self, klass, name, top, source_port ):
        """ create a ValueNoopActor, tie it's input to source's output """
        noop = klass( name )
        noop.container = top
        top.connect( source_port, noop.get_port('in'), name )
        return noop
            
    def perf_value_source_to_sink_chain( self ) :
        """ test a long source to sink chain
        [ source, noop*, sink ]
        """
        top = CompositeActor('top')
        source_actor = ValueSourceActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = top
        
        sink_actor = ValueSinkActor("sink_actor")
        sink_actor.get_port('in').token = "default value"
        sink_actor.container = top

        previous_actor = source_actor
        for i in xrange( self.num_noops ):
            name = 'noop%2d' % i
            noop_actor = self._add_noop_actor( ValueNoopActor, name, top, previous_actor.get_port('out') )
            previous_actor = noop_actor
         
        top.connect( noop_actor.get_port("out"), sink_actor.get_port("in") )
        
        d = SimpleSdfDirector()
        d.container = top

        self.assertEqual( None, sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    

        self._run( top, 'Value nodes')
        
        self.assertEqual( "source actor value", sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    

        return

    def perf_classic_source_to_sink_chain( self ) :
        """ test a long source to sink chain
        [ source, noop*, sink ]
        """
        top = CompositeActor('top')
        source_actor = OutputActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = top
        
        sink_actor = InputActor("sink_actor")
        sink_actor.get_port('in').token = "default value"
        sink_actor.container = top

        previous_actor = source_actor
        for i in xrange( self.num_noops ):
            name = 'noop%2d' % i
            noop_actor = self._add_noop_actor( NoopActor, name, top, previous_actor.get_port('out') )
            previous_actor = noop_actor
         
        top.connect( noop_actor.get_port("out"), sink_actor.get_port("in") )
        
        d = SdfDirector()
        d.container = top

        self.assertEqual( None, sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    

        self._run( top, 'Classic nodes' )
        
        self.assertEqual( "source actor value", sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    

        return


    def _run( self, top, title ):
        for i in [1, 16, 64, 256, 1024 ]:
            top.preinitialize()
            top.initialize()
        
            self._time_run( top, i, title )
        return
        
    def _time_run( self, top, num_iterations, title ):
        
        start_time = time.clock()
        
        top.iterate( num_iterations )
        
        end_time = time.clock()
        
        num_actors = len(top.nodes)
        num_firings = num_actors * num_iterations
        num_transfers = (num_actors + 1) * num_iterations
        delta_time = (end_time - start_time)*1000
        time_per_firing = delta_time / num_firings
        time_per_transfer = delta_time / num_transfers
        
        print '\n', title
        print 'Actors = %d, iterations = %d, firings = %d, transfers = %d' \
            % (num_actors, num_iterations, num_firings, num_transfers)
        print 'Total time = %.2f s, time/firing = %.2f ms, time/transfer = %.2f ms' \
            % (delta_time/1000, time_per_firing, time_per_transfer)
        
        top.wrapup()
        return
        


### EOF ######################################################################
