import unittest

from enthought.logger import logger

from enthought.graph import IORelation
from enthought.graph import IOPort
from enthought.graph.fifo_receiver import FifoReceiver
from enthought.graph.fifo_history_receiver import FifoHistoryReceiver
from enthought.graph.fifo_auto_history_receiver import FifoAutoHistoryReceiver

class ReceiverTestCase( unittest.TestCase ) :
        
    def test_receiver_new(self) :
        receiver = FifoReceiver()
        self.assertEqual( False, receiver.has_tokens(), 
            "new receiver has tokens")
        self.assertEqual( False, receiver.has_tokens(2), 
            "new receiver has tokens")
        
        self.assertEqual( True, receiver.has_room(), 
            "new receiver has no room")
        self.assertEqual( True, receiver.has_room(2), 
            "new receiver has no room")
        return
        
    def test_receiver_put_get(self) :
        receiver = FifoReceiver()

        self.failUnlessRaises(ValueError, receiver.get)

        self.failUnlessRaises(ValueError, receiver.get, 2)

        receiver.put('first')
        
        self.assertEqual( True, receiver.has_tokens(),
            "receiver has no tokens after put")
        self.assertEqual( True, receiver.has_tokens(1),
            "receiver does not have 1 token after single put")
        self.assertEqual( False, receiver.has_tokens(2), 
            "receiver has 2 or more tokens after single put")

        token = receiver.get()
        self.assertEqual( 'first', token, 
            "get(put('token')) returned %s" % token)
        
        self.assertEqual( False, receiver.has_tokens(), 
            "after get(put()) receiver has tokens")
            
        self.failUnlessRaises(ValueError, receiver.get)

        receiver.put('first')
        receiver.put(['second', 'third'])
        
        self.assertEqual( True, receiver.has_tokens(2), 
            "receiver does not have 2 tokens as it should")

        token = receiver.get()
        self.assertEqual( 'first', token, "expected 'first' got %s" % token)

        self.assertEqual( True, receiver.has_tokens(1), 
            "receiver does not have 1 token as it should")

        token = receiver.get()
        self.assertEqual( ['second', 'third'], token, "expected 'second' got %s" % token)
        
        receiver.put('first')
        receiver.put(['second', 'third'])
        self.assertEqual( True, receiver.has_tokens(2), 
            "receiver does not have 2 tokens as it should")
        receiver.clear()
        self.assertEqual( False, receiver.has_tokens(), "not empty after clear") 

        return

    def test_iorelation(self) :
        r = IORelation('ior')
        ports = r.linked_destination_ports()
        
        self.assertEqual( ports, [], "destination port list not empty" )
        
        iop = IOPort('ioport', True )
        self.assertEqual( iop.is_input, True, "IOPort is_input not correct.")
        return
        
    def test_fifo_history( self ):
        fifo = FifoHistoryReceiver()
        fifo.put( 1 )

        self.assertEqual( True, fifo.has_tokens(1) )
        self.assertEqual( False, fifo.has_history_tokens(1) )
        
        data = fifo.get()
        self.assertEqual( 1, data )

        self.assertEqual( False, fifo.has_tokens(1) )
        self.assertEqual( True, fifo.has_history_tokens(1) )
        
        hdata = fifo.get_history()
        self.assertEqual( data, hdata )
        
        return
        
    def test_fifo_auto_history( self ):
        fifo = FifoAutoHistoryReceiver()
        fifo.put( 1 )

        self.assertEqual( True, fifo.has_tokens(1) )
        self.assertEqual( True, fifo.has_unused_tokens(1) )
        self.assertEqual( False, fifo.has_history_tokens(1) )
        
        self.assertEqual( 1, fifo.available_tokens() )
        self.assertEqual( 1, fifo.available_unused_tokens() )
        self.assertEqual( 0, fifo.available_history_tokens() )

        data = fifo.get()
        self.assertEqual( 1, data )

        self.assertEqual( True, fifo.has_tokens(1) )
        self.assertEqual( False, fifo.has_unused_tokens(1) )
        self.assertEqual( True, fifo.has_history_tokens(1) )
        
        self.assertEqual( 1, fifo.available_tokens() )
        self.assertEqual( 0, fifo.available_unused_tokens() )
        self.assertEqual( 1, fifo.available_history_tokens() )
        
        hdata = fifo.get_history()
        self.assertEqual( data, hdata )

        hdata = fifo.get()
        self.assertEqual( data, hdata )
        
        return
        
    def test_fifo_auto_history_2( self ):
        fifo = FifoAutoHistoryReceiver( depth=2 )
        fifo.put( 1 )
        fifo.put( 2 )

        self.assertEqual( True, fifo.has_tokens(2) )
        self.assertEqual( True, fifo.has_unused_tokens(2) )
        self.assertEqual( False, fifo.has_history_tokens(2) )
        
        self.assertEqual( 2, fifo.available_tokens() )
        self.assertEqual( 2, fifo.available_unused_tokens() )
        self.assertEqual( 0, fifo.available_history_tokens() )

        data = fifo.get()
        self.assertEqual( 1, data )

        self.assertEqual( True, fifo.has_tokens(1) ) # from unused
        self.assertEqual( False, fifo.has_tokens(2) ) # only 1 in unused
        self.assertEqual( True, fifo.has_unused_tokens(1) )
        self.assertEqual( False, fifo.has_unused_tokens(2) )
        self.assertEqual( True, fifo.has_history_tokens(1) )
        self.assertEqual( False, fifo.has_history_tokens(2) )
        
        self.assertEqual( 1, fifo.available_tokens() )
        self.assertEqual( 1, fifo.available_unused_tokens() )
        self.assertEqual( 1, fifo.available_history_tokens() )
        
        hdata = fifo.get_history()
        self.assertEqual( data, hdata )

        data2 = fifo.get()
        self.assertEqual( 2, data2 )
        
        self.assertEqual( True, fifo.has_tokens(1) ) # from history
        self.assertEqual( True, fifo.has_tokens(2) ) # from history
        self.assertEqual( False, fifo.has_unused_tokens(1) )
        self.assertEqual( False, fifo.has_unused_tokens(2) )
        self.assertEqual( True, fifo.has_history_tokens(1) )
        self.assertEqual( True, fifo.has_history_tokens(2) )
        
        self.assertEqual( 2, fifo.available_tokens() )
        self.assertEqual( 0, fifo.available_unused_tokens() )
        self.assertEqual( 2, fifo.available_history_tokens() )

        hdata2 = fifo.get( 2 )
        self.assertEqual( [1,2], hdata2 )
        
        fifo.put( 3 )

        self.assertEqual( True, fifo.has_tokens(1) ) # from actove
        self.assertEqual( False, fifo.has_tokens(2) ) # only 1 in unused
        self.assertEqual( True, fifo.has_unused_tokens(1) )
        self.assertEqual( False, fifo.has_unused_tokens(2) )
        self.assertEqual( True, fifo.has_history_tokens(1) )
        self.assertEqual( True, fifo.has_history_tokens(2) )

        self.assertEqual( 1, fifo.available_tokens() )
        self.assertEqual( 1, fifo.available_unused_tokens() )
        self.assertEqual( 2, fifo.available_history_tokens() )

        data3 = fifo.get()
        self.assertEqual( 3, data3 )

        self.assertEqual( True, fifo.has_tokens(1) ) # from history
        self.assertEqual( True, fifo.has_tokens(2) ) # from history
        self.assertEqual( False, fifo.has_unused_tokens(1) )
        self.assertEqual( False, fifo.has_unused_tokens(2) )
        self.assertEqual( True, fifo.has_history_tokens(1) )
        self.assertEqual( True, fifo.has_history_tokens(2) )
        
        self.assertEqual( 2, fifo.available_tokens() )
        self.assertEqual( 0, fifo.available_unused_tokens() )
        self.assertEqual( 2, fifo.available_history_tokens() )

        hdata3 = fifo.get( 2 )
        self.assertEqual( [2,3], hdata3 )

        fifo.clear()
        self.assertEqual( 0, fifo.available_tokens() )
        self.assertEqual( 0, fifo.available_unused_tokens() )
        self.assertEqual( 0, fifo.available_history_tokens() )
        
        return
        
    
### EOF ######################################################################
