import unittest

from enthought.logger import logger

from enthought.graph import Node
from enthought.graph import CompositeNode
from enthought.graph import Relation
       
class SimpleNodeTestCase( unittest.TestCase ) :

    def test_basics(self) :
        cnode = CompositeNode('composite node')
        
        n = Node('aNode')
        self.assertEqual( n.name, 'aNode', msg='Node failed to set name')

        n.container = cnode
        self.assertEqual( [n], cnode.nodes, 
            msg='Node unknown to container after set_container')
        
        port_count = 5
        port_list = []
        for i in range( port_count ) :
            p = n.new_port('Port %s' % i)
            p.container = n
            port_list.append(p)
            self.assertEqual( n, p.container, 
                msg='Port container incorrect after set_container')
        
        self.assertEqual( port_count, len(n.ports), 
            msg='Expected %s port in node' % port_count )
        self.assertEqual( port_list, n.ports, 
            msg='Node port list incorrect ')
        
        r = cnode.new_relation('Relation 1')
        r.container =None # undo the auto set_container
        r11 = Relation('Relation 1.1')
        
        p = n.ports[0]

        self.failUnlessRaises(ValueError, p.link, r)
        self.failUnlessRaises(ValueError, p.link, r11)
        
        r.container = cnode
        self.assertEqual( cnode, r.container, 
            msg="Relation container not correct" )
        
        p.link(r)

        self.assertEqual( 1, len(p.relations), 
            msg='Expected 1 relation in port' )
        self.assertEqual( 1, len(r.ports), 
            msg='Expected 1 port in relation' )
    
        p = n.ports[1]
        p.link(r)
        self.assertEqual( 1, len(p.relations), 
            msg='Expected 1 relation in port got %s' % len(p.relations) )
        self.assertEqual( 2, len(r.ports), 
            msg='Expected 2 ports in relation, got %s' % len(r.ports) )
        return

    def test_node_name_uniqueness(self) :
        cnode = CompositeNode('composite node')
        
        n = Node('aNode')
        n.container = cnode
        self.assertEqual( [n], cnode.nodes, 
            msg='Node unknown to container after set_container')

        n2 = Node('aNode')
        try:
            n2.container = cnode
            self.fail(msg='Duplicate actor name allowed.')
            
        except ValueError:
            self.assertEqual( None, n2.container )
            pass
        
        n2.name = 'different'
        n2.container = cnode

        try:
            n.name = 'different'
            self.fail(msg='Duplicate actor name allowed during rename.')
            
        except ValueError:
            self.assertEqual( 'aNode', n.name )
            pass

        return
        


    def test_port_by_name(self) :
        n = Node('aNode')

        port_list = []
        port_count = 3
        for i in range( port_count ) :
            p = n.new_port('Port %s' % i)
            p.container = n
            port_list.append( p )
        
        for port in port_list :
            node_port_by_name = n.get_port(port.name)
            self.assertEqual( port, node_port_by_name,
                "Node.get_port returns incorrect port: %s %s" 
                % (port.name, node_port_by_name ))
        return

    def test_port_relation_linking( self ) :
        cnode = CompositeNode('composite node')
        n = Node('aNode')
        
        ports=[]
        for i in range(4) :
            ports.append( n.new_port('port %s' % i ) )
            ports[i].container = n
        
        r1 = cnode.new_relation('Relation 1')
        r1.container = cnode
        r2 = cnode.new_relation('Relation 2')
        r2.container = cnode
        
        self.assertEqual( [], r1.linked_ports(), 
            "Relation 1 should have no linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [], r2.linked_ports(), 
            "Relation 2 should have no linked ports, has %s" 
            % len(r2.linked_ports()))
        
        ports[0].link(r1)
        self.assertEqual( [ports[0]], r1.linked_ports(), 
            "Relation 1 should have one linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [], r2.linked_ports(), 
            "Relation 2 should have no linked ports, has %s" 
            % len(r2.linked_ports()))

        ports[1].link(r2)
        self.assertEqual( [ports[0]], r1.linked_ports(), 
            "Relation 1 should have one linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [ports[1]], r2.linked_ports(), 
            "Relation 2 should have one linked ports, has %s" 
            % len(r2.linked_ports()))

        ports[2].link(r2)
        self.assertEqual( [ports[0]], r1.linked_ports(), 
            "Relation 1 should have one linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [ports[1], ports[2]], r2.linked_ports(), 
            "Relation 2 should have two linked ports, has %s" 
            % len(r2.linked_ports()))

        ports[3].link(r1)
        self.assertEqual( [ports[0], ports[3]], r1.linked_ports(), 
            "Relation 1 should have two linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [ports[1], ports[2]], r2.linked_ports(), 
            "Relation 2 should have two linked ports, has %s" 
            % len(r2.linked_ports()))

        ports[1].unlink( r1 )
        self.assertEqual( [ports[0], ports[3]], r1.linked_ports(), 
            "Relation 1 should have two linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [ports[1], ports[2]], r2.linked_ports(), 
            "Relation 2 should have two linked ports, has %s" 
            % len(r2.linked_ports()))
        
        ports[1].unlink( r2 )
        self.assertEqual( [ports[0], ports[3]], r1.linked_ports(), 
            "Relation 1 should have two linked ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [ports[2]], r2.linked_ports(), 
            "Relation 2 should have one linked ports, has %s" 
            % len(r2.linked_ports()))

        self.assertEqual( [ports[0], ports[3]], r1.linked_ports(), 
            "Relation 1 should have two connected ports, has %s" 
            % len(r1.linked_ports()))
        self.assertEqual( [ports[0]], r1.linked_ports(ports[3]), 
            "Relation 1 should have two connected ports, has %s" 
            % len(r1.linked_ports()))

        self.assertEqual( [ports[3]], ports[0].connected_ports(), 
            "One port should be connected to port[0], got %s" 
            % len(ports[0].connected_ports()))
        self.assertEqual( [ports[0]], ports[3].connected_ports(), 
            "One port should be connected to port[3], got %s" 
            % len(ports[3].connected_ports()))
        
        ports[1].link(r1)
        self.assertEqual( [ports[3], ports[1]], ports[0].connected_ports(), 
            "Two ports should be connected to port[0], got %s" 
            % len(ports[0].connected_ports()))
        self.assertEqual( [ports[0], ports[1]], ports[3].connected_ports(), 
            "Two ports should be connected to port[3], got %s" 
            % len(ports[3].connected_ports()))
        return
        
    def test_composite_node_nodes(self) :
        cnode = CompositeNode("container")
        n1 = Node(name="n1")
        n1.container = cnode
        n2 = Node("n2")
        n2.container = cnode
        
        self.assertEqual( [n1, n2], cnode.reachable_node_list(), 
            "reachable_node_list is incorrect, got: %s" 
            % cnode.reachable_node_list())
        
        cnode2 = CompositeNode("container 2")
        cnode.container = cnode2

        n3 = Node("n2")
        n3.container = cnode2
        
        self.assertEqual( [n1, n2, n3], cnode2.reachable_node_list(), 
            "nested reachable_node_list is incorrect, got: %s" 
            % cnode.reachable_node_list())
        
        n2.container = None
        self.assertEqual( [n1, n3], cnode2.reachable_node_list(), 
            "nested reachable_node_list is incorrect, got: %s" 
            % cnode.reachable_node_list())
        return
        
        
    def test_composite_node_relations(self) :
        cnode = CompositeNode("container")
        r1 = cnode.new_relation("r1")
        r1.container = cnode
        
        self.assertEqual( [r1], cnode.relations,
            "CompositeNode has incorrect relation list, got: %s" 
            % cnode.relations )

        r2 = cnode.new_relation("r2")
        r2.container = cnode
        
        self.assertEqual( [r1,r2], cnode.relations,
            "CompositeNode has incorrect relation list, got: %s"
            % cnode.relations )
        
        r1.container = None
        self.assertEqual( [r2], cnode.relations,
            "CompositeNode has incorrect relation list after remove, got: %s"
            % cnode.relations )
        return
        
    def test_composite_port(self) :
        n = Node("node")
        cnode = CompositeNode("composite node")
        p = n.new_port("port")
        
        try :
            p.container = cnode
            self.fail()
            
        except TypeError:
            pass
            
        cport = cnode.new_port("composite port")
        cport.container = cnode
        
        #fixme: should this throw?
        cport.container = n
        return

    def test_auto_set_container(self) :
        cnode = CompositeNode("composite node")
        p = cnode.new_port("port")
        self.assertEqual( p.container, cnode, 
            "composite node not set as port container")
        
        n = cnode.new_node("node")
        self.assertEqual( p.container, cnode, 
            "composite node not set as node container")
        
        p = n.new_port("port")
        self.assertEqual( p.container, n, 
            "node not set as port container")
        return
        
    def test_connect(self) :
        cnode = CompositeNode("composite node")
        
        n1 = cnode.new_node("node 1")
        p1 = n1.new_port("port 1")
        n2 = cnode.new_node("node 2")
        p2 = n2.new_port("port 2")
        
        relation = cnode.connect(p1, p2, 'relation' )
        self.assertNotEqual( relation, None, 'No relation returned' )
        self.assertEqual( relation.name, 'relation', 
            'relation not named correctly got: %s' % relation.name )

        relation = cnode.connect(p1, p2 )
        self.assertNotEqual( relation, None, 'No relation returned' )
        self.assertEqual( relation.name, p2.name, 
            'relation not named correctly from port 2 got: %s' % relation.name )
        
        self.failUnlessRaises(ValueError, cnode.connect, p1, p1)

        self.failUnlessRaises(ValueError, cnode.connect, p1, None)
        
        self.failUnlessRaises(ValueError, cnode.connect, None, p2)
        return

    def test_unique_names(self):
        cnode = CompositeNode("composite node")
        
        r = cnode.new_relation('duplicate relation name')
        self.failUnlessRaises( ValueError, cnode.new_relation, 
            'duplicate relation name')

        n = cnode.new_node('duplicate node name')
        self.failUnlessRaises( ValueError, cnode.new_node, 
            'duplicate node name')
        
        n2 = cnode.new_node('node')
        self.failUnlessRaises( ValueError, 
            cnode.connect, n, n2, 'duplicate relation name')
        
        p = n.new_port('duplicate port name')
        self.failUnlessRaises( ValueError, n.new_port, 'duplicate port name')
        
        p = cnode.new_port('duplicate port name')
        self.failUnlessRaises( ValueError, cnode.new_port, 'duplicate port name')
        return

    
    
### EOF ######################################################################
