import unittest

from enthought.util.testingx import assert_array_almost_equal

from enthought.logger import logger
from enthought.graph   import CompositeActor, SdfDirector
from enthought.graph.exceptions import FireError

from enthought.graph.tests.tsd_actors  import *

class TsdNoSubgroup :

    def setUp(self) :
        logger.debug("-------------------------------------------------------")

        # create the actors that use rock physics
        self._createActors()

        # make connections between ports
        self._makeConnections()

        # install director, only in top composite actor
        self.director = SdfDirector()
        self.director.container = self.top

        logger.debug("test execution of TestTsdNoSubgroup begins...")
        
        self.top.preinitialize()
        self.top.initialize()
        return

    def _createActors(self) :
        """ The actors created in this function are the ...Actor class
        instances. These DO implement rock physics functions. Inputs
        are scalars or short sequences as demanded.
        """

        from tsd_actors import TsdDataInActor
        from tsd_actors import TempPressLinearActor
        from tsd_actors import BrinePropertiesActor
        from tsd_actors import VsWetTrendsActor
        from tsd_actors import BiotCoeffRhobActor
        from tsd_actors import TsdModelRhobActor

        self.top = CompositeActor('composite')

        self.tsd_data_in_actor = TsdDataInActor('tsd_data_in_actor')
        self.tsd_data_in_actor.container = self.top

        self.temp_press_linear_actor = \
            TempPressLinearActor('temp_press_linear_actor')
        self.temp_press_linear_actor.container = self.top

        self.brine_prop_sand_actor = \
            BrinePropertiesActor('brine_prop_sand_actor')
        self.brine_prop_sand_actor.container = self.top

        self.brine_prop_shale_actor = \
            BrinePropertiesActor('brine_prop_shale_actor')
        self.brine_prop_shale_actor.container = self.top

        self.vs_wet_trends_actor = \
            VsWetTrendsActor('vs_wet_trends_actor')
        self.vs_wet_trends_actor.container = self.top

        self.biot_coeff_rhob_actor = \
            BiotCoeffRhobActor('biot_coeff_rhob_actor')
        self.biot_coeff_rhob_actor.container = self.top

        self.biot_coeff_rhob_shale_actor = \
            BiotCoeffRhobActor('biot_coeff_rhob_shale_actor')
        self.biot_coeff_rhob_shale_actor.container = self.top

        self.tsdmodel_rhob_actor = \
            TsdModelRhobActor('tsdmodel_rhob_actor')
        self.tsdmodel_rhob_actor.container = self.top

        return

    def _makeConnections(self) :

        # temp_press_linear_actor inputs
        self.top.connect( self.tsd_data_in_actor.get_port('watdth_out'), 
                        self.temp_press_linear_actor.get_port('watdth_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('oftemp_out'), 
                        self.temp_press_linear_actor.get_port('oftemp_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('tgrad_out'), 
                        self.temp_press_linear_actor.get_port('tgrad_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('pgrad_out'), 
                        self.temp_press_linear_actor.get_port('pgrad_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('depth_out'), 
                        self.temp_press_linear_actor.get_port('depth_in'))

        # brine_prop_sand_actor inputs
        rel_sandbrine_tform = \
        self.top.connect( self.temp_press_linear_actor.get_port('tform_out'),
                        self.brine_prop_sand_actor.get_port('tform_in'))
        rel_sandbrine_ppore = \
        self.top.connect( self.temp_press_linear_actor.get_port('ppore_out'), 
                        self.brine_prop_sand_actor.get_port('ppore_in'))
        # have to give explicit name to relation to avoid conflict with the
        # salw_in to brine_prop_shale_actor below
        self.top.connect( self.tsd_data_in_actor.get_port('salw_ss_out'), 
                        self.brine_prop_sand_actor.get_port('salw_in'),
                        'sand_salw_rel')

        # brine_prop_shale_actor inputs
        # tform_in: relation already created
        self.brine_prop_shale_actor.get_port('tform_in').link(rel_sandbrine_tform)
        # ppore_in: relation already created
        self.brine_prop_shale_actor.get_port('ppore_in').link(rel_sandbrine_ppore)
        # have to give explicit name to relation to avoid conflict with the
        # salw_in to brine_prop_sand_actor above
        self.top.connect( self.tsd_data_in_actor.get_port('salw_sh_out'), 
                        self.brine_prop_shale_actor.get_port('salw_in'),
                        'shale_salw_rel')

        # vs_wet_trends_actor inputs
        rel_vp_ss_out = \
            self.top.connect( self.tsd_data_in_actor.get_port('vp_ss_out'), 
                            self.vs_wet_trends_actor.get_port('vp_ss_in'),
                            'vswettrends_vp_ss_rel')
        rel_vp_sh_out = \
            self.top.connect( self.tsd_data_in_actor.get_port('vp_sh_out'), 
                        self.vs_wet_trends_actor.get_port('vp_sh_in'),
                        'vswettrends_vp_sh_rel')

        # biot_coeff_rhob inputs
        self.top.connect( self.brine_prop_sand_actor.get_port('k_wat_out'), 
                        self.biot_coeff_rhob_actor.get_port('k_wat_ss_in'))
        rel_rho_wat_ss_out = \
        self.top.connect( self.brine_prop_sand_actor.get_port('rho_wat_out'), 
                        self.biot_coeff_rhob_actor.get_port('rho_wat_ss_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('kg_ss_out'), 
                        self.biot_coeff_rhob_actor.get_port('kg_ss_in'),
                        'biot_coeff_kg_ss_rel')
        self.top.connect( self.tsd_data_in_actor.get_port('rhog_ss_out'), 
                        self.biot_coeff_rhob_actor.get_port('rhog_ss_in'),
                        'biot_coeff_rhog_ss_rel')
        self.top.connect( self.tsd_data_in_actor.get_port('rhob_ss_out'), 
                        self.biot_coeff_rhob_actor.get_port('rhob_ss_in'),
                        'biot_coeff_rhob_ss_rel')
        self.top.connect( self.tsd_data_in_actor.get_port('vp_ss_out'), 
                        self.biot_coeff_rhob_actor.get_port('vp_ss_in'))
        rel_vs_ss_out = \
        self.top.connect( self.vs_wet_trends_actor.get_port('vs_ss_out'), 
                        self.biot_coeff_rhob_actor.get_port('vs_ss_in'))
        self.top.connect( self.brine_prop_shale_actor.get_port('k_wat_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('k_wat_ss_in'), 'k_wat_ss_in_shale')
        rel_rho_wat_sh_out = \
        self.top.connect( self.brine_prop_shale_actor.get_port('rho_wat_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('rho_wat_ss_in'), 'rho_wat_ss_in_shale')
        rel_kg_sh_out = \
        self.top.connect( self.tsd_data_in_actor.get_port('kg_sh_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('kg_ss_in'), 'kg_ss_in_shale')
        rel_rhog_sh_out = \
        self.top.connect( self.tsd_data_in_actor.get_port('rhog_sh_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('rhog_ss_in'), 'rhog_ss_in_shale')
        rel_rhob_sh_out = \
        self.top.connect( self.tsd_data_in_actor.get_port('rhob_sh_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('rhob_ss_in'), 'rhob_ss_in_shale')
        self.top.connect( self.tsd_data_in_actor.get_port('vp_sh_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('vp_ss_in'), 'vp_ss_in_shale')
        rel_vs_sh_out = \
        self.top.connect( self.vs_wet_trends_actor.get_port('vs_sh_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('vs_ss_in'), 'vs_ss_in_shale')

        # tsdmodel_rhob_actor inputs
        # vp_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vp_ss_in').link(rel_vp_ss_out)
        # vs_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vs_ss_in').link(rel_vs_ss_out)
        # have to give explicit name to relation to avoid name conflict
        self.top.connect( self.tsd_data_in_actor.get_port('rhob_ss_out'), 
                        self.tsdmodel_rhob_actor.get_port('rhob_ss_in'),
                        'tsdmodel_rhob_ss_rel')
        # have to give explicit name to relation to avoid name conflict
        self.top.connect( self.tsd_data_in_actor.get_port('rhog_ss_out'), 
                        self.tsdmodel_rhob_actor.get_port('rhog_ss_in'),
                        'tsdmodel_rhog_ss_rel')
        # have to give explicit name to relation to avoid name conflict
        self.top.connect( self.tsd_data_in_actor.get_port('kg_ss_out'), 
                        self.tsdmodel_rhob_actor.get_port('kg_ss_in'),
                        'tsdmodel_kg_ss_rel')
        self.top.connect( self.tsd_data_in_actor.get_port('gg_ss_out'), 
                        self.tsdmodel_rhob_actor.get_port('gg_ss_in'))

        # vp_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vp_sh_in').link(rel_vp_sh_out)
        # vs_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vs_sh_in').link(rel_vs_sh_out)
        # rhob_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rhob_sh_in').link(rel_rhob_sh_out)
        # rhog_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rhog_sh_in').link(rel_rhog_sh_out)
        # kg_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('kg_sh_in').link(rel_kg_sh_out)

        self.top.connect( self.tsd_data_in_actor.get_port('phit_out'), 
                        self.tsdmodel_rhob_actor.get_port('phit_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('xsh_out'), 
                        self.tsdmodel_rhob_actor.get_port('xsh_in'))
        # rho_wat_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rho_wat_ss_in').link(rel_rho_wat_ss_out)
        # rho_wat_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rho_wat_sh_in').link(rel_rho_wat_sh_out)

        return

    def tearDown( self ) :
        logger.debug("test execution ends.")
        self.top.wrapup()
        return

    def test_compare_matlab( self ) :
        logger.debug("----> test_compare_matlab")

        # set input arrays in TsdDataInActor
        num_repeats = 0
        increment = .6096
        self.tsd_data_in_actor.build_input_arrays(num_repeats, increment)

        # do one iteration of compute graph
        try:
            self.top.iterate(1)
        except FireError, e:
            import testoob.exceptions
            if (e.etype == testoob.exceptions.SkipTestException):
                raise e.value

        # test equality: computations in chip TempPressLinearActor
        #     output: tform
        known_tform = array((14.0794, 14.0916, 14.1038, 14.1160, 14.1283))
        result = self.temp_press_linear_actor.value[0]
        assert_array_almost_equal(known_tform, result, decimal=3)
        #     output: ppore
        known_ppore = array((20.8405, 20.8469, 20.8534, 20.8598, 20.8662))
        result = self.temp_press_linear_actor.value[1]
        assert_array_almost_equal(known_ppore, result, decimal=3)

        # test equality: computations in chip BrinePropertiesActor (sand)
        #     output: rho_wat_ss
        known_rho_wat_ss = array((1.0291, 1.0291, 1.0291, 1.0291, 1.0291))
        result = self.brine_prop_sand_actor.value[0]
        assert_array_almost_equal(known_rho_wat_ss, result, decimal=3)
        #     output: k_wat_ss
        known_k_wat_ss = array((2.4184, 2.4186, 2.4187, 2.4189, 2.4190))
        result = self.brine_prop_sand_actor.value[1]
        assert_array_almost_equal(known_k_wat_ss, result, decimal=3)

        # test equality: computations in chip BrinePropertiesActor (shale)
        #     output: rho_wat_sh
        known_rho_wat_sh = array((1.0187, 1.0187, 1.0187, 1.0187, 1.0187))
        result = self.brine_prop_shale_actor.value[0]
        assert_array_almost_equal(known_rho_wat_sh, result, decimal=3)
        #     output: k_wat_sh
        known_k_wat_sh = array((2.3355, 2.3357, 2.3359, 2.3360, 2.3362))
        result = self.brine_prop_shale_actor.value[1]
        assert_array_almost_equal(known_k_wat_sh, result, decimal=3)

        # test equality: computations in chip VsWetTrendsActor
        #     output: vs_ss
        known_vs_ss = array((1.6097, 1.6097, 1.6985, 1.5447, 1.6886))
        result = self.vs_wet_trends_actor.value[0]
        assert_array_almost_equal(known_vs_ss, result, decimal=3)
        #     output: vs_sh
        known_vs_sh = array((1.1666, 1.1666, 1.2455, 1.0873, 1.1270))
        result = self.vs_wet_trends_actor.value[1]
        assert_array_almost_equal(known_vs_sh, result, decimal=3)

        # test equality: computations in chip BiotCoeffRhobActor
        #     output: biot_ss
        known_biot_ss = array((0.8318, 0.8319, 0.8255, 0.9081, 0.9507))
        result = self.biot_coeff_rhob_actor.value[0]
        assert_array_almost_equal(known_biot_ss, result, decimal=3)
        #     output: biotmax_ss
        known_biotmax_ss = array((0.8784, 0.8784, 0.8638, 0.9028, 0.8965))
        result = self.biot_coeff_rhob_actor.value[1]
        assert_array_almost_equal(known_biotmax_ss, result, decimal=3)

        # test equality: computations in chip TsdModelRhobActor
        #     output: vp_wet
        known_vp_wet = array((2.5119, 2.5111, 2.4166, 2.1738, 3.3268))
        result = self.tsdmodel_rhob_actor.value[0]
        assert_array_almost_equal(known_vp_wet, result, decimal=3)
        #     output: vs_wet
        known_vs_wet = array((1.1175, 1.1168, 1.0428, 0.8316, 1.3270))
        result = self.tsdmodel_rhob_actor.value[1]
        assert_array_almost_equal(known_vs_wet, result, decimal=3)
        #     output: rhob_wet
        known_rhob_wet = array((2.3742, 2.3742, 2.3470, 2.3591, 2.1056))
        result = self.tsdmodel_rhob_actor.value[2]
        assert_array_almost_equal(known_rhob_wet, result, decimal=3)
        #     output: vp_sm
        known_vp_sm = array((2.2285, 2.2262, 1.9883, 1.7346, 3.0104))
        result = self.tsdmodel_rhob_actor.value[3]
        assert_array_almost_equal(known_vp_sm, result, decimal=3)
        #     output: vs_sm
        known_vs_sm = array((1.0382, 1.0363, 0.8329, 0.6158, 1.9272))
        result = self.tsdmodel_rhob_actor.value[4]
        assert_array_almost_equal(known_vs_sm, result, decimal=3)
        #     output: phi_sm
        known_phi_sm = array((0.3936, 0.3943, 0.3859, 0.3872, 3.3983))
        result = self.tsdmodel_rhob_actor.value[5]
        assert_array_almost_equal(known_phi_sm, result, decimal=3)
        #     output: rhob_sm
        known_rhob_sm = array((2.0210, 2.0200, 2.0304, 2.0236, -1.7265))
        result = self.tsdmodel_rhob_actor.value[6]
        assert_array_almost_equal(known_rhob_sm, result, decimal=3)
        #     output: phie_sm
        known_phie_sm = array((0.4056, 0.4063, 0.4006, 0.3983, 2.9566))
        result = self.tsdmodel_rhob_actor.value[7]
        assert_array_almost_equal(known_phie_sm, result, decimal=3)
        #     output: rhog_sm
        known_rhog_sm = array((2.6649, 2.6649, 2.6596, 2.6520, 2.1781))
        result = self.tsdmodel_rhob_actor.value[8]
        assert_array_almost_equal(known_rhog_sm, result, decimal=3)
        #     output: kg_sm
        known_kg_sm = array((39.9917, 40.0025, 38.4136, 41.578, 46.3220))
        result = self.tsdmodel_rhob_actor.value[9]
        assert_array_almost_equal(known_kg_sm, result, decimal=3)
        #     output: vp_hs
        known_vp_hs = array((4.4659, 4.4659, 4.6819, 4.5726, 4.9115))
        result = self.tsdmodel_rhob_actor.value[10]
        assert_array_almost_equal(known_vp_hs, result, decimal=3)
        #     output: vs_hs
        known_vs_hs = array((2.6330, 2.6330, 2.8428, 2.6840, 2.9657))
        result = self.tsdmodel_rhob_actor.value[11]
        assert_array_almost_equal(known_vs_hs, result, decimal=3)
        #     output: phi_hs
        known_phi_hs = array((0.0137, 0.0137, 0.0120, 0.0084, -0.0203))
        result = self.tsdmodel_rhob_actor.value[12]
        assert_array_almost_equal(known_phi_hs, result, decimal=3)
        #     output: xlam
        known_xlam = array((0.7374, 0.7379, 0.7074, 0.7163, 0.9054))
        result = self.tsdmodel_rhob_actor.value[13]
        assert_array_almost_equal(known_xlam, result, decimal=3)
        #     output: xdis
        known_xdis = array((-0.0498, -0.0499, -0.0644, -0.0643, -0.2674))
        result = self.tsdmodel_rhob_actor.value[14]
        assert_array_almost_equal(known_xdis, result, decimal=3)
        #     output: phi_ss
        known_phi_ss = array((0.2159, 0.2159, 0.1806, 0.1718, 0.1300))
        result = self.tsdmodel_rhob_actor.value[15]
        assert_array_almost_equal(known_phi_ss, result, decimal=3)
        #     output: phi_sh
        known_phi_sh = array((0.0632, 0.0632, 0.0665, 0.0491, -0.1563))
        result = self.tsdmodel_rhob_actor.value[16]
        assert_array_almost_equal(known_phi_sh, result, decimal=3)

        return
                
class TsdNoSubgroupTestCase( TsdNoSubgroup, unittest.TestCase ) :
    pass
    
### EOF ######################################################################
