import unittest

from enthought.util.testingx import assert_array_almost_equal

from enthought.logger import logger
from enthought.graph   import CompositeActor, SdfDirector

from tsd_no_subgroup_test_case import TsdNoSubgroup

def actorFactory(anActorClass, container, *args) :
    actor = apply(anActorClass, args)
    actor.container = container

    return actor

def connectPorts(container, out_port, in_port, name=None) :

    # get list of relations associated with output port
    if out_port.container.container == in_port.container.container :
        # this is an outside link to a relation
        relList = out_port.relations_list()
    else :
        # this is an inside link to a relation
        relList = out_port.inside_relations_list()

    # are there any relations?
    if len(relList) == 0 :
        # no relations exist on this output port: do a connect

        # code to build a name for the relation????
        # maybe the output port's name concatenated with the
        # actor containing the port?
        if name is None:
            name = "rel_" + out_port.container.name + "->" + out_port.name
        relation = container.connect(out_port, in_port, name)

    else :
        # relations exist on the output port: do a link

        # big assumption here: output ports can have more than
        # one relation, but one is sufficient, so use the first
        # one.
        relation = relList[0]

        # link this input port to output port using its relation
        in_port.link(relation)

    return relation
   
class TsdScriptSubgroup2( TsdNoSubgroup ) :

    def setUp(self) :
        logger.debug("-------------------------------------------------------")

        # create the actors that use rock physics
        self._buildGraph()

        # install director, only in top composite actor
        self.director = SdfDirector()
        self.director.container = self.top

        logger.debug("test execution of TestTsdWithSubgroup2 begins...")
        
        self.top.preinitialize()
        self.top.initialize()
        return

    def _buildGraph(self) :
        """ The actors created in this function are the ...Actor class
        instances. These DO implement rock physics functions. Inputs
        are scalars or short sequences as demanded.
        """
        
        from tsd_actors import TsdDataInActor
        from tsd_actors import TempPressLinearActor
        from tsd_actors import BrinePropertiesActor
        from tsd_actors import TsdSubgroup2Actor
        from tsd_actors import VsWetTrendsActor
        from tsd_actors import BiotCoeffRhobActor
        from tsd_actors import TsdModelRhobActor

        self.top = CompositeActor('composite')

        self.tsd_data_in_actor = actorFactory(TsdDataInActor, self.top,
                                     'tsd_data_in_actor')

        self.temp_press_linear_actor = actorFactory(TempPressLinearActor,
                                           self.top, 'temp_press_linear_actor')

        self.brine_prop_sand_actor = actorFactory(BrinePropertiesActor,
                                         self.top, 'brine_prop_sand_actor')

        self.brine_prop_shale_actor = actorFactory(BrinePropertiesActor,
                                          self.top, 'brine_prop_shale_actor')

        self.vs_wet_trends_actor = actorFactory(VsWetTrendsActor,
                                       self.top, 'vs_wet_trends_actor')


        # build the composite actor for the subgroup
        self.subgroup_ca = actorFactory(TsdSubgroup2Actor,
                               self.top, 'tsd_subgroup_composite_actor')

        # build two nodes contained by the subgroup composite actor
        self.biot_coeff_rhob_actor = actorFactory(BiotCoeffRhobActor,
                                         self.subgroup_ca,
                                         'biot_coeff_rhob_actor')

        self.biot_coeff_rhob_shale_actor = actorFactory(BiotCoeffRhobActor,
                                         self.subgroup_ca,
                                         'biot_coeff_rhob_shale_actor')

        self.tsdmodel_rhob_actor = actorFactory(TsdModelRhobActor,
                                       self.subgroup_ca,
                                       'tsdmodel_rhob_actor')

        # temp_press_linear_actor inputs
        connectPorts(self.top, self.tsd_data_in_actor.get_port('watdth_out'), 
                     self.temp_press_linear_actor.get_port('watdth_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('oftemp_out'), 
                     self.temp_press_linear_actor.get_port('oftemp_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('tgrad_out'), 
                     self.temp_press_linear_actor.get_port('tgrad_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('pgrad_out'), 
                     self.temp_press_linear_actor.get_port('pgrad_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('depth_out'), 
                     self.temp_press_linear_actor.get_port('depth_in'))

        # brine_prop_sand_actor inputs
        connectPorts(self.top, self.temp_press_linear_actor.get_port('tform_out'),
                     self.brine_prop_sand_actor.get_port('tform_in'))
        connectPorts(self.top, self.temp_press_linear_actor.get_port('ppore_out'), 
                        self.brine_prop_sand_actor.get_port('ppore_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('salw_ss_out'), 
                        self.brine_prop_sand_actor.get_port('salw_in'))

        # brine_prop_shale_actor inputs
        connectPorts(self.top, self.temp_press_linear_actor.get_port('tform_out'),
                     self.brine_prop_shale_actor.get_port('tform_in'))
        connectPorts(self.top, self.temp_press_linear_actor.get_port('ppore_out'),
                     self.brine_prop_shale_actor.get_port('ppore_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('salw_sh_out'), 
                     self.brine_prop_shale_actor.get_port('salw_in'))

        # vs_wet_trends_actor inputs
        connectPorts(self.top, self.tsd_data_in_actor.get_port('vp_ss_out'), 
                     self.vs_wet_trends_actor.get_port('vp_ss_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('vp_sh_out'), 
                     self.vs_wet_trends_actor.get_port('vp_sh_in'))

        # now do the connections to the subgroup composite actor
        # ca_vp_ss_in: relation already created
        connectPorts(self.top, self.tsd_data_in_actor.get_port('vp_ss_out'),
                     self.subgroup_ca.get_port('ca_vp_ss_in'))
        connectPorts(self.top, self.vs_wet_trends_actor.get_port('vs_ss_out'),
                        self.subgroup_ca.get_port('ca_vs_ss_in'))
        # ca_vp_sh_in: relation already created
        connectPorts(self.top, self.tsd_data_in_actor.get_port('vp_sh_out'),
                     self.subgroup_ca.get_port('ca_vp_sh_in'))
        connectPorts(self.top, self.vs_wet_trends_actor.get_port('vs_sh_out'),
                        self.subgroup_ca.get_port('ca_vs_sh_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('rhob_ss_out'),
                     self.subgroup_ca.get_port('ca_rhob_ss_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('rhog_ss_out'),
                     self.subgroup_ca.get_port('ca_rhog_ss_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('kg_ss_out'),
                     self.subgroup_ca.get_port('ca_kg_ss_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('gg_ss_out'),
                     self.subgroup_ca.get_port('ca_gg_ss_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('rhob_sh_out'),
                     self.subgroup_ca.get_port('ca_rhob_sh_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('rhog_sh_out'),
                     self.subgroup_ca.get_port('ca_rhog_sh_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('kg_sh_out'),
                     self.subgroup_ca.get_port('ca_kg_sh_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('phit_out'),
                     self.subgroup_ca.get_port('ca_phit_in'))
        connectPorts(self.top, self.tsd_data_in_actor.get_port('xsh_out'),
                     self.subgroup_ca.get_port('ca_xsh_in'))
        connectPorts(self.top, self.brine_prop_sand_actor.get_port('k_wat_out'),
                     self.subgroup_ca.get_port('ca_k_wat_ss_in'))
        connectPorts(self.top, self.brine_prop_sand_actor.get_port('rho_wat_out'),
                     self.subgroup_ca.get_port('ca_rho_wat_ss_in'))
        connectPorts(self.top, self.brine_prop_shale_actor.get_port('k_wat_out'),
                     self.subgroup_ca.get_port('ca_k_wat_sh_in'))
        connectPorts(self.top, self.brine_prop_shale_actor.get_port('rho_wat_out'),
                     self.subgroup_ca.get_port('ca_rho_wat_sh_in'))

        # now make connections from subgroup inputs to enclosed nodes

        # biot_coeff_rhob inputs
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_k_wat_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('k_wat_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rho_wat_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('rho_wat_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_kg_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('kg_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhog_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('rhog_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhob_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('rhob_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vp_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('vp_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vs_ss_in'), 
                     self.biot_coeff_rhob_actor.get_port('vs_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_k_wat_sh_in'), 
                     self.biot_coeff_rhob_shale_actor.get_port('k_wat_ss_in'), 'k_wat_ss_in_shale')
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rho_wat_sh_in'), 
                     self.biot_coeff_rhob_shale_actor.get_port('rho_wat_ss_in'), 'rho_wat_ss_in_shale')
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_kg_sh_in'), 
                     self.biot_coeff_rhob_shale_actor.get_port('kg_ss_in'), 'kg_ss_in_shale')
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhog_sh_in'), 
                     self.biot_coeff_rhob_shale_actor.get_port('rhog_ss_in'), 'rhog_ss_in_shale')
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhob_sh_in'), 
                     self.biot_coeff_rhob_shale_actor.get_port('rhob_ss_in'), 'rhob_ss_in_shale')
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vp_sh_in'), 
                     self.biot_coeff_rhob_shale_actor.get_port('vp_ss_in'), 'vp_ss_in_shale')
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vs_sh_in'),
                     self.biot_coeff_rhob_shale_actor.get_port('vs_ss_in'), 'vs_ss_in_shale')

        # tsdmodel_rhob_actor inputs
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vp_ss_in'),
                     self.tsdmodel_rhob_actor.get_port('vp_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vs_ss_in'),
                     self.tsdmodel_rhob_actor.get_port('vs_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhob_ss_in'),
                     self.tsdmodel_rhob_actor.get_port('rhob_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhog_ss_in'),
                     self.tsdmodel_rhob_actor.get_port('rhog_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_kg_ss_in'),
                     self.tsdmodel_rhob_actor.get_port('kg_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_gg_ss_in'), 
                     self.tsdmodel_rhob_actor.get_port('gg_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vp_sh_in'),
                     self.tsdmodel_rhob_actor.get_port('vp_sh_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_vs_sh_in'),
                     self.tsdmodel_rhob_actor.get_port('vs_sh_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhob_sh_in'),
                     self.tsdmodel_rhob_actor.get_port('rhob_sh_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rhog_sh_in'),
                     self.tsdmodel_rhob_actor.get_port('rhog_sh_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_kg_sh_in'),
                     self.tsdmodel_rhob_actor.get_port('kg_sh_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_phit_in'), 
                     self.tsdmodel_rhob_actor.get_port('phit_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_xsh_in'), 
                     self.tsdmodel_rhob_actor.get_port('xsh_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rho_wat_ss_in'),
                     self.tsdmodel_rhob_actor.get_port('rho_wat_ss_in'))
        connectPorts(self.subgroup_ca, self.subgroup_ca.get_port('ca_rho_wat_sh_in'),
                     self.tsdmodel_rhob_actor.get_port('rho_wat_sh_in'))

        # connect outputs of internal nodes to composite outputs
        connectPorts(self.subgroup_ca,
                     self.biot_coeff_rhob_actor.get_port('biot_ss_out'), 
                     self.subgroup_ca.get_port('ca_biot_ss_out'))
        connectPorts(self.subgroup_ca,
                     self.biot_coeff_rhob_actor.get_port('biotmax_ss_out'), 
                     self.subgroup_ca.get_port('ca_biotmax_ss_out'))
        connectPorts(self.subgroup_ca,
                     self.biot_coeff_rhob_shale_actor.get_port('biot_ss_out'), 
                     self.subgroup_ca.get_port('ca_biot_sh_out'))
        connectPorts(self.subgroup_ca,
                     self.biot_coeff_rhob_shale_actor.get_port('biotmax_ss_out'), 
                     self.subgroup_ca.get_port('ca_biotmax_sh_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('vp_wet_out'), 
                     self.subgroup_ca.get_port('ca_vp_wet_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('vs_wet_out'), 
                     self.subgroup_ca.get_port('ca_vs_wet_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('rhob_wet_out'), 
                     self.subgroup_ca.get_port('ca_rhob_wet_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('vp_sm_out'), 
                     self.subgroup_ca.get_port('ca_vp_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('vs_sm_out'), 
                     self.subgroup_ca.get_port('ca_vs_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('phi_sm_out'), 
                     self.subgroup_ca.get_port('ca_phi_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('rhob_sm_out'), 
                     self.subgroup_ca.get_port('ca_rhob_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('phie_sm_out'), 
                     self.subgroup_ca.get_port('ca_phie_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('rhog_sm_out'), 
                     self.subgroup_ca.get_port('ca_rhog_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('kg_sm_out'), 
                     self.subgroup_ca.get_port('ca_kg_sm_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('vp_hs_out'), 
                     self.subgroup_ca.get_port('ca_vp_hs_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('vs_hs_out'), 
                     self.subgroup_ca.get_port('ca_vs_hs_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('phi_hs_out'), 
                     self.subgroup_ca.get_port('ca_phi_hs_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('xlam_out'), 
                     self.subgroup_ca.get_port('ca_xlam_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('xdis_out'), 
                     self.subgroup_ca.get_port('ca_xdis_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('phi_ss_out'), 
                     self.subgroup_ca.get_port('ca_phi_ss_out'))
        connectPorts(self.subgroup_ca,
                     self.tsdmodel_rhob_actor.get_port('phi_sh_out'), 
                     self.subgroup_ca.get_port('ca_phi_sh_out'))

        return

    def tearDown( self ) :
        logger.debug("test execution ends.")
        self.top.wrapup()
        return

    def test_compare_matlab( self ) :
        # don't need to do this with inheritance, but want to explicitly
        # see this function without duplicating the code from
        # TsdNoSubgroup
        TsdNoSubgroup.test_compare_matlab(self)
        
        return

class TsdScriptSubgroup2TestCase( TsdScriptSubgroup2, unittest.TestCase ) :
    pass
    
    
### EOF ######################################################################
