import unittest

from enthought.util.testingx import assert_array_almost_equal

from enthought.logger import logger
from enthought.graph   import CompositeActor, SdfDirector

from tsd_no_subgroup_test_case import TsdNoSubgroup

        
class TsdWithSubgroup1( TsdNoSubgroup ) :

    def setUp(self) :
        logger.debug("-------------------------------------------------------")

        # create the actors that use rock physics
        self._createActors()

        # make connections between ports
        self._makeConnections()

        # install director, only in top composite actor
        self.director = SdfDirector()
        self.director.container = self.top

        logger.debug("test execution of TestTsdWithSubgroup1 begins...")
        
        self.top.preinitialize()
        self.top.initialize()
        return

    def _createActors(self) :
        """ The actors created in this function are the ...Actor class
        instances. These DO implement rock physics functions. Inputs
        are scalars or short sequences as demanded.
        """

        from tsd_actors import TsdDataInActor
        from tsd_actors import TempPressLinearActor
        from tsd_actors import BrinePropertiesActor
        from tsd_actors import TsdSubgroup1Actor
        from tsd_actors import VsWetTrendsActor
        from tsd_actors import BiotCoeffRhobActor
        from tsd_actors import TsdModelRhobActor
        
        # create the actors that use rock physics
        self.top = CompositeActor('composite')

        self.tsd_data_in_actor = TsdDataInActor('tsd_data_in_actor')
        self.tsd_data_in_actor.container = self.top

        self.temp_press_linear_actor = \
            TempPressLinearActor('temp_press_linear_actor')
        self.temp_press_linear_actor.container = self.top

        self.brine_prop_sand_actor = \
            BrinePropertiesActor('brine_prop_sand_actor')
        self.brine_prop_sand_actor.container = self.top

        self.brine_prop_shale_actor = \
            BrinePropertiesActor('brine_prop_shale_actor')
        self.brine_prop_shale_actor.container = self.top

        # build the composite actor for the subgroup
        self.subgroup_ca = \
            TsdSubgroup1Actor('tsd_subgroup_composite_actor')
        self.subgroup_ca.container = self.top

        # build three nodes contained by the subgroup composite actor
        self.vs_wet_trends_actor = \
            VsWetTrendsActor('vs_wet_trends_actor')
        self.vs_wet_trends_actor.container = self.subgroup_ca

        self.biot_coeff_rhob_actor = \
            BiotCoeffRhobActor('biot_coeff_rhob_actor')
        self.biot_coeff_rhob_actor.container = self.subgroup_ca

        self.biot_coeff_rhob_shale_actor = \
            BiotCoeffRhobActor('biot_coeff_rhob_shale_actor')
        self.biot_coeff_rhob_shale_actor.container = self.subgroup_ca

        self.tsdmodel_rhob_actor = \
            TsdModelRhobActor('tsdmodel_rhob_actor')
        self.tsdmodel_rhob_actor.container = self.subgroup_ca

        return

    def _makeConnections(self) :

        # temp_press_linear_actor inputs
        self.top.connect( self.tsd_data_in_actor.get_port('watdth_out'), 
                        self.temp_press_linear_actor.get_port('watdth_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('oftemp_out'), 
                        self.temp_press_linear_actor.get_port('oftemp_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('tgrad_out'), 
                        self.temp_press_linear_actor.get_port('tgrad_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('pgrad_out'), 
                        self.temp_press_linear_actor.get_port('pgrad_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('depth_out'), 
                        self.temp_press_linear_actor.get_port('depth_in'))

        # brine_prop_sand_actor inputs
        rel_sandbrine_tform = \
        self.top.connect( self.temp_press_linear_actor.get_port('tform_out'),
                        self.brine_prop_sand_actor.get_port('tform_in'))
        rel_sandbrine_ppore = \
        self.top.connect( self.temp_press_linear_actor.get_port('ppore_out'), 
                        self.brine_prop_sand_actor.get_port('ppore_in'))
        # have to give explicit name to relation to avoid conflict with the
        # salw_in to brine_prop_shale_actor below
        self.top.connect( self.tsd_data_in_actor.get_port('salw_ss_out'), 
                        self.brine_prop_sand_actor.get_port('salw_in'),
                        'sand_salw_rel')

        # brine_prop_shale_actor inputs
        # tform_in: relation already created
        self.brine_prop_shale_actor.get_port('tform_in').link(rel_sandbrine_tform)
        # ppore_in: relation already created
        self.brine_prop_shale_actor.get_port('ppore_in').link(rel_sandbrine_ppore)
        # have to give explicit name to relation to avoid conflict with the
        # salw_in to brine_prop_sand_actor above
        self.top.connect( self.tsd_data_in_actor.get_port('salw_sh_out'), 
                        self.brine_prop_shale_actor.get_port('salw_in'),
                        'shale_salw_rel')

        # now do the connections to the subgroup composite actor
        self.top.connect( self.tsd_data_in_actor.get_port('vp_ss_out'),
                          self.subgroup_ca.get_port('ca_vp_ss_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('vp_sh_out'),
                          self.subgroup_ca.get_port('ca_vp_sh_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('rhob_ss_out'),
                          self.subgroup_ca.get_port('ca_rhob_ss_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('rhog_ss_out'),
                          self.subgroup_ca.get_port('ca_rhog_ss_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('kg_ss_out'),
                          self.subgroup_ca.get_port('ca_kg_ss_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('gg_ss_out'),
                          self.subgroup_ca.get_port('ca_gg_ss_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('rhob_sh_out'),
                          self.subgroup_ca.get_port('ca_rhob_sh_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('rhog_sh_out'),
                          self.subgroup_ca.get_port('ca_rhog_sh_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('kg_sh_out'),
                          self.subgroup_ca.get_port('ca_kg_sh_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('phit_out'),
                          self.subgroup_ca.get_port('ca_phit_in'))
        self.top.connect( self.tsd_data_in_actor.get_port('xsh_out'),
                          self.subgroup_ca.get_port('ca_xsh_in'))
        self.top.connect( self.brine_prop_sand_actor.get_port('k_wat_out'),
                          self.subgroup_ca.get_port('ca_k_wat_ss_in'))
        self.top.connect( self.brine_prop_sand_actor.get_port('rho_wat_out'),
                          self.subgroup_ca.get_port('ca_rho_wat_ss_in'))
        self.top.connect( self.brine_prop_shale_actor.get_port('k_wat_out'),
                          self.subgroup_ca.get_port('ca_k_wat_sh_in'))
        self.top.connect( self.brine_prop_shale_actor.get_port('rho_wat_out'),
                          self.subgroup_ca.get_port('ca_rho_wat_sh_in'))

        # now make connections from subgroup inputs to enclosed nodes

        # vs_wet_trends_actor inputs
        rel_vp_ss_out = \
            self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_vp_ss_in'), 
                            self.vs_wet_trends_actor.get_port('vp_ss_in'),
                            'vswettrends_vp_ss_rel')
        rel_vp_sh_out = \
            self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_vp_sh_in'), 
                            self.vs_wet_trends_actor.get_port('vp_sh_in'),
                            'vswettrends_vp_sh_rel')

        # biot_coeff_rhob inputs
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_k_wat_ss_in'), 
                        self.biot_coeff_rhob_actor.get_port('k_wat_ss_in'))
        rel_rho_wat_ss_out = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_rho_wat_ss_in'), 
                        self.biot_coeff_rhob_actor.get_port('rho_wat_ss_in'))
        rel_sub_kg_ss = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_kg_ss_in'), 
                        self.biot_coeff_rhob_actor.get_port('kg_ss_in'),
                        'biot_coeff_kg_ss_rel')
        rel_sub_rhog_ss = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_rhog_ss_in'), 
                        self.biot_coeff_rhob_actor.get_port('rhog_ss_in'),
                        'biot_coeff_rhog_ss_rel')
        rel_sub_rhob_ss = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_rhob_ss_in'), 
                        self.biot_coeff_rhob_actor.get_port('rhob_ss_in'),
                        'biot_coeff_rhob_ss_rel')
        # vp_ss_in: relation already created
        self.biot_coeff_rhob_actor.get_port('vp_ss_in').link(rel_vp_ss_out)
        rel_vs_ss_out = \
        self.subgroup_ca.connect( self.vs_wet_trends_actor.get_port('vs_ss_out'), 
                        self.biot_coeff_rhob_actor.get_port('vs_ss_in'))
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_k_wat_sh_in'), 
                        self.biot_coeff_rhob_shale_actor.get_port('k_wat_ss_in'), 'k_wat_ss_in_shale')
        rel_rho_wat_sh_out = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_rho_wat_sh_in'), 
                        self.biot_coeff_rhob_shale_actor.get_port('rho_wat_ss_in'), 'rho_wat_ss_in_shale')
        rel_kg_sh_out = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_kg_sh_in'), 
                        self.biot_coeff_rhob_shale_actor.get_port('kg_ss_in'), 'kg_ss_in_shale')
        rel_rhog_sh_out = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_rhog_sh_in'), 
                        self.biot_coeff_rhob_shale_actor.get_port('rhog_ss_in'), 'rhog_ss_in_shale')
        rel_rhob_sh_out = \
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_rhob_sh_in'), 
                        self.biot_coeff_rhob_shale_actor.get_port('rhob_ss_in'), 'rhob_ss_in_shale')
        # vp_sh_in: relation already created
        self.biot_coeff_rhob_shale_actor.get_port('vp_ss_in').link(rel_vp_sh_out)
        rel_vs_sh_out = \
        self.subgroup_ca.connect( self.vs_wet_trends_actor.get_port('vs_sh_out'), 
                        self.biot_coeff_rhob_shale_actor.get_port('vs_ss_in'), 'vs_ss_in_shale')

        # tsdmodel_rhob_actor inputs
        # vp_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vp_ss_in').link(rel_vp_ss_out)
        # vs_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vs_ss_in').link(rel_vs_ss_out)
        # rhob_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rhob_ss_in').link(rel_sub_rhob_ss)
        # rhoh_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rhog_ss_in').link(rel_sub_rhog_ss)
        # kg_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('kg_ss_in').link(rel_sub_kg_ss)
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_gg_ss_in'), 
                        self.tsdmodel_rhob_actor.get_port('gg_ss_in'))

        # vp_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vp_sh_in').link(rel_vp_sh_out)
        # vs_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('vs_sh_in').link(rel_vs_sh_out)
        # rhob_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rhob_sh_in').link(rel_rhob_sh_out)
        # rhog_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rhog_sh_in').link(rel_rhog_sh_out)
        # kg_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('kg_sh_in').link(rel_kg_sh_out)

        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_phit_in'), 
                        self.tsdmodel_rhob_actor.get_port('phit_in'))
        self.subgroup_ca.connect( self.subgroup_ca.get_port('ca_xsh_in'), 
                        self.tsdmodel_rhob_actor.get_port('xsh_in'))
        # rho_wat_ss_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rho_wat_ss_in').link(rel_rho_wat_ss_out)
        # rho_wat_sh_in: relation already created
        self.tsdmodel_rhob_actor.get_port('rho_wat_sh_in').link(rel_rho_wat_sh_out)

        # connect outputs of internal nodes to composite outputs
        self.subgroup_ca.connect( self.biot_coeff_rhob_actor.get_port('biot_ss_out'), 
                        self.subgroup_ca.get_port('ca_biot_ss_out'))
        self.subgroup_ca.connect( self.biot_coeff_rhob_actor.get_port('biotmax_ss_out'), 
                        self.subgroup_ca.get_port('ca_biotmax_ss_out'))
        self.subgroup_ca.connect( self.biot_coeff_rhob_shale_actor.get_port('biot_ss_out'), 
                        self.subgroup_ca.get_port('ca_biot_sh_out'))
        self.subgroup_ca.connect( self.biot_coeff_rhob_shale_actor.get_port('biotmax_ss_out'), 
                        self.subgroup_ca.get_port('ca_biotmax_sh_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('vp_wet_out'), 
                        self.subgroup_ca.get_port('ca_vp_wet_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('vs_wet_out'), 
                        self.subgroup_ca.get_port('ca_vs_wet_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('rhob_wet_out'), 
                        self.subgroup_ca.get_port('ca_rhob_wet_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('vp_sm_out'), 
                        self.subgroup_ca.get_port('ca_vp_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('vs_sm_out'), 
                        self.subgroup_ca.get_port('ca_vs_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('phi_sm_out'), 
                        self.subgroup_ca.get_port('ca_phi_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('rhob_sm_out'), 
                        self.subgroup_ca.get_port('ca_rhob_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('phie_sm_out'), 
                        self.subgroup_ca.get_port('ca_phie_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('rhog_sm_out'), 
                        self.subgroup_ca.get_port('ca_rhog_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('kg_sm_out'), 
                        self.subgroup_ca.get_port('ca_kg_sm_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('vp_hs_out'), 
                        self.subgroup_ca.get_port('ca_vp_hs_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('vs_hs_out'), 
                        self.subgroup_ca.get_port('ca_vs_hs_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('phi_hs_out'), 
                        self.subgroup_ca.get_port('ca_phi_hs_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('xlam_out'), 
                        self.subgroup_ca.get_port('ca_xlam_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('xdis_out'), 
                        self.subgroup_ca.get_port('ca_xdis_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('phi_ss_out'), 
                        self.subgroup_ca.get_port('ca_phi_ss_out'))
        self.subgroup_ca.connect( self.tsdmodel_rhob_actor.get_port('phi_sh_out'), 
                        self.subgroup_ca.get_port('ca_phi_sh_out'))

        return

    def tearDown( self ) :
        logger.debug("test execution ends.")
        self.top.wrapup()
        return

    def test_compare_matlab( self ) :
        # don't need to do this with inheritance, but want to explicitly
        # see this function without duplicating the code from
        # TestTsdNoSubgroup
        TsdNoSubgroup.test_compare_matlab(self)
        
        return

class TsdWithSubgroup1TestCase( TsdWithSubgroup1, unittest.TestCase ) :
    pass

### EOF ######################################################################
