""" Testing of units conversion during dataflow execution """
import unittest
from os.path import abspath, dirname, join

from enthought.util.numerix import array, arange
from enthought.util.testingx import assert_almost_equal

from enthought.logger import logger

from enthought.units.unit_parser import unit_parser

from enthought.traits.api import Any

from enthought.units.scalar import Scalar

from enthought.graph import SimpleSdfDirector
from enthought.graph import InsufficientInputError
from enthought.graph import CompositeActor
from enthought.graph.vpl.port_factory import PortFactory
from enthought.graph.any_type_value_port import AnyTypeValuePort

#from cp.log.log import Log
#from cp.log.log_index import LogIndex
#from cp.plugin.graph.log_index_array_value_port import LogIndexArrayValuePort
#from cp.plugin.graph.log_array_value_port import LogArrayValuePort
from enthought.resource_type.scalar_float_value_port import ScalarFloatValuePort
from enthought.resource_type.scalar_value_port import ScalarValuePort


class ValueReceiversTestCase( unittest.TestCase ) :

    def setUp(self) :
        try:
            from enthought.graph.function import ActorPrototypeFactory
        except ImportError, e:
            import testoob.testing
            testoob.testing.skip("Import Error: " + str(e))
            
        logger.debug("-------------------------------------------------------")                
         
        self.apf = ActorPrototypeFactory()
        self.filename = join( abspath(dirname(__file__)), "log_array_noop.py")
        self.actor = self.apf.from_filename( self.filename )
        self.filename = join( abspath(dirname(__file__)), "log_array_noop_imperial.py")
        self.actor_imperial = self.apf.from_filename( self.filename )
        self.filename = join( abspath(dirname(__file__)), "scalar_noop.py")
        self.actor_scalar = self.apf.from_filename( self.filename )
        return
        
    def test_units_single_actor_no_conversion( self ) :
        """ test use of no-op function with no units conversion necessary.
        """

        self.test_index = LogIndex(arange(0.0, 10.0, 1.0), units='m', family_name='depth')
        self.test_log = Log( array( [1.0]*10, 'f'), name="vp", units='km/s', index=self.test_index)

        self.do_test_units_single_actor( 1.0, 'km/s', 1.0, 'km/s', 'm', 'm')
        return
        
    def test_units_single_actor_conversion( self ) :
        """ test use of no-op function with units conversion necessary.
        """

        self.test_index = LogIndex(arange(0.0, 10.0, 1.0), units='ft', family_name='depth')
        self.test_log = Log( array( [10000.0]*10, 'f'), name="vp", units='ft/s', index=self.test_index)

        self.do_test_units_single_actor( 10000.0, 'ft/s', 3.048, 'km/s', 'ft', 'm')
        return


    def test_units_actor_chain_conversion( self ) :
        """ test use of no-op function with units conversion necessary.
        """

        self.test_index = LogIndex(arange(0.0, 10.0, 1.0), units='ft', family_name='depth')
        self.test_log = Log( array( [10000.0]*10, 'f'), name="vp", units='ft/s', index=self.test_index)

        self.do_test_units_actor_chain( 10000.0, 'ft/s', 10000.0, 'ft/s', 'ft', 'ft')
        return

    def test_units_scalar_actor_no_conversion(self):
        self.test_scalar = Scalar( 60.0, units='ft', family_name='depth' )
        
        self.do_test_scalar_actor( 60.0, 'ft', 60.0, 'ft')
        return
        
    def test_units_scalar_actor_conversion(self):
        self.test_scalar = Scalar( 100.0, units='m', family_name='depth' )
        
        self.do_test_scalar_actor( 100.0, 'm', 328.08399, 'ft')
        return

    def do_test_scalar_actor(self, expected_value_in, expected_units_in,
                                   expected_value_out, expected_units_out):
                                       
        self.top = CompositeActor('top')
        self.actor_scalar.container = self.top
        
        x_in = self.actor_scalar.get_port('x_in')
        x_in.token = self.test_scalar
        
        self.d = SimpleSdfDirector()
        self.d.container = self.top

        self.top.preinitialize()
        self.top.initialize()
        self.top.iterate(1)
        
        x_out = self.actor_scalar.get_port('x_out')
        self.assert_( x_out.token is not None )
        out_scalar = x_out.token

        self.assertNotEqual( id(self.test_scalar), id(out_scalar) )
        assert_almost_equal( expected_value_in, self.test_scalar.data, decimal=3 )
        assert_almost_equal( expected_value_out, out_scalar.data, decimal=3 )
        self.assertEqual( out_scalar.units, 
                         unit_parser.parse_unit(expected_units_out) )
        self.assertEqual( self.test_scalar.units, 
                         unit_parser.parse_unit(expected_units_in) )
        self.assertEqual( self.test_scalar.family_name, out_scalar.family_name )


        self.top.wrapup()
        return


    def do_test_units_single_actor(self, expected_value_in, expected_units_in,
                                   expected_value_out, expected_units_out,
                                   index_units_in, index_units_out):
                                       
        self.top = CompositeActor('top')
        self.actor.container = self.top
        
        vp_in = self.actor.get_port('vp_in')
        vp_in.token = self.test_log
        
        self.d = SimpleSdfDirector()
        self.d.container = self.top

        self.top.preinitialize()
        self.top.initialize()
        self.top.iterate(1)
        
        vp_out = self.actor.get_port('vp_out')
        self.assert_( vp_out.token is not None )
        out_log = vp_out.token

        self.units_tests_asserts( out_log, expected_value_in, expected_units_in,
                                   expected_value_out, expected_units_out,
                                   index_units_in, index_units_out )
        self.top.wrapup()
        return

    def do_test_units_actor_chain(self, expected_value_in, expected_units_in,
                                   expected_value_out, expected_units_out,
                                   index_units_in, index_units_out):
                                       
        self.top = CompositeActor('top')
        self.actor.container = self.top
        self.actor_imperial.container = self.top
        self.top.connect( self.actor.get_port("vp_out"), self.actor_imperial.get_port("vp_in") )
        
        vp_in = self.actor.get_port('vp_in')
        vp_in.token = self.test_log
        
        self.d = SimpleSdfDirector()
        self.d.container = self.top

        self.top.preinitialize()
        self.top.initialize()
        self.top.iterate(1)
        
        vp_out = self.actor_imperial.get_port('vp_out')
        self.assert_( vp_out.token is not None )
        out_log = vp_out.token

        self.units_tests_asserts( out_log, expected_value_in, expected_units_in,
                                   expected_value_out, expected_units_out,
                                   index_units_in, index_units_out )
        self.top.wrapup()
        return

    def units_tests_asserts(self, out_log, expected_value_in, expected_units_in,
                                   expected_value_out, expected_units_out,
                                   index_units_in, index_units_out):
        

        self.assertNotEqual( id(self.test_log), id(out_log) )
        assert_almost_equal( expected_value_in, self.test_log.data[0] )
        assert_almost_equal( expected_value_out, out_log.data[0] )
        self.assertEqual( out_log.units, 
                         unit_parser.parse_unit(expected_units_out) )
        self.assertEqual( self.test_log.units, 
                         unit_parser.parse_unit(expected_units_in) )
        self.assertEqual( self.test_log.family_name, out_log.family_name )

        self.assertEqual( self.test_index.units,
                         unit_parser.parse_unit( index_units_in) )
        self.assertEqual( out_log.index.units,
                         unit_parser.parse_unit( index_units_out) )
        # LogIndex == is special!
        #self.assertEqual( self.test_index, out_log.index )
        
        #print 'test_index id, units, data', id(self.test_index), self.test_index.units, self.test_index.data
        #print 'out_log.index id, units, data', id(out_log.index), out_log.index.units, out_log.index.data

        return
        
### EOF #######################################################################
