import unittest

from enthought.logger import logger

from enthought.traits.api import Any

from enthought.graph import SimpleSdfDirector, ValueActor
from enthought.graph import InsufficientInputError
from enthought.graph import CompositeActor

from actors import *
        
class SinkActor( ValueActor ):
    """ An actor with an input but no output.  Most recent input in 'value'. """
    
    value = Any
    
    def __init__(self, *args, **kw) :
        super( SinkActor, self).__init__(*args, **kw)
        self._create_ports()
        self.value = None
        return
        
    def fire(self) :
        p = self.get_port("in")
        self.value = p.get()
        #logger.debug( "%s.fire: get() = %s" % (self.name, self.value) )
        return

    def _create_ports(self) :
        self.new_port("in", True )
        return


class SourceActor( ValueActor ):
    """ An actor with attribute defined output using value ports.
    The actor will put the value in the 'value' trait.
    """
    
    value = Any

    def __init__(self, *args, **kw) :
        super( SourceActor, self).__init__(*args, **kw)
        self._create_ports()
        self.value = "SourceActor default"
        return
        
    def fire(self) :
        p = self.get_port("out")
        self.value = p.put( self.value )
        #logger.debug( "%s.fire: get() = %s" % (self.name, self.value) )
        return

    def _create_ports(self) :
        self.new_port("out", False )
        return

    
class ValueReceiversTestCase( unittest.TestCase ) :

    def setUp(self) :
        logger.debug("-------------------------------------------------------")
        return
        
    def test_single_actor( self ) :
        """ test use of default value for unbound input
        [sink_actor]
        """
        top = CompositeActor('top')
        sink_actor = SinkActor("sink_actor")
        sink_actor.get_port('in').token = "default value"
        sink_actor.container = top
        d = SimpleSdfDirector()
        d.container = top

        self.assertEqual( None, sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.preinitialize()
        top.initialize()
        top.iterate(1)
        self.assertEqual( "default value", sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.wrapup()
        return
        
    def test_source_to_sink( self ) :
        """ test source to sink
        [ source, sink ]
        """
        top = CompositeActor('top')
        source_actor = SourceActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = top
        
        sink_actor = SinkActor("sink_actor")
        sink_actor.get_port('in').token = "default value"
        sink_actor.container = top

        top.connect( source_actor.get_port("out"), sink_actor.get_port("in") )
        
        d = SimpleSdfDirector()
        d.container = top

        self.assertEqual( None, sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.preinitialize()
        top.initialize()
        top.iterate(1)
        self.assertEqual( "source actor value", sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.wrapup()
        return
        
    def test_source_two_sink( self ) :
        """ test source to two sinks connected to the source via two relations.
        [ source, sink1, sink2 ]
        """
        top = CompositeActor('top')
        source_actor = SourceActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = top
        
        sink1_actor = SinkActor("sink1_actor")
        sink1_actor.get_port('in').token = "default value sink1"
        sink1_actor.container = top
        
        sink2_actor = SinkActor("sink2_actor")
        sink2_actor.get_port('in').token = "default value sink2"
        sink2_actor.container = top

        top.connect( source_actor.get_port("out"), sink1_actor.get_port("in"),
                    "source to sink1" )
        top.connect( source_actor.get_port("out"), sink2_actor.get_port("in"),
                    "source to sink2" )
        
        d = SimpleSdfDirector()
        d.container = top

        top.preinitialize()
        top.initialize()
        top.iterate(1)
        self.assertEqual( "source actor value", sink1_actor.value, \
            "sink1_actor has unexpected value: %s" % sink1_actor.value )    
        self.assertEqual( "source actor value", sink2_actor.value, \
            "sink2_actor has unexpected value: %s" % sink2_actor.value )    
        top.wrapup()
        return
        
    
    def test_source_two_sink_fanout( self ) :
        """ test source to two sinks with a single relation out of the source
        [ source, sink1, sink2 ]
        """
        top = CompositeActor('top')
        source_actor = SourceActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = top
        
        sink1_actor = SinkActor("sink1_actor")
        sink1_actor.get_port('in').token = "default value sink1"
        sink1_actor.container = top
        
        sink2_actor = SinkActor("sink2_actor")
        sink2_actor.get_port('in').token = "default value sink2"
        sink2_actor.container = top

        relation = top.connect( source_actor.get_port("out"), 
                                sink1_actor.get_port("in"),
                                "sourct to sinks" )
        sink2_actor.get_port("in").link( relation )
        
        d = SimpleSdfDirector()
        d.container = top

        top.preinitialize()
        top.initialize()
        top.iterate(1)
        self.assertEqual( "source actor value", sink1_actor.value, \
            "sink1_actor has unexpected value: %s" % sink1_actor.value )    
        self.assertEqual( "source actor value", sink2_actor.value, \
            "sink2_actor has unexpected value: %s" % sink2_actor.value )    
        top.wrapup()
        return

    def test_source_composite_sink( self ) :
        """ test crossing through composite boundary for input
        [source [ sink ] ]
        """
        top = CompositeActor('top')

        sub = CompositeActor('sub')
        sub.new_port("in", True)
        sub.container = top
        
        sink_actor = SinkActor("sink_actor")
        sink_actor.get_port('in').token = "default value"
        sink_actor.container = sub

        source_actor = SourceActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = top
        
        top.connect( source_actor.get_port("out"), sub.get_port("in") )
        sub.connect( sub.get_port("in"), sink_actor.get_port("in") )
        
        d = SimpleSdfDirector()
        d.container = top

        self.assertEqual( None, sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.preinitialize()
        top.initialize()
        top.iterate(1)
        self.assertEqual( "source actor value", sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.wrapup()
        return
        

    def test_composite_source_composite_sink( self ) :
        """ test crossing up through composite input and down through composite
        output.
        [ [ source ] [ sink ] ]
        """
        top = CompositeActor('top')

        source_sub = CompositeActor('source_sub')
        source_sub.new_port("out", False)
        source_sub.container = top

        source_actor = SourceActor("source_actor")
        source_actor.value = "source actor value"
        source_actor.container = source_sub
        
        sink_sub = CompositeActor('sink_sub')
        sink_sub.new_port("in", True)
        sink_sub.container = top
        
        sink_actor = SinkActor("sink_actor")
        sink_actor.get_port('in').token = "default value"
        sink_actor.container = sink_sub
        
        top.connect( source_sub.get_port("out"), sink_sub.get_port("in"))
        source_sub.connect( source_actor.get_port("out"), source_sub.get_port("out") )
        sink_sub.connect( sink_sub.get_port("in"), sink_actor.get_port("in") )
        
        d = SimpleSdfDirector()
        d.container = top

        self.assertEqual( None, sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.preinitialize()
        top.initialize()
        top.iterate(1)
        self.assertEqual( "source actor value", sink_actor.value, \
            "sink_actor has unexpected value: %s" % sink_actor.value )    
        top.wrapup()
        return


### EOF ######################################################################
