"""An IOPort where get can return a token even when not connected and
put retains the most recent output.
"""

import sys, traceback

from enthought.traits.api import Any, Property

from ioport import IOPort

class ValuePort( IOPort ):
    """An IOPort where get can return a token even when not connected and
    put retains the most recent output.  Must be used with "value" receivers
    and SimpleSdfDirector as put does not propagate token to receivers.
    
    When the port is an input port, the token in the token attribute is the
    default return value from get, otherwise the token is retrieved from the
    ancestor actor.  The SimpleSdfDirector chooses an appropriate receiver
    based on the connectivity of this port. If connected, a ValuePortReceiver
    is used otherwise a DefalutValueReceiver is used.
    
    When the port is an output port, the most recent output token is retained.
    Dependent actors are required to use ValuePortReceiver which retrieves
    the value from this ValuePort.
    
    In the output case, the port can be primed by setting value prior to 
    graph execution.
    
    ValuePort is only valid on non-composite actors.
    """
    
    # The token - most recent output or default input
    token = Property
    
    # Protected reference to the actual token object.
    _token = Any
    
    def __init__(self, *args, **kw):
        self._token = None
        super( ValuePort, self ).__init__( *args, **kw )
        
        self.reset_token()
        return
        
    def _set_token( self, token ):
        """ token property setter """
        self._token = token
        
        return
        
    def _get_token( self ):
        """ token property getter """
        return self._token
        
    #### IOPort methods #######################################################
    
    def get( self, n=1 ):
        """ Return token from local receiver. """
        return self._local_receiver.get(n)


    def has_tokens( self, n=1 ):
        """ Return true if the token attribute is not None or super.has_tokens.
        """
        return self._local_receiver.has_tokens(n)
        
    def put( self, token ):
        """ Save the token in the token attribute.
        """
        self.token = token
        return
        
    def has_room( self, n=1 ) :
        """ ValuePort can always receive, but only holds the last value.
        """
        return True

    #### ValuePort methods ####################################################
    
    def source_port(self) :
        """ When used with SimpleSdfDirector, this method is used to retrieve
        the upstream output port that provides data for this input port.
        Assumes that the upstream port is also a ValuePort and only one
        relation per port.
        """
        relation = self.relations[0]
        port = relation.linked_source_ports( exclude=self )[0]
        
        # walk through composite actor edges
        while not port.is_opaque:
            
            # if composite port is an input direction is from inside to outside
            if port.is_input:
                relation = port.relations[0]
            
            # else direction is from outside to inside
            else:
                relation = port.inside_relations[0]
                
            port = relation.linked_source_ports( exclude=port )[0]
        
        return port

    def reset_token(self):
        """ Reset token to the default value. """
        self.token = None
        return
        
#### EOF ######################################################################
    
