import sys
import os
import re

from enthought.graph.vpl.graph_actor_folder import GraphActorFolder, GraphActor
from enthought.graph.vpl.converge_actor import ConvergeActor
from enthought.graph.composite_actor import CompositeActor
from enthought.envisage.project.project_plugin import ProjectPlugin
from enthought.traits.api import HasTraits, Event, Dict
from enthought.logger import logger

from enthought.graph.function import ActorPrototypeFactory

from chip_change_event import ChipChangeEvent
from chip_validator import ChipValidator
from graph_canvas import GraphCanvas

class _ChipManager(HasTraits):
    """ Allow ability to add and remove actors from actor_classes. 
        
        Added actors will show up in the UI.
        
        Singleton, use ChipManager module function below to get instance.
    """
    
    # singleton instance
    instance = None
    
    # Dictionary of actor classes
    actor_classes = Dict
    
    # Fired when a user defined script has been converted to a chip and added.
    chip_changed = Event(ChipChangeEvent)
    
    def __init__(self):
            
        return
        
    ###########################################################################
    # Public interface.
    ###########################################################################
    
    def lookup(self, id):
        """ lookup the ConvergeActor associated with the id.
        """
        
        actor = None
        try:
            actor = self.actor_classes[id]
        except:
            # Unable to find the actor, log the exception and return None
            logger.warning('Unable to find chip for id = ' + id)
           
            
        return actor
        
    def lookup_with_path(self, path):
        """ lookup the ConvergeActor associated with the given absolute path.
        """
        
        module_name = self._find_module_from_absolute_path(path)
        
        actor = None
        if module_name is not None:
            actor = self.lookup(module_name)
            
        return actor
        
    def remove(self, id):
        """ Remove the ConvergeActor with the given id.
        """
        
        try:
            del self.actor_classes[id]
        except:
            # Ignore the exception id doesn't exist
            pass
            
        return
        
    def remove_with_path(self, path):
        """ Remove the ConvergeActor associated with the given absolute path.
        """
        
        module_name = self._find_module_from_absolute_path(path)
        
        if module_name is not None:
            self.remove(module_name)
            
        return
          
    def generate(self, path):
        """ Import the actor at the given path.  Actor should register itself with the ChipManager.
        """
                
        self._import(path)
        
        return
        
    def register(self, function, xml_metadata):
        """ Register the given function with the chip manager
        """
                
        factory = ActorPrototypeFactory()
        actor = factory.from_function(xml_metadata, function)
        
        if actor is not None:
            # import the module this chip belongs to in order to get its file name.
            imported_module = __import__(function.__module__, globals(), locals(), [''])
                        
            file_name = imported_module.__file__
            
            if file_name.endswith('.pyc'):
                # make it reference the .py
                file_name = file_name[:-1]   
                
            # set the file name
            actor.file_name = file_name
            self._add(function.__module__, actor)
            self.fire_chip_changed(actor)
    
        return
        
    def validate(self, id, *actors):
        """ Validate that the actor matches the current prototype.
        """
        
        chipValidator = ChipValidator()
        
        for actor in actors:
            if isinstance(actor, ConvergeActor):
                if id is None:
                    chipValidator.validate(actor)
                elif actor.id == id:
                    chipValidator.validate(actor)
            elif isinstance(actor, CompositeActor):
                # recursive call to traverse through actors contained in this composite.
                self.validate(id, *actor.nodes)
    
        return   
             
    def load_chips_from_path(self, path):
        """ Load chips from the given directory and all subdirectories 
        """
        
        logger.debug('Load actions from path - ' + path)
        
        # try to generate an individual chip
        if path.endswith('.py'):
            path_tuple = os.path.split(path)
            head = path_tuple[0]
            tail = path_tuple[1]
            
            absolute_path = self._find_abs_path(head)
            absolute_file_path = absolute_path + '/' + tail
            absolute_file_path = os.path.normpath(absolute_file_path)
            
            try:
                self._import(absolute_file_path)
            except:
                logger.exception('Error generating chip: ' + absolute_file_path)
        # try to generate all chips under the directory
        else:
            absolute_path = self._find_abs_path(path)
            
            if absolute_path is not None:
                # walk the subdirectories
                for root, dirs, files in os.walk(absolute_path):
                    for name in files:
                        if name.endswith('.py') and name != '__init__.py':
                            absolute_file_path = root + '/' + name
                            absolute_file_path = os.path.normpath(absolute_file_path)
                            
                            try:
                                self._import(absolute_file_path)
                            except:
                                logger.exception('Error generating chip: ' + absolute_file_path)
                                
                    if '.svn' in dirs:
                        # don't visit svn directories
                        dirs.remove('.svn')  
                    if 'tests' in dirs:
                        # don't visit test directories
                        dirs.remove('tests') 
                
        return         
     
    def load_user_defined_chips(self):
        """ Convert all user defined scripts to chips that were previously converted successfully. 
        
            This method is intended to be run on startup to load user defined chips into the chips tab.
        """
        
        self.load_chips_from_path('my_actions')
                
        return 
        
    def validate_canvas(self):
        """ Validate previously persisted canvases.  Add change listeners to them.
        """
        
        workspace = ProjectPlugin.instance.workspace
        
        bindings = workspace.find_bindings(self._canvas_visitor)
        
        for binding in bindings:
            obj = binding.obj
            
            self.validate(None, obj.data)
            
            self.add_listener(obj)
            
        return
               
    def fire_chip_changed(self, chip):
        """ Fires the chip changed event. """

        self.chip_changed = ChipChangeEvent(id = chip.id)
        
        return
        
    def add_listener(self, listener):
        """ Add a listener for chip events.
        """
          
        self.on_trait_change(listener.on_chip_changed, 'chip_changed')
        
        return
        
    def remove_listener(self, listener):
        """ Remove a listener for chip events.
        """
        
        self.on_trait_change(listener.on_chip_changed, 'chip_changed', remove=True)
        
        return
        
    ###########################################################################
    # Private interface.
    ###########################################################################
    
    def _add(self, module, instance = None):
        """ Add an actor to the actor_classes. 
            
            This method will overwrite any actors already existing for the given module name.  
         """
        
        # Need to add an actor.
        if instance is not None:    
            actor = GraphActor(instance = instance)
              
            self.actor_classes[module] = actor
                     
            # Set the id on the instance         
            instance.id = module
                             
        return True
                                      
    def _canvas_visitor(self, bindings, attributes):
        """ Return bindings that are instances of GraphCanvas.
        """
        
        matchedBindings = []
        
        for binding in bindings:
            obj = binding.obj
            
            if isinstance(obj, GraphCanvas): 
                matchedBindings.append(binding)
        
        return matchedBindings
      
    def _find_abs_path(self, relative_path):
        """ Find the absolute path for the given relative path.
        """
        
        absolute_path = None
        
        module_name = relative_path.replace('/', '.')
        module_name = module_name.replace('\\', '.')
              
        try:
            # find the module for the given path.
            module = __import__(module_name, globals(), locals(), [''])
             
            # remove init.pyc from the end
            temp_path = os.path.dirname(module.__file__)
            
            # in case this is relative path somehow.
            absolute_path = os.path.abspath(temp_path)
        except:
            logger.exception('Unable to find absolute path for ' + relative_path)
        
        return absolute_path
        
    def _find_module_from_absolute_path(self, absolute_path):
        """ find a module name from the absolute path to file
        """

        working_path = os.path.abspath(os.path.splitext(absolute_path)[0])
        
        found_path = None
        for system_path in sys.path:
            if working_path.startswith(system_path):
                found_path = system_path
                break
                
        module_name = None
        if found_path is not None:
            module_path = working_path[len(found_path) + 1:]
            module_name = module_path.replace('/', '.')
            module_name = module_name.replace('\\', '.')
            
        if module_name is None:
            logger.info('Could not find module for path - ' + absolute_path)
            
        return module_name
        
    def _import(self, path):
        """ Import the actor class for the function listed at the given path.
        
            The imported chip should call register to register itself with the ChipManager.
        """
                
        module_name = self._find_module_from_absolute_path(path)
        
        if module_name is not None:
            
            module = None
            try:
                module = sys.modules[module_name]
            except:
                # Ignore the exception, the lookup was unable to find the module
                pass
            
            if module is None:
                imported_module = __import__(module_name, globals(), locals(), [''])
            else:
                reload(module)
        
        return
        
def ChipManager():
    """ Ensure a singleton instance of the _ChipManager class.
        
        Call this function to get the instance.
    """
    
    if _ChipManager.instance == None:
        _ChipManager.instance = _ChipManager()
    
    return _ChipManager.instance
        
                
#### EOF ######################################################################                
