from enthought.traits.api import HasTraits, Dict, List

from enthought.graph.vpl.converge_actor import bundle_suffix_map

class ChipValidator(HasTraits):
    """ Validate that a specific chip is up to date with respect to its prototype.
    """
    
    
    # track ports that have been added.
    _added_ports = Dict
    
    # track ports that have been removed.
    _removed_ports = Dict
    
    # all original bundle ports.
    _original_bundle_ports = List

    ###########################################################################
    # Public interface.
    ###########################################################################

    def validate(self, actor):
        """ Make sure the actor is up to date with respect to its prototype.
        """
 
        from chip_manager import ChipManager        
        cm = ChipManager()
        graph_actor = cm.lookup(actor.id)
        
        if graph_actor is not None:
            prototype = graph_actor.instance
            
            actor.function = prototype.function
            actor.description = prototype.description
            actor.help = prototype.help
            actor.history = prototype.history
            actor.is_hidden = prototype.is_hidden
            actor.file_name = prototype.file_name

            self._initialize_port_lists(actor.instance_input_bundles, actor.instance_output_bundles)
            
            self._fix_ports(actor, prototype.ports, actor.ports)
            
            self._fix_io_ports(prototype.input_ports, actor.input_ports)
            self._fix_io_ports(prototype.output_ports, actor.output_ports)
            
            self._fix_unbound_ports(actor.input_ports, actor.unbound_input_ports)
            
            self._fix_bundles(prototype.instance_input_bundles, actor.instance_input_bundles)
            self._fix_bundles(prototype.instance_output_bundles, actor.instance_output_bundles)
            
            # Set the changed event on the actor.  The main expectation is that the
            # canvas will will listen for this and update accordingly.
            actor.changed = actor
                      
        return
        
    ###########################################################################
    # Private interface.
    ###########################################################################
                
    def _fix_ports(self, actor, prototypePorts, ports):
        """ Ensure ports have the same entries as prototype ports.

            Add or delete ports as necessary to make sure ports has the same number of ports
            and ports of the same name as prototypePorts.
        """
            
        # remove ports that no longer exist in prototype     
        portsCopy = ports[:]
        for port in portsCopy:
            prototypePort = self._find_object_by_name(prototypePorts, port.name, False)
            
            if prototypePort is None:
                self._unlink(port)
                ports.remove(port)
                self._removed_ports[port.name] = port
                                           
        # add ports that exist in prototype, but not in the instance object.
        for prototypePort in prototypePorts:
            port = self._find_object_by_name(ports, prototypePort.name, False)
            
            if port is None:
                clonePort = prototypePort.clone_traits(traits = 'all', copy = 'deep')
                
                # Note: setting the container appends this port to the ports list
                clonePort.container = actor
                # ports.append(clonePort)
                self._added_ports[clonePort.name] = clonePort
        
        # reorder ports in correct sequence.
        for index in range(len(prototypePorts)):
            prototypePort = prototypePorts[index]
            
            portIndex = self._find_index_by_name(ports, prototypePort.name, False)
            
            if index != portIndex:
                port = ports[portIndex]
                
                # remove the port and insert it at the correct index.
                ports.remove(port)
                ports.insert(index, port)
   
        return
        
    def _fix_bundles(self, prototypeBundles, bundles):
        """ Ensure bundles have the same entries as prototypeBundles.

            Add or delete bundles as necessary to make sure bundles has the same number of bundles
            and bundles of the same name as prototypeBundles.
        """

        # remove bundles that no longer exist in prototype
        bundlesCopy = bundles[:]
        for bundle in bundlesCopy:
            prototypeBundle = self._find_object_by_name(prototypeBundles, bundle.original_name, True)
            
            if prototypeBundle is None:
                bundles.remove(bundle)
                      
        # add bundles that exist in prototype, but not in the instance object.      
        for prototypeBundle in prototypeBundles:
            bundle = self._find_object_by_name(bundles, prototypeBundle.original_name, True)
            
            if bundle is None:
                cloneBundle = prototypeBundle.clone_traits(traits = 'all', copy = 'deep')
                
                # The ports will get filled in by the when we fix the ports for the bundle below..
                # FIXME: May want to find a different way to do this, so that we don't have to clone the ports
                # FIXME: only to make the list empty.
                cloneBundle.ports = []
                bundles.append(cloneBundle) 

        # reorder bundles in correct sequence and fill in values that may have changed.
        for index in range(len(prototypeBundles)):
            prototypeBundle = prototypeBundles[index]
            
            bundleIndex = self._find_index_by_name(bundles, prototypeBundle.original_name, True)
            
            bundle = bundles[index]
            
            if index != bundleIndex:
                # remove the bundle and insert it at the correct index.
                bundles.remove(bundle)
                bundles.insert(index, bundle)
                
            bundle.name = prototypeBundle.name
            
        # fix the ports contained in each of the bundles.
        for index in range(len(prototypeBundles)):
            prototypeBundlePorts = prototypeBundles[index].ports
            bundlePorts = bundles[index].ports
            
            self._fix_bundle_ports(prototypeBundlePorts, bundlePorts)
                        
        return 
        
    def _fix_bundle_ports(self, prototypePorts, ports):
        """ Ensure bundle ports have the same entries as prototype ports.

            Add or delete ports as necessary to make sure ports has the same number of ports
            and ports of the same name as prototypePorts.
        """
     
        # remove ports that no longer exist in prototype
        portsCopy = ports[:]
        for port in portsCopy:
            prototypePort = self._find_object_by_name(prototypePorts, port.original_name, True)
            
            if prototypePort is None:
                ports.remove(port)
                                         
        # add ports that exist in prototype, but not in the instance object.
        for prototypePort in prototypePorts:
            
            port = self._find_object_by_name(ports, prototypePort.original_name, True)
            
            if port is None:
                existingPort = self._find_object_by_name(self._original_bundle_ports, prototypePort.original_name, True)
                
                if existingPort is not None:
                    ports.append(existingPort)
                else:
                    clonePort = prototypePort.clone_traits(traits = 'all', copy = 'deep')
                    ports.append(clonePort)
        
        # reorder ports in correct sequence and fill in values that may have changed.
        for index in range(len(prototypePorts)):
            prototypePort = prototypePorts[index]
            
            portIndex = self._find_index_by_name(ports, prototypePort.original_name, True)
            
            port = ports[portIndex]
            
            if index != portIndex:
                # remove the port and insert it at the correct index.
                ports.remove(port)
                ports.insert(index, port)
            
            port.basename = prototypePort.basename
            port.modifier = prototypePort.modifier
            port.bundle = prototypePort.bundle
            port.name = prototypePort.name
            port.units = prototypePort.units
            port.family_name = prototypePort.family_name
            port.units_system_name = prototypePort.units_system_name
            port.description = prototypePort.description
            port.external_type = prototypePort.external_type
            port.internal_type = prototypePort.internal_type
            port.is_index = prototypePort.is_index
            port.is_optional = prototypePort.is_optional
            port.arg_name = prototypePort.arg_name
            port.index = prototypePort.index
            port.visible = prototypePort.visible
   
        return
        
    def _fix_io_ports(self, prototypePorts, ports):
        """ Ensure input/output ports have the same entries as prototype ports.
        
            Add or delete ports as necessary to make sure ports has the same number of ports
            and ports of the same name as prototypePorts.
        """
        
        # remove ports that no longer exist in prototype
        portsCopy = ports[:]
        for port in portsCopy:
            if self._removed_ports.has_key(port.name):
                ports.remove(port)
                    
        # add ports that exist in prototype, but not in the instance object.
        for port in prototypePorts:
            if self._added_ports.has_key(port.name):
                portToAdd = self._added_ports[port.name]
                ports.append(portToAdd)
                
        # reorder ports in correct sequence.
        for index in range(len(prototypePorts)):
            prototypePort = prototypePorts[index]
            
            portIndex = self._find_index_by_name(ports, prototypePort.name, False)
            
            if index != portIndex:
                port = ports[portIndex]
                
                # remove the port and insert it at the correct index.
                ports.remove(port)
                ports.insert(index, port)
                
        return
        
    def _fix_unbound_ports(self, inputPorts, unboundPorts):
        """ Ensure unbound are specified correctly.
        """
        
        # remove ports
        unboundPortsCopy = unboundPorts[:]
        for port in unboundPortsCopy:
            if self._removed_ports.has_key(port.name):
                unboundPorts.remove(port)
       
        # add ports
        for key in self._added_ports.keys():
            port = self._find_object_by_name(inputPorts, key, False)
            if port is not None:
                unboundPorts.append(self._added_ports[key])
             
        return
        
    def _unlink(self, port):
        """ Remove the specified ports link to another graph object.
        """
        
        for relation in port.relations:
            connectedPorts = relation.linked_ports(exclude=port)
            
            for connectedPort in connectedPorts:
                connectedPort.unlink(relation)
                
                if len(connectedPort.relations) == 0:
                    self._set_unbound_input_port(connectedPort)
                
            port.unlink(relation)
            
        if len(port.relations) == 0:
            self._set_unbound_input_port(port)
        
        return
        
    def _set_unbound_input_port(self, port):
        """ append this port to the list of unbound input ports if it is in fact an input port.
        """
        
        for inputPort in port.container.input_ports:
            if port is inputPort:
                port.container.unbound_input_ports.append(port)
                break
                
        return
                 
    def _find_object_by_name(self, objects, name, originalName = False):
        """ Find the object with the specified name in the list.
        """
        
        for obj in objects:
            
            if originalName:
                objName = obj.original_name
            else:
                objName = obj.name

            if objName == name:
                return obj
            
        return None
        
    def _find_index_by_name(self, objects, name, originalName = False):
        """ Find the index of the object with the specified name in the list.
        """
        
        for index in range(len(objects)):
            obj = objects[index]
            
            if originalName:
                objName = obj.original_name
            else:
                objName = obj.name
                
            if objName == name:
                return index
            
        return -1
        
    def _initialize_port_lists(self, inputBundles, outputBundles):
        """ Initialize class instance lists and dictionaries.
        
            Put all original bundle ports into list.
        """
        
        self._added_ports = {}
        self._removed_ports = {}
        self._original_bundle_ports = []
        
        for bundle in inputBundles:
            self._original_bundle_ports.extend(bundle.ports)
            
        for bundle in outputBundles:
            self._original_bundle_ports.extend(bundle.ports)
            
        return
    