#-------------------------------------------------------------------------------
#
#  Defines the ConvergeActor class.
#
#  Written by: David C. Morrill
#
#  Date: 03/31/2005
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from sys \
    import maxint

from enthought.util.numerix \
    import array

from enthought.envisage import get_application

from enthought.naming.api \
    import Binding

from enthought.graph \
    import CompositeValueActor, Port, ValueActor, ValuePort

from enthought.traits.api \
    import HasTraits, HasStrictTraits, HasPrivateTraits, Str, Int, Float, List, \
           Tuple, Callable, ReadOnly, Undefined, Regex, Property, TraitError, \
           true, false, Dict, Any, CInt, Enum, Bool, Delegate, Event, Instance

from enthought.traits.ui.api \
    import View, Group, Item, TableEditor, EnumEditor, CheckListEditor, \
           DropEditor, ButtonEditor, Handler, TextEditor

from enthought.traits.ui.table_column \
    import ObjectColumn

from enthought.traits.ui.table_filter \
    import TableFilter, EvalFilterTemplate, RuleFilterTemplate, \
           MenuFilterTemplate

from enthought.traits.ui.menu \
    import Menu, Action, Separator

from enthought.traits.ui.wx.constants \
    import DropColor, EditableColor

from enthought.units.unit_parser \
    import unit_parser

from enthought.units.unit_traits \
    import unit_family_trait

from enthought.units.unit_manager \
    import unit_manager

from enthought.units.quantity \
    import Quantity

from enthought.units.scalar \
    import Scalar

from cp.log \
    import Log, LogSuite, LogIndex

from cp.log.naming_log \
    import NamingIndex

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

noise_words = ( '_pert', '_reg', '_' )

suffixes    = ( '_ss', '_sh', '_wat', '_hs', '_sm', '_bh', '_watsh', '_f',
                '_sub', '_wet', '_lith', '_nom', 'phit', 'xsh', 'xlam', 'xdis'
)

bundle_suffix_map  = {
     'ss':     'sand',
     'sh':     'shale',
     'wat':    'water',
     'hs':     'hard shale',
     'sm':     'sandy member',
     'bh':     'borehole',
     'watsh':  'shale water',
     'f':      'fluid',
     'sub':    'mixed',
     'wet':    'wet',
     'lith':   'lith',
     'nom':    'nom',
     'phit':   'total',
     'xsh':    'total',
     'xlam':   'total',
     'xdis':   'total',
     'ppore':  'env',
     'tform':  'env',
     'repeat': 'repeat',
}

# Default value for 'pvalue' trait of a ConvergePort:
pvalue_default = " <Drag/drop a log here or click 'Create Log...'>"

#-------------------------------------------------------------------------------
#  'ConvergePort' class:
#-------------------------------------------------------------------------------

class ConvergePort ( HasPrivateTraits ):
    """ ConvergePorts hold the metadata that describes a graph port
    for a dataflow actor.  It is not an enthought.graph.Port.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Basename: e.g. vp, k, rhob.
    basename = Str

    # Name modifier: e.g. 'wat' in k_wat_ss.
    modifier = Str

    # Bundle name: e.g. 'ss' in k_wat_ss.
    bundle = Str

    # Full name of the port: e.g. 'k_wat_sh'.
    name = Str

    # Name used for automatic matching:
    # FIXME: use of match_name should be refactored to use basename, modifier...
    match_name = Property

    # Original full name of the port:
    original_name = ReadOnly

    # Actor this port is part of:
    actor = Any

    # Units: e.g. km/s.
    units = Str

    # Units family name: e.g. pvelocity.
    family_name = unit_family_trait

    # Units system name: e.g. KGS
    units_system_name = Str('KGS')

    # Short description of argument purpose.
    description = Str

    # Type of the value expected (input) or produced (output) by the actor port.
    external_type = Str

    # Type of the value expected (input) or produced (input) by the function.
    internal_type = Str

    # Is the argument an index?
    # FIXME: Probably don't need this: external_type == LogIndex
    is_index = Bool()

    # Is the argument optional?
    is_optional = Bool()

    # Name of the function input argument
    arg_name = Str

    # Argument/Result index:
    index = Int

    # Is the port visible in the user interface:
    visible = true

    # Default value for this port
    # FIXME: this could be any kind of object described by the metadata
    # FIXME: specifically it could be a Log
    value = Any( 1.0 )

#-- Traits specifically related to input/output table display ------------------

    # Canvas this port's chip is part of:
    canvas = Instance( 'enthought.graph.vpl.graph_canvas.GraphCanvas',
                       copy = 'ref' )

    # Is port connected? (as boolean)
    is_connected = Property

    # Is the port value a Log?
    is_log = Property

    # Is port connected? (as string)
    connected = Property

    # User provided 'notes' on what the 'data_source' represents:
    guidance = Str

    # The category of the chip this port is part of:
    component_category = Delegate( 'actor', 'category' )

    # The description of the component this port is part of:
    component_description = Delegate( 'actor', 'description' )

    # Name of the component this port is part of:
    component_name = Delegate( 'actor', 'name' )

    # Value assigned to the input port:
    pvalue = Property

    # User specified units associated with the value assigned to the input port:
    pvalue_units = Property

    # Event fired when user wishes to plot the log bound to the input port:
    plot = Event

    # Event fired when user wishes to create log bound to the input port:
    create_log = Event

    # The actor port associated with this port:
    actor_port = Property

#-- HasTraits overrides --------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Returns the list of all trait names, including implicitly defined
    #  traits:
    #
    #  fixme: This method is a hack, because it crashes if we try to copy an
    #  object and the 'component_name' delegated trait is copied.
    #---------------------------------------------------------------------------

    def all_trait_names ( self ):
        """ Returns the list of all trait names, including implicitly defined
            traits.
        """
        result = self.__class_traits__.keys()
        for name in [ 'component_category', 'component_name',
                      'component_description' ]:
            result.remove( name )
        return result

#-- Property implementations ---------------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles the 'name' trait being changed:
    #---------------------------------------------------------------------------

    def _name_changed ( self ):
        if self.original_name is Undefined:
            self.original_name = self.name
        self._match_name = None

    #---------------------------------------------------------------------------
    #  Implementation of the 'match_name' property:
    #---------------------------------------------------------------------------

    def _get_match_name ( self ):
        # FIXME: this method is not quite right 'k_wat_sh' ends with '_sh'
        # but name.find('_') returns the index of the '_' before 'wat' and
        # thus the method truncates '_wat_sh'.
        name = getattr( self, '_match_name', None )
        if name is None:
            name = remove_noise( self.name )

            for suffix in suffixes:
                if ends_with( name, suffix ):
                    if name != suffix:
                        col = name.find( '_' )
                        if col > 0:
                            name = name[ : col ]
                        else:
                            name = name[ : -len( suffix ) ]
                    break
            self._match_name = name
        return name

    #---------------------------------------------------------------------------
    #  Implementation of the 'is_connected' property:
    #---------------------------------------------------------------------------

    def _get_is_connected ( self ):
        from graph_helper import is_input_port_connected, find_actor_port

        return is_input_port_connected( find_actor_port( self, self.actor ) )

    #---------------------------------------------------------------------------
    #  Implementation of the 'is_log' property:
    #---------------------------------------------------------------------------

    def _get_is_log ( self ):
        return self.external_type in [ 'cp.log.log.Log', 'cp.log.log_index.LogIndex']

    #---------------------------------------------------------------------------
    #  Implementation of the 'connected' property:
    #---------------------------------------------------------------------------

    def _get_connected ( self ):
        if self.is_connected:
            return '==>'
        return ''

    #---------------------------------------------------------------------------
    #  Implementation of the 'pvalue' property:
    #---------------------------------------------------------------------------

    def _get_pvalue ( self ):
        if self.is_connected:
            return ''

        if self._pvalue is not None:
            return self._pvalue

        if self.is_log:
            return pvalue_default

        try:
            return self.actor_port.token.value.data
        except:
            return ''

    def _set_pvalue ( self, value ):
        type_value  = type( value )
        token_value = value
        if (value is not None) and (type_value is not str):
            if type_value is float:
                token_value = Scalar( value, self.units,
                                      family_name = self.family_name )
            elif ((not isinstance( value, Binding ))      or
                  (not isinstance( value.obj, LogIndex )) or
                  (not unit_manager.are_compatible_families( self.family_name,
                                                    value.obj.family_name ) ) ):
                return
            else:
                token_value = value = value.namespace_name

        old, self._pvalue = self._pvalue, value
        if value is None:
            self.actor_port.reset_token()
        elif value != pvalue_default:
            self.actor_port.token.value = token_value
        if old != value:
            self.trait_property_changed( 'pvalue', old, value )

    def _validate_pvalue ( self, value ):
        type_value = type( value )
        if (value is None) or isinstance( value, ( float, basestring ) ):
            return value
        if (((not isinstance( value, Binding ))      or
             (not isinstance( value.obj, LogIndex )) or
             (not unit_manager.are_compatible_families( self.family_name,
                                                  value.obj.family_name ) ) ) ):
            raise TraitError
        return value

    #---------------------------------------------------------------------------
    #  Implementation of the 'pvalue_units' property:
    #---------------------------------------------------------------------------

    def _get_pvalue_units ( self ):
        if self.actor_port.token is None:
            return ''
        else:
            value = self.actor_port.token.value
            if value is None:
                return ''
            return str( value.units )

    def _set_pvalue_units ( self, value ):
        quantity = self.actor_port.token.value
        self.actor_port.token.value = value
        self.trait_property_changed( 'pvalue_units', quantity, value )

    def _validate_pvalue_units ( self, value ):
        try:
            quantity = self.actor_port.token.value
            units    = unit_parser.parse_unit( value )
            if quantity.units.can_convert( units ):
                return Scalar( quantity.data,
                               units, family_name = quantity.family_name )
        except:
            pass
        raise TraitError

    #---------------------------------------------------------------------------
    #  Implementation of the 'actor_port' property:
    #---------------------------------------------------------------------------

    def _get_actor_port ( self ):
        if self._actor_port is None:
            from graph_helper import find_actor_port

            self._actor_port = find_actor_port( self, self.actor )
        return self._actor_port

#-------------------------------------------------------------------------------
#  'ConvergePortHandler' class:
#-------------------------------------------------------------------------------

class ConvergePortHandler ( Handler ):
    """ Handles the user clicking the 'Create Log' button:
    """

    def object_create_log_changed ( self, info ):
        """ Handles the user clicking the 'Create Log' button.
        """
        if info.initialized:
            port = info.object
            cnl  = CreateNewLog( port              = port,
                                 log_name          = port.name,
                                 family_name       = port.family_name,
                                 units_system_name = port.units_system_name )
            cnl.units = port.pvalue_units
            cnl.edit_traits( parent = info.ui.control )

#-------------------------------------------------------------------------------
#  Traits view definitions:
#-------------------------------------------------------------------------------

input_port_log_view = View( [
    [ 'name{Current selection}~^', '-' ],
    [ [ Item( name    = 'pvalue',
              label   = 'Log name',
              tooltip = 'Drag and drop a log here',
              editor  = DropEditor( klass = LogIndex, binding = True ),
              width   = 240 ),
        '-'
      ],
      [ Item( 'create_log',
              tooltip = 'Click to create a new log',
              editor  = ButtonEditor( label = 'Create Log...' ) ),
        '-<>' ],
      [ 'units~',
        'family_name{Family}~',
        'units_system_name{Units system}~',
        '-' ],
      '-'
    ] ],
    handler = ConvergePortHandler()
)

#-------------------------------------------------------------------------------
#  'InputTableHandler' class:
#-------------------------------------------------------------------------------

class InputTableHandler ( Handler ):

    #---------------------------------------------------------------------------
    #  Gets a specified View object:
    #---------------------------------------------------------------------------

    def trait_view_for ( self, info, view, object, object_name, trait_name ):
        """ Gets a specified View object.
        """
        if object.is_log:
            return input_port_log_view

        return View( [
            [ 'name{Current selection}~^', '-' ],
            [ [ [ 'description{Input Description}~',
                  'family_name{Family}~' ],
                [ [ Item( name    = 'pvalue',
                          label   = 'Value',
                          tooltip = 'Enter a scalar value here',
                          editor  = TextEditor( evaluate = float ) ),
                    '-' ],
                  [ Item( name    = 'pvalue_units',
                          tooltip = 'Enter the units the scalar value is in here',
                          editor  = EnumEditor(
                              values   = unit_manager.get_valid_unit_strings(
                                                          object.family_name ),
                              evaluate = str ) ),
                    '-<>' ], '-' ],
                [ Item( 'guidance@', width = 350, height = 60 ) ],
              ],
              [ '1' ],
              [ 'component_name{Chip Name}~',
                'bundle~',
                'name{Input Name}~',
                'external_type{External Type}~',
                'internal_type{Internal Type}~',
                '|' ],
            '-' ] ]
        )

input_table_handler = InputTableHandler()

#-------------------------------------------------------------------------------
#  'ConvergeInputPort' class:
#-------------------------------------------------------------------------------

class ConvergeInputPort ( ConvergePort ):
    dimension = Str

#-------------------------------------------------------------------------------
#  'ConvergeOutputPort' class:
#-------------------------------------------------------------------------------

class ConvergeOutputPort ( ConvergePort ):
    pass

#-------------------------------------------------------------------------------
#  'InputPortFilter' class:
#-------------------------------------------------------------------------------

class InputPortFilter ( TableFilter ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Partial match string for port name:
    port_name = Str

    # Partial match string for the component name:
    component_name = Str

    # Partial match string for the user guidance:
    guidance = Str

    # Connection state value:
    connected = Enum( "Don't care", "Connected", "Not connected" )

    #---------------------------------------------------------------------------
    #  Traits view definitions:
    #---------------------------------------------------------------------------

    filter_view = Group( 'component_name{Function name contains}',
                         'port_name{Input name contains}',
                         'guidance',
                         'connected',
                         help_id = 'enlib|HID_Edit_Filter_For_Inputs_Dlg',)

    #---------------------------------------------------------------------------
    #  Returns whether an object passes the filter or not:
    #---------------------------------------------------------------------------

    def filter ( self, port ):
        """ Returns whether an object passes the filter or not.
        """
        conn = self.connected
        if conn == 'Connected':
            conn = '==>'
        elif conn == 'Not connected':
            conn = ''
        return (
          (port.name.lower().find( self.port_name.lower() ) >= 0) and
          (port.component_name.lower().find( self.component_name.lower() )
              >= 0) and
          (port.guidance.lower().find( self.guidance.lower() ) >= 0) and
          ((conn == "Don't care") or (conn == port.connected))
        )

    #---------------------------------------------------------------------------
    #  Returns a user readable description of what the filter does:
    #---------------------------------------------------------------------------

    def description ( self ):
        """ Returns a user readable description of what the filter does.
        """
        if self.port_name != '':
            result = "Input name contains '%s'" % self.port_name
        else:
            result = ''
        if self.component_name != '':
            if result != '':
                result += ' and function name'
            else:
                result = 'Function name'
            result += " contains '%s'" % self.component_name
        if self.guidance != '':
            if result != '':
                result += ' and guidance'
            else:
                result = 'Guidance'
            result += " contains '%s'" % self.guidance
        conn = self.connected
        if conn != "Don't care":
            if result != '':
                result += ' and %s' % conn.lower()
            else:
                result = conn
        return result

#-------------------------------------------------------------------------------
#  'OutputPortFilter' class:
#-------------------------------------------------------------------------------

class OutputPortFilter ( TableFilter ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Partial match string for port name:
    port_name = Str

    # Partial match string for the component name:
    component_name = Str

    #---------------------------------------------------------------------------
    #  Traits view definitions:
    #---------------------------------------------------------------------------

    filter_view = Group( 'component_name{Function name contains}',
                         'port_name{Output name contains}',
                         help_id = 'enlib|HID_Edit_Filter_For_Outputs_Dlg',)

    #---------------------------------------------------------------------------
    #  Returns whether an object passes the filter or not:
    #---------------------------------------------------------------------------

    def filter ( self, port ):
        """ Returns whether an object passes the filter or not.
        """
        return (
          (port.name.lower().find( self.port_name.lower() ) >= 0) and
          (port.component_name.lower().find( self.component_name.lower() )
              >= 0)
        )

    #---------------------------------------------------------------------------
    #  Returns a user readable description of what the filter does:
    #---------------------------------------------------------------------------

    def description ( self ):
        """ Returns a user readable description of what the filter does.
        """
        if self.port_name != '':
            result = "Output name contains '%s'" % self.port_name
        else:
            result = ''
        if self.component_name != '':
            if result != '':
                result += ' and function name'
            else:
                result = 'Function name'
            result += " contains '%s'" % self.component_name
        return result

#-------------------------------------------------------------------------------
#  'PortValueColumn' class:
#-------------------------------------------------------------------------------

class PortValueColumn ( ObjectColumn ):

    #---------------------------------------------------------------------------
    #  Returns the value of the column for a specified object:
    #---------------------------------------------------------------------------

    def get_value ( self, object ):
        """ Returns whether the column is editable for a specified object.
        """
        value = getattr( object, self.name )
        if value == pvalue_default:
            return ''

        if object.is_log and type(value) == str:
            return value[ value.rfind( '/' ) + 1: ]

        return str( value )

    #---------------------------------------------------------------------------
    #  Returns whether the column is editable for a specified object:
    #---------------------------------------------------------------------------

    def is_editable ( self, object ):
        """ Returns whether the column is editable for a specified object.
        """
        return (not object.is_log)

    #---------------------------------------------------------------------------
    #  Returns whether a specified value is valid for dropping on the column
    #  for a specified object:
    #---------------------------------------------------------------------------

    def is_droppable ( self, object, value ):
        """ Returns whether a specified value is valid for dropping on the
            column for a specified object.
        """
        if object.is_connected or (not object.is_log):
            return False

        if isinstance( value, LogIndex ):
            return unit_manager.are_compatible_families( object.family_name,
                                                         value.family_name )

        return (isinstance( value, Binding )      and
                isinstance( value.obj, LogIndex ) and
                unit_manager.are_compatible_families( object.family_name,
                                                      value.obj.family_name ) )

    #---------------------------------------------------------------------------
    #  Returns the cell background color for the column for a specified object:
    #---------------------------------------------------------------------------

    def get_cell_color ( self, object ):
        """ Returns the cell background color for the column for a specified
            object.
        """
        if object.is_connected:
            return self.read_only_cell_color_

        if object.is_log:
            return DropColor

        return EditableColor

#-------------------------------------------------------------------------------
#  'PortValueUnitsColumn' class:
#-------------------------------------------------------------------------------

class PortValueUnitsColumn ( ObjectColumn ):

    #---------------------------------------------------------------------------
    #  Returns whether the column is editable for a specified object:
    #---------------------------------------------------------------------------

    def is_editable ( self, object ):
        """ Returns whether the column is editable for a specified object.
        """
        return (not object.is_log)

    #---------------------------------------------------------------------------
    #  Gets the editor for the column of a specified object:
    #---------------------------------------------------------------------------

    def get_editor ( self, object ):
        """ Gets the editor for the column of a specified object.
        """
        return EnumEditor( values   = unit_manager.get_valid_unit_strings(
                                                          object.family_name ) )
#                          evaluate = str )

#-------------------------------------------------------------------------------
#  Define ConvergePort related table columns and editors:
#-------------------------------------------------------------------------------

converge_inport_table_columns = [
    ObjectColumn( name      = 'component_name',
                  label     = 'Chip Name',
                  editable  = False ),
    ObjectColumn( name      = 'name',
                  label     = 'Input Name',
                  editable  = False ),
    PortValueColumn( name   = 'pvalue',
                  label     = 'Value',
                  droppable = True,
                  editor    = TextEditor( evaluate = float ) ),
    PortValueUnitsColumn( name = 'pvalue_units',
                  label     = 'Units' ),
    ObjectColumn( name      = 'family_name',
                  label     = 'Measure Of',
                  editable  = False ),
    ObjectColumn( name      = 'external_type',
                  label     = 'Input Type',
                  editable  = False ),
    ObjectColumn( name      = 'description',
                  label     = 'Input Description',
                  editable  = False ),
]

converge_outport_table_columns = [
    ObjectColumn( name      = 'component_name',
                  label     = 'Chip Name',
                  editable  = False ),
    ObjectColumn( name      = 'name',
                  label     = 'Output Name',
                  editable  = False ),
    ObjectColumn( name      = 'units',
                  editable  = False ),
    ObjectColumn( name      = 'family_name',
                  label     = 'Measure Of',
                  editable  = False ),
    ObjectColumn( name      = 'external_type',
                  label     = 'Output Type',
                  editable  = False ),
    ObjectColumn( name      = 'description',
                  label     = 'Output Description',
                  editable  = False ),
]

other_converge_output_port_table_columns = [
    ObjectColumn( name      = 'bundle',
                  label     = 'Chip Group',
                  editable  = False ),
    ObjectColumn( name      = 'units_system_name',
                  label     = 'Units System',
                  editable  = False ),
    ObjectColumn( name      = 'component_description',
                  label     = 'Chip Description',
                  editable  = False ),
    ObjectColumn( name      = 'internal_type',
                  label     = 'Internal Type',
                  editable  = False ),
]

other_converge_input_port_table_columns = \
    other_converge_output_port_table_columns + [
    ObjectColumn( name      = 'units',
                  label     = 'Internal Units',
                  editable  = False ),
    ObjectColumn( name      = 'guidance',
                  editable  = True ),
    ObjectColumn( name     = 'connected',
                  editable = False,
                  horizontal_alignment = 'center' ),
]

# Input/Output ports table menu:
input_table_menu = Menu(
    Action( name    = 'Plot',
            action  = 'on_plot_ports' ),
    Action( name    = 'Plot As...',
            action  = 'on_plot_as_ports',
            enabled = False ),
    Separator(),
    Action( name    = 'Spreadsheet',
            action  = 'on_spreadsheet_ports',
            enabled = False ),
    Separator(),
    Action( name   = 'Clear',
            action = 'on_clear_ports' )
)

output_table_menu = Menu(
    Action( name    = 'Plot',
            action  = 'on_plot_ports' ),
    Action( name    = 'Plot As...',
            action  = 'on_plot_as_ports',
            enabled = False ),
    Separator(),
    Action( name    = 'Spreadsheet',
            action  = 'on_spreadsheet_ports',
            enabled = False )
)

# Input ports table filters:
default_input_port_filter = InputPortFilter( connected = "Not connected" )

input_port_filters = [ EvalFilterTemplate,
                       RuleFilterTemplate,
                       MenuFilterTemplate,
                       InputPortFilter( name     = 'Filter template',
                                        template = True ),
                       default_input_port_filter ]

# Input ports table editor:
input_table_editor = TableEditor(
    columns           = converge_inport_table_columns,
    other_columns     = other_converge_input_port_table_columns,
    editable          = True,
    orientation       = 'vertical',
    filter            = default_input_port_filter,
    filters           = input_port_filters,
    edit_view         = 'input_port',
    edit_view_height  = 180,
    edit_view_handler = input_table_handler,
    search            = InputPortFilter(),
    menu              = input_table_menu
)

# Output ports table filters:
output_port_filters = [ EvalFilterTemplate,
                        RuleFilterTemplate,
                        MenuFilterTemplate,
                        OutputPortFilter( name     = 'Filter template',
                                          template = True ) ]

# Output ports table editor:
output_table_editor = TableEditor(
    columns       = converge_outport_table_columns,
    other_columns = other_converge_output_port_table_columns,
    editable      = False,
    orientation   = 'vertical',
    filters       = output_port_filters,
    search        = OutputPortFilter(),
    menu          = output_table_menu
)

#-------------------------------------------------------------------------------
#  'CreateNewLog' class:
#-------------------------------------------------------------------------------

class CreateNewLog ( Handler ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # The port this log is being created for:
    port = Instance( ConvergePort )

    # The index the new log should use:
    index = NamingIndex

    # The naming system name of the log suite the new log should be added to:
    log_suite = Str

    # The name of the new log within the log suite:
    log_name  = Regex( regex = '^[a-zA-Z0-9_]+$' )

    # The value to assign to all index values in the log:
    value = Float

    # The units for the new log:
    units = Property

    # The units family for the new log:
    family_name = unit_family_trait

    # The units system name for the new log:
    units_system_name = Str('KGS')

    # Current status message:
    status = Str

    #---------------------------------------------------------------------------
    #  Traits view definitions:
    #---------------------------------------------------------------------------

    def trait_view ( self, view ):
        return View( [
            [ 'index`Drag and drop a depth index from a log suite here`',
              'log_suite`The name of the log suite the log will be added to`~',
              'log_name`Enter the name of the new log here`',
              'value`Enter the initial value for all index values here`',
              Item( name    = 'units',
                    tooltip = 'The units of the new log',
                    editor  = EnumEditor(
                                values   = unit_manager.get_valid_unit_strings(
                                                             self.family_name ),
                                evaluate = str ) ),
              'family_name{Family}`The units family of the new log`~',
              'units_system_name{Units system}`The units system of the new '
                  'log`~' ],
            [ 'status~',
              '|<>' ] ],
            title   = 'Create a new log',
            kind    = 'live',
            width   = 0.25,
            undo    = False,
            revert  = False,
            help    = True,
            help_id = "enlib|HID_Create_a_New_Log_Dlg",
        )

    #---------------------------------------------------------------------------
    #  Implementation of the 'units' property:
    #---------------------------------------------------------------------------

    def _get_units ( self ):
        if self._units is None:
            return ''
        return self._units

    def _set_units ( self, value ):
        if isinstance(value, basestring):
            value = unit_parser.parse_unit( value )
        if not self.port.actor_port.token.value.units.can_convert( value ):
            raise TraitError
        self._units = value

    #---------------------------------------------------------------------------
    #  Handles the 'port' trait being changed:
    #---------------------------------------------------------------------------

    def _port_changed ( self ):
        """ Handles the 'port' trait being changed.
        """
        self.index = self.port.canvas.default_index

    #---------------------------------------------------------------------------
    #  Handles the 'index' trait being changed:
    #---------------------------------------------------------------------------

    def _index_changed ( self, value ):
        """ Handles the 'index' trait being changed.
        """
        if value[:1] == '<':
            return

        self.status    = ''
        col            = value.rfind( '/' )
        self.log_suite = value[ : col ]

        workspace      = get_application().service_registry.get_service(
                             'enthought.envisage.project.IWorkspace' )
        self._ls = ls  = workspace.lookup( self.log_suite )

        # Try to make sure the current log name does not exist in the log suite:
        log_name = self.log_name
        if ls.has_log( log_name ):
            for i in range( 2, 100 ):
                new_name = '%s_%d' % ( log_name, i )
                if not ls.has_log( new_name ):
                    self.log_name = new_name
                    break

    #---------------------------------------------------------------------------
    #  Handles a request to close the dialog by the user:
    #---------------------------------------------------------------------------

    def close ( self, info, is_ok ):
        """ Handles a request to close the dialog by the user.
        """
        # If the user hit 'cancel', then let the dialog close:
        if not is_ok:
            return True

        # Make sure a log suite has been specified:
        log_suite = self._ls
        if log_suite is None:
            self.status = ("Please drag a log suite index to the 'Index' "
                           "field.")
            return False

        # Make sure the log name is new within the log suite:
        log_name = self.log_name
        if log_suite.has_log( log_name ):
            self.status = ("A log with that name already exists in the log "
                           "suite.")
            return False

        # Create the Log and store it in the log suite:

        family_name = self.family_name
        index       = log_suite.index
        log_suite.add_log( Log( array( [ self.value ] * len( index.data ) ),
                                index       = index,
                                units       = self.units,
                                name        = log_name,
                                family_name = family_name ) )

        # Finally, assign the full name as the value of the port:
        self.port.pvalue = '%s/%s' % (self.log_suite, log_name )

        # Allow the dialog to close:
        return True

#-------------------------------------------------------------------------------
#  'ConvergeBundle' class:
#-------------------------------------------------------------------------------

class ConvergeBundle ( HasStrictTraits ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Name of the bundle:
    name          = Regex( value = 'connection', regex = r'\S' )

    # Original name of the bundle:
    original_name = Str( Undefined )

    # Name of the group the bundle belongs to:
    group_name    = Str( 'default' )

    # Ports contained in the bundle:
    ports         = List( ConvergePort )

    # Can the bundle and all its ports be deleted?
    deletable     = false

    # Is the bundle visible in the user interface:
    visible       = true

    #---------------------------------------------------------------------------
    #  Handles the 'name' trait being changed:
    #---------------------------------------------------------------------------

    def _name_changed ( self ):
        if self.original_name is Undefined:
            self.original_name = self.name

#-------------------------------------------------------------------------------
#  'ConvergeInputBundle' class:
#-------------------------------------------------------------------------------

class ConvergeInputBundle ( ConvergeBundle ):
    pass

#-------------------------------------------------------------------------------
#  'ConvergeOutputBundle' class:
#-------------------------------------------------------------------------------

class ConvergeOutputBundle ( ConvergeBundle ):
    pass

#-------------------------------------------------------------------------------
#  'ConvergeRepeatBundle' class:
#-------------------------------------------------------------------------------

class ConvergeRepeatBundle ( ConvergeBundle ):
    """ Bundle that can be repeated """

    # When there is no limit use UNLIMITED for max_num:
    UNLIMITED = maxint

    # Names sequentially used for naming the instances of the repeat group:
    names   = List( Str )

    # Minimum number of occurrences of the repeat group:
    min_num = CInt( 1 )

    # Maximum number of occurrences of the repeat group:
    max_num = CInt( UNLIMITED )

#-------------------------------------------------------------------------------
#  'ConvergeRepeatButtonBundle' class:
#-------------------------------------------------------------------------------

class ConvergeRepeatButtonBundle ( ConvergeBundle ):

    #---------------------------------------------------------------------------
    #  Trait definition overrides:
    #---------------------------------------------------------------------------

    # Name of the bundle:
    name = ''

#-------------------------------------------------------------------------------
#  'ConvergeActorTraits' class:
#
#  This class is used as a mix-in to add the traits needed to be used on a
#  Converge canvas.
#-------------------------------------------------------------------------------

class ConvergeActorTraits ( HasTraits ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Name of the chip (overrides the definition in the base class):
    name = Property( Str )

    # List of input bundles (editable version):
    instance_input_bundles = List( ConvergeBundle )

    # List of output bundles (editable version):
    instance_output_bundles = List( ConvergeBundle )

    # List of input bundles (current version):
    input_bundles = List( ConvergeBundle )

    # List of output bundles (current version):
    output_bundles = List( ConvergeBundle )

    # List of input ports (in function argument order):
    input_ports = List( Port )

    # List of unbound (i.e. unconnected) input ports:
    unbound_input_ports = List( Port )

    # Unbound input tokens.
    unbound_input_tokens = Property(Dict)

    # Output tokens.
    output_tokens = Property(Dict)

    # List of output ports (in function result tuple order):
    output_ports = List( Port )

    # Short description of function's purpose.
    description = Str

    # Help text description of the chip:
    help = Str

    # History of change:
    history = Str

    # Should the function normally be present in user interfaces?
    is_hidden = Bool()

    # Unique id of this ConvergeActor used to find a prototype in the
    # ChipManager:
    id = Str

    # The file containing function definition and metadata for this actor if
    # known:
    file_name = Str

    # Thrown when the validator believes it has made changes to the
    # ConvergeActor.
    changed = Event

    #--------------------------------------------------------------------------
    # protected traits used during execution cycle
    # FIXME: factor this out to a subclass used for Actors that use Logs
    # defined as Any to avoid dependency on cp.log
    _output_log_index = Any

    # Index into input_port of a port that is a Log from which the LogIndex
    # is taken for use with output Logs.
    _index_for_log_index = Int(-1)

    #---------------------------------------------------------------------------
    #  'ConvergeActorTraits' interface.
    #---------------------------------------------------------------------------

    def _get_unbound_input_tokens ( self ):
        """ Returns the unbound input tokens. """

        unbound_input_tokens = {}
        for port in self.unbound_input_ports:
            unbound_input_tokens[port.name] = port.token

        return unbound_input_tokens

    def _get_output_tokens ( self ):
        """ Returns the unbound tokens. """

        output_tokens = {}
        for port in self.output_ports:
            output_tokens[port.name] = port.token

        return output_tokens

    #---------------------------------------------------------------------------
    #  Binds an unbound input port:
    #---------------------------------------------------------------------------

    def bind_input_port ( self, port ):
        """ Binds an unbound input port.
        """
        self.unbound_input_ports.remove( port )

    #---------------------------------------------------------------------------
    #  Unbinds a previously bound input port:
    #---------------------------------------------------------------------------

    def unbind_input_port ( self, port ):
        """ Unbinds a previously bound input port.
        """
        unbound_ports = self.unbound_input_ports
        n = len( unbound_ports )
        i = 0
        for j, input_port in enumerate( self.input_ports ):
            if port is input_port:
                unbound_ports.insert( i, port )
                break
            for k in range( i, min( j + 1, n ) ):
                if input_port is unbound_ports[k]:
                    i = k + 1
                    break

    #---------------------------------------------------------------------------
    #  Event handlers:
    #---------------------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles a set of input bundles being assigned to the actor:
    #---------------------------------------------------------------------------

    def _input_bundles_changed ( self, bundles ):
        """ Handles a set of input bundles being assigned to the actor.
        """
        for bundle in bundles:
            for port in bundle.ports:
                port.actor = self

    #---------------------------------------------------------------------------
    #  Handles a set of output bundles being assigned to the actor:
    #---------------------------------------------------------------------------

    def _output_bundles_changed ( self, bundles ):
        """ Handles a set of input bundles being assigned to the actor.
        """
        self._input_bundles_changed( bundles )

#-------------------------------------------------------------------------------
#  'ConvergeActor' class:
#-------------------------------------------------------------------------------

class ConvergeActor ( ValueActor, ConvergeActorTraits ):

    #---------------------------------------------------------------------------
    #  Input/Output 'template' bundles (overridden in subclass):
    #---------------------------------------------------------------------------

    class_input_bundles  = []
    class_output_bundles = []

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Name of the chip (overrides the definition in the base class):
    name = Property( Str )

    # Evaluation function:
    function = Tuple( Callable )

    #---------------------------------------------------------------------------
    #  Initializes the actor:
    #---------------------------------------------------------------------------

    def __init__ ( self, *args, **traits ):
        """ Initializes the actor.
        """
        super( ConvergeActor, self ).__init__( *args, **traits )

        # Create copies of the 'master' values in case they are edited:
        self.instance_input_bundles  = [ x.clone_traits()
                                         for x in self.class_input_bundles  ]
        self.instance_output_bundles = [ x.clone_traits()
                                         for x in self.class_output_bundles ]

        # self.input_ports - list of actor ports in the order that corresponds
        # to the argument order for self.funcion
        # self.unbound_input_ports - list of unconnected input actor ports
        # (which is initially the same as the set of input ports, but changes
        # as connections with other actors are made and broken).
        self.unbound_input_ports = self.input_ports = self.init_ports(
                                            self.instance_input_bundles, True  )

        # self.output_ports - list of actor ports in the order that corresponds
        # to the return value order for self.funcion
        self.output_ports = self.init_ports( self.instance_output_bundles,
                                             False )

    def __getstate__ ( self ):
        """ return dictionary of attributes to pickle. """
#        print '\n\nconverge_actor.__getstate__', self.name
        attributes = {}
        attributes.update(self.__dict__)

        # Make sure function is not pickled
        for name in [ 'function', '_output_log_index' ]:
            if attributes.has_key(name):
                del attributes[name]

#        # DEBUG
#        print '\n\nconverge_actor.__getstate__'
#        for key, value in attributes.items():
#            print '    ', key, ':', value
        return attributes

    #---------------------------------------------------------------------------
    #  Returns a list of ports from a list of bundles:
    #---------------------------------------------------------------------------

    def init_ports ( self, bundles, is_input ):
        """ Returns a list of ports from a list of bundles.
        """
        ports = []
        for bundle in bundles:
            ports.extend( bundle.ports )
        ports.sort( lambda l, r: cmp( l.index, r.index ) )
        return [ self.new_port( port.name, is_input, port )
                 for port in ports
        ]

    #---------------------------------------------------------------------------
    #  Runs the function associated with the actor:
    #---------------------------------------------------------------------------

    def fire ( self ):
        """ Runs the function associated with the actor.
        """
        # build argument list, convert units as needed
        if self._index_for_log_index > -1 and len(self.output_ports) > 0:
            # If actor has log inputs and outputs, get the LogIndex the outputs
            # will need:
            port_for_index = self.input_ports[self._index_for_log_index]

            args = []
            for port in self.input_ports:

                if port is port_for_index:
                    self._output_log_index = port_for_index.get_outside()
                    if isinstance( self._output_log_index, Log ):
                        log = self._output_log_index
                        self._output_log_index = self._output_log_index.index
                    else:
                        log=self._output_log_index
                    arg = port.get( from_outside=log )

                else:
                    arg = port.get()
                args.append(arg)

        else:
            # normal case, no need for special index extraction
            args    = [ p.get() for p in self.input_ports ]


        # call the chip function
        results = self.function[0]( *args )

        # move the function results into the chip output ports
        if len( self.output_ports ) == 1:
            self.output_ports[0].put( results )

        else:
            for i, port in enumerate( self.output_ports ):
                port.put( results[i] )

        self._output_log_index = None
        return

    #---------------------------------------------------------------------------
    #  Implementation of the 'name' property:
    #---------------------------------------------------------------------------

    def _get_name ( self ):
        return getattr( self, '_name', 'unknown' )

    def _set_name ( self, name ):
        if name.strip() == '':
            raise TraitError, 'Chip name cannot be empty or only white space'
        container = self.container
        if container is not None:
            for node in container.nodes:
                if (node is not self) and (node.name == name):
                    raise TraitError, \
                          'Conflict with existing chip name: %s' % name

        old        = self.name
        self._name = name
        if name != old:
            self.trait_property_changed( 'name', old, name )

        return

    ### Node methods ##########################################################
    def new_port ( self, name, is_input, meta ):
        """ Create a port compatible with this node from the meta description.
        """
        return _new_port( self, name, is_input, meta )

#-------------------------------------------------------------------------------
#  'ConvergeCompositeActor' class:
#-------------------------------------------------------------------------------

class ConvergeCompositeActor ( CompositeValueActor, ConvergeActorTraits ):

    ### Node methods ##########################################################
    def new_port ( self, name, is_input, meta ):
        """ Create a port compatible with this node from the meta description.
        """
        return _new_port( self, name, is_input, meta )


#-------------------------------------------------------------------------------
#  'ConvergeCommentActor' class:
#-------------------------------------------------------------------------------

class ConvergeCommentActor ( ConvergeActor ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # User defined comment:
    comment = Str( 'comment' )


#### Private functions ########################################################


def remove_noise ( string ):
    """  Removes any noise words from a specified string. """
    while True:
        for word in noise_words:
            if ends_with( string, word ):
                string = string[ : -len( word ) ]
                break
        else:
            return string


def ends_with ( string, suffix ):
    """ Returns whether one strings ends with another string. """
    return (suffix == string[ -len( suffix ): ])


def _new_port ( actor, name, is_input, meta ):
    """ Create a port compatible 'actor' node from the meta description.
    """
    actor._check_port_name( meta.name )
    units=unit_parser.parse_unit(meta.units, suppress_warnings=True)

    kw = { 'token':             meta.value,
           'internal_type':     meta.internal_type,
           'external_type':     meta.external_type,
           'description':       meta.description,
           'units':             units,
           'family_name':       meta.family_name,
           'units_system_name': meta.units_system_name }

    pf=get_application().service_registry.get_service(
                                            'enthought.graph.vpl.PortFactory')

    port = pf.port( meta.external_type, meta.internal_type, meta.name, is_input, **kw )
    port.container = actor

    #FIXME: this should not be here....also, last one in wins
    if is_input and meta.external_type \
                        in [ 'cp.log.log.Log', 'cp.log.log_index.LogIndex']:
        actor._index_for_log_index = meta.index

    return port

#### EOF ######################################################################
