#-------------------------------------------------------------------------------
#  
#  Defines the ConvergeOptimizer class for optimizing a converge dataflow.  
#  
#  Written by: David C. Morrill
#  
#  Date: 04/14/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

from enthought.traits.api \
    import HasPrivateTraits, HasStrictTraits, Range, List, Instance, Any, \
           Event, Bool
    
from enthought.traits.ui.api \
    import View, Item, DropEditor, ButtonEditor
    
from graph_helper \
    import DraggablePort
    
from enthought.envisage.ui \
    import View as UIViewImpl
    
#-------------------------------------------------------------------------------
#  Trait definitions:  
#-------------------------------------------------------------------------------
        
# Defines a port which can be set using drag and drop:
ActorPort = Instance( DraggablePort, 
                      allow_none = True, 
                      editor     = DropEditor( klass = DraggablePort ) )
                      
#-------------------------------------------------------------------------------
#  'OptimizationTarget' class:  
#-------------------------------------------------------------------------------

class OptimizationTarget ( HasStrictTraits ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Port supplying measured data to optimize against:
    measured = ActorPort
    
    # Port defining the computed optimization values:
    computed = ActorPort
    
    # Weighting factor applied to this target:
    weight = Range( 0.0, 1.0, 1.0 )
    
    #---------------------------------------------------------------------------
    #  Traits view definitions:  
    #---------------------------------------------------------------------------
        
    traits_view = View( 'measured', 'computed', 'weight', '_' )
    
#-------------------------------------------------------------------------------
#  'StateParameter' class:   
#-------------------------------------------------------------------------------
        
class StateParameter ( HasStrictTraits ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Port defining the state parameter:
    port = ActorPort
    
    #---------------------------------------------------------------------------
    #  Traits view definitions:  
    #---------------------------------------------------------------------------
        
    traits_view = View( [ 'port', '|<>' ] )
    
#-------------------------------------------------------------------------------
#  Trait definitions:  
#-------------------------------------------------------------------------------
        
AnOptimizationTarget = Instance( OptimizationTarget, () )

AStateParameter      = Instance( StateParameter, () )

#-------------------------------------------------------------------------------
#  'ConvergeOptimizer' class:  
#-------------------------------------------------------------------------------

class ConvergeOptimizer ( HasPrivateTraits ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
    
    # Event fired when user wants to run optimization:
    optimize = Event( Bool, editor = ButtonEditor( label = 'Optimize' ) )
    
    # Lists of optimization targets:
    targets = List( AnOptimizationTarget, rows = 5 )
    
    # List of optimization state parameters:
    state_parameters = List( AStateParameter, rows = 10 )
    
    #---------------------------------------------------------------------------
    #  Trait view definitions:  
    #---------------------------------------------------------------------------
        
    traits_view = View( [ [ ' ', '_', 'optimize',  '|{Run}<>' ],
                          '|{Run}<>' ],
                        [ [ ' ', '_', 'targets@', '|{Targets}<>' ], 
                          '|{Targets}<>' ],
                        [ [ ' ', '_', 'state_parameters@', 
                            '|{State Parameters}<>' ],
                          '|{State Parameters}<>' ],
                        resizable = True,
                        nb_bottom = True )
        
    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------
    
    def __init__ ( self, **traits ):
        self.targets          = [ OptimizationTarget() ]
        self.state_parameters = [ StateParameter() ]
        super( ConvergeOptimizer, self ).__init__( **traits )

#-------------------------------------------------------------------------------
#  'ConvergeOptimizerEditor' class:  
#-------------------------------------------------------------------------------

class ConvergeOptimizerEditor ( UIViewImpl ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # The view object:
    optimizer = Instance( ConvergeOptimizer )
    
    # The plugin object:
    plugin    = Any
    
    #---------------------------------------------------------------------------
    #  Traits view definitions:  
    #---------------------------------------------------------------------------
        
    traits_view = View( [ Item( 'optimizer@', resizable = True ), '|<>' ], 
                        kind = 'subpanel' )

    #---------------------------------------------------------------------------
    #  Creates the control associated with the view:  
    #---------------------------------------------------------------------------
        
    def _create_contents( self, parent ):
        """ Creates the control associated with the view.
        """
        from enthought.graph.vpl.graph_editor_plugin import GraphEditorPlugin
        
        self.optimizer = ConvergeOptimizer()
        self.plugin    = GraphEditorPlugin.instance
        return self.edit_traits( parent = parent ).control
        
