#-------------------------------------------------------------------------------
#
#  Defines the GraphCanvas subclass of OMCanvas used with the graph package.
#
#  Written by: David C. Morrill
#
#  Date: 05/24/2005
#
#  (c) Copyright 2005 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from enthought.pyface.timer.api \
    import do_later

from enthought.traits.ui.key_bindings \
    import KeyBindings

from enthought.enable.om \
    import OMCanvas, OMMenuController

from enthought.traits.api \
    import Instance, Any, Event, false, Property

from enthought.traits.ui.api \
    import UndoHistory, View, Group, Item, Handler, CustomEditor, Tabbed

from enthought.traits.ui.menu \
    import NoButtons

from enthought.enable \
    import Scrolled

from enthought.enable.wx \
    import Window

from enthought.envisage \
    import get_application

from enthought.envisage.project \
    import CookieManager

from enthought.plugins.chaco.cookie.plot_cookie \
    import PlotCookie

from enthought.logger \
    import logger

from chip_listener \
    import ChipListener

from converge_actor \
    import input_table_editor, output_table_editor

from cp.log.naming_log \
    import NamingIndex

#-------------------------------------------------------------------------------
#  Creates a window for editing a workflow canvas:
#-------------------------------------------------------------------------------

def create_window ( parent, editor ):
    """ Creates a window for editing a workflow canvas.
    """
    from graph_canvas_toolbar import GraphCanvasToolbar

    editor.scrollable = True

    panel = wx.Panel( parent, -1 )
    sizer = wx.BoxSizer( wx.VERTICAL )

    canvas         = editor.object
    canvas.toolbar = GraphCanvasToolbar( parent = panel, canvas = canvas )
    window         = Window( panel, component   = Scrolled( canvas ),
                                    focus_owner = canvas )

    sizer.Add( canvas.toolbar.control, 0, wx.EXPAND )
    sizer.Add( wx.StaticLine( panel, -1 ), 0, wx.EXPAND | wx.TOP, 3 )
    sizer.Add( window.control, 1, wx.EXPAND )
    panel.SetSizer( sizer )

    return panel

#-------------------------------------------------------------------------------
#  'CanvasHandler' class:
#-------------------------------------------------------------------------------

class CanvasHandler ( Handler ):

    #---------------------------------------------------------------------------
    #  Handles a request to 'Plot' some ports:
    #---------------------------------------------------------------------------

    def on_plot_ports ( self, info, ports ):
        """ Handles a request to 'Plot' some ports.
        """
        bindings = self._get_bindings_for( info, ports )
        cm       = CookieManager()

        while len( bindings ) > 0:
            cookie = cm.get_cookie( PlotCookie, bindings[0].obj )

            # If there is a plot cookie for the first binding, invoke it:
            if cookie is not None:
                starting_length = len( bindings )
                cookie.plot( info.ui, bindings )

                # Protect ourselves from a cookie that does not remove the
                # bindings that it consumed.  Warn the user and remove the
                # first binding which was used to locate the cookie:
                if starting_length == len( bindings ):
                    logger.warning(
                        'PlotCookie %s failed to remove bindings.' % cookie )
                    del bindings[0]

            # Otherwise, if there is not a plot cookie for the first binding,
            # then just remove it from the set of bindings.
            else:
                logger.debug("Plot unavailable for %s" % bindings[0].name)
                del bindings[0]

    #---------------------------------------------------------------------------
    #  Handles a request to 'Plot as...' some ports:
    #---------------------------------------------------------------------------

    def on_plot_as_ports ( self, info, ports ):
        """ Handles a request to 'Plot as...' some ports.
        """
        pass

    #---------------------------------------------------------------------------
    #  Handles a request to 'spreadsheet' some ports:
    #---------------------------------------------------------------------------

    def on_spreadsheet_ports ( self, info, ports ):
        """ Handles a request to 'spreadsheet' some ports.
        """
        pass

    #---------------------------------------------------------------------------
    #  Handles a request to 'Clear' the contents of some ports:
    #---------------------------------------------------------------------------

    def on_clear_ports ( self, info, ports ):
        """ Handles a request to 'Clear' the contents of some ports.
        """
        for port in ports:
            port.pvalue = None

#-- Private methods ------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Gets the name system bindings for a list of ports:
    #---------------------------------------------------------------------------

    def _get_bindings_for ( self, info, ports ):
        """ Gets the name system bindings for a list of ports.
        """
        workspace = get_application().service_registry.get_service(
                        'enthought.envisage.project.IWorkspace' )
        prefix    = ('/'.join( workspace.search( info.object )[0].split(
                         '/' )[:-1] ) + '/')
        bindings  = []
        for port in ports:
            ap   = port.actor_port
            name = ap.full_name.replace( '.', '/' )
            if ap.is_input:
                name += '/value'
            bindings.append( workspace.lookup_binding( prefix + name ) )
        return bindings

#-------------------------------------------------------------------------------
#  'GraphCanvas' class:
#-------------------------------------------------------------------------------


class GraphCanvas ( OMCanvas, ChipListener ):

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Event fired whenever a change occurs that requires the dataflow to be run:
    run                = Event( bool )

    # Should the dataflow be run automatically when the canvas is modified?
    auto_run           = false

    # Key bindings for mapping keyboard events to actions:
    key_bindings       = Instance( KeyBindings )

    # Toolbar for the canvas:
    #toolbar           = Instance( GraphCanvasToolbar )
    toolbar            = Any

    # Controller used to handle toolbar events:
    toolbar_controller = Instance( OMMenuController )

    # Default index log associated with the canvas:
    default_index      = NamingIndex

    # A list of all input ports at the top level of the canvas:
    input_ports        = Property

    # A list of all output ports at the top level of the canvas:
    output_ports       = Property

    #---------------------------------------------------------------------------
    #  Traits view definitions:
    #---------------------------------------------------------------------------

    traits_view = View(
                      Tabbed(
                          Group(
                              Item( 'canvas',
                                    editor = CustomEditor( create_window ) ),
                              '|<>{Dataflow}',
                              export = '' ), #'DockShellWindow' ),
                          Group( [ [ Item( 'default_index',
                                      label   = 'Depth Index',
                                      tooltip = 'Drag and drop an index log here',
                                      width   = 240 ), '-' ],
                              [ Item( 'input_ports',
                                      id     = 'inputs_table',
                                      editor = input_table_editor,
                                      dock   = 'horizontal',
                                      height = 500 ),
                                '|<>' ],
                              '|' ],
                            '|<>{Inputs}',
                            export = 'DockShellWindow' ),
                          Group(
                              Item( 'output_ports',
                                    id     = 'outputs_table',
                                    editor = output_table_editor ),
                              '|<>{Outputs}',
                              export = 'DockShellWindow' ),
                          id = 'notebook' ),
                        id      = 'cp.converge.io_ports_view',
                        dock    = 'tab',
                        handler = CanvasHandler(),
                        buttons = NoButtons )

    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, *components, **traits ):
        OMCanvas.__init__(self, *components, **traits)

        from chip_manager \
            import ChipManager

        ChipManager().add_listener( self )

#-- Property definitions -------------------------------------------------------

    #---------------------------------------------------------------------------
    #  'input_ports' property:
    #---------------------------------------------------------------------------

    def _get_input_ports ( self ):
        """ Returns a list of all current input port on the canvas.
        """
        ports = self._input_ports
        if ports is not None:
            return ports

        # Extract just the input ports from each contact:
        port_lists = [ c.data[1].ports for c in self.contacts
                       if c.category == 'input' ]
        self._input_ports = ports = []

        # Convert to one big list:
        for list in port_lists:
            ports.extend( list )

        # Make sure each port knows it's part of this canvas:
        for port in ports:
            port.canvas = self

        return ports

    #---------------------------------------------------------------------------
    #  'output_ports' property:
    #---------------------------------------------------------------------------

    def _get_output_ports ( self ):
        """ Returns a list of all current output port on the canvas.
        """
        ports = self._output_ports
        if ports is not None:
            return ports

        # Extract just the output ports from each contact:
        port_lists = [ c.data[1].ports for c in self.contacts
                       if c.category == 'output' ]

        # Convert to one big list:
        self._output_ports = ports = []
        for list in port_lists:
            ports.extend( list )

        # Make sure each port knows it's part of this canvas:
        for port in ports:
            port.canvas = self

        return ports

#-- Event handlers -------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles components being added to or deleted from the canvas:
    #---------------------------------------------------------------------------

    def _components_items_changed ( self ):
        old, self._input_ports = self._input_ports or [], None
        self.trait_property_changed( 'input_ports',  old, self.input_ports )

        old, self._output_ports = self._output_ports or [], None
        self.trait_property_changed( 'output_ports', old, self.output_ports )

    #---------------------------------------------------------------------------
    #  Handles the canvas being updated in a way that requires the dataflow to
    #  run:
    #---------------------------------------------------------------------------

    def _run_changed ( self ):
        """ Handles the canvas being updated in a way that requires the
            dataflow to run.
        """
        if self.auto_run:
            do_later( self.controller.run, self )

    #---------------------------------------------------------------------------
    #  Returns the state of the canvas to be used when pickling:
    #---------------------------------------------------------------------------

    def __getstate__ ( self ):
        """ Returns the state of the canvas to be used when pickling.
        """
#        print '\n\n\n\nGraphCanvas.__getstate__: dict copy...'
        dic = self.__dict__.copy()

        for name in [ 'container', 'toolbar', 'toolbar_controller',
                      'controller', 'history', 'key_bindings', 'style',
                      '_input_ports', '_output_ports',
                      'selection',
                    ]:
            if dic.has_key( name ):
                del dic[ name ]
#        # DEBUG
#        print 'GraphCanvas.__getstate__ returns:'
#        for key, value in dic.items():
#            print '    ', key, ':', value
        return dic

    #---------------------------------------------------------------------------
    #  Restores the states of a previously pickled canvas:
    #---------------------------------------------------------------------------

    def __setstate__ ( self, state ):
        """ Restores the states of a previously pickled canvas.
        """
        import graph_canvas_controller
        self.__dict__.update( state )

        self._input_ports = self._output_ports = None

        self.controller   = graph_canvas_controller.graph_canvas_controller
        self.key_bindings = graph_canvas_controller.graph_canvas_key_bindings
        self.history      = UndoHistory()
        self.style        = graph_canvas_controller.GraphCanvasStyle()
        return

    #---------------------------------------------------------------------------
    # 'ChipListener' interface.
    #---------------------------------------------------------------------------

    def on_chip_changed ( self, event ):
        """ Called when a chip prototype has been changed.
        """
        from chip_manager \
            import ChipManager

        # Validate the chip, bring it up to date:
        ChipManager().validate( event.id, self.data )

