#-------------------------------------------------------------------------------
#  
#  Defines the OMCanvasController subclass used to manage a GraphCanvas 
#  containing graph Actor objects.
#  
#  Written by: David C. Morrill
#  
#  Date: 03/03/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

import wx
import traceback

from os.path \
    import abspath
    
from cp.log \
    import LogSuite
    
from converge_actor \
    import ConvergeCompositeActor, ConvergeCommentActor, \
           ConvergeInputPort, ConvergeOutputPort, ConvergeInputBundle, \
           ConvergeOutputBundle
           
from graph_undo_item \
    import GraphUndoItem, GraphBoundsUndoItem, GraphExpansionUndoItem, \
           GraphMoveUndoItem, GraphSelectionUndoItem, \
           GraphCreateCompositeUndoItem, GraphComponentUndoItem, \
           GraphConnectionUndoItem
           
from graph_component \
    import GraphComponent
    
from cp.rockphysics.converge.plugins.log_suite_sink \
    import LogSuiteSink
    
from enthought.envisage.project.project_plugin \
    import ProjectPlugin

from enthought.envisage.project \
    import CookieManager
    
from enthought.envisage.project.action.open_cookie \
    import OpenCookie
    
from enthought.envisage.project.wizard.new_file_wizard \
    import FileFactory, NewFileWizard
    
from enthought.envisage.ui \
    import UIPlugin
    
from enthought.pyface.image_resource \
    import ImageResource
    
from enthought.pyface.api \
    import OK
    
from enthought.naming.binding \
    import Binding
    
from enthought.naming.exception \
    import NameNotFoundError
    
from enthought.enable.om \
    import OMCanvasController, OMCanvasStyle, OMTrackContactManager, OMContact,\
           OMComponent, OMComponentStyle, OMOuterComponent, \
           OMOuterComponentStyle, OMLabel, OMLink
           
from enthought.enable.base_container \
    import default_container
    
from enthought.traits.ui.key_bindings \
    import KeyBinding, KeyBindings

from enthought.graph.vpl.chip_manager \
    import ChipManager

from enthought.graph.vpl.cookie.resource_sink_cookie \
    import ResourceSinkCookie
    
from enthought.graph.vpl.cookie.resource_source_cookie \
    import ResourceSourceCookie
               
from enthought.traits.api \
    import Instance, Str

from enthought.traits.ui.api \
     import UndoHistory
    
from enthought.traits.ui.menu \
    import Menu, Action, Separator, ToolBar

from enthought.graph import FireError

from enthought.io.file import File

from enthought.logger import logger

from graph_actor_folder \
    import GraphActor
    
from graph_component_controller \
    import graph_component_controller, graph_component_style
    
from graph_comment_controller \
    import graph_comment_controller
    
from graph_helper \
    import auto_connect, connect_ports, NominalColor, graph_actor_style, \
           graph_comment_style, set_component_style, recompute_links, \
           disconnect_links, new_port_name, unique_actor_name
    
#-------------------------------------------------------------------------------
#  Constants:  
#-------------------------------------------------------------------------------

# Action for toggling the canvas grid on or off:
toggle_grid_action = Action( name   = 'Visible',
                             action = 'on_toggle_grid',
                             style  = 'toggle' )
                             
# Action for toggling the canvas 'snap to grid' feature on or off:
toggle_snap_action = Action( name   = 'Snap to grid',
                             action = 'on_toggle_snap',
                             style  = 'toggle' )
                             
# Actions for setting the grid size:
grid_size_actions = ( Action( name   = '16 x 16',
                              action = 'on_grid_size_16x16',
                              style  = 'toggle' ),
                      Action( name   = '32 x 32',
                              action = 'on_grid_size_32x32',
                              style  = 'toggle' ),
                      Action( name   = '48 x 48',
                              action = 'on_grid_size_48x48',
                              style  = 'toggle' ),
                      Action( name   = '64 x 64',
                              action = 'on_grid_size_64x64',
                              style  = 'toggle' ) )
                   
# Toggle 'autorun' on/off action:                              
auto_run_action = Action( name   = 'Auto run',
                          action = 'on_autorun',
                          style  = 'toggle' )

# List of all arrow keys:
arrow_keys = [ 'Left', 'Right', 'Up', 'Down' ]

# Help template for a composite component:
composite_help = 'A composite component containing:\t%s'

#-------------------------------------------------------------------------------
#  Canvas key bindings:  
#-------------------------------------------------------------------------------

graph_canvas_key_bindings = KeyBindings( 
                
    KeyBinding( binding1    = 'z', 
                description = 'Undo the most recent change',
                method_name = 'undo' ),
                
    KeyBinding( binding1    = 'y', 
                description = 'Redo the most recently undone change',
                method_name = 'redo' ),
                
    KeyBinding( binding1    = 'Ctrl-a',
                binding2    = 'a',
                description = 'Select all components',
                method_name = 'select_all' ),
                
    KeyBinding( binding1    = 'Tab', 
                description = 'Select next component',
                method_name = 'select_next_component' ),
                
    KeyBinding( binding1    = 'Shift-Tab', 
                description = 'Select previous component',
                method_name = 'select_previous_component' ),
                
    KeyBinding( binding1    = 'Esc', 
                binding2    = 'A',
                description = 'Unselect all selected components',
                method_name = 'clear_selection' ),
                
    KeyBinding( binding1    = 'Up', 
                description = 'Move selected components up',
                method_name = 'move_up_slow' ),
                
    KeyBinding( binding1    = 'Down', 
                description = 'Move selected components down',
                method_name = 'move_down_slow' ),
                
    KeyBinding( binding1    = 'Left', 
                description = 'Move selected components left',
                method_name = 'move_left_slow' ),
                
    KeyBinding( binding1    = 'Right', 
                description = 'Move selected components right',
                method_name = 'move_right_slow' ),
                
    KeyBinding( binding1    = 'Alt-Up', 
                description = 'Move selected components up faster',
                method_name = 'move_up_fast' ),
                
    KeyBinding( binding1    = 'Alt-Down', 
                description = 'Move selected components down faster',
                method_name = 'move_down_fast' ),
                
    KeyBinding( binding1    = 'Alt-Left', 
                description = 'Move selected components left faster',
                method_name = 'move_left_fast' ),
                
    KeyBinding( binding1    = 'Alt-Right', 
                description = 'Move selected components right faster',
                method_name = 'move_right_fast' ),
                
    KeyBinding( binding1    = 'Shift-Left', 
                description = 'Left align selected components',
                method_name = 'align_left' ),
                
    KeyBinding( binding1    = 'Shift-Right', 
                description = 'Right align selected components',
                method_name = 'align_right' ),
                
    KeyBinding( binding1    = 'Shift-Up', 
                description = 'Top align selected components',
                method_name = 'align_top' ),
                
    KeyBinding( binding1    = 'Shift-Down', 
                description = 'Bottom align selected components',
                method_name = 'align_bottom' ),
                
    KeyBinding( binding1    = '\\', 
                binding2    = '|',
                description = 'Align horizontal centers of selected components',
                method_name = 'align_horizontally_centered' ),
                
    KeyBinding( binding1    = '-', 
                binding2    = '_',
                description = 'Align vertical centers of selected components',
                method_name = 'align_vertically_centered' ),
                
    KeyBinding( binding1    = '"', 
                binding2    = '\'',
                description = 'Distribute selected components horizontally',
                method_name = 'distribute_horizontally' ),
                
    KeyBinding( binding1    = '=', 
                binding2    = '+',
                description = 'Distribute selected components vertically',
                method_name = 'distribute_vertically' ),
                
    KeyBinding( binding1    = '1', 
                description = 'Minimize selected components',
                method_name = 'collapse_selected_components' ),
                
    KeyBinding( binding1    = '2', 
                description = 'Normalize selected components',
                method_name = 'normalize_selected_components' ),
                
    KeyBinding( binding1    = '3', 
                description = 'Maximize selected components',
                method_name = 'expand_selected_components' ),
                
    KeyBinding( binding1    = '!', 
                description = 'Minimize all components',
                method_name = 'collapse_all_components' ),
                
    KeyBinding( binding1    = '@', 
                description = 'Normalize all components',
                method_name = 'normalize_all_components' ),
                
    KeyBinding( binding1    = '#', 
                description = 'Maximize all components',
                method_name = 'expand_all_components' ),
                
    KeyBinding( binding1    = 'Ctrl-Up',
                binding2    = 'Page Up',
                description = 'Increase all components one display level',
                method_name = 'expand_by_1_all_components' ),
                
    KeyBinding( binding1    = 'Ctrl-Down', 
                binding2    = 'Page Down',
                description = 'Decrease all components one display level',
                method_name = 'contract_by_1_all_components' ),
                
    KeyBinding( binding1    = 'Space', 
                binding2    = 'r',
                description = 'Execute the dataflow',
                method_name = 'run' ),
                
    KeyBinding( binding1    = 't', 
                binding2    = 'T',
                description = "Toggle 'auto run' mode",
                method_name = 'toggle_auto_run' ),
                
    KeyBinding( binding1    = 'e', 
                binding2    = 'E',
                description = "Edit a component's contacts",
                method_name = 'edit_component' ),
                
    KeyBinding( binding1    = 'd', 
                binding2    = 'D',
                description = 'Disconnect contact(s) from a components',
                method_name = 'disconnect_contact' ),
                
    KeyBinding( binding1    = 'Delete', 
                description = 'Delete selected components',
                method_name = 'delete_component' ),

    KeyBinding( binding1    = ',', 
                description = 'Condense selected components to simplify',
                method_name = 'collapse_composite_simple' ),
                
    KeyBinding( binding1    = '<', 
                description = 'Condense selected components to reuse',
                method_name = 'collapse_composite_full' ),
                
    KeyBinding( binding1    = '.', 
                binding2    = '>',
                description = 'Expand selected composites into their components',
                method_name = 'expand_composite_in_place' ),
                
    KeyBinding( binding1    = 'g', 
                binding2    = 'G',
                description = 'Toggle the canvas grid on or off',
                method_name = 'toggle_grid' ),
                
    KeyBinding( binding1    = 's', 
                binding2    = 'S',
                description = "Toggle canvas 'snap to grid' on or off",
                method_name = 'toggle_snap' ),
                
    KeyBinding( binding1    = 'b', 
                binding2    = 'B',
                description = 'Edit the keyboard bindings',
                method_name = 'edit_bindings' ),
                
    KeyBinding( binding1    = 'c', 
                binding2    = 'C',
                description = 'Add a comment',
                method_name = 'add_comment' ),
                
    suffix = '_key_action' )

#-------------------------------------------------------------------------------
#  'GraphCanvasStyle' class:  
#-------------------------------------------------------------------------------

class GraphCanvasStyle ( OMCanvasStyle ):
    
    menu = Instance( Menu )
    
    #---------------------------------------------------------------------------
    #  Override the implementation of the 'menu' property:  
    #---------------------------------------------------------------------------
    
    def _menu_default ( self ):
        return Menu(
            Action( name   = 'Add comment',
                    action = 'on_add_comment' ),
            Separator(),
            Menu( 
                Action( name   = 'Minimized',
                        action = 'on_collapse_view' ),
                Action( name   = 'Normal',
                        action = 'on_normal_view' ),
                Action( name   = 'Maximized',
                        action = 'on_expand_view' ),
                name = 'Show' 
            ),
            Menu( 
                toggle_grid_action,
                toggle_snap_action,
                Menu( name = 'Size', *grid_size_actions ),
                name = 'Grid' 
            ),
            Action( name   = 'Key bindings...',
                    action = 'on_key_bindings' ),
            auto_run_action,
            name = 'popup'
        )
        
graph_canvas_style = GraphCanvasStyle()        

#-------------------------------------------------------------------------------
#  'GraphCanvasController' class:
#-------------------------------------------------------------------------------

class GraphCanvasController ( OMCanvasController ):
    
    #---------------------------------------------------------------------------
    #  Trait Definitions:  
    #---------------------------------------------------------------------------
        
    # File the canvas is stored in:
    file_name = Str
    
    #---------------------------------------------------------------------------
    #  Runs the underlying graph model:  
    #---------------------------------------------------------------------------
        
    # fixme: Need to execute this method when changes to the canvas occur...
    def run ( self, canvas ):
        actor = canvas.data
        
        try:
            actor.preinitialize()
            actor.initialize()
            actor.iterate( 1 )
        except FireError, e:
            print e
        
        actor.wrapup()

    #---------------------------------------------------------------------------
    #  Returns whether or not the selection for a specified canvas can be 
    #  cleared:  
    #---------------------------------------------------------------------------
    
    #def can_clear_selection ( self, canvas ):
    #    return super( GraphCanvasController, self ).can_clear_selection( 
    #                                                                   canvas )
        
    #---------------------------------------------------------------------------
    #  Returns whether or not a specified canvas should allow drag selection:  
    #---------------------------------------------------------------------------
                
    #def can_drag_select ( self, canvas ):
    #    return super( GraphCanvasController, self ).can_drag_select( canvas )
        
    #---------------------------------------------------------------------------
    #  Called whenever a canvas's selection is changed:  
    #---------------------------------------------------------------------------
        
    #def selection_changed ( self, canvas ):
    #    super( GraphCanvasController, self ).selection_changed( canvas )
        
    #---------------------------------------------------------------------------
    #  Edit the properties of the specified canvas:  
    #---------------------------------------------------------------------------
    
    def edit_canvas ( self, canvas ):
        self.run( canvas )
        
    #---------------------------------------------------------------------------
    #  Returns whether or not a specified object can be dropped on the canvas:
    #---------------------------------------------------------------------------
    
    def can_drop ( self, canvas, object, event ):
        
        if isinstance(object, File) and object.path.endswith('.py'):
            # Check if this is registered as a chip
            if ChipManager().lookup_with_path(object.path) is not None:
                return True
        if (isinstance( object, GraphActor ) or 
                isinstance( object, LogSuite )):
            return True
                
        return False
    #---------------------------------------------------------------------------
    #  Handles a specified object being dropped on a canvas:
    #---------------------------------------------------------------------------
    
    def dropped ( self, canvas, object, event ):

        # FIXME: this method needs to be refactored.  There should really
        # FIXME: be a method on the canvas to add an object at some x,y
        
        uname = unique_actor_name( object.name, canvas.data )
        actor = None
        
        if isinstance( object, File ) and object.path.endswith('.py'):
            
            prototype = ChipManager().lookup_with_path(object.path)
            
            if prototype is not None:
                # Clone the prototype actor class:
                actor = prototype.instance.clone_traits( traits = 'all', 
                                                      copy   = 'deep' )
                
                # Function is not cloned, set it explicitly:
                if ((prototype.instance.function is not None) and 
                    (prototype.instance.function[0] is not None)):
                    actor.function = prototype.instance.function
    
                actor.chip_name = uname
                actor.name      = uname
                actor.container = canvas.data
            else:
                logger.info("Unable to find actor for %s" % object.path)
        else:
            if event.copy:
                required_cookie = ResourceSinkCookie
            else:
                required_cookie = ResourceSourceCookie
            
            cookie_mgr = CookieManager()
            
            cookie = cookie_mgr.get_cookie( required_cookie, object )
            if cookie is not None:
                actor = cookie.drop( name      = uname,
                                     resource  = object,
                                     chip_name = uname,
                                     container = canvas.data )
            else:
                raise ValueError, ("No ResourceSinkCookie for %s" % 
                                   object.name)
        
        component = None
        if actor is not None:    
            component        = self.create_component( actor )
            x, y, dx, dy     = component.bounds
            cx, cy, cdx, cdy = canvas.bounds
            component.origin = ( int( event.x - cx - (dx / 2) ), 
                                 int( event.y - cy - (dy / 2) ) )
            canvas.add( component )
            canvas.add_undo( GraphComponentUndoItem( canvas    = canvas,
                                                     component = component ) )
            canvas.extend_undo( GraphSelectionUndoItem(
                                    canvas        = canvas,
                                    new_selection = [ component ] ) )
            canvas.select_component( component )
            recompute_links( component, updated = True )
        
        return component
        
    #---------------------------------------------------------------------------
    #  Handles a keyboard event for a canvas:  
    #---------------------------------------------------------------------------
                
    def key_pressed ( self, canvas, event ):
        canvas.key_bindings.do( event.event, self, event, canvas )
        
    #---------------------------------------------------------------------------
    #  Components were dragged from one location to another:  
    #---------------------------------------------------------------------------
    
    def components_dragged ( self, canvas, components, original_bounds ):
        """ Components were dragged from one location to another.
        """
        canvas.add_undo( GraphMoveUndoItem( 
                             components = components,
                             old_bounds = original_bounds,
                             new_bounds = [ c.bounds for c in components ] ) )
        
    #---------------------------------------------------------------------------
    #  Components were selected on the canvas:  
    #---------------------------------------------------------------------------
                
    def components_selected ( self, canvas, components, toggle = False ):
        """ Components were selected on the canvas.
        """
        undo_item = GraphSelectionUndoItem( canvas = canvas )
        if super( GraphCanvasController, self ).components_selected( 
                      canvas, components, toggle ):
            canvas.add_undo( undo_item.finalize() )
        
    #---------------------------------------------------------------------------
    #  Displays the canvas's context menu:  
    #---------------------------------------------------------------------------
                        
    def popup_menu ( self, canvas, event ):
        auto_run_action.checked    = canvas.auto_run
        toggle_grid_action.checked = canvas.grid_visible
        toggle_snap_action.checked = canvas.snap_to_grid
        
        for action in grid_size_actions:
            action.checked = False
        grid_size_actions[ (canvas.grid_width / 16) - 1 ].checked = True
        
        super( GraphCanvasController, self ).popup_menu( canvas, event )
        
#-- Menu Handlers --------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Edit the currently selected component:    
    #---------------------------------------------------------------------------
    
    def on_edit ( self, canvas ):
        if len( canvas.selection ) == 1:
            component = canvas.selection[0] 
            component.controller.edit_component_action( component )
            
    #---------------------------------------------------------------------------
    #  Collapse/Expand the definition of a composite component:    
    #---------------------------------------------------------------------------
    
    def on_collapse_composite ( self, canvas, force_visible = False ):
        selection = canvas.selection
        if len( selection ) >= 2:
            self.create_composite_action( canvas, selection, 
                                          force_visible = force_visible )
                                          
    def on_expand_composite ( self, canvas ):
        composites = [ c for c in canvas.selection 
                       if isinstance( c.data, ConvergeCompositeActor ) ]
        if len( composites ) > 0:
            composites[0].controller.expand_composite_action( *composites )

    #---------------------------------------------------------------------------
    #  Change the expansion state of all chips on the canvas at once:  
    #---------------------------------------------------------------------------
    
    def on_collapse_view ( self, canvas = None ):
        self.expand_all_components_action( canvas or self._canvas, 0 )
    
    def on_normal_view ( self, canvas = None ):
        self.expand_all_components_action( canvas or self._canvas, 1 )
    
    def on_expand_view ( self, canvas = None ):
        self.expand_all_components_action( canvas or self._canvas, 2 )
        
    #---------------------------------------------------------------------------
    #  Toggles the canvas grid on or off:  
    #---------------------------------------------------------------------------
                
    def on_toggle_grid ( self ):
        """ Toggles the canvas grid on or off.
        """
        self.toggle_grid_action( self._canvas )
        
    #---------------------------------------------------------------------------
    #  Toggles the 'snap to grid' feature on or off:  
    #---------------------------------------------------------------------------
                
    def on_toggle_snap ( self ):
        """ Toggles the 'snap to grid' feature on or off.
        """
        self.toggle_snap_action( self._canvas )
        
    #---------------------------------------------------------------------------
    #  Set the grid size to various sizes:  
    #---------------------------------------------------------------------------
                
    def on_grid_size_16x16 ( self ):
        self.set_grid_size_action( self._canvas, 16, 16 )
                
    def on_grid_size_32x32 ( self ):
        self.set_grid_size_action( self._canvas, 32, 32 )
                
    def on_grid_size_48x48 ( self ):
        self.set_grid_size_action( self._canvas, 48, 48 )
                
    def on_grid_size_64x64 ( self ):
        self.set_grid_size_action( self._canvas, 64, 64 )
        
    #---------------------------------------------------------------------------
    #  Align the currently selected components in various ways:    
    #---------------------------------------------------------------------------
                
    def on_align_left ( self, canvas ):
        self.align_selected_components_action( canvas, 0 )
                
    def on_align_right ( self, canvas ):
        self.align_selected_components_action( canvas, 1 )
                
    def on_align_top ( self, canvas ):
        self.align_selected_components_action( canvas, 2 )
                
    def on_align_bottom ( self, canvas ):
        self.align_selected_components_action( canvas, 3 )
                
    def on_align_hcenter ( self, canvas ):
        self.align_selected_components_action( canvas, 4 )
                
    def on_align_vcenter ( self, canvas ):
        self.align_selected_components_action( canvas, 5 )
        
    #---------------------------------------------------------------------------
    #  Distribute selected components horizontally or vertically:    
    #---------------------------------------------------------------------------
                
    def on_distribute_horizontally ( self, canvas ):
        self.distribute_selected_components_action( canvas, 0 )
                
    def on_distribute_vertically ( self, canvas ):
        self.distribute_selected_components_action( canvas, 1 )
        
    #---------------------------------------------------------------------------
    #  Undo/Redo the most recent action:    
    #---------------------------------------------------------------------------
                
    def on_undo ( self, canvas ):
        canvas.undo()
        
    def on_redo ( self, canvas ):
        canvas.redo()
        
    #---------------------------------------------------------------------------
    #  Edits the current set of key bindings for the canvas:    
    #---------------------------------------------------------------------------
                
    def on_key_bindings ( self ):
        self._canvas.key_bindings.edit_traits()
        
    #---------------------------------------------------------------------------
    #  Toggles the state of the 'autorun' flag:  
    #---------------------------------------------------------------------------
                
    def on_autorun ( self ):
        """ Toggles the state of the 'autorun' flag.
        """
        self.toggle_auto_run_action( self._canvas )
        
    #---------------------------------------------------------------------------
    #  Adds a comment to the canvas:  
    #---------------------------------------------------------------------------
    
    def on_add_comment ( self ):
        """ Adds a comment to the canvas.
        """
        self.add_comment_action( self._canvas, self._event.x, self._event.y ) 
                
#-- Public Methods -------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Sets the current expansion state of all components to the specified 
    #  state:  
    #---------------------------------------------------------------------------
                
    def expand_all_components ( self, canvas, expansion ):
        """ Sets the current expansion state of all components to the specified 
            state.
        """
        self.expand_components( self.all_components_for( canvas ), expansion )
        
    #---------------------------------------------------------------------------
    #  Sets the current expansion state of a set of components to the specified 
    #  state:  
    #---------------------------------------------------------------------------
                
    def expand_components ( self, components, expansion ):
        """ Sets the current expansion state of a set of components to the 
            specified state.
        """
        if type( expansion ) is int:
            for component in components:
                self.expand_component( component, expansion )
        else:
            for i, component in enumerate( components ):
                self.expand_component( component, expansion[i] )
            
        for component in components:
            recompute_links( component, updated = False )
            
    #---------------------------------------------------------------------------
    #  Expand a component to the specified state:    
    #---------------------------------------------------------------------------
                        
    def expand_component ( self, component, expansion ): 
        controller = component.controller
        controller.set_expansion_state( component, expansion, update = False )
        controller.build_component( component )
 
    #---------------------------------------------------------------------------
    #  Expands all components by the minimum expansion state + 1 (if possible):  
    #---------------------------------------------------------------------------
                        
    def expand_by_1_all_components ( self, canvas ):
        """ Expands all components by the minimum expansion state + 1 (if 
            possible).
        """
        expansion = 3
        for component in self.all_components_for( canvas ):
            expansion = min( expansion, 
                             component.controller.get_expansion_state(
                                 component ) )
        if expansion <= 1:
            self.expand_all_components( canvas, expansion + 1 )
            return expansion + 1
            
        return None
 
    #---------------------------------------------------------------------------
    #  Contracts all components by the maximum expansion state - 1 (if 
    #  possible):  
    #---------------------------------------------------------------------------
                        
    def contract_by_1_all_components ( self, canvas ):
        """ Contracts all components by the maximum expansion state - 1 (if 
            possible).
        """
        expansion = -1
        for component in self.all_components_for( canvas ):
            expansion = max( expansion, 
                             component.controller.get_expansion_state(
                                 component ) )
        if expansion > 0:
            self.expand_all_components( canvas, expansion - 1 )
            return expansion - 1
            
        return None
            
    #---------------------------------------------------------------------------
    #  Moves a set of components by the specified horizontal and vertical 
    #  amounts:  
    #---------------------------------------------------------------------------
                        
    def move_components ( self, canvas, components, dx, dy ):
        """ Moves a set of components by the specified horizontal and vertical 
            amounts.
        """
        # Scale the move amount by the current grid_size:
        dx *= canvas.grid_width
        dy *= canvas.grid_height
            
        # Do a trial move of all components to find the maximum amount we can
        # move without bumping into an edge:
        for component in components:
            cx, cy, cdx, cdy = bounds = component.bounds
            self.move_component( component, dx, dy )
            nx, ny, cdx, cdy = component.bounds
            component.bounds = bounds
            dx = nx - cx
            dy = ny - cy
        
        # Now do the actual moved based on the maximum legal move amounts:
        for component in components:
            self.move_component( component, dx, dy )
            
        # Force the canvas to be redrawn (makes sure the links get redrawn):
        canvas.redraw()
            
    #---------------------------------------------------------------------------
    #  Moves a component by the specified horizontal and vertical amounts:  
    #---------------------------------------------------------------------------
                        
    def move_component ( self, component, dx, dy ):
        """ Moves a component by the specified horizontal and vertical amounts.
        """
        canvas           = component.container
        cx, cy, cdx, cdy = component.bounds
        component.bounds = ( cx + dx, cy + dy, cdx, cdy )
        canvas._check_bounds( component )
        
    #---------------------------------------------------------------------------
    #  Selects the next component on the canvas:  
    #---------------------------------------------------------------------------
                
    def select_next_component ( self, canvas, is_next ):
        """ Selects the next component on the canvas.
        """
        components = self.all_components_for( canvas )
        n          = len( components )
        if n > 0:
            components.sort( geometric_compare ) 
            if len( canvas.selection ) != 1:
                component = components[ is_next - 1 ]
            else:
                index     = components.index( canvas.selection[0] )
                component = components[ (index - 1 + 2 * is_next) % n ]
            canvas.select_component( component )
            
    #---------------------------------------------------------------------------
    #  Aligns all selected components:   
    #---------------------------------------------------------------------------
                        
    def align_selected_components ( self, canvas, align_type ):
        """ Aligns all selected components.
        """
        components = canvas.selection
        if len( components ) > 1:
            ac = components[0]
            for component in components[1:]:
                if align_type == 0:
                    component.x = ac.x
                elif align_type == 1:
                    component.x = ac.right - component.width
                elif align_type == 2:
                    component.y = ac.top - component.height
                elif align_type == 3:
                    component.y = ac.y
                elif align_type == 4:
                    component.x = ac.x - (component.width - ac.width) / 2.0
                elif align_type == 5:
                    component.y = ac.y - (component.height - ac.height) / 2.0
                canvas._check_bounds( component )
            canvas.redraw()
            
    #---------------------------------------------------------------------------
    #  Distribute all selected components:   
    #---------------------------------------------------------------------------
                        
    def distribute_selected_components ( self, canvas, distribution_type ):
        """ Distributes all selected components.
        """
        components = canvas.selection
        if len( components ) > 1:
            if distribution_type == 0:
                self.distribute_components_horizontally( canvas, components )
            else:
                self.distribute_components_vertically(   canvas, components )
            for component in components:
                canvas._check_bounds( component )
            canvas.redraw()
            
    #---------------------------------------------------------------------------
    #  Distribute a specified set of components vertically:  
    #---------------------------------------------------------------------------
                        
    def distribute_components_vertically ( self, canvas, components ):
        """ Distribute a specified set of components vertically.
        """
        n            = len( components )
        min_y        =  999999
        max_y        = -999999
        total_height = 0
        
        for component in components:
            x, y          = component.origin
            min_y         = min( min_y, y )
            max_y         = max( max_y, y + component.height )
            total_height += component.height
            
        avail_height  = max_y - min_y
        needed_height = total_height + ((n - 1) * canvas.grid_height)
        if needed_height > avail_height:
            delta        = needed_height - avail_height
            avail_height = needed_height
            min_y        = max( 0, min_y - (delta / 2) )
            
        components.sort( vertical_compare )
        delta = float( avail_height - total_height ) / (n - 1)
        
        for component in components:
            x, y             = component.origin
            component.origin = ( x, int( min_y ) )
            min_y           += component.height + delta
            
    #---------------------------------------------------------------------------
    #  Distribute a specified set of components horizontally:  
    #---------------------------------------------------------------------------
                        
    def distribute_components_horizontally ( self, canvas, components ):
        """ Distribute a specified set of components horizontally.
        """
        n           = len( components )
        min_x       =  999999
        max_x       = -999999
        total_width = 0
        
        for component in components:
            x, y         = component.origin
            min_x        = min( min_x, x )
            max_x        = max( max_x, x + component.width )
            total_width += component.width
            
        avail_width  = max_x - min_x
        needed_width = total_width + ((n - 1) * canvas.grid_width)
        if needed_width > avail_width:
            delta       = needed_width - avail_width
            avail_width = needed_width
            min_x       = max( 0, min_x - (delta / 2) )
            
        components.sort( horizontal_compare )
        delta = float( avail_width - total_width ) / (n - 1)
        
        for component in components:
            x, y             = component.origin
            component.origin = ( int( min_x ), y )
            min_x           += component.width + delta
            
    #---------------------------------------------------------------------------
    #  Toggles the grid for a specified canvas on or off:  
    #---------------------------------------------------------------------------
                        
    def toggle_grid ( self, canvas ):
        """ Toggles the grid for a specified canvas on or off.
        """
        canvas.grid_visible = not canvas.grid_visible
        canvas.redraw()
            
    #---------------------------------------------------------------------------
    #  Toggles the 'snap to grid' feature for a specified canvas on or off:  
    #---------------------------------------------------------------------------
                        
    def toggle_snap ( self, canvas ):
        """ Toggles the 'snap to grid' feature for a specified canvas on or off.
        """
        canvas.snap_to_grid = not canvas.snap_to_grid
        
    #---------------------------------------------------------------------------
    #  Set a canvas's grid size to a specified size:  
    #---------------------------------------------------------------------------
                
    def set_grid_size ( self, canvas, width, height ):
        """ Set a canvas's grid size to a specified size.
        """
        canvas.grid_width  = width
        canvas.grid_height = height
        canvas.redraw()
        
    #---------------------------------------------------------------------------
    #  Creates a composite component from a set of components:  
    #---------------------------------------------------------------------------
                
    def create_composite ( self, canvas, components, force_visible = False ):
        """ Creates a composite component from a set of components.
        """
        # Create a unique name for the composite:
        name = unique_actor_name( 'Composite', canvas.data )

        # Create the new composite:
        composite = ConvergeCompositeActor( 
                        name      = name,
                        chip_name = name,
                        help      = composite_help % ('\t'.join(
                                        [ ('  - ' + c.inner_component.text) 
                                          for c in components ] )) )
        
        # Add it to the same composite actor the components belong to:
        composite.container = canvas_actor = components[0].data.container
        
        # Move all of the components into the new composite:
        input_bundles                 = []
        output_bundles                = []
        composite_input_ports         = []
        composite_unbound_input_ports = []
        composite_output_ports        = []
        actors                        = [ c.data for c in components ]
        for component in components:
            cname           = component.inner_component.text
            actor           = component.data
            actor.container = composite
            input_ports     = actor.input_ports
            output_ports    = actor.output_ports
            
            for bundle in actor.input_bundles:
                cips = []
                for port in bundle.ports:
                    actor_port = input_ports[ port.index ]
                    include    = True
                    relation   = None
                    for relation in actor_port.relations_list():
                        for other_actor_port in relation.linked_ports( 
                                                                   actor_port ):
                            if other_actor_port.container in actors:
                                include = False
                                
                    if include:
                        new_meta_port = port.clone_traits( traits = 'all' )
                        new_meta_port.name = '%s %s' % ( cname, port.name )
                        new_meta_port.index = len( composite_input_ports )
                        cips.append( new_meta_port )
                              
                        nip = composite.new_port( new_port_name(), True, new_meta_port)
                        composite_input_ports.append( nip )
                        
                        if relation is None:
                            # Add the port to the unbound port list:
                            composite_unbound_input_ports.append( nip )
                            actor.bind_input_port(  actor_port )
                        else:
                            # Discard the old relation:
                            relation.container = None
                        
                            # Connect the old port via the new port:
                            connect_ports( canvas_actor, other_actor_port, nip )
                        
                        # Connect the composite port to the inner actor port:
                        connect_ports( composite, nip, actor_port )
                        
                # If any input ports were created, then add their corresponding
                # input bundle:
                if len( cips ) > 0:
                    input_bundles.append( ConvergeInputBundle( 
                        name       = '%s %s' % ( cname, bundle.name ),
                        group_name = bundle.group_name,
                        ports      = cips ) )
                        
            for bundle in actor.output_bundles:
                cops = []
                for port in bundle.ports:
                    
                    actor_port = output_ports[ port.index ]
                    
                    cop = port.clone_traits( traits='all' )
                    cop.name = '%s %s' % ( cname, port.name )
                    cop.index = len( composite_output_ports )
                    cop.visible = force_visible
                    cops.append( cop )
                    
                    nop = composite.new_port( new_port_name(), False, cop)
                    composite_output_ports.append( nop )
                    
                    for relation in actor_port.relations_list():
                        for other_actor_port in relation.linked_ports( 
                                                               actor_port ):
                            # If there is a connection to a component outside
                            # of the new composite:
                            if other_actor_port.container not in actors:
                                # Mark the port as visible:
                                cop.visible = True
                                
                                # Discard the old relation:
                                relation.container = None
                                
                                # Connect the old ports via the new port:
                                connect_ports( canvas_actor, nop, 
                                               other_actor_port )
                    
                    # Connect the inner actor output port to the composite port:
                    connect_ports( composite, actor_port, nop )
                        
                # Add the corresponding output bundle (if necessary):
                output_bundles.append( ConvergeOutputBundle( 
                        name       = '%s %s' % ( cname, bundle.name ),
                        group_name = bundle.group_name,
                        ports      = cops ) )
        
        # Save the computed bundles and ports as part of the composite actor:
        composite.unbound_input_ports     = composite_unbound_input_ports
        composite.input_ports             = composite_input_ports
        composite.output_ports            = composite_output_ports
        composite.instance_input_bundles  = input_bundles
        composite.instance_output_bundles = output_bundles
        
        # Compute the 'center of gravity' of the old components and remove
        # any links connecting the old components:
        cx = cy = 0
        for component in components:
            x, y = component.origin
            cx  += x + (int( component.width ) / 2)
            cy  += y + (int( component.height ) / 2)
            disconnect_links( component )
        n  = len( components )
        cx = int( cx / n )
        cy = int( cy / n )
        
        # Attach the relative distance from each component's center to the new
        # composite's center to the actor as an aid when expanding the
        # composite back into individual components:
        for component in components:
            x, y = component.origin
            component.data._canvas_origin = ( 
                x + (int( component.width )  / 2) - cx,
                y + (int( component.height ) / 2) - cy )
            
        # Remove the old components from the canvas:
        canvas.remove( *components )
        
        # Remove any links that are no longer needed from the canvas:
        canvas.remove( *[ link for link in canvas.components
                          if isinstance( link, OMLink )              and 
                             (link.contact1.container in components) and
                             (link.contact2.container in components) ] )
        
        # Create a component for the newly created composite actor:
        component = self.create_component( composite )
        
        # Set its initial position on the canvas:
        component.origin = ( cx - (component.width  / 2), 
                             cy - (component.height / 2) )
        
        # Add the new composite component to the canvas, select it and add the
        # links back in:
        canvas.add( component )
        canvas.select_component( component )
        recompute_links( component, updated = True )
        
        # Return the composite component that was created:
        return component
        
    #---------------------------------------------------------------------------
    #  Generates all components on the specified canvas:  
    #---------------------------------------------------------------------------
                        
    def all_components_for ( self, canvas ):
        """ Generates all components on the specified canvas.
        """
        return [ c for c in canvas.components if isinstance( c, OMComponent ) ]

#-- Undoable actions -----------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Toggles the state of the 'autorun' flag: 
    #---------------------------------------------------------------------------

    def toggle_auto_run_action ( self, canvas ):
        """ Toggles the state of the 'autorun' flag.
        """
        canvas.add_undo( GraphUndoItem( object    = canvas,
                                        name      = 'auto_run',
                                        old_value = canvas.auto_run,
                                        new_value = not canvas.auto_run ) )
        canvas.auto_run = not canvas.auto_run

    #---------------------------------------------------------------------------
    #  Undoable action that toggles the grid for a specified canvas on or off:  
    #---------------------------------------------------------------------------
                        
    def toggle_grid_action ( self, canvas ):
        """ Undoable action that toggles the grid for a specified canvas on or off.
        """
        canvas.add_undo(
            GraphUndoItem( object    = canvas, 
                           name      = 'grid_visible', 
                           old_value = canvas.grid_visible,
                           new_value = not canvas.grid_visible ) )
        self.toggle_grid( canvas )
            
    #---------------------------------------------------------------------------
    #  Undoable action that toggles the 'snap to grid' feature for a specified 
    #  canvas on or off:  
    #---------------------------------------------------------------------------
                        
    def toggle_snap_action ( self, canvas ):
        """ Undoable action that toggles the 'snap to grid' feature for a 
            specified canvas on or off.
        """
        canvas.add_undo( 
            GraphUndoItem( object    = canvas, 
                           name      = 'snap_to_grid', 
                           old_value = canvas.snap_to_grid,
                           new_value = not canvas.snap_to_grid ) )
        self.toggle_snap( canvas )
        
    #---------------------------------------------------------------------------
    #  Undoable action that sets a canvas's grid size to a specified size:  
    #---------------------------------------------------------------------------
                
    def set_grid_size_action ( self, canvas, width, height ):
        """ Undoable action that sets a canvas's grid size to a specified size.
        """
        """ Set a canvas's grid size to a specified size.
        """
        canvas.add_undo( 
            GraphUndoItem( object    = canvas, 
                           name      = 'grid_width', 
                           old_value = canvas.grid_width,
                           new_value = width ) )
        canvas.extend_undo( 
            GraphUndoItem( object    = canvas, 
                           name      = 'grid_height', 
                           old_value = canvas.grid_height,
                           new_value = height ) )
        self.set_grid_size( canvas, width, height )
            
    #---------------------------------------------------------------------------
    #  Undoable action that aligns all selected components:   
    #---------------------------------------------------------------------------
                        
    def align_selected_components_action ( self, canvas, align_type ):
        """ Aligns all selected components.
        """
        undo_item = GraphBoundsUndoItem( components = canvas.selection )
        self.align_selected_components( canvas, align_type )
        canvas.add_undo( undo_item.finalize() )
        
    #---------------------------------------------------------------------------
    #  Undoable action that distributes the currently selected components:  
    #---------------------------------------------------------------------------
                
    def distribute_selected_components_action ( self, canvas, 
                                                distribution_type ):
        """ Undoable action that distributes the currently selected components.
        """
        undo_item = GraphBoundsUndoItem( components = canvas.selection )
        self.distribute_selected_components( canvas, distribution_type )
        canvas.add_undo( undo_item.finalize() )
            
    #---------------------------------------------------------------------------
    #  Undoable action that moves all components by the specified horizontal and 
    #  vertical amounts:  
    #---------------------------------------------------------------------------
                        
    def move_all_components_action ( self, canvas, dx, dy ):
        """ Undoable action that moves all components by the specified 
            horizontal and vertical amounts.
        """
        # Move the selection if there is one; otherwise move all the components:
        if len( canvas.selection ) > 0:
            components = canvas.selection
        else:
            components = self.all_components_for( canvas )
            
        undo_item = GraphMoveUndoItem( components = components )
        self.move_components( canvas, components, dx, dy )
        canvas.add_undo( undo_item.finalize() )
        
    #---------------------------------------------------------------------------
    #  Undoable action for expanding a set of components to a specified state:    
    #---------------------------------------------------------------------------
                
    def expand_components_action ( self, components, expansion ):
        """ Undoable action for expanding a set of components to a specified 
            state.
        """
        if len( components ) > 0:
            canvas = components[0].container
            canvas.add_undo( GraphExpansionUndoItem( components = components, 
                                                     expansion  = expansion ) )
            self.expand_components( components, expansion )
        
    #---------------------------------------------------------------------------
    #  Undoable action for expanding all components to a specified state:    
    #---------------------------------------------------------------------------
                
    def expand_all_components_action ( self, canvas, expansion ):
        """ Undoable action for expanding all components to a specified state.
        """
        canvas.add_undo( GraphExpansionUndoItem( 
                 components = self.all_components_for( canvas ), 
                 expansion  = expansion ) )
        self.expand_all_components( canvas, expansion )
        
    #---------------------------------------------------------------------------
    #  Undoable action for expanding all components by 1 expansion state:
    #---------------------------------------------------------------------------
                
    def expand_by_1_all_components_action ( self, canvas ):
        """ Undoable action for expanding all components by 1 expansion state.
        """
        undo_item = GraphExpansionUndoItem( 
                        components = self.all_components_for( canvas ) )
        expansion = self.expand_by_1_all_components( canvas )
        if expansion is not None:
            canvas.add_undo( undo_item.set( expansion = expansion ) )
        
    #---------------------------------------------------------------------------
    #  Undoable action for contracting all components by 1 expansion state:
    #---------------------------------------------------------------------------
                
    def contract_by_1_all_components_action ( self, canvas ):
        """ Undoable action for contracting all components by 1 expansion state.
        """
        undo_item = GraphExpansionUndoItem( 
                        components = self.all_components_for( canvas ) )
        expansion = self.contract_by_1_all_components( canvas )
        if expansion is not None:
            canvas.add_undo( undo_item.set( expansion = expansion ) )
            
    #---------------------------------------------------------------------------
    #  Undoable action that selects all components on the canvas:    
    #---------------------------------------------------------------------------
                        
    def select_all_action ( self, canvas ):
        """ Undoable action that selects all components on the canvas.
        """
        undo_item = GraphSelectionUndoItem( canvas = canvas )
        canvas.select_all()
        canvas.add_undo( undo_item.finalize() )
             
    #---------------------------------------------------------------------------
    #  Undoable action that selects a set of components on the canvas:    
    #---------------------------------------------------------------------------
                        
    def select_components_action ( self, canvas, components ):
        """ Undoable action that selects a set of components on the canvas.
        """
        canvas.add_undo( GraphSelectionUndoItem( canvas        = canvas, 
                                                 new_selection = components ) )
        canvas.select_components( components )
        
    #---------------------------------------------------------------------------
    #  Undoable action that selects the next component on the canvas:  
    #---------------------------------------------------------------------------
                
    def select_next_component_action ( self, canvas, is_next ):
        """ Undoable action that selects the next component on the canvas.
        """
        undo_item = GraphSelectionUndoItem( canvas = canvas )
        self.select_next_component( canvas, is_next )
        canvas.add_undo( undo_item.finalize() )
            
    #---------------------------------------------------------------------------
    #  Undoable action that unselects all components on the canvas:    
    #---------------------------------------------------------------------------
                        
    def clear_selection_action ( self, canvas ):
        """ Undoable action that unselects all components on the canvas.
        """
        undo_item = GraphSelectionUndoItem( canvas = canvas )
        canvas.clear_selection()
        canvas.add_undo( undo_item.finalize() )
        
    #---------------------------------------------------------------------------
    #  Undoable action that creates a composite component from a set of 
    #  components:  
    #---------------------------------------------------------------------------
                
    def create_composite_action ( self, canvas, components, 
                                        force_visible = False ):
        """ Undoable action that creates a composite component from a set of
            components.
        """
        undo_item = GraphCreateCompositeUndoItem( canvas     = canvas, 
                                                  components = components )
                        #force_visible = force_visible )
        undo_item.composite = self.create_composite( canvas, components, 
                                                     force_visible )
        canvas.add_undo( undo_item )

    #---------------------------------------------------------------------------
    #  Undoable action that deletes all currently selected components from the 
    #  canvas:  
    #---------------------------------------------------------------------------

    def delete_action ( self, canvas ):
        """ Undoable action that deletes all currently selected components from
            the canvas.
        """
        selection = canvas.selection
        if len( selection ) > 0:
            connections = []
            for component in selection:
                connections.extend( component.controller.delete( component ) )
                
            canvas.add_undo( GraphConnectionUndoItem( 
                                 canvas        = canvas,
                                 connections   = connections,
                                 is_connection = False ) )
                                 
            canvas.extend_undo( GraphSelectionUndoItem( canvas        = canvas,
                                                        new_selection = [] ) )
                                    
            for component in selection:
                canvas.extend_undo( GraphComponentUndoItem( 
                                        canvas    = canvas,
                                        component = component,
                                        is_add    = False ) )
            del selection[:]
            canvas.updated = True
            canvas.run     = True
            
    #---------------------------------------------------------------------------
    #  Undoable action that adds a new comment to the canvas at a specified 
    #  point:  
    #---------------------------------------------------------------------------
    
    def add_comment_action ( self, canvas, x, y ):
        """ Undoable action that adds a new comment to the canvas at a specified
            point.
        """
        uname = unique_actor_name( 'comment', canvas.data )
        actor = ConvergeCommentActor( chip_name = uname,
                                      name      = uname,
                                      container = canvas.data )
        component        = self.create_component( actor )
        x0, y0, dx, dy   = component.bounds
        cx, cy, cdx, cdy = canvas.bounds
        component.origin = ( int( x - cx - (dx / 2) ), 
                             int( y - cy - (dy / 2) ) )
           
        canvas.add( component )
        ignore, connections = graph_comment_controller.edit_component( 
                                                                     component )
                                                                     
        # fixme: There's a wierd interaction in the 'edit_traits' method
        # invoked by 'edit_component' that causes the 'name' trait to get set
        # to 'unknown', which the following line fixes by setting it to the 
        # correct value again.
        actor.name = uname
        
        if connections is None:
            canvas.remove( component )
            canvas.redraw()
            actor.container = None
            return
            
        canvas.add_undo( GraphComponentUndoItem( canvas    = canvas,
                                                 component = component ) )
        canvas.extend_undo( GraphSelectionUndoItem(
                                canvas        = canvas,
                                new_selection = [ component ] ) )
        canvas.select_component( component )
            
                    
#-- Key bindable actions -------------------------------------------------------

    def collapse_composite_simple_key_action ( self, event, canvas ):
        self.on_collapse_composite( canvas, force_visible = False )

    def collapse_composite_full_key_action ( self, event, canvas ):
        self.on_collapse_composite( canvas, force_visible = True )
        
    def expand_composite_in_place_key_action ( self, event, canvas ):        
        if not self._on_component(  canvas, event, 'expand_composite_action' ):
            self.on_expand_composite( canvas )
            
    def align_left_key_action ( self, event, canvas ):            
        self.align_selected_components_action( canvas, 0 ) 
            
    def align_right_key_action ( self, event, canvas ):            
        self.align_selected_components_action( canvas, 1 ) 
            
    def align_top_key_action ( self, event, canvas ):            
        self.align_selected_components_action( canvas, 2 ) 
            
    def align_bottom_key_action ( self, event, canvas ):            
        self.align_selected_components_action( canvas, 3 ) 
            
    def align_horizontally_centered_key_action ( self, event, canvas ):            
        self.align_selected_components_action( canvas, 4 ) 
            
    def align_vertically_centered_key_action ( self, event, canvas ):            
        self.align_selected_components_action( canvas, 5 )
        
    def distribute_horizontally_key_action ( self, event, canvas ):
        self.distribute_selected_components_action( canvas, 0 )
        
    def distribute_vertically_key_action ( self, event, canvas ):
        self.distribute_selected_components_action( canvas, 1 )
            
    def select_next_component_key_action ( self, event, canvas ):
        self.select_next_component_action( canvas, True )
            
    def select_previous_component_key_action ( self, event, canvas ):
        self.select_next_component_action( canvas, False )
        
    def select_all_key_action ( self, event, canvas ):
        self.select_all_action( canvas )
        
    def clear_selection_key_action ( self, event, canvas ):
        self.clear_selection_action( canvas )
        
    def undo_key_action ( self, event, canvas ):
        canvas.undo()
        
    def redo_key_action ( self, event, canvas ):
        canvas.redo()
        
    def run_key_action ( self, event, canvas ):
        self.run( canvas )
        
    def toggle_auto_run_key_action ( self, event, canvas ):        
        self.toggle_auto_run_action( canvas )
        
    def toggle_grid_key_action ( self, event, canvas ):        
        self.toggle_grid_action( canvas )
        
    def toggle_snap_key_action ( self, event, canvas ):        
        self.toggle_snap_action( canvas )
        
    def expand_by_1_all_components_key_action ( self, event, canvas ):
        self.expand_by_1_all_components_action( canvas )
        
    def contract_by_1_all_components_key_action ( self, event, canvas ):
        self.contract_by_1_all_components_action( canvas )
        
    def collapse_all_components_key_action ( self, event, canvas ):
        self.expand_all_components_action( canvas, 0 )
        
    def normalize_all_components_key_action ( self, event, canvas ):
        self.expand_all_components_action( canvas, 1 )
        
    def expand_all_components_key_action ( self, event, canvas ):
        self.expand_all_components_action( canvas, 2 )
        
    def collapse_selected_components_key_action ( self, event, canvas, 
                                                  expansion = 0 ):
        if not self._on_component( canvas, event, 'expand_component_action', 
                                   expansion ):
            self.expand_components_action( canvas.selection, expansion )
            
    def normalize_selected_components_key_action ( self, event, canvas ):
        self.collapse_selected_components_key_action( event, canvas, 1 )
            
    def expand_selected_components_key_action ( self, event, canvas ):
        self.collapse_selected_components_key_action( event, canvas, 2 )
        
    def edit_component_key_action ( self, event, canvas ):
        if ((not self._on_contact( canvas, event, 'edit_contact_action' )) and
            (not self._on_component( canvas, event, 'edit_component_action' ))):
            self.on_edit( canvas )
        
    def disconnect_contact_key_action ( self, event, canvas ):
        if not self._on_contact( canvas, event, 'disconnect_contact_action' ):
            self._on_component(  canvas, event, 'disconnect_contacts_action' )
            
    def delete_component_key_action ( self, event, canvas ):
        if not self._on_component( canvas, event, 'delete_component_action' ):
            self.delete_action( canvas )
            
    def move_up_slow_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, 0, 1 )
            
    def move_down_slow_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, 0, -1 )
            
    def move_left_slow_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, -1, 0 )
            
    def move_right_slow_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, 1, 0 )
            
    def move_up_fast_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, 0, 5 )
            
    def move_down_fast_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, 0, -5 )
            
    def move_left_fast_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, -5, 0 )
            
    def move_right_fast_key_action ( self, event, canvas ):
        self.move_all_components_action( canvas, 5, 0 )
            
    def edit_bindings_key_action ( self, event, canvas ):
        canvas.key_bindings.edit_traits()
            
    def add_comment_key_action ( self, event, canvas ):
        self.add_comment_action( canvas, event.x, event.y )
        
#-- Private Methods ------------------------------------------------------------
        
    #---------------------------------------------------------------------------
    #  Creates a component:  
    #---------------------------------------------------------------------------
                    
    def create_component ( self, actor ):
        """ Creates a component.
        """
        if isinstance( actor, ConvergeCommentActor ):
            # Build a 'comment' component:
            component = OMComponent( 
                style           = graph_component_style,
                outer_component = OMOuterComponent( 
                                       style = graph_comment_style ),
                inner_component = OMLabel( font          = 'Arial 8',
                                           text_position = 'top left' ),
                controller      = graph_comment_controller,
                contact_manager = OMTrackContactManager(),
                data            = actor )
        else:    
            # Build an 'actor' component:
            
            # Set the correct label border color based upon whether or not it 
            # is a  composite actor:
            border_color = ( 0.0, 0.0, 0.0, 1.0 )
            if isinstance( actor, ConvergeCompositeActor ):
                border_color = ( 1.0, 0.0, 0.0, 1.0 )
            
            component  = GraphComponent( 
                style           = graph_component_style,
                outer_component = OMOuterComponent( style = graph_actor_style ),
                inner_component = OMLabel( font           = 'Arial 8',
                                           text_position  = 'top',
                                           bg_color       = NominalColor,
                                           border_color   = border_color,
                                           border_size    = 1,
                                           margin_top     = 4 ),
                controller      = graph_component_controller,
                contact_manager = OMTrackContactManager(),
                data            = actor )
            
        # Finish building the component:                
        component.controller.build_component( component )
        set_component_style( component )
        
        return component
        
    #---------------------------------------------------------------------------
    #  Performs a requested action if the pointer is over a component:  
    #---------------------------------------------------------------------------
                
    def _on_component ( self, canvas, event, method_name, *args ):
        """ Performs a requested action if the pointer is over a component.
        """
        components = canvas.components_at( event.x, event.y )  
        for i in range( len( components ) - 1, -1, -1 ):
            c = components[i]
            if isinstance( c, OMComponent ):
                getattr( c.controller, method_name )( *( ( c, ) + args )  )
                return True
                
        return False
        
    #---------------------------------------------------------------------------
    #  Performs a requested action if the pointer is over a contact:  
    #---------------------------------------------------------------------------
                
    def _on_contact ( self, canvas, event, method_name, *args ):
        """ Performs a requested action if the pointer is over a contact.
        """
        components = canvas.components_at( event.x, event.y )  
        for i in range( len( components ) - 1, -1, -1 ):
            c = components[i]
            if isinstance( c, OMContact ):
                getattr( c.controller, method_name )( *( ( c, ) + args )  )
                return True
                
        return False          
                
# Standard canvas controller:
graph_canvas_controller = GraphCanvasController()

#-------------------------------------------------------------------------------
#  Compares the position of two components geometrically:  
#-------------------------------------------------------------------------------

def geometric_compare ( c1, c2 ):
    """ Compares the position of two components geometrically.
    """
    cx1, cy1, cdx1, cdy1 = c1.bounds
    cx2, cy2, cdx2, cdy2 = c2.bounds
    cmpy = cmp( cy2 + cdy2 / 2.0, cy1 + cdy1 / 2.0 )
    if cmpy != 0:
        return cmpy
    return cmp( cx1 + cdx1 / 2.0, cx2 + cdx2 / 2.0 )
    
#-------------------------------------------------------------------------------
#  Compares the vertical position of two components:  
#-------------------------------------------------------------------------------
        
def vertical_compare ( c1, c2 ):
    cmp_top = cmp( c1.top, c2.top )
    if cmp_top != 0:
        return cmp_top
    return cmp( c1.bottom, c2.bottom )
    
#-------------------------------------------------------------------------------
#  Compares the horizontal position of two components:  
#-------------------------------------------------------------------------------
        
def horizontal_compare ( c1, c2 ):
    cmp_left = cmp( c1.left, c2.left )
    if cmp_left != 0:
        return cmp_left
    return cmp( c1.right, c2.right )
    
