#-------------------------------------------------------------------------------
#  
#  Defines the GraphComponentController subclass of the Enable 'om' 
#  OMComponentController class for managing 'actor chips'.
#  
#  Written by: David C. Morrill
#  
#  Date: 03/14/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

from converge_actor \
    import ConvergeInputBundle, ConvergeOutputBundle, ConvergeRepeatBundle, \
           ConvergeRepeatButtonBundle, ConvergeCompositeActor
           
from enthought.util.wx.drag_and_drop \
    import PythonDropSource
    
from enthought.enable.om \
    import OMComponentController, OMComponentStyle, OMLink, OMTrack, OMContact,\
           OMContactStyle, OMDragContact, OMLabel, OMTrackGroupStyle, \
           default_track_group_style
           
from enthought.envisage.project \
    import CookieManager
    
from enthought.envisage.project.action.open_cookie \
    import OpenCookie
    
from enthought.envisage.ui \
    import UIPlugin
    
from enthought.naming.binding \
    import Binding
    
from enthought.io.file \
    import File
           
from graph_helper \
    import find_port_contact, set_contact_state, recompute_links, \
           auto_connect, auto_connectable, get_components, CollapsedColor, \
           NominalColor, ExpandedColor, GroupColor, disconnect_links, \
           connect_ports, unique_actor_name
    
from graph_contact_controller \
    import graph_contact_controller, graph_contact_style, repeat_contact_style
    
from graph_undo_item \
    import GraphExpandCompositeUndoItem, GraphConnectionUndoItem, \
           GraphComponentUndoItem, GraphEditComponentUndoItem
    
from repeat_contact_controller \
    import repeat_contact_controller
    
from enthought.traits.ui.menu \
    import Menu, Action, Separator
    
#-------------------------------------------------------------------------------
#  Constants:  
#-------------------------------------------------------------------------------

# Component menu actions:
show_actions = ( Action( name   = 'Minimized',
                         action = 'on_collapse_view',
                         style  = 'toggle' ),
                 Action( name   = 'Normal',
                         action = 'on_normal_view',
                         style  = 'toggle' ),
                 Action( name   = 'Maximized',
                         action = 'on_expand_view',
                         style  = 'toggle' ) )
                         
alignment_actions = ( Action( name   = 'Left',
                              action = 'on_align_left' ),
                      Action( name   = 'Right',
                              action = 'on_align_right' ),
                      Action( name   = 'Horizontally centered',
                              action = 'on_align_hcenter' ),
                      Action( name   = 'Top',
                              action = 'on_align_top' ),
                      Action( name   = 'Bottom',
                              action = 'on_align_bottom' ),
                      Action( name   = 'Vertically centered',
                              action = 'on_align_vcenter' ) )
                              
distribution_actions = ( Action( name   = 'Horizontally',
                                 action = 'on_distribute_horizontally' ),
                         Action( name   = 'Vertically',
                                 action = 'on_distribute_vertically' ) )
                         
create_simple_composite_action = Action( name   = 'Condense', ###'To simplify',
                                         action = 'on_create_simple_composite' )
                         
###create_full_composite_action   = Action( name   = 'To re-use',
###                                         action = 'on_create_full_composite' )

expand_in_place_action         = Action( name   = 'In place',
                                         action = 'on_expand_in_place' )
                                     
expand_on_new_canvas_action    = Action( name   = 'On new canvas',
                                         action = 'on_expand_on_new_canvas' )
                          
# Popup menu for a component (i.e. chip):
component_menu = Menu( 
    Menu( name = 'Show', *show_actions ),
    Menu( name = 'Align', 
          *(alignment_actions[:3] + ( Separator(), ) + alignment_actions[3:]) ),
    Menu( name = 'Distribute', *distribution_actions ),
    Action( name   = 'Edit...',
            action = 'on_edit_connections' ),
    Action( name   = 'Edit chip...',
            action = 'on_edit_chip' ),
    Separator(),
    create_simple_composite_action,
### Menu( create_simple_composite_action,
###       create_full_composite_action,
###       name = 'Condense' ),
    Menu( expand_in_place_action,
          expand_on_new_canvas_action,
          name = 'Expand' ),
    Separator(),
    Action( name   = 'Disconnect',
            action = 'disconnect_contacts_action' ),
    Separator(),
    Action( name   = 'Delete',
            action = 'on_delete' ),
    name = 'popup'
)
                        
input_contact_image  = '=contact_black'                        
repeat_contact_image = '=contact_add_0_6'                        
output_contact_image = '=contact_black'  

# Track group style for chips in the 'expanded' state:
expanded_track_group_style = OMTrackGroupStyle(
                                 bg_color     = GroupColor,
                                 border_color = ( 0.0, 0.0, 0.0, 1.0 ),
                                 padding      = 4, 
                                 visible      = True )
                                 
#-------------------------------------------------------------------------------
#  'GraphComponentStyle' class:  
#-------------------------------------------------------------------------------
                               
class GraphComponentStyle ( OMComponentStyle ):
    
    #---------------------------------------------------------------------------
    #  Override of the 'menu' property:  
    #---------------------------------------------------------------------------
        
    def get_menu ( self ):
        return component_menu

# 'Chip' component style:
graph_component_style = GraphComponentStyle( inner_component_position = 'top' )

#-------------------------------------------------------------------------------
#  'GraphComponentController' class:
#-------------------------------------------------------------------------------

class GraphComponentController ( OMComponentController ):
    
    #---------------------------------------------------------------------------
    #  Returns whether or not a specified component can be dragged:  
    #---------------------------------------------------------------------------
        
    #def can_drag_component ( self, component ):
    #    return True
        
    #---------------------------------------------------------------------------
    #  Returns whether or not a specified component can be selected/unselected:  
    #---------------------------------------------------------------------------
    
    #def can_select_component ( self, component ):
    #    return True
        
    #---------------------------------------------------------------------------
    #  Edits the traits of the specified component:  
    #---------------------------------------------------------------------------
    
    def edit_component ( self, component ):
        """ Edits the traits of the specified component.
        """
        from graph_bundle_editor import GraphBundleEditor
        
        editor = GraphBundleEditor( actor = component.data )
        ui     = editor.edit_traits()
        if not ui.result:
            return ( None, None )
            
        self.update_component( component )
        return ( ui.history, auto_connect( component ) )
        
    #---------------------------------------------------------------------------
    #  Processes the user clicking a component's label:  
    #---------------------------------------------------------------------------
    
    def label_clicked ( self, component, event ):
        """ Processes the user clicking a component's label.
        """
        if event.alt_down:
            expansion = (self.get_expansion_state( component ) + 1) % 3
        elif event.control_down:
            if event.shift_down:
                expansion = 1
            else:
                expansion = 0
        elif event.shift_down:
            expansion = 2
        else:
            if component.controller.can_select_component( component ):
                canvas = component.container
                canvas.controller.select_components_action( 
                                      canvas, [ component ] )
            return
            
        selection = component.container.selection
        if component in selection:
            for component in selection:
                self.set_expansion_state( component, expansion )
        else:
            self.set_expansion_state( component, expansion )
        
    #---------------------------------------------------------------------------
    #  Begins a drag operation when the 'control' key is pressed:  
    #---------------------------------------------------------------------------
                
    def begin_control_drag ( self, component, event ):
        """ Begins a drag operation when the 'control' key is pressed.
        """
        self._component  = component
        self._components = components = []
        for c in get_components( component ):
            if auto_connectable( component, c ):
                ic = c.inner_component
                components.append( ( ic, ic.bg_color ) )
                ic.bg_color = ( 0.0, 1.0, 0.0, 1.0 )
                c.redraw()
                
        canvas = component.container
        component.window.mouse_owner = None
        ex, ey = event.x, event.y
        dc     = OMDragContact( contact   = component,
                                container = canvas,
                                bounds    = ( ex, ey, 1, 1 ),
                                anchor    = ( int( ex ), int( ey ) ) ) 
        dc.on_trait_change( self.drag_component_done, 'resized' )
        canvas.window.drag_resize( dc, canvas.bounds, event,
                                   drag_handler = self.drag_update )
        
    #---------------------------------------------------------------------------
    #  Begins a drag operation when the 'shift' key is pressed:  
    #---------------------------------------------------------------------------
                
    #def begin_shift_drag ( self, component, event ):
    #    """ Begins a drag operation when the 'shift' key is pressed.
    #    """
    #    return False
        
    #---------------------------------------------------------------------------
    #  Begins a drag operation when the 'alt' key is pressed:  
    #---------------------------------------------------------------------------
                
    def begin_alt_drag ( self, component, event ):
        """ Begins a drag operation when the 'alt' key is pressed.
        """
        PythonDropSource( component.window.control, component ) 
        
    #---------------------------------------------------------------------------
    #  Displays the component's context menu:  
    #---------------------------------------------------------------------------
    
    def popup_menu ( self, component, event ):
        """ Displays the component's context menu.
        """
        # Get the current selection and whether the context component is part of
        # it:
        selection    = component.container.selection
        in_selection = (component in selection)
        multi_select = (len( selection ) >= 2)
        
        # 'Check' the current expansion state of the context component:
        for action in show_actions:
            action.checked = False
        show_actions[ self.get_expansion_state( component ) ].checked = True
        
        # Allow/Disallow the alignment and distribution options:
        alignable = (in_selection and multi_select)
        for action in alignment_actions:
            action.enabled = alignable
        for action in distribution_actions:
            action.enabled = alignable
        
        # Allow creating a composite if context component is part of the
        # current selection, and the selection has at least two components:
        create_simple_composite_action.enabled = (in_selection and multi_select) 
###     create_full_composite_action.enabled   = (in_selection and multi_select) 
        
        # Allow expanding composite components if the context component is a
        # composite and the component is either not in the selection, or if it
        # is, all components in the selection are composites:
        enabled = False                                                  
        if isinstance( component.data, ConvergeCompositeActor ):
            enabled = True
            if in_selection:
                for comp in selection:
                    if not isinstance( comp.data, ConvergeCompositeActor ):
                        enabled = False
                        break
        expand_in_place_action.enabled = enabled
        
        # fixme: Implement this...
        expand_on_new_canvas_action.enabled = False
        
        # Allow the superclass to do the actual display of the menu:
        super( GraphComponentController, self ).popup_menu( component, event )
                                
#-- Menu Event Handlers --------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Change the expansion state of a component:  
    #---------------------------------------------------------------------------
    
    def on_collapse_view ( self ):
        self.expand_component_action( self._component, 0 )
    
    def on_normal_view ( self ):
        self.expand_component_action( self._component, 1 )
    
    def on_expand_view ( self ):
        self.expand_component_action( self._component, 2 )
        
    #---------------------------------------------------------------------------
    #  Align all currently selected components:  
    #---------------------------------------------------------------------------
                
    def on_align_left ( self ):
        self.align_components( 0 )
                
    def on_align_right ( self ):
        self.align_components( 1 )
                
    def on_align_top ( self ):
        self.align_components( 2 )
                
    def on_align_bottom ( self ):
        self.align_components( 3 )
                
    def on_align_hcenter ( self ):
        self.align_components( 4 )
                
    def on_align_vcenter ( self ):
        self.align_components( 5 )
        
    def align_components ( self, align ):
        canvas = self._component.container
        canvas.controller.align_selected_components_action( canvas, align )
        
    #---------------------------------------------------------------------------
    #  Distribute all currently selected components:  
    #---------------------------------------------------------------------------
                
    def on_distribute_horizontally ( self ):
        canvas = self._component.container
        canvas.controller.distribute_selected_components_action( canvas, 0 )
        
    def on_distribute_vertically ( self ):
        canvas = self._component.container
        canvas.controller.distribute_selected_components_action( canvas, 1 )
        
    #---------------------------------------------------------------------------
    #  Create and expand composite components:  
    #---------------------------------------------------------------------------
        
    def on_create_simple_composite ( self ):
        canvas = self._component.container
        canvas.controller.create_composite_action( canvas, canvas.selection,
                                                   force_visible = False )
        
    def on_create_full_composite ( self ):
        canvas = self._component.container
        canvas.controller.create_composite_action( canvas, canvas.selection, 
                                                   force_visible = True )
        
    def on_expand_in_place ( self ):
        component  = self._component
        canvas     = component.container
        components = [ c for c in canvas.selection
                       if isinstance( c, ConvergeCompositeActor ) ]
        if component not in components:
            components = [ component ]
        self.expand_composite_action( *components ) 
        
    def on_expand_on_new_canvas ( self ):
        pass

    #---------------------------------------------------------------------------
    #  Displays the connection editor for the component:  
    #---------------------------------------------------------------------------
    
    def on_edit_connections ( self ):
        """ Displays the connection editor for the component.
        """
        self.edit_component_action( self._component )
      
      
    #---------------------------------------------------------------------------
    #  Displays the script editor for the component:  
    #---------------------------------------------------------------------------
    
    def on_edit_chip ( self ):
        """ Edit the scipt for the chip
        """
        
        if self._component.data is not None:
            if self._component.data.file_name is not None:
                file_name = self._component.data.file_name
        
        if file_name is not None:
            file = File(file_name) 
            
            cookieManager = CookieManager()
            
            cookie = cookieManager.get_cookie(OpenCookie, file)
            
            # FIXME: Cookie implementation expects a binding,
            # fake up a binding here, but there should be a better way.
            fake_binding = Binding()
            fake_binding.obj = file
            
            workbench_window = UIPlugin.instance.active_window
            
            cookie.open(workbench_window, fake_binding )
        
    #---------------------------------------------------------------------------
    #  Deletes a component from the canvas:  
    #---------------------------------------------------------------------------
                                           
    def on_delete ( self ):
        """ Deletes a component from the canvas.
        """
        self.delete_component_action( self._component )
            
    #---------------------------------------------------------------------------
    #  Deletes a specified component from the canvas:  
    #---------------------------------------------------------------------------
                        
    def delete ( self, component, update_model = True ):            
        """ Deletes a specified component from the canvas.
        """
        canvas = component.container
        actor  = component.data
            
        for item in canvas.components[:]:
            if isinstance( item, OMLink ):
                other = None
                if item.contact1.data[0] is actor:
                    other = item.contact2
                elif item.contact2.data[0] is actor:
                    other = item.contact1
                if other is not None:
                    canvas.remove( item )
                    set_contact_state( other )
        
        connections = []
        if update_model:
            canvas_actor = canvas.data
            for relation in actor.linked_relations():
                output = [ p for p in relation.linked_source_ports()
                           if not p.is_input ][0]
                input  = [ p for p in relation.linked_destination_ports()
                           if p.is_input ][0]
                connections.append( ( canvas_actor, output, input ) )
                input.container.unbind_input_port( input )
                relation.container = None
            actor.container = None
            
        canvas.remove( component )
        canvas.redraw()
            
        # Return all of the connections that were deleted:
        return connections
                
#--- Helper methods ------------------------------------------------------------
        
    #---------------------------------------------------------------------------
    #  Expands a composite chip in place on the canvas:  
    #---------------------------------------------------------------------------
                
    def expand_composite ( self, component ):
        """ Expands a composite chip in place on the canvas.
        """            
        # Save the center point of the composite as the origin for the 
        # components that will replace it:
        cx, cy = component.origin
        cx    += (int( component.width )  / 2)
        cy    += (int( component.height ) / 2)
            
        # Disconnect any graphical links the component has on the canvas:
        disconnect_links( component )
        
        # Remove the composite from the canvas altogether:
        canvas = component.container
        canvas.remove( component )
        
        # Set up the composite, the canvas it is contained in, and the actors
        # contained within the composite:
        composite    = component.data
        canvas_actor = composite.container
        actors       = composite.nodes[:]
        connections  = []
        
        # Reconnect all of the input ports:
        for actor_port in composite.input_ports:
            relation           = actor_port.inside_relations_list()[0]
            input              = relation.linked_ports( actor_port )[0]
            relation.container = None
            for relation in actor_port.relations_list():
                output = relation.linked_ports( actor_port )[0]
                relation.container = None
                connections.append( ( output, input ) )
                break
            else:
                input.container.unbind_input_port( input )
                
        # Reconnect all of the output ports:
        for actor_port in composite.output_ports:
            relation           = actor_port.inside_relations_list()[0]
            output             = relation.linked_ports( actor_port )[0]
            relation.container = None
            for relation in actor_port.relations_list():
                input = relation.linked_ports( actor_port )[0]
                relation.container = None
                connections.append( ( output, input ) )
            
        # Disconnect the composite from the canvas's graph:
        composite.container = None            
            
        # Reparent each of the composite's internal actors back to the canvas's
        # composite actor:
        for actor in actors:
            actor.name       = unique_actor_name( actor.name, canvas_actor )
            actor.container  = canvas_actor
                
        # Now, we can reconnect all of the ports that need to be connected
        # (Note that we had to delay it till now in order to avoid getting
        # graph-level errors generated by 'connect_ports' when the two ports
        # are not at the same level in the graph hierarchy).
        for output, input in connections:
            connect_ports( canvas_actor, output, input )
            
        # Build a new component for each actor and add it to the canvas at 
        # a reasonable location on the canvas:
        components = []
        for actor in actors:
            actor.container      = canvas_actor
            new_component        = canvas.controller.create_component( actor )
            x, y                 = actor._canvas_origin
            new_component.origin = (
                cx + x - (int( new_component.width )  / 2),
                cy + y - (int( new_component.height ) / 2) ) 
            canvas.add( new_component )
            components.append( new_component )
            
        # Recompute all of the links between the old and new components:
        for new_component in components:
            recompute_links( new_component, updated = True )
            
        # Returns all of the newly created components:
        return components

    #---------------------------------------------------------------------------
    #  Disconnects all links to contacts on a component:  
    #---------------------------------------------------------------------------

    def disconnect_contacts ( self, component ):
        """ Disconnects all links to contacts on a component.
        """
        components  = []
        connections = []
        for contact in component.contacts:
            components2, connections2 = contact.controller.disconnect_contact( 
                                                       contact, relink = False )
            components.extend(  components2 )
            connections.extend( connections2 )
        handled = []
        for component in components:
            if component not in handled:
                recompute_links( component, updated = True, run = True )
                handled.append( component )
                
        # Return all connections that were broken:        
        return connections                
  
    #---------------------------------------------------------------------------
    #  Handles a drag update event while performing a drag link component
    #  operation:  
    #---------------------------------------------------------------------------

    def drag_update ( self, event, drag_bounds ):
        """ Handles a drag update event while performing a drag link 
            component operation.
        """
        canvas     = self._component.container
        components = canvas.components_at( event.x, event.y )
        for i in range( len( components ) - 1, -1, -1 ):
            ic = components[i]
            if isinstance( ic, OMLabel ):
                if ic in [ item[0] for item in self._components ]:
                    canvas.pointer = 'bullseye'
                else:
                    canvas.pointer = 'no entry'
                break
        else:
            canvas.pointer = 'cross'
            
        return drag_bounds
        
    #---------------------------------------------------------------------------
    #  Handles the completion of an OMDragContact dragging operation:  
    #---------------------------------------------------------------------------
                
    def drag_component_done ( self, drag_contact, old, event ):
        # Restore any modified components:
        for ic, bg_color in self._components:
            ic.bg_color = bg_color
            ic.container.redraw()
        self._components = None
            
        component           = drag_contact.contact
        canvas              = drag_contact.container
        items               = canvas.components_at( *drag_contact.anchor2 )
        drag_contact.bounds = ( 99999.0, 99999.0, 1.0, 1.0 )
        items.reverse()
        for item in items:
            if isinstance( item, OMLabel ):
                auto_connect( component, item.container )
                break
        canvas.pointer = 'arrow'
        canvas.redraw()
      
    #---------------------------------------------------------------------------
    #  Completely updates a component after a structural change:  
    #---------------------------------------------------------------------------
    
    def update_component ( self, component, updated = True, run = True ):
        """ Completely updates a component after a structural change.
        """
        self.build_component( component )
        recompute_links( component, updated = updated, run = run )

    #---------------------------------------------------------------------------
    #  Builds a component using the contents of its associated actor:  
    #---------------------------------------------------------------------------
                    
    def build_component ( self, component ):
        """ Builds a component using the contents of its associated actor.
        """
        
        # Build the bundle list for the component based on its current state:
        self.build_bundles( component )
        
        actor = component.data
        cm    = component.contact_manager
        
        # Label the chip using the actor name:
        component.inner_component.text = actor.name
        
        # Determine the right track group style to use:
        if self.get_expansion_state( component ) == 2:
            track_group_style = expanded_track_group_style
        else:
            track_group_style = default_track_group_style
        
        # Discard the current contents of the track manager:
        cm.remove_tracks()
        
        # Add the input and output tracks:
        cm.add_tracks( 
            OMTrack( position     = 'left',  
                     category     = 'input',
                     right_margin = 1,
                     group_style  = track_group_style ),
            OMTrack( position     = 'right', 
                     category     = 'output',
                     left_margin  = 1,
                     group_style  = track_group_style ) )
         
        cc = getattr( actor, 'contact_controller', graph_contact_controller )
        if cc is None:
            cc = graph_contact_controller

        if hasattr( actor, 'input_bundles' ):
            if hasattr( actor, 'help' ):
                help = actor.help
                while True:
                    new_help = help.replace( '\n ', '\n' )
                    if len( help ) == len( new_help ):
                        break
                    help = new_help
                help = help.replace( '\n\n', '\r' )
                help = help.replace( '\n', ' ' )
                help = help.replace( '\r', '\n\n' )
                help = help.replace( '\t', '\n' )
                component.inner_component.tooltip = help
            cm.add_contacts( 
                *[ OMContact( label      = self._label_for( bundle ),
                              group      = bundle.group_name,
                              tooltip    = self._tooltip_for( bundle ),
                              category   = 'input',
                              controller = self._contact_controller_for( bundle, 
                                                                         cc ),
                              image      = self._contact_image_for( bundle ),
                              style      = self._contact_style_for( bundle ), 
                              data       = ( actor, bundle, True ) )
                   for bundle in actor.input_bundles
                   if self._should_create_contact_for( bundle ) ] )
                   
            cm.add_contacts( 
                *[ OMContact( label      = self._label_for( bundle ),
                              group      = bundle.group_name,
                              tooltip    = self._tooltip_for( bundle ),
                              category   = 'output',
                              controller = self._contact_controller_for( bundle, 
                                                                         cc ),
                              image      = self._contact_image_for( bundle ),
                              style      = self._contact_style_for( bundle ), 
                              data       = ( actor, bundle, False ) )
                   for bundle in actor.output_bundles 
                   if self._should_create_contact_for( bundle ) ] )
        else:
            cm.add_contacts( 
                *[ OMContact( label      = port.name, 
                              category   = 'input',
                              controller = cc,
                              image      = input_contact_image,
                              style      = graph_contact_style,
                              data       = port )
                   for port in actor.input_port_list() ] )
                   
            cm.add_contacts( 
                *[ OMContact( label      = port.name, 
                              category   = 'output',
                              controller = cc,
                              image      = output_contact_image,
                              style      = graph_contact_style,
                              data       = port )
                   for port in actor.output_port_list() ] )
               
        cm.layout_contacts()
        
    #---------------------------------------------------------------------------
    #  Aliases one component to another:
    #---------------------------------------------------------------------------
                
    def alias_component ( self, component, component2 ):
        """ Aliases one component to another.
        """
        component.data = component2.data
        contacts       = component2.contacts
        for i, contact in enumerate( component.contacts ):
            contact.data = contacts[i].data
        
    #---------------------------------------------------------------------------
    #  Returns the current expansion state of a specified component:
    #---------------------------------------------------------------------------
                
    def get_expansion_state ( self, component ):
        """ Returns the current expansion state of a specified component.
        """
        color = component.inner_component.bg_color_
        if color is CollapsedColor:
            return 0
        if color is NominalColor:
            return 1
        return 2
                
    #---------------------------------------------------------------------------
    #  Sets the current expansion state of the component:  
    #---------------------------------------------------------------------------
                        
    def set_expansion_state ( self, component, expansion, update = True ):
        """ Sets the current expansion state of the component.
        """
        if expansion == 0:
            component.inner_component.bg_color = CollapsedColor
            component.inner_component_position = 'center'
        else:
            component.inner_component_position = 'top'
            if expansion == 1:
                component.inner_component.bg_color = NominalColor
            else:
                component.inner_component.bg_color = ExpandedColor
        
        if update:
            self.update_component( component, updated = False )
        
    #---------------------------------------------------------------------------
    #  Builds the bundle lists for the specified component:  
    #---------------------------------------------------------------------------
    
    def build_bundles ( self, component ):
        """ Builds the bundle lists for the specified component.
        """
        actor     = component.data
        expansion = self.get_expansion_state( component )
        if expansion == 0:
            actor.input_bundles  = self.super_bundles( 
                                       actor.instance_input_bundles,
                                       ConvergeInputBundle )
            actor.output_bundles = self.super_bundles( 
                                       actor.instance_output_bundles,
                                       ConvergeOutputBundle )
        elif expansion == 1:
            actor.input_bundles  = actor.instance_input_bundles    
            actor.output_bundles = actor.instance_output_bundles    
        else:
            actor.input_bundles  = self.expand_bundles( 
                                       actor.instance_input_bundles,
                                       ConvergeInputBundle )
            actor.output_bundles = self.expand_bundles( 
                                       actor.instance_output_bundles,
                                       ConvergeOutputBundle )

    #---------------------------------------------------------------------------
    #  Builds an expanded bundle set from the normal bundle set:  
    #---------------------------------------------------------------------------
    
    def expand_bundles ( self, bundles, klass ):
        """ Builds an expanded bundle set from the normal bundle set.
        """
        result = []
        for bundle in bundles:
            if bundle.visible:
                for port in bundle.ports:
                    if port.visible:
                        result.append( klass( name       = port.name, 
                                              ports      = [ port ],
                                              group_name = bundle.name ) )
        return result

    #---------------------------------------------------------------------------
    #  Builds a 'super' bundle containing all ports:
    #---------------------------------------------------------------------------
    
    def super_bundles ( self, bundles, klass ):
        """ Builds a 'super' bundle containing all ports.
        """
        ports = []
        for bundle in bundles:
            if bundle.visible:
                for port in bundle.ports:
                    if port.visible:
                        ports.append( port )
        return [ klass( name = '.', ports = ports ) ]
        
#-- Undoable actions -----------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Undoable action that sets the expansion state of one or more components:
    #---------------------------------------------------------------------------
                                           
    def expand_component_action ( self, component, expansion ):
        """ Sets the expansion state of one or more components.
        """
        canvas = component.container
        if component.selected:
            canvas.controller.expand_components_action( canvas.selection,
                                                        expansion )
        else:
            canvas.controller.expand_components_action( [ component ],
                                                        expansion )
        
    #---------------------------------------------------------------------------
    #  Undoable action that expands one or more composite chips in place on the 
    #  canvas:  
    #---------------------------------------------------------------------------
                
    def expand_composite_action ( self, *composites ):
        """ Undoable action that expands one or more composite chips in place 
            on the canvas.
        """
        components = []
        canvas     = None
        for i, composite in enumerate( composites ):
            canvas    = composite.container
            undo_item = GraphExpandCompositeUndoItem( composite = composite )
            undo_item.components = self.expand_composite( composite )
            components.extend( undo_item.components )
            canvas.add_undo( undo_item, i > 0 )
        if canvas is not None:
            canvas.select_components( components )

    #---------------------------------------------------------------------------
    #  Undoable action that disconnects all links to contacts on a component:  
    #---------------------------------------------------------------------------

    def disconnect_contacts_action ( self, component = None ):
        """ Undoable action that disconnects all links to contacts on a component.
        """
        if component is None:
            component = self._component
            
        connections = self.disconnect_contacts( component )
        canvas      = component.container
        canvas.add_undo( GraphConnectionUndoItem( canvas        = canvas,
                                                  connections   = connections,
                                                  is_connection = False ) )
    
    #---------------------------------------------------------------------------
    #  Undoable action that deletes one or more components: 
    #---------------------------------------------------------------------------

    def delete_component_action ( self, component ):
        """ Undoable action that deletes one or more components.
        """
        canvas = component.container
        if component.selected:
            canvas.controller.delete_action( canvas )
        else:
            connections = self.delete( component )
            canvas.add_undo( GraphConnectionUndoItem( 
                                 canvas        = canvas,
                                 connections   = connections,
                                 is_connection = False ) )
            canvas.extend_undo( GraphComponentUndoItem( 
                                 canvas    = canvas,
                                 component = component,
                                 is_add    = False ) )
            canvas.updated = True
            canvas.run     = True
        
    #---------------------------------------------------------------------------
    #  Undoable action that edits the traits of the specified component:  
    #---------------------------------------------------------------------------
    
    def edit_component_action ( self, component ):
        """ Undoable action that edits the traits of the specified component.
        """
        history, connections = self.edit_component( component )
        if (history is not None) and history.can_undo:
            canvas = component.container
            canvas.add_undo( GraphEditComponentUndoItem( component = component,
                                                         history   = history ) )
            if len( connections ) > 0:
                canvas.extend_undo( GraphConnectionUndoItem( 
                                        canvas      = canvas,
                                        connections = connections ) )
                                                  
#-- Private methods ------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Returns the proper contact style to use for a specified bundle:  
    #---------------------------------------------------------------------------

    def _contact_style_for ( self, bundle ):
        """ Returns the proper contact style to use for a specified bundle.
        """
        if isinstance( bundle, ConvergeRepeatBundle ):
            return repeat_contact_style
        return graph_contact_style
        
    #---------------------------------------------------------------------------
    #  Returns the proper contact image to use for a specified bundle:  
    #---------------------------------------------------------------------------

    def _contact_image_for ( self, bundle ):
        """ Returns the proper contact image to use for a specified bundle.
        """
        if isinstance( bundle, ConvergeRepeatButtonBundle ):
            return repeat_contact_image
        return input_contact_image
        
    #---------------------------------------------------------------------------
    #  Returns the contact controller to use for a specified bundle:  
    #---------------------------------------------------------------------------
                
    def _contact_controller_for ( self, bundle, default ):
        """ Returns the contact controller to use for a specified bundle.
        """
        if isinstance( bundle, ConvergeRepeatButtonBundle ):
            return repeat_contact_controller
        return default
        
    #---------------------------------------------------------------------------
    #  Returns whether or not a contact should be created for a specified bundle:  
    #---------------------------------------------------------------------------
                
    def _should_create_contact_for ( self, bundle ):
        """ Returns whether or not a contact should be created for a specified 
            bundle.
        """
        if (bundle.visible and 
            (isinstance( bundle, ConvergeRepeatButtonBundle ) or
            (len( bundle.ports ) > 0))):
           for port in bundle.ports:
               if port.visible:
                   return True
        return False
                
    #---------------------------------------------------------------------------
    #  Returns the contact label for a specified bundle:  
    #---------------------------------------------------------------------------
                                
    def _label_for ( self, bundle ):
        return bundle.name
        
    #---------------------------------------------------------------------------
    #  Returns the tooltip for a specified contact 'bundle':  
    #---------------------------------------------------------------------------
    
    def _tooltip_for ( self, bundle ):
        """ Returns the tooltip for a specified contact 'bundle'.
        """
        if isinstance( bundle, ConvergeRepeatButtonBundle ):
            return 'Click to add a new repeat section'
        result = []
        maxlen = 0
        for port in bundle.ports:
            if port.visible:
                text   = '%s (%s): %s' % ( 
                         port.name, port.units, port.description )
                maxlen = max( maxlen, len( text ) )
                result.append( text )
        n = len( result[0] )
        result[0] = result[0].ljust( 2 * maxlen - n )
        return '\n'.join( result )
           
# Standard component controller:
graph_component_controller = GraphComponentController()
        
