#-------------------------------------------------------------------------------
#  
#  Defines the GraphContactController subclass of the Enable 'om' package 
#  OMContactController class used for managing the contacts of an 'actor chip'.
#
#  Written by: David C. Morrill
#  
#  Date: 03/14/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

from enthought.enable.om \
    import OMContactController, OMLink, OMContact, OMContactStyle
           
from enthought.util.wx.drag_and_drop \
    import PythonDropSource

from graph_actor_folder \
    import GraphActor

from converge_actor \
    import ConvergeRepeatBundle
    
from graph_helper \
    import set_contact_state, recompute_links, find_contact_bundle, \
           find_contact_bundles, auto_connect, find_actor_port, DraggablePort, \
           connect_ports, find_port, is_input_port_connected, \
           is_input_port_connected_or_optional, set_component_style
           
from graph_undo_item \
    import GraphConnectionUndoItem, GraphEditComponentUndoItem, \
           GraphEditContactUndoItem
    
from enthought.traits.ui.menu \
    import Menu, Action, Separator
       
#-------------------------------------------------------------------------------
#  Constants:  
#-------------------------------------------------------------------------------

# Action used to delete a 'repeat' contact:
delete_contact_action = Action( name   = 'Delete',
                                action = 'on_delete_contact' )
                                
# Popup menu for contact:
contact_menu = Menu( Action( name   = 'Edit connection...',
                             action = 'on_edit_contact' ),
                     Action( name   = 'Plot',
                             action = 'on_plot' ),
                     Separator(),
                     Action( name   = 'Disconnect',
                             action = 'on_disconnect' ),
                     Separator(),
                     delete_contact_action,
                     name = 'popup' )

# Map from contact 'state's to contact 'image's:
contact_image_map = {
    'unconnected':       '=contact_black',
    'connected':         '=contact_byellow',
    'connectable1':      '=contact_primary',
    'connectable2':      '=contact_secondary',
    'connectable3':      '=contact_tertiary',
    'partly_connected':  '=contact_bxyellow',
}

#-------------------------------------------------------------------------------
#  'GraphContactStyle' class:  
#-------------------------------------------------------------------------------

class GraphContactStyle ( OMContactStyle ):
    
    #---------------------------------------------------------------------------
    #  Override of the 'menu' property:  
    #---------------------------------------------------------------------------
        
    def get_menu ( self ):
        return contact_menu
    
# Contact style used for normal bundles:        
graph_contact_style = GraphContactStyle( 
    label_position = 'inside center',
    image_map      = contact_image_map )
                                         
# Contact style used for repeat bundles:    
repeat_contact_style = GraphContactStyle( 
    label_position = 'inside center',
    image_map      = contact_image_map,
    label_color    = ( 0.0, 0.0, 1.0, 1.0 ) )

#-------------------------------------------------------------------------------
#  'GraphContactController' class:
#-------------------------------------------------------------------------------

class GraphContactController ( OMContactController ):
    
    #---------------------------------------------------------------------------
    #  Returns whether or not a Contact can accept a specified Link object:  
    #---------------------------------------------------------------------------

    #def can_accept_link ( self, contact, link ):
    #    """ Returns whether or not a Contact can accept a specified Link object.
    #    """
    #    return True

    #---------------------------------------------------------------------------
    #  Returns whether or not a Contact can accept a specified Contact object:  
    #---------------------------------------------------------------------------
    
    #def can_accept_contact ( self, contact, other_contact ):
    #    """ Returns whether or not a Contact can accept a specified Contact 
    #        object.
    #    """
    #    return True
        
    #---------------------------------------------------------------------------
    #  Returns whether or not a Contact is allowed to begin a drag Link 
    #  operation:  
    #---------------------------------------------------------------------------
                
    #def can_drag_link ( self, contact ):
    #    """ Returns whether or not a Contact is allowed to begin a drag Link 
    #        operation.
    #    """
    #    return True
        
    #---------------------------------------------------------------------------
    #  Begins a drag link operation by returning a Link for the specified
    #  Contact:
    #---------------------------------------------------------------------------
                
    def begin_drag_link ( self, contact ):
        """ Begins a drag link operation by returning a Link for the specified 
            Contact.
        """
        self._contacts = contacts = []
        for contact2 in contact.container.container.contacts:
            connectable = self.connectable( contact, contact2 )
            if connectable:
                contact2.push_state( 'connectable%d' % connectable )
                contacts.append( contact2 )
        
        return
        
    #---------------------------------------------------------------------------
    #  Ends a drag link operation by completing the link between the two 
    #  specified contacts:
    #---------------------------------------------------------------------------
    
    def end_drag_link ( self, contact1, contact2 ):
        """ Ends a drag link operation by completing the link between the two
            specified contacts.
        """
        # _contacts may not exist if this controller didn't start the drag:
        for contact in self._contacts or []:
            contact.restore_state()
        self._contacts = None
        
        if contact2 is None:
            return
        
        connections   = 0
        canvas        = contact1.container.container
        output, input = contact1, contact2
        output_actor, output_bundle, output_dir = output.data
        input_actor,  input_bundle,  input_dir  = input.data

        if (output_actor is not input_actor) and (output_dir != input_dir):
            if output_dir:
                output_actor,  input_actor  = input_actor,  output_actor
                output_bundle, input_bundle = input_bundle, output_bundle
                output, input               = input, output
            
            # Let the input side do the work:
            if ((output.controller is not input.controller) and 
                (self is output.controller)):
                return input.controller.end_drag_link( output, input )
            
            # Get the dataflow ports:
            opl = output_actor.output_port_list()

            # Get the dataflow ports:
            # Note that 'log' may have zero dataflow ports, that's expected
            # they are created on the fly as needed.
            ipl       = input_actor.input_port_list()
            num_ports = len( ipl )

            # Connect as many ports as possible:
            connections = self._connect_ports( output_actor, output_bundle, opl, 
                              input_actor, input_bundle, ipl, canvas )

            # If any connections where made, determine the new state:
            if len( connections ) > 0:
                input.container.controller.build_bundles( input.container )
                
                canvas.add( OMLink( contact1 = contact1, contact2 = contact2 ) )
                
                # Create an undo entry for the connections made:
                canvas.add_undo( GraphConnectionUndoItem( 
                                     canvas      = canvas,
                                     connections = connections ) )
        
                # New ports may have been added to the actor:
                ipl = input_actor.input_port_list()
        
                # Count the number of existing and required connections:
                connections = required = 0
                for port_meta in input_bundle.ports:
                    port = ipl[ port_meta.index ]
                    if not port.is_optional:
                        required += 1
                        if len( port.relations_list() ) > 0:
                            connections += 1
                        
                if connections == required:
                    input.event_state = 'connected'
                else:
                    input.event_state = 'partly_connected'
                
                output.event_state = 'unconnected'
                
                # Mark the canvas as having been updated:
                canvas.updated = True
                canvas.run     = True
                
                input_component = input.container
                
                # If creating a connection changed the number of inputs, then
                # we have to reconfigure the component accordingly:
                input_component.controller.update_component( input_component )
                
                # Update the visual style of the 'input' side of the connection:
                set_component_style( input_component )
        
    #---------------------------------------------------------------------------
    #  Handles an update event while performing a drag link operation:  
    #---------------------------------------------------------------------------
                
    def drag_update ( self, contact, event ):
        canvas     = contact.container.container
        components = canvas.components_at( event.x, event.y )
        for i in range( len( components ) - 1, -1, -1 ):
            contact2 = components[i]
            if isinstance( contact2, OMContact ) and (contact2 is not contact):
                break
        else:
            contact.pointer = 'cross'
            return None
            
        if self.connectable( contact, contact2 ):
            contact.pointer = 'bullseye'
            return contact2
            
        contact.pointer = 'no entry'
        return None
        
    #---------------------------------------------------------------------------
    #  Handles the completion of an OMDragContact dragging operation:  
    #---------------------------------------------------------------------------
                
    #def drag_contact_done ( self, drag_contact, old, event ):
    #    pass
        
    #---------------------------------------------------------------------------
    #  Begins a drag operation when the 'control' key is pressed:  
    #---------------------------------------------------------------------------
                
    def begin_control_drag ( self, contact, event ):
        """ Begins a drag operation when the 'control' key is pressed.
        """
        # Get the bundle corresponding to this contact:
        actor, bundle, is_input = contact.data
        
        # If it only has a single output port contained in it, then allow the 
        # port to be dragged:
        if len( bundle.ports ) == 1:
            if not is_input:
                port       = bundle.ports[0]
                actor_port = find_actor_port( port, actor )
                PythonDropSource( contact.window.control, 
                                  DraggablePort( port      = actor_port,
                                                 chip_name = actor.name,
                                                 port_name = port.name ) )
        
    #---------------------------------------------------------------------------
    #  Begins a drag operation when the 'shift' key is pressed:  
    #---------------------------------------------------------------------------
                
    #def begin_shift_drag ( self, contact, event ):
    #    """ Begins a drag operation when the 'shift' key is pressed.
    #    """
    #    pass
        
    #---------------------------------------------------------------------------
    #  Begins a drag operation when the 'alt' key is pressed:  
    #---------------------------------------------------------------------------
                
    #def begin_alt_drag ( self, contact, event ):
    #    """ Begins a drag operation when the 'alt' key is pressed.
    #    """
    #    pass
        
    #---------------------------------------------------------------------------
    #  Returns the name of the image to use for a specified Contact state:  
    #---------------------------------------------------------------------------
                
    #def get_contact_image ( self, contact, state ):
    #    """ Returns the name of the image to use for a specified Contact state.
    #    """
    #    return contact.image_map.get( state, contact.image )
        
    #---------------------------------------------------------------------------
    #  Handles a contact being clicked:  
    #---------------------------------------------------------------------------
                
    def clicked_contact ( self, contact, event ):
        """ Handles a contact being clicked.
        """
        if event.alt_down:
            self.disconnect_contact_action( contact )
        
    #---------------------------------------------------------------------------
    #  Display the contact's context menu:  
    #---------------------------------------------------------------------------
    
    def popup_menu ( self, contact, event ):
        component, bundle, is_input   = contact.data
        delete_contact_action.enabled = bundle.deletable
        super( GraphContactController, self ).popup_menu( contact, event )
                                
#-- Menu Event Handlers --------------------------------------------------------

    class PlaceboEvent ( object ):
        
        def __init__(self, x, y):
            self.x = x
            self.y = y
    
    def on_plot ( self ):
        """ connect the output to a plot chip for plotting """
        # FIXME: this action needs to be pluggable so that this controller
        # FIXME: is not dependent on cp.rockphysics....
        from cp.rockphysics.converge.plugins.plot_logs \
            import plot_logs_prototype
        
        ga = GraphActor( name='Plot Logs', instance=plot_logs_prototype )
        canvas = self._contact.container.container
        canvas_controller = canvas.controller

        # no idea where there is white space so just move to the right a bit
        x, y, dx, dy = self._contact.bounds
        x += 200
        
        # pretend we are the gui... see FIXME in dropped
        event = self.PlaceboEvent( int(x), int(y) )
        plot_component = canvas_controller.dropped( canvas, ga, event )
        
        # find the contacts that should be automatically hooked up
        plot_log_contact   = None
        plot_depth_contact = None
        my_depth_contact   = None
        
        # FIXME: hard coded names from PlotLogs
        for contact in plot_component.contacts:
            if contact.label == 'log':
                plot_log_contact = contact
            elif contact.label == 'depth':
                plot_depth_contact = contact
        
        for contact in self._contact.container.contacts:
            if contact.label == 'depth':
                my_depth_contact = contact
                break

        # hookup what we can
        if plot_log_contact is not None:        
            self.end_drag_link( self._contact, plot_log_contact )

        if my_depth_contact is not None and plot_depth_contact is not None:
            self.end_drag_link( my_depth_contact, plot_depth_contact )

        return
        
    #---------------------------------------------------------------------------
    #  Disconnects the links connected to a contact:
    #---------------------------------------------------------------------------
                                           
    def on_disconnect ( self ):
        self.disconnect_contact_action( self._contact )

    #---------------------------------------------------------------------------
    #  Displays the contact editor for the contact:  
    #---------------------------------------------------------------------------
    
    def on_edit_contact ( self ):
        """ Displays the contact editor for the contact.
        """
        self.edit_contact_action( self._contact )
        
    #---------------------------------------------------------------------------
    #  Deletes the current contact and all of its associated ports:  
    #---------------------------------------------------------------------------
                
    def on_delete_contact ( self ):
        """ Deletes the current contact and all of its associated ports.
        """
        self.delete_contact( self._contact )

#-- Public methods -------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Disconnects the links connected to a contact:
    #---------------------------------------------------------------------------
        
    def disconnect_contact ( self, contact, relink = True ):        
        """ Disconnects the links connected to a contact.
        """
        component               = contact.container
        actor, bundle, is_input = contact.data

        if is_input:
            port_list = actor.input_ports
            contact.event_state = 'unconnected'
        else:
            port_list = actor.output_ports
            
        return self.disconnect_bundle( bundle, port_list, component, relink )

    #---------------------------------------------------------------------------
    #  Disconnects the links connected to a specified bundle:
    #---------------------------------------------------------------------------
        
    def disconnect_bundle ( self, bundle, port_list, component, relink = True ):        
        """ Disconnects the links connected to a specified bundle.
        """
        canvas      = component.container
        actor       = canvas.data
        components  = [ component ]
        connections = []
        for port in bundle.ports:
            actor_port = port_list[ port.index ]
            for relation in actor_port.relations_list():
                for other_actor_port in relation.linked_ports( actor_port ):
                    if other_actor_port.is_input:
                        connections.append( ( actor, actor_port, 
                                                     other_actor_port ) )
                        other_actor_port.container.unbind_input_port( 
                                                       other_actor_port )
                    else:
                        connections.append( ( actor, other_actor_port, 
                                                     actor_port ) )
                    other_component, other_bundle, other_port = \
                        find_port( other_actor_port, canvas )
                    if other_component not in components:
                        components.append( other_component )
                relation.container = None
            if actor_port.is_input:
                actor_port.container.unbind_input_port( actor_port )
         
        # If the caller wants the links recomputed now, do so for all affected
        # components:
        if relink:
            for component in components:
                component.controller.update_component( component )
                
        # Return the affected components so the update can be done later.
        # Also returns the connections broken so that they can be undone later:
        return components, connections
        
    #---------------------------------------------------------------------------
    #  Deletes the specified contact and all of its associated ports:  
    #---------------------------------------------------------------------------
                
    def delete_contact ( self, contact ):
        """ Deletes the specified contact and all of its associated ports.
        """
        actor, input_bundle, is_input = contact.data
        component = contact.container
        components, connections = self.disconnect_contact( contact, 
                                                           relink = False )
            
        # Get the index of the bundle being deleted:
        input_bundles = actor.instance_input_bundles
        input_index   = input_bundles.index( input_bundle )
        
        # Compute the index of the corresponding repeat output bundle, if any,
        # taking into account the fact that the input side has a 
        # ConvergeRepeatButtonBundle that the output side does not have:
        output_bundles = actor.instance_output_bundles
        output_index   = (len( output_bundles ) + input_index - 
                          len( input_bundles ) + 1)
        if output_index >= 0:
            output_bundle = output_bundles[ output_index ]
            
            # If the output side has a corresponding 'repeat' section, 
            # disconnect it and remove the bundle and actor ports:
            if isinstance( output_bundle, ConvergeRepeatBundle ):
                components2, connections2 = self.disconnect_bundle( 
                                                output_bundle,
                                                actor.output_ports, component,
                                                relink = False )
                components.extend(  components2 )
                connections.extend( connections2 )
                self._delete_contact( output_bundle, output_bundles, 
                                      output_index, actor.output_ports )
       
        # Finish removing the input side bundle and actor ports:                                      
        self._delete_contact( input_bundle, input_bundles, input_index,
                              actor.input_ports )
        
        # Update each affected component:
        component.controller.update_component( component )
        for cur_component in components:
            if cur_component is not component:
                recompute_links( cur_component, updated = True )
        
    #---------------------------------------------------------------------------
    #  Deletes the specified contact and all of its associated ports:  
    #---------------------------------------------------------------------------
                
    def _delete_contact ( self, bundle, bundles, index, actor_ports ):
        """ Deletes the specified contact and all of its associated ports.
        """
        ports      = bundle.ports
        n          = len( ports )
        base_index = 99999
        
        # Find the index of the first actor port to delete:
        for port in ports:
            base_index = min( base_index, port.index )
        
        # Delete the bundle:
        del bundles[ index ]
        
        # Delete the associated actor ports:
        for port in actor_ports[ base_index: base_index + n ]:
            port.container = None
        del actor_ports[ base_index: base_index + n ]
       
        # Adjust the bundle port indices to remove the deleted actor port gap:
        for bundle in bundles[ index: ]:
            for port in bundle.ports:
                port.index -= n
 
    #---------------------------------------------------------------------------
    #  Returns whether two contacts are connectable or not:  
    #---------------------------------------------------------------------------
        
    def connectable ( self, contact1, contact2 ):
        """  Returns whether two contacts are connectable or not:  """
    
        # Output and input must be different actors and opposite w.r.t. 
        # dataflow direction:
        output, input = contact1, contact2
        output_actor, output_bundle, output_dir = output.data
        input_actor,  input_bundle,  input_dir  = input.data
        
        if (output_actor is input_actor) or (output_dir == input_dir):
            return 0

        # Work out direction of dataflow such that output and input make 
        # sense:
        if output_dir:
            output, input = input, output
        
        # Let the input side make the decision:
        if ((output.controller is not input.controller) and 
            (self is output.controller)):
            return input.controller.connectable( output, input )
            
        return self._connectable( output, input )
        
#-- Undoable actions -----------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Undoable action that disconnects all connections to a specified contact:    
    #---------------------------------------------------------------------------

    def disconnect_contact_action ( self, contact ):
        """ Undoable action that disconnects all connections to a specified 
            contact.
        """
        canvas = contact.container.container
        components, connections = self.disconnect_contact( contact )
        canvas.add_undo( GraphConnectionUndoItem( canvas        = canvas,
                                                  connections   = connections,
                                                  is_connection = False ) )
        
    #---------------------------------------------------------------------------
    #  Undoable action that edits the properties of a specified contact:  
    #---------------------------------------------------------------------------
                
    def edit_contact_action ( self, contact ):
        """ Undoable action that edits the properties of a specified contact.
        """
        component  = contact.container
        canvas     = component.container
        controller = component.controller
        expansion  = controller.get_expansion_state( component )
        if expansion == 0:
            controller.edit_component_action( component )
            return
            
        successful = False
        if expansion == 1:
            from graph_contact_editor import GraphContactEditor
                
            bundle = find_contact_bundle( contact )
            editor = GraphContactEditor( bundle = bundle )
            ui     = editor.edit_traits()
            if ui.result:
                history    = ui.history
                successful = ((history is not None) and history.can_undo)
                if successful:
                    canvas.add_undo( GraphEditComponentUndoItem( 
                                         component = component,
                                         history   = history ) )
        else:
            from graph_port_editor import GraphPortEditor
                
            bundles    = find_contact_bundles( contact )
            editor     = GraphPortEditor( name = contact.label )
            ui         = editor.edit_traits()
            successful = (ui.result and (editor.name != contact.label))
            if successful:
                done = False
                for bundle in bundles:
                    for port in bundle.ports:
                        if port.name == contact.label:
                            canvas.add_undo( GraphEditContactUndoItem(
                                component = component,
                                object    = port,
                                name      = 'name',
                                old_value = port.name,
                                new_value = editor.name ) )
                            port.name = editor.name
                            done      = True
                            break
                    if done:
                        break
                                        
        if successful:
            controller.update_component( component )
            connections = auto_connect( component )
            if len( connections ) > 0:
                canvas.extend_undo( GraphConnectionUndoItem( 
                                        canvas      = canvas,
                                        connections = connections ) )

#-- Protected methods ----------------------------------------------------------

    def _connectable ( self, output, input ):
        """ Return connectability of output and input.
            1 = best match (usually full name matches)
            2 = likely match (usually short, match name matches)
            3 = possible match (usually only returned when anything is allowed)
            0 = no match
            Protected method called by 'connectable' after it has worked out 
            which is which.
        """
        output_actor, output_bundle, output_dir = output.data
        input_actor,  input_bundle,  input_dir  = input.data
   
        output_ports = output_bundle.ports
        input_ports  = input_bundle.ports
        ipl          = input_actor.input_port_list()
        opl          = output_actor.output_port_list()
        
        if (len( output_ports ) == 1) and (len( input_ports ) == 1):
            if not is_input_port_connected_or_optional( ipl[ input_ports[0].index ] ):
                if output_ports[0].name == input_ports[0].name:
                    return 1
                if output_ports[0].match_name == input_ports[0].match_name:
                    return 2
                return 3
       
        result = 0
        for output_port in output_ports:
            for input_port in input_ports:
                if not is_input_port_connected_or_optional( ipl[ input_port.index ] ):
                    if output_port.name == input_port.name:
                        return 1
                    if output_port.match_name == input_port.match_name:
                        result = 2

        return result
        
    #---------------------------------------------------------------------------
    #  Connect as many ports as possible from output to input:    
    #---------------------------------------------------------------------------

    def _connect_ports ( self, output_actor, output_bundle, opl, 
                               input_actor, input_bundle, ipl, canvas ):
        """ Connect as many ports as possible from output to input.
        """
        output_ports = output_bundle.ports
        input_ports  = input_bundle.ports
        connections  = []
        if (len( output_ports ) == 1) and (len( input_ports ) == 1):
            input_port = input_ports[0]
            ip         = ipl[ input_port.index ]
            if not is_input_port_connected( ip ):
                output_port = output_ports[0]
                connection  = ( canvas.data, opl[ output_port.index ], ip ) 
                connections.append( connection )
                connect_ports( *connection )
                input_actor.bind_input_port( ip )
        else:
            unmatched_output_ports = []
            for output_port in output_ports:
                for input_port in input_ports:
                    if output_port.name == input_port.name:
                        ip = ipl[ input_port.index ]
                        if not is_input_port_connected( ip ):
                            connection = ( canvas.data,
                                           opl[ output_port.index ], ip )
                            connections.append( connection )
                            connect_ports( *connection )
                            input_actor.bind_input_port( ip )
                            break
                else:
                    unmatched_output_ports.append( output_port )
                    
            for output_port in unmatched_output_ports:
                match_ip = None
                for input_port in input_ports:
                    if output_port.match_name == input_port.match_name:
                        ip = ipl[ input_port.index ]
                        if not is_input_port_connected( ip ):
                            connection = ( canvas.data, 
                                           opl[ output_port.index ], ip )
                            connections.append( connection )
                            connect_ports( *connection )
                            input_actor.bind_input_port( ip )
                            break

        return connections
        
# Standard contact controller:
graph_contact_controller = GraphContactController()

