#-------------------------------------------------------------------------------
#  
#  The graph editor plugin.  
#  
#  Written by: David C. Morrill
#  
#  Date: 04/07/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

import os
import sys

from types import FunctionType

from enthought.io.file \
    import File
    
from enthought.traits.api \
    import List, HasPrivateTraits, Instance
    
from enthought.traits.ui.api \
    import Item, TreeNode, TreeEditor, View, Handler
    
from enthought.traits.ui.menu \
    import Menu, Action, Separator
    
from enthought.traits.ui.wx.tree_editor \
    import CutAction, CopyAction, PasteAction, DeleteAction, RenameAction
    
from enthought.envisage \
    import get_application, Plugin
    
from enthought.envisage.ui \
    import View as UIViewImpl, UIPlugin

from enthought.envisage.project.action.save_action import SaveAction
from enthought.envisage.project.action.save_as_action import SaveAsAction
from enthought.envisage.project.action.open_cookie import OpenCookie
from enthought.envisage.project.wizard.new_file_wizard \
    import FileFactory, NewFileWizard
    
from enthought.envisage.project \
    import CookieManager
    
from enthought.envisage.core.import_manager import ImportManager

from enthought.graph.vpl.graph_actor_folder \
    import GraphActorFolder, GraphActor

from enthought.graph.function import MetadataDefaults

from enthought.pyface.api \
    import OK

from enthought.logger import logger

from enthought.naming.binding \
    import Binding
    
from graph_editor import GraphEditor

from graph_editor_plugin_definition import ActorClassLocations
from graph_editor_plugin_definition import TypedPorts
from graph_editor_plugin_definition import FunctionDefaultsSpecification

from chip_manager import ChipManager
from port_factory import PortFactory

#-------------------------------------------------------------------------------
#  'GraphEditorPlugin' class:  
#-------------------------------------------------------------------------------

class GraphEditorPlugin ( Plugin ):
    """ The graph editor plugin.
    """
        
    # The shared plugin instance:
    instance = None
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # List of all active GraphEditor objects:
    graph_editors = List( GraphEditor ) 
    
    # path to the my_actions folder
    my_actions_path = None

    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------

    def __init__ ( self, **traits ):
        """ Creates a new plugin. """

        # Base-class constructor.
        super( GraphEditorPlugin, self ).__init__( **traits )

        # Set the shared instance.
        GraphEditorPlugin.instance = self

    #---------------------------------------------------------------------------
    #  'Plugin' interface:  
    #---------------------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Starts the plugin:  
    #---------------------------------------------------------------------------
        
    def start ( self, application ):
        """ Starts the plugin.
        """
        key_bindings = self.preferences.get( 'key_bindings' )
        if key_bindings is not None:
            import graph_canvas_controller
            graph_canvas_controller.graph_canvas_key_bindings.merge( 
                                                                  key_bindings )

        logger.debug("plugin enthought.graph.vpl load function defaults.")
        self._load_function_defaults()

        logger.debug("plugin enthought.graph.vpl load type port definitions.")
        self._load_typed_port_definitions()

        logger.debug("plugin enthought.graph.vpl create my chips location.")
        self._create_my_chips_location()

        logger.debug("plugin enthought.graph.vpl load actor definitions.")
        self._load_actor_definitions()
        
        logger.debug("plugin enthought.graph.vpl validate existing project models against actor definitions.")
        self._validate_existing_models()
        

        return
    
    #---------------------------------------------------------------------------
    #  Stops the plugin:  
    #---------------------------------------------------------------------------
                
    def stop ( self, application ):
        """ Stops the plugin.
        """
        import graph_canvas_controller
        self.preferences.set( 
                 'key_bindings',
                 graph_canvas_controller.graph_canvas_key_bindings )
        self.save_preferences()
        
#-- GraphEditorPlugin interface ------------------------------------------------

    #---------------------------------------------------------------------------
    #  Opens a new graph editor:  
    #---------------------------------------------------------------------------

    def open_graph_editor ( self, parent, canvas = None, binding = None ):
        """ Opens a new graph editor.
        """
        
        if binding is not None:
            title         = binding.name
            resource_name = binding.namespace_name
        else: 
            title         = 'Graph Editor' 
            resource_name = ''

        if canvas is not None:
            from graph_canvas_controller import GraphCanvasStyle
            canvas.style = GraphCanvasStyle()
            
        graph_editor = GraphEditor(
            parent        = parent,
            title         = title,
            canvas        = canvas, 
            resource_name = resource_name,
        )
        self._add_save_actions( canvas, graph_editor )
        graph_editor.open()

        # Set the size of the editor to be a specific percentage of the MDI
        # area.
        #
        # fixme: We should have a real API for this!  Also, to do this
        # properly we would have to take into account the position of any
        # splitters in the window!
        mdi_child_frame = graph_editor.control
        
        mdi_parent_frame = mdi_child_frame.GetParent()
        width, height = mdi_parent_frame.GetSize()
        mdi_child_frame.SetSize((width * .65, height * .55))

        self.graph_editors.append( graph_editor )
        
    #---------------------------------------------------------------------------
    #  Marks an existing graph editor as closed:  
    #---------------------------------------------------------------------------
                
    def close_graph_editor ( self, graph_editor ):
        """ Marks an existing graph editor as closed.
        """
        self.graph_editors.remove( graph_editor )
       
        
    #---------------------------------------------------------------------------
    #  Add the save and save as actions to the canvas menu:  
    #---------------------------------------------------------------------------

    def _add_save_actions( self, canvas, editor ):
        """ Add the save and save as actions to the canvas menu.
        """
        
        canvas.menu.append(SaveAction(editor=editor))
        canvas.menu.append(SaveAsAction(editor=editor))
        
        return

### Protected Methods #########################################################
    def _load_function_defaults( self ):
        """ populate the function defaults """
        # Create and register the port factory as a service.
        metadata_defaults = MetadataDefaults()
        self.register_service( 'enthought.graph.function.MetadataDefaults',
                              metadata_defaults )
        
        im = ImportManager()

        extensions = get_application().extension_registry.get_extensions( 
                                                FunctionDefaultsSpecification )

        for fds in extensions:
            metadata_defaults.load_from_filename( fds.filename )
            
        # Load defaults defined through the application by the user.
        # TODO: It might be good to be able to specify this location in the plugin definition
        # and ensure that it is loaded last to overwrite any defaults already defined with
        # the same name.
        default_file_location = os.path.abspath(self.state_location + '/defaults.xml')
        if os.path.exists(default_file_location):
            metadata_defaults.load_from_filename(default_file_location)

        return


    def _load_typed_port_definitions( self ):
        """ populate the PortFactory with port classes """
        # Create and register the port factory as a service.
        port_factory = PortFactory()
        self.register_service( 'enthought.graph.vpl.PortFactory', port_factory )
        
        im = ImportManager()

        extensions = get_application().extension_registry.get_extensions( TypedPorts )

        for typed_ports in extensions:
            for name_spec in typed_ports.class_names:
                if type(name_spec) is tuple:
                    name, is_default = name_spec
                else:
                    name, is_default = name_spec, False
                    
                klass = im.import_symbol( name )
                port_factory.add( klass, is_default )

        return
         
    def _load_actor_definitions( self ):
        # load the actor classes for display in the ui.
        id = ActorClassLocations

        extensions = get_application().extension_registry.get_extensions(id)
        
        chipManager = ChipManager()

        for actorClassLocations in extensions:
            for actorClassLocation in actorClassLocations.class_locations:
                
                # actorClassLocation specifies a path to load chips from or an instance
                # of a chip to add.
                if (actorClassLocation.path is not None
                     and actorClassLocation.path != '' ):
                              
                     chipManager.load_chips_from_path(actorClassLocation.path)
#                else:
#                    # if an instance is given explicitly add it to the chip manager. 
#                    instance = actorClassLocation.instance 
#                    chipManager.add(
#                        get_application().import_manager.import_symbol( instance ) )
                
        chipManager.load_user_defined_chips()
        return
    
    def _validate_existing_models(self):
        """ Cross check and correct, if possible, the models in the project
        against the current definitions of the actors.
        """
        chipManager = ChipManager()
        chipManager.validate_canvas()
        return
        
    def _create_my_chips_location(self):
        """ Create my chips directory if needed.
        """
        
        action_python_path = os.path.abspath(self.state_location + '/user_action_path')
        
        # create directories if they do not exist. 
        if not os.path.exists(action_python_path):
            os.mkdir(action_python_path)
         
        user_chips_location = os.path.abspath(action_python_path + '/my_actions')
        if not os.path.exists(user_chips_location):
            os.mkdir(user_chips_location)
            
        self.my_actions_path = user_chips_location
        
        # create __init__.py file if one does not exist.    
        init_file = os.path.abspath(user_chips_location + '/__init__.py')
        if not os.path.exists(init_file):
            file = open(init_file, 'w')
            file.close()
            
        # append directory to python path
        sys.path.append(action_python_path)
          
        return    
    

#### EOF ######################################################################        
