#-------------------------------------------------------------------------------
#  
#  Envisage plugin definition file for the enthought.graph.vpl editor plugin.  
#  
#  Written by: David C. Morrill
#  
#  Date: 03/01/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------
from os.path import abspath, dirname, join

from enthought.traits.api import Str, Any, List, Trait, Tuple, Bool

from enthought.envisage.core.core_plugin_definition \
    import ExtensionItem, ExtensionPoint, PluginDefinition

from enthought.envisage.ui.ui_plugin_definition \
    import Action

from enthought.envisage.ui.ui_plugin_definition \
    import UIActions, UIViews, View

from enthought.envisage.project.project_plugin_definition \
     import ResourceType, ResourceTypes

from enthought.envisage.project.project_plugin_definition \
     import ResourceWizard, ResourceWizards

from enthought.envisage.project.project_plugin_definition \
     import CookieImplementation, CookieImplementations
     
# The plugin's globally unique identifier (also used as the prefix for all
# identifiers defined in this module).
ID = "enthought.graph.vpl"

#-------------------------------------------------------------------------------
#  Extensions:
#-------------------------------------------------------------------------------

###############################################################################
# Extension Points.
###############################################################################

class ActorClassLocation(ExtensionItem):
    """ The definition of a single resource type. """
    
    # instance of class to register
    instance = Any
    
    # root path of functions to register
    # only instance or path should be specified.
    path = Str
    
class ActorClassLocations(ExtensionPoint):
    """ Allows other plugins to contribute new resource types. """
    
    class_locations = List(ActorClassLocation)

class TypedPorts( ExtensionPoint ):
    """ Adds the named IOPort subclass to the PortFactory """
    # List of full class names of the IOPort subclasses.
    # When the name is accompanied by a Boolean value True, that class
    # will become the default port class.
    class_names = List( Trait( None, Str, Tuple(Str,Bool)) )

class FunctionDefaultsSpecification( ExtensionPoint ):    
    """ Adds function metadata defaults from specified files. """
    # Complete path to xml file containing default definitions.
    filename = Str
    
###############################################################################
# Extension Contributions.
###############################################################################

function_defaults = FunctionDefaultsSpecification(
    filename = join( abspath(dirname(__file__) + "/../function"),
                     "defaults.xml")
    )
    
### Typed Ports ###############################################################
typed_ports = TypedPorts(
    class_names = [ 
        ("enthought.graph.AnyTypeValuePort", True),
        "enthought.units.scalar_float_value_port.ScalarFloatValuePort",
        "enthought.units.scalar_value_port.ScalarValuePort",
        ]
    )


#-- ActorClassLocations --------------------------------------------------------

actor_class_locations = ActorClassLocations(
    class_locations = [
        ActorClassLocation( instance = 'enthought.graph.vpl.expression_actor.expression_actor_prototype' ),
    ])

#-- Resource types -------------------------------------------------------------
BASE = 'enthought.envisage.resource'

# fixme: put these 'built-in' resource types into a separate resource plugin?
INSTANCE_RESOURCE_TYPE = BASE + '.instance_resource_type.InstanceResourceType'

resource_types = ResourceTypes(
    resource_types = [
        ResourceType(
            class_name = ID + ".resource_type" \
            + ".dataflow_resource_type.DataflowResourceType",
            precedes = [INSTANCE_RESOURCE_TYPE]
        ),
        
        ResourceType(
            class_name = ID + ".resource_type" \
            + ".composite_actor_resource_type.CompositeActorResourceType",
            precedes = [INSTANCE_RESOURCE_TYPE]
        ),

        ResourceType(
            class_name = ID + ".resource_type" \
            + ".converge_actor_resource_type.ConvergeActorResourceType",
            precedes = [INSTANCE_RESOURCE_TYPE]
        ),
    ]
)

resource_wizards = ResourceWizards(
    resource_wizards = [
        ResourceWizard(
            category    = 'Dataflow Modeling',
            name        = 'Empty Dataflow',
            class_name  = ID + '.wizard.new_graph_wizard.NewGraphWizard'
        )
        
    ]
)

#-- Cookies --------------------------------------------------------------------

cookies = CookieImplementations(
    implementations = [
        CookieImplementation(
            resource_type = ID + ".resource_type" \
            + ".dataflow_resource_type.DataflowResourceType",
            
            cookie_interface = "enthought.envisage.project.action" \
            + ".open_cookie.OpenCookie",
            
            cookie_implementation = ID + ".cookie" \
            + ".dataflow_resource_open_cookie.DataflowResourceOpenCookie"
        ),
    ]
)
    
#-------------------------------------------------------------------------------
#  Plug definition:  
#-------------------------------------------------------------------------------

PluginDefinition(
    # The plugin's globally unique identifier:
    id = ID,

    # The name of the class that implements the plugin:
    class_name = ID + ".graph_editor_plugin.GraphEditorPlugin",

    # General information about the plugin:
    name          = "Graph Editor Plugin",
    version       = "1.0.0",
    provider_name = "Enthought Inc",
    provider_url  = "www.enthought.com",
    autostart     = True,
    enabled       = True,

    # The Id's of the plugins that this plugin requires:
    requires = [
        "enthought.envisage.ui",
        "enthought.envisage.ui.preference",
        "enthought.envisage.project",
        "cp.lab.plugin.action_plugin.ActionPlugin",
        "Units"
    ],

    # The extension points offered by this plugin:
    extension_points = [ActorClassLocations, TypedPorts, 
                        FunctionDefaultsSpecification],
    
    # The contributions that this plugin makes to extension points offered by
    # either itself or other plugins:
    extensions = [
        resource_types, resource_wizards,
        cookies, actor_class_locations,
        typed_ports,
        function_defaults
    ]
)

