#-------------------------------------------------------------------------------
#  
#  Miscellaneous help functions used to implement the graph editor.  
#  
#  Written by: David C. Morrill
#  
#  Date: 04/07/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

from enthought.traits.api \
    import HasTraits, Str, Any
    
from enthought.enable.om \
    import OMComponent, OMLink, OMOuterComponentStyle
    
from converge_actor \
    import ConvergeActor, ConvergeInputPort
    
#-------------------------------------------------------------------------------
#  Constants:  
#-------------------------------------------------------------------------------
        
# Colors used to represent a chip's expanded/nominal/collapsed states:
CollapsedColor = ( 0.93, 0.91, 0.87, 1.0 )
NominalColor   = ( 0.82, 0.99, 0.86, 1.0 )
ExpandedColor  = ( 0.82, 0.88, 0.96, 1.0 )
GroupColor     = ( 0.93, 0.95, 0.98, 1.0 ) 

graph_actor_style = OMOuterComponentStyle( 
                        horizontally_symmetric = False,
                        image = '=blue_t5_l5_r5_b5_%d.png' )

not_ready_graph_actor_style = OMOuterComponentStyle( 
                                  horizontally_symmetric = False,
                                  image = '=orange_t5_l5_r5_b5_%d.png' )

graph_comment_style = OMOuterComponentStyle( 
                        horizontally_symmetric = False,
                        image = '=yellow_t3_l3_r3_b3_%d.png' )
                        
#-------------------------------------------------------------------------------
#  Called when an internal state error is detected:  
#-------------------------------------------------------------------------------
                                                
def internal_state_error ( ):
    raise AssertionError, ("An inconsistent model internal state has been "
       "detected. Based on previous experience, a call to a component "
       "controller's 'update_component' method may be missing somewhere.") 
    
#-------------------------------------------------------------------------------
#  Return a list of all components contained in a specified canvas:  
#-------------------------------------------------------------------------------
                                                                    
def canvas_components ( canvas ):
    return [ c for c in canvas.components if isinstance( c, OMComponent ) ]
    
#-------------------------------------------------------------------------------
#  Finds the component, bundle and port associated with a specified actor port:  
#-------------------------------------------------------------------------------
                    
def find_port ( port, canvas ):
    """ Finds the component, bundle and port associated with a specified actor
        port.
    """
    is_input = port.is_input
    for component in canvas_components( canvas ):
        actor = component.data
        if is_input:
            bundles = actor.input_bundles
            ports   = actor.input_ports
        else:
            bundles = actor.output_bundles
            ports   = actor.output_ports
        for bundle in bundles:
            for aport in bundle.ports:
                if ports[ aport.index ] is port:
                    return ( component, bundle, aport )
                        
    # Should never happen:
    internal_state_error()
    
#-------------------------------------------------------------------------------
#  Finds the actor port associated with a specified (converge) port:  
#-------------------------------------------------------------------------------
        
def find_actor_port ( port, actor ):
    """ Finds the actor port associated with a specified (converge) port.
    """
    if isinstance( port, ConvergeInputPort ):
        ports = actor.input_ports
    else:
        ports = actor.output_ports
    return ports[ port.index ]
        
#-------------------------------------------------------------------------------
#  Finds the contact associated with a specified actor port:  
#-------------------------------------------------------------------------------
                    
def find_port_contact ( port, canvas ):
    """ Finds the contact associated with a specified actor port.
    """
    component, bundle, ignore = find_port( port, canvas )
    return find_bundle_contact( bundle, component )
                
#-------------------------------------------------------------------------------
#  Returns all instance bundles associated with a specified contact:  
#-------------------------------------------------------------------------------
                                
def find_contact_bundles ( contact ):
    actor, bundle, is_input = contact.data
    if is_input:
        return actor.instance_input_bundles
    return actor.instance_output_bundles
                
#-------------------------------------------------------------------------------
#  Returns the instance bundle associated with a specified contact:  
#-------------------------------------------------------------------------------
                                
def find_contact_bundle ( contact ):
    name = contact.label
    for bundle in find_contact_bundles( contact ):
        if name == bundle.name:
            return bundle
            
    internal_state_error()
            
#-------------------------------------------------------------------------------
#  Finds the contact associated with a specified bundle:  
#-------------------------------------------------------------------------------
                        
def find_bundle_contact ( bundle, container ):            
    """ Finds the contact associated with a specified bundle.
    """
    if isinstance( container, OMComponent ):
        components = [ container ]
    else:
        components = canvas_components( container )
        
    for component in components:
        for contact in component.contacts:
            if contact.data[1] is bundle:
                return contact
    
    return None                
             
#-------------------------------------------------------------------------------
#  Sets a specified contact's state based upon its degree of connectivity:  
#-------------------------------------------------------------------------------
            
def set_contact_state ( contact ):
    """ Sets a specified contact's state based upon its degree of connectivity.
    """
    actor, bundle, is_input = contact.data
    if is_input:
        port_list   = actor.input_port_list()
        connections = 0
        for port in bundle.ports:
            actor_port = port_list[ port.index ]
            if (is_input_port_connected_or_optional( actor_port ) or 
                (actor_port.token != port.value)):
                connections += 1
        if connections == 0:
            contact.event_state = 'unconnected'
        elif connections == len( bundle.ports ):
            contact.event_state = 'connected'
        else:
            contact.event_state = 'partly_connected'
            
#-------------------------------------------------------------------------------
#  Recomputes all of the links to a specified component:  
#-------------------------------------------------------------------------------
                                                
def recompute_links ( component, updated = False, run = False, redraw = True ):
    disconnect_links( component )
    reconnect_links(  component )
    set_component_style( component )
    canvas = component.container
    if updated:
        canvas.updated = True
    if run:
        canvas.run = True
    if redraw:
        canvas.redraw()

#-------------------------------------------------------------------------------
#  Disconnect all links to the specified component:  
#-------------------------------------------------------------------------------
            
def disconnect_links ( component ):
    canvas = component.container
    for item in canvas.components[:]:
        if (isinstance( item, OMLink ) and
            ((item.contact1.container is component) or 
             (item.contact2.container is component))):
            canvas.remove( item )
            
#-------------------------------------------------------------------------------
#  Reconnect all links to the specified component:  
#-------------------------------------------------------------------------------
                            
def reconnect_links ( component ):
    actor = component.data
    reconnect_bundles( component, actor.input_bundles,  actor.input_ports, 
                       'input'  )
    reconnect_bundles( component, actor.output_bundles, actor.output_ports, 
                       'output' )
            
#-------------------------------------------------------------------------------
#  Reconnect all links to the specified set of input or ouput bundles:
#-------------------------------------------------------------------------------
                            
def reconnect_bundles ( component, bundles, ports, category ):
    canvas = component.container
    for bundle in bundles:
        contact1 = find_bundle_contact( bundle, component )
        if contact1 is not None:
            contacts = []
            for port in bundle.ports:
                actor_port = ports[ port.index ]
                for relation in actor_port.relations_list():
                    if actor_port.is_input:
                        linked_ports = relation.linked_source_ports(
                                           actor_port )
                    else:
                        linked_ports = relation.linked_destination_ports(
                                           actor_port )
                    for linked_port in linked_ports:
                        contact2 = find_port_contact( linked_port, canvas )
                        if contact2 not in contacts:
                            contacts.append( contact2 )
                            canvas.add( OMLink( contact1 = contact1, 
                                                contact2 = contact2 ) )
                            set_contact_state( contact2 )
            set_contact_state( contact1 )
            
#-------------------------------------------------------------------------------
#  Sets the correct style for a specified component:  
#-------------------------------------------------------------------------------
                        
def set_component_style ( component ):
    """ Sets the correct style for a specified component.
    """
    if component.outer_component.style is graph_comment_style:
        return
        
    actor       = component.data
    input_ports = actor.input_ports
    for bundle in actor.input_bundles:
        for port in bundle.ports:
            if not is_input_port_connected_or_optional( 
                                                    input_ports[ port.index ] ):
                component.outer_component.style = not_ready_graph_actor_style
                return
    component.outer_component.style = graph_actor_style
            
#-------------------------------------------------------------------------------
#  Finds the output ports that match a specified input bundle and port:
#  Note: Returns a list of tuples, with one tuple for each match. The format
#  of each tuple is: ( bundle.port, component, output_bundle, output_port,
#  match_type ), where match_type = 1 means a strong match, and 2 = weak match.
#-------------------------------------------------------------------------------
                    
def find_output_ports ( input_bundle, input_port, components, instance = False):
    """ Finds the output ports that match a specified input bundle and port.
    """
    matches = []
    for component in components:
        actor = component.data
        if instance:
            output_bundles = actor.instance_output_bundles
        else:
            output_bundles = actor.output_bundles
        for output_bundle in output_bundles:
            if ((input_bundle is None) or 
                (input_bundle.name == output_bundle.name)):
                for output_port in output_bundle.ports:
                    if input_port.name == output_port.name:
                        match = 1   # Strong match
                    elif input_port.match_name == output_port.match_name:
                        match = 2   # Weak match
                    else:
                        continue
                    matches.append( ( '%s.%s' % ( actor.name, 
                                                  output_port.name ),
                                      component, output_bundle, output_port, 
                                      match ) )
    matches.sort( lambda l, r: cmp( l[3], r[3] ) )
    return matches
    
#-------------------------------------------------------------------------------
#  Returns the components to be used:  
#-------------------------------------------------------------------------------
        
def get_components ( component, components = None ):
    canvas = component.container
    if components is None:
        return [ c for c in canvas.components
                 if isinstance( c, OMComponent )        and
                    isinstance( c.data, ConvergeActor ) and
                    (c is not component) ]
                    
    if not isinstance( components, ( list, tuple ) ):
        return [ components ]
        
    return components
    
#-------------------------------------------------------------------------------
#  Automatically connect a specified component into the current dataflow graph:  
#-------------------------------------------------------------------------------
        
def auto_connect ( component, components = None ):
    """ Automatically connect a specified component into the current dataflow
        graph.
    """
    return []
    #components     = get_components( component, components )
    #connections    = _auto_connect(  component, components )
    #new_components = [ component ]                
    #for other_component in components:
    #    connections.extend( _auto_connect( other_component, new_components ) )
    #            
    #if len( connections ) > 0:
    #    recompute_links( component, updated = True )
    #    
    #return connections
             
#def _auto_connect ( component, components ):
#    connections = []
#    actor       = component.data
#    for input_bundle in actor.instance_input_bundles:
#        for input_port in input_bundle.ports:
#            input_actor_port = actor.input_ports[ input_port.index ]
#            if not is_input_port_connected_or_optional( input_actor_port ):
#                matches = find_output_ports( input_bundle, input_port, 
#                                             components, instance = True )
#                if ((len( matches ) == 1) or 
#                    ((len( matches ) > 1) and 
#                     (matches[0][-1] != matches[1][-1]))):
#                    name, output_component, output_bundle, output_port, \
#                        match = matches[0]
#                    output_actor_port = output_component.data.output_ports[ 
#                                            output_port.index ]
#                    connection = ( component.container.data, 
#                                   output_actor_port, input_actor_port )
#                    connect_ports( *connection )
#                    actor.bind_input_port( input_actor_port )
#                    connections.append( connection )
#    
#    return connections
    
#-------------------------------------------------------------------------------
#  Returns whether a component can be auto-connected to one or more other 
#  components:
#-------------------------------------------------------------------------------

def auto_connectable ( component, components = None ):
    """ Returns whether a component can be auto-connected to one or more other
        components.
    """
    components = get_components( component, components )
    if _auto_connectable( component, components ):
        return True
    new_components = [ component ]                
    for other_component in components:
        if _auto_connectable( other_component, new_components ):
            return True
    
    return False            
             
def _auto_connectable ( component, components ):
    actor = component.data
    for input_bundle in actor.instance_input_bundles:
        for input_port in input_bundle.ports:
            input_actor_port = actor.input_ports[ input_port.index ]
            if not is_input_port_connected_or_optional( input_actor_port ):
                matches = find_output_ports( input_bundle, input_port, 
                                             components, instance = True )
                if ((len( matches ) == 1) or 
                    ((len( matches ) > 1) and 
                     (matches[0][-1] != matches[1][-1]))):
                     return True
    
    return False

def connect_ports ( actor, output, input, name = None ):
    """ Within the CompositeActor, actor, connect the output and input ports.
    If name is not specified, a unique name is generated.
    """
    if name is not None:
        actor.connect( output, input, name )
    else:
        for attempt in range( 2000 ):
            try:
                actor.connect( output, input, _connection_name() )
                return
            except ValueError:
                pass
        raise ValueError, 'Could not create unique connection name'
    
#-------------------------------------------------------------------------------
#  Determines if an actor input port is connected to something or not:  
#-------------------------------------------------------------------------------
        
def is_input_port_connected_or_optional ( port ):
    """ Determines if an actor input port is connected to something
    or is optional.
    """
    return (port.is_optional or (len( port.relations_list() ) > 0))	
        
def is_input_port_connected ( port ):
    """ Determines if an actor input port is connected to something.
    """
    return len( port.relations_list() ) > 0

#-------------------------------------------------------------------------------
#  Define a connection (i.e. relation) name: 
#-------------------------------------------------------------------------------
   
connection_id = 0

def _connection_name ( ):
    global connection_id
    
    connection_id += 1
    return 'connection_%d' % connection_id      
    
#-------------------------------------------------------------------------------
#  Returns a new unique actor port name:  
#-------------------------------------------------------------------------------
        
port_id = 0

def new_port_name ( ):
    """ Returns a new unique actor port name.
    """
    global port_id
    
    port_id += 1
    return 'actor_port_%d' % port_id   
    
#-------------------------------------------------------------------------------
#  Returns a unique actor name:  
#-------------------------------------------------------------------------------
            
def unique_actor_name ( name, container ) :
    """ Returns a unique actor name within the specified container based on 
        name. If the container already contains an actor with the given name,
        a new name is created by appending a number from the range(2,100)
        until a unique name is found.  The search for a unique name is
        arbitrarily stopped at 100.
    """
    uname = name
    for i in range( 2, 100 ):
        try:
            container._check_node_name( uname )
            break
        except ValueError:
            uname = '%s %d' % ( name, i )
    
    return uname
                
#-------------------------------------------------------------------------------
#  'DraggablePort' class:  
#-------------------------------------------------------------------------------
                
class DraggablePort ( HasTraits ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Actor port being dragged:
    port      = Any
    
    # Name of the chip that the dragged port belongs to:
    chip_name = Str
    
    # Name of the port being dragged:
    port_name = Str
    
    #---------------------------------------------------------------------------
    #  Returns the port being dragged:  
    #---------------------------------------------------------------------------
    
    #def drop_editor_value ( self ):
    #    return self.port
        
    #---------------------------------------------------------------------------
    #  Returns the text representation of the port:  
    #---------------------------------------------------------------------------
    
    #def drop_editor_update ( self, control ):
    #    control.SetValue( '%s.%s' % ( self.chip_name, self.port_name ) ) 
     
    #---------------------------------------------------------------------------
    #  Returns the string value used to represent the associated port object:  
    #---------------------------------------------------------------------------
        
    def __str__ ( self ):
        return '%s.%s' % ( self.chip_name, self.port_name ) 
