#-------------------------------------------------------------------------------
#  
#  Tool for editing the inputs to a Converge component.  
#  
#  Written by: David C. Morrill
#  
#  Date: 04/07/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------
    
from enthought.traits.api \
    import HasPrivateTraits, Trait, Str, Instance, Any, TraitHandler, TraitError
    
from enthought.traits.ui.api \
    import View, Group, Item, EnumEditor
    
from enthought.enable.om \
    import OMComponent        
    
from enthought.envisage.ui \
    import View as UIViewImpl
    
from enthought.pyface.timer.api \
    import do_later
    
from graph_canvas \
    import GraphCanvas
    
from graph_helper \
    import find_port, recompute_links, connect_ports, find_output_ports
    
from graph_undo_item \
    import GraphConnectionUndoItem, GraphConstantInputUndoItem
    
#-------------------------------------------------------------------------------
#  Constants:  
#-------------------------------------------------------------------------------
        
NoConnection = ''
NotConnected = 'Not connected'
    
#-------------------------------------------------------------------------------
#  'InputTraitHandler' class:  
#-------------------------------------------------------------------------------
        
class InputTraitHandler ( TraitHandler ):
    
    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------
        
    def __init__ ( self, matches ):
        self.matches = matches
        
    def validate ( self, object, name, value ):
        if value == NotConnected:
            return NoConnection
        if type( value ) is float:
            return value
        value = value.strip()
        if value in self.matches:
            return value
        raise TraitError
    
#-------------------------------------------------------------------------------
#  'GraphInput' class:  
#-------------------------------------------------------------------------------
        
class GraphInput ( HasPrivateTraits ):
    pass

#-------------------------------------------------------------------------------
#  'GraphInputEditor' class:  
#-------------------------------------------------------------------------------

class GraphInputEditor ( UIViewImpl ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # The view object:
    graph_input = Instance( GraphInput )
    
    # The current canvas:
    canvas      = Instance( GraphCanvas, allow_none = True )
    
    # The plugin object:
    plugin      = Any
    
    #---------------------------------------------------------------------------
    #  Traits view definitions:  
    #---------------------------------------------------------------------------
        
    traits_view = View( [ Item( 'graph_input@', resizable = True ), '|<>' ], 
                        kind = 'subpanel' )

    #---------------------------------------------------------------------------
    #  Creates the control associated with the view:  
    #---------------------------------------------------------------------------
        
    def _create_contents ( self, parent ):
        """ Creates the control associated with the view.
        """
        from enthought.graph.vpl.graph_editor_plugin import GraphEditorPlugin
        
        self.plugin = GraphEditorPlugin.instance
        self.plugin.on_trait_change( self.on_editors_changed, 
                                     'graph_editors_items' )
                                     
        return self.edit_traits( parent = parent ).control
        
    #---------------------------------------------------------------------------
    #  Updates the graph input editor when something changes:  
    #---------------------------------------------------------------------------
                
    def update_editor ( self, canvas ):
        do_later( self._update_editor, canvas )
        
    def _update_editor ( self, canvas ):      
        if canvas is not self.canvas:
            if self.canvas is not None:
                self.canvas.on_trait_change( self.on_canvas_updated, 'updated',
                                             remove = True )
            self.canvas = canvas
            if canvas is not None:
                canvas.on_trait_change( self.on_canvas_updated, 'updated' )
                
        if self.graph_input is not None:
            self.graph_input.on_trait_change( self.on_connection_changed,
                                              remove = True )
                                              
        if (canvas is None) or (len( canvas.selection ) != 1):
            self.graph_input = None
        else:
            gi            = GraphInput()
            component     = canvas.selection[0]
            actor         = component.data
            components    = [ c for c in canvas.components
                              if isinstance( c, OMComponent ) and
                                 (c is not component)         and
                                 hasattr( c.data, 'output_bundles' ) ]
            groups        = []
            gi._map = map = {}
            input_bundles = getattr( actor, 'instance_input_bundles', [] )
            for input_bundle in input_bundles:
                items = []
                for input_port in input_bundle.ports:
                    name    = input_port.name
                    matches = find_output_ports( None, input_port, components )
                    values  = [ NotConnected ] + [ m[0] for m in matches ]
                    value = self.get_connected_port( actor, input_port, canvas )
                    if value == NoConnection:
                        value = actor.input_ports[ input_port.index ].token
                        if value is None:
                            value = NoConnection
                    port_map    = {}
                    map[ name ] = ( component, input_bundle, input_port, 
                                    port_map ) 
                    for match in matches:
                        port_map[ match[0] ] = match
                    gi.add_trait( name, Trait( value,
                                               InputTraitHandler( values ) ) )
                    items.append(
                        Item( name   = name,
                              label  = name,
                              editor = EnumEditor( values   = values,
                                                   evaluate = float ) ) )
                groups.append( Group( label       = input_bundle.name,
                                      show_border = True,
                                      *items ) )
            gi.on_trait_change( self.on_connection_changed )
            gi.trait_view( 'traits_view', View( Group( label = actor.name,
                                                       *groups ) ) )
            self.graph_input = gi

    #---------------------------------------------------------------------------
    #  Returns the name of the port (if any) the specified input port is 
    #  connected to:  
    #---------------------------------------------------------------------------
                
    def get_connected_port ( self, actor, input_port, canvas ):
        """ Returns the name of the port (if any) the specified input port is
            connected to.
        """
        actor_port = actor.input_ports[ input_port.index ]
        for relation in actor_port.relations_list():
            for linked_port in relation.linked_source_ports( actor_port ):
                component, bundle, output_port = find_port( linked_port, 
                                                            canvas )
                return '%s.%s' % ( component.data.name, output_port.name )
        return NoConnection
        
#-- Event handlers -------------------------------------------------------------

    #---------------------------------------------------------------------------
    #  Handles graph editors being opened and closed:  
    #---------------------------------------------------------------------------

    def on_editors_changed ( self, event ):
        """ Handles graph editors being opened and closed.
        """
        for editor in event.removed:
            editor.canvas.on_trait_change( self.on_selection_changed,
                                           'selection_items', remove = True )
            editor.on_trait_change( self.on_activated_changed,
                                    'activated', remove = True )
            
            # Fire a 'deactivated' event:
            self.on_activated_changed( editor, False, False )
            
        for editor in event.added:
            editor.canvas.on_trait_change( self.on_selection_changed,
                                           'selection_items') 
            editor.on_trait_change( self.on_activated_changed, 'activated' )
            
            # Fire the 'activated' event, since we already missed the real one:
            self.on_activated_changed( editor, False, True )

    #---------------------------------------------------------------------------
    #  Handles the selection of a canvas being changed:  
    #---------------------------------------------------------------------------
                                                        
    def on_selection_changed ( self, canvas, old, event ):
        """ Handles the selection of a canvas being changed.
        """
        self.update_editor( canvas )
        
    #---------------------------------------------------------------------------
    #  Handles an editor being activated/deactivated:  
    #---------------------------------------------------------------------------
                
    def on_activated_changed ( self, editor, old, event ):
        """ Handles an editor being activated/deactivated.
        """
        if event:
            self.update_editor( editor.canvas )
        else:
            self.update_editor( None )
            
    #---------------------------------------------------------------------------
    #  Handles the current canvas being updated:  
    #---------------------------------------------------------------------------
                        
    def on_canvas_updated ( self ):
        """ Handles the current canvas being updated.
        """
        self.update_editor( self.canvas )
        
    #---------------------------------------------------------------------------
    #  Handles the user changing one of the editor connections:  
    #---------------------------------------------------------------------------
    
    def on_connection_changed ( self, gi, name, old, new ):
        """ Handles the user changing one of the editor connections.
        """
        input_component, input_bundle, input_port, port_map = gi._map[ name ]
        canvas           = input_component.container
        canvas_actor     = canvas.data
        actor            = input_component.data
        input_actor_port = actor.input_ports[ input_port.index ]
        
        # Break any existing links:
        connection_undo_item = None
        for relation in input_actor_port.relations_list():
            output_actor_port    = relation.linked_ports( input_actor_port )[0]
            connection_undo_item = GraphConnectionUndoItem( 
                canvas        = canvas,
                connections   = [ ( canvas_actor, output_actor_port, 
                                                  input_actor_port ) ],
                is_connection = False )
            relation.container = None
        
        if new == NoConnection:
            # The user has broken the connection:
            if connection_undo_item is not None:
                canvas.add_undo( connection_undo_item )
                actor.unbind_input_port( input_actor_port )
            elif input_actor_port.token is not None:
                canvas.add_undo( GraphConstantInputUndoItem(
                                     component = input_component,
                                     index     = input_port.index,
                                     new_value = None ) )
                input_actor_port.token = input_port.value
            recompute_links( input_component, run = True )
            return
            
        try:
            # The user has connected to a different port:
            ignore, output_component, output_bundle, output_port, match = \
                port_map[ new ]
            output_actor_port = output_component.data.output_ports[ 
                                    output_port.index ]
            connection = ( canvas.data, output_actor_port, input_actor_port )
            connect_ports( *connection )
            recompute_links( input_component, run = True )
            undo_item = GraphConnectionUndoItem( 
                            canvas      = canvas,
                            connections = [ connection ] )
            if connection_undo_item is not None:
                canvas.add_undo( connection_undo_item )
                canvas.extend_undo( undo_item )
            else:
                canvas.add_undo( undo_item )
                actor.bind_input_port( input_actor_port )
        except:
            pass
        
        # The user has typed in a constant float value:
        canvas.add_undo( GraphConstantInputUndoItem(
                             component = input_component,
                             index     = input_port.index,
                             new_value = new,
                             undo_item = connection_undo_item ) )
        input_actor_port.token = new
        recompute_links( input_component, run = True )
        if connection_undo_item is not None:
            actor.unbind_input_port( input_actor_port )

