#-------------------------------------------------------------------------------
#  
#  Defines a graph canvas specific UndoHistory item.
#  
#  Written by: David C. Morrill
#  
#  Date: 05/12/2005
#  
#  (c) Copyright 2005 by Enthought, Inc.
#  
#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
#  Imports:  
#-------------------------------------------------------------------------------

from enthought.traits.api \
    import List, Int, Bool, Instance, Any, true
    
from enthought.traits.ui.api \
    import AbstractUndoItem, UndoItem, UndoHistoryUndoItem
    
from enthought.enable.om \
    import OMComponent
    
from converge_actor \
    import ConvergeInputPort
    
from graph_canvas \
    import GraphCanvas
    
from graph_helper \
    import recompute_links, connect_ports, find_port

#-------------------------------------------------------------------------------
#  'GraphUndoItem' class:  
#
#  Note: This class assumes 'object' is a 'GraphCanvas' object.
#-------------------------------------------------------------------------------

class GraphUndoItem ( UndoItem ):
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        super( GraphUndoItem, self ).undo()
        self.object.redraw()
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        super( GraphUndoItem, self ).redo()
        self.object.redraw()
        
#-------------------------------------------------------------------------------
#  'GraphSelectionUndoItem' class:  
#-------------------------------------------------------------------------------
                
class GraphSelectionUndoItem ( AbstractUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Canvas the selection applies to:
    canvas        = Instance( GraphCanvas )
    
    # Components previously selected:
    old_selection = List( OMComponent )
    
    # Components now selected:
    new_selection = List( OMComponent )
    
    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------
        
    def __init__ ( self, **traits ):
        """ Initializes the object.
        """
        super( GraphSelectionUndoItem, self ).__init__( **traits )
        self.old_selection = self.canvas.selection
        
    #---------------------------------------------------------------------------
    #  Finalizes the state of the undo item:  
    #---------------------------------------------------------------------------
                
    def finalize ( self ):
        """ Finalizes the state of the undo item.
        """
        self.new_selection = self.canvas.selection
        return self
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        self.canvas.select_components( self.old_selection )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        self.canvas.select_components( self.new_selection )
    
    #---------------------------------------------------------------------------
    #  Merges two undo items if possible:    
    #---------------------------------------------------------------------------
                
    def merge_undo ( self, undo_item ):
        """ Merges two undo items if possible.
        """
        if (isinstance( undo_item, self.__class__ ) and
            (self.canvas is undo_item.canvas)):
            self.new_selection = undo_item.new_selection
            return True
        return False
    
#-------------------------------------------------------------------------------
#  'GraphBoundsUndoItem' class:
#-------------------------------------------------------------------------------

class GraphBoundsUndoItem ( AbstractUndoItem ):
 
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
    
    # List of components affected by the operation:
    components = List( OMComponent )
    
    # List of original bounds:
    old_bounds = List
    
    # List of new bounds:
    new_bounds = List
    
    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------
        
    def __init__ ( self, **traits ):
        """ Initializes the object.
        """
        super( GraphBoundsUndoItem, self ).__init__( **traits )
        if len( self.old_bounds ) == 0:
            self.old_bounds = [ c.bounds for c in self.components ]
        
    #---------------------------------------------------------------------------
    #  Finalizes the state of the undo item:  
    #---------------------------------------------------------------------------
                
    def finalize ( self ):
        """ Finalizes the state of the undo item.
        """
        self.new_bounds = [ c.bounds for c in self.components ]
        return self
        
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        self._do( self.old_bounds )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        self._do( self.new_bounds )
        
    #---------------------------------------------------------------------------
    #  Apply the specified changes to the components:
    #---------------------------------------------------------------------------
        
    def _do ( self, bounds ):
        components = self.components
        if len( components ) > 0:
            for i, component in enumerate( components ):
                component.bounds = bounds[i]
            components[0].container.redraw()
            
#-------------------------------------------------------------------------------
#  'GraphMoveUndoItem' class:  
#-------------------------------------------------------------------------------
          
class GraphMoveUndoItem ( GraphBoundsUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Merges two undo items if possible:    
    #---------------------------------------------------------------------------
                
    def merge_undo ( self, undo_item ):
        """ Merges two undo items if possible.
        """
        if (isinstance( undo_item, self.__class__ ) and
            (self.components == undo_item.components)):
            self.new_bounds = undo_item.new_bounds
            return True
        return False
        
#-------------------------------------------------------------------------------
#  'GraphExpansionUndoItem' class:  
#-------------------------------------------------------------------------------

class GraphExpansionUndoItem ( AbstractUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:    
    #---------------------------------------------------------------------------
    
    # List of components affected by the operation:
    components     = List( OMComponent )
    
    # List of original expansion states:
    old_expansions = List
    
    # New expansion state:
    expansion      = Int
    
    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------
        
    def __init__ ( self, **traits ):
        """ Initializes the object.
        """
        super( GraphExpansionUndoItem, self ).__init__( **traits )
        self.old_expansions = [ c.controller.get_expansion_state( c )
                                for c in self.components ]
         
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        components = self.components
        if len( components ) > 0:
            components[0].container.controller.expand_components( 
                                               components, self.old_expansions )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        components = self.components
        if len( components ) > 0:
            components[0].container.controller.expand_components( 
                                               components, self.expansion )
        
#-------------------------------------------------------------------------------
#  'GraphCreateCompositeUndoItem' class:  
#-------------------------------------------------------------------------------
                
class GraphCreateCompositeUndoItem ( AbstractUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Canvas the composite is being added to:
    canvas        = Instance( GraphCanvas )
    
    # Components used to create the composite:
    components    = List( OMComponent )
    
    # Should unused outputs be made visible or not by default?
    #force_visible = Bool
    
    # Resulting composite component:
    composite     = Instance( OMComponent )
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        composite  = self.composite
        components = composite.controller.expand_composite( composite )
        canvas     = self.canvas
        canvas.clear_selection()
        for component in components:
            component.controller.delete( component, update_model = False )
        canvas.add( *self.components )
        for component in self.components:
            recompute_links( component )
        canvas.select_components( self.components )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        canvas     = self.canvas
        composite  = canvas.controller.create_composite( canvas, 
                                                         self.components )
        controller = composite.controller
        canvas.clear_selection()
        old_actor = self.composite.data
        controller.alias_component( self.composite, composite )
        new_actor = composite.data
        new_actor.instance_input_bundles  = old_actor.instance_input_bundles
        new_actor.instance_output_bundles = old_actor.instance_output_bundles
        controller.delete( composite, update_model = False )
        canvas.add( self.composite )
        controller.update_component( self.composite )
        canvas.select_component( self.composite )
        
#-------------------------------------------------------------------------------
#  'GraphExpandCompositeUndoItem' class:  
#-------------------------------------------------------------------------------
                
class GraphExpandCompositeUndoItem ( GraphCreateCompositeUndoItem ):

    #---------------------------------------------------------------------------
    #  Initializes the component:  
    #---------------------------------------------------------------------------
    
    def __init__ ( self, **traits ):
        super( GraphExpandCompositeUndoItem, self ).__init__( **traits )
        self.canvas = self.composite.container
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        super( GraphExpandCompositeUndoItem, self ).redo()
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        super( GraphExpandCompositeUndoItem, self ).undo()

#-------------------------------------------------------------------------------
#  'GraphComponentUndoItem'  
#-------------------------------------------------------------------------------

class GraphComponentUndoItem ( AbstractUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:    
    #---------------------------------------------------------------------------
    
    # Canvas the componet is to be added to or deleted from:
    canvas    = Instance( GraphCanvas )
    
    # Component to be added or deleted:
    component = Instance( OMComponent )
    
    # Is this an addition (or deletion)?
    is_add    = true
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        self._do( not self.is_add )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        self._do( self.is_add )
    
    #---------------------------------------------------------------------------
    #  Add or delete the component from the canvas:   
    #---------------------------------------------------------------------------
    
    def _do ( self, is_add ):
        canvas    = self.canvas
        component = self.component
        actor     = component.data
        if is_add:
            actor.container = canvas.data
            canvas.add( component )
        else:
            canvas.remove( component )
            actor.container = None
            
        canvas.redraw()
        canvas.updated = True
        canvas.run     = True
        
#-------------------------------------------------------------------------------
#  'GraphConnectionUndoItem' class:  
#-------------------------------------------------------------------------------
                
class GraphConnectionUndoItem ( AbstractUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:    
    #---------------------------------------------------------------------------
    
    # Canvas the connections are made/broken on:
    canvas        = Instance( GraphCanvas )
    
    # List of ( actor, output, input ) tuples that form the connection:
    connections   = List
    
    # Is this a connection being made (or broken)?
    is_connection = true
        
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        self._do( not self.is_connection )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        self._do( self.is_connection )
    
    #---------------------------------------------------------------------------
    #  Connects or disconnects the specified set of connections:
    #---------------------------------------------------------------------------
    
    def _do ( self, is_connection ):
        if is_connection:
            for actor, output, input in self.connections:
                connect_ports( actor, output, input )
                input.container.bind_input_port( input )
        else:
            for actor, output, input in self.connections:
                for relation in input.relations_list():
                     if relation.linked_source_ports( input )[0] is output:
                         relation.container = None
                input.container.unbind_input_port( input )
                         
        components = []
        canvas     = self.canvas
        for actor, output, input in self.connections:
            component, bundle, port = find_port( output, canvas )
            if component not in components:
                components.append( component )
            component, bundle, port = find_port( input, canvas )
            if component not in components:
                components.append( component )
                
        for component in components:
            component.controller.update_component( component )
            
        canvas.updated = True
        canvas.run     = True
            
#-------------------------------------------------------------------------------
#  'GraphEditComponentUndoItem' class:  
#-------------------------------------------------------------------------------
                        
class GraphEditComponentUndoItem ( UndoHistoryUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Component being edited:
    component = Instance( OMComponent )
        
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        super( GraphEditComponentUndoItem, self ).undo()
        component = self.component
        component.controller.update_component( component )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        super( GraphEditComponentUndoItem, self ).redo()
        component = self.component
        component.controller.update_component( component )
    
#-------------------------------------------------------------------------------
#  'GraphEditContactUndoItem' class:  
#-------------------------------------------------------------------------------
                
class GraphEditContactUndoItem ( UndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
        
    # Component whose contact is being edited:
    component = Instance( OMComponent )
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        super( GraphEditContactUndoItem, self ).undo()
        component = self.component
        component.controller.update_component( component )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        super( GraphEditContactUndoItem, self ).redo()
        component = self.component
        component.controller.update_component( component )
    
#-------------------------------------------------------------------------------
#  'GraphConstantInputUndoItem' class:  
#-------------------------------------------------------------------------------
                
class GraphConstantInputUndoItem ( AbstractUndoItem ):
    
    #---------------------------------------------------------------------------
    #  Trait definitions:  
    #---------------------------------------------------------------------------
    
    # Component whose port is being set to a constant input value:
    component = Instance( OMComponent )
    
    # Index of input port being set:
    index     = Int
    
    # Previous value of port's constant input:
    old_value = Any
    
    # New value of port's constant input:
    new_value = Any
    
    # Linked undo item:
    undo_item = Any
        
    #---------------------------------------------------------------------------
    #  Initializes the object:  
    #---------------------------------------------------------------------------
        
    def __init__ ( self, **traits ):
        super( GraphConstantInputUndoItem, self ).__init__( **traits )
        actor          = self.component.data
        self.old_value = actor.input_ports[ self.index ].token
    
    #---------------------------------------------------------------------------
    #  Undoes the change:
    #---------------------------------------------------------------------------
    
    def undo ( self ):
        """ Undoes the change.
        """
        actor = self.component.data
        actor.input_ports[ self.index ].token = self.old_value
        if self.undo_item:
            self.undo_item.undo()
        else:
            recompute_links( self.component, updated = True, run = True )
    
    #---------------------------------------------------------------------------
    #  Re-does the change:
    #---------------------------------------------------------------------------
    
    def redo ( self ):
        """ Re-does the change.
        """
        if self.undo_item:
            self.undo_item.redo()
        actor = self.component.data
        actor.input_ports[ self.index ].token = self.new_value
        recompute_links( self.component, updated = True, run = True )
    
    #---------------------------------------------------------------------------
    #  Merges two undo items if possible:    
    #---------------------------------------------------------------------------
                
    def merge_undo ( self, undo_item ):
        """ Merges two undo items if possible.
        """
        if (isinstance( undo_item, self.__class__ ) and
            (self.component is undo_item.component) and
            (self.index     == undo_item.index)):
            self.new_value = undo_item.new_value
            return True
        return False
        
