""" Port factory for ConvergeActor ports. """

from types import ClassType
from enthought.traits.api import HasTraits, Dict, Instance, Type
from enthought.graph import IOPort, TypedPort

class PortFactory( HasTraits ):
    """ Port factory for ConvergeActor ports. 
    
    Produce an IOPort subclass with TypedPort given an external type
    and internal type as keys.
    """
    
    # Mapping from Tuple( external_type, internal_type ) to IOPort subclass.
    _port_klasses = Dict
    
    # When no mapping is defined, use the default class for port creation.
    # FIXME: Trait should really be something like Class(TypedPort)
    default_klass = Type
    
    def port_class( self, external_type, internal_type ):
        """ Returns the subclass of IOPort appropriate for the input types.
        
        external_type - name of the require datatype at the dataflow level.
        internal_type - name of the require datatype required by the function.
        """
        key = (external_type, internal_type)
        if self._port_klasses.has_key( key ):
            klass = self._port_klasses[ key ]
        
        else:
            klass = self.default_klass
        
        return klass
    
    def port( self, ex_type, in_type, name, is_input=False, **kw ):
        """ Returns a port compatible with the specified types.
        """
        klass = self.port_class( ex_type, in_type )
        if klass is not None:
            port = klass( name, is_input, **kw )
            return port
        
        return None
        
    def add( self, klass, is_default=False ):
        """ Add klass to the list of known port classes.
        
        Klass is required to be an IOPort with TypedPort.
        If is_default is True, klass will be used when no other klass matches
        a request in port_klass and port.  The default must be able to handle
        all external and internal types.
        """
        if issubclass( klass, IOPort ) and issubclass( klass, TypedPort ):
            key = (klass.external_type, klass.internal_type)
            self._port_klasses[ key ] = klass
            
            if is_default:
                self.default_klass = klass
        
        else:
            raise ValueError, "%s required to be an IOPort and TypedPort"\
                                 % klass.__name__

        return
    
    def remove( self, external_type, internal_type ):
        """ Remove the port klass for the specified types.
        """        
        key = (external_type, internal_type)
        if self._port_klasses.has_key( key ):
            if self.default_klass == self._port_klasses[ key ]:
                raise ValueError, "Cannot remove default port klass."
            del self._port_klasses[ key ]
        
        return
        
        
        
#### EOF ######################################################################
            