""" Context adapter for converge actors. """


# Enthought library imports.
from enthought.graph.vpl.converge_actor import ConvergeActor
from enthought.naming.api import Binding, Context
from enthought.naming.adapter import InstanceContextAdapter
from enthought.traits.api import Any, Dict, HasTraits, Instance, List, Property
from enthought.traits.api import Str


class ValuePortContext(Context):
    """ A context implementation over a list of value ports. """

    #### 'Context' interface ##################################################

    # The name of the context within its own namespace.
    namespace_name = Property(Str)

    #### 'ValuePortContext' interface #########################################
    
    # The actor.
    actor = Any

    # The contex that we are part of.
    context = Any
    
    # The trait name.
    trait_name = Str

    # The name of the context.
    name = Str
    
    #### Protected 'Context' interface #######################################

    # The bindings in the context.
    _bindings = Property(Dict(Str, Any))

    ###########################################################################
    # 'Context' interface.
    ###########################################################################

    def _get_namespace_name(self):
        """ Returns the name of the context within its own namespace. """

        return self.context.namespace_name + '/' + self.name

    ###########################################################################
    # Protected 'Context' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get__bindings(self):
        """ Returns the bindings. """

        ports = getattr(self.actor, self.trait_name)
        
        bindings = {}
        for port in ports:
            bindings[port.name] = port.token

        return bindings

    #### Methods ##############################################################
    
    def _bind(self, name, obj):
        """ Binds a name to an object in this context. """

        ports = getattr(self.actor, self.trait_name)
        
        for port in ports:
            if port.name == name:
                port.token = obj

        return

    #### Trait event handlers #################################################

    def _context_changed(self):
        """ Called when the context is changed. """

        self.environment = self.context.environment

        return
    

class ConvergeActorHelper(HasTraits):
    """ Helper wrapper for converge actors. """

    #### 'ConvergeActorHelper' interface ######################################
    
    # All unbound input tokens.
    inputs = Property(List)

    # All output tokens.
    outputs = Property(List)

    # The name of the actor.
    name = Property(Str)
    
    #### 'Private' interface ##################################################
    
    # The actor that we are wrapping.
    _actor = Any

    # The context that we are in.
    _context = Any
    
    # Shadow trait for the 'inputs' property.
    _inputs = Any

    # Shadow trait for the 'outputs' property.
    _outputs = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, context, actor):
        """ Creates a new helper. """

        self._actor = actor
        self._context = context

        return

    ###########################################################################
    # 'ConvergeActorHelper' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_name(self):
        return self._actor.name

    def _set_name(self, name):
        self._actor.name = name
        return
    
    def _get_inputs(self):
        """ Returns the actor's unbound input tokens. """

        if self._inputs is None:
            self._inputs = ValuePortContext(
                context     = self._context,
                actor       = self._actor,
                trait_name  = 'unbound_input_ports',
                name        = 'inputs'
            )

        return self._inputs

    def _get_outputs(self):
        """ Returns the actor's output tokens. """

        if self._outputs is None:
            self._outputs = ValuePortContext(
                context     = self._context,
                actor       = self._actor,
                trait_name  = 'output_ports',
                name        = 'outputs'
            )

        return self._outputs


class ConvergeActorContextAdapter(InstanceContextAdapter):
    """ Context adapter for converge actors. """

    # fixme: Hack to make sure we only ever create one helper for an actor.
    # We need to re-visit this whole part of the tree!
    __helpers = {}

    #### 'Context' interface ##################################################

    # The name of the context within its own namespace.
    namespace_name = Property(Str)

    #### 'ContextAdapter' interface ###########################################

    # The resource that we are adapting.
    adaptee = Property(Instance(ConvergeActor))

    #### 'InstanceContextAdapter' interface ###################################

    # By default every public attribute of an instance is exposed. Use the
    # following traits to either include or exclude attributes as appropriate.
    #
    # Regular expressions that describe the names of attributes to include.
    include = ['inputs', 'outputs']

    #### Private interface ####################################################

    # Shadow for the 'adaptee' trait.
    _adaptee = Any

    ###########################################################################
    # 'Context' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_namespace_name(self):
        """ Returns the name of the context within its own namespace. """

        base = self.context.namespace_name
        
        # fixme: This is a bit dodgy 'cos we actually return a name that can
        # be looked up, and not the file system name...
        namespace_name = base + '/' + self.adaptee._actor.name
        
        return namespace_name
    
    ###########################################################################
    # 'ContextAdapter' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_adaptee(self):
        """ Returns the object that we are adapting. """

        helper = self.__helpers.get(id(self._adaptee), None)
        if helper is None:
            helper = ConvergeActorHelper(self, self._adaptee)
            self.__helpers[id(self._adaptee)] = helper

        return helper

    def _set_adaptee(self, adaptee):
        """ Sets the object that we are adapting. """

        self._adaptee = adaptee

        return
    
##### EOF #####################################################################
