#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Greg Rogers
#
#-----------------------------------------------------------------------------

""" An Expression object has a string representation of a python expression
and provides methods for extracting identifiers, compiling, applying, etc.
"""

# Standard library imports.
import logging
import parser
import symbol
import token
import types

# Alias builtin python functions so well known names can be used as
# eval arguments.
python_locals = locals
python_globals = globals

# Enthought library imports
from enthought.traits.api import Any, HasStrictTraits, List, Property, String

# Local imports
from expression_trait import ExpressionTrait


# Setup a logger for this module.
logger = logging.getLogger(__name__)


class Expression(HasStrictTraits):
    """ An Expression object has a string representation of a python expression
    and provides methods for extracting identifiers, compiling, applying, etc.
    """

    ###########################################################################
    # Public Expression traits
    ###########################################################################

    # The expression
    expression = ExpressionTrait

    # Read only: The compiled version of the expression.
    compiled_expression = Property

    # Read only: Identifiers that are callables (name followed by '(').
    callables = Property

    # Read only: Identifiers in the expression (variables and callables).
    identifiers = Property

    # Read only: Identifiers in the expression that are not callables.
    variables = Property

    ###########################################################################
    # Protected Expression traits
    ###########################################################################

    # Transient: The compiled version of the expression.
    _compiled_expression = Any

    # Transient: Identifiers that are callables (name followed by '(').
    _callables = List(String)

    # Transient: Names of identifers in the expression.
    _identifiers = List(String)

    # Transient: Identifiers in the expression that are not callables.
    _variables = List(String)

    ##########################################################################
    # 'object' interface
    ##########################################################################

    def __getstate__(self):
        """ Returns the state of this object for pickling. """
        state = super(Expression, self).__getstate__()
        # Don't include some attributes in our pickle.
        for key in ['_compiled_expression', '_identifiers',
                    '_variables', '_callables']:
            if key in state:
                del state[key]

        return state

    def __setstate__(self, state):
        """ Restores the state of this object during unpickling. """
        super(Expression, self).__setstate__(state)

        # Always recompile our expression if we have one.
        expression = state.get("expression")
        if expression is not None:
            self._expression_changed(expression)

        return

    ###########################################################################
    # HasTraits interface
    ###########################################################################

    def trait_view(self, name=None, view_element=None):
        """ Returns a View """
        if (name or view_element) != None:
            return super(Expression, self).trait_view( name=name,
                                                view_element=view_element )

        from enthought.python.ui.expression_view import ExpressionView
        return ExpressionView()

    ###########################################################################
    # Expression methods
    ###########################################################################

    def eval(self, globals=None, locals=None):
        """ Evaluate the expression using globals and locals using the
        builtin function eval.
        """
        if globals is None:
            globals = python_globals()
        if locals is None:
            locals = python_locals()

        if self._compiled_expression is None and self.expression is not None:
            self._expression_changed(self.expression)


        self._pre_eval( globals, locals )
        value = eval( self._compiled_expression, globals, locals )
        value = self._post_eval(value, globals, locals)
        return value

    ###########################################################################
    # Expression property methods
    ###########################################################################

    def _get_compiled_expression(self):
        """ Returns the compiled version of the expression. """
        return self._compiled_expression

    def _get_callables(self):
        """ Returns a copy of the list of the callables named in the experssion.
        """
        return self._callables[:]

    def _get_identifiers(self):
        """ Returns a copy of the list the identifiers, variables and callables,
        named in the expression.
        """
        return self._identifiers[:]

    def _get_variables(self):
        """ Returns a copy of the list the variables named in the expression.
        """
        return self._variables[:]

    ###########################################################################
    # Expression Event handlers
    ###########################################################################

    def _expression_changed(self, new):
        """ Expression has changed, parse to update identifiers and variables
        and compile.__call__"""

        if new is None or new.strip() == '':
            self._callables = []
            self._variables = []
            self._identifiers = []
            self._compiled_expression = None

        else:
            ast = parser.expr(new)
            self._compiled_expression = ast.compile('eval')
            t = ast.totuple()

            # All atom names are considered identifiers if they are not a
            # python reserved word.
            all_names = self._find_identifiers(t)
            self._identifiers = all_names.keys()

            # Variables are identifers that where not identified as callables.
            self._variables = [ key for key, value in all_names.items()
                                       if not value ]


            self._callables = [ key for key, value in all_names.items()
                                       if value ]

        return

    ###########################################################################
    # Expression protected methods
    ###########################################################################

    def _find_identifiers(self, ast_tuple, identifiers=None, is_func=False):
        """ Fills in identifiers with the variables in ast_tuple.  Then key is
        the variable name and the value is set to True if the variables is
        determined to be a callable and False otherwise.  Returns identifiers.

        This method assumes that ast_tuple is keyword free.  True if
        it was created with parser.expr(expression).totuple().

        is_func is a flag that says the next variable is function.
        This is determined by look ahead for a name followed by a '('.
        This is not the only way a function call can be coded, but it
        is the only one that is recognized.
        """
        # Initialize identifiers, if needed
        if identifiers is None:
            identifiers = {}

        # Get the id
        id = ast_tuple[0]
        if id == symbol.trailer and ast_tuple[1][0] == token.DOT:

            # We need to ignore the name after the dot, as it is not
            # really a variable.  Process the rest of the pieces
            # of the trailer. e.g.: foo.bar(baz) ignore bar but not baz
            for t in ast_tuple[2:]:
                if type(t) == types.TupleType:
                    self._find_identifiers(t, identifiers)

        # If we have an atom of type name then return the name
        elif id == symbol.atom and ast_tuple[1][0] == token.NAME:
            assert(len(ast_tuple)==2)
            name = ast_tuple[1][1]
            identifiers[name] = is_func

        # Otherwise recursively find the variables and concatenate the lists.
        else:
            remainder = ast_tuple[1:]
            last = len(remainder) -1

            for i, t in enumerate(remainder):
                if type(t) == types.TupleType:

                    # Look ahead for name followed by (
                    if t[0] == symbol.atom and t[1][0] == token.NAME \
                        and (i < last) \
                        and remainder[i+1][0] == symbol.trailer \
                        and remainder[i+1][1][0] == token.LPAR:
                        self._find_identifiers(t, identifiers, is_func=True)

                    else:
                        self._find_identifiers(t, identifiers)

        return identifiers


    def _pre_eval(self, globals, locals):
        """ Immediately prior to eval. """
        return

    def _post_eval(self, value, globals, locals):
        """ Protected method called just after eval actually evaluates the
        expression.  This method further validates and/or transforms the
        raw evaluation value.  That value is returned as the eval result.
        If value does not pass the validation requirements or cannot be
        transformed to do so, an exception is raised.

        This default implementation returns the input value with no checks
        or transformations.

        It is anticpated that subclasses will override this method.
        """
        return value


### EOF

