#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2006 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------

from cPickle import dumps, loads
import unittest

from enthought.python.api import Expression
from enthought.traits.api import TraitError


class ExpressionTestCase(unittest.TestCase):

    def test_default(self):
        exp = Expression()
        self.failUnlessEqual( exp.identifiers, [] )
        self.failUnlessEqual( exp.variables, [] )
        self.failUnlessEqual( exp.callables, [] )
        return

    def test_none(self):
        self.do_example( None, [], [] )
        return

    def test_empty(self):
        self.do_example( '', [], [] )
        return

    def test_simple_failure(self):
        exp = Expression()
        self.failUnlessRaises( TraitError, setattr, exp, 'expression', ';' )
        return

    def test_no_change_on_failure(self):
        exp = Expression(expression='a')
        self.failUnlessEqual( exp.identifiers, ['a'] )
        self.failUnlessEqual( exp.variables, ['a'] )
        self.failUnlessEqual( exp.callables, [] )

        self.failUnlessRaises( TraitError, setattr, exp, 'expression', ';' )

        self.failUnlessEqual( exp.expression, 'a' )
        self.failUnlessEqual( exp.identifiers, ['a'] )
        self.failUnlessEqual( exp.variables, ['a'] )
        self.failUnlessEqual( exp.callables, [] )

        return

    def test_simple(self):
        self.do_example( 'a', ['a'], [] )
        return


    def test_identifiers_not_variables(self):
        self.do_example( 'foo(a)', ['a'], ['foo'] )
        return

    def test_objects_are_variables(self):
        self.do_example( 'foo.bar(a) + baz.blargle / func(2.4)',
                         ['a', 'baz', 'foo'], ['func'] )
        return

    def test_functions_not_discovered_parens(self):
        # These are valid expressions where foo is a function, but
        # Expression parsing is not able to discover that fact.
        self.do_example( '(foo)(a)', ['a', 'foo'], [] )
        return

    def test_functions_not_discovered_from_list(self):
        # These are valid expressions where foo is a function, but
        # Expression parsing is not able to discover that fact.
        self.do_example( '[foo][0](a)', ['a', 'foo'], [] )
        return

    def test_example_1(self):
        self.do_example( 'a+b*c/d()', ['a', 'b', 'c'], ['d'] )
        return

    def test_example_2(self):
        self.do_example( 'a.b', ['a'], [] )
        return

    def test_example_3(self):
        self.do_example( 'a.b.c', ['a'], [] )
        return

    def test_example_4(self):
        self.do_example( 'a.b(c.d)', ['a', 'c'], [] )
        return

    def test_example_5(self):
        self.do_example( "lookup('../bar').v", [], ['lookup'] )
        return

    def test_pickle(self):
        exp = Expression(expression='a+b*c/d()')
        self.do_check(exp, ['a', 'b', 'c'], ['d'] )

        exp2 = loads( dumps( exp ) )

        self.do_check(exp2, ['a', 'b', 'c'], ['d'] )

    def do_example(self, expression, expected_variables, expected_callables):
        """ Create an Expression object from the expression string and
        check against the expected variables and callables.
        """
        exp = Expression(expression=expression)
        self.do_check(exp, expected_variables, expected_callables)
        return

    def do_check(self, expression, expected_variables, expected_callables):
        """ Check the expression (an Expression object). """

        expected_identifers = expected_variables + expected_callables
        expected_identifers.sort()

        identifiers = expression.identifiers
        identifiers.sort()
        self.failUnlessEqual( identifiers, expected_identifers )

        variables = expression.variables
        variables.sort()
        self.failUnlessEqual( variables, expected_variables )

        callables = expression.callables
        callables.sort()
        self.failUnlessEqual( callables, expected_callables )

        return

    def double(self, x):
        return 2*x

    def test_eval(self):
        a = 3.14159
        exp = Expression( expression='a * foo.double(a) * self.double(a)' )

        mylocals = locals()
        mylocals['foo'] = self
        result = exp.eval( globals=globals(), locals=mylocals )

        self.failUnlessEqual( a*2*a*2*a, result )

        return

    def test_eval_constant(self):
        exp = Expression( expression='54.321' )

        mylocals = locals()
        result = exp.eval( globals=globals(), locals=mylocals )

        self.failUnlessEqual( 54.321, result )

        return

    def test_eval_no_globals(self):
        a = 3.14159
        exp = Expression( expression='a * foo.double(a) * self.double(a)' )

        locals()['foo'] = self
        result = exp.eval(locals=locals())

        self.failUnlessEqual( a*2*a*2*a, result )

        return

    def ui_simple(self):
        exp = Expression()
        exp.configure_traits( kind='modal' )
        exp.print_traits()
        return

if __name__ == "__main__":
    unittest.main()

### EOF
