#-----------------------------------------------------------------------------
#
#  Copyright (c) 2005-2006 by Enthought, Inc.
#  All rights reserved.
#
#-----------------------------------------------------------------------------

from cPickle import dumps, loads
import unittest

from enthought.python.api import QuantityExpression
from enthought.traits.api import TraitError
from enthought.units.api import Quantity


class QuantityExpressionTestCase(unittest.TestCase):

    def test_default(self):
        exp = QuantityExpression()
        self.failUnlessEqual( exp.identifiers, [] )
        self.failUnlessEqual( exp.variables, [] )
        self.failUnlessEqual( exp.callables, [] )
        self.failUnlessEqual( exp.name, '' )
        self.failUnlessEqual( exp.units.label, 'm' )
        self.failUnlessEqual( exp.family_name, 'length')
        return

    def test_none(self):
        self.do_example( None, [], [], 'none', 'none', 'none' )
        return

    def test_empty(self):
        self.do_example( '', [], [], 'empty', 'm', 'length' )
        return

    def test_simple_failure(self):
        exp = QuantityExpression()
        self.failUnlessRaises( TraitError, setattr, exp, 'expression', ';' )
        return

    def test_no_change_on_failure(self):
        exp = QuantityExpression(expression='a', name='foo', units='hour', family_name='time')
        self.failUnlessEqual( exp.identifiers, ['a'] )
        self.failUnlessEqual( exp.variables, ['a'] )
        self.failUnlessEqual( exp.callables, [] )
        self.failUnlessEqual( exp.name, 'foo' )
        self.failUnlessEqual( exp.units.label, 'hour' )
        self.failUnlessEqual( exp.family_name, 'time' )

        self.failUnlessRaises( TraitError, setattr, exp, 'expression', ';' )

        self.failUnlessEqual( exp.expression, 'a' )
        self.failUnlessEqual( exp.identifiers, ['a'] )
        self.failUnlessEqual( exp.variables, ['a'] )
        self.failUnlessEqual( exp.callables, [] )

        return

    def test_simple(self):
        self.do_example( 'a', ['a'], [], 'foo', 'msec', 'time' )
        return


    def test_identifiers_not_variables(self):
        self.do_example( 'foo(a)', ['a'], ['foo'], 'foo', 'msec', 'time'  )
        return

    def test_objects_are_variables(self):
        self.do_example( 'foo.bar(a) + baz.blargle / func(2.4)',
                         ['a', 'baz', 'foo'], ['func'], 'foo', 'msec', 'time'  )
        return

    def test_functions_not_discovered_parens(self):
        # These are valid expressions where foo is a function, but
        # QuantityExpression parsing is not able to discover that fact.
        self.do_example( '(foo)(a)', ['a', 'foo'], [], 'foo', 'msec', 'time'  )
        return

    def test_functions_not_discovered_from_list(self):
        # These are valid expressions where foo is a function, but
        # QuantityExpression parsing is not able to discover that fact.
        self.do_example( '[foo][0](a)', ['a', 'foo'], [], 'foo', 'msec', 'time'  )
        return

    def test_example_1(self):
        self.do_example( 'a+b*c/d()', ['a', 'b', 'c'], ['d'], 'foo', 'msec', 'time'  )
        return

    def test_example_2(self):
        self.do_example( 'a.b', ['a'], [], 'foo', 'msec', 'time'  )
        return

    def test_example_3(self):
        self.do_example( 'a.b.c', ['a'], [], 'foo', 'msec', 'time'  )
        return

    def test_example_4(self):
        self.do_example( 'a.b(c.d)', ['a', 'c'], [], 'foo', 'msec', 'time'  )
        return

    def test_example_5(self):
        self.do_example( "lookup('../bar').v", [], ['lookup'], 'foo', 'msec', 'time'  )
        return

    def test_pickle(self):
        exp = QuantityExpression(expression='a+b*c/d()', name='foo', units='msec', family_name='time' )
        self.do_check(exp, ['a', 'b', 'c'], ['d'], 'foo', 'msec', 'time'  )

        exp2 = loads( dumps( exp ) )

        self.do_check(exp2, ['a', 'b', 'c'], ['d'], 'foo', 'msec', 'time'  )

    def do_example(self, expression, expected_variables, expected_callables, name, units, family_name):
        """ Create an QuantityExpression object from the expression string and
        check against the expected variables and callables.
        """
        exp = QuantityExpression(expression=expression, name=name, units=units, family_name=family_name)
        self.do_check(exp, expected_variables, expected_callables, name, units, family_name)
        return

    def do_check(self, expression, expected_variables, expected_callables, name, units, family_name):
        """ Check the expression (an QuantityExpression object). """

        expected_identifers = expected_variables + expected_callables
        expected_identifers.sort()

        identifiers = expression.identifiers
        identifiers.sort()
        self.failUnlessEqual( identifiers, expected_identifers )

        variables = expression.variables
        variables.sort()
        self.failUnlessEqual( variables, expected_variables )

        callables = expression.callables
        callables.sort()
        self.failUnlessEqual( callables, expected_callables )

        self.failUnlessEqual( name, expression.name )
        self.failUnlessEqual( units, expression.units.label )
        self.failUnlessEqual( family_name, expression.family_name )
        return

    def double(self, x):
        return 2*x

    def test_eval(self):
        a = 3.14159
        exp = QuantityExpression( expression='a * foo.double(a) * self.double(a)',
                                  name='bar', family_name='length', units='ft' )

        mylocals = locals()
        mylocals['foo'] = self
        result = exp.eval( globals=globals(), locals=mylocals )

        self.failUnless( isinstance( result, Quantity ) )
        self.failUnlessEqual( a*2*a*2*a, result.data )
        self.failUnlessEqual( result.units.label, 'ft' )
        self.failUnlessEqual( result.family_name, 'length' )
        self.failUnlessEqual( result.name, 'bar' )

        return

    def ui_simple(self):
        exp = QuantityExpression()
        exp.configure_traits( kind='modal' )
        exp.print_traits()
        return

if __name__ == "__main__":
    unittest.main()

### EOF
