""" Keeps track of shared resource libraries available to the application.

    (Copyright 2006 by Enthought, Inc.)
"""   

# Third-party package imports
import pysvn
import logging

# Enthought library imports
from enthought.traits.api import HasTraits, Dict, Str, Instance, Password, Bool, Any
from enthought.traits.ui.api import View, Group, Item

from enthought.envisage.project.project_plugin import ProjectPlugin

# Local imports
from library import Library


logger = logging.getLogger(__name__)


class LoginInfo(HasTraits):
    id = Str
    password = Password
    save_info = Bool

    traits_view = View(Group('id',
                             'password',
                             Group(Item('save_info', label="Remember me"),
                                   show_left=False)),
                       buttons=['OK', 'Cancel'],
                       title='Authorization Required',
                       width=200)


def get_login(realm, username, maysave):
    info = LoginInfo()
    ui = info.edit_traits(kind='livemodal') 
    return(ui.result, info.id, info.password, info.save_info)


############################################################################
# LibraryManager class
############################################################################
class LibraryManager(HasTraits):
    """ Keeps track of shared resource libraries available to the application.

        For the SVN version of Resource Sharing, the LibraryManager simply 
        keeps track of which libraries are defined.  However, other 
        implementations may require explicit bookkeeping (as SVN does not)
        to keep track of which resource comes from which repository/location.
        This class could then be used to implement such bookkeeping 
        transparently.
    """

    # The shared file library to be copied from
    current_library = Instance(Library)

    # The shared object library to be copied from
    current_object_library = Instance(Library)

    # A dictionary of all shared file libraries that the app knows about, 
    # keyed by a (hopefully!) descriptive name.
    available_libraries = Dict
    available_object_libraries = Dict

    svn_client = Any
    object_buffer = Str


    def __init__(self):
        self.reset()


    def reset(self):
        """ Set up a new SVN Client for this LibraryManager. 

            (This is required on startup because Client objects can't be
             pickled for some arcane reason.)
        """

        self.svn_client = pysvn.Client()
        self.svn_client.callback_get_login = get_login

        return


    def mothball(self):
        """ Prepare the LibraryManager for pickling. """

        self.svn_client = None

        return


    #####################################################################
    # LibraryManager API.  If the class is re-implemented to use a
    # different back end, the names and signatures of these methods should
    # not change.
    #####################################################################
    
    def post_changes(self, resource, comment = " "):
        """ Save any changes of 'resource' to the appropriate library.

            The resource should already exist in the library.
        """
        self.svn_client.checkin([resource], comment) 
        logger.info('Posted changes in ' + resource + ' to shared library')

        return


    def get_changes(self, resource):
        """ Update the local resource. 

            Update the local resource with any changes that have been
            made to the library version since the resource was copied
            or last updated.
        """
        self.svn_client.update(resource)
        logger.info('Updated ' + resource + ' from shared library')
        return


    def cancel_changes(self, resource):
        """ Cancel all changes to the local resource.

            Cancel all changes to the local resource since the last
            update from the library.  (This may or may not bring the
            resource in sync with the library version, depending on
            whether anyone else has made changes in the meantime.)
        """
        self.svn_client.revert(resource)
        logger.info('Reverted ' + resource + ' to library version')

        return

    
    def add_library(self, name, location):
        """ Add a new library to the list of available shared libraries.

            Given a library name and a location key (URL in SVN), add a
            new library to the list of availables.
            FIXME: Doesn't check whether library exists.
        """
        newlib = Library(name, location, self)

        self.available_libraries[name] = newlib

        if self.current_library == None:
            self.current_library = name
         
        return


    def remove_library(self, name):
        """ Remove the given library from the list of available libraries.

            Remove the library of the given name from 'available_libraries'.
            For the present, it is assumed that this function will only be
            called with 'name' values that are actually in the dictionary.
            This may be revisited later.
        """
        del self.available_libraries[name]
        if self.current_library.name == name:
            if len(self.available_libraries) == 0:
                self.current_library == None
            else:
                new_current_library = self.available_libraries.keys()[0]
                self.current_library = self.available_libraries[new_current_library]

        return


    def set_current_library(self, name):
        """ Set the current library to the library of the given name.

            At present, it is assumed that this name is chosen from the
            keys of 'available_libraries' and is therefore the name of a
            valid library.  This may be revisited.
        """
        self.current_library = self.available_libraries[name]

        return

    def set_current_object_library(self, name):
        """ Set the current object library to the library of the given name.

            At present, it is assumed that this name is chosen from the
            keys of 'available_object_libraries' and is therefore the name of a
            valid library.  This may be revisited.
        """
        self.current_object_library = self.available_object_libraries[name]

        return


    def is_library(self, location):

        return self.svn_client.is_url(location)



    def get_object_library(self, item):

        libname = item.context.name
        if libname in self.available_object_libraries:
            return self.available_object_libraries[libname]
        else:
            return None

#### EOF ######################################################################
