""" Classes for modeling a shared library as a hierarchy of names and types.

    This module defines a set of classes for modeling a shared library
    in an abstract hierarchical format that can be displayed and browsed
    by the Resource Sharing tool, which therefore does not need to be
    directly aware of the underlying library implementation (in this case,
    SVN.)

    (Copyright 2006 by Enthought, Inc.)
"""

# Third-party package imports
import pysvn

# Enthought library imports
from enthought.traits.api import HasTraits, Str, List, Instance

############################################################################
# Utility Functions
#
# FIXME: There is no real reason for these to be module-level functions.
# Make them methods of an appropriate class or separate them into a
# "utilities" module.
############################################################################

def strip_prefix(prefix, string):
    """ Strip the given prefix from the string """

    suffix = string[len(prefix):]
#    if suffix[0] == '/':
#        suffix = suffix[1:]

    return suffix


def simple_file_name(path):
    """ Return the file name at the end of a full or partial path. """

    index = path.rfind('/')
    if index == -1:
        return path
    else:
        return path[index+1:]

    return    


def parent_path(resource):
    """ Return the path to the folder containing the resource. """

    simple_name = simple_file_name(resource)
    name_index = resource.rfind(simple_name)
    if name_index == 0:
        return ""
    else:
        return resource[:name_index-1]

############################################################################
# Library modeling classes
############################################################################

class ResourceTag(HasTraits):
    """ Object to represent a generic resource.

        Represents both containers (e.g. folders) and files.
    """
    name = Str
    location = Str
    

class FileTag(ResourceTag):
    """ Object to represent an individual file."""

    pass


class ObjectTag(ResourceTag):
    """ Object to represent a resource. """
    pass

class ContainerTag(ResourceTag):
    """ Object to represent a container (e.g., folder). """

    resources = List(ResourceTag)



class LibraryStruct(ContainerTag):
    """ Special container representing the entire shared library. """

    def __init__(self, library, objects):
        if objects:
            self.init_for_objects(library)
            return

        self.name = library.longname()
        self.location = "" 

        listing = library.lib_manager.svn_client.ls(library.location, recurse=True)

        container_dictionary = {}

        for resource in listing:
            wholenewname = strip_prefix(library.location, str(resource["name"]))
            newname_path = parent_path(wholenewname)
            newname = simple_file_name(wholenewname)

            if resource["kind"] == pysvn.node_kind.file:
                newres = FileTag(name=newname, location=wholenewname)
            else:
                newres = ContainerTag(name=newname, location=wholenewname)
                container_dictionary[wholenewname] = newres

            if newname_path == "":
                self.resources.append(newres)
            else:
                parent_container = container_dictionary[newname_path]
                parent_container.resources.append(newres)

        return


    def init_for_objects(self, library):
        self.name = library.longname()
        self.location = "" 

        listing = library.lib_manager.svn_client.ls(library.location, recurse=True)
        print "Library location: " + library.location

        # For the present, we assume a flat folder structure for object libraries
        for resource in listing:
            if resource["kind"] != pysvn.node_kind.file:
                newloc = strip_prefix(library.location, str(resource["name"]))
                print newloc
                newname = simple_file_name(newloc)
                newres = ObjectTag(name=newname, location=newloc)
                self.resources.append(newres)

        return
        
