""" The set of Actions called from Navigator object context menus."""

# General imports
import wx
import os

# Enthought library imports.
from enthought.envisage.ui import WorkbenchAction
from enthought.envisage.project import Project
from enthought.traits.api import Str, HasTraits, List, Dict
from enthought.traits.ui.api import View, Item, Group, EnumEditor
from enthought.traits.ui.message import message
from enthought.envisage import get_application

# Local imports
from enthought.sharing.plugin import start_wait, end_wait
from enthought.sharing.log_struct import LogStruct
from library_browser import LibraryBrowser
from library_selector import LibrarySelector


############################################################################
# Context Menu action classes
############################################################################

class CommentBuffer(HasTraits):
    """ Class used to prompt the user for a log comment.
    """    

    buffer = Str

    traits_view = View( 
                     Group(Item(label="Please enter a description of change"),
                     Group(Item(name='buffer', style='custom'),
                           show_labels=False)),
                     width = 300,
                     buttons = ['OK', 'Cancel'],
                     title = "Log Entry"
                  )

class CheckoutAction(WorkbenchAction):
    """ Project-level menu action to create and populate a folder from library.
    """

    def perform( self ):
        """ Copy all files in the shared library to the selected project(s)
        """
        selection = self.window.selection

        lm = get_application().get_service('enthought.sharing.ILibraryManager')
        ls = LibrarySelector()
        libname = ls.get_library()

        if libname != None:
            lm.set_current_library(libname)

            start_wait()

            # copy the library to the appropriately named shared folder under
            # each project (creating folders if needed)
            for binding in selection:
                project = binding.obj

                shared_folder_name = os.path.join(project.path, "Shared files")

                if os.path.isdir(shared_folder_name):
                    newloc = project.lookup("Shared files")
                else:
                    newloc = project.create_folder("Shared files")
            
                if os.path.isdir(os.path.join(shared_folder_name, libname)):
                    newloc = newloc.lookup(libname)
                else:
                    newloc = newloc.create_subcontext(libname)

                try:
                    lm.current_library.get_shared_resource("", newloc.path)
                except:
                    message("Could not open " + lm.current_library.name)

                newloc.refresh()

            end_wait()

        return


class UpdateAction(WorkbenchAction):
    """ Menu action to update the selected object(s) from appropriate library.
    """
    
    def perform( self ):
        """ Update the selected objects from the library they came from.
        """
        selection = self.window.selection

        lm = get_application().get_service('enthought.sharing.ILibraryManager')

        start_wait()

        for binding in selection:
            resource = binding.obj

            lm.get_changes(resource.path)


        end_wait()

        return


class PostAction(WorkbenchAction):
    """ Menu action to update the library version of the selected object(s).
    """

    def perform( self ):
        """ Update the library from the selected object(s).
        """
        selection = self.window.selection

        comment = CommentBuffer(buffer="")
        while comment.buffer == "":
            if comment.edit_traits(kind='livemodal').result == False:
                message("No Changes Posted")
                return

        lm = get_application().get_service('enthought.sharing.ILibraryManager')

        start_wait()

        for binding in selection:
            resource = binding.obj

            lm.post_changes(resource.path, comment.buffer)

        end_wait()

        return


class AddAction(WorkbenchAction):
    """ Menu action to add the selected object(s) to the appropriate library.
    """
    def perform( self ):
        """ Add the object to the library its folder came from.
        """
        selection = self.window.selection

        lm = get_application().get_service('enthought.sharing.ILibraryManager')

        start_wait()

        for binding in selection:
            resource = binding.obj

            lm.current_library.add_new_resource(resource.path)

        end_wait()

        return


class ShowLogAction(WorkbenchAction):
    """ Menu action to show the log for the calling resource
    """
    def perform( self ):
        """ Show the log for the resource
        """
        selection = self.window.selection

        lm = get_application().get_service('enthought.sharing.ILibraryManager')

        start_wait()

        for binding in selection:
            resource = binding.obj

            log_struct = LogStruct(resource.path, lm)
            log_struct.edit_traits(kind='livemodal')

        end_wait()

        return



class BrowseAction(WorkbenchAction):
    """ Menu action to browse the current repository
    """
    def perform(self):

        # Prompt for the library to be displayed
        lm = get_application().get_service('enthought.sharing.ILibraryManager')
        ls = LibrarySelector()

        selection = self.window.selection
        binding = selection[0]
        project = binding.obj

        libname = ls.get_library()

        if libname != None:
            lm.set_current_library(libname)

            # Construct and display a browser for the selected library
            start_wait()

            try:
                lb = LibraryBrowser(lm.current_library, project)
            except:
                message("Unable to open " + libname)
                lb = None

            end_wait()

            if lb != None:
                lb.edit_traits()

        return


class ToyAction(WorkbenchAction):
    """ Toy action for development purposes.  Should be removed.
    """
    def perform(self):

        message("It worked!")

        return
#### EOF ######################################################################
