""" Class and editor to build and display a library diagram

    The LibraryBrowser is a TraitsUI-based hierarchical diagram of
    the shared library with which it is initialized.  The hierarchical
    model itself is a LibraryStruct object.

    (Copyright 2006 by Enthought, Inc.)
"""

# General imports
import os
import wx
import pickle
from os import mkdir, makedirs
from os.path import join, isdir, sep

# Enthought library imports
from enthought.traits.api import HasTraits, Instance, Str, Any
from enthought.traits.ui.api import TreeEditor, TreeNode, View, Group, Item
from enthought.traits.ui.menu import Menu, Action
from enthought.traits.ui.api import Handler
from enthought.traits.ui.menu import Action, CloseAction
from enthought.envisage.resource import ResourceContext
from enthought.envisage import get_application

# Local imports
from enthought.sharing import Library
from enthought.sharing.plugin import start_wait, end_wait
from enthought.sharing.library_struct import LibraryStruct, \
                                             ContainerTag, FileTag, ObjectTag


############################################################################
# Handler class for LibraryBrowser
############################################################################
class BrowserHandler(Handler):

    def folder_copy_action(self, info, object):

        start_wait()

        if not isdir(info.object.local_folder):
            mkdir(info.object.local_folder)

        dest_folder = info.object.local_resource_location(object.location)
        info.object.library.get_shared_resource(object.location, dest_folder)
        info.object.folder_context.refresh()

        end_wait()

        return


    def object_copy_action(self, info, object):
        start_wait()

        if not isdir(info.object.local_folder):
            info.object.folder_context.create_folder("Shared Objects")

        partial_context = info.object.folder_context.lookup("Shared Objects")
        visible_folder = join(partial_context.path, info.object.library.name)

        if not isdir(visible_folder):
            partial_context.create_folder(info.object.library.name)

        dest_context = partial_context.lookup(info.object.library.name)


        lm = get_application().get_service('enthought.sharing.ILibraryManager')

        hidden_object_folder = join(lm.object_buffer, info.object.library.name)
        if not isdir(hidden_object_folder):
            info.object.library.make_marker(hidden_object_folder)

        dest_folder = join(hidden_object_folder, object.name)

        info.object.library.get_shared_resource(object.location,
                                                dest_folder)

        newobj_loc = join(hidden_object_folder, object.name)
        newobj_storage = object.name + ".pcl"
        newobj_filepath = join(newobj_loc, newobj_storage)
        newobj_file = open(newobj_filepath, "rb")
        newobj = pickle.load(newobj_file)
        dest_context.create_resource(object.name, newobj)

        end_wait()

        return


############################################################################
# Context menus for library browser tree nodes
############################################################################
null_menu = Menu(*[])

container_menu = Menu(Action(name='Copy to shared folder',
                             action='folder_copy_action'))

object_menu = Menu(Action(name='Copy to shared folder',
                          action='object_copy_action'))

############################################################################
# The TraitsUI editor used to display the library hierarchy
############################################################################

no_view = View()
tree_editor = TreeEditor(
    editable = False,
    hide_root = False,
    orientation = 'vertical',

    nodes = [ TreeNode(
                  node_for = [ContainerTag],
                  children = 'resources',
                  label = "name",
                  auto_open = True,
                  delete = False,
                  insert = False,
                  copy = False,
                  rename = False,
                  menu = container_menu,
                  view = no_view,
              ),
              TreeNode(
                  node_for = [FileTag],
                  children = "",
                  label = "name",
                  delete = False,
                  insert = False,
                  copy = False,
                  rename = False,
                  menu = null_menu,
                  view = no_view,
              ),
              TreeNode(
                  node_for = [ObjectTag],
                  children = "",
                  label = "name",
                  delete = False,
                  insert = False,
                  copy = False,
                  icon_item = 'object_node',
                  rename = False,
                  menu = object_menu,
                  view = no_view,
              )
             ]
)

############################################################################
# The LibraryBrowser class (a GUI wrapper for the LibraryStruct)
############################################################################

class LibraryBrowser(HasTraits):
    library = Instance(Library)
    library_diagram = Instance(LibraryStruct)
    folder_context = Instance(ResourceContext)
    local_folder = Str

    traits_view = View( Group( Item( name = 'library_diagram',
                                     editor = tree_editor,
                                     resizable = True ),
                               show_labels=False ),
                        title = 'Repository Browser',
                        buttons = [ CloseAction ],
                        resizable = True,
                        style='custom',
                        width = .3,
                        height = .6,
                        handler = BrowserHandler()
                       )



    def __init__(self, library, context, objects=False):
        self.library = library
        self.library_diagram = LibraryStruct(library=self.library, objects=objects)
        self.folder_context = context
        if objects:
            self.local_folder = join(self.folder_context.path, "Shared Objects")
        else:
            self.local_folder = join(self.folder_context.path, "Shared Files")

        return


    def local_resource_location(self, resource_location):
        """ Returns the folder into which the resource should be copied.

            Creates the folder (and other folders in its path) if necessary.
        """

        nextfolder = self.local_folder
        locstring = resource_location.replace('/', os.sep)

        if locstring != "":
            if locstring[0] == '\\':
                locstring = locstring[1:]
            nextfolder = join(nextfolder, locstring)

        if not isdir(nextfolder):
            mkdir(nextfolder)

        return nextfolder

