#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Travis N. Vaught
# Date: 05/25/2005
# Description: Define dictionary of unit conversion methods by object type
#
# Symbols defined: default_unit_converters, [add_unit_converter]
#
#------------------------------------------------------------------------------

#############################################################################
# Imports:
#############################################################################
# Standard library imports.
import logging

# Enthought library imports.
from enthought.units import convert as units_convert
from enthought.util.numerix import typecode


logger = logging.getLogger(__name__)


# TODO: move this to cp.log.log_type_converters
def convert_log_index(log_index, unit_system=None, to_unit=None, family_name=None):
    """ Function to convert the units of a LogIndex

        Parameters
        ----------
        log_index
            index to be converted
        unit_system
            the unit system of the current log_index units, defaults to the
            unit_managers default system (current unit_system):
        to_units
            new_units to convert to
        family_name
            provided in cases where the family_name attribute is not present
            in the provided log_index object
        """


    if family_name==None:
        family_name=log_index.family_name or log_index.name

    if to_unit==None:
        unit_system = _get_unit_system(unit_system)
        try:
            to_unit = unit_system.units(family_name)

        except KeyError:
            logger.exception("Could not convert LogIndex: %s to system: %s" % \
                              (log_index, unit_system))
            return log_index.clone()

    if log_index.units == to_unit:
        new_log_index = log_index.clone()
        new_log_index.family_name = family_name # TODO: not sure if I need this
    else:
#        print "To_units", to_unit, log_index.units, family_name, unit_system
        data = units_convert(log_index.data, log_index.units, to_unit)
        new_log_index = log_index.clone(data=data)
        new_log_index.family_name = family_name
        new_log_index.units = to_unit

    return new_log_index


def convert_log(log, unit_system=None, to_unit=None, family_name=None):
    """ Function to convert the units of a Log

        Parameters
        ----------
        log
            Log to be converted
        unit_system
            the unit system of the current log units, defaults to the
            unit_managers default system (current unit_system):
        to_units
            new_units to convert to
        family_name
            provided in cases where the family_name attribute is not present
            in the provided log object
        """

    from enthought.util.numerix import PyObject

    if family_name==None:
        family_name=log.family_name or log.name

    unit_system = _get_unit_system(unit_system)
    new_index = convert_log_index(log.index, unit_system=unit_system)

    # Do not convert unknown log values
    if family_name == 'unknown':
        return log.clone(index=new_index)

    if to_unit==None:
        try:
            to_unit = unit_system.units(family_name)
        except KeyError:
            logger.exception("Could not convert Log: %s to system: %s" \
                              % (log,unit_system))
            return log.clone(index=new_index)

    if typecode(log.data) == PyObject:
        # handle string logs
        new_log = log.clone(index=new_index)
    else:
        new_log = convert_log_index(log, family_name=family_name, to_unit=to_unit)
        new_log.index = new_index
    return new_log


def convert_log_suite(log_suite, unit_system=None, to_unit=None, family_name=None):
    from cp.log.log_suite import LogSuite
    
    new_index = convert_log_index(log_suite.index, unit_system)
    new_logs = []
    for log in log_suite.get_logs():
        new_logs.append(convert_log(log))
    return LogSuite(log_suite.name, new_index, *new_logs)



def convert_quantity_old(q, unit_system):
    # I think importing within the function is needed to avoid circularity
    from enthought.units.quantity import Quantity

    try:
        units = unit_system.units(q.family_name)
    except KeyError:
        logger.exception("Could not convert quantity: %s to system: %s" % (q,unit_system))
        return q.clone()

    if q.units == units:
        q = q.clone()

    else:
        data = units_convert(q.data, q.units, units)
        q = Quantity(data, units=units, name=q.name or q.family_name, family_name=q.family_name)
    return q

def convert_action_variable(q, unit_system=None, to_unit=None, family_name=None):
    import copy
    
    if family_name==None:
        family_name=q.family_name

    if to_unit==None:
        unit_system = _get_unit_system(unit_system)
        try:
            to_unit = unit_system.units(family_name)

        except KeyError:
            logger.exception("Could not convert quantity: %s to system: %s" % (q,unit_system))
            return copy.deepcopy(q)

    q = copy.deepcopy(q)
    if q.units == to_unit.get_label():
        return q

    else:
        q.units = to_unit.get_label()
        
    return q

def convert_quantity(q, unit_system=None, to_unit=None, family_name=None):
    from enthought.units.quantity import Quantity

    if family_name==None:
        family_name=q.family_name

    if to_unit==None:
        unit_system = _get_unit_system(unit_system)
        try:
            to_unit = unit_system.units(family_name)

        except KeyError:
            logger.exception("Could not convert quantity: %s to system: %s" % (q,unit_system))
            return q.clone()

    if q.units == to_unit:
        q = q.clone()

    else:
        data = units_convert(q.data, q.units, to_unit)
        # create a new object of the same type as the input Quantity object
        # could be a subclass, so use __class__
        q = q.__class__(data, units=to_unit, name=q.name or family_name, family_name=family_name)
    return q



def _get_unit_system(unit_system=None):
    from enthought.units.unit_manager import unit_manager
    return unit_manager.get_unit_system(unit_system)

# The dict of defaults
default_unit_converters = {"<class 'enthought.units.quantity.Quantity'>": convert_quantity,
                           "<class 'enthought.units.scalar.Scalar'>": convert_quantity,
                           "<class 'cp.log.log_index.LogIndex'>": convert_log_index,
                           "<class 'cp.log.log.Log'>": convert_log,
                           "<class 'cp.log.log_suite_proxy.LogProxy'>": convert_log,
                           "<class 'cp.log.log_suite.LogSuite'>": convert_log_suite,
                           "<class 'cp.lab.action_variable.ActionVariable'>": convert_action_variable,
                            }

#### EOF ######################################################################
